package com.jgoodies.animation;

/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.
 * Use is subject to license terms.
 *
 */

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * An abstract class that minimizes the effort required to implement 
 * the {@link Animation} interface. Defines the duration and freezed state
 * and provides a listener list.
 *
 * @author Karsten Lentzsch
 */
public abstract class AbstractAnimation implements Animation {

    private final long      duration;
    private final boolean   freezed;
    private final List      listenerList = new LinkedList();

    private boolean active = false;


    /**
     * Constructs an <code>Animation</code> with the specified duration.
     * 
     * @param duration  the animation's duration
     */
    protected AbstractAnimation(long duration) {
        this(duration, false);
    }

    /**
     * Constructs an <code>Animation</code> with the specified duration 
     * and freezed mode.
     * 
     * @param duration  the animation's duration
     * @param freezed   true indicates that the effect will be retained after
     * the animation is finished, false resets the effect to the time 0
     */
    protected AbstractAnimation(long duration, boolean freezed) {
        this.duration = duration;
        this.freezed  = freezed;
    }

    /**
     * Answers the animation duration.
     * 
     * @return the animation's duration
     */
    public long duration() {
        return duration;
    }

    /**
     * Returns whether the animation effect should be freezed after 
     * we exceeded the animation's duration. If this is not the case,
     * the animation effect of time 0 will be set.
     * 
     * @return true indicates that the effect will be retained if the 
     * animation duration exceeded; false indicates that the effect 
     * will be reset
     */
    public boolean isFreezed() {
        return freezed;
    }

    /**
     * Applies the animation effect for the given time to the animation target.
     * 
     * @param time   the time used to determine the animation effect
     */
    abstract protected void applyEffect(long time);

    /**
     * Perform the animation at the given time: applies the animation
     * effect to the animation target, fires appropriate events, 
     * and resets the effect if we exceeded the animations duration.
     * 
     * @param time   the time used to determine the animation effect
     */
    public void animate(long time) {
        if (time >= duration) {
            if (active) {
                applyEffect(isFreezed() ? duration - 1 : 0);
                fireAnimationStopped(time);
                active = false;
            }
            return;
        }

        if (!active) {
            active = true;
            fireAnimationStarted(time);
        }
        applyEffect(time);
    }

    /**
     * Adds an <code>AnimationListener</code> to this animation.
     * 
     * @param listener  the <code>AnimationListener</code> to add
     */
    public void addAnimationListener(AnimationListener listener) {
        listenerList.add(listener);
    }

    /**
     * Removes an <code>AnimationListener</code> to this animation.
     * 
     * @param listener  the <code>AnimationListener</code> to remove
     */
    public void removeAnimationListener(AnimationListener listener) {
        listenerList.remove(listener);
    }

    /**
     * Fires an event that indicates that the animation has been started 
     * at the specified time.
     * 
     * @param time   the time that will be reported in the event
     */
    protected void fireAnimationStarted(long time) {
        AnimationEvent e =
            new AnimationEvent(this, AnimationEvent.STARTED, time);
        for (Iterator i = listenerList.iterator(); i.hasNext();) {
            AnimationListener listener = (AnimationListener) i.next();
            listener.animationStarted(e);
        }
    }

    /**
     * Fires an event that indicates that the animation has been stopped
     * at the specified time.
     * 
     * @param time   the time that will be reported in the event
     */
    protected void fireAnimationStopped(long time) {
        AnimationEvent e =
            new AnimationEvent(this, AnimationEvent.STOPPED, time);
        for (Iterator i = listenerList.iterator(); i.hasNext();) {
            AnimationListener listener = (AnimationListener) i.next();
            listener.animationStopped(e);
        }
    }

    /**
     * Returns a string representation for this animation.
     * 
     * @return a string representation for this animation
     */
    public String toString() {
        return "["
            + getClass().getName()
            + "; duration="
            + duration
            + "; active="
            + active;
    }
}