package com.jgoodies.animation;

/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.
 * Use is subject to license terms.
 *
 */

/**
 * Describes events appropriate for animations: start or stop.
 *
 * @author Karsten Lentzsch
 */
public final class AnimationEvent {

    public static final Type STARTED = new Type("Started");
    public static final Type STOPPED = new Type("Stopped");
    
    private final Animation source;
    private final Type type;
    private final long time;

    /**
     * Constructs an <code>AnimationEvent</code> for the 
     * initiating animation, event type, and time.
     * 
     * @param source   the <code>Animation</code> that has originated the event
     * @param type     the event type: start or stop
     * @param time     the time the event has happened, which is likely 
     * before the event has been created
     */
    AnimationEvent(Animation source, Type type, long time) {
        this.source = source;
        this.type = type;
        this.time = time;
    }

    /**
     * Returns the animation the has created this event.
     * 
     * @return the animation that has originated this event
     */
    public Animation getSource() {
        return source;
    }

    /**
     * Returns the event type: start or stop.
     * 
     * @return the event type: start or stop
     */
    public Type type() {
        return type;
    }

    /**
     * Returns the time when this event has been created.
     * 
     * @return the event creation time
     */
    public long time() {
        return time;
    }

    /**
     * Returns an appropriate string representation.
     * 
     * @return a string representation for this event
     */
    public String toString() {
        return "[type= "
            + type
            + "; time= "
            + time
            + "; source="
            + source
            + ']';
    }
    
    // Helper Class *********************************************************
    
    /**
     * A typesafe enumeration for the event types.
     */
    private static final class Type {
        private final String name;

        private Type(String name) {
            this.name = name;
        }

        public String toString() {
            return name;
        }

    }

    
}