/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.
 * Use is subject to license terms.
 *
 */

package com.jgoodies.animation;

import java.awt.Color;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

/**
 * This class consists only of static methods that construct and operate on 
 * {@link AnimationFunction}s.
 *
 * @author Karsten Lentzsch
 */
public final class AnimationFunctions {
        
    /**
     * A constant {@link AnimationFunction} that returns 
     * <code>1</code> all the time.
     */
    public static final AnimationFunction ONE =
        constant(Integer.MAX_VALUE, new Float(1.0f));

    /**
     * A constant {@link AnimationFunction} that returns 
     * <code>0.0f</code> all the time.
     */
    public static final AnimationFunction ZERO =
        constant(Integer.MAX_VALUE, new Float(0.0f));


    // Overrides the default constructor; prevents instantiation.
    private AnimationFunctions() {}

    /**
     * Creates and answers an animation function, that answers time-based 
     * sRGB colors that are built from a given base color and 
     * an animation function of alpha values.
     * <p>
     * Useful for fading effects.
     * 
     * @param f  the animation function of alpha values
     * @param baseColor  the base color
     * @return an animation function of colors
     */
    public static AnimationFunction alphaColor(
        AnimationFunction f,
        Color baseColor) {
        return new AlphaColorAnimationFunction(f, baseColor);
    }

    /**
     * Creates a time-based function that wraps the given Float-based animation
     * function to return the corresponding float values.
     * 
     * @param f  the underlying animation function of floats
     * @return an animation function the answers Float objects
     */
    public static FloatFunction asFloat(AnimationFunction f) {
        return new FloatFunction(f);
    }

    /**
     * Concatenates the fiven animation functions and answers a compound
     * animation function that represents the concatenation.
     * 
     * @param first     the concatenation's first AnimationFunction
     * @param second    the concatenation's second AnimationFunction
     * @return the concatenated animation function
     */
    public static AnimationFunction concat(
        AnimationFunction first,
        AnimationFunction second) {
        List sequence = new LinkedList();
        sequence.add(first);
        sequence.add(second);
        return new SequencedAnimationFunction(sequence);
    }

    /**
     * Creates and answers an animation function that returns a constant value
     * over the given duration.
     * 
     * @param duration  the function's duration
     * @param value     the Object that will be returned all the time
     * @return a constant animation function
     */
    public static AnimationFunction constant(long duration, Object value) {
        return discrete(duration, new Object[] { value });
    }

    /**
     * Creates and answers a discrete animation function for the 
     * given duration and values. The values are equally distributed
     * over the duration.
     * 
     * @param duration  the function's duration
     * @param values    an array of discrete result values
     * @return a discrete animation function
     */
    public static AnimationFunction discrete(
        long duration,
        Object[] values) {
        return discrete(duration, values, null);
    }

    /**
     * Creates and answers a discrete animation function for the given duration,
     * values and interpolation key times.
     * 
     * @param duration  the function's duration
     * @param values    an array of discrete result values
     * @param keyTimes  an array of key times used to distribute the 
     *   result values over the time
     * @return a discrete animation function
     */
    public static AnimationFunction discrete(
        long duration,
        Object[] values,
        float[] keyTimes) {
        return new InterpolatedAnimationFunction(
            duration,
            values,
            keyTimes,
            InterpolatedAnimationFunction.MODE_DISCRETE);
    }

    /**
     * Creates and answers a linear animation function for the given duration 
     * that returns Float in interval [from, from + by].
     * 
     * @param duration   the animation duration
     * @param from       the initial result value
     * @param by         the difference that is added to the initial value
     * @return a linear animation function with values in [from, from + by]
     */
    public static AnimationFunction fromBy(
        long duration,
        float from,
        float by) {
        return fromTo(duration, from, from + by);
    }

    /**
     * Ceates and answers a linear animation function with the given duration.
     * The function values are Floats in the intervall [from, to].
     * 
     * @param duration   the animation duration
     * @param from       the initial result value
     * @param to         the last result value
     * @return a linear animation function with values in [from, to]
     */
    public static AnimationFunction fromTo(
        long duration,
        float from,
        float to) {
        return linear(
            duration,
            new Float[] { new Float(from), new Float(to)});
    }

    /**
     * Creates and answers a linear animation function, that is defined
     * by an array of numeric values; these are distributed equally 
     * over the duration.
     * 
     * @param duration   the animation duration
     * @param values     an array of values
     * @return a linear animation function for the given values
     */
    public static AnimationFunction linear(long duration, Object[] values) {
        return linear(duration, values, null);
    }

    /**
     * Creates and answers a linear animation function, that is defined
     * by an array of numeric values and an array of relative key times.
     * 
     * @param duration   the animation duration
     * @param values     an array of values
     * @param keyTimes   an array of key times used to distribute the 
     * result values over the time
     * @return a linear animation function for the given values
     */
    public static AnimationFunction linear(
        long duration,
        Object[] values,
        float[] keyTimes) {
        return new InterpolatedAnimationFunction(
            duration,
            values,
            keyTimes,
            InterpolatedAnimationFunction.MODE_LINEAR);
    }

    /**
     * Creates and answers an animation function, that answers random values
     * from the interval [min, max] with a given change probability.
     * 
     * @param min                the minimum result value
     * @param max                the maximum result value
     * @param changeProbability  the probability that the value changes
     * @return an animation function with random values in [min, max]
     */
    public static AnimationFunction random(
        int min,
        int max,
        float changeProbability) {
        return new RandomAnimationFunction(min, max, changeProbability);
    }

    /**
     * Creates and answers an animation function, that is defined 
     * by repeating the specified animation function.
     * 
     * @param f            the animation function to repeat
     * @param repeatTime   the time to repeat the function
     * @return the repeated animation function
     */
    public static AnimationFunction repeat(
        AnimationFunction f,
        long repeatTime) {
        return new RepeatedAnimationFunction(f, repeatTime);
    }

    /**
     * Creates and answers an animation function that is defined 
     * by reverting the given animation function in time.
     * 
     * @param f     the animation function to reverse
     * @return the reversed animation function
     */
    public static AnimationFunction reverse(AnimationFunction f) {
        return new ReversedAnimationFunction(f);
    }
    
    
    // Helper Classes *******************************************************

    // Helper class that provides code common to operations on animation functions.
    static abstract class AbstractAnimationFunction
        implements AnimationFunction {
        private final long duration;

        protected AbstractAnimationFunction(long duration) {
            if (duration < 0)
                throw new IllegalArgumentException("The duration must not be negative.");

            this.duration = duration;
        }

        protected void checkTimeRange(long time) {
            if ((time < 0) || (time > duration()))
                throw new IllegalArgumentException(
                    "The time must be larger than 0 and smaller than "
                        + duration()
                        + ".");
        }

        public long duration() {
            return duration;
        }

    }

    // Helper class for animation functions that answer colors.
    private static class AlphaColorAnimationFunction
        implements AnimationFunction {
        private final Color baseColor;
        private final AnimationFunction fAlpha;

        private AlphaColorAnimationFunction(
            AnimationFunction fAlpha,
            Color baseColor) {
            this.fAlpha = fAlpha;
            this.baseColor = baseColor;
        }

        public long duration() {
            return fAlpha.duration();
        }

        // Constructs colors from the sRGB color space.
        public Object valueAt(long time) {
            return new Color(
                baseColor.getRed(),
                baseColor.getGreen(),
                baseColor.getBlue(),
                ((Float) fAlpha.valueAt(time)).intValue());
        }
    }

    /**
     * Helper class that wraps a Float-based animation function to answer floats.
     */
    public static class FloatFunction {
        private final AnimationFunction f;

        FloatFunction(AnimationFunction f) {
            this.f = f;
        }

        public long duration() {
            return f.duration();
        }
        public float valueAt(long time) {
            return ((Number) f.valueAt(time)).floatValue();
        }
    }

    // Helper class for interpolation based animation functions.
    private static class InterpolatedAnimationFunction
        extends AbstractAnimationFunction {
        static final int MODE_DISCRETE = 1;
        static final int MODE_LINEAR   = 2;
        private final float[] keyTimes;
        private final int mode;

        /* Left for long winter nights in northern Germany.
        static final int MODE_PACED	     = 4; 
        static final int MODE_SPLINE	 = 8;  
        */

        private final Object[] values;

        private InterpolatedAnimationFunction(
            long duration,
            Object[] values,
            float[] keyTimes,
            int mode) {
            super(duration);
            this.values = values;
            this.keyTimes = keyTimes;
            this.mode = mode;
            checkValidKeyTimes(values.length, keyTimes);
        }

        private void checkValidKeyTimes(int valuesLength, float[] theKeyTimes) {
            if (theKeyTimes == null)
                return;

            if (valuesLength < 2 || valuesLength != theKeyTimes.length)
                throw new IllegalArgumentException("The values and key times arrays must be non-empty and must have equal length.");

            for (int index = 0; index < theKeyTimes.length - 2; index++) {
                if (theKeyTimes[index] >= theKeyTimes[index + 1])
                    throw new IllegalArgumentException("The key times must be increasing.");
            }
        }

        private Object discreteValueAt(long time) {
            return values[indexAt(time, values.length)];
        }

        private int indexAt(long time, int intervalCount) {
            // Gleichlange Zeitabschnitte
            if (keyTimes == null) {
                return (int) (time * intervalCount / duration());
            }
            for (int index = keyTimes.length - 2; index > 0; index--) {
                if (time > duration() * keyTimes[index])
                    return index;
            }
            return 0;
        }

        /**
         * Currently, we provide only linear interpolations that are based on floats.
         */
        private Object interpolateLinear(
            Object value1,
            Object value2,
            long time,
            long duration) {
            float f1 = ((Number) value1).floatValue();
            float f2 = ((Number) value2).floatValue();
            float value = f1 + (f2 - f1) * time / duration;
            return new Float(value);
        }

        private Object linearValueAt(long time) {
            int segments = values.length - 1;
            int beginIndex = indexAt(time, segments);
            int endIndex = beginIndex + 1;
            long beginTime =
                keyTimes == null
                    ? beginIndex * duration() / segments
                    : (long) (keyTimes[beginIndex] * duration());
            long endTime =
                keyTimes == null
                    ? endIndex * duration() / segments
                    : (long) (keyTimes[endIndex] * duration());

            return interpolateLinear(
                values[beginIndex],
                values[endIndex],
                time - beginTime,
                endTime - beginTime);
        }

        public Object valueAt(long time) {
            checkTimeRange(time);
            switch (mode) {
                case MODE_DISCRETE :
                    return discreteValueAt(time);

                case MODE_LINEAR :
                    return linearValueAt(time);

                default :
                    throw new IllegalStateException("Unsupported interpolation mode.");
            }
        }
    }

    // Helper class for animation function that answer random values.
    private static class RandomAnimationFunction
        implements AnimationFunction {
        private final float changeProbability;
        private final int max;
        private final int min;
        private final Random random;

        private Object value;

        private RandomAnimationFunction(
            int min,
            int max,
            float changeProbability) {
            this.random = new Random();
            this.min = min;
            this.max = max;
            this.changeProbability = changeProbability;
        }

        public long duration() {
            return Integer.MAX_VALUE;
        }

        public Object valueAt(long time) {
            if ((value == null)
                || (random.nextFloat() < changeProbability)) {
                value = new Integer(min + random.nextInt(max - min));
            }
            return value;
        }
    }

    // Helper class used to repeat or sequence an animation function.
    private static class RepeatedAnimationFunction
        extends AbstractAnimationFunction {
        private final AnimationFunction f;
        private final long simpleDuration;

        private RepeatedAnimationFunction(
            AnimationFunction f,
            long repeatTime) {
            super(repeatTime);
            this.f = f;
            this.simpleDuration = f.duration();
        }

        public Object valueAt(long time) {
            return f.valueAt(time % simpleDuration);
        }
    }

    // Helper class for reversing an animation function.
    private static class ReversedAnimationFunction
        extends AbstractAnimationFunction {
        private final AnimationFunction f;

        private ReversedAnimationFunction(AnimationFunction f) {
            super(f.duration());
            this.f = f;
        }

        public Object valueAt(long time) {
            return f.valueAt(duration() - time);
        }
    }

    // Helper class to compose an animation functions from a sequences of such functions.
    private static class SequencedAnimationFunction
        implements AnimationFunction {
        private final List functions;

        private SequencedAnimationFunction(List functions) {
            this.functions = Collections.unmodifiableList(functions);
            if (this.functions.isEmpty())
                throw new IllegalArgumentException("The list of functions must not be empty.");
        }

        public long duration() {
            long cumulatedDuration = 0;
            for (Iterator i = functions.iterator(); i.hasNext();) {
                AnimationFunction f = (AnimationFunction) i.next();
                cumulatedDuration += f.duration();
            }
            return cumulatedDuration;
        }

        public Object valueAt(long time) {
            if (time < 0)
                throw new IllegalArgumentException("The time must be positive.");

            long begin = 0;
            long end;
            for (Iterator i = functions.iterator(); i.hasNext();) {
                AnimationFunction f = (AnimationFunction) i.next();
                end = begin + f.duration();
                if (time < end)
                    return f.valueAt(time - begin);
                begin = end;
            }
            throw new IllegalArgumentException("The time must be smaller than the total duration.");
        }
    }
}