package com.jgoodies.animation;

/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.
 * Use is subject to license terms.
 *
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.Timer;

/**
 * This class starts and stops an animation and triggers
 * the animation at a given frame rate.
 *
 * @author Karsten Lentzsch
 */
public final class Animator implements ActionListener {

    private final Animation animation;
    private final Timer     timer;
    private final int       framesPerSecond;

    private long startTime;
    private long elapsedTime  = 0;

    /**
     * Constructs an <code>Animator</code> for the given animation and 
     * frame rate.
     * 
     * @param animation         the animation to animate
     * @param framesPerSecond   the desired frame rate
     */
    public Animator(Animation animation, int framesPerSecond) {
        if (animation == null)
            throw new NullPointerException("The animation must not be null.");

        if (framesPerSecond <= 0)
            throw new IllegalArgumentException("The frame rate must be positive.");

        this.animation = animation;
        this.framesPerSecond = framesPerSecond;
        this.timer = createTimer(framesPerSecond);
    }

    /**
     * Creates and configures a <code>Timer</code> object.
     * 
     * @param fps    the frames per second
     * @param a <code>Timer</code> with the specified frame rate
     */
    private Timer createTimer(int fps) {
        int delay = 1000 / fps;

        Timer aTimer = new Timer(delay, this);
        aTimer.setInitialDelay(0);
        aTimer.setCoalesce(true);
        return aTimer;
    }

    /**
     * Returns the animator's animation.
     * 
     * @return the animator's animation 
     */
    public Animation animation() {
        return animation;
    }

    /**
     * Returns the desired frame rate.
     * 
     * @return the desired frame rate per second
     */
    public int framesPerSecond() {
        return framesPerSecond;
    }

    /**
     * Returns the elapsed time since animation start.
     * 
     * @return time elapsed since the animation start
     */
    public long elapsedTime() {
        return timer.isRunning()
            ? System.currentTimeMillis() - startTime + elapsedTime
            : elapsedTime;
    }

    /**
     * Starts the animator and in turn the animation.
     */
    public void start() {
        if (!timer.isRunning()) {
            registerStopListener();
            startTime = System.currentTimeMillis();
            timer.start();
        } /*else {
               	System.out.println("Animator is already running");
        		} */
    }

    /**
     * Stops the animator.
     */
    public void stop() {
        if (timer.isRunning()) {
            elapsedTime = elapsedTime();
            timer.stop();
        }
    }

    /**
     * Registers a listener that stops the animator if the animation stopped.
     */
    private void registerStopListener() {
        animation.addAnimationListener(new AnimationAdapter() {
            public void animationStopped(AnimationEvent e) {
                //System.out.println("All animations stopped.");
                stop();
                //animation.animate(animation.duration());
            }
        });
    }

    /**
     * Implements the ActionListener interface for use by 
     * the <code>Timer</code>.
     * 
     * @param e  the action event
     */
    public void actionPerformed(ActionEvent e) {
        animation.animate(elapsedTime());
    }

    /**
     * Returns a string representation for the animator.
     * 
     * @return a string representation for the animator
     */
    public String toString() {
        return "elapsedTime=" + elapsedTime() + "; fps=" + framesPerSecond;
    }
}