package com.jgoodies.animation.animations;

/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */

import java.awt.Color;

import com.jgoodies.animation.*;
import com.jgoodies.animation.components.BasicTextLabel;

/**
 * A basic text animation that fades in a text; changes the x and y scaling,
 * the glyph space and its position.
 * 
 * @author Karsten Lentzsch
 */
public final class BasicTextAnimation extends AbstractAnimation {

	private final BasicTextLabel label;
	private final String text;
	private final AnimationFunction colorFunction;
	private final AnimationFunctions.FloatFunction offsetXFunction;
	private final AnimationFunctions.FloatFunction offsetYFunction;
	private final AnimationFunctions.FloatFunction scaleXFunction;
	private final AnimationFunctions.FloatFunction scaleYFunction;
	private final AnimationFunctions.FloatFunction spaceFunction;

	private boolean offsetEnabled = false;

	/**
	 * Constructs a text animation, that fades in a text, scales it and 
	 * fades it out. Uses the given color and scaling functions.
	 * 
	 * @param label           the animation target component
	 * @param duration        the animation duration 
	 * @param text            the text to fade in
	 * @param colorFunction   the animation function for the color
	 * @param scaleXFunction  the animation function for the horizontal scale
	 * @param scaleYFunction  the animation function for the vertical scale
	 * @param spaceFunction   the animation function for the glyph space
	 */
	public BasicTextAnimation(
		BasicTextLabel label,
		long duration,
		String text,
		AnimationFunction colorFunction,
		AnimationFunction scaleXFunction,
		AnimationFunction scaleYFunction,
		AnimationFunction spaceFunction,
		AnimationFunction offsetXFunction,
		AnimationFunction offsetYFunction) {

		super(duration);
		this.label = label;
		this.text = text;
		this.colorFunction =
			colorFunction != null ? colorFunction : defaultFadeColorFunction(duration, Color.darkGray);
		this.scaleXFunction =
			AnimationFunctions.asFloat(scaleXFunction != null ? scaleXFunction : AnimationFunctions.ONE);
		this.scaleYFunction =
			AnimationFunctions.asFloat(scaleYFunction != null ? scaleYFunction : AnimationFunctions.ONE);
		this.spaceFunction =
			AnimationFunctions.asFloat(spaceFunction != null ? spaceFunction : AnimationFunctions.ZERO);
		this.offsetXFunction =
			AnimationFunctions.asFloat(offsetXFunction != null ? offsetXFunction : AnimationFunctions.ZERO);
		this.offsetYFunction =
			AnimationFunctions.asFloat(offsetYFunction != null ? offsetYFunction : AnimationFunctions.ZERO);
	}

	/**
	 * Creates and answers an animation, that provides a text fade-in 
	 * and -out and scales the text while fading out.
	 * 
	 * @param label           the animation target component
	 * @param duration        the animation duration 
	 * @param text            the text to fade in
	 * @param baseColor       the base color for the fade effect
	 * @return an animation with a default scaling text effect
	 */
	public static BasicTextAnimation defaultScale(
		BasicTextLabel label,
		long duration,
		String text,
		Color baseColor) {

		return new BasicTextAnimation(
			label,
			duration,
			text,
			defaultScaleColorFunction(duration, baseColor),
			defaultScaleFunction(duration),
			defaultScaleFunction(duration),
			null,
			null,
			null);
	}

	/**
	 * Creates and answers an animation, that provides a text fade-in and -out.
	 * 
	 * @param label           the animation target component
	 * @param duration        the animation duration 
	 * @param text            the text to fade in
	 * @param baseColor       the base color for the fade effect
	 * @return an animation with a default text fade
	 */
	public static BasicTextAnimation defaultFade(
		BasicTextLabel label,
		long duration,
		String text,
		Color baseColor) {

		return new BasicTextAnimation(
			label,
			duration,
			text,
			cinemaFadeColorFunction(duration, baseColor),
			null,
			null,
			null,
			null,
			null);
	}

	/**
	 * Creates and answers an animation, that provides a text fade-in 
	 * and -out and increases the glyph spacing. 
	 * 
	 * @param label           the animation target component
	 * @param duration        the animation duration 
	 * @param text            the text to fade in
	 * @param baseColor       the base color for the fade effect
	 * @return an animation with a default glyph spacing effect
	 */
	public static BasicTextAnimation defaultSpace(
		BasicTextLabel label,
		long duration,
		String text,
		Color baseColor) {

		return new BasicTextAnimation(
			label,
			duration,
			text,
			defaultSpaceColorFunction(duration, baseColor),
			null,
			null,
			defaultSpaceFunction(duration),
			null,
			null);
	}

	/**
	 * Creates and answers the color animation function for the default fade.
	 * 
	 * @param duration        the animation duration 
	 * @param baseColor       the base color for the fade effect
	 * @return a Color-valued animation function for the default fade
	 */
	public static AnimationFunction defaultFadeColorFunction(long duration, Color baseColor) {
		return AnimationFunctions.alphaColor(
			AnimationFunctions.linear(
				duration,
				new Integer[] { new Integer(0), new Integer(255), new Integer(255), new Integer(0)},
				new float[] { 0.0f, 0.3f, 0.7f, 1.0f }),
			baseColor);
	}

	/**
	 * Creates and answers the color animation function for the default fade.
	 * 
	 * @param duration        the animation duration 
	 * @param baseColor       the base color for the fade effect
	 * @return a Color-valued animation function for the default fade
	 */
	public static AnimationFunction cinemaFadeColorFunction(long duration, Color baseColor) {
		return AnimationFunctions.alphaColor(
			AnimationFunctions.linear(
				duration,
				new Integer[] { new Integer(0), new Integer(255), new Integer(255), new Integer(0)},
				new float[] { 0.0f, 0.3f, 0.85f, 1.0f }),
			baseColor);
	}

	/**
	 * Creates and answers the animation function for the default scaling.
	 * 
	 * @param duration        the animation duration 
	 * @param baseColor       the base color for the fade effect
	 * @return a Color-valued animation function for the default scaling
	 */
	public static AnimationFunction defaultScaleColorFunction(long duration, Color baseColor) {
		return AnimationFunctions.alphaColor(
			AnimationFunctions.linear(
				duration,
				new Integer[] { new Integer(0), new Integer(255), new Integer(255), new Integer(0)},
				new float[] { 0.0f, 0.2f, 0.85f, 1.0f }),
			baseColor);
	}

	/**
	 * Creates and answers the color animation function for 
	 * the default spacing animation.
	 * 
	 * @param duration        the animation duration 
	 * @param baseColor       the base color for the fade effect
	 * @return a Color-valued animation function for the default spacing
	 */
	public static AnimationFunction defaultSpaceColorFunction(long duration, Color baseColor) {
		return AnimationFunctions.alphaColor(
			AnimationFunctions.linear(
				duration,
				new Integer[] { new Integer(0), new Integer(255), new Integer(255), new Integer(0)},
				new float[] { 0.0f, 0.2f, 0.8f, 1.0f }),
			baseColor);
	}

	/** 
	 * Creates and answers the default scaling animation function.
	 * 
	 * @param duration    the animation duration
	 * @return an animation function for the default scaling effect
	 */
	public static AnimationFunction defaultScaleFunction(long duration) {
		return AnimationFunctions.linear(
			duration,
			new Float[] { new Float(1.0f), new Float(1.0f), new Float(1.8f)},
			new float[] { 0.0f, 0.85f, 1.0f });
	}

	/** 
	 * Creates and answers the default spacing animation function.
	 * 
	 * @param duration    the animation duration
	 * @return an animation function for the default spacing effect
	 */
	public static AnimationFunction defaultSpaceFunction(long duration) {
		return AnimationFunctions.fromTo(duration, 0, 10);
	}

	/**
	 * Applies the effect: sets color, spacing, scaling and offset,
	 * the latter only if enabled.
	 * 
	 * @param time    the render time
	 */
	protected void applyEffect(long time) {
		label.setText(time == 0 ? " " : text);
		label.setColor((Color)colorFunction.valueAt(time));
		label.setScaleX(scaleXFunction.valueAt(time));
		label.setScaleY(scaleYFunction.valueAt(time));
		label.setSpace(spaceFunction.valueAt(time));
		if (isOffsetEnabled()) {
			label.setOffsetX(offsetXFunction.valueAt(time));
			label.setOffsetY(offsetYFunction.valueAt(time));
		}
	}

	/**
	 * Returns whether the random position offset is enabled.
	 * 
	 * @return true indicates offset enabled, false disabled 
	 */
	public boolean isOffsetEnabled() {
		return offsetEnabled;
	}

	/**
	 * Enables or disables the random position offset
	 * 
	 * @param b  the new enablement
	 */
	public void setOffsetEnabled(boolean b) {
		offsetEnabled = b;
	}

}