package com.jgoodies.animation.animations;

/*
 * Copyright (c) 2002 Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */
 
import java.awt.Color;
import java.util.LinkedList;
import java.util.List;

import com.jgoodies.animation.Animation;
import com.jgoodies.animation.AnimationUtils;
import com.jgoodies.animation.Animations;
import com.jgoodies.animation.components.BasicTextLabel;


/**
 * Provides a text animation that shows an overlapping sequence of 
 * texts using a bunch of different effects: color fade, scaling, glyph spacing.
 * 
 * @author Karsten Lentzsch
 */
public final class BasicTextAnimations {
	
	private static final int FADE_TYPE	= 0;
	private static final int SCALE_TYPE	= 1;
	private static final int SPACE_TYPE	= 2;
	
	
	// Override default constructor; prevents instantiation.
	private BasicTextAnimations() {}
	

	/**
	 * Creates and answers the default color fade text sequence.
     * 
     * @param label1           a text label used to blend over
     * @param label2           a second text label used to blend over
     * @param singleDuration   the duration of a single animation
     * @param beginOffset      an initial animation time offset
     * @param separatedTexts   a sequence of texts in a string separated
     * by the | character
     * @param baseColor        the base color for the fade
     * @return a default fade animation
	 */
	public static Animation defaultFade(
		BasicTextLabel label1,
		BasicTextLabel label2,
		long singleDuration,
		long beginOffset,
		String separatedTexts,
		Color baseColor) {
			
		return createTextSequence(label1, label2, 
								   singleDuration, beginOffset, 
								   separatedTexts, baseColor, 
								   FADE_TYPE);
	}
	

	/**
	 * Creates and answers the default scaling text sequence.
     * 
     * @param label1           a text label used to blend over
     * @param label2           a second text label used to blend over
     * @param singleDuration   the duration of a single animation
     * @param beginOffset      an initial animation time offset
     * @param separatedTexts   a sequence of texts in a string separated
     * by the | character
     * @param baseColor        the base color for the fade
     * @return a default scaling blend over animation
	 */
	public static Animation defaultScale(
		BasicTextLabel label1,
		BasicTextLabel label2,
		long singleDuration,
		long beginOffset,
		String separatedTexts,
		Color baseColor) {
			
		return createTextSequence(label1, label2, 
								   singleDuration, beginOffset, 
								   separatedTexts, baseColor, 
								   SCALE_TYPE);
	}


	/**
	 * Creates and answers the default glyph spacing text sequence.
     * 
     * @param label1           a text label used to blend over
     * @param label2           a second text label used to blend over
     * @param singleDuration   the duration of a single animation
     * @param beginOffset      an initial animation time offset
     * @param separatedTexts   a sequence of texts in a string separated
     * by the | character
     * @param baseColor        the base color for the fade
     * @return a default space blend over animation
	 */
	public static Animation defaultSpace(
		BasicTextLabel label1,
		BasicTextLabel label2,
		long singleDuration,
		long beginOffset,
		String separatedTexts,
		Color baseColor) {
			
		return createTextSequence(label1, label2, 
								   singleDuration, beginOffset, 
								   separatedTexts, baseColor, 
								   SPACE_TYPE);
	}
	

	// Private Helper Code *************************************************************

	/**
	 * Creates and answers the default glyph spacing text sequence.
	 */
	private static Animation createTextSequence(
		BasicTextLabel label1,
		BasicTextLabel label2,
		long singleDuration,
		long beginOffset,
		String separatedTexts,
		Color baseColor,
		int type) {
			
		Animation animation;
		String[] texts  = AnimationUtils.splitTexts(separatedTexts);
		List animations = new LinkedList();
		long beginTime	= 0;
		
		BasicTextLabel label = label1;
		for (int i = 0; i < texts.length; i++) {
			label = i % 2 == 0 ? label1 : label2;
			animation = animation(label, singleDuration, texts[i], baseColor, type);
			animations.add(Animations.offset(beginTime, animation));
			beginTime += singleDuration + beginOffset;
		}

		return Animations.parallel(animations);
	}
	
	
	private static Animation animation(BasicTextLabel label, 
								 long duration, 
								 String text, 
								 Color baseColor, 
								 int type) {
		switch (type) {
			case FADE_TYPE :
				return BasicTextAnimation.defaultFade(label, duration, text, baseColor);

			case SCALE_TYPE :
				return BasicTextAnimation.defaultScale(label, duration, text, baseColor);

			case SPACE_TYPE :
				return BasicTextAnimation.defaultSpace(label, duration, text, baseColor);

			default :
				return null;
		}
	}

	
}