package com.jgoodies.animation.animations;

/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */
 
import com.jgoodies.animation.AbstractAnimation;
import com.jgoodies.animation.AnimationFunction;
import com.jgoodies.animation.AnimationFunctions;
import com.jgoodies.animation.components.FanComponent;


/**
 * An animation that rotates a fan, that is a set of translucent sectors.
 * 
 * @author Karsten Lentzsch
 */
public final class FanAnimation extends AbstractAnimation {
	
	public  static final double DEFAULT_CLOCKWISE_ROTATION     =  0.01; // rotations per second
	public  static final double DEFAULT_ANTICLOCKWISE_ROTATION = -0.02;
	
	
	private final FanComponent fan;
	private final AnimationFunctions.FloatFunction rotationFunction;
	
	
	/**
	 * Constructs an animation that rotates a fan using the given fan component,
	 * duration and rotation animation function.
     * 
     * @param fan                the fan component animation target
     * @param duration           the animation duration
     * @param rotationFunction   the rotation animation function
	 */
	public FanAnimation(FanComponent fan, long duration, AnimationFunction rotationFunction) {
		super(duration);	
		this.fan = fan;
		this.rotationFunction  = 
			AnimationFunctions.asFloat(rotationFunction != null 
			? rotationFunction
			: defaultRotationFunction(duration));
	}


	/**
	 * Creates and answers the default fan animation.
     * 
     * @param fan        the fan component animation target
     * @param duration   the animation duration
     * @return a default fan animation
	 */
	public static FanAnimation defaultFan(FanComponent fan, long duration) {
		return new FanAnimation(fan, duration, null);
	}


	/** 
	 * Creates and answers an animation function for the default rotation.
     * 
     * @param duration   the animation duration
     * @return an animation function for the default rotation
	 */
	public static AnimationFunction defaultRotationFunction(long duration) {
		return AnimationFunctions.fromTo(duration, 0f, (float) (Math.PI * 2.0f));
	}


	/**
	 * Applies the effect: sets the time-based rotation.
     * 
     * @param time    the render time
	 */
	protected void applyEffect(long time) {
		fan.setRotation(rotationFunction.valueAt(time));
	}
	
}