package com.jgoodies.animation.components;

/*
 * Copyright (c) 2002 Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;

import javax.swing.JComponent;

import com.jgoodies.animation.AnimationFunction;
import com.jgoodies.animation.AnimationFunctions;
import com.jgoodies.animation.renderer.GlyphRenderer;
import com.jgoodies.animation.renderer.HeightMode;

/**
 * A Swing component, that can transform a text's individual glyphs.
 *
 * @author Karsten Lentzsch
 */
public final class GlyphLabel extends JComponent {

	private final GlyphRenderer renderer;

	/**
	 * Erzeugt ein skalierbares Label zu dem gegebenen <code>text</code>.
	 */
	public GlyphLabel(String text, long duration, long glyphDelay) {
		this(text, duration, glyphDelay, Color.darkGray);
	}

	/**
	 * Erzeugt ein skalierbares Label zu dem gegebenen <code>text</code>.
	 */
	public GlyphLabel(String text, long duration, long glyphDelay, Color baseColor) {
		this(
			new GlyphRenderer(
				text,
				defaultScaleFunction(duration),
				AnimationFunctions.ZERO,
				defaultColorFunction(duration, baseColor),
				glyphDelay));
	}

	public GlyphLabel(GlyphRenderer glyphRenderer) {
		renderer = glyphRenderer;
	}

	/** 
	 * Liefert die Standard-Skalierungsfunktion.	
	 */
	public static AnimationFunction defaultScaleFunction(long duration) {
		return AnimationFunctions.linear(
			duration,
			new Float[] { new Float(5.0f), new Float(0.8f), new Float(1.0f), new Float(1.0f)},
			new float[] { 0.0f, 0.1f, 0.12f, 1.0f });
	}

	/**
	 * Liefert die Standard-Einblendfunktion.
	 */
	public static AnimationFunction defaultColorFunction(long duration, Color baseColor) {
		return AnimationFunctions.alphaColor(
			AnimationFunctions.linear(
				duration,
				new Integer[] { new Integer(0), new Integer(255), new Integer(255)},
				new float[] { 0.0f, 0.15f, 1.0f }),
			baseColor);
	}

	public HeightMode getHeightMode() {
		return renderer.getHeightMode();
	}

	public String getText() {
		return renderer.getText();
	}

	public void setHeightMode(HeightMode heightMode) {
		renderer.setHeightMode(heightMode);
	}

	public void setText(String newText) {
		renderer.setText(newText);
		repaint();
	}

	public void setTime(long time) {
		renderer.setTime(time);
		repaint();
	}

	/**
	 * Paints the component. Sets high-fidelity rendering hints, 
	 * then invoke the renderer to render the glyphs.
	 */
	public void paintComponent(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;

		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

		renderer.setFont(getFont());
		renderer.render(g2, getWidth(), getHeight());
	}

}