package com.jgoodies.animation.demo;

/*
 * Copyright (c) 2003 Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */
 
import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.GridLayout;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

import com.jgoodies.animation.*;
import com.jgoodies.animation.Animation;
import com.jgoodies.animation.AnimationAdapter;
import com.jgoodies.animation.AnimationEvent;
import com.jgoodies.animation.Animations;
import com.jgoodies.animation.Animator;
import com.jgoodies.animation.animations.BasicTextAnimation;
import com.jgoodies.animation.components.BasicTextLabel;

/** 
 * A test class for a bunch of animations.
 * 
 * @author Karsten Lentzsch
 */
public final class AnimationDemo {
	
	private JLabel		 	reportLabel;
	private BasicTextLabel label1;
	private BasicTextLabel label2;
	private BasicTextLabel label3;

	
	public static void main(String[] args) {
		// Try to get the framerate from the command-line.
		int fps = 30;
		if (args.length > 0) {
			try {
				fps = Integer.parseInt(args[0]);
			} catch (NumberFormatException e) {}
		}
		System.out.println("The desired framerate is " + fps + '.');	
		new AnimationDemo(fps);
	}
	
	
	/**
	 * Constructs an animation demo using the specified frame rate.
	 */
	private AnimationDemo(int fps) {
		createComponents();
		buildFrame();
		
		Animator animator = new Animator(createComposedAnimation(), fps);
		animator.start();
	}


	// Building *************************************************************************
	
	/**
	 * Creates and configures the components used by the animations.
	 */
	private void createComponents() {
		Font font = new Font("dialog", Font.BOLD, 16);

		reportLabel = new JLabel();

		label1 = new BasicTextLabel(" ");
		label1.setFont(font);

		label2 = new BasicTextLabel(" ");
		label2.setFont(font);
		
		label3 = new BasicTextLabel(" ");
		label3.setFont(font);
	}
	

	/**
	 * Builds and opens the demo frame.
	 */	
	private void buildFrame() {
		JFrame frame = new JFrame();
		frame.setContentPane(buildContent());
		frame.setSize(400, 300);
		frame.setTitle("Animation Test");
		frame.setDefaultCloseOperation(3); // EXIT_ON_CLOSE
		frame.setVisible(true);
	}
	

	/**
	 * Builds and answers the frame's content pane.
	 */
	private Container buildContent() {
		JPanel panel = new JPanel(new GridLayout(4, 0));
		panel.add(reportLabel);
		panel.add(label1);
		panel.add(label2);
		panel.add(label3);
		panel.setBorder(BorderFactory.createEmptyBorder(12, 12, 12, 12));
		return panel;
	}


	// Animation Creation ***************************************************************
		
	/**
	 * Creates and answers a parallel animation that is composed of four animations: 
	 * report, fade, scaling, and spacing.
	 */
	private Animation createComposedAnimation() {
		List animations = new LinkedList();
		
		animations.add(Animations.repeat(5f, new SimpleReportAnimation(reportLabel, 2000)));
		animations.add(BasicTextAnimation.defaultFade(label1, 
													  5000,
													  "Default Fade", 
													  Color.darkGray));
		animations.add(BasicTextAnimation.defaultScale(label2,  
													   4000,
													   "Default Scale", 
													   Color.darkGray));
		animations.add(BasicTextAnimation.defaultSpace(label3, 3000,
													   "Default Space", 
													   Color.darkGray));
		
		int count = 1;	
		for (Iterator i = animations.iterator(); i.hasNext();) {
			Animation element = (Animation) i.next();
			element.addAnimationListener(new AnimationLogger("Animation " + count++));
		}
		
		Animation composedAnimation = Animations.parallel(animations);
		composedAnimation.addAnimationListener(new AnimationLogger("Composed animation"));
		return composedAnimation;
	}


	// Helper Class *********************************************************************
	
	private static class SimpleReportAnimation extends AbstractAnimation {
	
		private final JLabel target;
		
		private long frameNumber =  0;
		private long startTime   = -1;
	
		/**
		 * Constructs a simple report animation.
		 */
		private SimpleReportAnimation(JLabel target, long duration) {
			super(duration);
			this.target = target;
		}
	
		/**
		 * Applies the animation effect: sets a text for the time, total time, 
		 * iteration, frame number, and - after a second - the total frame rate.
		 */
		protected void applyEffect(long time) {
			frameNumber++;
			long  totalTime = totalTime();  
			float fps       = totalTime < 1000 ? 0 : frameNumber * 1000 / totalTime;
			int   iteration = (int) (totalTime / duration());

			target.setText(
				"time=" + time / 1000 + "s"
				+ "; total=" + totalTime / 1000 + "s"
				+ "; iteration=" + iteration 
				+ "; frame=" + frameNumber
				+ "; fps= " + fps);
		}
		
		private long totalTime() {
			if (startTime == -1) {
				startTime = System.currentTimeMillis();
			}
			return System.currentTimeMillis() - startTime;
		}
		
	}
	
	
	// An AnimationListener that logs animation events to the console.
	private static class AnimationLogger extends AnimationAdapter {
		private final String name;
		
		AnimationLogger(String name) { this.name = name; }
		
		public void animationStarted(AnimationEvent e) {
			System.out.println(name + " started.");
		}
		public void animationStopped(AnimationEvent e) {
			System.out.println(name + " stopped.");
		}
	}
	

}