/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */
 
package com.jgoodies.animation.renderer;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.Random;

import com.jgoodies.animation.AnimationRenderer;


/**
 * Paints two colored and often translucent fans, that can be turned around.
 *
 * @author Karsten Lentzsch
 */
public final class FanRenderer implements AnimationRenderer {
	
	private static final Random random = new Random(System.currentTimeMillis());
	
	
	private final Triangle[]	triangles;
	
	private Point2D origin;
	private double rotation; 

	
	public FanRenderer(Triangle[] triangles) {
		this.triangles   = triangles;
	}


	public FanRenderer(int triangleCount, Color baseColor) {
		this(createSectors(triangleCount, baseColor));
	}
	
	
	public static Triangle[] createSectors(int count, Color baseColor) {
		Triangle[] result = new Triangle[count];
		double sectorAngle = Math.PI * 2 / count;

		for (int i = 0; i < count; i++) {
			double rotation = i * sectorAngle + (random.nextFloat() - 0.5) * Math.PI/10;
			double angle    = sectorAngle * (0.2 + random.nextFloat() * 0.4);
			result[i] = new Triangle(rotation, angle, nextColor(baseColor));
		}
		return result;
	}
	
	private static Color nextColor(Color baseColor) {
		float[] hsb = new float[3];
		Color.RGBtoHSB(baseColor.getRed(), baseColor.getGreen(), baseColor.getBlue(), hsb);
		float brightness = 0.8f + random.nextFloat() * 0.2f;
		return Color.getHSBColor(hsb[0], hsb[1], brightness);
	}
	
	
	public Point2D getOrigin()						{ return origin;			}
	public void   setOrigin(Point2D origin)		{ this.origin = origin;	}
	public double getRotation() 					{ return rotation; 		}
	public void   setRotation(double rotation)	{ this.rotation = rotation; }
	
	
	public void render(Graphics2D g2, int width, int height) {
		double radius = Math.sqrt(width * width + height * height);
		
		Point2D p = getOrigin() != null ? getOrigin() : getDefaultOrigin(width, height);
		
		g2.translate(p.getX(), p.getY());
		g2.rotate(rotation);
		for (int i = 0; i < triangles.length; i++) {
			triangles[i].render(g2, radius);
		}
		g2.rotate(-rotation);
		g2.translate(-p.getX(), -p.getY());
	}
	
	
	private Point2D getDefaultOrigin(int width, int height) {
		return new Point2D.Double(width * 0.75, height * 0.75);
	}
	

	// Helper Class *********************************************************************
	
		
	// A helper class that models and renders a single sector.
	private static class Triangle {
		
		private final double rotation;
		private final double angle;
		private final Color   color;
		
		private Triangle(double rotation, double angle, Color color) {
			this.rotation = rotation;
			this.angle    = angle;
			this.color    = color;
		}
		
		private static Shape createPolygon(double rotation, double angle, double radius) {
			double startAngle   = rotation - angle / 2;
			double stopAngle    = startAngle + angle;
			double hypothenusis = radius / Math.cos(angle / 2);
		
			float x0 = 0.0f;
			float y0 = 0.0f;
			float x1 = (float) (x0 - hypothenusis * Math.cos(startAngle));
			float y1 = (float) (y0 - hypothenusis * Math.sin(startAngle));
			float x2 = (float) (x0 - hypothenusis * Math.cos(stopAngle));
			float y2 = (float) (y0 - hypothenusis * Math.sin(stopAngle));
			
			GeneralPath polygon = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 3);
			polygon.moveTo(x0, y0);
			polygon.lineTo(x1, y1);
			polygon.lineTo(x2, y2);
			polygon.closePath();
			
			return polygon;
		}
		
		void render(Graphics2D g2, double radius) {
			g2.setColor(color);
			g2.fill(createPolygon(rotation, angle, radius));
		}
	}
	
}