/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * This software is the proprietary information of JGoodies Karsten Lentzsch.  
 * Use is subject to license terms.
 *
 */

package com.jgoodies.animation.renderer;

import java.awt.Color;
import java.awt.Graphics2D;

import com.jgoodies.animation.*;

/**
 * Renders the glyphs of a string with individual scaling, transform, and color.
 *
 * @author Karsten Lentzsch
 */
public final class GlyphRenderer extends AbstractTextRenderer {

	private final AnimationFunction colorFunction;
	private final AnimationFunctions.FloatFunction scaleFunction;
	private final AnimationFunctions.FloatFunction offsetXFunction;
	private final AnimationFunctions.FloatFunction offsetYFunction;

	private final long glyphDelay;

	private long time;

	/**
	 * Constructs a <code>GlyphRenderer</code> that paints 
	 * individual glyphs with different transforms.
	 */
	public GlyphRenderer(
		String text,
		AnimationFunction scaleFunction,
		AnimationFunction translateXFunction,
		AnimationFunction translateYFunction,
		AnimationFunction colorFunction,
		long glyphDelay) {
		super(text);
		this.scaleFunction = AnimationFunctions.asFloat(scaleFunction);
		this.offsetXFunction = AnimationFunctions.asFloat(translateXFunction);
		this.offsetYFunction = AnimationFunctions.asFloat(translateYFunction);
		this.colorFunction = colorFunction;
		this.glyphDelay = glyphDelay;
		this.time = 0;
	}

	/**
	 * Constructs a <code>GlyphRenderer</code> that paints 
	 * individual glyphs with different transforms.
	 */
	public GlyphRenderer(
		String text,
		AnimationFunction scaleFunction,
		AnimationFunction translateFunction,
		AnimationFunction colorFunction,
		long glyphDelay) {
		this(text, scaleFunction, translateFunction, translateFunction, colorFunction, glyphDelay);
	}

	public void setTime(long time) {
		this.time = time;
	}

	private long relativeTime(int glyphIndex) {
		return Math.max(0, time - glyphDelay * glyphIndex);
	}

	private float scaleAt(int glyphIndex) {
		return scaleFunction.valueAt(relativeTime(glyphIndex));
	}

	private float offsetXAt(int glyphIndex) {
		return offsetXFunction.valueAt(relativeTime(glyphIndex));
	}

	private float offsetYAt(int glyphIndex) {
		return offsetYFunction.valueAt(relativeTime(glyphIndex));
	}

	private Color colorAt(int glyphIndex) {
		return (Color)colorFunction.valueAt(relativeTime(glyphIndex));
	}

	/**
	 * Renders the text. Firstly, ensures a valid cache,
	 * then sets the color, and finally paints the cached glyph shaped,
	 * using individual transforms.
	 */
	public void render(Graphics2D g2, int width, int height) {
		ensureValidCache(g2);

		int glyphCount = cachedGlyphShapes.length;
		float offsetX = (width - cachedTextWidth) / 2.0f;
		float offsetY = (height + cachedTextHeight) / 2.0f - getAdjustedDescent();

		g2.translate(offsetX, offsetY);
		for (int i = glyphCount - 1; i >= 0; i--) {
			float scale = scaleAt(i);
			float translateX = offsetXAt(i);
			float translateY = offsetYAt(i);

			g2.setColor(colorAt(i));
			g2.translate(translateX, translateY);

			double glyphX = cachedGlyphVector.getGlyphPosition(i).getX();
			double glyphY = cachedGlyphVector.getGlyphVisualBounds(i).getBounds2D().getHeight();
			double adjustX = -glyphX * (scale - 1.0f);
			double adjustY = glyphY * (scale - 1.0f) / 2.0f;
			g2.translate(adjustX, adjustY);
			g2.scale(scale, scale);
			if (!AnimationUtils.IS_14_ENVIRONMENT) {
				g2.translate(glyphX, 0);
				g2.fill(cachedGlyphShapes[i]);
				g2.translate(-glyphX, 0);
			} else {
				g2.fill(cachedGlyphShapes[i]);
			}
			g2.scale(1.0f / scale, 1.0f / scale);
			g2.translate(-adjustX, -adjustY);
			g2.translate(-translateX, -translateY);
		}
		g2.translate(-offsetX, -offsetY);
	}
}