package de.uniwue.ki.gui.mas.application;

import jargs.gnu.CmdLineParser;
import jargs.gnu.CmdLineParser.*;

import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.*;
import java.util.logging.*;

import javax.swing.*;

import junit.framework.Assert;
import de.uniwue.ki.gui.mas.model.IGUISeSAmDeclaration;
import de.uniwue.ki.gui.mas.model.XMLIniReaderForGUI;
import de.uniwue.ki.gui.mas.sim.simulation.ThreadedSimulationWithNamePanel;
import de.uniwue.ki.gui.util.SwingMoreUtilities;
import de.uniwue.ki.gui.util.lookAndFeel.LookAndFeelManager;
import de.uniwue.ki.gui.util.lookAndFeel.LookAndFeelManagerAuxiliary;
import de.uniwue.ki.mas.application.SeSAmCmd;
import de.uniwue.ki.mas.basics.feature.IFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.featureList.FeatureClassDeclarationListManager;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.type.*;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.sim.experiment.*;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.xml.reader.XMLReadException;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.VerySimpleFormatter;
import de.uniwue.ki.util.jarclassloader.ClassLoaderExtender;
import de.uniwue.ki.util.jarclassloader.JarClassLoader;

/**
 * Creation date: (10.10.2000 18:53:07)
 * 
 * @author Christoph Oechslein
 *  
 */
public final class SeSAm {

    private File _modelFile;

    private File _iniFile;

    private String _lookAndFeelClassName = LookAndFeelManager
            .getDefaultLookAndFeelManager()
            .getLookAndFeelClassNameFromPreferences();

    private String _skinName = LookAndFeelManager
            .getDefaultLookAndFeelManager()
            .getLookAndFeelSkinNameFromPreferences();

    private ArrayList _commandLineFeatures = new ArrayList();

    private String _runSimulation = null;

    private Dimension _rootWindowDimension = ((Toolkit.getDefaultToolkit()
            .getScreenSize().getWidth() >= 1600) && (Toolkit
            .getDefaultToolkit().getScreenSize().getHeight() >= 1200)) ? new Dimension(
            1024, 800)
            : new Dimension(800, 600);

    private static final String[] _classesForBackgroundLoading = new String[] {
            "de.uniwue.ki.mas.entity.objectclass.agentclass.ActivityAgentClassDeclarationFactory",
            "de.uniwue.ki.mas.entity.objectclass.worldclass.ActivityWorldClassDeclarationFactory",
            "de.uniwue.ki.gui.mas.entity.objectclass.ActivityAgentClassDeclarationPanel",
            "de.uniwue.ki.gui.mas.behaviour.activityGraph.ActivityGraphDeclarationGraphPanel",
            "de.uniwue.ki.gui.mas.behaviour.activityGraph.ActivityGraphDeclarationGraph",
            "de.uniwue.ki.gui.mas.behaviour.activityGraph.ActivityGraphDeclarationGraphEditor",
            "de.uniwue.ki.gui.mas.sim.simulation.SimulationDeclarationPanel"};

    private void loadClassesInBackground(final String[] classNames) {
        Thread backgroundClassLoadingThread = new Thread(new Runnable() {

            public void run() {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException ie) {
                }
                for (int i = 0; i < classNames.length; i++) {
                    try {
                        Class.forName(classNames[i]);
                        Thread.yield();
                        Thread.sleep(10);
                    } catch (ClassNotFoundException exception) {
                        exception.printStackTrace(System.out);
                        System.exit(2);
                    } catch (InterruptedException ie) {
                    }
                }
            }
        }, "Background Class loading");
        backgroundClassLoadingThread.setPriority(Thread.MIN_PRIORITY);
        backgroundClassLoadingThread.start();
    }

    public static void main(String[] args) {
        if (Utils.isJavaVersionOK(1, 4, 0)) {
            extendClasspath();
            System.setProperty("skinlf.themepack", LookAndFeelManagerAuxiliary
                    .getDefaultLandFSkin());
            new SeSAm().start(args);
        } else {
            JOptionPane.showMessageDialog(null,
                    "Need Java Version greater or equal than 1.4.0!");
        }
    }

    private static void extendClasspath() {
        try {
            ClassLoaderExtender.addToClassPath(ClassLoaderExtender
                    .getArchives(new File(LookAndFeelManagerAuxiliary
                            .getLandFDirectory())));
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
    }

    private void parseArguments(java.lang.String[] args) {
        CmdLineParser parser = new CmdLineParser();
        CmdLineParser.Option help = parser.addBooleanOption('h', "help",
                "print usage info");
        CmdLineParser.Option version = parser.addBooleanOption('v', "version",
                "show version number");
        CmdLineParser.Option model = parser.addStringOption('m', "model",
                "load model file");
        CmdLineParser.Option runSimilation = parser.addStringOption('r', "run",
                "run simulation");
        CmdLineParser.Option ini = parser.addStringOption('i', "ini",
                "load ini file");
        CmdLineParser.Option feature = parser.addStringOption('f', "feature",
                "");
        CmdLineParser.Option lookAndFeel = parser.addStringOption('l',
                "lookFeel", "use given look&feel");
        CmdLineParser.Option dimensionOption = new CmdLineParser.Option('s',
                "size", true, "window size") {

            protected Object parseValue(String arg)
                    throws IllegalOptionValueException {
                int index = arg.indexOf('x');
                if (index == -1) { throw new IllegalOptionValueException(this,
                        arg); }
                try {
                    return new java.awt.Dimension(Integer.parseInt(arg
                            .substring(0, index)), Integer.parseInt(arg
                            .substring(index + 1)));
                } catch (NumberFormatException e) {
                    throw new IllegalOptionValueException(this, arg);
                }
            }
        };
        parser.addOption(dimensionOption);
        try {
            parser.parse(args);
        } catch (OptionException e) {
            System.err.println(e.getMessage());
            printUsage();
            System.exit(2);
        }
        parseIniFileOption(parser, ini);
        if (parser.getOptionValue(help) != null) {
            printUsage();
            System.exit(0);
        }
        if (parser.getOptionValue(version) != null) {
            printVersion();
            System.exit(2);
        }
        if (parser.getOptionValue(model) != null) {
            File modelFile = new File((String) parser.getOptionValue(model));
            if (modelFile.canRead()) {
                _modelFile = modelFile;
            } else {
                System.out.println("Could not read model file: " + modelFile);
            }
        }
        if (parser.getOptionValue(runSimilation) != null)
                if (_modelFile != null)
                    _runSimulation = (String) parser
                            .getOptionValue(runSimilation);
                else {
                    System.out
                            .println("Need to specify model file with \"-m\"");
                }
        if (parser.getOptionValue(lookAndFeel) != null) {
            _lookAndFeelClassName = (String) parser.getOptionValue(lookAndFeel);
        }
        if ((parser.getOptionValue(feature) != null)
                || (System.getProperty("FEATURES") != null)) {
            String commandLineStr = (parser.getOptionValue(feature) != null) ? (String) parser
                    .getOptionValue(feature)
                    : System.getProperty("FEATURES");
            if (commandLineStr.indexOf(';') == -1) {
                _commandLineFeatures.add(commandLineStr);
            } else {
                StringTokenizer it = new StringTokenizer(commandLineStr, ";");
                while (it.hasMoreElements()) {
                    _commandLineFeatures.add(it.nextElement());
                }
            }
        }
        if (parser.getOptionValue(dimensionOption) != null) {
            _rootWindowDimension = (Dimension) parser
                    .getOptionValue(dimensionOption);
        }
    }

    private void parseIniFileOption(CmdLineParser parser, Option ini) {
        _iniFile = SeSAmCmd.parseIniFileOption(parser, ini);
    }

    private static void printUsage() {
        printVersion();
        System.out.println("usage: SeSAm [{-h,--help}]" + " [{-v, --version}]"
                + " [{-s,--size} widthxheight]"
                + " [{-f,--feature} featureclass1;featureclass2]"
                + " [{-i,--ini} ini file name]"
                + " [{-m,--model} model_file_name]"
                + " [{-r,--run} simulation]"
                + " [{-l,--lookFeel} look&Feel Class Name]");
    }

    private static void printVersion() {
        System.out.println("SeSAm version: " + version());
    }

    private void start(String args[]) {
        initialize();
        SeSAmSplashScreen aSeSAmSplashScreen = createSplashScreen();
        getUserDefinedDimension();
        parseArguments(args);
        LookAndFeelManager.getDefaultLookAndFeelManager().setLookAndFeel(
                _lookAndFeelClassName, _skinName);
        final SeSAmMainDesktopFrame seSAmMainDesktopFrame = createSeSAmMainDesktopFrame();
        loadCommandLineFeatures();
        loadIniFiles(seSAmMainDesktopFrame);
        
        if (_runSimulation == null) {
            seSAmMainDesktopFrame.show();
            if (_modelFile != null) 
            seSAmMainDesktopFrame.loadModel(_modelFile);
        }
        else
        {
            if (_modelFile != null) {
                try {
                    seSAmMainDesktopFrame.getGUISeSAmDeclaration().openModel(_modelFile);
                } catch (XMLReadException e) {
                    e.printStackTrace();
                } catch (IOException e) {
                    e.printStackTrace();
                }
                
                if (_runSimulation != null) {
                    openSimulationWindow(seSAmMainDesktopFrame);
                }
            }
            openSimulationWindow(seSAmMainDesktopFrame);
        }
            
        

        /* load some classes in background */
        loadClassesInBackground(_classesForBackgroundLoading);
        aSeSAmSplashScreen.dispose();
        Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {

            public void run() {
                IGUISeSAmDeclaration guiSeSAmDeclaration = seSAmMainDesktopFrame
                        .getGUISeSAmDeclaration();
                if (guiSeSAmDeclaration != null) {
                    ISeSAmDeclaration sesamDeclaration = guiSeSAmDeclaration
                            .getSeSAmDeclaration();
                    if (sesamDeclaration != null) {
                        synchronized (sesamDeclaration) {
                            sesamDeclaration.clear();
                        }
                    }
                }
            }
        }, "Shutdown"));
        ToolTipManager.sharedInstance().setDismissDelay(1 * 60 * 1000);
        ToolTipManager.sharedInstance().setInitialDelay(1 * 1000);
        testIsomorphicEquals();
    }

    private void openSimulationWindow(
            final SeSAmMainDesktopFrame seSAmMainDesktopFrame) {
        ISeSAmDeclaration sesamDeclaration = seSAmMainDesktopFrame
                .getGUISeSAmDeclaration().getSeSAmDeclaration();
        Iterator it = sesamDeclaration.getSeSAmModelDeclaration()
                .getSimulationElementsDeclaration()
                .getSimulationDeclarationListAndFactory()
                .getSimulationDeclarationList().iterator();
        while (it.hasNext()) {
            ISimulationDeclaration simulationDeclaration = (ISimulationDeclaration) it
                    .next();
            try {
                IExperiment experiment = ExperimentFactory.createExperiment(
                        sesamDeclaration, simulationDeclaration);
                if (experiment instanceof OneSimulationExperiment) {
                    ThreadedSimulationWithNamePanel simulationPanel = new ThreadedSimulationWithNamePanel();
                    simulationPanel
                            .setGUISeSAmDeclaration(seSAmMainDesktopFrame
                                    .getGUISeSAmDeclaration());
                    simulationPanel
                            .setThreadedSimulation(((OneSimulationExperiment) experiment)
                                    .getThreadedSimulation());
                    JFrame frame = new JFrame("SeSAm simulation run of "+_modelFile.getName());
                    frame.setSize(_rootWindowDimension);
                    frame.setContentPane(simulationPanel);
                    frame.addWindowListener(new WindowAdapter() {
        				public void windowClosing(WindowEvent e) {
        					System.exit(0);
        				};
        			});
                    frame.show();
                }
            } catch (ModelCompilingException e) {
                e.printStackTrace();
                System.exit(0);
            }
        }
    }

    private void testIsomorphicEquals() {
        if (Assert.ASSERTION_ON) {
            Iterator it = TypeManager.getInstance().getAllTypes();
            while (it.hasNext()) {
                IType element = (IType) it.next();
                testIsomorphicEqualsOfType(element);
            }
        }
    }

    private void testIsomorphicEqualsOfType(IType element) {
        if (element instanceof ISpecificType) {
            IFunctionCallInternDeclaration ifcid1 = ((ISpecificType) element)
                    .createDefaultFunctionCallInternDeclaration();
            IFunctionCallInternDeclaration ifcid2 = ((ISpecificType) element)
                    .createDefaultFunctionCallInternDeclaration();
            Assert.assertTrue("Error two instances of "
                    + ifcid1.getSpecificOutputType()
                    + "createDefaultFunctionCallInternDeclaration() differ!",
                    ifcid1.isomorphicEquals(ifcid2));
            Assert
                    .assertEquals(
                            "Error two instances of "
                                    + ifcid1.getSpecificOutputType()
                                    + "createDefaultFunctionCallInternDeclaration() differ in hashcode!",
                            ifcid1.isomorphicHashCode(), ifcid2
                                    .isomorphicHashCode());
        }
        if (element instanceof IGenericType) {
            IFunctionCallInternDeclaration ifcid1 = ((IGenericType) element)
                    .createDefaultSpecificType()
                    .createDefaultFunctionCallInternDeclaration();
            IFunctionCallInternDeclaration ifcid2 = ((IGenericType) element)
                    .createDefaultSpecificType()
                    .createDefaultFunctionCallInternDeclaration();
            Assert.assertTrue("Error two instances of "
                    + ifcid1.getSpecificOutputType()
                    + "createDefaultFunctionCallInternDeclaration() differ!",
                    ifcid1.isomorphicEquals(ifcid2));
            Assert
                    .assertEquals(
                            "Error two instances of "
                                    + ifcid1.getSpecificOutputType()
                                    + "createDefaultFunctionCallInternDeclaration() differ in hashcode!",
                            ifcid1.isomorphicHashCode(), ifcid2
                                    .isomorphicHashCode());
        }
    }

    private void loadIniFiles(final SeSAmMainDesktopFrame seSAmMainDesktopFrame) {
        try {
            if (_iniFile != null) {
                createXMLIniReaderForGUI(SeSAm.class.getClassLoader(),
                        seSAmMainDesktopFrame).parseIniFile(_iniFile);
            }
            createXMLIniReaderForGUI(JarClassLoader.getInstance(),
                    seSAmMainDesktopFrame).parseIniReaders(
                    JarClassLoader.getInstance().getIniFilesForPlugins());
        } catch (Exception exception) {
            seSAmMainDesktopFrame.getGUISeSAmDeclaration().showException(
                    "Error during loading ini file", exception,
                    seSAmMainDesktopFrame);
        }
    }

    private XMLIniReaderForGUI createXMLIniReaderForGUI(
            ClassLoader classLoader,
            final SeSAmMainDesktopFrame seSAmMainDesktopFrame) {
        return new XMLIniReaderForGUI(classLoader, seSAmMainDesktopFrame
                .getGUISeSAmDeclaration(), seSAmMainDesktopFrame,
                seSAmMainDesktopFrame.getSeSAmMainDesktopFrameMenuBar()
                        .getJMenuPlugins());
    }

    private void loadCommandLineFeatures() {
        /*
         * Set Featuresclasses given at the Commandline
         */
        Iterator it = _commandLineFeatures.iterator();
        while (it.hasNext()) {
            String newFeatureClassName = (String) it.next();
            IFeatureClassDeclaration featureDecl = null;
            try {
                Class featureClass = getClass().getClassLoader().loadClass(
                        newFeatureClassName);
                featureDecl = (IFeatureClassDeclaration) featureClass
                        .getMethod("getInstance", new Class[] {}).invoke(null,
                                null);
            } catch (Exception e) {
                System.out.println("Error loading feature "
                        + newFeatureClassName);
                e.printStackTrace();
            }
            FeatureClassDeclarationListManager.getInstance()
                    .getBuildinFeatureClassDeclarationList().add(featureDecl);
        }
    }

    private SeSAmMainDesktopFrame createSeSAmMainDesktopFrame()
            throws HeadlessException {
        SeSAmMainDesktopFrame seSAmMainDesktopFrame = new SeSAmMainDesktopFrame();
        seSAmMainDesktopFrame.pack();
        seSAmMainDesktopFrame.setUserDefinedDimension(
                _rootWindowDimension.width, _rootWindowDimension.height);
        return seSAmMainDesktopFrame;
    }

    private void getUserDefinedDimension() {
        Dimension dim = SeSAmMainDesktopFrame.getWidthAndHeight(
                _rootWindowDimension.width, _rootWindowDimension.height);
        _rootWindowDimension.width = dim.width;
        _rootWindowDimension.height = dim.height;
    }

    private SeSAmSplashScreen createSplashScreen() {
        /* Create the splash screen */
        SeSAmSplashScreen aSeSAmSplashScreen = new SeSAmSplashScreen();
        aSeSAmSplashScreen.pack();
        /* Center splash screen */
        SwingMoreUtilities.centerWindow(aSeSAmSplashScreen);
        aSeSAmSplashScreen.setVisible(true);
        return aSeSAmSplashScreen;
    }

    private void initialize() throws SecurityException {
        // set to english
        Locale.setDefault(Locale.ENGLISH);
        System.setProperty("user.language", "");
        // set to Desktop
        String oldUserHome = System.getProperty("user.home");
        String newUserHome = oldUserHome + "\\desktop";
        if (new File(newUserHome).isDirectory()) {
            System.setProperty("user.home", newUserHome);
        }
        StreamHandler streamHandler = new StreamHandler(System.out,
                new VerySimpleFormatter()) {

            public synchronized void publish(LogRecord record) {
                super.publish(record);
                flush();
            }
        };
        streamHandler.setLevel(Level.ALL);
        Handler[] handlers = Logger.getLogger("").getHandlers();
        for (int i = 0; i < handlers.length; i++) {
            Logger.getLogger("").removeHandler(handlers[i]);
        }
        Logger.getLogger("").addHandler(streamHandler);
        Logger.getLogger("").setLevel(Level.WARNING);
    }

    public static String version() {
        return SeSAmCmd.version();
    }
}