package de.uniwue.ki.gui.mas.application.remote.server;

import java.io.IOException;
import java.text.*;
import java.util.*;

import javax.swing.event.*;

import org.jfree.ui.SortableTableModel;

import de.uniwue.ki.mas.application.remote.ISimulationControlConnection;
import de.uniwue.ki.mas.application.remote.RemoteClientState;
import de.uniwue.ki.mas.application.remote.server.StateCacheSimulationControlConnection;
import de.uniwue.ki.mas.xml.castor.v01.remote.XMLRemoteInfo;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.comparator.ReverseComparator;
import de.uniwue.ki.util.list.DefaultMutableListModel;

/**
 * Creation date: 10.09.2002, 15:25:38
 * @author oechslein
 */
public class SimulationControlConnectionsTableModel extends SortableTableModel {

	private class ValueAsDateStringComparator extends ValueAsFormatStringComparator {
		private ValueAsDateStringComparator(int columnIndex) {
			super(columnIndex);
		}

		protected int compareRealParsedObject(Object o1, Object o2) {
			Date date1 = (Date) o1;
			Date date2 = (Date) o2;

			return date1.compareTo(date2);
		}
		protected Format getFormat() {
			return DateFormat.getDateTimeInstance();
		}
	}

	private abstract class ValueAsFormatStringComparator extends ValueStringComparator {

		private ValueAsFormatStringComparator(int columnIndex) {
			super(columnIndex);
		}

		protected final int compareRealObject(Object o1, Object o2) {
			try {
				Object realObj1 = getFormat().parseObject((String) o1);
				Object realObj2 = getFormat().parseObject((String) o2);

				return compareRealParsedObject(realObj1, realObj2);

			} catch (ParseException e) {
				return 0;
			}
		}

		protected abstract int compareRealParsedObject(Object o1, Object o2);
		protected abstract Format getFormat();
	}

	private class ValueAsNumberStringComparator extends ValueAsFormatStringComparator {
		private ValueAsNumberStringComparator(int columnIndex) {
			super(columnIndex);
		}

		protected int compareRealParsedObject(Object o1, Object o2) {
			Number number1 = (Number) o1;
			Number number2 = (Number) o2;

			return Double.compare(number1.doubleValue(), number2.doubleValue());
		}
		protected Format getFormat() {
			return NumberFormat.getInstance();
		}
	}

	private class ValueAsStringComparator extends ValueStringComparator {
		private ValueAsStringComparator(int columnIndex) {
			super(columnIndex);
		}
		protected int compareRealObject(Object o1, Object o2) {
			String value1 = (String) o1;
			String value2 = (String) o2;

			return Collator.getInstance().compare(value1, value2);
		}
	}
	private abstract class ValueStringComparator implements Comparator {
		private final int _columnIndex;

		private ValueStringComparator(int columnIndex) {
			_columnIndex = columnIndex;
		}
		/**
		 * @param o1 Object
		 * @param o2 Object
		 * @return int
		 * @see java.util.Comparator#compare(Object, Object)
		 */
		public int compare(Object o1, Object o2) {
			StateCacheSimulationControlConnection simulationControlConnection1 =
				(StateCacheSimulationControlConnection) o1;
			StateCacheSimulationControlConnection simulationControlConnection2 =
				(StateCacheSimulationControlConnection) o2;
			try {
				Object value1 = getValueAt(simulationControlConnection1, _columnIndex);
				Object value2 = getValueAt(simulationControlConnection2, _columnIndex);

				if (value1 == null) {
					if (value2 == null) {
						return 0;
					} else {
						return -1;
					}
				} else if (value2 == null) {
					return 1;

				} else {
					return compareRealObject(value1, value2);
				}
			} catch (IOException e) {
				return 0;
			}
		}

		protected abstract int compareRealObject(Object o1, Object o2);
	}

	private static final String[] COLUMN_IDENTIFIERS =
		new String[] {
			"Schedule",
			"Connection",
			"State",
			"Model",
			"Experiment",
			"Simulation",
			"Ticks",
			"ObjectCount",
			"Hours per kticks",
			"StartDate" };

	private final transient ChangeListener _changeListener = new ChangeListener() {
		public void stateChanged(ChangeEvent changeEvent) {
			fireTableDataChanged();
		}
	};

	private final Comparator[] _comparators =
		new Comparator[] {
			null,
			new ValueAsStringComparator(1),
			new ValueAsStringComparator(2),
			new ValueAsStringComparator(3),
			new ValueAsStringComparator(4),
			new ValueAsStringComparator(5),
			new ValueAsNumberStringComparator(6),
			new ValueAsNumberStringComparator(7),
			new ValueAsNumberStringComparator(8),
			new ValueAsDateStringComparator(9)};

	private final DefaultMutableListModel _connections;
	private boolean _isAlreadyInSorting = false;

	private final ListDataListener _listDataListener = new ListDataListener() {

		public synchronized void contentsChanged(ListDataEvent e) {
			updateChangeListener();
			fireTableDataChanged();
		}
		public synchronized void intervalAdded(ListDataEvent e) {
			updateChangeListener();
			fireTableRowsInserted(e.getIndex0(), e.getIndex1());
		}

		public synchronized void intervalRemoved(ListDataEvent e) {
			updateChangeListener();
			fireTableRowsDeleted(e.getIndex0(), e.getIndex1());
		}
	};

	private final Comparator[] _reversedComparators = new Comparator[COLUMN_IDENTIFIERS.length];

	public SimulationControlConnectionsTableModel(DefaultMutableListModel connections) {
		super();

		_connections = connections;

		addAllChangeListener();
		fireTableRowsInserted(0, Math.max(0, _connections.size() - 1));

		_connections.addListDataListener(_listDataListener);
	}

	private void addAllChangeListener() {
		Iterator it = _connections.iterator();
		while (it.hasNext()) {
			StateCacheSimulationControlConnection conn = (StateCacheSimulationControlConnection) it.next();
			conn.addChangeListener(_changeListener);
		}
	}

	/**
	 * @param e TableModelEvent
	 * @see javax.swing.table.AbstractTableModel#fireTableChanged(TableModelEvent)
	 */
	public void fireTableChanged(TableModelEvent e) {
		super.fireTableChanged(e);
		updateSorting();
	}

	public Class getColumnClass(int columnIndex) {
		if (columnIndex == 0)
			return Boolean.class;
		else
			return super.getColumnClass(columnIndex);
	}

	public int getColumnCount() {
		return COLUMN_IDENTIFIERS.length;
	}

	/**
	 * @param columnIndex int
	 * @return String
	 * @see javax.swing.table.TableModel#getColumnName(int)
	 */
	public String getColumnName(int columnIndex) {
		return COLUMN_IDENTIFIERS[columnIndex];
	}

	private Comparator getComparator(int column) {
		return _comparators[column];
	}

	private Comparator getComparator(int column, boolean ascending) {
		if (ascending)
			return getComparator(column);
		else
			return getReverseComparator(column);
	}

	private Comparator getReverseComparator(int column) {
		if (_reversedComparators[column] == null) {
			_reversedComparators[column] = new ReverseComparator(getComparator(column));
		}

		return _reversedComparators[column];
	}

	public int getRowCount() {
		return _connections.size();
	}

	public Object getValueAt(int rowIndex, int columnIndex) {
		if (rowIndex >= _connections.size())
			return null;

		StateCacheSimulationControlConnection simulationControlConnection =
			(StateCacheSimulationControlConnection) _connections.get(rowIndex);

		try {
			return getValueAt(simulationControlConnection, columnIndex);

		} catch (IOException e) {
			e.printStackTrace(System.err);
			_connections.remove(rowIndex);
			return "ERROR: " + e.getMessage();
		}
	}

	protected Object getValueAt(
		StateCacheSimulationControlConnection simulationControlConnection,
		int columnIndex)
		throws IOException {

		if (columnIndex == 0) {
			return Boolean.valueOf(simulationControlConnection.isAutSchedule());

		} else {
			XMLRemoteInfo xmlRemoteInfo = simulationControlConnection.getXMLRemoteInfo();

			if (xmlRemoteInfo == null) {
				if (columnIndex == 1) {
					return simulationControlConnection.getName();
				} else {
					return "ERROR";
				}

			} else if (columnIndex == 1) {
				return xmlRemoteInfo.getMachineName();

			} else if (columnIndex == 2) {
				return RemoteClientState.getStateName(xmlRemoteInfo.getState());

			} else if (columnIndex == 3) {
				return xmlRemoteInfo.getModelFile();

			} else if (columnIndex == 4) {
				return xmlRemoteInfo.getExperimentName();

			} else if (columnIndex == 5) {
				return xmlRemoteInfo.getSimulationName();

			} else if (columnIndex == 6) {
				int ticks = xmlRemoteInfo.getTicks();
				if (ticks == -1)
					return null;
				else
					return NumberFormat.getIntegerInstance().format(ticks);

			} else if (columnIndex == 7) {
				int objectCount = xmlRemoteInfo.getObjectCount();
				if (objectCount == -1)
					return null;
				else
					return NumberFormat.getIntegerInstance().format(Math.max(0, objectCount));

			} else if (columnIndex == 8) {
				if (!xmlRemoteInfo.hasSecondsPerTick())
					return null;
				else if (xmlRemoteInfo.getSecondsPerTick() == -1)
					return null;
				else
					return NumberFormat.getInstance().format(
						xmlRemoteInfo.getSecondsPerTick() * 1000 / (60 * 60 * 1000));

			} else if (columnIndex == 9) {
				Date simStartDate = xmlRemoteInfo.getSimStartDate();
				if (simStartDate == null)
					return null;
				else
					return DateFormat.getDateTimeInstance().format(simStartDate);

			} else {
				return "ERROR";
			}
		}
	}

	private int indexOf(ISimulationControlConnection connection) {
		for (int i = 0; i < _connections.size(); i++) {
			if (Utils.equals(_connections.get(i), connection))
				return i;
		}

		return -1;
	}

	public boolean isCellEditable(int row, int column) {
		return (column == 0);
	}

	/**
	 * @param column int
	 * @return boolean
	 */
	public boolean isSortable(int column) {
		return (column != 0);
	}

	public void refresh(int row) {
		if (row != -1) {
			fireTableRowsUpdated(row, row);
		}
	}

	public void refresh(ISimulationControlConnection simulationControlConnection) {
		refresh(indexOf(simulationControlConnection));
	}

	private void removeAllChangeListener() {
		Iterator it = _connections.iterator();
		while (it.hasNext()) {
			StateCacheSimulationControlConnection conn = (StateCacheSimulationControlConnection) it.next();
			conn.removeChangeListener(_changeListener);
		}
	}

	public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
		if (rowIndex >= _connections.size())
			return;

		StateCacheSimulationControlConnection simulationControlConnection =
			(StateCacheSimulationControlConnection) _connections.get(rowIndex);

		setValueAt(aValue, simulationControlConnection, columnIndex);
	}

	private void setValueAt(
		Object aValue,
		StateCacheSimulationControlConnection simulationControlConnection,
		int columnIndex) {
		if (columnIndex == 0) {
			simulationControlConnection.setAutSchedule(((Boolean) aValue).booleanValue());
		}
	}

	public void sortByColumn(int column, boolean ascending) {
		super.sortByColumn(column, ascending);

		if (isSortable(column)) {
			_connections.sort(getComparator(column, ascending));
		}
	}

	protected void updateChangeListener() {
		removeAllChangeListener();
		addAllChangeListener();
	}

	private void updateSorting() {
		if (!_isAlreadyInSorting) {
			try {
				_isAlreadyInSorting = true;
				int index = getSortingColumn();
				if (index != -1) {
					sortByColumn(index, getAscending());
				}
			} finally {
				_isAlreadyInSorting = false;
			}
		}
	}
}