package de.uniwue.ki.gui.mas.basics.feature.spatial;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.uniwue.ki.gui.util.SwingMoreUtilities;
import de.uniwue.ki.gui.util.actions.AbstractBoundAction;
import de.uniwue.ki.mas.basics.feature.spatial.ISpatialInfo;
import de.uniwue.ki.util.Position2D;
import de.uniwue.ki.util.event.StateChangeMulticaster;

/**
 * @author oechslein
 * 
 * Creation date: 24.06.2003
 */
public class SelectSpatialMapPanelState extends AbstractSpatialMapPanelState implements MouseMotionListener {
	private transient ChangeListener _changeListener;
	private Point _currPoint = null;
	private boolean _inSelectMode = false;
	private boolean _dragged = false;
	private final LinkedList _selectedSpatialInfos = new LinkedList();
	private Point _startPoint = null;
	private Rectangle2D.Double _lastSelection;
	
	public SelectSpatialMapPanelState(){
		super();
		_lastSelection = new Rectangle2D.Double();		
	}
			
	public void addChangeListener(ChangeListener changeListener) {
		_changeListener = StateChangeMulticaster.add(_changeListener, changeListener);
	}

	protected void clearSelectMode() {
		getSpatialMapPanel().repaint();
		_inSelectMode = false;
		_currPoint = null;
		_startPoint = null;
		_dragged = false;
	}

	private void fireStateChanged() {
		if (_changeListener != null) {
			_changeListener.stateChanged(new ChangeEvent(this));
		}
	}

	public Icon getIcon() {
		return SwingMoreUtilities.getJavaSunIcon(24, "general/Select24.gif");
	}

	public String getName() {
		return "Select";
	}

	public Iterator getSelectedSpatialInfos() {
		refreshSelectedSpatialInfos();
		return _selectedSpatialInfos.iterator();
	}

	public void mouseDragged(MouseEvent e) {
		if (_inSelectMode) {
			repaint(_startPoint, _currPoint);
			_currPoint = e.getPoint();
			repaint(_startPoint, _currPoint);
			_dragged = true;
		}
	}

	public void mouseMoved(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
		startSelectMode(e);
	}

	public void mouseReleased(MouseEvent e) {
		if (_inSelectMode && !_dragged){
			Point temp = new Point();
			temp.setLocation(_startPoint.x + 1, _startPoint.y + 1);
			setSelectedSpatialInfos(_startPoint,  temp);
		}
		else if (_inSelectMode) {
			setSelectedSpatialInfos(_startPoint,  e.getPoint());
		}
		clearSelectMode();
	}

	public void paintComponent(Graphics2D g2d) {
		if (_inSelectMode && (_currPoint != null) && (_startPoint != null)) {
			g2d.drawRect(
				Math.min(_currPoint.x, _startPoint.x),
				Math.min(_currPoint.y, _startPoint.y),
				Math.abs(_currPoint.x - _startPoint.x),
				Math.abs(_currPoint.y - _startPoint.y));
		}

		Color oldColor = g2d.getColor();
		g2d.setColor(Color.blue);
		
		refreshSelectedSpatialInfos();
	
		for (Iterator iter = _selectedSpatialInfos.iterator(); iter.hasNext();) {
			ISpatialInfo spInfo = (ISpatialInfo) iter.next();
			g2d.draw(getSpatialMapPanel().calculateRectangle(spInfo.getPosition().x, spInfo.getPosition().y, spInfo.getRadius()));
		}
	
		g2d.setColor(oldColor);
	}

	public void removeChangeListener(ChangeListener changeListener) {
		_changeListener = StateChangeMulticaster.remove(_changeListener, changeListener);
	}

	private void repaint(Point currPoint, Point startPoint) {
		if (_inSelectMode && (startPoint != null) && (currPoint != null)) {
			getSpatialMapPanel().repaint(
				Math.min(startPoint.x, currPoint.x),
				Math.min(startPoint.y, currPoint.y),
				Math.abs(startPoint.x - currPoint.x) + 1,
				Math.abs(startPoint.y - currPoint.y) + 1);
		}
	}
	
	private void setSelectedSpatialInfos(Point startPoint, Point endPoint) {
		Rectangle2D.Double rectangle =
			new Rectangle2D.Double(
				getSpatialMapPanel().convertPanelPositionX(startPoint.x),
				getSpatialMapPanel().convertPanelPositionY(startPoint.y),
				0,
				0);

		rectangle.add(
			getSpatialMapPanel().convertPanelPositionX(endPoint.x),
			getSpatialMapPanel().convertPanelPositionY(endPoint.y));
		
		_lastSelection.setRect(rectangle);
		
		setSelectedSpatialInfos(rectangle);
	}

	private void setSelectedSpatialInfos(Rectangle2D rectangle) {
		_selectedSpatialInfos.clear();

		Iterator it = getSpatialMapPanel().getSpatialMapModel().getAllSpatialInfoList().iterator();
		while (it.hasNext()) {
			ISpatialInfo spatialInfo = (ISpatialInfo)it.next();
			Position2D position = spatialInfo.getPosition();
			double radius = Math.max(5, spatialInfo.getRadius());
			if (rectangle.intersects(position.x - radius, position.y - radius, radius * 2, radius * 2)) {
				_selectedSpatialInfos.add(spatialInfo);
				if (!_dragged) break;
			}
		}

		if (!_selectedSpatialInfos.isEmpty()
			|| ((rectangle.getWidth() > 10) && (rectangle.getHeight() > 10)))
			fireStateChanged();
	}
	
	public void refreshSelectedSpatialInfos(){
		_selectedSpatialInfos.clear();
		
		if (getSpatialMapPanel() != null){
			
			Iterator it = getSpatialMapPanel().getSpatialMapModel().getAllSpatialInfoList().iterator();
			while (it.hasNext()) {
				ISpatialInfo spatialInfo = (ISpatialInfo)it.next();
				Position2D position = spatialInfo.getPosition();
				double radius = Math.max(5, spatialInfo.getRadius());
				if (_lastSelection.intersects(position.x - radius, position.y - radius, radius * 2, radius * 2)) {
					_selectedSpatialInfos.add(spatialInfo);

				}
			}
		}
		
	}

	public void setSpatialMapPanel(JSpatialMapPanel spatialMapPanel) {
		if (getSpatialMapPanel() != null) {
			getSpatialMapPanel().removeMouseMotionListener(this);
			getSpatialMapPanel().unregisterKeyboardAction(KeyStroke.getKeyStroke('A', InputEvent.CTRL_MASK));
			getSpatialMapPanel().unregisterKeyboardAction(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
		}

		super.setSpatialMapPanel(spatialMapPanel);

		if (spatialMapPanel != null) {
			spatialMapPanel.addMouseMotionListener(this);
			spatialMapPanel.registerKeyboardAction(new AbstractBoundAction() {
				public void doit() {
					_selectedSpatialInfos.clear();
					_selectedSpatialInfos.addAll(
						getSpatialMapPanel().getSpatialMapModel().getAllSpatialInfoList());
					fireStateChanged();
				}
			}, KeyStroke.getKeyStroke('A', InputEvent.CTRL_MASK), JComponent.WHEN_IN_FOCUSED_WINDOW);
			spatialMapPanel.registerKeyboardAction(new AbstractBoundAction() {
				public void doit() {
					clearSelectMode();
				}
			}, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_IN_FOCUSED_WINDOW);
		}
	}

	private void startSelectMode(MouseEvent e) {
		_inSelectMode = true;
		_startPoint = e.getPoint();
	}
	
	protected void clearSelection(){
		_lastSelection.setRect(0,0,0,0);
		_selectedSpatialInfos.clear();
	}
	
	protected LinkedList getSelectedSpatialInfosAsList(){
		return _selectedSpatialInfos;
	}

}
