package de.uniwue.ki.gui.mas.refactoring;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;

import junit.framework.TestCase;
import de.uniwue.ki.mas.basics.functionCall.IComposedFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.model.ISeSAmModelDeclaration;
import de.uniwue.ki.mas.model.SeSAmDeclarationFactory;
import de.uniwue.ki.mas.refactoring.AbstractRefactoring;
import de.uniwue.ki.mas.refactoring.RefactoringFactory;
import de.uniwue.ki.mas.sim.experiment.ExperimentFactory;
import de.uniwue.ki.mas.sim.experiment.IExperiment;
import de.uniwue.ki.mas.sim.simulation.ISimulation;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.xml.XMLReader;
import de.uniwue.ki.mas.xml.XMLWriter;
import de.uniwue.ki.mas.xml.reader.XMLReadException;

/**
 * @author Michael Scherer
 *
 *  Einige Methoden zum Laden der Sesam Models, die bei jedem abgeleiteten
 *  Testfall bentigt werden.
 */
public abstract class AbstractModelTestCase extends TestCase {
	private ISeSAmDeclaration seSAmDeclaration;
	private ISeSAmModelDeclaration seSAmModelDeclaration;

	private static final String TESTMODEL_DIRECTORY = "/TestModels/";
	
	protected File getFullPath(String fileName) {
		return (
			new File(
				AbstractModelTestCase.class.getPackage().getName()
				.replace('.', '/') + TESTMODEL_DIRECTORY + fileName
			)
		);
	}
	
	protected void loadTestModel(File file) throws Throwable {
		if (seSAmDeclaration != null) seSAmDeclaration.remove();

		seSAmDeclaration = loadModelFile(file);	
		seSAmModelDeclaration = seSAmDeclaration.getSeSAmModelDeclaration();
	}
	
	protected void runTestModel() throws Throwable {
		createSimulations(seSAmDeclaration, true);
	}
	
	protected ISeSAmDeclaration getSeSAmDeclaration() {
		return seSAmDeclaration;
	}

	protected ISeSAmModelDeclaration getSeSAmModelDeclaration() {
		return seSAmModelDeclaration;
	}

	/**
	 * Ldt ein SeSAm-Modell, das als XML-Datei vorliegt.
	 *  
	 * @see de.uniwue.ki.mas.test.TestXMLTestModels 
	 * @param file  Zu ladende XML-Datei
	 * @return      SeSAmDeclaration mit dem geladenen Modell
	 * @throws XMLReadException
	 * @throws IOException
	 * @throws Throwable
	 */
	private ISeSAmDeclaration loadModelFile(File file)
	throws XMLReadException, IOException, Throwable {

		ISeSAmDeclaration seSAmDeclaration = SeSAmDeclarationFactory.createSeSAmDeclaration(false);

		System.out.println("Load file '" + file.getName() + "'");

		XMLReader.readModelInto(file, seSAmDeclaration, null, true);

		return seSAmDeclaration;
	}

	/**
	 * Speichert das aktuelle SeSAm-Modell in eine Datei.
	 *  
	 * @see de.uniwue.ki.mas.test.TestXMLTestModels 
	 * @param file  Zieldatei der Speicherung
	 * @throws IOException
	 * @throws Throwable
	 */
	protected void saveModelFile(File file)
	throws IOException, Throwable {

		System.out.println("Write file '" + file.getName() + "'");
		XMLWriter.writeModel(file, seSAmDeclaration, null);
	}
	
	private int createSimulations(ISeSAmDeclaration seSAmDeclaration, boolean runSimulations)
	throws Throwable {

		int count = 0;

		Iterator it =
		seSAmDeclaration
		.getSeSAmModelDeclaration()
		.getSimulationElementsDeclaration()
		.getSimulationDeclarationListAndFactory()
		.getSimulationDeclarationList()
		.iterator();

		while (it.hasNext()) {
			ISimulationDeclaration simulationDeclaration = (ISimulationDeclaration)it.next();

			System.out.println("Create Simulation '" + simulationDeclaration.getName() + "'");

			IExperiment exp = ExperimentFactory.createExperiment(seSAmDeclaration, simulationDeclaration);

			if (runSimulations) {
				System.out.println("Simulate Simulation '" + exp.getName() + "'");

				simulateExperiment(exp);
			}

			exp.remove();
			count++;
		}

		return count;
	}	

	private void simulateExperiment(IExperiment exp) throws Throwable {
		long startTime = System.currentTimeMillis();
		exp.startInThread();

		do {
			synchronized (this) {
				try {
					wait(1 * 1000);
				} catch (InterruptedException exception) {
				}
			}

			if (System.currentTimeMillis() - startTime > 1 * 60 * 1000) {
				System.out.print("Terminated experiment: " + exp.getName());
				ISimulation currentSimulation = exp.getCurrentSimulation();
				if (currentSimulation != null)
					System.out.println(
							" at tick " + currentSimulation.getSituation().getTimeDurationInTicks() + ".");
				else
					System.out.println(" curr sim == null!");
				exp.terminateInThread();
				break;
			}

		} while (exp.isRunning());

		if (exp.getException() != null)
			throw exp.getException();
	}

	protected AbstractRefactoring getRefactoring(
			IFunctionCallInternDeclaration refactoringSubject,
			Class classOfRefactoring) {

		// Ermittle alle anwendbaren Refactorings
		Iterator refactorings = 
		RefactoringFactory.getInstance()
		.getApplicableRefactorings(getSeSAmDeclaration(), refactoringSubject)
		.iterator();

		// Suche zu testendes Refactoring
		AbstractRefactoring refactoring = null;
		while (refactorings.hasNext()) {
			AbstractRefactoring currItem = (AbstractRefactoring) refactorings.next();
			if (currItem.getClass() == classOfRefactoring) {
				refactoring = currItem;
				break;
			}
		}
		return refactoring;
	}

	protected IComposedFunctionCallInternDeclaration getAndNode() {
		// Suche UserFunction "Assert"
		IUserFunctionDeclaration assertFunction = 
		(IUserFunctionDeclaration) getSeSAmDeclaration()
		.getSeSAmModelDeclaration()
		.getSeSAmModelBasicsDeclaration()
		.getUserFunctionDeclarationListAndFactory()
		.getUserFunctionDeclarationList()
		.getDeclarationByName("Assert");
		
		// Suche "AND"-Knoten in UserFunction
		IComposedFunctionCallInternDeclaration andNode = 
		(IComposedFunctionCallInternDeclaration) assertFunction
		.getParametricFunctionCallDeclaration()
		.getFunctionCallDeclaration().getFunctionCallInternDeclaration();

		return andNode;
	}
}
