package de.uniwue.ki.gui.mas.sim.simulation;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.PrintWriter;

import javax.swing.*;

import junit.framework.Assert;
import adavid.io.JTextAreaWriter;
import de.uniwue.ki.gui.mas.model.IGUISeSAmDeclaration;
import de.uniwue.ki.gui.mas.model.JPanelWithGUISeSAmDeclaration;
import de.uniwue.ki.gui.util.NamedTreeCellRenderer;
import de.uniwue.ki.gui.util.actions.AbstractBoundAction;
import de.uniwue.ki.gui.util.tree.JTreeTable;
import de.uniwue.ki.mas.sim.logging.*;
import de.uniwue.ki.mas.sim.simulation.ISimulation;
import de.uniwue.ki.mas.sim.simulation.IThreadedSimulation;

public class SimulationLoggingPanel extends JPanelWithGUISeSAmDeclaration {
	private final StreamOutputLogger _streamOutputLogger = new StreamOutputLogger();
	private final SimulationStatisticLogger _simulationStatisticLogger = new SimulationStatisticLogger();
	private final StatisticTreeTableModel _statisticTreeTableModel =
		new StatisticTreeTableModel(_simulationStatisticLogger);

	private ISimulation _simulation;
	private boolean _logging = false;

	private final JTextArea _textArea = new JTextArea();
	private final JTextAreaWriter _textAreaWriter = new JTextAreaWriter(_textArea, 255);

	private final JCheckBox _loggingCheckBox = new JCheckBox("Logging");

	public SimulationLoggingPanel() {
		super();

		initialize();
	}

	private void initialize() {
		_streamOutputLogger.setPrintWriter(new PrintWriter(_textAreaWriter));

		setLayout(new BorderLayout(0, 0));

		JTabbedPane tabbedPane = new JTabbedPane(SwingConstants.BOTTOM, JTabbedPane.SCROLL_TAB_LAYOUT);

		add(tabbedPane, BorderLayout.CENTER);

		//_textArea.setPreferredSize(new Dimension(400, 300));

		tabbedPane.add(
			"Logging",
			new JScrollPane(
				_textArea,
				ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
				ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED));

		JTreeTable treeTable = new JTreeTable(_statisticTreeTableModel);

		treeTable.getTree().setRootVisible(false);
		treeTable.getTree().setCellRenderer(new NamedTreeCellRenderer());
		treeTable.getTree().setShowsRootHandles(true);

		tabbedPane.add("Profiling", new JScrollPane(treeTable));

		JPanel panel = new JPanel();
		panel.setLayout(new FlowLayout(FlowLayout.RIGHT, 5, 3));

		JButton refreshButton = new JButton(new AbstractBoundAction("Refresh") {
			public void doit() {
				_textAreaWriter.flush();
				repaint();
			}
		});

		_loggingCheckBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				updateListeners();
			}
		});

		panel.add(_loggingCheckBox);
		panel.add(refreshButton);

		add(panel, BorderLayout.SOUTH);
	}

	public ISimulation getSimulation() {
		return _simulation;
	}

	public void setSimulation(ISimulation simulation) {
		_simulation = simulation;
	}

	public boolean isLogging() {
		return _logging;
	}

	public void setLogging(boolean logging) {
		Assert.assertNotNull(getSimulation());

		if (isLogging() != logging) {
			getRootPane().getParent().setVisible(logging);
			_logging = logging;
		}

		updateListeners();
	}

	private void updateListeners() {
		removeListeners(_simulationStatisticLogger);
		removeListeners(_streamOutputLogger);

		if (isLogging()) {
			if (_loggingCheckBox.isSelected())
				addListeners(_streamOutputLogger);

			addListeners(_simulationStatisticLogger);
		}
	}

	private void removeListeners(AbstractLogger logger) {
		IThreadedSimulation threadedSimulation = (IThreadedSimulation) getSimulation();

		logger.setSimulation(null);

		threadedSimulation.getSimulationFunctionArgs().removeAgentsListener(logger);
		threadedSimulation.getSimulationFunctionArgs().removeAgentsActionListener(logger);
		threadedSimulation.getSimulationFunctionArgs().removeAgentsActivityListener(logger);
		threadedSimulation.getSimulationFunctionArgs().removeObjectsListener(logger);
		threadedSimulation.getSimulationFunctionArgs().removeObjectsVariableListener(logger);
		threadedSimulation.getSimulationFunctionArgs().removeAnalysisListener(logger);

		threadedSimulation.removeSimulationListener(logger);
		threadedSimulation.removeStartPauseFinishListener(logger);
	}

	private void addListeners(AbstractLogger logger) {
		IThreadedSimulation threadedSimulation = (IThreadedSimulation) getSimulation();

		logger.setSimulation(threadedSimulation);

		threadedSimulation.getSimulationFunctionArgs().addAgentsListener(logger);
		threadedSimulation.getSimulationFunctionArgs().addAgentsActionListener(logger);
		threadedSimulation.getSimulationFunctionArgs().addAgentsActivityListener(logger);
		threadedSimulation.getSimulationFunctionArgs().addObjectsListener(logger);
		threadedSimulation.getSimulationFunctionArgs().addObjectsVariableListener(logger);
		threadedSimulation.getSimulationFunctionArgs().addAnalysisListener(logger);

		threadedSimulation.addSimulationListener(logger);
		threadedSimulation.addStartPauseFinishListener(logger);
	}

	public void setGUISeSAmDeclaration(IGUISeSAmDeclaration guiSeSAmDeclaration) {
		super.setGUISeSAmDeclaration(guiSeSAmDeclaration);

		if (guiSeSAmDeclaration != null)
			_simulationStatisticLogger.setSeSAmDeclaration(guiSeSAmDeclaration.getSeSAmDeclaration());
		else
			_simulationStatisticLogger.setSeSAmDeclaration(null);

		_statisticTreeTableModel.reload();
	}

}