package de.uniwue.ki.gui.mas.sim.test;

import java.text.Collator;
import java.util.*;

import javax.swing.Icon;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;

import org.jfree.ui.SortableTableModel;

import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.sim.experiment.ExperimentDeclarationList;
import de.uniwue.ki.mas.sim.simulation.SimulationDeclarationList;
import de.uniwue.ki.mas.util.declaration.*;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.comparator.ReverseComparator;
import de.uniwue.ki.util.functions.IMapFunction;
import de.uniwue.ki.util.functions.IPredicateFunction;
/**
 * @author oechslein
 * 
 * Creation date: 28.08.2003
 */
class TestRunnerTableModel extends SortableTableModel {

	static class Entry {
		private final ITestableDeclaration _testableDeclaration;
		private boolean _active = true;
		private Throwable _lastRunError;
		private boolean _runFinished;

		public boolean isRunFinished() {
			return _runFinished;
		}

		public void setRunFinished(boolean runFinished) {
			_runFinished = runFinished;
		}

		public Throwable getLastRunError() {
			return _lastRunError;
		}

		public void setLastRunError(Throwable lastRunError) {
			_lastRunError = lastRunError;
		}

		public Entry(ITestableDeclaration testableDeclaration) {
			_testableDeclaration = testableDeclaration;
		}

		public ITestableDeclaration getTestableDeclaration() {
			return _testableDeclaration;
		}

		private void setActive(boolean active) {
			_active = active;
		}

		private boolean isActive() {
			return _active;
		}
	}

	private final ArrayList _entries = new ArrayList();
	private ExperimentDeclarationList _experimentDeclarationList;
	private ISeSAmDeclaration _seSAmDeclaration;
	private SimulationDeclarationList _simulationDeclarationList;
	private Entry _currEntry;

	private final transient ChangeListener _testableDeclarationChangeListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			ITestableDeclaration simulationDeclaration = (ITestableDeclaration)e.getSource();

			removeEntry(simulationDeclaration);
			if (simulationDeclaration.isTestSimulationDeclaration()) {
				addEntry(simulationDeclaration);
			}
			fireTableDataChanged();
		}
	};

	private final transient DeclarationListListener _testableDeclarationListener =
		new DeclarationListListener() {
		public void declarationAdded(DeclarationListEvent event) {
			ITestableDeclaration testableDeclaration = (ITestableDeclaration)event.getDeclaration();

			if (testableDeclaration.isTestSimulationDeclaration()) {
				addEntry(testableDeclaration);
				fireTableDataChanged();
			}

			testableDeclaration.addChangeListener(_testableDeclarationChangeListener);
		}

		public void declarationRemoved(DeclarationListEvent event) {
			ITestableDeclaration testableDeclaration = (ITestableDeclaration)event.getDeclaration();

			removeEntry(testableDeclaration);
			fireTableDataChanged();

			testableDeclaration.removeChangeListener(_testableDeclarationChangeListener);
		}
	};

	public Entry getCurrEntry() {
		return _currEntry;
	}

	public void setCurrEntry(Entry currEntry) {
		_currEntry = currEntry;
	}

	public TestRunnerTableModel() {
	}

	private void addEntry(ITestableDeclaration testableDeclaration) {
		_entries.add(new Entry(testableDeclaration));
	}

	public Iterator getActiveTestDeclarations(final boolean withReset) {
		return CollectionsHelper.map(getActiveEntries(), new IMapFunction() {
			public Object map(Object obj) throws Exception {
				Entry entry = ((Entry)obj);
				if (withReset) {
					entry.setLastRunError(null);
					entry.setRunFinished(false);
				}

				return entry.getTestableDeclaration();
			}
		});
	}

	public Iterator getActiveEntries() {
		return CollectionsHelper.select(_entries.iterator(), new IPredicateFunction() {
			public boolean isElementOk(Object obj) throws Exception {
				return ((Entry)obj).isActive();
			}
		});
	}

	public Class getColumnClass(int columnIndex) {
		if (columnIndex == 0)
			return Boolean.class;
		else if (columnIndex == 1)
			return String.class;
		else
			return Icon.class;
	}

	public int getColumnCount() {
		return 3;
	}

	public String getColumnName(int columnIndex) {
		if (columnIndex == 0)
			return "Active";
		else if (columnIndex == 1)
			return "Simulation/Experiment";
		else
			return "Result";
	}

	Entry getEntry(int rowIndex) {
		if ((rowIndex < _entries.size()) && (rowIndex >= 0))
			return (Entry)_entries.get(rowIndex);
		else
			return null;
	}

	Entry getEntry(ITestableDeclaration testableDeclaration) {
		Iterator it = _entries.iterator();

		while (it.hasNext()) {
			Entry currEntry = (Entry)it.next();
			if (testableDeclaration == currEntry._testableDeclaration)
				return currEntry;
		}
		return null;
	}

	public ExperimentDeclarationList getExperimentDeclarationList() {
		return _experimentDeclarationList;
	}

	public int getRowCount() {
		return _entries.size();
	}

	public ISeSAmDeclaration getSeSAmDeclaration() {
		return _seSAmDeclaration;
	}

	private SimulationDeclarationList getSimulationDeclarationList() {
		return _simulationDeclarationList;
	}

	public Object getValueAt(int rowIndex, int columnIndex) {
		if (columnIndex == 0)
			return Boolean.valueOf(getEntry(rowIndex).isActive());
		else if (columnIndex == 1)
			return getEntry(rowIndex).getTestableDeclaration().getName();
		else
			return getIconFor(getEntry(rowIndex));
	}
	
	
	private Object getIconFor(Entry entry) {
		if (_currEntry == entry) {
			return TestRunner.CURRENT_TEST_ICON;
		} else if (!entry.isRunFinished()) {
			return null;
		} else if (entry.getLastRunError() == null) {
			return TestRunner.POSITIVE_TEST_ICON; 
		} else {
			return TestRunner.NEGATIVE_TEST;
		}
	}

	public boolean isCellEditable(int row, int column) {
		return (column == 0);
	}

	public boolean isSortable(int column) {
		return (column != 0);
	}

	public void remove() {
		setSeSAmDeclaration(null);
		_entries.clear();
	}

	private void removeEntry(ITestableDeclaration testableDeclaration) {
		Iterator entrys = _entries.iterator();
		while (entrys.hasNext()) {
			Entry entry = (Entry)entrys.next();
			if (entry.getTestableDeclaration() == testableDeclaration)
				entrys.remove();
		}
	}

	public void setExperimentDeclarationList(ExperimentDeclarationList newExperimentDeclarationList) {
		ExperimentDeclarationList oldExperimentDeclarationList = getExperimentDeclarationList();

		if (oldExperimentDeclarationList != newExperimentDeclarationList) {
			_experimentDeclarationList = newExperimentDeclarationList;

			updateTestDeclarationListListeners(newExperimentDeclarationList, oldExperimentDeclarationList);
		}
	}

	public void setSeSAmDeclaration(ISeSAmDeclaration seSAmDeclaration) {
		_seSAmDeclaration = seSAmDeclaration;

		_entries.clear();

		if (seSAmDeclaration == null) {
			setSimulationDeclarationList(null);
			setExperimentDeclarationList(null);

		} else {
			setSimulationDeclarationList(
				seSAmDeclaration
					.getSeSAmModelDeclaration()
					.getSimulationElementsDeclaration()
					.getSimulationDeclarationListAndFactory()
					.getSimulationDeclarationList());
			setExperimentDeclarationList(
				seSAmDeclaration
					.getSeSAmModelDeclaration()
					.getSimulationElementsDeclaration()
					.getExperimentDeclarationListAndFactory()
					.getExperimentDeclarationList());
		}
	}
	private void setSimulationDeclarationList(SimulationDeclarationList newSimulationDeclarationList) {
		SimulationDeclarationList oldSimulationDeclarationList = getSimulationDeclarationList();

		if (oldSimulationDeclarationList != newSimulationDeclarationList) {
			_simulationDeclarationList = newSimulationDeclarationList;

			updateTestDeclarationListListeners(newSimulationDeclarationList, oldSimulationDeclarationList);
		}
	}

	public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
		Assert.assertEquals(0, columnIndex);

		getEntry(rowIndex).setActive(((Boolean)aValue).booleanValue());
	}

	public void sortByColumn(int column, boolean ascending) {
		super.sortByColumn(column, ascending);

		if (isSortable(column)) {
			if (ascending)
				Collections.sort(_entries, new ReverseComparator(Collator.getInstance()));
			else
				Collections.sort(_entries);
		}
	}

	private void updateTestDeclarationListListeners(
		IDeclarationList newDeclarationList,
		IDeclarationList oldDeclarationList) {
		if (oldDeclarationList != null) {
			oldDeclarationList.removeDeclarationListListener(_testableDeclarationListener);
		}

		if (newDeclarationList != null) {
			Iterator it = newDeclarationList.iterator();
			while (it.hasNext()) {
				ITestableDeclaration element = (ITestableDeclaration)it.next();
				if (element.isTestSimulationDeclaration())
					addEntry(element);
			}
			newDeclarationList.addDeclarationListListener(_testableDeclarationListener);
		}
	}
}