package de.uniwue.ki.gui.util.lookAndFeel;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.prefs.Preferences;

import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JToggleButton;
import javax.swing.LookAndFeel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import junit.framework.Assert;
import de.uniwue.ki.util.jarclassloader.JarClassLoader;

/**
 * DISCLAIMER - No guarantees. Use at your own risk.
 * A simple manager to manager look and feel of all or registered frames.
 *
 * This class acts as a factory for a menu based and toggle button based
 * control panel for controlling the look and feel of the application.
 * The menu and the button control panel is built based on all the
 * available look and feels. When the user selects the menu or toggle button
 * the look and feel of the frames (and their descendents) is updated to the selected
 * look and feel. Whether all or only registered frames (and their descendents)
 * are updated is controlled by the 'frameUpdatePolicy' property setting.
 * When the frame is registered whether it's look and feel is updated
 * immediately is controlled by the 'updateImmediately' property.
 * The programmer may force an update of all qualified frames
 * using the 'update()' call.
 *
 * @author Sandip Chitale (schitale@selectica.com)
 * @version 1.0 (Mar 2000)
 * @since JDK1.2
 */
public final class LookAndFeelManager {

	/** Update all frames policy constant */
	public static final String ALL_FRAMES = "ALL_FRAMES";

	/** Update only registered frames policy constant */
	public static final String REGISTERED_FRAMES = "REGISTERED_FRAMES";

	public static final String REGISTRATION_NAME_SKINNABLE_LOOKANDFEEL = "SkinLF";

	private static final Insets ZERO_INSETS = new Insets(0, 0, 0, 0);

	private static LookAndFeelManager _lookAndFeelManager;

	private boolean _updateImmediately = false;

	private String _frameUpdatePolicy = REGISTERED_FRAMES;

	private UpdateLookAndFeelAction _updateLookAndFeelAction;

	private JMenu _lookAndFeelMenu;

	private JPanel _lookAndFeelButtonPanel;

	private ArrayList _registeredFrames;

	// ###################
	public static final String PLASTIC_3D_LOOK_AND_FEEL_CLASS_NAME =
		"com.jgoodies.plaf.plastic.Plastic3DLookAndFeel";
	public static final String SKIN_LOOK_AND_FEEL_CLASS_NAME = "com.l2fprod.gui.plaf.skin.SkinLookAndFeel";
	public static final String KUNSTSTOFF_LOOK_AND_FEEL_CLASS_NAME =
		"com.incors.plaf.kunststoff.KunststoffLookAndFeel";
	public static final String METAL_LOOK_AND_FEEL_CLASS_NAME = "javax.swing.plaf.metal.MetalLookAndFeel";

	public static final String THEMEPACK_SEPARATOR = "::";

	private static final String DEFAULT__LOOK_AND_FEEL = PLASTIC_3D_LOOK_AND_FEEL_CLASS_NAME;

	private static final String PREFERENCE_KEY__LANDF__SKIN_NAME = "lookAndFeelSkinName";
	private static final String PREFERENCE_KEY__LANDF__CLASS_NAME = "lookAndFeelClassName";
	//	###################

	private LookAndFeelManagerAuxiliary _auxiliary;

	private final class UpdateLookAndFeelAction extends AbstractAction {

		public UpdateLookAndFeelAction() {
			super("UpdateLookAndFeelAction");
		}

		public void actionPerformed(ActionEvent ae) {
			setLookAndFeel(((AbstractButton)ae.getSource()).getActionCommand());
			JOptionPane.showMessageDialog(
				null,
				"To view the full new Look and Feel\n" + "you have to restart application!");
		}
	}

	private LookAndFeelManager() {

		_updateLookAndFeelAction = new UpdateLookAndFeelAction();
		_registeredFrames = new ArrayList();

		// adding custom LookAndFeels
		_auxiliary = LookAndFeelManagerAuxiliary.getInstance();
	}

	private LookAndFeelEntry[] uniquifyInstalledLnFListAndMixWithRepository(
		UIManager.LookAndFeelInfo[] list) {
		// die Liste, die UIManager.LookAndFeelInfo[] zurckliefert enthlt evtl. doppelte Eintrge:
		// das zuerst aktivierte LnF ist immer doppelt. 
		ArrayList result = new ArrayList();
		ArrayList resultNames = new ArrayList();

		List repList = _auxiliary.getRepository().asList();
		for (Iterator iter = repList.iterator(); iter.hasNext();) {
			LookAndFeelEntry element = (LookAndFeelEntry)iter.next();

			if (!resultNames.contains(element.getName())) {
				addSortedIntoArrayList(result, element);
				resultNames.add(element.getName());
			}
		}

		for (int index = 0; index < list.length; index++) {
			UIManager.LookAndFeelInfo info = list[index];

			if (!resultNames.contains(info.getName())) {
				addSortedIntoArrayList(
					result,
					new LookAndFeelEntry(null, info.getClassName(), info.getName()));
				resultNames.add(info.getName());
			}
		}

		return (LookAndFeelEntry[])result.toArray(new LookAndFeelEntry[1]);
	}

	private void addSortedIntoArrayList(ArrayList list, LookAndFeelEntry entry) {

		int index = 0;
		while (index < list.size()
			&& ((LookAndFeelEntry)list.get(index)).getName().compareTo(entry.getName()) < 0)
			index++;

		list.add(index, entry);

	}

	private void createLookAndFeelControls() {
		_lookAndFeelMenu = new JMenu("Look And Feel");
		_lookAndFeelButtonPanel = new JPanel();
		_lookAndFeelButtonPanel.setLayout(new BoxLayout(_lookAndFeelButtonPanel, BoxLayout.X_AXIS));
		_lookAndFeelButtonPanel.setBorder(BorderFactory.createEtchedBorder());

		LookAndFeelEntry landfInfo[] =
			uniquifyInstalledLnFListAndMixWithRepository(UIManager.getInstalledLookAndFeels());

		ButtonGroup landfMIGroup = new ButtonGroup();
		for (int i = 0; i < landfInfo.length; i++) {

			JMenuItem mi = null;
			if (landfInfo[i].getName().equals(REGISTRATION_NAME_SKINNABLE_LOOKANDFEEL)) {

				mi = new JMenu(REGISTRATION_NAME_SKINNABLE_LOOKANDFEEL);
				String[] strings = LookAndFeelManagerAuxiliary.getThemepackNames();
				for (int index = 0; index < strings.length; index++) {
					String themepack = strings[index];
					String themepackName = themepack.substring(0, themepack.indexOf(".zip"));

					JRadioButtonMenuItem item = new JRadioButtonMenuItem(themepackName);
					item.setActionCommand(landfInfo[i].getClassName() + THEMEPACK_SEPARATOR + themepackName);
					item.addActionListener(_updateLookAndFeelAction);

					landfMIGroup.add(item);
					item.setSelected(isCurrentTheme(themepackName));

					mi.add(item);
				}
			} else {

				mi = new JRadioButtonMenuItem(landfInfo[i].getName(), false);
				mi.setActionCommand(landfInfo[i].getClassName());
				mi.addActionListener(_updateLookAndFeelAction);
				landfMIGroup.add(mi);
				mi.setSelected(
					UIManager.getLookAndFeel().getClass().getName().equals(landfInfo[i].getClassName()));
			}

			_lookAndFeelMenu.add(mi);
			JToggleButton b = new JToggleButton(landfInfo[i].getName());

			b.setToolTipText(landfInfo[i].getName());
			b.setMargin(ZERO_INSETS);
			b.setModel(mi.getModel());
			_lookAndFeelButtonPanel.add(b);
		}
	}

	/**
	 * @return LookAndFeelManager
	 */
	public static LookAndFeelManager getDefaultLookAndFeelManager() {
		if (_lookAndFeelManager == null) {
			_lookAndFeelManager = new LookAndFeelManager();
		}

		return _lookAndFeelManager;
	}

	/**
	 * Get current 'setFrameUpdatePolicy' policy.
	 *
	 * @return a value of type 'String'
	 */
	public String getFrameUpdatePolicy() {
		return _frameUpdatePolicy;
	}

	/**
	 * Get toggle button control panel.
	 *
	 * @return a value of type 'JPanel'
	 */
	public JPanel getLookAndFeelButtonPanel() {
		if (_lookAndFeelButtonPanel == null) {
			createLookAndFeelControls();
		}
		return _lookAndFeelButtonPanel;
	}

	/**
	 * Get the menu.
	 *
	 * @return a value of type 'JMenu'
	 */
	public JMenu getLookAndFeelMenu() {
		if (_lookAndFeelMenu == null) {
			createLookAndFeelControls();
		}

		return _lookAndFeelMenu;
	}

	/**
	 * Get the current 'updateImmidiately' policy.
	 *
	 * @return a value of type 'boolean'
	 */
	public boolean isUpdateImmediately() {
		return _updateImmediately;
	}

	/**
	 * Register a frame. Only registered frames are updated if REGISTERED_FRAMES
	 * frameUpdatePolicy is in effect.
	 *
	 * @param frame a value of type 'Frame'
	 */
	public void registerFrame(Frame frame) {
		Assert.assertNotNull(frame);
		if (!_registeredFrames.contains(frame)) {
			_registeredFrames.add(frame);
			if (isUpdateImmediately()) {
				updateLookAndFeelOfChildWindows(frame);
			}
		}
	}

	/**
	 * Set 'setFrameUpdatePolicy' policy.
	 *
	 * @param frameUpdatePolicy a value of type 'String'
	 */
	public void setFrameUpdatePolicy(String frameUpdatePolicy) {
		if (!frameUpdatePolicy.equals(ALL_FRAMES) && !frameUpdatePolicy.equals(REGISTERED_FRAMES)) {
			throw new IllegalArgumentException(
				"The frameUpdatePolicy must be one of (" + ALL_FRAMES + "," + REGISTERED_FRAMES + ")");
		}
		this._frameUpdatePolicy = frameUpdatePolicy;
		if (frameUpdatePolicy.equals(ALL_FRAMES)) {
			if (isUpdateImmediately()) {
				updateImmediately();
			}
		}

	}

	/**
	 * Set 'updateImmidiately' policy.
	 *
	 * @param updateImmediately a value of type 'boolean'
	 */
	public void setUpdateImmediately(boolean updateImmediately) {
		this._updateImmediately = updateImmediately;
		if (isUpdateImmediately()) {
			updateImmediately();
		}
	}

	/**
	 * Unregister a frame.
	 *
	 * @param frame a value of type 'Frame'
	 */
	public void unregisterFrame(Frame frame) {
		_registeredFrames.remove(frame);
	}

	private void updateLookAndFeel(LookAndFeel landf) {

		try {
			UIManager.setLookAndFeel(landf);
			update();
		} catch (Exception e) {
			System.err.println("Could not switch look and feel to " + landf.getName());
			e.printStackTrace();
		}
	}

	private void setLookAndFeel(String lookAndFeelClassName) {
		int index = lookAndFeelClassName.indexOf(THEMEPACK_SEPARATOR);

		// TODO lade LookAndFeels falls das noch nicht geschehen ist!

		if (index >= 0) {
			Assert.assertEquals(SKIN_LOOK_AND_FEEL_CLASS_NAME, lookAndFeelClassName.substring(0, index));

			String theme = lookAndFeelClassName.substring(index + THEMEPACK_SEPARATOR.length());

			setLookAndFeel(SKIN_LOOK_AND_FEEL_CLASS_NAME, theme);
		} else {

			setLookAndFeel(lookAndFeelClassName, "");
		}

	}

	public void setLookAndFeel(String lookAndFeelClassName, String theme) {
		Assert.assertNotNull(lookAndFeelClassName);

		if (_auxiliary.getRepository().installLookAndFeelByMainClass(lookAndFeelClassName)) {

			System.out.println(
				"LookAndFeelManager.setLookAndFeel(): loaded LnF into memory: " + lookAndFeelClassName);
		}

		if (lookAndFeelClassName.indexOf(SKIN_LOOK_AND_FEEL_CLASS_NAME) >= 0) {

			setSkinLookAndFeel(theme);
			rememberCurrentLookAndFeel(lookAndFeelClassName, theme);

		} else {

			try {
				updateNoSkinLookAndFeel(lookAndFeelClassName);
			} catch (UnsupportedLookAndFeelException e) {
				//wenn das gewnschte L&F nicht untersttzt wird, soll das Standard-L&F gewhlt werden
				// -> fallback  falls schon beim Programmstart ein nicht vorhandenes LnF gewhlt wurde
				setLookAndFeel(METAL_LOOK_AND_FEEL_CLASS_NAME, "");
			}

			rememberCurrentLookAndFeel(lookAndFeelClassName, "");
		}
	}

	private void setSkinLookAndFeel(String theme) {
		try {
			Class skinlfClass = JarClassLoader.getInstance().loadClass(SKIN_LOOK_AND_FEEL_CLASS_NAME);
			LookAndFeel lnf = (LookAndFeel)skinlfClass.newInstance();
			Method setSkinMethod =
				skinlfClass.getMethod(
					"setSkin",
					new Class[] { JarClassLoader.getInstance().loadClass("com.l2fprod.gui.plaf.skin.Skin")});

			Method loadThemePackMethod = skinlfClass.getMethod("loadThemePack", new Class[] { String.class });

			URL resource = LookAndFeelManagerAuxiliary.getInstance().getResource(theme + ".zip");
			if (resource != null) {

				Object themepackObject = loadThemePackMethod.invoke(lnf, new Object[] { resource.getFile()});
				setSkinMethod.invoke(lnf, new Object[] { themepackObject });

			} else {

				setSkinMethod.invoke(
					lnf,
					new Object[] {
						 loadThemePackMethod.invoke(
							lnf,
							new Object[] { LookAndFeelManagerAuxiliary.getDefaultLandFSkin()})
				});
			}

			updateLookAndFeel(lnf);

		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
	}

	private void updateNoSkinLookAndFeel(String lookAndFeelClassName)
		throws UnsupportedLookAndFeelException {

		try {
			Class lfClass = JarClassLoader.getInstance().loadClass(lookAndFeelClassName);
			updateLookAndFeel((LookAndFeel)lfClass.newInstance());

		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			throw new UnsupportedLookAndFeelException(lookAndFeelClassName + " is not supported.");
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
	}

	private void rememberCurrentLookAndFeel(String look, String theme) {
		Preferences preferences = Preferences.userNodeForPackage(this.getClass());
		preferences.put(PREFERENCE_KEY__LANDF__CLASS_NAME, look);
		preferences.put(PREFERENCE_KEY__LANDF__SKIN_NAME, theme);
	}

	public String getLookAndFeelClassNameFromPreferences() {
		Preferences preferences = Preferences.userNodeForPackage(this.getClass());
		return preferences.get(PREFERENCE_KEY__LANDF__CLASS_NAME, DEFAULT__LOOK_AND_FEEL);
	}

	public String getLookAndFeelSkinNameFromPreferences() {
		Preferences preferences = Preferences.userNodeForPackage(this.getClass());
		return preferences.get(
			PREFERENCE_KEY__LANDF__SKIN_NAME,
			LookAndFeelManagerAuxiliary.DEFAULT__SKIN_NAME);
	}

	public boolean isCurrentTheme(String theme) {
		return getLookAndFeelSkinNameFromPreferences().equals(theme);
	}

	public void update() {
		if (getFrameUpdatePolicy().equals(REGISTERED_FRAMES)) {
			Frame[] frames = (Frame[])_registeredFrames.toArray(new Frame[0]);
			if (frames != null) {
				updateLookAndFeelOfChildWindows(frames);
			}

		} else {
			updateLookAndFeelOfChildWindows(Frame.getFrames());
		}
	}

	private void updateImmediately() {
		updateLookAndFeelOfChildWindows(Frame.getFrames());
	}

	private static void updateLookAndFeelOfChildWindows(Window[] windows) {
		for (int i = 0; i < windows.length; i++) {
			updateLookAndFeelOfChildWindows(windows[i]);
		}
	}

	private static void updateLookAndFeelOfChildWindows(Window window) {
		Assert.assertNotNull(window);

		SwingUtilities.updateComponentTreeUI(window);

		updateLookAndFeelOfChildWindows(window.getOwnedWindows());
	}

	//	private static void updateComponentTreeUI(Component c) {
	//		updateComponentTreeUI0(c);
	//		c.invalidate();
	//		c.validate();
	//		c.repaint();
	//	}
	//
	//	private static void updateComponentTreeUI0(Component c) {
	//		if (c instanceof JTree) {
	//			TreeCellRenderer treeCellRenderer = ((JTree) c).getCellRenderer();
	//			if (treeCellRenderer instanceof DefaultTreeCellRenderer) {
	//				updateTreeCellRenderUI((DefaultTreeCellRenderer) treeCellRenderer);
	//			}
	//		}
	//		if (c instanceof JComponent) {
	//			((JComponent) c).updateUI();
	//		}
	//		Component[] children = null;
	//		if (c instanceof JMenu) {
	//			children = ((JMenu) c).getMenuComponents();
	//		} else if (c instanceof Container) {
	//			children = ((Container) c).getComponents();
	//		}
	//		if (children != null) {
	//			for (int i = 0; i < children.length; i++) {
	//				updateComponentTreeUI0(children[i]);
	//			}
	//		}
	//	}
	//
	//	private static void updateTreeCellRenderUI(DefaultTreeCellRenderer cellRenderer) {
	//		cellRenderer.setLeafIcon(cellRenderer.getDefaultLeafIcon());
	//		cellRenderer.setClosedIcon(cellRenderer.getDefaultClosedIcon());
	//		cellRenderer.setOpenIcon(cellRenderer.getDefaultOpenIcon());
	//
	//		cellRenderer.setTextSelectionColor(UIManager.getColor("Tree.selectionForeground"));
	//		cellRenderer.setTextNonSelectionColor(UIManager.getColor("Tree.textForeground"));
	//		cellRenderer.setBackgroundSelectionColor(UIManager.getColor("Tree.selectionBackground"));
	//		cellRenderer.setBackgroundNonSelectionColor(UIManager.getColor("Tree.textBackground"));
	//		cellRenderer.setBorderSelectionColor(UIManager.getColor("Tree.selectionBorderColor"));
	//		//		Object value = UIManager.get("Tree.drawsFocusBorderAroundIcon");
	//		//		cellRenderer.drawsFocusBorderAroundIcon = (value != null && ((Boolean)value).
	//		//						  booleanValue());
	//	}
}