package de.uniwue.ki.gui.util.mdi;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyVetoException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDesktopPane;
import javax.swing.JInternalFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.WindowConstants;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

import de.uniwue.ki.gui.util.SwingMoreUtilities;
import de.uniwue.ki.gui.util.actions.AbstractBoundAction;
/**
 * Menu component that handles the functionality expected of a standard
 * "Windows" menu for MDI applications.
 */
public class MDIWindowMenu extends JMenu {
	private final List _beforeMenuList = new LinkedList();
	private final List _afterMenuList = new LinkedList();
	/* This JCheckBoxMenuItem descendant is used to track the child frame that corresponds
	   to a give menu. */
	private static class ChildMenuItem extends JCheckBoxMenuItem {
		private JInternalFrame _frame;
		private static final ActionListener _action = new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JInternalFrame frame = ((ChildMenuItem) ae.getSource()).getFrame();
				frame.moveToFront();
				try {
					frame.setSelected(true);
					frame.setIcon(false);
				} catch (PropertyVetoException e) {
					e.printStackTrace(System.err);
				}
			}
		};

		public ChildMenuItem(JInternalFrame frame, boolean selected) {
			super(frame.getTitle(), frame.getFrameIcon());
			setSelected(selected);
			_frame = frame;
			AbstractBoundAction.setActionAndFont(this, _action, true);
		}

		public JInternalFrame getFrame() {
			return _frame;
		}
	}

	public MDIWindowMenu(final MDIDesktopPane desktop) {
		setText("Window");
		_beforeMenuList
			.add(
				AbstractBoundAction
				.createJMenuItem(new AbstractBoundAction(
					"Cascade",
					SwingMoreUtilities.getJavaSunIcon(16, "general/TileCascade")) {
			public void doit() {
				cascadeFrames(desktop);
			}
		}, true));
		_beforeMenuList
			.add(
				AbstractBoundAction
				.createJMenuItem(new AbstractBoundAction(
					"Tile Hor&Vert",
					SwingMoreUtilities.getJavaSunIcon(16, "general/TileHorVert")) {
			public void doit() {
				tileHorAndVertFrames(desktop);
			}
		}, true));
		_beforeMenuList
			.add(
				AbstractBoundAction
				.createJMenuItem(new AbstractBoundAction(
					"Tile Horizontal",
					SwingMoreUtilities.getJavaSunIcon(16, "general/TileHor")) {
			public void doit() {
				tileHorFrames(desktop);
			}
		}, true));
		_beforeMenuList
			.add(
				AbstractBoundAction
				.createJMenuItem(new AbstractBoundAction(
					"Tile Vertical",
					SwingMoreUtilities.getJavaSunIcon(16, "general/TileVert")) {
			public void doit() {
				tileVertFrames(desktop);
			}
		}, true));
		_afterMenuList.add(AbstractBoundAction.createJMenuItem(new AbstractBoundAction("Close All") {
			public void doit() {
				closeAllFrames(desktop);
			}
		}, true));
		_afterMenuList.add(AbstractBoundAction.createJMenuItem(new AbstractBoundAction("Minimize All") {
			public void doit() {
				minimizeAllFrames(desktop, true);
			}
		}, true));
		addMenuListener(new MenuListener() {
			public void menuCanceled(MenuEvent e) {
			}
			public void menuDeselected(MenuEvent e) {
				removeAll();
			}
			public void menuSelected(MenuEvent e) {
				buildChildMenus(desktop);
			}
		});
	}
	/* Sets up the children menus depending on the current desktop state */
	private void buildChildMenus(MDIDesktopPane desktop) {
		JInternalFrame[] array = desktop.getAllFrames();
		boolean oneFrameVisible = false;
		for (int i = 0; i < array.length; i++) {
			if (array[i].isVisible()) {
				oneFrameVisible = true;
				break;
			}
		}
		Iterator it = _beforeMenuList.iterator();
		while (it.hasNext()) {
			JMenuItem currJMenuItem = (JMenuItem) it.next();
			currJMenuItem.setEnabled(oneFrameVisible);
			add(currJMenuItem);
		}
		if (oneFrameVisible)
			addSeparator();
		for (int i = 0; i < array.length; i++) {
			if (array[i].isVisible())
				add(new ChildMenuItem(array[i], i == 0));
		}
		addSeparator();
		it = _afterMenuList.iterator();
		while (it.hasNext()) {
			JMenuItem currJMenuItem = (JMenuItem) it.next();
			currJMenuItem.setEnabled(oneFrameVisible);
			add(currJMenuItem);
		}
	}

	public static void closeAllFrames(JDesktopPane desktop) {
		JInternalFrame[] array = desktop.getAllFrames();
		for (int i = 0; i < array.length; i++) {
			if (array[i].isVisible())
				try {
					switch (array[i].getDefaultCloseOperation()) {
						case WindowConstants.DISPOSE_ON_CLOSE :
							array[i].setClosed(true);
							break;

						case WindowConstants.DO_NOTHING_ON_CLOSE :
							break;

						case WindowConstants.EXIT_ON_CLOSE :
							array[i].setClosed(true);
							break;

						case WindowConstants.HIDE_ON_CLOSE :
							array[i].hide();
							array[i].repaint();
							break;

						default :
							break;
					}

				} catch (PropertyVetoException e) {
				}
		}
		desktop.repaint();
	
	}

	public static void minimizeAllFrames(JDesktopPane desktop, boolean iconState) {
		JInternalFrame[] array = desktop.getAllFrames();
		for (int i = 0; i < array.length; i++) {
			if (array[i].isVisible())
				try {
					array[i].setIcon(iconState);
				} catch (PropertyVetoException e) {
				}
		}
	}

	private static double calculateQuality(
		double cols,
		double rows,
		double width,
		double height,
		JInternalFrame[] allFrames) {
		double freeSpaceRatio = (allFrames.length / (cols * rows));
		// lower than 1 (cols * rows always greater than allFrames.length)
		double aspectRatio = 2 / (width / height + height / width);
		// lower than 1 (1.0 means perfect aspect ratio)
		return aspectRatio * aspectRatio * freeSpaceRatio * freeSpaceRatio * freeSpaceRatio;
	}
	/**
	 * Cascade all internal frames
	 */
	private static void cascadeFrames(MDIDesktopPane desktop) {
		final int FRAME_OFFSET = 20;
		desktop.getMDIDesktopManager().setNormalSize();
		JInternalFrame[] allFrames = desktop.getAllFrames();
		java.awt.Insets desktopInsets = desktop.getInsets();
		int desktopWidth = desktop.getWidth() - desktopInsets.left - desktopInsets.right;
		int desktopHeight = desktop.getHeight() - desktopInsets.top - desktopInsets.bottom;
		int frameHeight = desktopHeight - (allFrames.length - 1) * FRAME_OFFSET;
		int frameWidth = desktopWidth - (allFrames.length - 1) * FRAME_OFFSET;
		int x = 0;
		int y = 0;
		for (int i = allFrames.length - 1; i >= 0; i--) {
			allFrames[i].setSize(frameWidth, frameHeight);
			allFrames[i].setLocation(x, y);
			try {
				allFrames[i].setIcon(false);
			} catch (PropertyVetoException exception) {
			}
			x += FRAME_OFFSET;
			y += FRAME_OFFSET;
		}
	}
	/**
	 * Tile all internal frames
	 */
	private static void tileHorAndVertFrames(MDIDesktopPane desktop) {
		JInternalFrame[] allFrames = desktop.getAllFrames();
		double currMaxQuality = Double.MIN_VALUE;
		double currMaxCols = -1;
		double currMaxRows = -1;
		for (double cols = 1; cols <= allFrames.length; cols++) {
			for (double rows = 1; rows <= allFrames.length; rows++) {
				if ((cols * rows) >= allFrames.length) {
					double width = desktop.getWidth() / cols;
					double height = desktop.getHeight() / rows;
					double quality = calculateQuality(cols, rows, width, height, allFrames);
					if (quality > currMaxQuality) {
						currMaxCols = cols;
						currMaxRows = rows;
						currMaxQuality = quality;
					}
				}
			}
		}
		tileInColAndRows(desktop, currMaxCols, currMaxRows);
	}
	private static void tileHorFrames(MDIDesktopPane desktop) {
		tileInColAndRows(desktop, 1, desktop.getAllFrames().length);
	}
	/**
	 * Tile all internal frames
	 */
	private static void tileInColAndRows(MDIDesktopPane desktop, final double cols, final double rows) {
		desktop.getMDIDesktopManager().setNormalSize();
		JInternalFrame[] allFrames = desktop.getAllFrames();
		java.awt.Insets desktopInsets = desktop.getInsets();
		int desktopWidth = desktop.getWidth() - desktopInsets.left - desktopInsets.right;
		int desktopHeight = desktop.getHeight() - desktopInsets.top - desktopInsets.bottom;
		final int width = (int) Math.floor(desktopWidth / cols);
		final int height = (int) Math.floor(desktopHeight / rows);
		int y = 0;
		int i = 0;
		for (int row = 0; row < rows; row++) {
			int x = 0;
			for (int col = 0; col < cols; col++) {
				allFrames[i].setLocation(x, y);
				allFrames[i].setSize(width, height);
				try {
					allFrames[i].setIcon(false);
				} catch (PropertyVetoException exception) {
				}
				x += width + 1;
				i++;
				if (i >= allFrames.length)
					return;
			}
			y += height + 1;
		}
	}
	private static void tileVertFrames(MDIDesktopPane desktop) {
		tileInColAndRows(desktop, desktop.getAllFrames().length, 1);
	}

	public void updateUI() {
		if (_beforeMenuList != null)
			updateMenuItemUI(_beforeMenuList.iterator());
		if (_afterMenuList != null)
			updateMenuItemUI(_afterMenuList.iterator());

		super.updateUI();
	}
	private void updateMenuItemUI(Iterator menuItems) {
		while (menuItems.hasNext()) {
			JMenuItem item = (JMenuItem) menuItems.next();
			item.updateUI();
		}
	}
}