/*
 * @(#)TreeTableModelAdapter.java	1.2 98/10/27
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package de.uniwue.ki.gui.util.tree;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreePath;

import org.jfree.ui.SortableTableModel;

import de.uniwue.ki.util.tree.TreeTableModel;

/**
 * This is a wrapper class takes a TreeTableModel and implements 
 * the table model interface. The implementation is trivial, with 
 * all of the event dispatching support provided by the superclass: 
 * the AbstractTableModel. 
 *
 * @version 1.2 10/27/98
 *
 * @author Philip Milne
 * @author Scott Violet
 */
public class TreeTableModelAdapter extends SortableTableModel {
	private JTree _tree;
	private TreeTableModel _treeTableModel;

	public TreeTableModelAdapter() {
		super();
	}

	public TreeTableModelAdapter(TreeTableModel treeTableModel, JTree tree) {
		super();

		setTree(tree);
		setTreeTableModel(treeTableModel);
	}

	public JTree getTree() {
		return _tree;
	}

	public TreeTableModel getTreeTableModel() {
		return _treeTableModel;
	}

	public void setTreeTableModel(TreeTableModel treeTableModel) {
		_treeTableModel = treeTableModel;
		// Install a TreeModelListener that can update the table when
		// tree changes. We use delayedFireTableDataChanged as we can
		// not be guaranteed the tree will have finished processing
		// the event before us.
		treeTableModel.addTreeModelListener(new TreeModelListener() {
			public void treeNodesChanged(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeNodesInserted(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeNodesRemoved(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeStructureChanged(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}
		});

		fireTableStructureChanged();
	}

	public void setTree(JTree tree) {
		_tree = tree;
		tree.addTreeExpansionListener(new TreeExpansionListener() {
			// Don't use fireTableRowsInserted() here; the selection model
			// would get updated twice. 
			public void treeExpanded(TreeExpansionEvent event) {
				fireTableDataChanged();
			}
			public void treeCollapsed(TreeExpansionEvent event) {
				fireTableDataChanged();
			}
		});

		fireTableStructureChanged();
	}

	// Wrappers, implementing TableModel interface. 

	public int getColumnCount() {
		if (getTreeTableModel() == null)
			return 0;
		else
			return getTreeTableModel().getColumnCount();
	}

	public String getColumnName(int column) {
		return getTreeTableModel().getColumnName(column);
	}

	public Class getColumnClass(int column) {
		return getTreeTableModel().getColumnClass(column);
	}

	public int getRowCount() {
		if (getTree() == null)
			return 0;
		else
			return getTree().getRowCount();
	}

	protected Object nodeForRow(int row) {
		TreePath treePath = getTree().getPathForRow(row);
		return treePath.getLastPathComponent();
	}

	public Object getValueAt(int row, int column) {
		return getTreeTableModel().getValueAt(nodeForRow(row), column);
	}

	public boolean isCellEditable(int row, int column) {
		return getTreeTableModel().isCellEditable(nodeForRow(row), column);
	}

	public void setValueAt(Object value, int row, int column) {
		getTreeTableModel().setValueAt(value, nodeForRow(row), column);
	}

	/**
	 * Invokes fireTableDataChanged after all the pending events have been
	 * processed. SwingUtilities.invokeLater is used to handle this.
	 */
	protected void delayedFireTableDataChanged() {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				fireTableDataChanged();
			}
		});
	}

	public boolean getAscending() {
		if (getTreeTableModel() == null)
			return true;
		else
			return getTreeTableModel().getAscending();
	}

	public int getSortingColumn() {
		if (getTreeTableModel() == null)
			return -1;
		else
			return getTreeTableModel().getSortingColumn();
	}

	public boolean isSortable(int sortable) {
		if (getTreeTableModel() == null)
			return false;
		else
			return getTreeTableModel().isSortable(sortable);
	}

	public void setAscending(boolean ascending) {
		getTreeTableModel().setAscending(ascending);
	}

	public void sortByColumn(int col, boolean ascending) {
		if (getTreeTableModel() != null)
			getTreeTableModel().sortByColumn(col, ascending);
	}
}
