package de.uniwue.ki.mas.application;

import jargs.gnu.CmdLineParser;

import java.io.*;
import java.text.NumberFormat;
import java.util.logging.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.action.debugging.FunctionActionDebuggingDeclarationFactory;
import de.uniwue.ki.mas.basics.functionDeclaration.FunctionDeclarationCompiler;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SeSAmVersion;
import de.uniwue.ki.mas.sim.experiment.*;
import de.uniwue.ki.mas.sim.simulation.*;
import de.uniwue.ki.mas.sim.situation.*;
import de.uniwue.ki.mas.xml.XMLIniFileException;
import de.uniwue.ki.mas.xml.XMLIniReader;
import de.uniwue.ki.mas.xml.reader.XMLReadException;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.VerySimpleFormatter;
import de.uniwue.ki.util.jarclassloader.JarClassLoader;
import de.uniwue.ki.util.math.DescriptiveStatistics;

/**
 * Creation date: (21.02.2002 15:21:24)
 * @author Christoph Oechslein
 */
public class SeSAmCmd {

	private class MyExperimentListener implements ExperimentListener {

		public void currentSimulationChanged(ExperimentCurrentSimulationChangedEvent event) {
			currentSimulationChanged(event.getOldSimulation(), event.getNewSimulation());
		}

		public void currentSimulationChanged(ISimulation oldSimulation, ISimulation newSimulation) {
			if (oldSimulation != null) {
				oldSimulation.removeStartPauseFinishListener(_mySimAndSituationListener);
				oldSimulation.removeSimulationListener(_mySimAndSituationListener);
			}

			_currentSimStartTime = System.currentTimeMillis();
			_lastTimeStep = _currentSimStartTime;

			if (newSimulation != null) {
				System.out.println(
					"\nCurrent Simulation is now '"
						+ newSimulation.getName()
						+ "'.\n"
						+ "One '.' are "
						+ DOT_TICK_STEP
						+ " ticks.");

				newSimulation.addStartPauseFinishListener(_mySimAndSituationListener);
				newSimulation.addSimulationListener(_mySimAndSituationListener);
			}
		}

		public void finished(StartPauseFinishEvent event) {
		}

		public void paused(StartPauseFinishEvent event) {
		}

		public void started(StartPauseFinishEvent event) {
		}
	}

	private class MySimAndSituationListener
		implements StartPauseFinishListener, SimulationListener, SituationListener {

		private ISituation _currSituation = null;

		private double _totalAmountOfObjects = 0;

		private void addSituationListener(ISituation situation) {
			if (!Utils.equals(_currSituation, situation)) {
				Assert.assertNotNull(situation);
				situation.addSituationListener(_mySimAndSituationListener);
				_currSituation = situation;
			}
		}

		public void finished(StartPauseFinishEvent event) {
			Throwable exceptionInExperiment = event.getException();

			if (exceptionInExperiment != null) {
				throw new RuntimeException(exceptionInExperiment);
			}

			int amountOfTicks = ((ISimulation)event.getSource()).getSituation().getTimeDurationInTicks();
			System.out.println(
				"Length of Run "
					+ amountOfTicks
					+ "\n"
					+ "Milli seconds per Tick (overall): "
					+ (System.currentTimeMillis() - _currentSimStartTime) / amountOfTicks);
		}

		public void nextTimeStep(SituationEvent event) {
			_totalAmountOfObjects += event.getSituation().getWorld().getObjectCount(false, false);

			if ((event.getTimeDurationInTicks() % DOT_TICK_STEP) == 0) {
				System.out.print('.');
			}

			if ((event.getTimeDurationInTicks() % INFORMATION_TICK_STEP) == 0) {
				double meanObjectCount = _totalAmountOfObjects / INFORMATION_TICK_STEP;

				System.out.println("\nTime: " + event.getTimeDurationInTicks());

				long currentTime = System.currentTimeMillis();

				System.out.print("Last " + INFORMATION_TICK_STEP + " ticks took");
				appendElapsedTime(currentTime - _lastTimeStep, System.out);

				System.out.println(
					"Mean Object count (of last "
						+ INFORMATION_TICK_STEP
						+ " ticks): "
						+ NumberFormat.getInstance().format(meanObjectCount));

				double msecs = currentTime - _lastTimeStep;
				double msecsPerTick = msecs / INFORMATION_TICK_STEP;
				double msecsPerObject = (meanObjectCount == 0) ? 0 : msecs / meanObjectCount;
				double msecsPerTickAndPerObject =
					(meanObjectCount == 0) ? 0 : msecsPerObject / meanObjectCount;

				System.out.print(NumberFormat.getInstance().format(msecsPerTick) + "msec (per tick) ");
				System.out.print(NumberFormat.getInstance().format(msecsPerObject) + "msec (per object) ");
				System.out.println(
					NumberFormat.getInstance().format(msecsPerTickAndPerObject)
						+ "msec (per tick and object).");

				_lastTimeStep = currentTime;
				_totalAmountOfObjects = 0;
			}
		}

		public void paused(StartPauseFinishEvent event) {
		}

		public void reseted(SimulationEvent event) {
			addSituationListener(event.getSimulation().getSituation());
		}

		public void started(StartPauseFinishEvent event) {
			addSituationListener(((ISimulation)event.getSource()).getSituation());
		}
	}

	private static final int DOT_TICK_STEP = 10;
	private static final int INFORMATION_TICK_STEP = 1000;

	private static void appendAmount(PrintStream writer, String word, long amount, boolean force) {
		if (!force && (amount == 0)) {
		} else if (amount == 1) {
			writer.print(" one ");
			writer.print(word);
		} else {
			writer.print(" ");
			writer.print(amount);
			writer.print(" ");
			writer.print(word);
			writer.print("s");
		}
	}

	public static void appendElapsedTime(long differenceInMilliSeconds, PrintStream writer) {
		long difference = differenceInMilliSeconds / 1000;

		long seconds = difference % 60;
		difference = (difference - seconds) / 60;

		long minutes = difference % 60;
		difference = (difference - minutes) / 60;

		long hours = difference % 24;
		difference = (difference - hours) / 24;

		long days = difference % 7;

		appendAmount(writer, "week", (difference - days) / 7, false);
		appendAmount(writer, "day", days, false);
		appendAmount(writer, "hour", hours, false);
		appendAmount(writer, "min", minutes, false);
		if (differenceInMilliSeconds / 3600 == 0) {
			appendAmount(writer, "sec", seconds, false);
			appendAmount(writer, "msec", differenceInMilliSeconds, true);
		} else {
			appendAmount(writer, "sec", seconds, false);
		}
		writer.println();
	}

	public static void loadIniFiles(ClassLoader classLoader, File iniFile) throws XMLIniFileException {
		if (iniFile != null) {
			new XMLIniReader(classLoader).parseIniFile(iniFile);
		}

		loadPluginIniFiles();
	}

	public static void loadPluginIniFiles() throws XMLIniFileException {
		new XMLIniReader(JarClassLoader.getInstance()).parseIniReaders(
			JarClassLoader.getInstance().getIniFilesForPlugins());
	}

	public static void main(String[] args) {
		new SeSAmCmd().start(args);
	}

	public static File parseIniFileOption(CmdLineParser parser, CmdLineParser.Option ini) {
		if (parser.getOptionValue(ini) != null) {
			String iniFileName = (String)parser.getOptionValue(ini);
			File iniFile = new File(iniFileName);

			if (iniFile.canRead()) {
				return iniFile;

			} else {
				String userHomeDirectoryName = System.getProperty("user.home");

				if (userHomeDirectoryName != null) {
					File userHomeIniFile = new File(userHomeDirectoryName + File.separator + iniFileName);

					if (userHomeIniFile.canRead()) {
						return userHomeIniFile;
					}
				}
			}

			System.out.println("Could not load ini File: " + iniFile);
		}

		return null;
	}

	private static void printUsage(CmdLineParser parser) {
		printVersion();
		System.out.println("usage: SeSAmCmd [options]\noptions:");
		parser.printOptionUsage();
	}

	private static void printVersion() {
		System.out.println("SeSAmCmd version: " + version());
	}

	public static String version() {
		return SeSAmVersion.version();
	}

	private boolean _assert = false;

	private boolean _benchmarking = false;
	private int _benchmarkRunLength = 10;

	protected long _currentSimStartTime;
	private String _experimentName;
	private File _iniFile;
	protected long _lastTimeStep;
	private File _modelFile;

	private final MyExperimentListener _myExperimentListener = new MyExperimentListener();
	protected final MySimAndSituationListener _mySimAndSituationListener = new MySimAndSituationListener();
	private boolean _noCompile = false;
	private boolean _noDebugFunctions = false;
	private boolean _noInline = false;

	protected final SeSAmSimulationRunner _seSAmSimulationRunner;

	public SeSAmCmd() {
		super();

		_seSAmSimulationRunner = new SeSAmSimulationRunner();
	}

	private void initialize() throws SecurityException {
		Assert.setAssertionFlag(_assert);
		FunctionDeclarationCompiler.setDefaultCompileOn(!_noCompile);
		FunctionDeclarationCompiler.setDefaultInlineOn(!_noInline);
		FunctionActionDebuggingDeclarationFactory.setCompileDebugFunctions(!_noDebugFunctions);

		Utils.setEnglishLocale();

		StreamHandler streamHandler = new StreamHandler(System.out, new VerySimpleFormatter());
		Handler[] handlers = Logger.getLogger("").getHandlers();
		for (int i = 0; i < handlers.length; i++) {
			Logger.getLogger("").removeHandler(handlers[i]);
		}

		if (!_benchmarking) {
			Logger.getLogger("").addHandler(streamHandler);
			Logger.getLogger("").setLevel(Level.WARNING);
		} else {
			Logger.getLogger("").setLevel(Level.OFF);
		}

		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			public void run() {
				if (_seSAmSimulationRunner != null)
					_seSAmSimulationRunner.clear();
			}
		}, "Shutdown"));
	}

	private void loadIniFilesAndModelFile() {
		try {
			loadIniFiles(getClass().getClassLoader(), _iniFile);

			_seSAmSimulationRunner.loadModelFile(_modelFile);

		} catch (XMLIniFileException exception) {
			exception.printStackTrace(System.out);
			System.exit(2);

		} catch (IOException exception) {
			exception.printStackTrace(System.out);
			System.exit(2);

		} catch (XMLReadException exception) {
			exception.printStackTrace(System.out);
			System.exit(2);
		}
	}

	private void outputSetting(boolean value, String name) {
		if (value) {
			System.out.print("+");
		} else {
			System.out.print("-");
		}
		System.out.print(name);
		System.out.print(" ");
	}

	private void parseArguments(java.lang.String[] args) {
		CmdLineParser parser = new CmdLineParser();

		CmdLineParser.Option help = parser.addBooleanOption('h', "help", "print usage info");
		CmdLineParser.Option version = parser.addBooleanOption('v', "version", "show version number");
		CmdLineParser.Option model = parser.addStringOption('m', "model", "load model file (required)");
		CmdLineParser.Option ini = parser.addStringOption('i', "ini", "load ini file");
		CmdLineParser.Option experiment = parser.addStringOption('e', "experiment", "simulate experiment");
		CmdLineParser.Option benchmarking =
			parser.addBooleanOption('b', "benchmarking", "ouput benchmarking results");
		CmdLineParser.Option benchmarkingRunLength =
			parser.addIntegerOption((char)0, "benchmarkRunLength", "run simulation n times");
		CmdLineParser.Option noCompile = parser.addBooleanOption((char)0, "noCompile", "set compile flag");
		CmdLineParser.Option noInline = parser.addBooleanOption((char)0, "noInline", "set inline flag");
		CmdLineParser.Option assertOption =
			parser.addBooleanOption((char)0, "assert", "set gloabl assert flag");
		CmdLineParser.Option noDebugFunctions =
			parser.addBooleanOption((char)0, "noDebugFunctions", "set no debug functions flag");

		try {
			parser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage(parser);
			System.exit(2);
		}

		parseIniFileOption(parser, ini);

		if (parser.getOptionValue(help) != null) {
			printUsage(parser);
			System.exit(0);
		}

		if (parser.getOptionValue(version) != null) {
			printVersion();
			System.exit(0);
		}

		_iniFile = parseIniFileOption(parser, ini);

		if (parser.getOptionValue(model) != null) {
			File modelFile = new File((String)parser.getOptionValue(model));

			if (modelFile.canRead()) {
				_modelFile = modelFile;
			} else {
				System.out.println("Could not locate file: '" + modelFile + "'!");
				System.exit(2);
			}

		} else {
			System.out.println("Missing modelFile!");
			printUsage(parser);
			System.exit(2);
		}

		if (parser.getOptionValue(experiment) != null) {
			_experimentName = (String)parser.getOptionValue(experiment);
		}

		if (parser.getOptionValue(benchmarking) != null) {
			_benchmarking = true;
		}

		if (parser.getOptionValue(benchmarkingRunLength) != null) {
			_benchmarkRunLength = ((Number)parser.getOptionValue(benchmarkingRunLength)).intValue();
		}

		if (parser.getOptionValue(noCompile) != null) {
			_noCompile = true;
		}

		if (parser.getOptionValue(noInline) != null) {
			_noInline = true;
		}

		if (parser.getOptionValue(assertOption) != null) {
			_assert = true;
		}

		if (parser.getOptionValue(noDebugFunctions) != null) {
			_noDebugFunctions = true;
		}
	}

	private boolean simulate() {

		System.out.print("Settings: ");
		outputSetting(_assert, "assert");
		outputSetting(!_noCompile, "compile");
		outputSetting(!_noInline, "inline");
		outputSetting(!_noDebugFunctions, "debugFunctions");
		System.out.println();

		if (!_benchmarking) {
			_seSAmSimulationRunner.setExperimentListener(_myExperimentListener);
		}

		_seSAmSimulationRunner.setVerbose(!_benchmarking);

		try {
			int counter = 1;
			double[] runLengths = new double[_benchmarkRunLength];

			do {
				long startTime = System.currentTimeMillis();

				IExperiment newExperiment;
				if (_experimentName != null)
					newExperiment = _seSAmSimulationRunner.simulate(_experimentName);
				else
					newExperiment = _seSAmSimulationRunner.simulate();

				if (newExperiment != null) {
					synchronized (newExperiment) {
						while (!newExperiment.isFinished() && !newExperiment.isFinishedWithErrors()) {
							try {
								newExperiment.wait(100);
								Thread.yield();
							} catch (InterruptedException exception) {
							}
						}
					}
				} else {
					return false;
				}

				long runLength = System.currentTimeMillis() - startTime;
				runLengths[counter - 1] = runLength;

				System.out.print("Run took");
				appendElapsedTime(runLength, System.out);

			}
			while (!_benchmarking || (counter++ < _benchmarkRunLength));

			if (_benchmarking) {
				double mean = DescriptiveStatistics.mean(runLengths);
				double variance = DescriptiveStatistics.sampleVariance(runLengths, mean);
				System.out.println("Run took mean " + mean + "msecs");
				System.out.println("Run took median " + DescriptiveStatistics.median(runLengths) + "msecs");
				System.out.println(
					"Run took std " + DescriptiveStatistics.standardDeviation(variance) + "msecs");
				System.out.println(
					"Run took stdErr "
						+ DescriptiveStatistics.standardError(runLengths.length, variance)
						+ "msecs");
				for (int i = 0; i < runLengths.length; i++) {
					System.out.println(runLengths[i]);
				}
			}

			return true;

		} catch (ModelCompilingException exception) {
			exception.printStackTrace();
			return false;

		} finally {
			_seSAmSimulationRunner.clear();
		}
	}

	public void start(String[] args) {
		parseArguments(args);

		initialize();

		loadIniFilesAndModelFile();

		if (simulate())
			System.exit(0);
		else
			System.exit(2);
	}
}