package de.uniwue.ki.mas.application;
import java.io.*;

import de.uniwue.ki.mas.model.*;
import de.uniwue.ki.mas.sim.experiment.*;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.sim.situation.ISituationDeclaration;
import de.uniwue.ki.mas.xml.XMLReader;
import de.uniwue.ki.mas.xml.reader.XMLReadException;

/**
 * Creation date: (21.02.2002 15:21:24)
 * @author Christoph Oechslein
 */
public class SeSAmSimulationRunner implements Serializable {
	private IExperiment _experiment = null;
	private ExperimentListener _experimentListener;
	private final ISeSAmDeclaration _seSAmDeclaration;
	private boolean _verbose = true;

	public SeSAmSimulationRunner() {
		this(SeSAmDeclarationFactory.createSeSAmDeclaration(false));
	}

	public SeSAmSimulationRunner(ISeSAmDeclaration seSAmDeclaration) {
		_seSAmDeclaration = seSAmDeclaration;
	}

	public synchronized void clear() {
		getSeSAmDeclaration().clear();
	}

	public IExperiment getExperiment() {
		return _experiment;
	}

	public final ExperimentListener getExperimentListener() {
		return _experimentListener;
	}

	public final ISeSAmDeclaration getSeSAmDeclaration() {
		return _seSAmDeclaration;
	}

	public boolean isVerbose() {
		return _verbose;
	}

	public void loadModelFile(File modelFile) throws IOException, XMLReadException {
		XMLReader.readModelInto(modelFile, getSeSAmDeclaration(), null, false);
	}

	public void loadModelStream(Reader reader) throws IOException, XMLReadException {
		XMLReader.readModelInto(reader, getSeSAmDeclaration(), null, false);
	}

	protected void setExperiment(IExperiment experiment) {
		IExperiment oldExperiment = getExperiment();
		if (oldExperiment != experiment) {
			_experiment = experiment;

			if (oldExperiment != null) {
				if (getExperimentListener() != null)
					oldExperiment.removeExperimentListener(getExperimentListener());

				oldExperiment.remove();
			}
		}
	}

	public final void setExperimentListener(ExperimentListener experimentListener) {
		_experimentListener = experimentListener;
	}

	public void setVerbose(boolean verbose) {
		_verbose = verbose;
	}

	public IExperiment simulate() throws ModelCompilingException {
		ISimulationElementsDeclaration simulationElementsDeclaration =
			getSeSAmDeclaration().getSeSAmModelDeclaration().getSimulationElementsDeclaration();

		if (!simulationElementsDeclaration
			.getExperimentDeclarationListAndFactory()
			.getExperimentDeclarationList()
			.isEmpty()) {
			return simulate(
				(IExperimentDeclaration) simulationElementsDeclaration
					.getExperimentDeclarationListAndFactory()
					.getExperimentDeclarationList()
					.get(0));
		} else if (
			!simulationElementsDeclaration
				.getSimulationDeclarationListAndFactory()
				.getSimulationDeclarationList()
				.isEmpty()) {
			if (isVerbose())
				System.out.println("Could not find any experiment!");
			return simulate(
				(ISimulationDeclaration) simulationElementsDeclaration
					.getSimulationDeclarationListAndFactory()
					.getSimulationDeclarationList()
					.get(0));
		} else if (
			!simulationElementsDeclaration
				.getSituationDeclarationListAndFactory()
				.getSituationDeclarationList()
				.isEmpty()) {
			if (isVerbose())
				System.out.println("Could not find any experiment or simulation!");
			return simulate(
				(ISituationDeclaration) simulationElementsDeclaration
					.getSituationDeclarationListAndFactory()
					.getSituationDeclarationList()
					.get(0));
		} else {
			System.out.println("Could not find any situation!");
			return null;
		}
	}

	private IExperiment simulate(IExperimentDeclaration experimentDeclaration)
		throws ModelCompilingException {
		if (isVerbose())
			System.out.println("Simulate experiment '" + experimentDeclaration.getName() + "'.");

		return simulateIntern(
			ExperimentFactory.createExperiment(getSeSAmDeclaration(), experimentDeclaration));
	}

	private IExperiment simulate(ISimulationDeclaration simulationDeclaration)
		throws ModelCompilingException {
		if (isVerbose())
			System.out.println("Simulate simulation '" + simulationDeclaration.getName() + "'.");

		return simulateIntern(
			ExperimentFactory.createExperiment(getSeSAmDeclaration(), simulationDeclaration));
	}

	private IExperiment simulate(ISituationDeclaration situationDeclaration) throws ModelCompilingException {
		if (isVerbose())
			System.out.println("Simulate situation '" + situationDeclaration.getName() + "'.");

		ISimulationDeclaration simulationDeclaration =
			getSeSAmDeclaration()
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getSimulationDeclarationListAndFactory()
				.getSimulationDeclarationFactory()
				.createSimulationDeclaration(situationDeclaration);

		return simulateIntern(
			ExperimentFactory.createExperiment(getSeSAmDeclaration(), simulationDeclaration));
	}

	public IExperiment simulate(String experimentName) throws ModelCompilingException {
		IExperimentDeclaration experimentDeclaration =
			(IExperimentDeclaration) getSeSAmDeclaration()
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getExperimentDeclarationListAndFactory()
				.getExperimentDeclarationList()
				.getDeclarationByName(experimentName);

		if (experimentDeclaration != null) {
			return simulate(experimentDeclaration);
		} else {
			if (isVerbose())
				System.out.println(
					"Could not find experiment with name '" + experimentName + "', trying simulations.");

			ISimulationDeclaration simulationDeclaration =
				(ISimulationDeclaration) getSeSAmDeclaration()
					.getSeSAmModelDeclaration()
					.getSimulationElementsDeclaration()
					.getSimulationDeclarationListAndFactory()
					.getSimulationDeclarationList()
					.getDeclarationByName(experimentName);

			if (simulationDeclaration != null) {
				return simulate(simulationDeclaration);
			} else {
				if (isVerbose())
					System.out.println(
						"Could not find simulation with name '" + experimentName + "', trying situations.");

				ISituationDeclaration situationDeclaration =
					(ISituationDeclaration) getSeSAmDeclaration()
						.getSeSAmModelDeclaration()
						.getSimulationElementsDeclaration()
						.getSituationDeclarationListAndFactory()
						.getSituationDeclarationList()
						.getDeclarationByName(experimentName);

				if (situationDeclaration != null) {
					return simulate(situationDeclaration);
				} else {
					System.out.println("Could not find situation with name '" + experimentName + "'!");
					return null;
				}
			}
		}
	}

	private synchronized IExperiment simulateIntern(IExperiment newExperiment) {
		IExperiment oldExperiment = getExperiment();

		setExperiment(newExperiment);

		if ((oldExperiment != newExperiment) && (getExperimentListener() != null)) {
			if (newExperiment.getCurrentSimulation() != null) {
				getExperimentListener().currentSimulationChanged(
					new ExperimentCurrentSimulationChangedEvent(
						newExperiment,
						null,
						newExperiment.getCurrentSimulation()));
			}

			newExperiment.addExperimentListener(_experimentListener);
		}

		newExperiment.startInThread();

		return newExperiment;
	}
}