package de.uniwue.ki.mas.application.remote;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.Assert;

import org.tanukisoftware.wrapper.WrapperListener;
import org.tanukisoftware.wrapper.WrapperManager;

import de.uniwue.ki.mas.application.SeSAmCmd;
import de.uniwue.ki.mas.application.remote.client.LocalSimulationControl;
import de.uniwue.ki.mas.model.SeSAmVersion;
import de.uniwue.ki.mas.xml.XMLIniFileException;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.file.FileUtilities;

/**
 * Creation date: 10.09.2002, 15:25:38
 * @author oechslein
 */
public class ClientSimulationControl implements WrapperListener {
	private static final Logger _logger = Logger.getLogger(ClientSimulationControl.class.getName());

	private LocalSimulationControl _localSimulationControl;

	/**
	 * The start method is called when the WrapperManager is signalled by the 
	 *	native wrapper code that it can start its application.  This
	 *	method call is expected to return, so a new thread should be launched
	 *	if necessary.
	 * If there are any problems, then an Integer should be returned, set to
	 *	the desired exit code.  If the application should continue,
	 *	return null.
	 * @param args String[]
	 * @return Integer
	 * @see com.silveregg.wrapper.WrapperListener#start(String[])
	 */
	public Integer start(String[] args) {
		try {
			_localSimulationControl = LocalSimulationControl.createLocalSimulationControl(args);
			_localSimulationControl.startServer();

			return null;

		} catch (IOException e) {
			if (_logger.isLoggable(Level.WARNING)) {
				_logger.log(Level.WARNING, "", e);
			}
			return new Integer(2);
		}
	}

	/**
	 * Called when the application is shutting down.
	 * @param exitCode int
	 * @return int
	 * @see com.silveregg.wrapper.WrapperListener#stop(int)
	 */
	public int stop(int exitCode) {
		if (WrapperManager.isDebugEnabled()) {
			System.out.println("SeSAm Client: stop(" + exitCode + ")");
		}

		if (exitCode != 0)
			return exitCode;

		try {
			_localSimulationControl.stopServer();

			File jarFile = _localSimulationControl.getRestartWithNewJarFile();

			if (jarFile != null) {
				FileUtilities.replaceJarFromClassPath(jarFile, "SeSAm.jar");
			}

		} catch (IOException e) {
			e.printStackTrace();
			return 1;
		}

		// Normally an application will be asked to shutdown here.  Standard Java applications do
		//  not have shutdown hooks, so do nothing here.  It will be as if the user hit CTRL-C to
		//  kill the application.
		return 0;
	}

	/**
	* Called whenever the native wrapper code traps a system control signal
	*  against the Java process.  It is up to the callback to take any actions
	*  necessary.  Possible values are: WrapperManager.WRAPPER_CTRL_C_EVENT, 
	*    WRAPPER_CTRL_CLOSE_EVENT, WRAPPER_CTRL_LOGOFF_EVENT, or 
	*    WRAPPER_CTRL_SHUTDOWN_EVENT
	* @param event int
	* @see com.silveregg.wrapper.WrapperListener#controlEvent(int)
	*/
	public void controlEvent(int event) {
		if (event == WrapperManager.WRAPPER_CTRL_LOGOFF_EVENT) {
			_localSimulationControl.userLogedOff();

		} else if (event == WrapperManager.WRAPPER_CTRL_SHUTDOWN_EVENT) {
			_localSimulationControl.shutDown();
		}

		if (WrapperManager.isControlledByNativeWrapper()) {
			if (WrapperManager.isDebugEnabled()) {
				System.out.println("SeSAm Client: controlEvent(" + event + ") Ignored");
			}
			// Ignore the event as the native wrapper will handle it.
		} else {
			if (WrapperManager.isDebugEnabled()) {
				System.out.println("SeSAm Client: controlEvent(" + event + ") Stopping");
			}

			// Not being run under a wrapper, so this isn't an NT service and should always exit.
			//  Handle the event here.
			WrapperManager.stop(0);
			// Will not get here.
		}
	}

	/*---------------------------------------------------------------
	 * Main Method
	 *-------------------------------------------------------------*/
	/**
	 * Used to Wrapper enable a standard Java application.  This main
	 *  expects the first argument to be the class name of the application
	 *  to launch.  All remaining arguments will be wrapped into a new
	 *  argument list and passed to the main method of the specified
	 *  application.
	 * @param args String[]
	 */
	public static void main(String args[]) {
		Assert.setAssertionFlag(false);

		Utils.setEnglishLocale();

		Logger.getLogger("de.uniwue.").setLevel(Level.WARNING);
		Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.FINER);

		try {
			SeSAmCmd.loadPluginIniFiles();
		} catch (XMLIniFileException e) {
			e.printStackTrace();
		}

		// Start the application.  If the JVM was launched from the native
		//  Wrapper then the application will wait for the native Wrapper to
		//  call the application's start method.  Otherwise the start method
		//  will be called immediately.
		WrapperManager.start(new ClientSimulationControl(), args);

		// This thread ends, the WrapperManager will start the application after the Wrapper has
		//  been propperly initialized by calling the start method above.
	}

	public static String version() {
		return SeSAmVersion.version();
	}
}