package de.uniwue.ki.mas.application.remote.client;

import java.io.*;
import java.util.*;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import junit.framework.Assert;

import org.exolab.castor.xml.MarshalException;
import org.exolab.castor.xml.ValidationException;

import de.uniwue.ki.mas.application.SeSAmCmd;
import de.uniwue.ki.mas.application.SeSAmSimulationRunner;
import de.uniwue.ki.mas.application.remote.IRemoteProtocolConstants;
import de.uniwue.ki.mas.application.remote.RemoteClientState;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.sim.experiment.*;
import de.uniwue.ki.mas.sim.simulation.*;
import de.uniwue.ki.mas.sim.situation.ISituation;
import de.uniwue.ki.mas.xml.castor.v01.remote.XMLRemoteTimeSchedule;
import de.uniwue.ki.mas.xml.castor.v01.remote.XMLRemoteTimeSlice;
import de.uniwue.ki.mas.xml.reader.XMLReadException;
import de.uniwue.ki.util.Utils;

class LocalSeSAmSimulationRunner extends SeSAmSimulationRunner {
	private static final String TIME_SCHEDULE_PREF_STRING = "timeSchedule";
	private transient Timer _timer;
	private String _remoteModelFileName = null;

	protected ISimulation _simulation = null;
	protected int _simulationCounter = -1;

	private boolean _disallowed = false;
	private XMLRemoteTimeSchedule _remoteTimeSchedule = null;

	private final ExperimentListener _experimentListener = new MyExperimentListener();

	private final class MyExperimentListener implements ExperimentListener, Serializable {
		public void currentSimulationChanged(ExperimentCurrentSimulationChangedEvent event) {
			ISimulation newSimulation = event.getNewSimulation();
			ISimulation oldSimulation = _simulation;

			if (newSimulation == oldSimulation)
				return;

			if (oldSimulation != null) {
				oldSimulation.removeStartPauseFinishListener(_simulationListener);
			}

			_simulation = newSimulation;

			if (newSimulation != null) {
				newSimulation.addStartPauseFinishListener(_simulationListener);
				_simulationPaused = false;
			}
		}

		public void finished(StartPauseFinishEvent event) {
			if (event.getException() != null) {
				event.getException().printStackTrace(System.err);
			}

			if (getCurrentSimulation() != null) {
				getCurrentSimulation().removeStartPauseFinishListener(_simulationListener);
				_simulation = null;
			}
		}

		public void paused(StartPauseFinishEvent event) {
		}

		public void started(StartPauseFinishEvent event) {
		}
	}

	protected boolean _simulationPaused = false;

	protected final MyStartPauseFinishListener _simulationListener = new MyStartPauseFinishListener();

	private final class MyStartPauseFinishListener implements StartPauseFinishListener, Serializable {

		private long _stopTime = -1;
		private long _sumOfStopTime = 0;

		public void finished(StartPauseFinishEvent event) {

			_stopTime = System.currentTimeMillis();

			ISimulation simulation = (ISimulation) event.getSource();

			System.out.print("Finished simulation: '");
			System.out.print(simulation.getName());
			System.out.print("' (ticks: ");
			if (simulation.getSituation() == null)
				System.out.print("null");
			else
				System.out.print(simulation.getSituation().getTimeDurationInTicks());

			System.out.print(") in");

			if (simulation.getStartDate() != null) {
				PrintStream writer = System.out;
				SeSAmCmd.appendElapsedTime(
					System.currentTimeMillis() - simulation.getStartDate().getTime(),
					writer);
			}

			System.out.println();

			if (event.getException() != null) {
				event.getException().printStackTrace(System.err);
			}

		}

		public void paused(StartPauseFinishEvent event) {
			_stopTime = System.currentTimeMillis();

			_simulationPaused = true;
		}

		public void started(StartPauseFinishEvent event) {
			if (_simulationPaused) {
				if (_stopTime != -1) {
					_sumOfStopTime += (System.currentTimeMillis() - _stopTime);
					_stopTime = -1;
				}

				_simulationPaused = false;

			} else {
				_simulationCounter++;

				ISimulation simulation = (ISimulation) event.getSource();
				System.out.println("Started simulation '" + simulation.getName() + "'.");

				_stopTime = -1;
				_sumOfStopTime = 0;
			}
		}

		public long getTotalTime() {
			synchronized (this) {
				if ((getCurrentSimulation() == null) || (getCurrentSimulation().getStartDate() == null)) {
					return 0;

				} else if (_stopTime != -1) {
					return _stopTime - getCurrentSimulation().getStartDate().getTime() - _sumOfStopTime;

				} else {
					return System.currentTimeMillis()
						- getCurrentSimulation().getStartDate().getTime()
						- _sumOfStopTime;
				}
			}
		}
	}

	public LocalSeSAmSimulationRunner() {
		super();

		_remoteTimeSchedule = readTimeSchedulePrefs();
		if ((_remoteTimeSchedule != null) && _remoteTimeSchedule.getAutomaticSchedule())
			checkTimer();

		setExperimentListener(_experimentListener);
	}

	public final ISimulation getCurrentSimulation() {
		return _simulation;
	}

	public final int getSimulationCounter() {
		return _simulationCounter;
	}

	public IExperiment simulate() throws ModelCompilingException {
		_simulationCounter = -1;

		return super.simulate();
	}

	public IExperiment simulate(String experimentName) throws ModelCompilingException {
		_simulationCounter = -1;

		return super.simulate(experimentName);
	}

	public String getRemoteModelFileName() {
		File modelFile = getSeSAmDeclaration().getFile();
		if (modelFile != null)
			return modelFile.toString();
		else
			return _remoteModelFileName;
	}

	public void loadModelStream(String remoteModelFileName, Reader reader)
		throws IOException, XMLReadException {

		Assert.assertNotNull(remoteModelFileName);
		loadModelStream(reader);
		_remoteModelFileName = remoteModelFileName;
	}

	public boolean setAllowSimulation(boolean allowSimulation) {
		if (isInState(IRemoteProtocolConstants.ERROR_STATE)) {
			return false;

		} else if (allowSimulation) {
			_disallowed = false;
			return unpause(true);

		} else {
			_disallowed = true;
			return pause(true);
		}
	}

	public boolean unpause() {
		return unpause(false);
	}

	private boolean unpause(boolean force) {
		if (force || isInState(IRemoteProtocolConstants.PAUSED_STATE)) {

			ISimulation simulation = getCurrentSimulation();
			if ((simulation != null) && (simulation instanceof IThreadedSimulation)) {
				IThreadedSimulation threadedSimulation = (IThreadedSimulation) simulation;
				if (!threadedSimulation.isRunning()) {
					System.out.println("Unpause Simulation " + simulation.getName());
					threadedSimulation.startInThread();
					return true;
				} else {
					return false;
				}
			}
		}

		return false;
	}

	public boolean pause() {
		return pause(false);
	}

	private boolean pause(boolean force) {
		if (force
			|| isInState(IRemoteProtocolConstants.RUNNING_STATE)
			|| isInState(IRemoteProtocolConstants.INITIALIZED_STATE)) {

			ISimulation simulation = getCurrentSimulation();
			if ((simulation != null) && (simulation instanceof IThreadedSimulation)) {
				IThreadedSimulation threadedSimulation = (IThreadedSimulation) simulation;
				if (threadedSimulation.isRunning()) {
					System.out.println("Pause Simulation " + simulation.getName());
					threadedSimulation.pauseInThread();
					return true;
				} else {
					return false;
				}
			}
		}

		return false;
	}

	public void userLogedOff() {
		if (isInState(IRemoteProtocolConstants.DISALLOW_SIMULATION_STATE))
			setAllowSimulation(true);
	}

	public boolean isInState(RemoteClientState state) {
		return Utils.equals(getState(), state);
	}

	public RemoteClientState getState() {
		if (_disallowed) {
			return IRemoteProtocolConstants.DISALLOW_SIMULATION_STATE;
		} else if (getExperiment() != null) {
			if (getExperiment().isRunning()) {
				return IRemoteProtocolConstants.RUNNING_STATE;
			} else if (getExperiment().isPaused()) {
				return IRemoteProtocolConstants.PAUSED_STATE;
			} else if (getExperiment().isFinished()) {
				return IRemoteProtocolConstants.FINISHED_STATE;
			} else if (getExperiment().isFinishedWithErrors()) {
				return IRemoteProtocolConstants.FINISHED_WITH_ERRORS_STATE;
			} else {
				return IRemoteProtocolConstants.INITIALIZED_STATE;
			}

		} else if (getCurrentSimulation() != null) {
			if (getCurrentSimulation().isRunning()) {
				return IRemoteProtocolConstants.RUNNING_STATE;
			} else if (getCurrentSimulation().isPaused()) {
				return IRemoteProtocolConstants.PAUSED_STATE;
			} else if (getCurrentSimulation().isFinished()) {
				return IRemoteProtocolConstants.FINISHED_STATE;
			} else if (getCurrentSimulation().isFinishedWithErrors()) {
				return IRemoteProtocolConstants.FINISHED_WITH_ERRORS_STATE;
			} else {
				return IRemoteProtocolConstants.INITIALIZED_STATE;
			}

		} else if (getRemoteModelFileName() != null) {
			return IRemoteProtocolConstants.HAS_MODEL_STATE;
		} else {
			return IRemoteProtocolConstants.NO_MODEL_STATE;
		}
	}

	public void setTimeSchedule(XMLRemoteTimeSchedule xmlRemoteTimeSchedule) {
		boolean oldAutomaticSchedule =
			(_remoteTimeSchedule != null) && _remoteTimeSchedule.getAutomaticSchedule();

		_remoteTimeSchedule = xmlRemoteTimeSchedule;

		if (xmlRemoteTimeSchedule.getAutomaticSchedule()) {
			checkTimer();
			checkTimeSchedule();

		} else if (oldAutomaticSchedule) {
			setAllowSimulation(true);
		}

		writeTimeSchedulePrefs();
	}

	private void writeTimeSchedulePrefs() {
		Preferences timeSchedulePrefs = getTimeSchedulePrefs();
		if (_remoteTimeSchedule == null) {
			timeSchedulePrefs.put(TIME_SCHEDULE_PREF_STRING, null);
		} else {
			try {
				StringWriter writer = new StringWriter();
				_remoteTimeSchedule.marshal(writer);
				timeSchedulePrefs.put(TIME_SCHEDULE_PREF_STRING, writer.toString());

			} catch (MarshalException e) {
				e.printStackTrace();
				timeSchedulePrefs.put(TIME_SCHEDULE_PREF_STRING, null);
			} catch (ValidationException e) {
				e.printStackTrace();
				timeSchedulePrefs.put(TIME_SCHEDULE_PREF_STRING, null);
			}
		}
	}

	private Preferences getTimeSchedulePrefs() {
		Preferences timeSchedulePrefs = Preferences.systemNodeForPackage(getClass());
		try {
			timeSchedulePrefs.flush();
			return timeSchedulePrefs;
		} catch (BackingStoreException e) {
			return Preferences.userNodeForPackage(getClass());
		}
	}

	private XMLRemoteTimeSchedule readTimeSchedulePrefs() {
		Preferences timeSchedulePrefs = getTimeSchedulePrefs();
		String xmlRemoteTimeSchedule = timeSchedulePrefs.get(TIME_SCHEDULE_PREF_STRING, null);
		if (xmlRemoteTimeSchedule != null) {
			try {
				return XMLRemoteTimeSchedule.unmarshalXMLRemoteTimeSchedule(
					new StringReader(xmlRemoteTimeSchedule));
			} catch (MarshalException e) {
				e.printStackTrace();
			} catch (ValidationException e) {
				e.printStackTrace();
			}
		}
		return null;
	}

	public XMLRemoteTimeSchedule getTimeSchedule() {
		return _remoteTimeSchedule;
	}

	protected synchronized void checkTimeSchedule() {
		if ((_remoteTimeSchedule != null) && _remoteTimeSchedule.getAutomaticSchedule()) {
			Calendar now = Calendar.getInstance();
			int dowNow = now.get(Calendar.DAY_OF_WEEK);

			XMLRemoteTimeSlice xmlRemoteTimeSlice = getSliceFor(dowNow);

			if (xmlRemoteTimeSlice == null) {
				System.out.println("Time slice of dow " + dowNow + " is missing!");
				setAllowSimulation(true);
				return;
			}

			int timeNow = now.get(Calendar.HOUR_OF_DAY) * 100 + now.get(Calendar.MINUTE);
			setAllowSimulation(
				timeNow < xmlRemoteTimeSlice.getStart() || timeNow >= xmlRemoteTimeSlice.getEnd());
		}
	}

	private XMLRemoteTimeSlice getSliceFor(int dow) {
		for (int i = 0; i < _remoteTimeSchedule.getSliceCount(); i++) {
			XMLRemoteTimeSlice slice = _remoteTimeSchedule.getSlice(i);
			if (slice.getDow() == dow)
				return slice;

		}

		return null;
	}

	public void checkTimer() {
		if (_timer == null) {
			_timer = new Timer(true);
			TimerTask timerTask = new TimerTask() {
				public void run() {
					try {
						checkTimeSchedule();

					} catch (Throwable e) {
						// don't mess up the whole Timer
						e.printStackTrace();
					}
				}
			};
			_timer.scheduleAtFixedRate(timerTask, 0, 1 * 60 * 1000);
		}
	}

	public void stopCurrentRunningExperimentAndSimulations() {
		if ((getExperiment() != null) && (getExperiment().isRunning())) {
			pauseWaitAndTerminate(getExperiment(), 5 * 60, false);
		}

		if ((getCurrentSimulation() != null)
			&& (getCurrentSimulation().isRunning())
			&& (getCurrentSimulation() instanceof IThreadedSimulation)) {

			pauseWaitAndTerminate((IThreadedSimulation) getCurrentSimulation(), 5 * 60, false);
		}
	}

	public void startCurrentPausedSimulation() {

		if ((getExperiment() != null) && (getExperiment().isPaused())) {
			getExperiment().startInThread();

			while (getExperiment().isPaused()) {
				synchronized (getExperiment()) {
					try {
						getExperiment().wait(100);
					} catch (InterruptedException e) {

					}
				}
			}
		}

		if ((getCurrentSimulation() != null)
			&& (getCurrentSimulation().isPaused())
			&& (getCurrentSimulation() instanceof IThreadedSimulation)) {

			IThreadedSimulation sim = (IThreadedSimulation) getCurrentSimulation();

			sim.startInThread();

			while (getCurrentSimulation().isPaused()) {
				synchronized (getCurrentSimulation()) {
					try {
						getCurrentSimulation().wait(100);
					} catch (InterruptedException e) {

					}
				}
			}
		}
	}

	public boolean stopSimulation() {
		if (isInState(IRemoteProtocolConstants.RUNNING_STATE)
			|| isInState(IRemoteProtocolConstants.FINISHED_STATE)
			|| isInState(IRemoteProtocolConstants.FINISHED_WITH_ERRORS_STATE)
			|| isInState(IRemoteProtocolConstants.PAUSED_STATE)) {

			IExperiment experiment = getExperiment();

			if (experiment != null) {
				System.out.println("Stop Experiment " + experiment.getName());

				experiment.removeExperimentListener(_experimentListener);

				pauseWaitAndTerminate(experiment, 10, true);

				experiment.remove();
			}

			ISimulation currentSimulation = getCurrentSimulation();
			if (currentSimulation != null) {
				System.out.println("Stop Simulation " + currentSimulation.getName());

				currentSimulation.removeStartPauseFinishListener(_simulationListener);
				if (currentSimulation instanceof IThreadedSimulation) {
					pauseWaitAndTerminate(((IThreadedSimulation) currentSimulation), 10, true);
				} else {
					waitSomeTimeForNotRunning(currentSimulation, 10);
				}

				currentSimulation.remove();
			}

			setExperiment(null);
			_simulation = null;

		}

		return true;
	}

	private void pauseWaitAndTerminate(
		IStartPauseFinishInThreadAble threadAble,
		int secondsToWait,
		boolean forceTerminate) {

		if (threadAble.isRunning()) {
			threadAble.pauseInThread();
			waitSomeTimeForNotRunning(threadAble, secondsToWait);
		}

		if (forceTerminate || threadAble.isRunning())
			threadAble.terminateInThread();
	}

	private void waitSomeTimeForNotRunning(IStartPauseFinishAble startPauseFinishAble, int secondsToWait) {
		int counter = 0;
		synchronized (startPauseFinishAble) {
			while (startPauseFinishAble.isRunning() && (counter++ < secondsToWait * 10)) {
				try {
					startPauseFinishAble.wait(100);
				} catch (InterruptedException e) {
					//e.printStackTrace();
				}
			}
		}
	}

	public double getSecondsPerTick() {
		ISimulation currentSimulation = getCurrentSimulation();
		if (currentSimulation == null)
			return -1;

		ISituation situation = currentSimulation.getSituation();
		if (situation == null)
			return 0;

		int ticks = situation.getTimeDurationInTicks();

		if (ticks == -1)
			return 0;
		else
			return (double) _simulationListener.getTotalTime() / (double) ticks;
	}

}