package de.uniwue.ki.mas.application.remote.client;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;

import de.uniwue.ki.mas.application.SeSAmCmd;
import de.uniwue.ki.mas.application.remote.IRemoteSimulationConstants;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.collection.CollectionsHelper;

/**
 * Creation date: 10.09.2002, 15:25:38
 * @author oechslein
 */
public class LocalSimulationControl {
	private final class MultiCastServerThread extends Thread implements Runnable {
		private DatagramSocket _socket;

		private MultiCastServerThread() throws SocketException {
			super("Listen for Datagram");

			_socket = new DatagramSocket(IRemoteSimulationConstants.DEFAULT_DATAGRAM_CLIENT_PORT);
		}

		public void run() {
			try {
				while (true) {
					try {
						waitForDatagram();

					} catch (IOException e) {
						if (!_socket.isClosed()) {
							if (LOGGER.isLoggable(Level.WARNING)) {
								LOGGER.warning(e.toString());
								e.printStackTrace(System.err);
							}
						}
					}
				}

			} finally {
				closeConnection();
			}
		}

		private void waitForDatagram() throws IOException {
			byte[] buf = new byte[256];
			DatagramPacket packet = new DatagramPacket(buf, buf.length);

			try {
				_socket.receive(packet);

			} catch (SocketException e) {
				if (_socket.isClosed()) {
					LOGGER.throwing(getClass().getName(), "waitForDatagram", e);
					return;
				} else {
					throw e;
				}
			}

			String received = new String(packet.getData(), packet.getOffset(), packet.getLength());
			if (LOGGER.isLoggable(Level.FINE)) {
				LOGGER.fine(
					"Cient '"
						+ System.identityHashCode(this)
						+ "' received: "
						+ received
						+ "("
						+ System.currentTimeMillis()
						+ ")");
			}

			if (received.startsWith(IRemoteSimulationConstants.DATAGRAM_ID_STRING)) {
				int afterPortIndex =
					received.indexOf(' ', IRemoteSimulationConstants.DATAGRAM_ID_STRING.length() + 1);

				if (afterPortIndex == -1) {

					System.out.println(
						"Wrong (old) server datagram at " + packet.getAddress().getHostAddress());

				} else {
					int serverID = Integer.parseInt(received.substring(afterPortIndex + 1));

					LocalSimulationControlConnection localSimulationControlConnection =
						getAlreadyConnectedServer(serverID);

					if (localSimulationControlConnection != null) {

						if (LOGGER.isLoggable(Level.FINE)) {
							LOGGER.fine("Already connected to " + localSimulationControlConnection.getName());
						}

					} else {
						int serverPort =
							Integer.parseInt(
								received.substring(
									IRemoteSimulationConstants.DATAGRAM_ID_STRING.length() + 1,
									afterPortIndex));

						try {
							enableConnectionToServer(packet.getAddress(), serverPort, serverID);

						} catch (IOException e) {
							if (LOGGER.isLoggable(Level.WARNING)) {
								LOGGER.warning(
									"Connection to "
										+ packet.getAddress().getHostAddress()
										+ ":"
										+ serverPort
										+ " refused.");
								LOGGER.throwing(getClass().getName(), "waitForDatagram", e);
							}
						}
					}
				}

			}
		}

		public void closeConnection() {
			_socket.close();
		}

		private LocalSimulationControlConnection getAlreadyConnectedServer(int serverID) {
			Iterator controlConnections = getClientSimulationControlConnectionList().iterator();
			Iterator it = CollectionsHelper.asLinkedList(controlConnections).iterator();
			while (it.hasNext()) {
				LocalSimulationControlConnection element = (LocalSimulationControlConnection) it.next();
				if (element.getServerID() == serverID) {
					if (element.isConnected()) {
						return element;
					} else {
						element.closeConnection();
					}
				}
			}
			return null;
		}
	}

	public static final Logger LOGGER = Logger.getLogger(LocalSimulationControl.class.getName());

	private MultiCastServerThread _multiCastServerThread;

	private final List _clientSimulationControlConnectionList =
		Collections.synchronizedList(new LinkedList());

	private final LocalSeSAmSimulationRunner _localSeSAmSimulationRunner;

	private File _restartWithNewJarFile;

	public LocalSimulationControl() {
		super();

		_localSeSAmSimulationRunner = readLocalSeSAmSimulationRunner();
		_localSeSAmSimulationRunner.checkTimer();
		_localSeSAmSimulationRunner.startCurrentPausedSimulation();
	}

	public synchronized void startServer() throws IOException {
		if (_multiCastServerThread == null) {
			_multiCastServerThread = new MultiCastServerThread();
		}

		if (!_multiCastServerThread.isAlive())
			_multiCastServerThread.start();
	}

	public synchronized void stopServer() throws IOException {
		saveRunningSimulations();

		getLocalSeSAmSimulationRunner().clear();

		_multiCastServerThread.closeConnection();
	}

	public void writeLocalSeSAmSimulationRunner() {
		writeLocalSeSAmSimulationRunner(getLocalSeSAmSimulationRunner());
	}

	protected void enableConnectionToServer(InetAddress serverAddress, int serverPort, int serverID)
		throws IOException {

		if ((serverPort < 0) || (serverPort > 0xFFFF)) {
			if (LOGGER.isLoggable(Level.WARNING))
				LOGGER.warning("Bad server port: " + serverAddress + ":" + serverPort);
			return;
		}

		if (LOGGER.isLoggable(Level.FINE))
			LOGGER.fine("Found new Server: " + serverAddress + ":" + serverPort);

		Socket clientSocket = new Socket(serverAddress, serverPort);

		try {
			LocalSimulationControlConnection clientSimulationControlConnection =
				new LocalSimulationControlConnection(this, clientSocket, serverID);

			if (!clientSimulationControlConnection.initServer()) {
				clientSocket.close();
			}

		} catch (IOException e) {
			clientSocket.close();
			if (!(e instanceof SocketTimeoutException) && !(clientSocket.isClosed())) {
				throw e;
			} else {
				LOGGER.throwing(getClass().getName(), "enableConnectionToServer", e);
			}
		}
	}

	public LocalSeSAmSimulationRunner getLocalSeSAmSimulationRunner() {
		return _localSeSAmSimulationRunner;
	}

	public List getClientSimulationControlConnectionList() {
		return _clientSimulationControlConnectionList;
	}

	public void userLogedOff() {
		_localSeSAmSimulationRunner.userLogedOff();
	}

	public void shutDown() {
		saveRunningSimulations();
	}

	private void saveRunningSimulations() {
		_localSeSAmSimulationRunner.stopCurrentRunningExperimentAndSimulations();
		writeLocalSeSAmSimulationRunner();
	}

	public ISeSAmDeclaration getSeSAmDeclaration() {
		return getLocalSeSAmSimulationRunner().getSeSAmDeclaration();
	}

	private static LocalSeSAmSimulationRunner readLocalSeSAmSimulationRunner() {
		Preferences prefs = Preferences.userNodeForPackage(LocalSimulationControl.class);

		String fileString = prefs.get("savedTo", null);

		if (fileString != null) {
			File file = new File(fileString);
			if (file.exists()) {
				try {
					file.deleteOnExit();
					ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
					Object obj = ois.readObject();
					ois.close();
					file.delete();
					if (obj instanceof LocalSeSAmSimulationRunner) {
						return (LocalSeSAmSimulationRunner) obj;
					}

				} catch (ClassNotFoundException e) {
					e.printStackTrace();

				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		return new LocalSeSAmSimulationRunner();
	}

	private static void writeLocalSeSAmSimulationRunner(LocalSeSAmSimulationRunner localSeSAmSimulationRunner) {
		Preferences prefs = Preferences.userNodeForPackage(LocalSimulationControl.class);

		try {
			File file = File.createTempFile("sesamClientSaveFile-save", null);

			ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file));
			oos.writeObject(localSeSAmSimulationRunner);
			oos.close();

			prefs.put("savedTo", file.toString());

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void setRestartWithNewJarFile(File restartWithNewJarFile) {
		_restartWithNewJarFile = restartWithNewJarFile;
	}

	public File getRestartWithNewJarFile() {
		return _restartWithNewJarFile;
	}

	public void appendInfo(StringBuffer buff, boolean withHTMLTags) {
		if (withHTMLTags)
			buff.append("<html>");
		buff.append("SeSAm version: ");
		buff.append(SeSAmCmd.version());

		appendNewline(buff, withHTMLTags);
		buff.append("Machine name: ");
		appendMachineName(buff);
		buff.append(", user name: ");
		buff.append(System.getProperty("user.name"));

		appendNewline(buff, withHTMLTags);
		buff.append("Memory: ");
		Utils.appendMemoryInfo(buff);

		appendNewline(buff, withHTMLTags);
		buff.append("Java Version: ");
		buff.append(System.getProperty("java.version"));

		appendNewline(buff, withHTMLTags);
		buff.append("OS Info: ");
		buff.append(System.getProperty("os.name"));
		buff.append(" (");
		buff.append(System.getProperty("os.version"));
		buff.append(", ");
		buff.append(System.getProperty("os.arch"));
		buff.append(")");

		if (withHTMLTags)
			buff.append("</html>");
	}

	private void appendNewline(StringBuffer buff, boolean withHTMLTags) {
		if (withHTMLTags)
			buff.append("<P>");
		else
			buff.append('\n');
	}

	private void appendMachineName(StringBuffer buff) {
		InetAddress localHostAdr;
		try {
			localHostAdr = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			buff.append("<Unknown>");
			return;
		}
		buff.append(localHostAdr.getHostName());
		buff.append(" [");
		buff.append(localHostAdr.getHostAddress());
		buff.append("]");
	}

	public static void main(String[] args) {
		try {
			createLocalSimulationControl(args).startServer();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static LocalSimulationControl createLocalSimulationControl(String[] args) throws IOException {
		if (args.length > 0) {
			if ("-INFO".equalsIgnoreCase(args[0])) {
				Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.INFO);
			} else if ("-WARN".equalsIgnoreCase(args[0])) {
				Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.WARNING);
			} else if ("-WARNING".equalsIgnoreCase(args[0])) {
				Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.WARNING);
			} else if ("-FINE".equalsIgnoreCase(args[0])) {
				Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.FINE);
			} else if ("-FINER".equalsIgnoreCase(args[0])) {
				Logger.getLogger("de.uniwue.ki.mas.application.remote").setLevel(Level.FINER);
			}

			System.out.println("Processed argument '" + args[0] + "'.");
		}

		LocalSimulationControl localSimulationControl = new LocalSimulationControl();

		StringBuffer buff = new StringBuffer();
		localSimulationControl.appendInfo(buff, false);
		System.out.println(buff.toString());

		return localSimulationControl;
	}
}