package de.uniwue.ki.mas.application.remote.client;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.tanukisoftware.wrapper.WrapperManager;

import de.uniwue.ki.mas.application.remote.*;
import de.uniwue.ki.mas.entity.body.IBody;
import de.uniwue.ki.mas.entity.object.world.IWorld;
import de.uniwue.ki.mas.entity.variable.IBodyVariable;
import de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.mas.sim.analysis.CSVAnalysisSaveAllFinished;
import de.uniwue.ki.mas.sim.experiment.IExperiment;
import de.uniwue.ki.mas.sim.experiment.OneSimulationExperiment;
import de.uniwue.ki.mas.sim.simulation.ISimulation;
import de.uniwue.ki.mas.sim.situation.ISituation;
import de.uniwue.ki.mas.xml.castor.v01.remote.*;
import de.uniwue.ki.mas.xml.reader.XMLReadException;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.file.ZipUtilities;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.name.AbstractNamed;

final class LocalSimulationControlConnection extends AbstractNamed implements ISimulationControlConnection {

	private static final Logger _logger = Logger.getLogger(LocalSimulationControlConnection.class.getName());

	private final LocalSimulationParser _parser = new LocalSimulationParser(this);

	private final Socket _clientSocket;
	private final DataOutputStream _dataOutputStream;
	private final DataInputStream _dataInputStream;

	private boolean _shouldQuitClient = false;
	private boolean _shouldCloseConnection = false;

	private final LocalSimulationControl _localSimulationControl;

	private final int _serverID;

	public LocalSimulationControlConnection(
		LocalSimulationControl localSimulationControl,
		Socket clientSocket,
		int serverID)
		throws IOException {

		super();

		_serverID = serverID;

		clientSocket.setKeepAlive(true);
		clientSocket.setSoTimeout(IRemoteSimulationConstants.CLIENT_TIMEOUT_IN_MILLIS);
		clientSocket.setSoLinger(true, IRemoteSimulationConstants.CLIENT_TIMEOUT_IN_MILLIS);

		_localSimulationControl = localSimulationControl;

		_clientSocket = clientSocket;

		_dataInputStream = new DataInputStream(new BufferedInputStream(clientSocket.getInputStream()));
		_dataOutputStream = new DataOutputStream(new BufferedOutputStream(clientSocket.getOutputStream()));

		setShouldQuitClient(false);
		setShouldCloseConnection(false);
	}

	public boolean initServer() throws IOException {
		getLocalSimulationControl().getClientSimulationControlConnectionList().add(this);
		listenInThread();
		return true;
	}

	public LocalSeSAmSimulationRunner getLocalSeSAmSimulationRunner() {
		return getLocalSimulationControl().getLocalSeSAmSimulationRunner();
	}

	private ISituation getSituation() {
		ISimulation simulation = getLocalSeSAmSimulationRunner().getCurrentSimulation();
		if (simulation != null)
			return simulation.getSituation();
		else
			return null;
	}

	private RemoteClientState getState() {
		return getLocalSeSAmSimulationRunner().getState();
	}

	public boolean isInState(RemoteClientState state) {
		return getLocalSeSAmSimulationRunner().isInState(state);
	}

	public boolean loadModelFile(File modelFile) {
		if (modelFile == null
			|| !modelFile.exists()
			|| isInState(IRemoteProtocolConstants.DISALLOW_SIMULATION_STATE)) {
			return false;
		} else {

			try {
				getLocalSeSAmSimulationRunner().loadModelFile(modelFile);
				return true;

			} catch (IOException e) {
				if (_logger.isLoggable(Level.WARNING)) {
					_logger.warning(e.toString());
					e.printStackTrace(System.err);
				}
				return false;

			} catch (XMLReadException e) {
				if (_logger.isLoggable(Level.WARNING)) {
					_logger.warning(e.toString());
					e.printStackTrace(System.err);
				}
				return false;
			}
		}
	}

	public boolean loadModelStream(String remoteModelFileName, Reader reader) {
		if (getState() == IRemoteProtocolConstants.DISALLOW_SIMULATION_STATE)
			return false;

		try {
			getLocalSeSAmSimulationRunner().loadModelStream(remoteModelFileName, reader);
			return true;

		} catch (IOException e) {
			if (_logger.isLoggable(Level.WARNING)) {
				_logger.warning(e.toString());
				e.printStackTrace(System.err);
			}
			return false;
		} catch (XMLReadException e) {
			if (_logger.isLoggable(Level.WARNING)) {
				_logger.warning(e.toString());
				e.printStackTrace(System.err);
			}
			return false;
		}
	}

	public boolean setXMLTimeSchedule(XMLRemoteTimeSchedule xmlRemoteTimeSchedule) {
		getLocalSeSAmSimulationRunner().setTimeSchedule(xmlRemoteTimeSchedule);
		return true;
	}

	public XMLRemoteTimeSchedule getXMLTimeSchedule() {
		return getLocalSeSAmSimulationRunner().getTimeSchedule();
	}

	public boolean pause() {
		return getLocalSeSAmSimulationRunner().pause();
	}

	public boolean setAllowSimulation(boolean allowSimulation) {
		return getLocalSeSAmSimulationRunner().setAllowSimulation(allowSimulation);
	}

	public boolean unpause() {
		return getLocalSeSAmSimulationRunner().unpause();
	}

	private String getInfo() {
		StringBuffer buff = new StringBuffer();
		_localSimulationControl.appendInfo(buff, true);
		return buff.toString();
	}

	private String getMachineName() {
		InetAddress localHostAdr;

		try {
			localHostAdr = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			return null;
		}

		return localHostAdr.getHostName();
	}

	public boolean stopSimulation() {
		return getLocalSeSAmSimulationRunner().stopSimulation();
	}

	public boolean startSimulation(String experimentName) {
		if (getState() == IRemoteProtocolConstants.DISALLOW_SIMULATION_STATE)
			return false;

		try {
			return getLocalSeSAmSimulationRunner().simulate(experimentName) != null;

		} catch (ModelCompilingException e) {
			if (_logger.isLoggable(Level.WARNING)) {
				_logger.warning(e.toString());
				e.printStackTrace(System.err);
			}
			return false;
		}
	}

	private int getTicks() {
		ISituation situation = getSituation();
		if (situation != null)
			return situation.getTimeDurationInTicks();
		else
			return -1;
	}

	private int getObjectCount() {
		ISituation situation = getSituation();
		if ((situation == null) || (situation.getWorld() == null))
			return -1;
		else
			return situation.getWorld().getObjectCount(false, false);
	}

	private String getSimulationName() {
		ISimulation simulation = getLocalSeSAmSimulationRunner().getCurrentSimulation();
		if (simulation != null)
			return simulation.getName();
		else
			return null;
	}

	private Date getSimulationStartDate() {
		ISimulation simulation = getLocalSeSAmSimulationRunner().getCurrentSimulation();
		if (simulation != null)
			return simulation.getStartDate();
		else
			return null;
	}

	private String getExperimentName() {
		IExperiment experiment = getLocalSeSAmSimulationRunner().getExperiment();

		if ((experiment != null) && (!(experiment instanceof OneSimulationExperiment)))
			return experiment.getName();
		else
			return null;
	}

	private void listenInThread() {
		new Thread(_parser, "Connection to Server: " + _clientSocket).start();
	}

	public boolean shouldCloseConnection() {
		return getShouldQuitClient() || getShouldCloseConnection();
	}

	public synchronized void closeConnectionImmediately() {
		if (_clientSocket != null)
			try {
				_dataInputStream.close();
				_dataOutputStream.close();
				_clientSocket.close();

			} catch (IOException e) {
				printException(e);
			}

		getLocalSimulationControl().getClientSimulationControlConnectionList().remove(this);
	}

	public void closeConnection() {
		setShouldCloseConnection(true);
	}

	public void quit() {
		setShouldQuitClient(true);
	}

	public void restart(File jarFile) {
		setShouldQuitClient(true);
		getLocalSimulationControl().setRestartWithNewJarFile(jarFile);
		WrapperManager.restart();
	}

	public XMLRemoteInfo getXMLRemoteInfo() {

		XMLRemoteInfo xmlRemoteInfo = new XMLRemoteInfo();

		xmlRemoteInfo.setState(getState().getId());
		xmlRemoteInfo.setModelFile(getLocalSeSAmSimulationRunner().getRemoteModelFileName());

		xmlRemoteInfo.setMachineName(getMachineName());
		xmlRemoteInfo.setExperimentName(getExperimentName());
		xmlRemoteInfo.setSimulationName(getSimulationName());
		xmlRemoteInfo.setTicks(getTicks());
		xmlRemoteInfo.setObjectCount(getObjectCount());
		xmlRemoteInfo.setSimStartDate(getSimulationStartDate());
		xmlRemoteInfo.setSecondsPerTick(getSecondsPerTick());

		return xmlRemoteInfo;
	}

	private double getSecondsPerTick() {
		return getLocalSeSAmSimulationRunner().getSecondsPerTick();
	}

	public XMLDetailedRemoteInfo getXMLDetailedRemoteInfo() {
		XMLDetailedRemoteInfo xmlDetailedRemoteInfo = new XMLDetailedRemoteInfo();

		xmlDetailedRemoteInfo.setDetailedInfo(getInfo());

		ISimulation simulation = getLocalSeSAmSimulationRunner().getCurrentSimulation();

		if (simulation != null) {
			IWorld world = simulation.getSituation().getWorld();
			IBody worldBody = world.getBody();

			for (int i = 0; i < worldBody.getBodyVariablesLength(); i++) {
				IBodyVariable bodyVariable = worldBody.getBodyVariableAt(i);
				IVariableClassDeclaration variableClassDeclaration =
					worldBody
						.getBodyVariableInstanceDeclarationFor(bodyVariable)
						.getBodyVariableClassDeclaration()
						.getVariableClassDeclaration();

				String varName = variableClassDeclaration.getName();
				Object value;
				try {
					value = bodyVariable.getValue();

				} catch (SimulationRuntimeException e) {
					value = e.getMessage();
					printException(e);
				}

				XMLRemoteVariableInfo xmlRemoteVariableInfo = new XMLRemoteVariableInfo();
				xmlRemoteVariableInfo.setName(varName);

				StringBuffer buff = new StringBuffer();
				variableClassDeclaration.getSpecificType().appendNameForArgumentObject(buff, value, 0, 30);
				xmlRemoteVariableInfo.setValue(buff.toString());

				xmlDetailedRemoteInfo.addVariableInfo(xmlRemoteVariableInfo);
			}
		}
		return xmlDetailedRemoteInfo;
	}

	public Socket getClientSocket() {
		return _clientSocket;
	}

	public String getName() {
		if (getClientSocket() != null) {
			return "Connection to "
				+ getClientSocket().getInetAddress().getHostAddress()
				+ ":"
				+ getClientSocket().getPort()
				+ " id "
				+ getServerID();
		} else {
			return "Closed connection";
		}
	}

	public boolean isConnected() {
		if (!getClientSocket().isConnected()
			|| getClientSocket().isInputShutdown()
			|| getClientSocket().isOutputShutdown()
			|| getShouldCloseConnection()
			|| getShouldQuitClient())
			return false;
		else
			return true;
	}

	private static void printException(Exception e) {
		if (_logger.isLoggable(Level.WARNING)) {
			_logger.warning(e.toString());
			e.printStackTrace(System.err);
		}
	}

	public synchronized void sendAnswer(RemoteAnswer answer) throws IOException {
		answer.writeObject(_dataOutputStream);
		_dataOutputStream.flush();
	}

	public synchronized RemoteCommand receiveCommand() throws IOException {
		return RemoteCommand.readObject(_dataInputStream);
	}

	public void setShouldQuitClient(boolean shouldQuitClient) {
		_shouldQuitClient = shouldQuitClient;
	}

	public boolean getShouldQuitClient() {
		return _shouldQuitClient;
	}

	public void setShouldCloseConnection(boolean shouldCloseConnection) {
		_shouldCloseConnection = shouldCloseConnection;
	}

	public boolean getShouldCloseConnection() {
		return _shouldCloseConnection;
	}

	protected final List _oldAnalysisList = new LinkedList();

	public File getAnalysisAsZipFile() throws IOException {
		synchronized (_oldAnalysisList) {
			Iterator it = CSVAnalysisSaveAllFinished.getAllFinishedFiles();
			if (CollectionsHelper.isEmpty(it)) {
				return null;

			} else {
				_oldAnalysisList.clear();

				File zipFile = File.createTempFile("analysis_", ".zip");
				zipFile.deleteOnExit();

				Iterator selectedFiles = CollectionsHelper.select(it, new IPredicateFunction() {
					public boolean isElementOk(Object obj) throws Exception {
						File file = (File) obj;
						if (file.canRead()) {
							_oldAnalysisList.add(file);
							return true;
						} else {
							return false;
						}
					}
				});

				if (CollectionsHelper.isEmpty(selectedFiles)) {
					return null;
				}

				ZipUtilities.zip(
					zipFile,
					selectedFiles,
					getLocalSeSAmSimulationRunner().getSeSAmDeclaration().getFile());

				return zipFile;
			}
		}
	}

	public void removeAnalysisFromLastZipFile() {
		synchronized (_oldAnalysisList) {
			Iterator it = _oldAnalysisList.iterator();

			while (it.hasNext()) {
				File file = (File) it.next();
				file.deleteOnExit();
				file.delete();
				CSVAnalysisSaveAllFinished.removeFinishedFile(file);
				it.remove();
			}
		}
		CSVAnalysisSaveAllFinished.cleanCreatedDirectories();
	}

	public void removeAllUnfinishedFiles() {
		Iterator it = CSVAnalysisSaveAllFinished.getAllStartedFiles();
		while (it.hasNext()) {
			File file = (File) it.next();
			if (file.delete()) {
				it.remove();
			}
		}
		CSVAnalysisSaveAllFinished.cleanCreatedDirectories();
	}

	public int getServerID() {
		return _serverID;
	}

	private LocalSimulationControl getLocalSimulationControl() {
		return _localSimulationControl;
	}
}