package de.uniwue.ki.mas.application.remote.client;

import java.io.*;
import java.net.SocketTimeoutException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.Assert;

import org.tanukisoftware.wrapper.WrapperManager;

import de.uniwue.ki.mas.application.remote.*;
import de.uniwue.ki.mas.xml.castor.v01.remote.XMLRemoteTimeSchedule;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.file.FileUtilities;

final class LocalSimulationParser implements Runnable {
	private static final Logger _logger = Logger.getLogger(LocalSimulationParser.class.getName());

	private static abstract class ParseCommand {
		private final String _commandString;

		public ParseCommand(String commandString) {
			_commandString = commandString;
		}

		public boolean matchCommand(RemoteCommand command) {
			if (command.getCommandId() instanceof String) {
				return _commandString.equals(command.getCommandId());
			} else
				return false;
		}

		public abstract RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command);
	}

	private static final List COMMAND_LIST = new LinkedList();

	static {
		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.REMOTE_INFO_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(command, conn.getXMLRemoteInfo());
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.DETAIL_REMOTE_INFO_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(command, conn.getXMLDetailedRemoteInfo());
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.LOAD_MODEL_FILE_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				if (!noModelOrHasModelOrPausedOrFinished(conn.getLocalSeSAmSimulationRunner().getState()))
					return new RemoteAnswer(command, false);

				String modelFileString = (String) command.getArgument();

				return new RemoteAnswer(command, conn.loadModelFile(new File(modelFileString)));
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.LOAD_MODEL_STREAM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				if (!noModelOrHasModelOrPausedOrFinished(conn.getLocalSeSAmSimulationRunner().getState()))
					return new RemoteAnswer(command, false);

				Object[] args = (Object[]) command.getArgument();
				String remoteModelFileName = (String) args[0];

				StringReader stringReader = new StringReader((String) args[1]);

				return new RemoteAnswer(command, conn.loadModelStream(remoteModelFileName, stringReader));
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.START_SIM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				if (!hasModelOrPausedOrFinished(conn.getLocalSeSAmSimulationRunner().getState()))
					return new RemoteAnswer(command, false);

				String simName = (String) command.getArgument();

				return new RemoteAnswer(command, conn.startSimulation(simName));
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.STOP_SIM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(command, conn.stopSimulation());
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.PAUSE_SIM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(command, conn.pause());
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.UNPAUSE_SIM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(command, conn.unpause());
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.ALLOW_SIM_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				return new RemoteAnswer(
					command,
					conn.setAllowSimulation(((Boolean) command.getArgument()).booleanValue()));
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.QUIT_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				conn.quit();
				return new RemoteAnswer(command, true);
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.CLOSE_CONNECTION_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				conn.closeConnection();
				return new RemoteAnswer(command, true);
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.RESTART_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				try {
					byte[] arg = (byte[]) command.getArgument();

					if (arg != null) {
						ByteArrayInputStream is = new ByteArrayInputStream(arg);
						File tempFile = File.createTempFile("SeSAm", ".jar");
						FileUtilities.copyInputStreamToFile(is, tempFile);

						conn.restart(tempFile);
						return new RemoteAnswer(command, true);

					} else {
						conn.restart(null);
						return new RemoteAnswer(command, true);
					}

				} catch (UnsupportedEncodingException e) {
					printException(e);
				} catch (IOException e) {
					printException(e);
				}

				return null;
			}

		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.GET_ANALYSIS_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				try {
					File jarFile = conn.getAnalysisAsZipFile();
					if ((jarFile != null) && jarFile.exists())
						return new RemoteAnswer(command, marshalFile(jarFile));
					else
						return new RemoteAnswer(command, true);

				} catch (IOException e) {
					printException(e);
					return new RemoteAnswer(command, false);
				}
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.REMOVE_ANALYSIS_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				conn.removeAnalysisFromLastZipFile();

				return new RemoteAnswer(command, true);
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.REMOVE_UNFINISHED_ANALYSIS_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {
				conn.removeAllUnfinishedFiles();

				return new RemoteAnswer(command, true);
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.SET_TIME_SCHEDULE_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {

				XMLRemoteTimeSchedule xmlTimeSchedule = (XMLRemoteTimeSchedule) command.getArgument();

				conn.setXMLTimeSchedule(xmlTimeSchedule);

				return new RemoteAnswer(command, true);
			}
		});

		COMMAND_LIST.add(new ParseCommand(IRemoteProtocolConstants.GET_TIME_SCHEDULE_STRING) {
			public RemoteAnswer parse(LocalSimulationControlConnection conn, RemoteCommand command) {

				return new RemoteAnswer(command, conn.getXMLTimeSchedule());
			}
		});
	}

	private final LocalSimulationControlConnection _localSimulationControlConnection;

	public LocalSimulationParser(LocalSimulationControlConnection localSimulationControlConnection) {
		_localSimulationControlConnection = localSimulationControlConnection;
	}

	public void run() {
		try {
			System.out.println(
				"Started Receiving Commands from " + _localSimulationControlConnection.getName());

			parseCommands();

		} catch (IOException e) {
			if (!(e instanceof SocketTimeoutException) && _localSimulationControlConnection.isConnected())
				printException(e);

		} finally {
			System.out.println(
				"Finished Receiving Commands from " + _localSimulationControlConnection.getName());

			if (_localSimulationControlConnection != null) {
				_localSimulationControlConnection.closeConnectionImmediately();
				if (_localSimulationControlConnection.getShouldQuitClient()) {
					WrapperManager.stop(0);
				}
			}
		}
	}

	private void parseCommands() throws IOException {
		while (!_localSimulationControlConnection.shouldCloseConnection()) {
			try {
				RemoteCommand command = _localSimulationControlConnection.receiveCommand();

				if (command != null) {
					try {
						RemoteAnswer answer = parseCommand(command);

						if (_logger.isLoggable(Level.FINER))
							_logger.fine("Got '" + command + "'! reponse with '" + answer + "'");

						if (answer != null)
							_localSimulationControlConnection.sendAnswer(answer);
						else
							_localSimulationControlConnection.sendAnswer(new RemoteAnswer(command, false));

					} catch (RuntimeException e) {
						e.printStackTrace();
						_localSimulationControlConnection.sendAnswer(new RemoteAnswer(command, false));
					}

				} else
					_localSimulationControlConnection.sendAnswer(new RemoteAnswer(command, false));

			} catch (SocketTimeoutException e) {
				//			} catch (SocketException e) {
				//				if (!_localSimulationControlConnection.isConnected())
				//					throw e;
			}
		}
	}

	private ParseCommand getParseCommand(RemoteCommand command) {
		Iterator it = COMMAND_LIST.iterator();
		while (it.hasNext()) {
			ParseCommand parseCommand = (ParseCommand) it.next();
			if (parseCommand.matchCommand(command)) {

				if (Assert.ASSERTION_ON) {
					while (it.hasNext()) {
						ParseCommand otherParseCommand = (ParseCommand) it.next();
						Assert.assertFalse(otherParseCommand.matchCommand(command));
					}
				}

				return parseCommand;
			}
		}

		return null;
	}

	private RemoteAnswer parseCommand(RemoteCommand command) {
		Assert.assertNotNull(_localSimulationControlConnection);
		
		ParseCommand parseCommand = getParseCommand(command);
		if (parseCommand == null)
			return new RemoteAnswer(command, false);
		else
			return parseCommand.parse(_localSimulationControlConnection, command);
	}

	protected static void printException(Exception e) {
		if (_logger.isLoggable(Level.WARNING))
			_logger.log(Level.WARNING,null,e);
		e.printStackTrace(System.err);
	}

	protected static boolean hasModelOrPausedOrFinished(RemoteClientState state) {
		return (
			Utils.equals(state, IRemoteProtocolConstants.HAS_MODEL_STATE)
				|| Utils.equals(state, IRemoteProtocolConstants.FINISHED_STATE)
				|| Utils.equals(state, IRemoteProtocolConstants.FINISHED_WITH_ERRORS_STATE)
				|| Utils.equals(state, IRemoteProtocolConstants.PAUSED_STATE));
	}

	protected static boolean noModelOrHasModelOrPausedOrFinished(RemoteClientState state) {
		return Utils.equals(state, IRemoteProtocolConstants.NO_MODEL_STATE)
			|| hasModelOrPausedOrFinished(state);
	}

	protected static byte[] marshalFile(File file) throws FileNotFoundException, IOException {
		Assert.assertTrue((file != null) && file.exists());

		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		FileUtilities.copyFileToOutputStream(file, outputStream, true);

		return outputStream.toByteArray();
	}
}