package de.uniwue.ki.mas.application.remote.server;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.Timer;

import junit.framework.Assert;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.list.DefaultMutableListModel;

/**
 * Creation date: 04.09.2002, 19:19:00
 * @author oechslein
 */
public class RemoteSimulationAutomaticControl implements IRemoveable {
	private static final Logger _logger = Logger.getLogger(RemoteSimulationAutomaticControl.class.getName());

	private final class CheckSimTaskListThread extends Thread implements Runnable {
		public CheckSimTaskListThread() {
			super("Schedule Sim Tasks");

			setPriority(Thread.MIN_PRIORITY);
			start();
		}

		private boolean _stop = false;
		public void run() {
			while (!_stop) {
				synchronized (this) {
					try {
						wait();
					} catch (InterruptedException e) {
					}
				}

				if (_stop)
					return;

				try {
					if (_scheduleListener != null)
						_scheduleListener.scheduleStarted();

					_remoteSimulationControl.refreshList(getConnectionListModel().iterator());
					updateScheduledConections(getConnectionListModel().iterator());
					startSimOnAllFreeConnectionsReally();

				} catch (Throwable e) {
					e.printStackTrace();

				} finally {
					if (_scheduleListener != null)
						_scheduleListener.scheduleFinished();

				}
			}
		}

		public void stopThread() {
			_stop = true;

			notifyThread();
		}

		public synchronized void notifyThread() {
			if (!_stop && !isAlive())
				start();

			notifyAll();
		}
	}

	private CheckSimTaskListThread _checkSimTaskListThread;

	protected transient ScheduleListener _scheduleListener;

	public interface ScheduleListener {
		void simulationStarted(String name);
		void scheduleStarted();
		void scheduleFinished();
	}

	private final DefaultMutableListModel _simulationTaskList = new DefaultMutableListModel();
	private javax.swing.Timer _timer = null;

	private boolean _automaticSchedule = false;
	private int _schedulePeriodInSecs = 60;
	protected final RemoteSimulationControl _remoteSimulationControl;

	public RemoteSimulationAutomaticControl(RemoteSimulationControl remoteSimulationControl) {
		super();

		_remoteSimulationControl = remoteSimulationControl;
	}

	public void addScheduleListener(ScheduleListener scheduleListener) {
		Assert.assertNull(_scheduleListener);
		_scheduleListener = scheduleListener;
	}

	public void removeScheduleListener(ScheduleListener scheduleListener) {
		Assert.assertSame(_scheduleListener, scheduleListener);
		_scheduleListener = null;
	}

	public DefaultMutableListModel getConnectionListModel() {
		return _remoteSimulationControl.getConnectionListModel();
	}

	private Iterator getAllFreeSimulationControlConnections() {
		//Iterator it = CollectionsHelper.asShuffledList(getConnectionListModel().iterator()).iterator();
		Iterator it = getAllSimulationControlConnections();
		return CollectionsHelper.select(it, new IPredicateFunction() {
			public boolean isElementOk(Object obj) throws Exception {
				StateCacheSimulationControlConnection simulationControlConnection =
					(StateCacheSimulationControlConnection) obj;
				try {
					return RemoteSimulationControl.isFree(simulationControlConnection);
				} catch (IOException e) {
					e.printStackTrace();
					return false;
				}
			}
		});
	}

	private boolean existsOneFreeSimulationControlConnections() {
		return !CollectionsHelper.isEmpty(getAllFreeSimulationControlConnections());
	}

	private Iterator getAllSimulationControlConnections() {
		return CollectionsHelper.asLinkedList(getConnectionListModel().iterator()).iterator();
	}

	public void startSimOnAllFreeConnections() {
		getCheckSimTaskListThread().notifyThread();
	}

	protected void startSimOnAllFreeConnectionsReally() {
		if (existsOneFreeSimulationControlConnections()) {

			synchronized (getSimulationTaskList()) {
				Iterator it = getSimulationTaskList().iterator();

				while (it.hasNext()) {
					IRemoteModelSimulationTask simTask = (IRemoteModelSimulationTask) it.next();
					String simulationName = simTask.getSimulationName();

					if (startSimulationOnFree(simTask, simulationName)) {
						it.remove();
					}

					if (!existsOneFreeSimulationControlConnections()) {
						break;
					}
				}
			}
		}
	}

	protected void updateScheduledConections(Iterator it) {
		while (it.hasNext()) {
			StateCacheSimulationControlConnection simulationControlConnection =
				(StateCacheSimulationControlConnection) it.next();
			try {
				simulationControlConnection.refresh();
			} catch (IOException e) {
			}
		}
	}

	private boolean startSimulationOnFree(IRemoteModelSimulationTask simTask, String simulationName) {
		Iterator it = getAllFreeSimulationControlConnections();
		while (it.hasNext()) {
			StateCacheSimulationControlConnection simulationControlConnection =
				(StateCacheSimulationControlConnection) it.next();
			try {
				int result = simTask.loadModelAndSimulate(simulationControlConnection);

				if (result == IRemoteModelSimulationTask.LOAD_MODEL_START_SIM_SUCCESS) {
					if (_scheduleListener != null)
						_scheduleListener.simulationStarted(simulationName);

					simulationControlConnection.refresh();
					return true;

				} else if (result == IRemoteModelSimulationTask.MODEL_FILE_LOAD_ERROR) {
					if (_logger.isLoggable(Level.INFO))
						_logger.info(
							"Cannot load modelFile "
								+ simTask.getModelFileName()
								+ " by client "
								+ simulationControlConnection.getName());

				} else if (result == IRemoteModelSimulationTask.START_SIMULATION_ERROR) {
					if (_logger.isLoggable(Level.INFO))
						_logger.info(
							"Cannot start simulation"
								+ simTask.getSimulationName()
								+ " by client "
								+ simulationControlConnection.getName());
				}
			} catch (IOException e) {
				if (_logger.isLoggable(Level.INFO))
					_logger.info(e.toString());
			}
		}

		return false;
	}

	private synchronized void updateTimerTaskThread() {
		stopTimer();

		if (isAutomaticSchedule()) {
			_timer = new Timer(getSchedulePeriodInSecs() * 1000, new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (isAutomaticSchedule()) {
						startSimOnAllFreeConnections();
					}
				}
			});
			_timer.setInitialDelay(0);
			_timer.setCoalesce(true);
			_timer.start();
			Assert.assertTrue(_timer.isRunning());
		}
	}

	private void stopTimer() {
		if (_timer != null) {
			_timer.stop();
			Assert.assertFalse(_timer.isRunning());
			_timer = null;
		}
	}

	public boolean isAutomaticSchedule() {
		return _automaticSchedule;
	}

	public void setAutomaticSchedule(boolean automaticSchedule) {
		boolean oldValue = isAutomaticSchedule();

		if (oldValue != automaticSchedule) {
			_automaticSchedule = automaticSchedule;

			updateTimerTaskThread();
		}
	}

	public int getSchedulePeriodInSecs() {
		return _schedulePeriodInSecs;
	}

	public void setSchedulePeriodInSecs(int updateDelay) {
		int oldValue = getSchedulePeriodInSecs();

		if (oldValue != updateDelay) {
			_schedulePeriodInSecs = updateDelay;

			updateTimerTaskThread();
		}
	}

	public DefaultMutableListModel getSimulationTaskList() {
		return _simulationTaskList;
	}

	public void remove() {
		stopTimer();

		if (_checkSimTaskListThread != null)
			getCheckSimTaskListThread().stopThread();
	}

	private CheckSimTaskListThread getCheckSimTaskListThread() {
		if (_checkSimTaskListThread == null) {
			_checkSimTaskListThread = new CheckSimTaskListThread();
		}

		return _checkSimTaskListThread;
	}
}