package de.uniwue.ki.mas.application.remote.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JOptionPane;

import junit.framework.Assert;
import de.uniwue.ki.mas.application.remote.*;
import de.uniwue.ki.mas.application.remote.event.RemoteSimulationControlConnectionEvent;
import de.uniwue.ki.mas.application.remote.event.RemoteSimulationControlConnectionListenerAdapter;
import de.uniwue.ki.mas.xml.castor.v01.remote.XMLRemoteInfo;
import de.uniwue.ki.util.DefaultThreadFactory;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.list.DefaultMutableListModel;
import edu.oswego.cs.dl.util.concurrent.*;

/**
 * Creation date: 04.09.2002, 19:19:00
 * @author oechslein
 */
public class RemoteSimulationControl implements IRemoveable {
	private static final Logger _logger = Logger.getLogger(RemoteSimulationControl.class.getName());

	protected final Executor _connectionExecutor =
		new ThreadedExecutor(new DefaultThreadFactory("Connection Thread", Thread.NORM_PRIORITY, true));

	private final int _serverID = (int) (Math.random() * Integer.MAX_VALUE);

	private final DefaultMutableListModel _remoteSimulationControlConnectionList =
		new DefaultMutableListModel(new SyncList(new LinkedList(), new WriterPreferenceReadWriteLock()));

	private final ServerThread _serverThread = new ServerThread();

	private final RemoteSimulationAutomaticControl _remoteSimulationAutomaticControl =
		new RemoteSimulationAutomaticControl(this);

	public RemoteSimulationControl() {
		super();
	}

	public DefaultMutableListModel getConnectionListModel() {
		return _remoteSimulationControlConnectionList;
	}

	private void sendDatagram() {
		try {
			sendDatagram(InetAddress.getByName("255.255.255.255"));
		} catch (IOException e) {
			e.printStackTrace();
		}

		try {
			sendDatagram(InetAddress.getLocalHost().getHostAddress());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}

		String ips = System.getProperty("SeSAmRemoteIPs");
		if (ips != null) {
			StringTokenizer stringTokenizer = new StringTokenizer(ips, ";");
			while (stringTokenizer.hasMoreTokens()) {
				sendDatagram(stringTokenizer.nextToken());
			}
		}
	}

	public void sendDatagrams(Iterator it) {
		while (it.hasNext()) {
			try {
				sendDatagram((InetAddress) it.next());

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void sendDatagram(InetAddress adress) throws SocketException, IOException {
		sendDatagram(adress, _serverThread.getServerPort(), getServerID());
	}

	public void sendDatagram(InetAddress[] adressArray) {
		for (int i = 0; i < adressArray.length; i++) {
			try {
				sendDatagram(adressArray[i]);

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void sendDatagram(String textualAdress) {
		try {
			sendDatagram(InetAddress.getAllByName(textualAdress));
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}

	public static void sendDatagram(InetAddress adress, int serverPort, int serverID)
		throws SocketException, IOException {
		Assert.assertTrue("ServerPort must be greater than 0", serverPort > 0);

		String sendStr = IRemoteSimulationConstants.DATAGRAM_ID_STRING + " " + serverPort + " " + serverID;

		byte[] buf = sendStr.getBytes();

		synchronized (RemoteSimulationControl.class) {
			DatagramSocket socket =
				new DatagramSocket(IRemoteSimulationConstants.DEFAULT_DATAGRAM_SERVER_PORT);

			// send it
			DatagramPacket packet =
				new DatagramPacket(
					buf,
					buf.length,
					adress,
					IRemoteSimulationConstants.DEFAULT_DATAGRAM_CLIENT_PORT);

			socket.send(packet);

			socket.close();
			//			
			//			try {
			//				Thread.sleep((long) (Math.random() * 5000));
			//			} catch (InterruptedException e) {
			//			}
			//
		}
	}

	private class ServerThread extends Thread {
		private boolean _stop = false;
		private ServerSocket _serverSocket;

		public ServerThread() {
			super("SeSAm Service Server");
		}

		public synchronized int getServerPort() {
			if (_serverSocket == null)
				return -1;
			else
				return _serverSocket.getLocalPort();
		}

		public void startServer() {
			start();
		}

		public void stopServer() {
			try {
				_stop = true;

				if (_serverSocket != null) {
					_serverSocket.close();
					_serverSocket = null;
				}

			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		public void run() {
			try {
				try {
					synchronized (this) {
						_serverSocket =
							createServerSocket(
								IRemoteSimulationConstants.DEFAULT_CONN_SERVER_PORT,
								IRemoteSimulationConstants.DEFAULT_CONN_SERVER_PORT + 10);
					}

					acceptConnections();

				} finally {
					if (_serverSocket != null)
						_serverSocket.close();
				}

			} catch (BindException e) {
				JOptionPane.showMessageDialog(
					null,
					"Server already running!",
					null,
					JOptionPane.WARNING_MESSAGE);
				return;

			} catch (IOException e) {
				printException(e);
				return;
			}
		}

		private void acceptConnections() throws IOException {
			Assert.assertNotNull(_serverSocket);

			while (!_stop) {
				try {
					Socket clientSocket = _serverSocket.accept();

					if (!_stop) {
						final InetAddress remoteInetAddress = clientSocket.getInetAddress();

						if (!hasConnectionToAddress(remoteInetAddress) && !_stop) {

							try {
								_connectionExecutor.execute(new ConnectionRunnable(clientSocket));

							} catch (InterruptedException e1) {
								e1.printStackTrace();
								Thread.currentThread().interrupt();
							}
						}
					}

				} catch (IOException e) {
					if (_stop || _serverSocket.isClosed())
						return;

					else if (!(e instanceof SocketTimeoutException)) {
						printException(e);
						return;
					}
				}
			}
		}

		private boolean hasConnectionToAddress(final InetAddress remoteInetAddress) {
			boolean found = false;
			Iterator conns = getConnectionListModel().iterator();
			while (conns.hasNext()) {
				StateCacheSimulationControlConnection conn =
					(StateCacheSimulationControlConnection) conns.next();
				InetAddress currInetAddress = conn.getSimulationControlConnection().getInetAddress();
				if (currInetAddress.equals(remoteInetAddress)) {
					if (conn.isConnected()) {
						found = true;
						break;
					} else {
						conns.remove();
					}

				}
			}
			return found;
		}

		private ServerSocket createServerSocket(int minPort, int maxPort) throws IOException {
			for (int currPort = minPort; currPort < maxPort; currPort++) {
				try {
					return new ServerSocket(currPort);
				} catch (BindException e) {
					if (currPort == maxPort - 1)
						throw e;
				}
			}
			// no serverPort found
			return null;
		}
	}

	private class ConnectionRunnable implements Runnable {
		private final Socket _clientSocket;

		public ConnectionRunnable(Socket clientSocket) throws SocketException {
			_clientSocket = clientSocket;
		}

		public void run() {
			try {
				RemoteSimulationControlConnection remoteSimulationControlConnection =
					new RemoteSimulationControlConnection(_clientSocket);

				final StateCacheSimulationControlConnection stateCacheSimulationControlConnection =
					new StateCacheSimulationControlConnection(remoteSimulationControlConnection);

				getConnectionListModel().add(stateCacheSimulationControlConnection);

				remoteSimulationControlConnection
					.addRemoteSimulationControlConnectionListener(
						new RemoteSimulationControlConnectionListenerAdapter() {
					public void closedConnection(RemoteSimulationControlConnectionEvent evt) {
						getConnectionListModel().remove(stateCacheSimulationControlConnection);
					}
				});

			} catch (IOException e) {
				if (e instanceof InvalidClassException) {
					e.printStackTrace();
				}

				// ignore exception, just close socket
				try {
					_clientSocket.close();
				} catch (IOException e2) {
				}
			}
		}
	}

	public void removeAllConnections() {
		Iterator it = CollectionsHelper.copyCollection(getConnectionListModel()).iterator();
		while (it.hasNext()) {
			StateCacheSimulationControlConnection simulationControlConnection =
				(StateCacheSimulationControlConnection) it.next();

			try {
				simulationControlConnection.closeConnection();

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		getConnectionListModel().clear();
	}

	public void refreshList() {
		refreshList(getConnectionListModel().iterator());
	}

	public void refreshList(Iterator conns) {
		while (conns.hasNext()) {
			StateCacheSimulationControlConnection conn = (StateCacheSimulationControlConnection) conns.next();
			if (!conn.isConnected()) {
				conns.remove();
			}
		}

		sendDatagram();
	}

	public void startServer() {
		_serverThread.startServer();
	}

	public void stopServer() {
		_serverThread.stopServer();
	}

	public StateCacheSimulationControlConnection getFreeSimulationControlConnection() {
		for (int i = 0; i < getConnectionListModel().getSize(); i++) {
			StateCacheSimulationControlConnection simulationControlConnection =
				(StateCacheSimulationControlConnection) getConnectionListModel().getElementAt(i);
			try {
				simulationControlConnection.refresh();
				if (isFree(simulationControlConnection))
					return simulationControlConnection;

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return null;
	}

	public static boolean isFree(StateCacheSimulationControlConnection simulationControlConnection)
		throws IOException {

		XMLRemoteInfo xmlRemoteInfo = simulationControlConnection.getXMLRemoteInfo();
		return simulationControlConnection.isAutSchedule()
			&& (xmlRemoteInfo != null)
			&& isFree(xmlRemoteInfo.getState());
	}

	private static boolean isFree(int state) {
		return (IRemoteProtocolConstants.NO_MODEL_STATE.equals(state))
			|| (IRemoteProtocolConstants.HAS_MODEL_STATE.equals(state))
			|| (IRemoteProtocolConstants.FINISHED_STATE.equals(state))
		//|| (IRemoteProtocolConstants.FINISHED_WITH_ERRORS_STATE.equals(state))
		;
	}

	public ISimulationControlConnection startSimulationOnFree(File modelFile, String simulationName) {
		ISimulationControlConnection simulationControlConnection = getFreeSimulationControlConnection();
		if (simulationControlConnection == null)
			return null;

		try {
			if (!simulationControlConnection
				.loadModelStream(modelFile.toString(), new FileReader(modelFile)))
				return null;

			if (simulationControlConnection.startSimulation(simulationName))
				return simulationControlConnection;
			else
				return null;

		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public RemoteSimulationAutomaticControl getRemoteSimulationAutomaticControl() {
		return _remoteSimulationAutomaticControl;
	}

	public void remove() {
		stopServer();
		removeAllConnections();
		getRemoteSimulationAutomaticControl().remove();
	}

	protected void printException(Throwable e) {
		if (_logger.isLoggable(Level.WARNING)) {
			_logger.log(Level.WARNING, null, e);
			e.printStackTrace(System.err);
		}
	}

	public int getServerID() {
		return _serverID;
	}
}