package de.uniwue.ki.mas.application.remote.server;

import java.io.*;
import java.net.InetAddress;
import java.net.Socket;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.Assert;
import de.uniwue.ki.mas.application.remote.*;
import de.uniwue.ki.mas.application.remote.event.IRemoteSimulationControlConnectionListener;
import de.uniwue.ki.mas.application.remote.event.RemoteSimulationControlConnectionEvent;
import de.uniwue.ki.mas.xml.castor.v01.remote.*;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.file.FileUtilities;
import de.uniwue.ki.util.name.AbstractNamed;

public final class RemoteSimulationControlConnection
	extends AbstractNamed
	implements ISimulationControlConnection {

	private final List _listeners = new LinkedList();

	protected static final Logger _logger =
		Logger.getLogger(RemoteSimulationControlConnection.class.getName());

	private final Socket _clientSocket;
	private final DataOutputStream _dataOutputStream;
	private final DataInputStream _dataInputStream;

	public RemoteSimulationControlConnection(Socket clientSocket) throws IOException {
		Assert.assertNotNull(clientSocket);

		clientSocket.setKeepAlive(true);
		clientSocket.setSoTimeout(IRemoteSimulationConstants.SERVER_NORMAL_TIMEOUT_IN_MILLIS);
		clientSocket.setSoLinger(true, IRemoteSimulationConstants.SERVER_NORMAL_TIMEOUT_IN_MILLIS);

		_clientSocket = clientSocket;

		_dataInputStream = new DataInputStream(new BufferedInputStream(clientSocket.getInputStream()));
		_dataOutputStream = new DataOutputStream(new BufferedOutputStream(clientSocket.getOutputStream()));
	}

	protected Socket getClientSocket() {
		return _clientSocket;
	}

	public InetAddress getInetAddress() {
		return getClientSocket().getInetAddress();
	}

	public void addRemoteSimulationControlConnectionListener(IRemoteSimulationControlConnectionListener newListener) {
		_listeners.add(newListener);
	}

	public void removeRemoteSimulationControlConnectionListener(IRemoteSimulationControlConnectionListener newListener) {
		_listeners.remove(newListener);
	}

	protected void fireClosedConnection() {
		if (!_listeners.isEmpty()) {
			RemoteSimulationControlConnectionEvent evt = new RemoteSimulationControlConnectionEvent(this);
			Iterator it = _listeners.iterator();
			while (it.hasNext()) {
				((IRemoteSimulationControlConnectionListener) it.next()).closedConnection(evt);
			}
		}
	}

	public String getName() {
		return getClientSocket().getInetAddress().getHostName()
			+ " ["
			+ getClientSocket().getInetAddress().getHostAddress()
			+ "]:"
			+ getClientSocket().getPort();
	}

	public final boolean isConnected() {
		if (!getClientSocket().isConnected() || getClientSocket().isClosed())
			return false;
		else
			try {
				return getXMLRemoteInfo() != null;
			} catch (IOException e) {
				return false;
			}
	}

	private synchronized void closeThisConnection() throws IOException {
		_dataInputStream.close();
		_dataOutputStream.close();

		getClientSocket().close();
		fireClosedConnection();

		_listeners.clear();
	}

	public XMLRemoteInfo getXMLRemoteInfo() throws IOException {
		return (XMLRemoteInfo) sendObjectCommand(
			new RemoteCommand(IRemoteProtocolConstants.REMOTE_INFO_STRING));
	}

	public XMLDetailedRemoteInfo getXMLDetailedRemoteInfo() throws IOException {
		return (XMLDetailedRemoteInfo) sendObjectCommand(
			new RemoteCommand(IRemoteProtocolConstants.DETAIL_REMOTE_INFO_STRING));
	}

	public boolean loadModelFile(File modelFile) throws IOException {
		try {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_LONG_TIMEOUT_IN_MILLIS);
			String commandString =
				IRemoteProtocolConstants.LOAD_MODEL_FILE_STRING + " " + modelFile.toString();

			return sendBoolCommand(new RemoteCommand(commandString));

		} finally {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_NORMAL_TIMEOUT_IN_MILLIS);
		}
	}

	public boolean loadModelStream(String remoteModelFileName, Reader reader) throws IOException {
		return loadModelStream(remoteModelFileName, FileUtilities.getContentsAsString(reader));
	}

	public boolean loadModelStream(String remoteModelFileName, String fileContents) throws IOException {
		try {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_LONG_TIMEOUT_IN_MILLIS);

			return sendBoolCommand(
				new RemoteCommand(
					IRemoteProtocolConstants.LOAD_MODEL_STREAM_STRING,
					new String[] { remoteModelFileName, fileContents }));

		} finally {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_NORMAL_TIMEOUT_IN_MILLIS);
		}
	}

	public boolean setXMLTimeSchedule(XMLRemoteTimeSchedule xmlRemoteTimeSchedule) throws IOException {
		return sendBoolCommand(
			new RemoteCommand(IRemoteProtocolConstants.SET_TIME_SCHEDULE_STRING, xmlRemoteTimeSchedule));
	}

	public XMLRemoteTimeSchedule getXMLTimeSchedule() throws IOException {
		return (XMLRemoteTimeSchedule) sendObjectCommand(
			new RemoteCommand(IRemoteProtocolConstants.GET_TIME_SCHEDULE_STRING));
	}

	public File getAnalysisAsZipFile() throws IOException {
		return getAnalysisAsZipFile(getAnalysisAsRemoteAnswer());
	}

	public File getAnalysisAsZipFile(RemoteAnswer answer) throws IOException, FileNotFoundException {
		if ((answer == null) || !answer.isSuccessful()) {
			return null;
		}

		File tempFile = File.createTempFile("analysis", ".zip");

		FileUtilities.copyInputStreamToFile(new ByteArrayInputStream((byte[]) answer.getAnswer()), tempFile);

		return tempFile;
	}

	public RemoteAnswer getAnalysisAsRemoteAnswer() throws IOException {
		return sendCommand(new RemoteCommand(IRemoteProtocolConstants.GET_ANALYSIS_STRING));
	}

	public void removeAnalysisFromLastZipFile() throws IOException {
		sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.REMOVE_ANALYSIS_STRING));
	}

	public void removeAllUnfinishedFiles() throws IOException {
		sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.REMOVE_UNFINISHED_ANALYSIS_STRING));
	}

	public boolean startSimulation(String name) throws IOException {
		try {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_LONG_TIMEOUT_IN_MILLIS);

			boolean result =
				sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.START_SIM_STRING, name));

			if (result)
				_logger.warning("Started simulation '" + name + "' on '" + getName() + "'.");

			return result;

		} finally {
			getClientSocket().setSoTimeout(IRemoteSimulationConstants.SERVER_NORMAL_TIMEOUT_IN_MILLIS);
		}
	}

	public boolean stopSimulation() throws IOException {
		boolean result = sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.STOP_SIM_STRING));

		if (result)
			_logger.warning("Stop simulation on '" + getName() + "'.");

		return result;
	}

	public boolean pause() throws IOException {
		return sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.PAUSE_SIM_STRING));
	}

	public boolean setAllowSimulation(boolean allowSimulation) throws IOException {
		return sendBoolCommand(
			new RemoteCommand(IRemoteProtocolConstants.ALLOW_SIM_STRING, Boolean.valueOf(allowSimulation)));
	}

	public boolean unpause() throws IOException {
		return sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.UNPAUSE_SIM_STRING));
	}

	public void closeConnection() throws IOException {
		try {
			sendBoolCommand(new RemoteCommand(IRemoteProtocolConstants.CLOSE_CONNECTION_STRING));

		} finally {
			closeThisConnection();
		}
	}

	public void quit() throws IOException {
		try {
			sendCommandWithoutAnswer(new RemoteCommand(IRemoteProtocolConstants.QUIT_STRING));

		} finally {
			closeThisConnection();
		}
	}

	public void restart(File jarFile) throws IOException {
		try {
			if (jarFile == null)
				sendCommandWithoutAnswer(new RemoteCommand(IRemoteProtocolConstants.RESTART_STRING));

			else {
				ByteArrayOutputStream os = new ByteArrayOutputStream();
				FileUtilities.copyFileToOutputStream(jarFile, os, true);
				sendCommandWithoutAnswer(
					new RemoteCommand(IRemoteProtocolConstants.RESTART_STRING, os.toByteArray()));
			}

		} finally {
			closeThisConnection();
		}
	}

	private synchronized boolean sendBoolCommand(RemoteCommand command) throws IOException {
		RemoteAnswer remoteAnswer = sendCommand(command);
		if (remoteAnswer != null)
			return remoteAnswer.isSuccessful();
		else
			return false;
	}

	private synchronized Object sendObjectCommand(RemoteCommand command) throws IOException {
		RemoteAnswer remoteAnswer = sendCommand(command);
		if ((remoteAnswer != null) && remoteAnswer.isSuccessful())
			return remoteAnswer.getAnswer();
		else
			return null;
	}

	private boolean _inSendCommand = false;

	private synchronized RemoteAnswer sendCommand(RemoteCommand command) throws IOException {
		Assert.assertFalse(_inSendCommand);
		_inSendCommand = true;
		try {
			command.writeObject(_dataOutputStream);
			_dataOutputStream.flush();

			return receiveAnswer(command);

		} finally {
			_inSendCommand = false;
		}
	}

	private synchronized void sendCommandWithoutAnswer(RemoteCommand command) throws IOException {
		Assert.assertFalse(_inSendCommand);
		_inSendCommand = true;
		try {
			command.writeObject(_dataOutputStream);
			_dataOutputStream.flush();

		} finally {
			_inSendCommand = false;
		}
	}

	private synchronized RemoteAnswer receiveAnswer(RemoteCommand command) throws IOException {
		RemoteAnswer remoteAnswer = RemoteAnswer.readObject(_dataInputStream);

		if (remoteAnswer == null)
			return null;

		else if (Utils.equals(remoteAnswer.getAnswerToMessageId(), command.getMessageId()))
			return remoteAnswer;

		else {
			if (_logger.isLoggable(Level.INFO))
				_logger.info(
					"Got wrong answer, expected: "
						+ command.getMessageId()
						+ " got "
						+ remoteAnswer.getAnswerToMessageId());
			return receiveAnswer(command);
		}
	}
}