package de.uniwue.ki.mas.basics.feature.evolution;

import java.util.*;

import junit.framework.Assert;

import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.function.iterator.ParametricFunctionHashedValueComparator;
import de.uniwue.ki.mas.basics.type.AbstractSpecificFunctionType;
import de.uniwue.ki.mas.model.*;
import de.uniwue.ki.util.math.MathHelper;
import de.uniwue.ki.util.math.RandomHelper;

/**
 * Insert Documentation string.
 * Creation date: 29.01.2003, 14:36:31
 * @author oechslein
 */
final class RouletteWheelFunction extends AbstractFunctionDynamic {
	private final IFunction _valueFunctionFunction;
	private final IFunction _iteratorFunction;
	private final IFunction _amountFunction;
	private final IFunction _duplicateAllowedFunction;

	private final SingletonObjectFunctionList _singletonObjectFunctionList =
		new SingletonObjectFunctionList();
	private final HashMap _valueMap = new HashMap();

	public RouletteWheelFunction(
		IFunction valueFunctionFunction,
		IFunction iteratorFunction,
		IFunction amountFunction,
		IFunction duplicateAllowedFunction) {
		super();

		_valueFunctionFunction = valueFunctionFunction;
		_iteratorFunction = iteratorFunction;
		_amountFunction = amountFunction;
		_duplicateAllowedFunction = duplicateAllowedFunction;
	}

	public Object execute(IExecuteFunctionArgs executeFunctionArgs) throws SimulationRuntimeException {
		try {
			return executeIntern(executeFunctionArgs);

		} catch (ModelCompilingException e) {
			throw new SimulationRuntimeException(e);

		} catch (SimulationRuntimeUncheckException e) {
			Throwable cause = e.getCause();
			if (cause instanceof SimulationRuntimeException)
				throw (SimulationRuntimeException)cause;
			else if (cause instanceof ModelCompilingException)
				throw new SimulationRuntimeException(cause);
			else
				throw e;
		}
	}

	private Object executeIntern(IExecuteFunctionArgs executeFunctionArgs)
		throws SimulationRuntimeException, ModelCompilingException {

		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs =
			executeFunctionArgs.createCreateAndExecuteFunctionArgs();

		Object valueFunction = _valueFunctionFunction.execute(executeFunctionArgs);

		// we are destructivly modifying the list so we have to copy it! 
		ArrayList list = createCopiedArrayList(executeFunctionArgs);

		Collections.sort(
			list,
			new ParametricFunctionHashedValueComparator(
				valueFunction,
				createAndExecuteFunctionArgs,
				_singletonObjectFunctionList,
				_valueMap));

		double sumOfValues = calcTotalSumOfValues(createAndExecuteFunctionArgs, valueFunction, list);

		List result = calcResultingList(createAndExecuteFunctionArgs, valueFunction, list, sumOfValues);

		_valueMap.clear();

		return result;
	}

	private ArrayList createCopiedArrayList(IExecuteFunctionArgs executeFunctionArgs)
		throws SimulationRuntimeException {

		Object listOrIterator = _iteratorFunction.execute(executeFunctionArgs);

		if (listOrIterator instanceof Iterator) {
			return ExecuteIteratorFactory.createArrayList((Iterator)listOrIterator);

		} else {
			return new ArrayList((List)listOrIterator);
		}
	}

	private List calcResultingList(
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
		Object valueFunction,
		ArrayList list,
		double sumOfValues)
		throws SimulationRuntimeException, ModelCompilingException {

		List result = new LinkedList();

		boolean duplicateAllowed = _duplicateAllowedFunction.executeForBoolean(createAndExecuteFunctionArgs);
		int amount = _amountFunction.executeForInt(createAndExecuteFunctionArgs);

		int realAmount = duplicateAllowed ? amount : Math.min(amount, list.size());

		for (int i = 0; i < realAmount; i++) {
			int randomObjectIndex =
				chooseObjectIndex(createAndExecuteFunctionArgs, valueFunction, list, sumOfValues);
			Object randomObject = list.get(randomObjectIndex);

			if (!duplicateAllowed) {
				list.remove(randomObjectIndex);
				sumOfValues -= calcValue(createAndExecuteFunctionArgs, valueFunction, randomObject);
			}

			result.add(randomObject);
		}

		return result;
	}

	private double calcTotalSumOfValues(
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
		Object valueFunction,
		ArrayList list)
		throws SimulationRuntimeException, ModelCompilingException {

		double sumOfValues = 0;

		for (int i = 0; i < list.size(); i++) {
			sumOfValues += calcValue(createAndExecuteFunctionArgs, valueFunction, list.get(i));
		}

		return sumOfValues;
	}

	private double calcValue(
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
		Object valueFunction,
		Object currObject)
		throws SimulationRuntimeException, ModelCompilingException {

		if (_valueMap.containsKey(currObject)) {
			return ((Double)_valueMap.get(currObject)).doubleValue();

		} else {
			_singletonObjectFunctionList.setFunctionObject(currObject);
			return AbstractSpecificFunctionType.getParametricFunction(valueFunction).executeForDouble(
				createAndExecuteFunctionArgs,
				_singletonObjectFunctionList);
		}
	}

	private int chooseObjectIndex(
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
		Object valueFunction,
		ArrayList list,
		double sumOfValues)
		throws SimulationRuntimeException, ModelCompilingException {

		int index = 0;
		double randomValue = RandomHelper.getRandomDouble(0, sumOfValues);
		double currSumOfValues = 0;
		for (int i = 0; i < list.size(); i++) {
			Object currObject = list.get(i);

			currSumOfValues += calcValue(createAndExecuteFunctionArgs, valueFunction, currObject);

			if (MathHelper.doubleGreaterEqual(currSumOfValues, randomValue)) {
				return index;
			}

			index++;
		}

		Assert.fail();

		return -1;
	}
}
