package de.uniwue.ki.mas.basics.feature.spatial;
import java.awt.Image;
import java.awt.geom.Point2D;
import java.io.File;
import java.util.Iterator;

import de.uniwue.ki.util.Position2D;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.math.MathHelper;

/**
 * Creation date: (30.10.2001 20:36:12)
 * @author Christoph Oechslein
 */

public abstract class AbstractSpatialMapModel implements ISpatialMapModel {

	private transient SpatialMapModelListenerHelper _spatialMapModelListenerHelper;

	public final void addSpatialMapModelListener(ISpatialMapModelListener spatialMapModelListener) {
		getSpatialMapModelListenerHelper().addSpatialMapModelListener(spatialMapModelListener);
	}

	public final void removeSpatialMapModelListener(ISpatialMapModelListener spatialMapModelListener) {
		getSpatialMapModelListenerHelper().removeSpatialMapModelListener(spatialMapModelListener);
	}

	protected final void fireSpatialInfoChanged(ISpatialInfo spatialInfo, double oldValue, int changedType) {
		getSpatialMapModelListenerHelper().fireSpatialInfoChanged(this, spatialInfo, oldValue, changedType);
	}

	protected final void fireSpatialInfoChanged(ISpatialInfo spatialInfo, Image oldImage, File oldImagePath) {
		getSpatialMapModelListenerHelper().fireSpatialInfoChanged(this, spatialInfo, oldImage, oldImagePath);
	}

	protected final void fireSpatialInfoChanged(ISpatialInfo spatialInfo, double oldPosX, double oldPosY) {
		getSpatialMapModelListenerHelper().fireSpatialInfoChanged(this, spatialInfo, oldPosX, oldPosY);
	}

	protected final void fireSpatialInfoAdded(ISpatialInfo spatialInfo) {
		getSpatialMapModelListenerHelper().fireSpatialInfoAdded(this, spatialInfo);
	}

	protected final void fireSpatialInfoRemoved(ISpatialInfo spatialInfo) {
		getSpatialMapModelListenerHelper().fireSpatialInfoRemoved(this, spatialInfo);
	}

	protected final void fireSpatialMapModelChanged(ISpatialMapModel spatialMapModel) {
		getSpatialMapModelListenerHelper().fireSpatialMapModelChanged(spatialMapModel);
	}
	
	public final void fireSpatialInfosMoved(ISpatialInfo[] movedSpatialInfos, Position2D[] oldPosition){
		getSpatialMapModelListenerHelper().fireSpatialInfosMoved(movedSpatialInfos, oldPosition);
	}

	protected final double angleWithTorus(double fromPosX, double fromPosY, double toPosX, double toPosY) {
		if (isTorus()) {
			boolean mirrorOnXAxis = false;
			boolean mirrorOnYAxis = false;

			if (toPosX > fromPosX) {
				if ((toPosX - fromPosX) > (fromPosX - 0) + (getWidth() - toPosX)) {
					mirrorOnXAxis = true;
				}

			} else {
				if ((fromPosX - toPosX) > (toPosX - 0) + (getWidth() - fromPosX)) {
					mirrorOnXAxis = true;
				}
			}

			if (toPosY > fromPosY) {
				if ((toPosY - fromPosY) > (fromPosY - 0) + (getHeight() - toPosY)) {
					mirrorOnYAxis = true;
				}

			} else {
				if ((fromPosY - toPosY) > (fromPosY - 0) + (getHeight() - toPosY)) {
					mirrorOnYAxis = true;
				}
			}

			double angle = Position2D.angle(fromPosX, fromPosY, toPosX, toPosY);

			if (mirrorOnXAxis == true) {
				if (mirrorOnYAxis == true) {
					return -angle;
				} else {
					return Math.toRadians(180) - angle;
				}
			} else {
				if (mirrorOnYAxis == true) {
					return Math.toRadians(360) - angle;
				} else {
					return angle;
				}
			}

		} else {
			return Position2D.angle(fromPosX, fromPosY, toPosX, toPosY);
		}
	}

	public void changeDirection(ISpatialInfo spatialInfo, double newDirectionInRad) {
		double oldDirection = spatialInfo.getDirection();
		if (!MathHelper.doubleEqual(newDirectionInRad, oldDirection)) {
			spatialInfo.setDirection(newDirectionInRad);

			fireSpatialInfoChanged(
				spatialInfo,
				oldDirection,
				SpatialMapModelChangeEvent.DIRECTION_CHANGED_TYPE);
		}
	}

	public void changeImage(ISpatialInfo spatialInfo, Image newImage) {
		Image oldImage = spatialInfo.getImage();
		File oldImagePath = spatialInfo.getImagePath();
		if (newImage != oldImage) {
			spatialInfo.setImage(newImage);

			fireSpatialInfoChanged(spatialInfo, oldImage, oldImagePath);
		}
	}

	public void changeImagePath(ISpatialInfo spatialInfo, File newImagePath) {
		if (!Utils.equals(spatialInfo.getImagePath(), newImagePath)) {
			Image oldImage = spatialInfo.getImage();
			File oldImagePath = spatialInfo.getImagePath();

			spatialInfo.setImagePath(newImagePath);

			fireSpatialInfoChanged(spatialInfo, oldImage, oldImagePath);
		}
	}

	public void changePosition(ISpatialInfo spatialInfo, double toPositionX, double toPositionY) {
		Position2D fromPosition = spatialInfo.getPosition();

		toPositionX = getCorrectedPositionX(toPositionX);
		toPositionY = getCorrectedPositionY(toPositionY);

		double oldPosX = getCorrectedPositionX(fromPosition.x);
		double oldPosY = getCorrectedPositionY(fromPosition.y);

		if (!MathHelper.doubleEqual(oldPosX, toPositionX) || !MathHelper.doubleEqual(oldPosY, toPositionY)) {

			fromPosition.x = toPositionX;
			fromPosition.y = toPositionY;

			fireSpatialInfoChanged(spatialInfo, oldPosX, oldPosY);

		} else {
			fromPosition.x = oldPosX;
			fromPosition.y = oldPosY;
		}
	}

	public final void changePosition(ISpatialInfo spatialInfo, Position2D toPosition) {
		changePosition(spatialInfo, toPosition.x, toPosition.y);
	}

	public void changeRadius(ISpatialInfo spatialInfo, double newRadius) {
		double oldRadius = spatialInfo.getRadius();
		if (!MathHelper.doubleEqual(newRadius, oldRadius)) {
			spatialInfo.setRadius(newRadius);

			fireSpatialInfoChanged(spatialInfo, oldRadius, SpatialMapModelChangeEvent.RADIUS_CHANGED_TYPE);
		}
	}

	public void changeSpeed(ISpatialInfo spatialInfo, double newSpeed) {
		double oldSpeed = spatialInfo.getSpeed();
		if (!MathHelper.doubleEqual(newSpeed, oldSpeed)) {
			spatialInfo.setSpeed(newSpeed);

			fireSpatialInfoChanged(spatialInfo, oldSpeed, SpatialMapModelChangeEvent.SPEED_CHANGED_TYPE);
		}
	}
	
	public void changeDrawPriority(ISpatialInfo spatialInfo, double newPriority){
		double oldPriority = spatialInfo.getDrawPriority();
		spatialInfo.setDrawPriority(newPriority);
		fireSpatialInfoChanged(spatialInfo, oldPriority, SpatialMapModelChangeEvent.DRAWPRIORITY_CHANGED_TYPE);
	
	}

	public double distance(ISpatialInfo spatialInfo1, ISpatialInfo spatialInfo2) {
		Position2D pos1 = spatialInfo1.getPosition();
		Position2D pos2 = spatialInfo2.getPosition();

		double posX1 = getCorrectedPositionX(pos1.x);
		double posY1 = getCorrectedPositionY(pos1.y);
		double posX2 = getCorrectedPositionX(pos2.x);
		double posY2 = getCorrectedPositionY(pos2.y);

		return distanceWithTorus(posX1, posY1, posX2, posY2, isTorus());
	}

	protected final double distanceSqWithTorus(
		double pos1X,
		double pos1Y,
		double pos2X,
		double pos2Y,
		boolean torus) {

		return distanceSqWithTorus(pos1X, pos1Y, pos2X, pos2Y, getWidth(), getHeight(), torus);
	}

	public static final double distanceSqWithTorus(
		double pos1X,
		double pos1Y,
		double pos2X,
		double pos2Y,
		ISpatialMapModel spatialMapModel) {

		return distanceSqWithTorus(
			pos1X,
			pos1Y,
			pos2X,
			pos2Y,
			spatialMapModel.getWidth(),
			spatialMapModel.getHeight(),
			spatialMapModel.isTorus());
	}

	public static final double distanceSqWithTorus(
		Position2D pos1,
		Position2D pos2,
		ISpatialMapModel spatialMapModel) {

		return distanceSqWithTorus(pos1.getX(), pos1.getY(), pos2.getX(), pos2.getY(), spatialMapModel);
	}

	public static double distanceSqWithTorus(
		double pos1X,
		double pos1Y,
		double pos2X,
		double pos2Y,
		double width,
		double height,
		boolean torus) {

		double returnvalue;
		if (torus) {
			double distancePosX, distancePosY;

			if (pos2X > pos1X) {
				distancePosX = Math.min(pos2X - pos1X, (pos1X - 0) + (width - pos2X));

			} else {
				distancePosX = Math.min(pos1X - pos2X, (pos2X - 0) + (width - pos1X));
			}

			if (pos2Y > pos1Y) {
				distancePosY = Math.min(pos2Y - pos1Y, (pos1Y - 0) + (height - pos2Y));

			} else {
				distancePosY = Math.min(pos1Y - pos2Y, (pos2Y - 0) + (height - pos1Y));
			}
			//double sicher machen
			returnvalue = distancePosX * distancePosX + distancePosY * distancePosY;
			if(MathHelper.doubleEqual(returnvalue,0)){returnvalue=0;}			
			return returnvalue;

		} else {
			//double sicher machen
			returnvalue = Point2D.distanceSq(pos1X, pos1Y, pos2X, pos2Y);
			if(MathHelper.doubleEqual(returnvalue,0)){returnvalue=0;}			
			return returnvalue;
		}
	}

	protected final double distanceWithTorus(
		double pos1X,
		double pos1Y,
		double pos2X,
		double pos2Y,
		boolean torus) {
		return Math.sqrt(distanceSqWithTorus(pos1X, pos1Y, pos2X, pos2Y, torus));
	}

	protected final double getCorrectedPositionX(double positionX) {
		if (positionX >= getWidth()) {
			if (isTorus()) {
				return positionX % getWidth();

			} else {
				return getWidth() - 1;
			}

		} else if (positionX < 0) {
			if (isTorus()) {
				// attention -1 % 4 = -1 (not 3)
				//return positionX % getWidth();
				return (positionX % getWidth()) + getWidth();

			} else {
				return 0;
			}

		} else {
			return positionX;
		}
	}

	protected final double getCorrectedPositionY(double positionY) {
		if (positionY >= getHeight()) {
			if (isTorus()) {
				return positionY % getHeight();

			} else {
				return getHeight() - 1;
			}

		} else if (positionY < 0) {
			if (isTorus()) {
				// attention -1 % 4 = -1 (not 3)
				//return positionY % getHeight();
				return (positionY % getHeight()) + getHeight();

			} else {
				return 0;
			}

		} else {
			return positionY;
		}
	}

	public final ISpatialInfo getSpatialInfoAt(Position2D pos, double radius, int index) {
		return (ISpatialInfo)CollectionsHelper.get(getSpatialInfos(pos, radius), index);
	}

	public final int getSpatialInfoCount(Position2D pos, double radius) {
		return CollectionsHelper.size(getSpatialInfos(pos, radius));
	}

	public final Iterator getSpatialInfos(Position2D pos, double radius) {

		// slow but easy implementation; good for large radius or small amount of objects

		final double radiusSq = radius * radius;

		final double posX = getCorrectedPositionX(pos.x);
		final double posY = getCorrectedPositionY(pos.y);
		final boolean torus = isTorus();

		return CollectionsHelper.select(getAllSpatialInfos(), new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				ISpatialInfo spatialInfo = (ISpatialInfo)obj;
				Position2D objPos = spatialInfo.getPosition();

				double objPosX = objPos.x;
				double objPosY = objPos.y;

				return distanceSqWithTorus(objPosX, objPosY, posX, posY, torus) <= radiusSq;
			}
		});

	}

	public final Iterator getSpatialInfos(
		Position2D pos,
		double radius,
		final double angleStartInRad,
		final double angleEndInRad) {

		final double posX = getCorrectedPositionX(pos.x);
		final double posY = getCorrectedPositionY(pos.y);

		return CollectionsHelper.select(getSpatialInfos(pos, radius), new IPredicateFunction() {
			public boolean isElementOk(Object anObject) {
				ISpatialInfo currSpatialInfo = (ISpatialInfo)anObject;

				final Position2D objPos = currSpatialInfo.getPosition();

				double objPosX = objPos.x;
				double objPosY = objPos.y;

				return isPositionInAngleWithTorus(
					posX,
					posY,
					objPosX,
					objPosY,
					angleStartInRad,
					angleEndInRad);
			}
		});
	}

	protected final boolean isPositionInAngleWithTorus(
		double fromPosX,
		double fromPosY,
		double toPosX,
		double toPosY,
		double angleStartInRad,
		double angleEndInRad) {

		if (MathHelper.doubleEqual(toPosX,fromPosX) 
			&& MathHelper.doubleEqual(toPosY, fromPosY)){
		//if((toPosX==fromPosX) && (toPosY==fromPosY)){
			return true;

		} else {
			double currAngle = angleWithTorus(fromPosX, fromPosY, toPosX, toPosY);
				
			//adapt angles to lie between 0 and 2 PI
			if(angleStartInRad < 0){
				angleStartInRad += 2 * Math.PI;
			}
			if(angleEndInRad > 2 * Math.PI){
				angleEndInRad -= 2 * Math.PI;
			}
									
			//switch because of only positive degrees!
			if(angleStartInRad <= angleEndInRad)
				return (angleStartInRad <= currAngle) && (currAngle <= angleEndInRad);
			else
				return ((angleStartInRad <= currAngle) || (currAngle <= angleEndInRad));
		}
	}

	public final boolean overlap(ISpatialInfo spatialInfo1, ISpatialInfo spatialInfo2) {
		double distance1To2 = distance(spatialInfo1, spatialInfo2);
		return distance1To2 <= (spatialInfo1.getRadius() + spatialInfo2.getRadius());
	}

	public void remove() {
		getSpatialMapModelListenerHelper().remove();
	}

	private synchronized SpatialMapModelListenerHelper getSpatialMapModelListenerHelper() {
		if (_spatialMapModelListenerHelper == null) {
			_spatialMapModelListenerHelper = new SpatialMapModelListenerHelper();
		}
		return _spatialMapModelListenerHelper;
	}
}