package de.uniwue.ki.mas.basics.function;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.model.*;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.collection.IteratorWithException;

/**
 * Creation date: (28.03.2002 19:41:30)
 * @author Christoph Oechslein
 */
public final class ExecuteIteratorFactory {
	public static final IExecuteIterator EMPTY_EXECUTE_ITERATOR = new AbstractExecuteIterator() {
		public boolean hasNext() {
			return false;
		}

		public Object nextExecute() throws SimulationRuntimeException {
			Assert.fail();
			return null;
		}
	};
	private static class DelegateExecuteIteratorWithException extends AbstractExecuteIterator {
		private final IteratorWithException _iteratorWithException;

		public DelegateExecuteIteratorWithException(IteratorWithException iteratorWithException) {
			_iteratorWithException = iteratorWithException;
		}

		public boolean hasNext() {
			return _iteratorWithException.hasNext();
		}

		public Object nextExecute() throws SimulationRuntimeException {
			try {
				return _iteratorWithException.nextWithException();

			} catch (Exception e) {
				if (e instanceof SimulationRuntimeException)
					throw (SimulationRuntimeException) e;

				else if (e instanceof ModelCompilingException)
					throw new SimulationRuntimeCausedByModelCompileException((ModelCompilingException) e);

				else
					throw new RuntimeException(e);
			}
		}
		public Object nextWithException() throws Exception {
			return _iteratorWithException.nextWithException();
		}
	}

	private static class DelegateExecuteIterator extends AbstractExecuteIterator {
		private final Iterator _iterator;

		public DelegateExecuteIterator(Iterator iterator) {
			_iterator = iterator;
		}

		public boolean hasNext() {
			return _iterator.hasNext();
		}

		public Object nextExecute() {
			return _iterator.next();
		}

		public Object nextWithException() {
			return _iterator.next();
		}
	}

	public ExecuteIteratorFactory() {
		super();
	}

	private static IExecuteIterator createExecuteIteratorIntern(Iterator it, boolean withCallingHasNext) {
		if (withCallingHasNext && !it.hasNext()) {
			return EMPTY_EXECUTE_ITERATOR;

		} else if (it instanceof IExecuteIterator) {
			return (IExecuteIterator) it;

		} else if (it instanceof IteratorWithException) {
			return new DelegateExecuteIteratorWithException((IteratorWithException) it);

		} else {
			return new DelegateExecuteIterator(it);
		}
	}

	public static final IExecuteIterator createExecuteIterator(Object anObject) {
		return createExecuteIterator(anObject, false);
	}

	public static final IExecuteIterator createExecuteIterator(Object anObject, boolean withCallingHasNext) {

		if (anObject instanceof Iterator) {
			return ExecuteIteratorFactory.createExecuteIteratorIntern(
				(Iterator) anObject,
				withCallingHasNext);

		} else if (anObject instanceof List) {
			if (((List) anObject).isEmpty()) {
				return ExecuteIteratorFactory.EMPTY_EXECUTE_ITERATOR;

			} else {
				return new DelegateExecuteIterator(((List) anObject).iterator());
			}

		} else {
			Assert.fail();
			return ExecuteIteratorFactory.EMPTY_EXECUTE_ITERATOR;
		}
	}

	public static final IExecuteIterator createExecuteIterator(
		IExecuteFunctionArgs executeFunctionArgs,
		IFunction function)
		throws SimulationRuntimeException {

		return createExecuteIterator(function.execute(executeFunctionArgs));
	}

	public static final LinkedList createLinkedList(Iterator anObject) throws SimulationRuntimeException {
		if (anObject instanceof IteratorWithException) {
			try {
				return CollectionsHelper.asLinkedListWithException((IteratorWithException) anObject);
			} catch (Exception e) {
				if (e instanceof SimulationRuntimeException)
					throw (SimulationRuntimeException) e;
				else if (e instanceof ModelCompilingException)
					throw new SimulationRuntimeCausedByModelCompileException((ModelCompilingException) e);
				else
					throw new RuntimeException(e);
			}

		} else {
			return CollectionsHelper.asLinkedList(anObject);
		}
	}

	public static final ArrayList createArrayList(Iterator anObject) throws SimulationRuntimeException {
		if (anObject instanceof IteratorWithException) {
			try {
				return CollectionsHelper.asArrayListWithException((IteratorWithException) anObject);
			} catch (Exception e) {
				if (e instanceof SimulationRuntimeException)
					throw (SimulationRuntimeException) e;
				else if (e instanceof ModelCompilingException)
					throw new SimulationRuntimeCausedByModelCompileException((ModelCompilingException) e);
				else
					throw new RuntimeException(e);
			}

		} else {
			return CollectionsHelper.asArrayList(anObject);
		}
	}

	public static final List createList(Object anObject) throws SimulationRuntimeException {
		if (anObject instanceof Iterator) {
			return createLinkedList((Iterator) anObject);

		} else if (anObject instanceof List) {
			return (List) anObject;

		} else {
			Assert.fail();
			return CollectionsHelper.EMPTY_LIST;
		}
	}

	public static final List createList(IExecuteFunctionArgs executeFunctionArgs, IFunction function)
		throws SimulationRuntimeException {
		return createList(function.execute(executeFunctionArgs));
	}
}