package de.uniwue.ki.mas.basics.function.bool;
import java.util.Iterator;
import java.util.logging.Level;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.functionCall.IComposedFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionDeclaration.AbstractSpecificFunctionUnlimitedArgsDeclaration;
import de.uniwue.ki.mas.basics.functionDeclaration.FunctionDeclarationCompiler;
import de.uniwue.ki.mas.basics.type.BooleanType;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * Creation date: (01.02.2002 13:59:00)
 * @author Christoph Oechslein
 */
final class AndSpecificFunctionDeclaration extends AbstractSpecificFunctionUnlimitedArgsDeclaration {

	public AndSpecificFunctionDeclaration() {
		super("And", BooleanType.getInstance(), BooleanType.getInstance(), true);

		setDocumentation(
			"Returns the verity content if its operands are both 'true' " + "e.g. And(TRUE, FALSE) = FALSE");
	}

	public IFunction createFunction(ICreateFunctionArgs createFunctionArgs, IFunctionList functions) {
		if (!createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()) {
			return new FunctionAnd(functions);
			
		} else {
			switch (functions.size()) {
				case 0 :
					return BooleanType.BOOLEAN_TYPE_TRUE_ARGUMENT;
				case 1 :
					return functions.getFunction(0);
				case 2 :
					return new FunctionAnd2Args(functions.getFunction(0), functions.getFunction(1));
				default :
					return new FunctionAnd(functions);
			}
		}
	}

	// just for faster dynamic calling
	public boolean executeForBoolean(
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
		IFunctionList functions)
		throws SimulationRuntimeException {

		switch (functions.size()) {
			case 0 :
				return true;
			case 1 :
				return functions.getFunction(0).executeForBoolean(createAndExecuteFunctionArgs);
			case 2 :
				return FunctionAnd2Args.executeForBoolean(
					functions.getFunction(0),
					functions.getFunction(1),
					createAndExecuteFunctionArgs);
			default :
				return FunctionAnd.executeForBoolean(functions, createAndExecuteFunctionArgs);
		}
	}

	private boolean getConstantValue(
		ICreateFunctionArgs createFunctionArgs,
		IFunctionCallInternDeclaration element)
		throws ModelCompilingException {

		try {
			return element.createFunction(createFunctionArgs).executeForBoolean(
				createFunctionArgs.createCreateAndExecuteFunctionArgs());

		} catch (SimulationRuntimeException e) {
			throw new ModelCompilingException(e);
		}
	}

	public IFunction createFunction(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (!createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()) {
			return super.createFunction(createFunctionArgs, composedFunctionCallInternDeclaration);

		} else {

			FunctionList functionList = new FunctionList();

			Iterator it = composedFunctionCallInternDeclaration.getChildren();

			boolean onlyConstantItems = true;
			boolean onlyReadOnlyItems = true;
			boolean transformed = false;

			while (it.hasNext()) {
				IFunctionCallInternDeclaration element = (IFunctionCallInternDeclaration) it.next();

				if (!element.isConstant(createFunctionArgs)) {
					onlyConstantItems = false;
					if (!element.isReadOnly(createFunctionArgs)) {
						onlyReadOnlyItems = false;
					}

					functionList.add(element.createFunction(createFunctionArgs));

				} else if (getConstantValue(createFunctionArgs, element) == true) {
					// We can ignore constant true values!
					transformed = true;

				} else if (onlyReadOnlyItems || onlyConstantItems) {
					// AND(... ,FALSE, ...) is always false!
					if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO)) {
						FunctionDeclarationCompiler.LOGGER.info(
							"Removed redundand AND '"
								+ composedFunctionCallInternDeclaration.getName()
								+ "'"
								+ " ( = FALSE)");
					}

					return BooleanType.BOOLEAN_TYPE_FALSE_ARGUMENT;

				} else {
					// we need to execute the other funktions since we don't know if
					// they use side-effects (like pop or progn, ...)
					functionList.add(BooleanType.BOOLEAN_TYPE_FALSE_ARGUMENT);

					// but we can ignore the next items
					transformed = true;
					break;
				}
			}

			if (transformed) {
				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO)) {
					FunctionDeclarationCompiler.LOGGER.info(
						"Transformed AND in '"
							+ composedFunctionCallInternDeclaration.getName()
							+ "' to "
							+ functionList.getName());
				}
			}

			return createFunction(createFunctionArgs, functionList);
		}
	}

	public boolean isConstant(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		Iterator it = composedFunctionCallInternDeclaration.getChildren();

		if (!it.hasNext())
			return true;

		while (it.hasNext()) {
			IFunctionCallInternDeclaration element = (IFunctionCallInternDeclaration) it.next();
			if (element.isConstant(createFunctionArgs)) {

				if (getConstantValue(createFunctionArgs, element) == false) {
					// AND(... ,FALSE, ...) is always false (and constant)!
					return true;
				}
			}
		}

		return super.isConstant(createFunctionArgs, composedFunctionCallInternDeclaration);
	}
}