package de.uniwue.ki.mas.basics.function.number;
import java.util.List;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.functionDeclaration.*;
import de.uniwue.ki.mas.basics.functionDeclaration.list.FunctionDeclarationList;
import de.uniwue.ki.mas.basics.functionDeclaration.list.IFunctionDeclarationList;
import de.uniwue.ki.mas.basics.type.*;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.collection.CollectionsHelper;

/**
 * Creation date: (29.01.2002 10:36:30)
 * @author Christoph Oechslein
 */

public final class FunctionDeclarationNumberFactory {

	public static final FloorSpecificFunctionDeclaration FLOOR_FUNCTION_DECLARATION =
		new FloorSpecificFunctionDeclaration();

	public static final IGenericFunctionDeclaration PLUS_FUNCTION_DECLARATION =
		new AbstractNumberGenericFunctionDeclaration("+") {
		{
			setDocumentation("Returns the sum of its operands e.g. +(4, 5) = 9");
		}
		protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
			return new SpecificFunctionDeclarationPlus(PLUS_FUNCTION_DECLARATION, outputType);
		}
	};

	public static final IGenericFunctionDeclaration MULTIPLY_FUNCTION_DECLARATION =
		new AbstractNumberGenericFunctionDeclaration("*") {
		{
			setDocumentation("Returns the sum of its operands e.g. *(4, 5) = 20");
		}
		protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
			return new SpecificFunctionDeclarationMultiply(MULTIPLY_FUNCTION_DECLARATION, outputType);
		}
	};

	public static final IGenericFunctionDeclaration MAX_FUNCTION_DECLARATION =
		new AbstractNumberGenericFunctionDeclaration("Max") {
		{
			setDocumentation("Returns the greatest of its operands e.g. Max(2, 5) = 5");
		}
		protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
			return new SpecificFunctionDeclarationMax(MAX_FUNCTION_DECLARATION, outputType);
		}
	};

	public static final IGenericFunctionDeclaration MIN_FUNCTION_DECLARATION =
		new AbstractNumberGenericFunctionDeclaration("Min") {
		{
			setDocumentation("Returns the greatest of its operands e.g. Min(2, 5) = 2");
		}
		protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
			return new SpecificFunctionDeclarationMin(MIN_FUNCTION_DECLARATION, outputType);
		}
	};

	private static final List _numberPrototypeOneArgList =
		CollectionsHelper.createList(DoubleType.getInstance());
	private static final List _numberPrototypeTwoArgsList =
		CollectionsHelper.createList(DoubleType.getInstance(), DoubleType.getInstance());

	private static IFunctionDeclarationList _cachedFunctionDeclarationList = new FunctionDeclarationList();

	static {
		insertUnlimitedFunctions();

		insertTwoAryFunctions();

		insertOneAryFunctions();
	}

	public static void insertFunctions(IFunctionDeclarationList functionDeclarationList) {
		functionDeclarationList.addAll(_cachedFunctionDeclarationList);
	}

	private static void insertOneAryFunctions() {

		AbstractSpecificFunctionArgsDeclaration functionDecl;

		// Unary -
		_cachedFunctionDeclarationList.add(new AbstractNumberGenericFunctionDeclaration("-Unary", 1) {
			{
				setDocumentation("Returns the negation of its operand e.g. -Unary(5) = -5");
			}
			protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
				return new MinusUnarySpecificFunctionDeclaration(outputType);
			}
		});

		// exp
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Exp",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionExp(functions.getFunction(0));
			}
			// just for faster dynamic calling
			public double executeForDouble(
				ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
				IFunctionList functions)
				throws SimulationRuntimeException {
				return Math.exp(functions.getFunction(0).executeForDouble(createAndExecuteFunctionArgs));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation(
			"Returns Euler's number e raised to the power of its operand e.g. exp(2) = e^2 ~ 7.4");
		_cachedFunctionDeclarationList.add(functionDecl);

		// ln
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Ln",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionLn(functions.getFunction(0));
			}
			// just for faster dynamic calling
			public double executeForDouble(
				ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs,
				IFunctionList functions)
				throws SimulationRuntimeException {
				return Math.log(functions.getFunction(0).executeForDouble(createAndExecuteFunctionArgs));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the natural logarithm (base e) of its operand e.g. Ln(e) = 1");
		_cachedFunctionDeclarationList.add(functionDecl);

		// abs
		_cachedFunctionDeclarationList.add(new AbstractNumberGenericFunctionDeclaration("Abs", 1, true) {
			{
				setDocumentation("Returns the absolute value of its operand e.g. abs(-5) = 5");
			}
			protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
				return new AbsSpecificFunctionDeclaration(outputType);
			}
		});

		// Sq
		_cachedFunctionDeclarationList.add(new AbstractNumberGenericFunctionDeclaration("Sq", 1, true) {
			{
				setDocumentation("Returns the square of its operand e.g. Sq(5) = 25");
			}
			protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
				return new SquareSpecificFunctionDeclaration(outputType);
			}
		});

		// Sqrt
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Sqrt",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionSqrt(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the positive square root of its operand e.g. Sqrt(25) = 5");
		_cachedFunctionDeclarationList.add(functionDecl);

		// IsNaN
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"IsNaN",
				BooleanType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());

				return new FunctionIsNaN(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns true if value is not a number, e.g. IsNaN(0 / 0) == true.");
		_cachedFunctionDeclarationList.add(functionDecl);

		// IsInfinity
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"IsInfinity",
				BooleanType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());

				return new FunctionIsInfinity(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation(
			"Returns true if value is infinity (negative or positive), e.g. IsInfinity(1 / 0) == true.");
		_cachedFunctionDeclarationList.add(functionDecl);

		insertRoundFunctions();

		insertTrigonometricFunctions();
	}

	private static void insertRoundFunctions() {
		// truncate
		AbstractSpecificFunctionArgsDeclaration functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Truncate",
				IntegerType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionTruncate(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Truncates the decimal places of its operand e.g. Truncate(5.89) = 5");
		_cachedFunctionDeclarationList.add(functionDecl);

		// ceil
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Ceil",
				IntegerType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionCeil(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation(
			"Returns the smallest (closest to negative infinity) value that is not less than the argument.");
		_cachedFunctionDeclarationList.add(functionDecl);

		// floor
		_cachedFunctionDeclarationList.add(FLOOR_FUNCTION_DECLARATION);

		// round
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Round",
				IntegerType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionRound(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation(
			"Returns the closest integer to the argument. The result is rounded to an integer by adding 1/2, taking the floor of the result.");
		_cachedFunctionDeclarationList.add(functionDecl);
	}

	private static void insertTrigonometricFunctions() {
		// cos
		AbstractSpecificFunctionArgsDeclaration functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Cos",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionCos(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the cosinus of the argument (in degree).");
		_cachedFunctionDeclarationList.add(functionDecl);

		// sin
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Sin",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionSin(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the sinus of the argument (in degree).");
		_cachedFunctionDeclarationList.add(functionDecl);

		// tan
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Tan",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionTan(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the tangens of the argument (in degree).");
		_cachedFunctionDeclarationList.add(functionDecl);

		// acos
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"ACos",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionACos(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the arc cosinus of the argument, result is in degrees.");
		_cachedFunctionDeclarationList.add(functionDecl);

		// asin
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"ASin",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionASin(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the arc sinus of the argument, result is in degrees.");
		_cachedFunctionDeclarationList.add(functionDecl);

		// atan
		functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"ATan",
				DoubleType.getInstance(),
				_numberPrototypeOneArgList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(1, functions.size());
				return new FunctionATan(functions.getFunction(0));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the arc tangens of the argument, result is in degrees.");
		_cachedFunctionDeclarationList.add(functionDecl);
	}

	private static void insertTwoAryFunctions() {

		// -
		_cachedFunctionDeclarationList.add(new AbstractNumberGenericFunctionDeclaration("-", 2) {
			{
				setDocumentation("Returns the difference of its operands where (Arg1) is the minuend and (Arg2) is the subtrahend e.g. -(20, 4) = 16");
			}
			protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
				return new MinusSpecificFunctionDeclaration(outputType);
			}
		});

		// /
		_cachedFunctionDeclarationList.add(new DivideSpecificFunctionDeclaration(true));

		// Div
		_cachedFunctionDeclarationList.add(new DivideSpecificFunctionDeclaration(false));

		// ^ = Power
		_cachedFunctionDeclarationList.add(new PowerSpecificFunctionDeclaration());

		// Log
		AbstractSpecificFunctionArgsDeclaration functionDecl =
			new AbstractSpecificFunctionArgsDeclaration(
				"Log",
				DoubleType.getInstance(),
				_numberPrototypeTwoArgsList) {
			public IFunction createFunction(
				ICreateFunctionArgs createFunctionArgs,
				IFunctionList functions) {
				Assert.assertEquals(2, functions.size());
				return new FunctionLog(functions.getFunction(0), functions.getFunction(1));
			}
		};
		functionDecl.setExpertFunctionDeclaration(true);
		functionDecl.setDocumentation("Returns the logarithm (base (Arg2)) of (Arg1) e.g. Log(100, 10) = 2");
		_cachedFunctionDeclarationList.add(functionDecl);

		// % = modulo
		_cachedFunctionDeclarationList.add(new AbstractNumberGenericFunctionDeclaration("%", 2, true) {
			{
				setDocumentation("Returns the remainder of its operands from an implied division where (Arg1) is the dividend and (Arg2) is the divisor e.g. %(7, 3) = 1");
			}
			protected ISpecificFunctionDeclaration createFunctionDeclarationIntern(ISpecificType outputType) {
				return new ModuloSpecificFunctionDeclaration(outputType);
			}
		});
	}

	private static void insertUnlimitedFunctions() {
		// +
		_cachedFunctionDeclarationList.add(PLUS_FUNCTION_DECLARATION);

		// *
		_cachedFunctionDeclarationList.add(MULTIPLY_FUNCTION_DECLARATION);

		// Max
		_cachedFunctionDeclarationList.add(MAX_FUNCTION_DECLARATION);

		// Min
		_cachedFunctionDeclarationList.add(MIN_FUNCTION_DECLARATION);
	}
}