package de.uniwue.ki.mas.basics.function.variable;
import junit.framework.Assert;
import de.uniwue.ki.mas.basics.action.AbstractFunctionAction;
import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.type.*;
import de.uniwue.ki.mas.entity.body.BodyVariableNotInitialized;
import de.uniwue.ki.mas.entity.body.IBody;
import de.uniwue.ki.mas.entity.object.IObject;
import de.uniwue.ki.mas.entity.object.IObjectExecuteFunctionArgs;
import de.uniwue.ki.mas.entity.object.agent.IAgentExecuteFunctionArgs;
import de.uniwue.ki.mas.entity.variable.IBodyVariable;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * Creation date: (04.09.2001 15:46:50)
 * @author Christoph Oechslein
 */

final class FunctionActionSetVariable extends AbstractFunctionAction implements IFunction {
	private ObjectClassVariableIndexCache _objectClassVariableIndexCache;

	private final int _valueType;

	private static final int OBJECT_TYPE = 0;
	private static final int ITERATOR_TYPE = 1;
	private static final int BOOLEAN_TYPE = 2;
	private static final int DOUBLE_TYPE = 3;
	private static final int INTEGER_TYPE = 4;

	private final IFunction _variableFunction;
	private final IFunction _valueFunction;

	public FunctionActionSetVariable(
		ISpecificType type,
		IFunction variableFunction,
		IFunction valueFunction) {
		super();

		_valueType = convertToTypeValue(type);
		_variableFunction = variableFunction;
		_valueFunction = valueFunction;
	}

	public void act(IExecuteFunctionArgs executeFunctionArgs) throws SimulationRuntimeException {
		IObject simObject = ((IAgentExecuteFunctionArgs) executeFunctionArgs).getObject();

		int index = getIndex(simObject, executeFunctionArgs);

		setValue(
			_valueType,
			(IObjectExecuteFunctionArgs) executeFunctionArgs,
			simObject,
			index,
			_valueFunction);
	}

	static int convertToTypeValue(ISpecificType type) {
		if (DoubleType.getInstance().equals(type))
			return DOUBLE_TYPE;

		else if (IntegerType.getInstance().equals(type))
			return INTEGER_TYPE;

		else if (BooleanType.getInstance().equals(type))
			return BOOLEAN_TYPE;

		else if (GenericIteratorType.getInstance().equals(type.getGenericType()))
			return ITERATOR_TYPE;

		else
			return OBJECT_TYPE;
	}

	private int getIndex(IObject fromObject, IExecuteFunctionArgs executeFunctionArgs)
		throws SimulationRuntimeException {
		Object obj = _variableFunction.execute(executeFunctionArgs);

		if (_objectClassVariableIndexCache != null) {
			return _objectClassVariableIndexCache.getIndexFor(fromObject, obj);

		} else if (obj instanceof Number) {
			return ((Number) obj).intValue();

		} else {
			_objectClassVariableIndexCache = new ObjectClassVariableIndexCache();
			return _objectClassVariableIndexCache.getIndexFor(fromObject, obj);
		}
	}

	public String getName() {
		return "SetVariable(" + _variableFunction.getName() + ", " + _valueFunction.getName() + ")";
	}

	static void setValue(
		int valueType,
		IObjectExecuteFunctionArgs objectExecuteFunctionArgs,
		IObject fromObject,
		int index,
		IFunction valueFunction)
		throws SimulationRuntimeException {

		IBody fromBody = fromObject.getBody();

		if (Assert.ASSERTION_ON && !fromBody.isBodyVariableInitialized(index))
			throw new BodyVariableNotInitialized(fromBody, index);

		IBodyVariable bodyVariable = fromBody.getBodyVariableAt(index);

		if (objectExecuteFunctionArgs.hasObjectsVariableListener()) {
			IObject settingAgent = objectExecuteFunctionArgs.getObject();
			Object oldValue = bodyVariable.getValue();

			setValueWithoutEvent(valueType, objectExecuteFunctionArgs, valueFunction, bodyVariable);

			Object newValue = bodyVariable.getValue();

			objectExecuteFunctionArgs.fireVariableValueChanged(
				fromObject,
				settingAgent,
				bodyVariable,
				oldValue,
				newValue);

		} else {

			setValueWithoutEvent(valueType, objectExecuteFunctionArgs, valueFunction, bodyVariable);
		}
	}

	private static void setValueWithoutEvent(
		int valueType,
		IObjectExecuteFunctionArgs objectExecuteFunctionArgs,
		IFunction valueFunction,
		IBodyVariable bodyVariable)
		throws SimulationRuntimeException {

		switch (valueType) {
			case OBJECT_TYPE :
				{
					bodyVariable.setValue(valueFunction.execute(objectExecuteFunctionArgs));
					break;
				}

			case DOUBLE_TYPE :
				{
					bodyVariable.setDoubleValue(valueFunction.executeForDouble(objectExecuteFunctionArgs));
					break;
				}

			case INTEGER_TYPE :
				{
					bodyVariable.setIntValue(valueFunction.executeForInt(objectExecuteFunctionArgs));
					break;
				}

			case BOOLEAN_TYPE :
				{
					bodyVariable.setBooleanValue(valueFunction.executeForBoolean(objectExecuteFunctionArgs));
					break;
				}

			case ITERATOR_TYPE :
				{
					bodyVariable.setValue(
						ExecuteIteratorFactory.createList(valueFunction.execute(objectExecuteFunctionArgs)));
					break;
				}

			default :
				Assert.fail();
		}
	}
}