package de.uniwue.ki.mas.basics.functionArgument;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.functionCall.*;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.TypeManager;
import de.uniwue.ki.mas.basics.type.argument.IArgument;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.AbstractDocumentableDeclaration;
import de.uniwue.ki.util.IRemoveListener;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (30.07.2001 14:17:41)
 * @author Christoph Oechslein
 */

class FunctionArgumentDeclaration
	extends AbstractDocumentableDeclaration
	implements IFunctionArgumentDeclaration {

	private final IFunctionArgument _functionArgument = new FunctionArgument(this);
	private ISpecificType _specificType;
	protected final List _referencedAtomicFunInternCalls = new ArrayList();

	private static boolean _insertReferences = true;

	public FunctionArgumentDeclaration(ISpecificType specificType) {
		super();

		_specificType = specificType;
	}

	public FunctionArgumentDeclaration(String name, ISpecificType specificType) {
		super();

		setShortDescription(name);
		setUsesShortDescription(true);

		_specificType = specificType;
	}

	private transient final IRemoveListener _removeListener = new IRemoveListener() {
		public void removed(IRemoveable removeable) {
			_referencedAtomicFunInternCalls.remove(removeable);
		}
	};

	private void addReferencedAtomicFunInternCallDecl(IAtomicFunctionCallInternDeclaration atomicCall) {
		if (isInsertReferences() && !_referencedAtomicFunInternCalls.contains(atomicCall)) {
			_referencedAtomicFunInternCalls.add(atomicCall);
			atomicCall.addRemoveListener(_removeListener);
		}
	}

	public boolean canBeRemoved() {
		if (!removingOkForListeners())
			return false;
		else
			return _referencedAtomicFunInternCalls.isEmpty();
	}

	public boolean canChangeSpecificTypeTo(ISpecificType newSpecificType) {
		Assert.assertNotNull(newSpecificType);

		Iterator it = _referencedAtomicFunInternCalls.iterator();
		while (it.hasNext()) {
			if (!canChangeSpecificTypeTo(newSpecificType,
				((IAtomicFunctionCallInternDeclaration)it.next()).getParent()))
				return false;
		}

		return true;
	}

	private boolean canChangeSpecificTypeTo(
		ISpecificType newSpecificType,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration) {

		if (composedFunctionCallInternDeclaration == null)
			return true;

		Iterator it = composedFunctionCallInternDeclaration.getChildren();
		int index = 0;
		while (it.hasNext()) {
			IFunctionCallInternDeclaration functionCallInternDeclaration =
				(IFunctionCallInternDeclaration)it.next();
			if (functionCallInternDeclaration instanceof IAtomicFunctionCallInternDeclaration) {
				if (((IAtomicFunctionCallInternDeclaration)functionCallInternDeclaration)
					.getArgumentDeclaration()
					== this) {
					ISpecificType neededType =
						composedFunctionCallInternDeclaration
							.getSpecificFunctionDeclaration()
							.getSpecificInputTypeAt(
							index);
					if (!TypeManager.getInstance().isMoreGeneral(neededType, newSpecificType))
						return false;
				}
			}
			index++;
		}

		return true;
	}

	public void copyInto(Object intoObject, Map objectMapping) {
		super.copyInto(intoObject, objectMapping);

		IFunctionArgumentDeclaration intoFunctionArgumentDeclaration =
			(IFunctionArgumentDeclaration)intoObject;

		intoFunctionArgumentDeclaration.setSpecificType(getSpecificType());
	}

	public IArgument createArgument(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException {
		return getArgument();
	}

	protected ICopyable createInstance(Map objectMapping) {
		if ((objectMapping != null) && (objectMapping.containsKey(this))) {
			return (ICopyable)objectMapping.get(this);

		} else {
			return this;
		}
	}

	public final IArgument getArgument() {
		return getFunctionArgument();
	}

	protected String getDefaultName() {
		return "Untitled";
	}

	public final IFunctionArgument getFunctionArgument() {
		return _functionArgument;
	}

	public String getName() {
		return getUsesShortDescription() ? getShortDescription() : getDefaultName();
	}

	public Object getObject() {
		throw new UnsupportedOperationException();
		//		if (getSpecificType().isAtomicDefaultValue())
		//			return getSpecificType().getArgumentDeclarationDefaultValue();
		//		else
		//			return null;
	}

	public int getReferencedAtomicFunInternCallDeclCount() {
		return _referencedAtomicFunInternCalls.size();
	}

	public ISpecificType getSpecificType() {
		return _specificType;
	}

	public boolean isConstant(ICreateFunctionArgs createFunctionArgs) {
		return false;
	}

	public boolean isomorphicEquals(IIsomorphicHashable obj) {
		return equals(obj);
	}

	public int isomorphicHasCode() {
		return hashCode();
	}

	private void removeAllReferencedAtomicFunInternCallDecl() {
		Iterator it = _referencedAtomicFunInternCalls.iterator();

		while (it.hasNext()) {
			IAtomicFunctionCallInternDeclaration element = (IAtomicFunctionCallInternDeclaration)it.next();
			element.removeRemoveListener(_removeListener);
			it.remove();
		}
	}

	public void setSpecificType(ISpecificType newSpecificType) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(canChangeSpecificTypeTo(newSpecificType));
		Assert.assertNotNull(newSpecificType);

		ISpecificType oldSpecificType = getSpecificType();
		_specificType = newSpecificType;
		firePropertyChange("specificType", oldSpecificType, newSpecificType);
		fireStateChanged();
	}

	public void setObject(Object anObject) {
		throw new UnsupportedOperationException();
	}

	public void remove() {
		if (canBeRemoved()) {
			removeAllReferencedAtomicFunInternCallDecl();

			super.remove();
		}
	}

	public IAtomicFunctionCallInternDeclaration createAtomicFunctionCallInternDeclaration() {
		IAtomicFunctionCallInternDeclaration newAtomicFunctionCallInternDeclaration =
			AtomicFunctionCallInternDeclarationFactory.createAtomicFunctionCallInternDeclaration(this);

		addReferencedAtomicFunInternCallDecl(newAtomicFunctionCallInternDeclaration);

		return newAtomicFunctionCallInternDeclaration;
	}

	static void setInsertReferences(boolean insertReferences) {
		_insertReferences = insertReferences;
	}

	static boolean isInsertReferences() {
		return _insertReferences;
	}
}
