package de.uniwue.ki.mas.basics.functionArgument;
import gnu.trove.THashSet;
import gnu.trove.TObjectBooleanHashMap;

import java.io.Serializable;
import java.util.*;
import java.util.logging.Level;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.functionCall.*;
import de.uniwue.ki.mas.basics.functionDeclaration.*;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.util.IRemoveable;

/**
* Insert the type's description here.
* Creation date: (12.06.2001 10:56:18)
* @author Christoph Oechslein
*/

public final class ParametricFunctionCallDeclarationCompiler implements IRemoveable, Serializable {

	/* 
	 * If turned on compilation may introduce errors:
	 * - Arguments which are unused are removed, but could change something (like pop, progn)
	 * - Inline Arguments could change order of read, write operationen (SetVar(x,1) before GetVar(x))
	 */
	static final boolean AGGRESSIVE_INLINING_AND_REMOVAL = false;

	private final boolean _inlineDuringConstantOn = true;
	private final boolean _inlineOn;

	private final Map _currentCompiledParametricFunCallDecls = new HashMap();
	private final THashSet _constantProcessingSet = new THashSet();

	private final TObjectBooleanHashMap _recursiveMap = new TObjectBooleanHashMap();

	private final ParametricFunctionCallInliner _parametricFunctionCallInliner;

	public ParametricFunctionCallDeclarationCompiler(boolean inlineOn) {
		_parametricFunctionCallInliner = new ParametricFunctionCallInliner(this);

		_inlineOn = inlineOn;
	}

	public IParametricFunction createParametricFunction(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration)
		throws ModelCompilingException {

		if (isRecursive(parametricFunctionCallDeclaration)) {
			if (createFunctionArgs.existsCompilationObject(parametricFunctionCallDeclaration, false)) {

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Retrieve cached recursive function definition for: '"
							+ parametricFunctionCallDeclaration.getName()
							+ "'");
				}

				return (IParametricFunction) createFunctionArgs.getCompilationObject(
					parametricFunctionCallDeclaration,
					false);

			} else {
				IParametricFunctionDeferred parametricFunction =
					createParametricFunctionDeferred(createFunctionArgs, parametricFunctionCallDeclaration);

				createFunctionArgs.addCompilationObject(
					parametricFunctionCallDeclaration,
					parametricFunction,
					false);

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Use special case for recursive function: '"
							+ parametricFunctionCallDeclaration.getName()
							+ "'");
				}

				IFunction function =
					parametricFunctionCallDeclaration.getFunctionCallDeclaration().createFunction(
						createFunctionArgs);
				parametricFunction.setRealFunction(function);

				return parametricFunction;
			}
		}

		if (_currentCompiledParametricFunCallDecls.containsKey(parametricFunctionCallDeclaration)) {
			// so this is a recursive parametric function call
			// but we can handle this
			IParametricFunctionDeferred parametricFunction =
				createParametricFunctionDeferred(createFunctionArgs, parametricFunctionCallDeclaration);

			List list = (List) _currentCompiledParametricFunCallDecls.get(parametricFunctionCallDeclaration);
			if (list == null) {
				list = new LinkedList();
				_currentCompiledParametricFunCallDecls.put(parametricFunctionCallDeclaration, list);
			}
			list.add(parametricFunction);

			return parametricFunction;

		} else {
			_currentCompiledParametricFunCallDecls.put(parametricFunctionCallDeclaration, null);

			IParametricFunction result =
				createParametricFunctionAndAddToRecursiveInvokations(
					createFunctionArgs,
					parametricFunctionCallDeclaration,
					parametricFunctionCallDeclaration.getFunctionCallDeclaration().createFunction(
						createFunctionArgs));

			return result;
		}
	}

	private IParametricFunction createParametricFunctionAndAddToRecursiveInvokations(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IFunction realFunction) {

		IParametricFunction parametricFunction;
		if (createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()
			&& hasOnlyUnusedyArguments(parametricFunctionCallDeclaration)) {

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Use special case for null arg parametric function: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "'");
			}

			parametricFunction =
				new NullArgumentsParametricFunction(parametricFunctionCallDeclaration, realFunction);

		} else if (realFunction instanceof IFunctionArgument) {

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Use special case for function argument only parametric function: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "'");
			}

			parametricFunction =
				new ParametricFunctionWithFunctionArgument(
					parametricFunctionCallDeclaration.getName(),
					(IFunctionArgument) realFunction);

		} else if (!_inlineOn || !parametricFunctionCallDeclaration.isInlinePossible()) {

			if (createFunctionArgs.existsCompilationObject(parametricFunctionCallDeclaration, false)) {
				parametricFunction =
					(IParametricFunction) createFunctionArgs.getCompilationObject(
						parametricFunctionCallDeclaration,
						false);

			} else {

				if (Assert.ASSERTION_ON)
					Assert.assertTrue(
						!createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()
							|| (parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().size()
								> 0));

				parametricFunction =
					new ParametricFunctionReuseable(
						parametricFunctionCallDeclaration.getName(),
						parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList(),
						realFunction);

				createFunctionArgs.addCompilationObject(
					parametricFunctionCallDeclaration,
					parametricFunction,
					false);
			}

		} else {
			parametricFunction =
				new ParametricFunctionNotReuseable(
					createFunctionArgs,
					parametricFunctionCallDeclaration.getName(),
					parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList(),
					realFunction);
		}

		addCreatedFunctionToRecursiveInvokations(parametricFunctionCallDeclaration, realFunction);

		return parametricFunction;
	}

	private void addCreatedFunctionToRecursiveInvokations(
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IFunction realFunction) {

		List list = (List) _currentCompiledParametricFunCallDecls.remove(parametricFunctionCallDeclaration);

		if (list != null) {
			Iterator it = list.iterator();

			while (it.hasNext()) {
				IParametricFunctionDeferred parametricFunctionDefered =
					(IParametricFunctionDeferred) it.next();
				parametricFunctionDefered.setRealFunction(realFunction);
				it.remove();
			}
		}
	}

	private IParametricFunctionDeferred createParametricFunctionDeferred(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration) {

		if (createFunctionArgs.existsCompilationObject(parametricFunctionCallDeclaration, false)) {
			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Use cached special case for recursive call to: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "'");
			}

			return (IParametricFunctionDeferred) createFunctionArgs.getCompilationObject(
				parametricFunctionCallDeclaration,
				false);

		} else if (
			createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()
				&& hasOnlyUnusedyArguments(parametricFunctionCallDeclaration)) {

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Use special case for null arg parametric function: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "'");
			}

			return new NullArgumentsParametricFunction(parametricFunctionCallDeclaration);

		} else {
			IParametricFunctionDeferred parametricFunction =
				new ParametricFunctionDeferred(parametricFunctionCallDeclaration);

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Use special case for recursive call to: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "'");
			}

			createFunctionArgs.addCompilationObject(
				parametricFunctionCallDeclaration,
				parametricFunction,
				false);

			return parametricFunction;
		}
	}

	public IFunction createFunction(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IFunctionList functions)
		throws ModelCompilingException {

		if (createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()
			&& hasOnlyUnusedyArguments(parametricFunctionCallDeclaration)) {
			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO)) {
				FunctionDeclarationCompiler.LOGGER.info(
					"Inlined: '"
						+ parametricFunctionCallDeclaration.getName()
						+ "' to '"
						+ parametricFunctionCallDeclaration.getFunctionCallDeclaration().getName()
						+ "'");
			}

			return parametricFunctionCallDeclaration.getFunctionCallDeclaration().createFunction(
				createFunctionArgs);

		} else {
			return createParametricFunction(
				createFunctionArgs,
				parametricFunctionCallDeclaration).createFunction(
				createFunctionArgs,
				functions);
		}
	}

	boolean hasOnlyUnusedyArguments(IParametricFunctionCallDeclaration parametricFunctionCallDeclaration) {
		int size = parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().getSize();

		if (size == 0)
			return true;
		else if (!AGGRESSIVE_INLINING_AND_REMOVAL)
			return false;
		else
			return size == countUnusedArguments(parametricFunctionCallDeclaration);
	}

	private final THashSet _currentReadOnlyProcessing = new THashSet();

	public boolean isReadOnly(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (!createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn())
			return false;

		if (_currentReadOnlyProcessing.contains(parametricFunctionCallDeclaration)) {
			return true;

		} else {
			_currentReadOnlyProcessing.add(parametricFunctionCallDeclaration);

			boolean result =
				parametricFunctionCallDeclaration.getFunctionCallDeclaration().isReadOnly(createFunctionArgs)
					&& AbstractSpecificFunctionDeclaration.isReadOnlyStatic(
						createFunctionArgs,
						composedFunctionCallInternDeclaration);

			_currentReadOnlyProcessing.remove(parametricFunctionCallDeclaration);

			return result;
		}
	}

	public boolean isConstant(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (!composedFunctionCallInternDeclaration.isReadOnly(createFunctionArgs)) {
			return false;
		}

		if (!_inlineDuringConstantOn
			|| !parametricFunctionCallDeclaration.isInlinePossible()
			|| isRecursive(parametricFunctionCallDeclaration)) {

			return isConstantOld(
				createFunctionArgs,
				parametricFunctionCallDeclaration,
				composedFunctionCallInternDeclaration);
		}

		Assert.assertNotNull(composedFunctionCallInternDeclaration);

		if (_constantProcessingSet.contains(parametricFunctionCallDeclaration)) {
			// only test arguments 
			return areArgumentValuesConstant(createFunctionArgs, composedFunctionCallInternDeclaration);

		} else {
			_constantProcessingSet.add(parametricFunctionCallDeclaration);

			// there are a few problems with the inlining approach:
			// - what if this is a recursive user function
			// - what if an argument is used mre than one time
			//   -> first occurance is inlined, so the parent could be changed of some of it children
			//   example: inlineTime (inlineTime (GetVar(a), 60))

			// first issue is solved above
			// next issue is solved by inlining a copy

			//return isConstantOld(createFunctionArgs, parametricFunctionCallDeclaration, composedFunctionCallInternDeclaration);

			boolean result =
				isConstantInlined(
					createFunctionArgs,
					parametricFunctionCallDeclaration,
					composedFunctionCallInternDeclaration);

			_constantProcessingSet.remove(parametricFunctionCallDeclaration);

			return result;
		}
	}

	private boolean _inConstantOld = false;

	private boolean isConstantInlined(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (_inConstantOld) {
			return isConstantOld(
				createFunctionArgs,
				parametricFunctionCallDeclaration,
				composedFunctionCallInternDeclaration);

		} else if (!_inlineOn || !parametricFunctionCallDeclaration.isInlinePossible()) {

			try {
				_inConstantOld = true;

				return isConstantOld(
					createFunctionArgs,
					parametricFunctionCallDeclaration,
					composedFunctionCallInternDeclaration);
			} finally {
				_inConstantOld = false;
			}

		} else {
			// instead test if a inlined (forced to inline all arguments) version is constant 
			IFunctionCallInternDeclaration inlinedFunctionCallInternDeclaration =
				_parametricFunctionCallInliner.inlineConstantFunctionCallInternDeclaration(
					createFunctionArgs,
					parametricFunctionCallDeclaration,
					composedFunctionCallInternDeclaration,
					true);

			try {
				_inConstantOld = true;
				if (inlinedFunctionCallInternDeclaration != composedFunctionCallInternDeclaration) {
					return inlinedFunctionCallInternDeclaration.isConstant(createFunctionArgs);

				} else {
					return isConstantOld(
						createFunctionArgs,
						parametricFunctionCallDeclaration,
						composedFunctionCallInternDeclaration);
				}
			} finally {
				_inConstantOld = false;
			}
		}
	}

	private final THashSet _currentConstantProcessing = new THashSet();

	private final boolean isConstantOld(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (_currentConstantProcessing.contains(parametricFunctionCallDeclaration)) {
			return true;

		} else {
			_currentConstantProcessing.add(parametricFunctionCallDeclaration);

			// I dont' know why the parametric function call should be constant AND
			// the specific invokation of the parametric function call ??
			// -> the arguments must be constant and the definition of the parametric function
			//    (normally only under the given variable bindings)

			boolean result =
				parametricFunctionCallDeclaration.getFunctionCallDeclaration().isConstant(createFunctionArgs)
					&& areArgumentValuesConstant(createFunctionArgs, composedFunctionCallInternDeclaration);

			_currentConstantProcessing.remove(parametricFunctionCallDeclaration);

			return result;
		}
	}

	private boolean areArgumentValuesConstant(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		Iterator functionCallInternDeclarations = composedFunctionCallInternDeclaration.getChildren();
		while (functionCallInternDeclarations.hasNext()) {
			if (!((IFunctionCallInternDeclaration) functionCallInternDeclarations.next())
				.isConstant(createFunctionArgs)) {
				return false;
			}
		}

		return true;
	}

	public IFunction createFunction(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		if (!createFunctionArgs.getFunctionDeclarationCompiler().isCompileOn()
			|| !_inlineOn
			|| isRecursive(parametricFunctionCallDeclaration)
			|| !parametricFunctionCallDeclaration.isInlinePossible()) {
			return AbstractParametricFunction.createFunctionStatic(
				createFunctionArgs,
				parametricFunctionCallDeclaration.createParametricFunction(createFunctionArgs),
				composedFunctionCallInternDeclaration);
		}

		IFunctionCallInternDeclaration newFunctionCallInternDeclaration =
			_parametricFunctionCallInliner.inlineConstantFunctionCallInternDeclaration(
				createFunctionArgs,
				parametricFunctionCallDeclaration,
				composedFunctionCallInternDeclaration,
				false);

		if (newFunctionCallInternDeclaration == composedFunctionCallInternDeclaration) {
			return AbstractParametricFunction.createFunctionStatic(
				createFunctionArgs,
				parametricFunctionCallDeclaration.createParametricFunction(createFunctionArgs),
				composedFunctionCallInternDeclaration);
		} else {
			return newFunctionCallInternDeclaration.createFunction(createFunctionArgs);
		}
	}

	public boolean isRecursive(IParametricFunctionCallDeclaration parametricFunctionCallDeclaration) {
		return isRecursive(new THashSet(), parametricFunctionCallDeclaration);
	}

	private boolean isRecursive(
		THashSet paramFunctions,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration) {

		if (paramFunctions.contains(parametricFunctionCallDeclaration)) {
			return true;

		} else if (_recursiveMap.containsKey(parametricFunctionCallDeclaration)) {
			return _recursiveMap.get(parametricFunctionCallDeclaration);

		} else {
			paramFunctions.add(parametricFunctionCallDeclaration);

			boolean recursive =
				isRecursiveIntern(
					paramFunctions,
					parametricFunctionCallDeclaration
						.getFunctionCallDeclaration()
						.getFunctionCallInternDeclaration());

			paramFunctions.remove(parametricFunctionCallDeclaration);

			_recursiveMap.put(parametricFunctionCallDeclaration, recursive);

			return recursive;
		}
	}

	private boolean isRecursiveIntern(
		THashSet paramFunctions,
		IFunctionCallInternDeclaration functionCallInternDeclaration) {

		if (functionCallInternDeclaration instanceof IComposedFunctionCallInternDeclaration) {
			IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration =
				(IComposedFunctionCallInternDeclaration) functionCallInternDeclaration;

			ISpecificFunctionDeclaration specificFunctionDeclaration =
				composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration();

			if (specificFunctionDeclaration instanceof ISpecificFunctionDeclarationWithArgDeclBinds) {
				if (isRecursiveIntern(paramFunctions,
					(ISpecificFunctionDeclarationWithArgDeclBinds) specificFunctionDeclaration))
					return true;
			}

			for (int i = 0; i < composedFunctionCallInternDeclaration.getChildCount(); i++) {
				if (isRecursiveIntern(paramFunctions, composedFunctionCallInternDeclaration.getChild(i)))
					return true;
			}

			return false;

		} else {
			IAtomicFunctionCallInternDeclaration atomicFunCallDecl =
				(IAtomicFunctionCallInternDeclaration) functionCallInternDeclaration;

			if (!(atomicFunCallDecl.getArgumentDeclaration() instanceof IFunctionArgumentDeclaration)
				&& (atomicFunCallDecl.getArgumentDeclaration().getObject()
					instanceof ISpecificFunctionDeclarationWithArgDeclBinds)) {
				return isRecursiveIntern(
					paramFunctions,
					(ISpecificFunctionDeclarationWithArgDeclBinds) atomicFunCallDecl
						.getArgumentDeclaration()
						.getObject());

			} else {
				return false;
			}
		}
	}

	private boolean isRecursiveIntern(
		THashSet paramFunctions,
		ISpecificFunctionDeclarationWithArgDeclBinds argDeclBinds) {

		return isRecursive(paramFunctions, argDeclBinds.getParametricFunctionCallDeclaration());
	}

	private boolean isUnusedArgument(IFunctionArgumentDeclaration functionArgumentDeclaration) {
		return functionArgumentDeclaration.canBeRemoved();
	}

	private int countUnusedArguments(IParametricFunctionCallDeclaration parametricFunctionCallDeclaration) {
		int count = 0;

		final boolean shouldLog =
			Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO);

		final List unusedArgumentList = shouldLog ? new LinkedList() : null;

		Iterator it = parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().iterator();
		while (it.hasNext()) {
			IFunctionArgumentDeclaration currFunctionArgumentDeclaration =
				(IFunctionArgumentDeclaration) it.next();

			if (isUnusedArgument(currFunctionArgumentDeclaration)) {
				count++;
				if (shouldLog) {
					unusedArgumentList.add(currFunctionArgumentDeclaration);
				}
			}
		}
		if (shouldLog && (count > 0)) {
			StringBuffer buff = new StringBuffer();
			buff.append("Detected ");
			buff.append(count);
			buff.append(" unused arguments [");

			Iterator unusedArguments = unusedArgumentList.iterator();
			while (unusedArguments.hasNext()) {
				IFunctionArgumentDeclaration currFunctionArgumentDeclaration =
					(IFunctionArgumentDeclaration) unusedArguments.next();

				buff.append(currFunctionArgumentDeclaration.getName());

				if (unusedArguments.hasNext()) {
					buff.append(", ");
				}
			}

			buff.append("] in ");
			buff.append(parametricFunctionCallDeclaration.getName());

			FunctionDeclarationCompiler.LOGGER.info(buff.toString());
		}

		return count;
	}

	public void remove() {
		_parametricFunctionCallInliner.remove();
		_currentCompiledParametricFunCallDecls.clear();
		_recursiveMap.clear();
	}
}