package de.uniwue.ki.mas.basics.functionArgument;

import gnu.trove.THashMap;

import java.io.Serializable;
import java.util.*;
import java.util.logging.Level;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.functionCall.*;
import de.uniwue.ki.mas.basics.functionDeclaration.FunctionDeclarationCompiler;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.basics.type.argument.IArgumentDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.IDeclaration;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.hash.IsomorphicHashTableWrapper;
import de.uniwue.ki.util.name.AbstractNamed;

/**
 * @author oechslein
 * 
 * Creation date: 16.05.2003
 */
final class ParametricFunctionCallInliner implements Serializable, IRemoveable {

	private final ParametricFunctionCallDeclarationCompiler _parametricFunCallDeclCompiler;

	private Map _inlinedMap = new IsomorphicHashTableWrapper(new HashMap());
	private Map _forcedInlinedMap = new IsomorphicHashTableWrapper(new HashMap());

	private final List _createdDeclarationsDuringInlining = new LinkedList();

	private static final class InlineArgument extends AbstractNamed implements Serializable {
		int _count;
		final IFunctionArgumentDeclaration _functionArgumentDeclaration;
		final IFunctionCallInternDeclaration _argumentFunCallDecl;

		public InlineArgument(
			IFunctionArgumentDeclaration functionArgumentDeclaration,
			IFunctionCallInternDeclaration argumentFunCallDecl) {
			super();

			_functionArgumentDeclaration = functionArgumentDeclaration;
			_argumentFunCallDecl = argumentFunCallDecl;
			_count = 0;
		}

		public String getName() {
			return _functionArgumentDeclaration.getName() + "=" + _argumentFunCallDecl.getName();
		}
	}

	private static final class InliningArguments extends AbstractNamed implements Serializable {
		final boolean _forceInlining;
		final ICreateFunctionArgs _createFunctionArgs;
		final Map _objectMap = new HashMap();
		final THashMap _inlineArguments = new THashMap();
		boolean _inlined = false;

		public InliningArguments(ICreateFunctionArgs createFunctionArgs, boolean forceInlining) {
			super();

			_createFunctionArgs = createFunctionArgs;
			_forceInlining = forceInlining;
		}

		boolean addInliningArgument(
			IFunctionArgumentDeclaration functionArgumentDeclaration,
			IFunctionCallInternDeclaration argumentFunCallDecl) {

			if (!contains(functionArgumentDeclaration)) {
				_inlineArguments.put(
					functionArgumentDeclaration,
					new InlineArgument(functionArgumentDeclaration, argumentFunCallDecl));
				return true;
			} else {
				return argumentFunCallDecl.isomorphicEquals(
					getArgumentFunCallDecl(functionArgumentDeclaration));
			}
		}

		int getInliningArgumentCount() {
			return _inlineArguments.size();
		}

		void removeInliningArgument(IFunctionArgumentDeclaration functionArgumentDeclaration) {

			Assert.assertTrue(contains(functionArgumentDeclaration));
			_inlineArguments.remove(functionArgumentDeclaration);
		}

		void incrementCount(IFunctionArgumentDeclaration functionArgumentDeclaration) {
			getInlineArgument(functionArgumentDeclaration)._count++;
			_inlined = true;
		}

		int getCount(IFunctionArgumentDeclaration functionArgumentDeclaration) {
			return getInlineArgument(functionArgumentDeclaration)._count;
		}

		IFunctionCallInternDeclaration getArgumentFunCallDecl(IFunctionArgumentDeclaration functionArgumentDeclaration) {
			return getInlineArgument(functionArgumentDeclaration)._argumentFunCallDecl;
		}

		boolean contains(IFunctionArgumentDeclaration functionArgumentDeclaration) {
			return _inlineArguments.containsKey(functionArgumentDeclaration);
		}

		private InlineArgument getInlineArgument(IFunctionArgumentDeclaration functionArgumentDeclaration) {
			Assert.assertTrue(
				"Does not contain function argument " + functionArgumentDeclaration.getName(),
				contains(functionArgumentDeclaration));
			return (InlineArgument)_inlineArguments.get(functionArgumentDeclaration);
		}

		boolean isEmpty() {
			return _inlineArguments.isEmpty();
		}

		public String getName() {
			return super.getNameIntern();
		}

		public void appendName(StringBuffer buff, int currDepth, int maxDepth) {
			buff.append("inlined: ");
			buff.append(_inlined);
			buff.append(" [");
			Iterator it = _inlineArguments.values().iterator();
			while (it.hasNext()) {
				InlineArgument inlineArgument = (InlineArgument)it.next();
				inlineArgument.appendName(buff, currDepth + 1, maxDepth);
				buff.append(", ");
			}
			buff.append("]");
		}
	}

	public ParametricFunctionCallInliner(ParametricFunctionCallDeclarationCompiler parametricFunctionCallDeclarationCompiler) {
		_parametricFunCallDeclCompiler = parametricFunctionCallDeclarationCompiler;
	}

	private IParametricFunctionCallDeclaration createInlinedParametricFunctionCallDeclaration(
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		IFunctionCallInternDeclaration oldFunctionCallInternDeclaration =
			parametricFunctionCallDeclaration.getFunctionCallDeclaration().getFunctionCallInternDeclaration();

		IFunctionCallInternDeclaration newFunctionCallInternDeclaration =
			createInlinedFunctionCallInternDeclaration(
				null,
				oldFunctionCallInternDeclaration,
				inliningArguments);

		Assert.assertNotSame(oldFunctionCallInternDeclaration, newFunctionCallInternDeclaration);

		ParametricFunctionCallDeclaration newParametricFunctionCallDeclaration =
			new ParametricFunctionCallDeclaration(
				parametricFunctionCallDeclaration.getFunctionCallDeclaration().getMostGeneralOutputType());

		Iterator it = parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().iterator();
		newParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().setComposition(false);
		while (it.hasNext()) {
			IFunctionArgumentDeclaration element = (IFunctionArgumentDeclaration)it.next();
			newParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().add(element);
		}

		newParametricFunctionCallDeclaration.setFunctionCallInternDeclaration(
			newFunctionCallInternDeclaration);

		return newParametricFunctionCallDeclaration;
	}

	private IFunctionCallInternDeclaration createInlinedAtomicFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration currParentComposedFunctionCallInternDeclaration,
		IAtomicFunctionCallInternDeclaration currAtomicFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		IFunctionCallInternDeclaration newFunctionCallInternDeclaration;

		IArgumentDeclaration currArgumentDeclaration =
			currAtomicFunctionCallInternDeclaration.getArgumentDeclaration();

		if (currArgumentDeclaration instanceof IFunctionArgumentDeclaration) {
			if (inliningArguments.contains((IFunctionArgumentDeclaration)currArgumentDeclaration)) {
				IFunctionArgumentDeclaration inlineFunctionArgumentDeclaration =
					(IFunctionArgumentDeclaration)currArgumentDeclaration;

				inliningArguments.incrementCount(inlineFunctionArgumentDeclaration);

				IFunctionCallInternDeclaration inlineFunctionCallInternDeclaration =
					inliningArguments.getArgumentFunCallDecl(inlineFunctionArgumentDeclaration);

				// do not try to inline argument, 
				// e.g. inc(inc(3)) with inc: [x] (+ 1 x) would result in endless recursion
				//				newFunctionCallInternDeclaration =
				//					createInlinedFunctionCallInternDeclaration(
				//						currParentComposedFunctionCallInternDeclaration,
				//						inlineFunctionCallInternDeclaration,
				//						inliningArguments);
				//
				//				Assert.assertNotSame(newFunctionCallInternDeclaration, inlineFunctionCallInternDeclaration);

				newFunctionCallInternDeclaration =
					(IFunctionCallInternDeclaration)inlineFunctionCallInternDeclaration.copy(
						inliningArguments._objectMap);

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINER)) {
					FunctionDeclarationCompiler.LOGGER.finer(
						"Inlined Argument '"
							+ inlineFunctionArgumentDeclaration.getName()
							+ "' with '"
							+ newFunctionCallInternDeclaration.getName()
							+ "'.");
				}

			} else {
				newFunctionCallInternDeclaration =
					(IFunctionCallInternDeclaration)currAtomicFunctionCallInternDeclaration.copy(
						inliningArguments._objectMap);
			}

		} else if (
			currArgumentDeclaration.getObject() instanceof ISpecificFunctionDeclarationWithArgDeclBinds) {
			ISpecificFunctionDeclarationWithArgDeclBinds currSpecificFunctionDeclarationWithArgDeclBinds =
				(ISpecificFunctionDeclarationWithArgDeclBinds)currArgumentDeclaration.getObject();

			ParametricFunctionCallDeclaration currParametricFunctionCallDeclaration =
				(ParametricFunctionCallDeclaration)currSpecificFunctionDeclarationWithArgDeclBinds
					.getParametricFunctionCallDeclaration();

			IParametricFunctionCallDeclaration newParametricFunctionCallDeclaration =
				createInlinedParametricFunctionCallDeclaration(
					currParametricFunctionCallDeclaration,
					inliningArguments);

			ISpecificFunctionDeclarationWithArgDeclBinds newSpecificFunctionDeclarationWithArgDeclBinds =
				new SpecificFunctionDeclarationWithArgDeclBinds(newParametricFunctionCallDeclaration);

			addToCreatedDeclarationsDuringInliningList(newSpecificFunctionDeclarationWithArgDeclBinds);

			newFunctionCallInternDeclaration =
				FunctionCallDeclarationFactory.getInstance().createAtomicFunctionCallInternDeclaration(
					currParentComposedFunctionCallInternDeclaration,
					currArgumentDeclaration.getSpecificType(),
					newSpecificFunctionDeclarationWithArgDeclBinds);

		} else {
			newFunctionCallInternDeclaration =
				(IFunctionCallInternDeclaration)currAtomicFunctionCallInternDeclaration.copy(
					inliningArguments._objectMap);
		}

		newFunctionCallInternDeclaration.setParent(currParentComposedFunctionCallInternDeclaration);

		return newFunctionCallInternDeclaration;
	}

	private IFunctionCallInternDeclaration createInlinedComposedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration currParentComposedFunctionCallInternDeclaration,
		IComposedFunctionCallInternDeclaration currComposedFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		// leads to problems if enabled (incrementTime(incrementTime(...)
		if (false
			&& currComposedFunctionCallInternDeclaration.getSpecificFunctionDeclaration()
				instanceof ISpecificFunctionDeclarationWithArgDeclBinds) {
			ISpecificFunctionDeclarationWithArgDeclBinds argDeclBinds =
				(ISpecificFunctionDeclarationWithArgDeclBinds)currComposedFunctionCallInternDeclaration
					.getSpecificFunctionDeclaration();

			IFunctionCallInternDeclaration newFunctionCallDecl =
				inlineConstantFunctionCallInternDeclarationIntern(
					currComposedFunctionCallInternDeclaration,
					argDeclBinds.getParametricFunctionCallDeclaration(),
					inliningArguments);

			if (newFunctionCallDecl == currComposedFunctionCallInternDeclaration) {

				IFunctionCallInternDeclaration newParametricFunCallDecl =
					(IFunctionCallInternDeclaration)argDeclBinds
						.getParametricFunctionCallDeclaration()
						.getFunctionCallDeclaration()
						.getFunctionCallInternDeclaration()
						.copy(inliningArguments._objectMap);
				newParametricFunCallDecl.setParent(null);
				IParametricFunctionCallDeclaration newParametricFunctionCallDeclaration =
					new ParametricFunctionCallDeclaration(
						argDeclBinds
							.getParametricFunctionCallDeclaration()
							.getFunctionCallDeclaration()
							.getMostGeneralOutputType(),
						newParametricFunCallDecl);

				newFunctionCallDecl =
					createNewComposedFunctionCallInternDeclaration(
						argDeclBinds.getParametricFunctionCallDeclaration(),
						currComposedFunctionCallInternDeclaration,
						newParametricFunctionCallDeclaration,
						inliningArguments);
			}

			newFunctionCallDecl.setParent(currParentComposedFunctionCallInternDeclaration);

			return newFunctionCallDecl;

		} else {
			IComposedFunctionCallInternDeclaration newComposedFunctionCallInternDeclaration =
				FunctionCallDeclarationFactory.getInstance().createComposedFunctionCallInternDeclaration(
					currParentComposedFunctionCallInternDeclaration,
					currComposedFunctionCallInternDeclaration.getSpecificFunctionDeclaration());

			for (int i = 0; i < currComposedFunctionCallInternDeclaration.getChildCount(); i++) {
				IFunctionCallInternDeclaration newChild =
					createInlinedFunctionCallInternDeclaration(
						newComposedFunctionCallInternDeclaration,
						currComposedFunctionCallInternDeclaration.getChild(i),
						inliningArguments);

				newComposedFunctionCallInternDeclaration.setChild(i, newChild);
			}

			int tooManyChilds =
				newComposedFunctionCallInternDeclaration.getChildCount()
					- currComposedFunctionCallInternDeclaration.getChildCount();

			for (int i = 0; i < tooManyChilds; i++) {
				newComposedFunctionCallInternDeclaration.removeChild(
					newComposedFunctionCallInternDeclaration.getChildCount() - 1);
			}

			if (Assert.ASSERTION_ON)
				newComposedFunctionCallInternDeclaration.checkInvariants();

			return newComposedFunctionCallInternDeclaration;
		}
	}

	private IFunctionCallInternDeclaration createInlinedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration currParentComposedFunctionCallInternDeclaration,
		IFunctionCallInternDeclaration currFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		IFunctionCallInternDeclaration result;

		if (currFunctionCallInternDeclaration instanceof IComposedFunctionCallInternDeclaration) {
			result =
				createInlinedComposedFunctionCallInternDeclaration(
					currParentComposedFunctionCallInternDeclaration,
					(IComposedFunctionCallInternDeclaration)currFunctionCallInternDeclaration,
					inliningArguments);

		} else {
			result =
				createInlinedAtomicFunctionCallInternDeclaration(
					currParentComposedFunctionCallInternDeclaration,
					(IAtomicFunctionCallInternDeclaration)currFunctionCallInternDeclaration,
					inliningArguments);
		}

		Assert.assertNotSame(result, currFunctionCallInternDeclaration);
		Assert.assertSame(currParentComposedFunctionCallInternDeclaration, result.getParent());

		return result;
	}

	IFunctionCallInternDeclaration inlineConstantFunctionCallInternDeclaration(
		ICreateFunctionArgs createFunctionArgs,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration,
		boolean forceInlining)
		throws ModelCompilingException {

		Assert.assertTrue(parametricFunctionCallDeclaration.isInlinePossible());

		if (_parametricFunCallDeclCompiler.isRecursive(parametricFunctionCallDeclaration)) {
			return composedFunctionCallInternDeclaration;

		} else {
			Assert.assertSame(
				((ISpecificFunctionDeclarationWithArgDeclBinds)composedFunctionCallInternDeclaration
					.getSpecificFunctionDeclaration())
					.getParametricFunctionCallDeclaration(),
				parametricFunctionCallDeclaration);

			Map usedMap = forceInlining ? _forcedInlinedMap : _inlinedMap;

			if (usedMap.containsKey(composedFunctionCallInternDeclaration)) {
				return (IFunctionCallInternDeclaration)usedMap.get(composedFunctionCallInternDeclaration);

			} else {
				InliningArguments inlineArgs = new InliningArguments(createFunctionArgs, forceInlining);
				IFunctionCallInternDeclaration inlinedFunCallInternDecl =
					inlineConstantFunctionCallInternDeclaration(
						composedFunctionCallInternDeclaration,
						inlineArgs);

				Assert.assertTrue(inlineArgs.isEmpty());

				if (forceInlining && (inlinedFunCallInternDecl == composedFunctionCallInternDeclaration)) {
					_forcedInlinedMap.put(composedFunctionCallInternDeclaration, inlinedFunCallInternDecl);
					_inlinedMap.put(composedFunctionCallInternDeclaration, inlinedFunCallInternDecl);
				} else {
					usedMap.put(composedFunctionCallInternDeclaration, inlinedFunCallInternDecl);
				}

				return inlinedFunCallInternDecl;
			}
		}
	}

	private IFunctionCallInternDeclaration inlineConstantFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		boolean oldInsertReferences = FunctionArgumentDeclaration.isInsertReferences();
		FunctionArgumentDeclaration.setInsertReferences(false);
		try {
			IFunctionCallInternDeclaration newFunctionCallInternDeclaration =
				inlineConstantFunctionCallInternDeclarationIntern(
					composedFunctionCallInternDeclaration,
					inliningArguments);

			if (newFunctionCallInternDeclaration == null) {
				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Aborted inlining of: '" + composedFunctionCallInternDeclaration.getName() + ".");
				}
				return composedFunctionCallInternDeclaration;

			} else if (newFunctionCallInternDeclaration != composedFunctionCallInternDeclaration) {
				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO)) {
					FunctionDeclarationCompiler.LOGGER.info(
						"Inlined: '"
							+ composedFunctionCallInternDeclaration.getName()
							+ "' to '"
							+ newFunctionCallInternDeclaration.getName()
							+ "'");
				}

				newFunctionCallInternDeclaration.setParent(composedFunctionCallInternDeclaration.getParent());
			}

			if (Assert.ASSERTION_ON) {
				composedFunctionCallInternDeclaration.checkInvariants();
				newFunctionCallInternDeclaration.checkInvariants();
			}

			return newFunctionCallInternDeclaration;

		} finally {
			FunctionArgumentDeclaration.setInsertReferences(oldInsertReferences);
		}
	}

	private IFunctionCallInternDeclaration inlineConstantFunctionCallInternDeclarationIntern(
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration =
			((ISpecificFunctionDeclarationWithArgDeclBinds)composedFunctionCallInternDeclaration
				.getSpecificFunctionDeclaration())
				.getParametricFunctionCallDeclaration();

		if (ParametricFunctionCallDeclarationCompiler.AGGRESSIVE_INLINING_AND_REMOVAL
			&& _parametricFunCallDeclCompiler.hasOnlyUnusedyArguments(parametricFunctionCallDeclaration)) {
			IFunctionCallInternDeclaration copyOfFunCallDecl =
				(IFunctionCallInternDeclaration)parametricFunctionCallDeclaration
					.getFunctionCallDeclaration()
					.getFunctionCallInternDeclaration()
					.copy(inliningArguments._objectMap);

			addToCreatedDeclarationsDuringInliningList(copyOfFunCallDecl);

			return copyOfFunCallDecl;

		} else {
			Assert.assertEquals(
				composedFunctionCallInternDeclaration.getChildCount(),
				parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().size());

			IFunctionCallInternDeclaration newFunctionCallDecl =
				inlineConstantFunctionCallInternDeclarationIntern(
					composedFunctionCallInternDeclaration,
					parametricFunctionCallDeclaration,
					inliningArguments);

			if ((newFunctionCallDecl == null)
				|| (newFunctionCallDecl == composedFunctionCallInternDeclaration)) {
				Assert.assertTrue(!inliningArguments._inlined);

				return newFunctionCallDecl;

			} else if (!inliningArguments._inlined) {

				newFunctionCallDecl.remove();

				return composedFunctionCallInternDeclaration;

			} else {

				addToCreatedDeclarationsDuringInliningList(newFunctionCallDecl);

				if ((newFunctionCallDecl instanceof IComposedFunctionCallInternDeclaration)
					&& (((IComposedFunctionCallInternDeclaration)newFunctionCallDecl)
						.getSpecificFunctionDeclaration()
						instanceof ISpecificFunctionDeclarationWithArgDeclBinds)) {

					IComposedFunctionCallInternDeclaration newComposedFunctionCallDecl =
						(IComposedFunctionCallInternDeclaration)newFunctionCallDecl;

					Assert.assertNotSame(
						newComposedFunctionCallDecl.getSpecificFunctionDeclaration(),
						composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration());

					// recursive Inlining
					inliningArguments._inlined = false;

					return inlineConstantFunctionCallInternDeclarationIntern(
						newComposedFunctionCallDecl,
						inliningArguments);

				} else {
					return newFunctionCallDecl;
				}
			}
		}
	}

	private IFunctionCallInternDeclaration inlineConstantFunctionCallInternDeclarationIntern(
		IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration,
		IParametricFunctionCallDeclaration oldParametricFunctionCallDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		if (Assert.ASSERTION_ON)
			oldComposedFunctionCallInternDeclaration.checkInvariants();

		int oldInliningArgumentCount = inliningArguments.getInliningArgumentCount();

		boolean shouldInline =
			addInliningArguments(
				oldComposedFunctionCallInternDeclaration,
				oldParametricFunctionCallDeclaration,
				inliningArguments);

		if (shouldInline) {

			IParametricFunctionCallDeclaration newParametricFunctionCallDeclaration =
				createInlinedParametricFunctionCallDeclaration(
					oldParametricFunctionCallDeclaration,
					inliningArguments);

			Assert.assertNotSame(oldParametricFunctionCallDeclaration, newParametricFunctionCallDeclaration);

			IFunctionCallInternDeclaration newFunctionCallInternDeclaration =
				createNewComposedFunctionCallInternDeclaration(
					oldParametricFunctionCallDeclaration,
					oldComposedFunctionCallInternDeclaration,
					newParametricFunctionCallDeclaration,
					inliningArguments);

			Assert.assertNotNull(newFunctionCallInternDeclaration);
			Assert.assertTrue(oldInliningArgumentCount >= inliningArguments.getInliningArgumentCount());

			oldComposedFunctionCallInternDeclaration.checkInvariants();

			return newFunctionCallInternDeclaration;

		} else {

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Skipped insertion of args from function call "
						+ oldComposedFunctionCallInternDeclaration.getName());
			}

			return oldComposedFunctionCallInternDeclaration;
		}

	}

	private IFunctionCallInternDeclaration createNewComposedFunctionCallInternDeclaration(
		IParametricFunctionCallDeclaration oldParametricFunctionCallDeclaration,
		IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration,
		IParametricFunctionCallDeclaration newParametricFunctionCallDeclaration,
		InliningArguments inliningArguments) {

		Assert.assertNotSame(newParametricFunctionCallDeclaration, oldParametricFunctionCallDeclaration);

		List childs = null;

		int i = -1;
		Iterator it = oldParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().iterator();
		while (it.hasNext()) {
			i++;
			IFunctionArgumentDeclaration currFunctionArgumentDeclaration =
				(IFunctionArgumentDeclaration)it.next();

			boolean inliningArgumentsContains = inliningArguments.contains(currFunctionArgumentDeclaration);
			boolean inliningArgumentsEqualCount =
				inliningArgumentsContains
					&& (inliningArguments.getCount(currFunctionArgumentDeclaration)
						>= currFunctionArgumentDeclaration.getReferencedAtomicFunInternCallDeclCount());

			if (!inliningArgumentsContains || !inliningArgumentsEqualCount) {

				if (childs == null) {
					childs = new LinkedList();
					newParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().clear();
				}
			}

			if (!inliningArgumentsContains) {
				childs.add(
					oldComposedFunctionCallInternDeclaration.getChild(i).copy(inliningArguments._objectMap));
				newParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().add(
					currFunctionArgumentDeclaration);

			} else if (!inliningArgumentsEqualCount) {

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINER)) {
					FunctionDeclarationCompiler.LOGGER.finer(
						"Missed a reference '"
							+ currFunctionArgumentDeclaration.getName()
							+ " was "
							+ inliningArguments.getCount(currFunctionArgumentDeclaration)
							+ ", but should be "
							+ currFunctionArgumentDeclaration.getReferencedAtomicFunInternCallDeclCount()
							+ ".");
				}

				childs.add(
					inliningArguments.getArgumentFunCallDecl(currFunctionArgumentDeclaration).copy(
						inliningArguments._objectMap));

				inliningArguments.removeInliningArgument(currFunctionArgumentDeclaration);
				newParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().add(
					currFunctionArgumentDeclaration);

			} else {
				inliningArguments.removeInliningArgument(currFunctionArgumentDeclaration);
			}

		}

		if (childs == null) {
			Assert.assertTrue(
				inliningArguments._inlined
					|| oldParametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().isEmpty());
			Assert.assertNotSame(oldParametricFunctionCallDeclaration, newParametricFunctionCallDeclaration);

			return newParametricFunctionCallDeclaration
				.getFunctionCallDeclaration()
				.getFunctionCallInternDeclaration();

		} else {
			ISpecificFunctionDeclarationWithArgDeclBinds newSpecificFunctionDeclarationWithArgDeclBinds =
				new SpecificFunctionDeclarationWithArgDeclBinds(newParametricFunctionCallDeclaration);

			addToCreatedDeclarationsDuringInliningList(newSpecificFunctionDeclarationWithArgDeclBinds);

			IComposedFunctionCallInternDeclaration newComposedFunctionCallInternDeclaration =
				FunctionCallDeclarationFactory.getInstance().createComposedFunctionCallInternDeclaration(
					oldComposedFunctionCallInternDeclaration.getParent(),
					newSpecificFunctionDeclarationWithArgDeclBinds,
					childs);

			newComposedFunctionCallInternDeclaration.checkInvariants();

			return newComposedFunctionCallInternDeclaration;
		}
	}

	private boolean addInliningArguments(
		IComposedFunctionCallInternDeclaration argumentComposedFunctionCallInternDeclaration,
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		if (!parametricFunctionCallDeclaration.isInlinePossible()
			|| _parametricFunCallDeclCompiler.isRecursive(parametricFunctionCallDeclaration)) {
			return false;
		}

		boolean shouldInline = false;

		Assert.assertFalse(_parametricFunCallDeclCompiler.isRecursive(parametricFunctionCallDeclaration));

		Iterator it = argumentComposedFunctionCallInternDeclaration.getChildren();
		Iterator it2 = parametricFunctionCallDeclaration.getFunctionArgumentDeclarationList().iterator();

		while (it.hasNext()) {
			IFunctionArgumentDeclaration currFunctionArgumentDeclaration =
				(IFunctionArgumentDeclaration)it2.next();
			IFunctionCallInternDeclaration currFunctionCallInternDeclaration =
				(IFunctionCallInternDeclaration)it.next();

			if (addInliningArgument(parametricFunctionCallDeclaration,
				currFunctionArgumentDeclaration,
				currFunctionCallInternDeclaration,
				inliningArguments)) {

				shouldInline = true;
			}
		}

		return shouldInline;
	}

	private boolean addInliningArgument(
		IParametricFunctionCallDeclaration parametricFunctionCallDeclaration,
		IFunctionArgumentDeclaration currFunctionArgumentDeclaration,
		IFunctionCallInternDeclaration currFunctionCallInternDeclaration,
		InliningArguments inliningArguments)
		throws ModelCompilingException {

		if ((currFunctionArgumentDeclaration.getReferencedAtomicFunInternCallDeclCount() == 0)
			&& (ParametricFunctionCallDeclarationCompiler.AGGRESSIVE_INLINING_AND_REMOVAL
				|| currFunctionCallInternDeclaration.isReadOnly(inliningArguments._createFunctionArgs))) {

			if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
				FunctionDeclarationCompiler.LOGGER.fine(
					"Remove unused and read only argument '"
						+ currFunctionArgumentDeclaration.getName()
						+ "' ("
						+ currFunctionCallInternDeclaration.getName()
						+ ").");
			}

		} else if (inliningArguments._forceInlining) {
			if (inliningArguments
				.addInliningArgument(currFunctionArgumentDeclaration, currFunctionCallInternDeclaration)) {
				return true;
			}

		} else if (
			(currFunctionCallInternDeclaration instanceof IAtomicFunctionCallInternDeclaration)
				&& ((IAtomicFunctionCallInternDeclaration)currFunctionCallInternDeclaration)
					.getArgumentDeclaration()
					instanceof IFunctionArgumentDeclaration) {

			IFunctionArgumentDeclaration containedFunctionArgumentDeclaration =
				(IFunctionArgumentDeclaration)
					((IAtomicFunctionCallInternDeclaration)currFunctionCallInternDeclaration)
					.getArgumentDeclaration();

			if (inliningArguments.contains(containedFunctionArgumentDeclaration)) {

				IFunctionCallInternDeclaration alreadyContainedFunctionCallInternDeclaration =
					inliningArguments.getArgumentFunCallDecl(containedFunctionArgumentDeclaration);

				if (inliningArguments
					.addInliningArgument(
						currFunctionArgumentDeclaration,
						alreadyContainedFunctionCallInternDeclaration)) {
					if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
						FunctionDeclarationCompiler.LOGGER.fine(
							"Inlined Argument '"
								+ currFunctionCallInternDeclaration.getName()
								+ "' with funCall from argument '"
								+ containedFunctionArgumentDeclaration.getName()
								+ "' (= "
								+ alreadyContainedFunctionCallInternDeclaration.getName()
								+ ").");
					}

					Assert.assertNotSame(
						currFunctionArgumentDeclaration,
						containedFunctionArgumentDeclaration);

					inliningArguments.incrementCount(containedFunctionArgumentDeclaration);

					return true;
				}

			} else {
				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Inlined Argument '"
							+ containedFunctionArgumentDeclaration.getName()
							+ "' with '"
							+ currFunctionArgumentDeclaration.getName()
							+ "'.");
				}

				if (inliningArguments
					.addInliningArgument(
						currFunctionArgumentDeclaration,
						currFunctionCallInternDeclaration)) {
					return true;
				}
			}

		} else if (
			(currFunctionArgumentDeclaration.getReferencedAtomicFunInternCallDeclCount() == 1)
				&& (ParametricFunctionCallDeclarationCompiler.AGGRESSIVE_INLINING_AND_REMOVAL
					|| (currFunctionCallInternDeclaration.isReadOnly(inliningArguments._createFunctionArgs)
						&& parametricFunctionCallDeclaration.getFunctionCallDeclaration().isReadOnly(
							inliningArguments._createFunctionArgs)))) {

			if (inliningArguments
				.addInliningArgument(currFunctionArgumentDeclaration, currFunctionCallInternDeclaration)) {

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Inlined one time referenced read only argument '"
							+ currFunctionArgumentDeclaration.getName()
							+ "' ("
							+ currFunctionCallInternDeclaration.getName()
							+ ").");
				}

				return true;
			}

		} else if (currFunctionCallInternDeclaration.isConstant(inliningArguments._createFunctionArgs)) {

			if (inliningArguments
				.addInliningArgument(currFunctionArgumentDeclaration, currFunctionCallInternDeclaration)) {

				if (Assert.ASSERTION_ON && FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
					FunctionDeclarationCompiler.LOGGER.fine(
						"Inlined constant call '" + currFunctionCallInternDeclaration.getName() + "'.");
				}

				return true;
			}
		}

		if (Assert.ASSERTION_ON
			&& !inliningArguments._forceInlining
			&& FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINER)) {

			FunctionDeclarationCompiler.LOGGER.finer(
				"Not inlined call '" + currFunctionCallInternDeclaration.getName() + "'.");
		}

		return false;
	}

	private void addToCreatedDeclarationsDuringInliningList(ISpecificFunctionDeclarationWithArgDeclBinds newSpecificFunctionDeclarationWithArgDeclBinds) {
		Assert.assertFalse(
			newSpecificFunctionDeclarationWithArgDeclBinds instanceof IUserFunctionDeclaration);
		_createdDeclarationsDuringInlining.add(newSpecificFunctionDeclarationWithArgDeclBinds);
	}

	private void addToCreatedDeclarationsDuringInliningList(IFunctionCallInternDeclaration copyOfFunCallDecl) {
		_createdDeclarationsDuringInlining.add(copyOfFunCallDecl);
	}

	public void remove() {
		_inlinedMap.clear();
		_forcedInlinedMap.clear();

		Iterator it = _createdDeclarationsDuringInlining.iterator();
		while (it.hasNext()) {
			Object currObj = it.next();
			if (currObj instanceof IDeclaration) {
				IDeclaration element = (IDeclaration)currObj;
				element.remove();
			} else {

				((IFunctionCallInternDeclaration)currObj).remove();
			}
			it.remove();
		}
	}
}
