package de.uniwue.ki.mas.basics.functionCall;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.basics.functionDeclaration.IGenericFunctionDeclaration;
import de.uniwue.ki.mas.basics.functionDeclaration.ISpecificFunctionDeclaration;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.TypeManager;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.NeverRemoveVetoListener;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.functions.IMapFunction;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (25.07.2001 16:09:50)
 * @author Christoph Oechslein
 */

final class ComposedFunctionCallInternDeclaration
	extends AbstractFunctionCallInternDeclaration
	implements IComposedFunctionCallInternDeclaration {

	private NeverRemoveVetoListener _neverRemoveVetoListener;
	private final ISpecificFunctionDeclaration _specificFunctionDeclaration;
	private final List _inputFunctionCallInternDeclarationList;

	public ComposedFunctionCallInternDeclaration(ISpecificFunctionDeclaration specificFunctionDeclaration) {
		this(null, specificFunctionDeclaration);
	}

	public ComposedFunctionCallInternDeclaration(
		ISpecificFunctionDeclaration specificFunctionDeclaration,
		Iterator inputFunctionCallInternDeclarations) {

		this(null, specificFunctionDeclaration, inputFunctionCallInternDeclarations);
	}

	public ComposedFunctionCallInternDeclaration(
		ISpecificFunctionDeclaration specificFunctionDeclaration,
		List inputFunctionCallInternDeclarationList) {

		this(null, specificFunctionDeclaration, inputFunctionCallInternDeclarationList);
	}

	public ComposedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration parent,
		ISpecificFunctionDeclaration specificFunctionDeclaration) {

		super(parent);

		Assert.assertNotNull(specificFunctionDeclaration);

		_specificFunctionDeclaration = specificFunctionDeclaration;
		_inputFunctionCallInternDeclarationList = generateFunctionCallInternDeclarations();

		addRemoveVetoListener();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public ComposedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration parent,
		ISpecificFunctionDeclaration specificFunctionDeclaration,
		IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration) {

		super(parent);

		Assert.assertNotNull(specificFunctionDeclaration);
		Assert.assertNotNull(oldComposedFunctionCallInternDeclaration);

		_specificFunctionDeclaration = specificFunctionDeclaration;
		_inputFunctionCallInternDeclarationList =
			generateFunctionCallInternDeclarations(oldComposedFunctionCallInternDeclaration);

		addRemoveVetoListener();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public ComposedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration parent,
		ISpecificFunctionDeclaration specificFunctionDeclaration,
		Iterator inputFunctionCallInternDeclarations) {

		this(
			parent,
			specificFunctionDeclaration,
			CollectionsHelper.asLinkedList(inputFunctionCallInternDeclarations));
	}

	public ComposedFunctionCallInternDeclaration(
		IComposedFunctionCallInternDeclaration parent,
		ISpecificFunctionDeclaration specificFunctionDeclaration,
		List inputFunctionCallInternDeclarationList) {

		super(parent);

		_specificFunctionDeclaration = specificFunctionDeclaration;
		_inputFunctionCallInternDeclarationList = inputFunctionCallInternDeclarationList;
		setParentForChildren();

		addRemoveVetoListener();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void addChild() {
		addChild(generateFunctionCallInternDeclaration(getChildCount()));
	}

	public void addChild(int index) {
		addChild(index, generateFunctionCallInternDeclaration(index));
	}

	public void addChild(int index, IFunctionCallInternDeclaration child) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(isArgumentLengthOk(getChildCount()));
		_inputFunctionCallInternDeclarationList.add(index, child);
		fireStateChanged();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void addChild(IFunctionCallInternDeclaration child) {
		addChild(getChildCount(), child);
	}

	public void addRemoveVetoListener() {
		if (getSpecificFunctionDeclaration() instanceof IUserFunctionDeclaration) {
			((IUserFunctionDeclaration)getSpecificFunctionDeclaration()).addRemoveVetoListener(
				getNeverRemoveVetoListener());
		}
	}

	public void checkInvariants() {
		if (!Assert.ASSERTION_ON)
			return;

		Assert.assertNotNull(
			"FunctionDeclaration is null in composedFunctionCallInternDeclaration",
			getSpecificFunctionDeclaration());
		Assert.assertTrue(
			"ArgumentSize doesn't fit functionDeclaration in composedFunctionCallInternDeclaration",
			isArgumentLengthOk(getChildCount()));

		for (int i = 0; i < getChildCount(); ++i) {
			IFunctionCallInternDeclaration childFunctionCallInternDeclaration = getChild(i);

			if (!Utils.equals(this, childFunctionCallInternDeclaration.getParent())) {
				Assert.fail("Parent is wrong in " + childFunctionCallInternDeclaration.getName());
			}

			if (!TypeManager
				.getInstance()
				.isMoreGeneral(
					getSpecificFunctionDeclaration().getSpecificInputTypeAt(i),
					childFunctionCallInternDeclaration.getSpecificOutputType())) {

				Assert.fail(
					"Type is wrong. Expected '"
						+ getSpecificFunctionDeclaration().getSpecificInputTypeAt(i).getName()
						+ "' but was '"
						+ childFunctionCallInternDeclaration.getSpecificOutputType().getName()
						+ "' (For arg '"
						+ childFunctionCallInternDeclaration.getName()
						+ "' in '"
						+ getName()
						+ "')");
			}

			childFunctionCallInternDeclaration.checkInvariants();
		}
	}

	public ICopyable copy(Map objectMapping) {

		IComposedFunctionCallInternDeclaration newComposedFunctionCallInternDeclaration =
			new ComposedFunctionCallInternDeclaration(getSpecificFunctionDeclaration());

		int index = 0;
		Iterator it = getChildren();
		while (it.hasNext()
			&& ((index < getSpecificFunctionDeclaration().getSpecificInputMaxArgsLength())
				|| (getSpecificFunctionDeclaration().getSpecificInputMaxArgsLength()
					== ISpecificFunctionDeclaration.UNLIMITED_ARGUMENT_COUNT))) {

			IFunctionCallInternDeclaration currFunctionCallInternDeclaration =
				(IFunctionCallInternDeclaration)it.next();
			IFunctionCallInternDeclaration copyCurrFunctionCallInternDeclaration =
				(IFunctionCallInternDeclaration)currFunctionCallInternDeclaration.copy(objectMapping);

			copyCurrFunctionCallInternDeclaration.setParent(newComposedFunctionCallInternDeclaration);
			newComposedFunctionCallInternDeclaration.setChild(index, copyCurrFunctionCallInternDeclaration);

			index++;
		}

		if (newComposedFunctionCallInternDeclaration.getChildCount() > index) {
			int amount = newComposedFunctionCallInternDeclaration.getChildCount() - index;
			for (int i = 0; i < amount; i++) {
				newComposedFunctionCallInternDeclaration.removeChild(index);
			}
		}

		super.copyInto(newComposedFunctionCallInternDeclaration, objectMapping);

		return newComposedFunctionCallInternDeclaration;
	}

	public void copyInto(Object intoObject, Map objectMapping) {
		throw new UnsupportedOperationException();
	}

	public IFunction createFunction(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException {
		return createFunctionArgs.getFunctionDeclarationCompiler().createFunction(createFunctionArgs, this);
	}

	private IFunctionCallInternDeclaration generateFunctionCallInternDeclaration(int index) {

		/*
		IFunctionCallInternDeclaration functionCallInternDeclaration =
		   getSpecificFunctionDeclaration().getSpecificInputTypeAt(index).createDefaultFunctionCallInternDeclaration();
		
		functionCallInternDeclaration.setParent(this);
		   */

		// I don't know why you should enforce this
		// e.g. for VoidType assertion is false !!
		/*
		Assert.assertTrue(
		getSpecificFunctionDeclaration().getSpecificInputTypeAt(index).createDefaultFunctionCallInternDeclaration()
		   instanceof IAtomicFunctionCallInternDeclaration);
		   */

		IFunctionCallInternDeclaration functionCallInternDeclaration =
			getSpecificFunctionDeclaration().createDefaultFunctionCallInternDeclarationAt(index);

		functionCallInternDeclaration.setParent(this);

		return functionCallInternDeclaration;
	}

	private IFunctionCallInternDeclaration generateFunctionCallInternDeclaration(
		int index,
		IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration,
		HashMap objectMapping) {

		// changed order of the isMoreGeneral expression since in check invariants
		if ((oldComposedFunctionCallInternDeclaration.getChildCount() > index)
			&& (TypeManager
				.getInstance()
				.isMoreGeneral(
					getSpecificFunctionDeclaration().getSpecificInputTypeAt(index),
					oldComposedFunctionCallInternDeclaration
						.getSpecificFunctionDeclaration()
						.getSpecificInputTypeAt(
						index)))) {

			//			// don't copy because of e.g. reused FunctionArgument Declarations which would lose its identity
			//			// IFunctionCallInternDeclaration functionCallInternDeclaration =
			//			//    (IFunctionCallInternDeclaration) oldComposedFunctionCallInternDeclaration.getChild(index).copy();
			//			IFunctionCallInternDeclaration functionCallInternDeclaration =
			//				oldComposedFunctionCallInternDeclaration.getChild(index);
			//

			// but we must do a copy since it is the child of the parent and a child can only have one parent
			// other idea would be to change the old parent!!
			IFunctionCallInternDeclaration functionCallInternDeclaration =
				(IFunctionCallInternDeclaration)oldComposedFunctionCallInternDeclaration.getChild(
					index).copy(
					objectMapping);
			functionCallInternDeclaration.setParent(this);
			return functionCallInternDeclaration;

		} else {
			return generateFunctionCallInternDeclaration(index);
		}
	}

	private List generateFunctionCallInternDeclarations() {
		return generateFunctionCallInternDeclarations(getSpecificFunctionDeclaration());
	}

	private List generateFunctionCallInternDeclarations(ISpecificFunctionDeclaration funcDecl) {
		final int size = funcDecl.getSpecificInputDefaultArgsLength();
		List args = new ArrayList(size);
		for (int i = 0; i < size; ++i) {
			args.add(generateFunctionCallInternDeclaration(i));
		}
		return args;
	}

	private List generateFunctionCallInternDeclarations(
		ISpecificFunctionDeclaration funcDecl,
		IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration) {

		int size;

		ISpecificFunctionDeclaration oldFunctionDeclaration =
			oldComposedFunctionCallInternDeclaration.getSpecificFunctionDeclaration();

		if (funcDecl.getSpecificInputMaxArgsLength()
			== ISpecificFunctionDeclaration.UNLIMITED_ARGUMENT_COUNT) {
			if (oldFunctionDeclaration.getSpecificInputMaxArgsLength()
				== ISpecificFunctionDeclaration.UNLIMITED_ARGUMENT_COUNT) {
				if (funcDecl
					.getSpecificInputTypeAt(0)
					.equals(oldFunctionDeclaration.getSpecificInputTypeAt(0))) {
					size =
						Math.max(
							funcDecl.getSpecificInputMinArgsLength(),
							oldComposedFunctionCallInternDeclaration.getChildCount());

				} else {
					size = funcDecl.getSpecificInputDefaultArgsLength();
				}

			} else {
				size = funcDecl.getSpecificInputDefaultArgsLength();
			}

		} else {
			size = funcDecl.getSpecificInputDefaultArgsLength();
		}

		ArrayList args = new ArrayList(size);

		HashMap objectMapping = new HashMap();

		for (int i = 0; i < size; ++i) {
			args.add(
				generateFunctionCallInternDeclaration(
					i,
					oldComposedFunctionCallInternDeclaration,
					objectMapping));
		}

		return args;
	}

	private List generateFunctionCallInternDeclarations(IComposedFunctionCallInternDeclaration oldComposedFunctionCallInternDeclaration) {
		return generateFunctionCallInternDeclarations(
			getSpecificFunctionDeclaration(),
			oldComposedFunctionCallInternDeclaration);
	}

	public java.util.Iterator getAllAtomicFunctionCallInternDeclarations() {

		Iterator allAtomics = CollectionsHelper.select(getChildren(), new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return (obj instanceof IAtomicFunctionCallInternDeclaration);
			}
		});
		Iterator allComposeds = CollectionsHelper.select(getChildren(), new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return (obj instanceof IComposedFunctionCallInternDeclaration);
			}
		});

		Iterator allAtomicsRecursive = CollectionsHelper.flatten(allComposeds, new IMapFunction() {
			public Object map(Object obj) {
				return ((IComposedFunctionCallInternDeclaration)obj)
					.getAllAtomicFunctionCallInternDeclarations();
			}
		});

		return CollectionsHelper.concatenate(allAtomics, allAtomicsRecursive);
	}

	public IFunctionCallInternDeclaration getChild(int index) {
		return (IFunctionCallInternDeclaration)_inputFunctionCallInternDeclarationList.get(index);
	}

	public int getChildCount() {
		if (_inputFunctionCallInternDeclarationList == null)
			return 0;
		else
			return _inputFunctionCallInternDeclarationList.size();
	}

	public Iterator getChildren() {
		return _inputFunctionCallInternDeclarationList.iterator();
	}

	public ISpecificFunctionDeclaration getSpecificFunctionDeclaration() {
		return _specificFunctionDeclaration;
	}

	public int getIndexOfChild(IFunctionCallInternDeclaration child) {
		return _inputFunctionCallInternDeclarationList.indexOf(child);
	}

	public String getName() {
		return getNameIntern();
	}

	public void appendName(StringBuffer buff, int depth, int maxDepth) {
		if (depth < maxDepth) {
			getSpecificFunctionDeclaration().appendName(buff, depth + 1, maxDepth);

			//			buff.append("@");
			//			buff.append(System.identityHashCode(this) % 100);
			buff.append("(");

			Iterator it = getChildren();
			while (it.hasNext()) {
				((IFunctionCallInternDeclaration)it.next()).appendName(buff, depth + 1, maxDepth);
				if (it.hasNext()) {
					buff.append(", ");
				}
			}
			buff.append(")");

		} else {
			buff.append("(...)");
		}
	}

	private NeverRemoveVetoListener getNeverRemoveVetoListener() {
		if (_neverRemoveVetoListener == null)
			_neverRemoveVetoListener = new NeverRemoveVetoListener(this);

		return _neverRemoveVetoListener;
	}

	public ISpecificType getSpecificOutputType() {
		return getSpecificFunctionDeclaration().getSpecificOutputType();
	}

	private final boolean isArgumentLengthOk(int size) {
		return (
			(getSpecificFunctionDeclaration().getSpecificInputMinArgsLength() <= size)
				&& (getSpecificFunctionDeclaration().getSpecificInputMaxArgsLength()
					== ISpecificFunctionDeclaration.UNLIMITED_ARGUMENT_COUNT))
			|| (getSpecificFunctionDeclaration().getSpecificInputMaxArgsLength() >= size);
	}

	public final boolean isConstant(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException {
		return createFunctionArgs.getFunctionDeclarationCompiler().isConstant(createFunctionArgs, this);
	}

	public final boolean isReadOnly(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException {
		return createFunctionArgs.getFunctionDeclarationCompiler().isReadOnly(createFunctionArgs, this);
	}



//	public boolean equals(Object arg0) {
//		if (arg0==null) return false;
//		if (arg0==this) return true;
//		
//		if (arg0 instanceof ComposedFunctionCallInternDeclaration) {
//			ComposedFunctionCallInternDeclaration cfc = (ComposedFunctionCallInternDeclaration)arg0;
//			return (
//				cfc.getSpecificFunctionDeclaration().equals(getSpecificFunctionDeclaration())
//					&& CollectionsHelper.isEqual(cfc.getChildren(), getChildren(), true));
//		} else
//			return false;
//	}
//		
//	public int hashCode() {
//		if (getSpecificFunctionDeclaration()!=null) 
//		   return getSpecificFunctionDeclaration().hashCode();
//		else return 0;
//	}

	public boolean isomorphicEquals(IIsomorphicHashable obj) {
		if (this == obj)
			return true;

		if (!(obj instanceof IComposedFunctionCallInternDeclaration)) {
			return false;
		}

		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration =
			(IComposedFunctionCallInternDeclaration)obj;

		if (getSpecificFunctionDeclaration() == null) {
			return equals(composedFunctionCallInternDeclaration);
		}

		if (composedFunctionCallInternDeclaration.getChildCount() != getChildCount()) {
			return false;
		}

		if (!getSpecificFunctionDeclaration()
			.isomorphicEquals(composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration())) {
			return false;
		}

		for (int i = 0; i < getChildCount(); i++) {
			if (!getChild(i).isomorphicEquals(composedFunctionCallInternDeclaration.getChild(i))) {
				return false;
			}
		}

		return true;
	}

	public int isomorphicHashCode() {
		ISpecificFunctionDeclaration specificFunctionDeclaration = getSpecificFunctionDeclaration();

		if (specificFunctionDeclaration == null) {
			return super.hashCode();

		} else {
			IGenericFunctionDeclaration genericFunctionDeclaration =
				specificFunctionDeclaration.getGenericFunctionDeclaration();
			if (genericFunctionDeclaration != null) {
				return genericFunctionDeclaration.isomorphicHashCode() * 3 + getChildCount();

			} else {
				return specificFunctionDeclaration.isomorphicHashCode() * 3 + getChildCount();
			}
		}
	}

	public void remove() {
		removeRemoveVetoListener();

		for (int i = 0; i < getChildCount(); i++) {
			getChild(i).remove();
		}

		if (_inputFunctionCallInternDeclarationList != null)
			_inputFunctionCallInternDeclarationList.clear();

		super.remove();
	}

	public void removeChild(int index) {
		Assert.assertTrue(isArgumentLengthOk(index));

		IFunctionCallInternDeclaration child = getChild(index);

		_inputFunctionCallInternDeclarationList.remove(index);

		child.remove();
		fireStateChanged();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void removeChild(IFunctionCallInternDeclaration child) {
		_inputFunctionCallInternDeclarationList.remove(child);

		child.remove();
		fireStateChanged();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void removeRemoveVetoListener() {
		if (getSpecificFunctionDeclaration() instanceof IUserFunctionDeclaration) {
			((IUserFunctionDeclaration)getSpecificFunctionDeclaration()).removeRemoveVetoListener(
				getNeverRemoveVetoListener());
		}
	}

	public void setChild(int index, IFunctionCallInternDeclaration child) {
		Assert.assertEquals(this, child.getParent());

		setChildCount(Math.max(index + 1, getChildCount()));

		Assert.assertTrue(index < getChildCount());

		IFunctionCallInternDeclaration oldChild =
			(IFunctionCallInternDeclaration)_inputFunctionCallInternDeclarationList.set(index, child);

		if ((oldChild != null) && (oldChild != child)) {
			oldChild.remove();
		}

		fireStateChanged();

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	private void setChildCount(int newChildCount) {
		if (newChildCount != getChildCount()) {

			Assert.assertTrue(
				"newChildCount (" + newChildCount + ") too big.",
				isArgumentLengthOk(newChildCount));

			if (newChildCount > getChildCount()) {
				for (int i = 0; i < newChildCount - getChildCount(); i++) {
					_inputFunctionCallInternDeclarationList.add(generateFunctionCallInternDeclaration(i));
				}

			} else
				// (newChildCount < getChildCount())
				{
				for (int i = getChildCount() - 1; i <= newChildCount; i++) {
					removeChild(i);
				}
			}
			fireStateChanged();

			if (Assert.ASSERTION_ON) {
				Assert.assertEquals(newChildCount, getChildCount());
				checkInvariants();
			}
		}
	}

	private void setParentForChildren() {
		Iterator it = _inputFunctionCallInternDeclarationList.iterator();
		while (it.hasNext()) {
			IFunctionCallInternDeclaration call = (IFunctionCallInternDeclaration)it.next();
			call.setParent(this);
		}
	}
}