package de.uniwue.ki.mas.basics.functionCall;
import gnu.trove.TObjectBooleanHashMap;

import java.awt.Image;
import java.io.Serializable;
import java.util.Iterator;
import java.util.logging.Level;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.functionArgument.SpecificFunctionDeclarationWithArgDeclBinds;
import de.uniwue.ki.mas.basics.functionDeclaration.FunctionDeclarationCompiler;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.VoidType;
import de.uniwue.ki.mas.basics.type.argument.IArgument;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.collection.CollectionsHelper;

public final class FunctionCallDeclarationCompiler implements IRemoveable, Serializable {
	private static final boolean _cacheComposedFunCallsOn = true;
	private static final boolean _cacheConstantsOn = true;
	private static final boolean _cacheReadOnlyOn = true;

	public static IFunctionList createFunctionList(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		int childCount = composedFunctionCallInternDeclaration.getChildCount();

		switch (childCount) {
			case 0 :
				return IFunctionList.EMPTY_FUNCTION_LIST;

			case 1 :
				return new SingletonFunctionList(
					composedFunctionCallInternDeclaration.getChild(0).createFunction(createFunctionArgs));

			case 2 :
				return new TwoFunctionList(
					composedFunctionCallInternDeclaration.getChild(0).createFunction(createFunctionArgs),
					composedFunctionCallInternDeclaration.getChild(1).createFunction(createFunctionArgs));

			default :
				FunctionList functions = new FunctionList(childCount);
				Iterator itCallFunctionDeclaration = composedFunctionCallInternDeclaration.getChildren();
				while (itCallFunctionDeclaration.hasNext()) {
					IFunctionCallInternDeclaration callDecl =
						(IFunctionCallInternDeclaration)itCallFunctionDeclaration.next();
					functions.add(callDecl.createFunction(createFunctionArgs));
				}

				return functions;
		}
	}

	private final boolean _compileOn;
	private boolean _shouldCompile = true;

	//TODO isomorphic Equal HashMaps couldbe faster ?!
	private final TObjectBooleanHashMap _constantHelperMap = new TObjectBooleanHashMap();
	private final TObjectBooleanHashMap _readOnlyHelperMap = new TObjectBooleanHashMap();

	public FunctionCallDeclarationCompiler(boolean compileOn) {
		super();

		_compileOn = compileOn;
	}

	private IArgument createConstantArgument(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		boolean oldShouldCompile = isShouldCompile();

		try {

			setShouldCompile(false);

			IFunction constantFunction =
				composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration().createFunction(
					createFunctionArgs,
					composedFunctionCallInternDeclaration);

			if (constantFunction instanceof IArgument) {
				return (IArgument)constantFunction;

			} else {
				Object result =
					constantFunction.execute(createFunctionArgs.createCreateAndExecuteFunctionArgs());

				if (Assert.ASSERTION_ON) {
					if ((result instanceof Iterator)) {
						Assert.assertTrue(
							CollectionsHelper.isEqual(
								(Iterator)result,
								(Iterator)constantFunction.execute(
									createFunctionArgs.createCreateAndExecuteFunctionArgs()),
								false));

						result =
							constantFunction.execute(createFunctionArgs.createCreateAndExecuteFunctionArgs());
					} else if (result instanceof Image) {
						// do not check

					} else {
						Assert.assertEquals(
							result,
							constantFunction.execute(
								createFunctionArgs.createCreateAndExecuteFunctionArgs()));
					}
				}

				return composedFunctionCallInternDeclaration.getSpecificOutputType().createArgument(
					createFunctionArgs,
					result);
			}

		} catch (ModelCompilingException exception) {
			throw new FunctionCallInternDeclCompileException(
				composedFunctionCallInternDeclaration,
				exception);

		} catch (SimulationRuntimeException exception) {
			throw new FunctionCallInternDeclCompileException(
				composedFunctionCallInternDeclaration,
				exception);

		} finally {
			setShouldCompile(oldShouldCompile);
		}
	}

	public IFunction createFunction(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		if (!_cacheComposedFunCallsOn) {
			return createNewFunction(createFunctionArgs, composedFunctionCallInternDeclaration);

		} else if (createFunctionArgs.existsCompilationObject(composedFunctionCallInternDeclaration, true)) {
			return (IFunction)createFunctionArgs.getCompilationObject(
				composedFunctionCallInternDeclaration,
				true);

		} else {
			IFunction newFunction =
				createNewFunction(createFunctionArgs, composedFunctionCallInternDeclaration);

			createFunctionArgs.addCompilationObject(composedFunctionCallInternDeclaration, newFunction, true);

			return newFunction;
		}
	}

	private IFunction createNewFunction(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		try {
			return createNewFunctionIntern(createFunctionArgs, composedFunctionCallInternDeclaration);

		} catch (ModelCompilingException exception) {
			if (exception instanceof FunctionCallInternDeclCompileException)
				throw (FunctionCallInternDeclCompileException)exception;
			else
				throw new FunctionCallInternDeclCompileException(
					composedFunctionCallInternDeclaration,
					exception);
		}
	}

	private IFunction createNewFunctionIntern(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		boolean shouldCompile = isCompileOn() && isShouldCompile();

		if (shouldCompile && isConstant(createFunctionArgs, composedFunctionCallInternDeclaration)) {

			IArgument argument =
				createConstantArgument(createFunctionArgs, composedFunctionCallInternDeclaration);

			if (Assert.ASSERTION_ON) {

				try {
					Object value = argument.execute(createFunctionArgs.createCreateAndExecuteFunctionArgs());

					if (FunctionDeclarationCompiler.LOGGER.isLoggable(Level.INFO)) {
						ISpecificType specificOutputType =
							composedFunctionCallInternDeclaration.getSpecificOutputType();
						StringBuffer buff = new StringBuffer();
						buff.append("Const expression detected: '");
						specificOutputType.appendNameForArgumentObject(buff, value, 0, 30);
						buff.append("' expression: ");
						composedFunctionCallInternDeclaration.appendName(buff, 0, 30);
						FunctionDeclarationCompiler.LOGGER.info(buff.toString());

						if (value instanceof Iterator) {
							value = argument.execute(createFunctionArgs.createCreateAndExecuteFunctionArgs());
						}
					}

				} catch (SimulationRuntimeException exception) {
					throw new FunctionCallInternDeclCompileException(
						composedFunctionCallInternDeclaration,
						exception);
				}
			}

			return argument;

		} else if (
			shouldCompile
				&& (composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration()
					instanceof SpecificFunctionDeclarationWithArgDeclBinds)) {

			SpecificFunctionDeclarationWithArgDeclBinds specificFunctionDeclarationWithArgDeclBinds =
				(SpecificFunctionDeclarationWithArgDeclBinds)composedFunctionCallInternDeclaration
					.getSpecificFunctionDeclaration();

			return createFunctionArgs.getFunctionDeclarationCompiler().createFunction(
				createFunctionArgs,
				specificFunctionDeclarationWithArgDeclBinds.getParametricFunctionCallDeclaration(),
				composedFunctionCallInternDeclaration);

		} else {
			return reallyCreateFunction(createFunctionArgs, composedFunctionCallInternDeclaration);
		}
	}

	public final boolean isCompileOn() {
		return _compileOn;
	}

	public boolean isConstant(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		if (!isCompileOn()) {
			return false;
		} else if (!_cacheConstantsOn) {
			return isConstantIntern(createFunctionArgs, composedFunctionCallInternDeclaration);

		} else if (_constantHelperMap.containsKey(composedFunctionCallInternDeclaration)) {
			return _constantHelperMap.get(composedFunctionCallInternDeclaration);

		} else {
			boolean value = isConstantIntern(createFunctionArgs, composedFunctionCallInternDeclaration);
			_constantHelperMap.put(composedFunctionCallInternDeclaration, value);
			return value;
		}
	}

	private boolean isConstantIntern(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		try {
			boolean value =
				composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration().isConstant(
					createFunctionArgs,
					composedFunctionCallInternDeclaration);

			if (Assert.ASSERTION_ON
				&& composedFunctionCallInternDeclaration.getSpecificOutputType().equals(
					VoidType.getInstance())) {
				Assert.assertFalse(value);
			}

			return value;

		} catch (ModelCompilingException e) {
			if (e instanceof FunctionCallInternDeclCompileException)
				throw (FunctionCallInternDeclCompileException)e;
			else
				throw new FunctionCallInternDeclCompileException(composedFunctionCallInternDeclaration, e);
		}
	}

	public boolean isReadOnly(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		if (!isCompileOn()) {
			return false;
		} else if (!_cacheReadOnlyOn) {
			return isReadOnlyIntern(createFunctionArgs, composedFunctionCallInternDeclaration);

		} else if (_readOnlyHelperMap.containsKey(composedFunctionCallInternDeclaration)) {
			return _readOnlyHelperMap.get(composedFunctionCallInternDeclaration);

		} else {
			boolean value = isReadOnlyIntern(createFunctionArgs, composedFunctionCallInternDeclaration);
			_readOnlyHelperMap.put(composedFunctionCallInternDeclaration, value);
			return value;
		}
	}

	private boolean isReadOnlyIntern(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws FunctionCallInternDeclCompileException {

		try {
			//don't call it -> endless recursion
			//return composedFunctionCallInternDeclaration.isReadOnly(createFunctionArgs);
			return composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration().isReadOnly(
				createFunctionArgs,
				composedFunctionCallInternDeclaration);

		} catch (ModelCompilingException e) {
			if (e instanceof FunctionCallInternDeclCompileException)
				throw (FunctionCallInternDeclCompileException)e;
			else
				throw new FunctionCallInternDeclCompileException(composedFunctionCallInternDeclaration, e);
		}
	}

	private boolean isShouldCompile() {
		return _shouldCompile;
	}

	private IFunction reallyCreateFunction(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException {

		IFunction newFunction =
			composedFunctionCallInternDeclaration.getSpecificFunctionDeclaration().createFunction(
				createFunctionArgs,
				composedFunctionCallInternDeclaration);

		if (FunctionDeclarationCompiler.LOGGER.isLoggable(Level.FINE)) {
			FunctionDeclarationCompiler.LOGGER.fine(
				"Compile '"
					+ composedFunctionCallInternDeclaration.getName()
					+ "' to '"
					+ newFunction.getName()
					+ "'");
		}

		return newFunction;
	}

	public void remove() {
	}

	private void setShouldCompile(boolean value) {
		_shouldCompile = value;
	}
}