package de.uniwue.ki.mas.basics.functionCall;

import java.beans.PropertyChangeListener;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.tree.TreeModel;

import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.basics.functionDeclaration.ISpecificFunctionDeclaration;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.argument.IArgumentDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.IDeclaration;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * A FunctionCallDeclaration is a tree which inner nodes are functions (see {@link ISpecificFunctionDeclaration})
 * with arguments (<code>IFunctionCallDeclaration</code>) as childrens
 * and its leafs are values (see {@link IArgumentDeclaration}).
 * The inner nodes are represented by {@link IComposedFunctionCallInternDeclaration}
 * and the leafs by {@link IAtomicFunctionCallInternDeclaration}.
 * <BR>Creation date: (31.03.00 13:40:55)
 *
 * @author Christoph Oechslein
 */

public interface IFunctionCallDeclaration extends IIsomorphicHashable, IDeclaration {

	/**
	 * Add a change listener, if something changes in this declaration
	 * (or in dependend declarations) a {@link ChangeEvent change event} is fired.
	 *
	 * @see# removeChangeListener(ChangeListener)
	 */
	//void addChangeListener(ChangeListener newListener);

	/**
	 * Adds property change listener.
	 *
	 * @see #removePropertyChangeListener(PropertyChangeListener)
	 */
	//void addPropertyChangeListener(PropertyChangeListener listener);

	/**
	 * Checks invariants of this function call declaration:
	 * </ul>
	 *   <li> Most general type is more general than type of root function call intern declaration.
	 *   <li> Parent in function call intern declaration is correct.
	 *   <li> Type of function call intern declaration are correct argument types of parent.
	 * </ul>
	 */
	void checkInvariants();

	/**
	 * Compiles a new {@link IFunction}, calling execute of the results executs the declared function call tree.
	 *
	 * @see IFunctionCallInternDeclaration#createFunction(ICreateFunctionArgs)
	 */
	IFunction createFunction(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException;

	/**
	 * Returns the root node.
	 */

	IFunctionCallInternDeclaration getFunctionCallInternDeclaration();

	/**
	 * The output type of the root function call must be this type
	 * or can be more specific.
	 */
	ISpecificType getMostGeneralOutputType();

	/**
	 * Returns the path of {@link IFunctionCallInternDeclaration IFunctionCallInternDeclarations}
	 * in the tree to <code>aNode</code>.
	 */
	IFunctionCallInternDeclaration[] getPathToRoot(IFunctionCallInternDeclaration aNode);

	/**
	 * The real output type of the root function call.
	 */
	ISpecificType getSpecificOutputType();

	/**
	 * Gets the corresponding TreeModel, could be <code>return this;</code>.
	 */
	TreeModel getTreeModel();

	/**
	 * @see IFunctionCallInternDeclaration#isConstant(ICreateFunctionArgs)
	 */
	boolean isConstant(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException;

	/**
	 * @see IFunctionCallInternDeclaration#isReadOnly(ICreateFunctionArgs)
	 */
	boolean isReadOnly(ICreateFunctionArgs createFunctionArgs) throws ModelCompilingException;

	/**
	 * Removes a change listener.
	 *
	 * @see addChangeListener(ChangeListener)
	 */
	void removeChangeListener(ChangeListener newListener);

	/**
	 * Removes property change listener.
	 *
	 * @see #addPropertyChangeListener(PropertyChangeListener)
	 */

	void removePropertyChangeListener(PropertyChangeListener listener);

	/**
	 * Sets a new root node, only if the output type of the <code>functionCallInternDeclaration</code>
	 * is same or more specific as {@link #getMostGeneralOutputType}
	 */
	void setFunctionCallInternDeclaration(IFunctionCallInternDeclaration functionCallInternDeclaration);

	/**
	 * Sets a new most general output type, if the output type of the root node isn't equal or
	 * more specific than the new type the root node is replace by a new default function call by
	 * calling {@link FunctionCallDeclarationFactory#createFunctionCallInternDeclaration}.
	 */
	void setMostGeneralOutputType(ISpecificType newOutputArgument);

	/**
	 * Replace the one argument of an inner node (function call) with a new node/leaf.
	 * newCall.getParent() == oldCall.getParent()
	 */
	void replaceFunctionCallInternDeclarationTo(
		IFunctionCallInternDeclaration newCall,
		IFunctionCallInternDeclaration oldCall);
}
