package de.uniwue.ki.mas.basics.functionDeclaration;
import java.util.List;

import de.uniwue.ki.mas.basics.function.*;
import de.uniwue.ki.mas.basics.functionCall.IComposedFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.type.ISpecificFunctionType;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.model.ModelCompilingException;
/**
 * This interface is a declaration for the declaration of a specific function, with specific input arguments.
 * <BR>Creation date: (30.03.00 17:48:12)
 *
 * @author Christoph Oechslein
 * @see IGenericFunctionDeclaration
 */
public interface ISpecificFunctionDeclaration extends IFunctionDeclaration {

	/**
	 * Constant for a function declaration with unlimited length of input arguments.
	 */
	int UNLIMITED_ARGUMENT_COUNT = -1;

	/**
	 * Returns the default length of the parameter list.
	 * <BR>Returning {@link #UNLIMITED_ARGUMENT_COUNT} means unlimited.
	 */
	int getSpecificInputDefaultArgsLength();

	/**
	 * Returning {@link #UNLIMITED_ARGUMENT_COUNT} means unlimited.
	 */
	int getSpecificInputMaxArgsLength();

	/**
	 * Returning {@link #UNLIMITED_ARGUMENT_COUNT} means unlimited.
	 */
	int getSpecificInputMinArgsLength();

	/**
	 * Returns the specific type at position 'index'.
	 */
	ISpecificType getSpecificInputTypeAt(int index);

	/**
	 * Returns the name of the input argument at position 'index'.
	 */
	String getSpecificInputTypeNameAt(int index);

	/**
	 * Returns the specific output type of this declaration.
	 */
	ISpecificType getSpecificOutputType();

	/**
	 * Returns true if the inputTypes and the outputTypes matches this
	 * SpecificFunctionDeclaration. Attention outputType could be null!
	 */
	boolean isSpecificFunctionDeclOk(ISpecificType outputType, List inputTypes);

	/**
	 * Creates a new function call for the specified position.
	 * Possible implementation:
	 * <code>return getSpecificInputTypeAt(index).createDefaultFunctionCallInternDeclaration();</code>
	 */
	IFunctionCallInternDeclaration createDefaultFunctionCallInternDeclarationAt(int index);

	/**
	 * Returns the function type of this function declaration..
	 */
	ISpecificFunctionType getSpecificFunctionType();

	/**
	 * Must return true if the resulting IFunction of createFunction is dependent on the execute binding argument
	 * (if it can return different results with a different execute binding argument)
	 * or if createFunction returns different IFunction (different in the sense of producing a
	 * different result with an other binding). 
	 * 
	 * <BR>E.g. the '+' functions returns true the 'getRandom' returns false.
	 */
	boolean isConstant(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException;

	boolean isConstant(ICreateFunctionArgs createFunctionArgs);

	/**
	 * Must return false if the resulting IFunction of createFunction has side effects, 
	 * i.e. changes something. All actions have side effects and some functions, like pop.
	 */
	boolean isReadOnly(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException;

	boolean isReadOnly(ICreateFunctionArgs createFunctionArgs);

	/**
	 * Returns the Generic Function it was created by, or null if wasn't created by a Generic Function.
	 */
	IGenericFunctionDeclaration getGenericFunctionDeclaration();

	/**
	 * Creates an IFunction, the IFunctions in the IFunctionList must be called for producing the real input objects.
	 * In createFunctionArgs is the current compiling binding, e.g. the simulation for this function.
	 *
	 * @input functions is list of IFunctions, the size of functionsmust be ok for the IFunctionDeclaration
	 * (between min and max and the input types of the declaration must the same as the output types of the
	 * IFunctions in functions).
	 * @output returns a new IFunction, which implements the FunctionDeclaration and the ParameterConstellation.
	 * it must be of type outputType
	 * 
	 * @see #createFunction(ICreateFunctionArgs, IComposedFunctionCallInternDeclaration)
	 * 
	 * @throws ModelCompilingException if during creation of a new IFunction an error occurs.
	 */
	IFunction createFunction(ICreateFunctionArgs createFunctionArgs, IFunctionList functions)
		throws ModelCompilingException;

	/**
	 * Creates an IFunction, the composedFunctionCallInternDeclaration encodes the call,
	 * normaly creates a {@link IFunctionList} of the children of the composed call 
	 * (via {@link IFunctionCallInternDeclaration#createFunction(ICreateFunctionArgs)}) und call {@link #createFunction(ICreateFunctionArgs, IFunctionList).
	 * 
	 * @see #createFunction(ICreateFunctionArgs, IFunctionList)
	 * 
	 * @throws ModelCompilingException if during creation of a new IFunction an error occurs.
	 */
	IFunction createFunction(
		ICreateFunctionArgs createFunctionArgs,
		IComposedFunctionCallInternDeclaration composedFunctionCallInternDeclaration)
		throws ModelCompilingException;

	IParametricFunction createParametricFunction(ICreateFunctionArgs createFunctionArgs)
		throws ModelCompilingException;

}