package de.uniwue.ki.mas.basics.functionDeclaration.list;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.functionDeclaration.*;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.TypeManager;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.name.AbstractNamed;

/**
 * Creation date: (08.11.2001 14:51:23)
 * @author Christoph Oechslein
 */
public final class FunctionDeclarationListHelper {

	private FunctionDeclarationListHelper() {
		super();
	}

	/** 
	 * Return a specificFunctionDeclaration with name <code>name</code>, with inputtypes <code>inputTypes</code> 
	 * and with outputType <code>outputType</code> 
	 * (with can be null see {@link IGenericFunctionDeclaration#needSpecificOutputTypeToCreateSpecificFunctionDeclaration()}).
	
	 */
	public static ISpecificFunctionDeclaration getSpecificFunctionDeclaration(
		Collection functionDeclarationList,
		String name,
		ISpecificType outputType,
		List inputTypeList) {

		Iterator it =
			CollectionsHelper.removeDuplicates(
				getFunctionDeclarationsByName(functionDeclarationList.iterator(), name));

		if (CollectionsHelper.isEmpty(it)) {
			return null;

		} else {
			ISpecificFunctionDeclaration foundSpecificFunctionDeclaration = null;
			while (it.hasNext()) {
				IFunctionDeclaration functionDeclaration = (IFunctionDeclaration) it.next();

				ISpecificFunctionDeclaration specificFunctionDeclaration =
					isFunctionDeclOk(functionDeclaration, outputType, inputTypeList);

				if (specificFunctionDeclaration != null) {
					Assert.assertNull(
						"Found two function declarations: "
							+ (foundSpecificFunctionDeclaration == null
								? null
								: foundSpecificFunctionDeclaration.getName())
							+ " and "
							+ specificFunctionDeclaration.getName(),
						foundSpecificFunctionDeclaration);

					foundSpecificFunctionDeclaration = specificFunctionDeclaration;

					if (!Assert.ASSERTION_ON)
						break;
				}
			}

			return foundSpecificFunctionDeclaration;
		}
	}

	private static ISpecificFunctionDeclaration isFunctionDeclOk(
		IFunctionDeclaration functionDeclaration,
		ISpecificType outputType,
		List inputTypes) {

		if (functionDeclaration instanceof ISpecificFunctionDeclaration) {
			ISpecificFunctionDeclaration specificFunctionDeclaration =
				(ISpecificFunctionDeclaration) functionDeclaration;

			if (specificFunctionDeclaration.isSpecificFunctionDeclOk(outputType, inputTypes))
				return specificFunctionDeclaration;
			else
				return null;

		} else {
			IGenericFunctionDeclaration genericFunctionDeclaration =
				(IGenericFunctionDeclaration) functionDeclaration;

			if ((outputType != null)
				&& (!genericFunctionDeclaration.canCreateFunctionDeclarationWith(outputType)))
				return null;

			ISpecificFunctionDeclaration specificFunctionDeclaration =
				genericFunctionDeclaration.createFunctionDeclaration(outputType, inputTypes.iterator());

			if (Assert.ASSERTION_ON
				&& (specificFunctionDeclaration != null)
				&& (!specificFunctionDeclaration.isSpecificFunctionDeclOk(outputType, inputTypes))) {

				StringBuffer buff = new StringBuffer();

				buff.append("specific Function Not Ok: ");
				specificFunctionDeclaration.appendName(buff, 0, 30);
				buff.append("(");
				AbstractNamed.append(inputTypes.iterator(), buff, 0, 30);
				buff.append(") ");
				buff.append(outputType + "");

				Assert.fail(buff.toString());
			}

			return specificFunctionDeclaration;
		}
	}

	public static Iterator getFunctionDeclarationsByName(Iterator functionDeclarations, final String name) {
		return CollectionsHelper.select(functionDeclarations, new IPredicateFunction() {
			public boolean isElementOk(Object obj) throws Exception {
				IFunctionDeclaration currentFunctionDeclaration = (IFunctionDeclaration) obj;
				return name.equalsIgnoreCase(currentFunctionDeclaration.getName());
			}
		});
	}

	/**
	  * Produces a list of all functions which produce results of type outputType. 
	*/
	public static Iterator getFunctionDeclarations(
		final List functionDeclarationList,
		ISpecificType outputType) {

		if (Assert.ASSERTION_ON)
			Assert.assertTrue(
				"Duplicates: " + CollectionsHelper.getDuplicateEntries(functionDeclarationList),
				!CollectionsHelper.hasDuplicateEntries(functionDeclarationList));

		Iterator specificFunctionDecls =
			getFunctionDeclarationsAtomicType(
				getSpecificFunctionDeclarations(functionDeclarationList.iterator()),
				outputType);

		return CollectionsHelper.concatenate(
			specificFunctionDecls,
			getFunctionDeclarationsGenericTypeNotInstantiated(
				getGenericFunctionDeclarations(functionDeclarationList.iterator()),
				outputType));
	}

	private static Iterator getFunctionDeclarationsAtomicType(
		Iterator specificFunctionDeclarations,
		final ISpecificType outputType) {

		IPredicateFunction filterFunction = new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return hasSameOutputType((ISpecificFunctionDeclaration) obj, outputType);
			}
		};
		return CollectionsHelper.select(specificFunctionDeclarations, filterFunction);
	}

	private static Iterator getFunctionDeclarationsGenericTypeNotInstantiated(
		Iterator genericFunctionDeclarations,
		final ISpecificType specificOutputType) {

		IPredicateFunction filterFunction = new IPredicateFunction() {
			public boolean isElementOk(Object currentElement) {
				IGenericFunctionDeclaration genericFunDecl = (IGenericFunctionDeclaration) currentElement;
				return getOutputTypeWithCanCreateFunctionDeclaration(genericFunDecl, specificOutputType)
					!= null;
			}
		};
		return CollectionsHelper.select(genericFunctionDeclarations, filterFunction);
	}

	public static Iterator getSpecificFunctionDeclarations(Iterator functionDeclarations) {
		return CollectionsHelper.select(functionDeclarations, new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return obj instanceof ISpecificFunctionDeclaration;
			}
		});
	}

	protected static boolean hasSameOutputType(
		ISpecificFunctionDeclaration functionDeclaration,
		ISpecificType outputType) {

		return TypeManager.getInstance().isMoreGeneral(
			outputType,
			functionDeclaration.getSpecificOutputType());
	}

	public static ISpecificType getOutputTypeWithCanCreateFunctionDeclaration(
		IGenericFunctionDeclaration genericFunDecl,
		ISpecificType specificOutputType) {

		if (genericFunDecl.canCreateFunctionDeclarationWith(specificOutputType)) {
			return specificOutputType;
			
		} else {
			return getOutputTypeWithCanCreateFunctionDeclaration(
				genericFunDecl,
				specificOutputType.getSubTypes());
		}
	}

	private static ISpecificType getOutputTypeWithCanCreateFunctionDeclaration(
		IGenericFunctionDeclaration genericFunDecl,
		Iterator specificOutputTypes) {
			
		while (specificOutputTypes.hasNext()) {
			ISpecificType specificOutputType = (ISpecificType) specificOutputTypes.next();
			ISpecificType specificType =
				getOutputTypeWithCanCreateFunctionDeclaration(genericFunDecl, specificOutputType);
			if (specificType != null) {
				return specificType;
			}
		}
		return null;
	}

	public static Iterator getGenericFunctionDeclarations(Iterator functionDeclarations) {
		return CollectionsHelper.select(functionDeclarations, new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return obj instanceof IGenericFunctionDeclaration;
			}
		});
	}

	public static final ISpecificFunctionDeclaration getSpecificFunctionDeclaration(
		Iterator functionDecls,
		String name,
		ISpecificType outputType) {

		Iterator it = getFunctionDeclarationsByName(functionDecls, name);

		if (it.hasNext())
			return (ISpecificFunctionDeclaration) it.next();
		else
			return null;
	}
}