package de.uniwue.ki.mas.basics.search;

import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.feature.user.IUserFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.functionArgument.*;
import de.uniwue.ki.mas.basics.functionCall.*;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.argument.IArgumentDeclaration;
import de.uniwue.ki.mas.basics.type.composed.IComposedTypeAttribute;
import de.uniwue.ki.mas.basics.type.composed.SpecificComposedTypeAttributeType;
import de.uniwue.ki.mas.basics.type.enum.EnumItem;
import de.uniwue.ki.mas.basics.type.user.IUserTypeDeclaration;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.IActionActivityGraphLeafDeclaration;
import de.uniwue.ki.mas.behaviour.rule.IActivityGraphRuleDeclaration;
import de.uniwue.ki.mas.entity.body.IBodyInstanceDeclaration;
import de.uniwue.ki.mas.entity.bodyclass.IBodyClassDeclaration;
import de.uniwue.ki.mas.entity.object.IRessourceInstanceDeclaration;
import de.uniwue.ki.mas.entity.object.agent.IActivityAgentInstanceDeclaration;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclaration;
import de.uniwue.ki.mas.entity.variable.IBodyVariableInstanceDeclaration;
import de.uniwue.ki.mas.entity.variableclass.*;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.model.tree.UpdateChildrenAbleNodeThread;
import de.uniwue.ki.mas.sim.analysis.IAnalysisDeclaration;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.sim.situation.ISituationDeclaration;
import de.uniwue.ki.mas.util.declaration.IDeclaration;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IMapFunction;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.tree.IObjectTreeNode;

/**
 * @author Rainer Herrler
 */

public class SearchReferencesHelper {

	public interface IUnwrapper {
		Object unWrap(Object obj);
	}

	private static final List _unwrappers = new LinkedList();

	static {
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				return unWrapArgumentDeclaration(obj);
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IVariableClassDeclaration)
					return ((IVariableClassDeclaration)obj).getDefaultValueFunctionCallDeclaration();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IVariableClassDeclaration)
					return ((IVariableClassDeclaration)obj).getNextValueFunctionCallDeclaration();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IBodyFeatureVariableClassDeclaration)
					return ((IBodyFeatureVariableClassDeclaration)obj)
						.getDefaultValueFunctionCallDeclaration();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IBodyVariableInstanceDeclaration)
					return ((IBodyVariableInstanceDeclaration)obj).getDefaultValueFunctionCallDeclaration();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IComposedFunctionCallInternDeclaration)
					return ((IComposedFunctionCallInternDeclaration)obj).getChildren();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IBodyClassDeclaration)
					return ((IBodyClassDeclaration)obj)
						.getBodyVariableClassDeclarationListAndFactory()
						.getDeclarationList()
						.iterator();
				else
					return null;
			}
		});
		_unwrappers.add(new IUnwrapper() {
			public Object unWrap(Object obj) {
				if (obj instanceof IBodyInstanceDeclaration)
					return ((IBodyInstanceDeclaration)obj)
						.getBodyVariableInstanceDeclarationList()
						.iterator();
				else
					return null;
			}
		});

		Collections.reverse(_unwrappers);
	}

	// gets the first editable definition in the path to the node
	private static IDeclaration getDefinition(IObjectTreeNode node) {
		if (isEditableDeclaration(node.getUserObject())) {
			return (IDeclaration) (node.getUserObject());
		}

		if (node.getParent() == null)
			return null;

		return getDefinition((IObjectTreeNode)node.getParent());
	}

	/**
	 * returns an iterator of IDeclarations 
	 */
	public static Iterator findReferences(
		final IDeclaration searchedDeclaration,
		ISeSAmDeclaration sesamDeclaration) {

		UpdateChildrenAbleNodeThread.getInstance().forceUpdateChildren();

		IObjectTreeNode treeModelRoot =
			(IObjectTreeNode)sesamDeclaration.getSeSAmDeclarationTreeModel().getRoot();
		
		return findReferencesFrom(searchedDeclaration,treeModelRoot,sesamDeclaration);
	}
	
	private static Iterator findReferencesFrom(
			final IDeclaration searchedDeclaration,final IObjectTreeNode treeModelName,
			ISeSAmDeclaration sesamDeclaration) {

		UpdateChildrenAbleNodeThread.getInstance().forceUpdateChildren();

		Assert.assertTrue(
				"Tried to search for " + searchedDeclaration + " which is not possible",
				isSearchAbleDeclaration(searchedDeclaration));

		Set foundRefSet = new HashSet();
		findReferenceRek(treeModelName, searchedDeclaration, foundRefSet);

		Iterator foundDeclarations =
			CollectionsHelper.select(CollectionsHelper.map(foundRefSet.iterator(), new IMapFunction() {
				public Object map(Object obj) throws Exception {
					return getDefinition((IObjectTreeNode)obj);
				}
			}), new IPredicateFunction() {
				public boolean isElementOk(Object obj) throws Exception {
					return (obj != null) && (!searchedDeclaration.equals(obj));
				}
			});
		return CollectionsHelper.removeDuplicates(foundDeclarations);
	}

	/**
	 * this declaration (anchestor of the searched obj) can be returned as result
	 */
	private static boolean isEditableDeclaration(Object declaration) {
		return (declaration instanceof IUserFunctionDeclaration)
			|| (declaration instanceof IUserFeatureClassDeclaration)
			|| (declaration instanceof IObjectClassDeclaration)
			|| (declaration instanceof IActionActivityGraphLeafDeclaration)
			|| (declaration instanceof IActivityGraphRuleDeclaration)
			|| (declaration instanceof ISituationDeclaration)
			|| (declaration instanceof ISimulationDeclaration)
			|| (declaration instanceof IUserTypeDeclaration)
			|| (declaration instanceof IAnalysisDeclaration);
	}

	// this types of declaration can be searched
	public static boolean isSearchAbleDeclaration(final Object searchedDeclaration) {
		return (
			(searchedDeclaration instanceof IUserFunctionDeclaration)
			|| (searchedDeclaration instanceof IVariableClassDeclaration)
			|| (searchedDeclaration instanceof IBodyVariableClassDeclaration)
			|| (searchedDeclaration instanceof IUserTypeDeclaration)

		//				|| (searchedDeclaration instanceof IActivityAgentClassDeclaration)
		//				|| (searchedDeclaration instanceof IRessourceClassDeclaration)
		|| (searchedDeclaration instanceof IObjectClassDeclaration)
		|| (searchedDeclaration instanceof ISituationDeclaration)
		|| (searchedDeclaration instanceof ISimulationDeclaration)
		|| (searchedDeclaration instanceof IAnalysisDeclaration)
		|| (searchedDeclaration instanceof EnumItem)
		|| (searchedDeclaration instanceof SpecificComposedTypeAttributeType));
	}

	/**
	 * Finds recursivly all nodes containing a certain declaration, these are added 
	 * to the given Hash-Set
	 */
	private static void findReferenceRek(
		IObjectTreeNode node,
		IDeclaration declarationToFind,
		Set foundRefSet) {

		if (nodeContainsDeclaration(declarationToFind, node.getUserObject(), new HashSet())) {
			//foundRefSet.add((IObjectTreeNode)node.getParent());
			foundRefSet.add((IObjectTreeNode)node);
		}

		//search deeper
		if (!node.isLeaf())
			for (int i = 0;(i < node.getChildCount()); i++) {
				findReferenceRek((IObjectTreeNode)node.getChildAt(i), declarationToFind, foundRefSet);
			}

		if (declarationToFind instanceof IVariableClassDeclaration) {
			findReferenceRek(
				node,
				((IVariableClassDeclaration)declarationToFind).getBodyVariableClassDeclarations(),
				foundRefSet);
		}
	}

	private static void findReferenceRek(
		IObjectTreeNode node,
		Iterator declarationsToFind,
		Set foundRefSet) {

		while (declarationsToFind.hasNext()) {
			findReferenceRek(node, (IDeclaration)declarationsToFind.next(), foundRefSet);
		}
	}

	/* wollen nur IBodyVariableclassDeclaration, IUserFunctionDeclaration */
	private static Object unWrapArgumentDeclaration(Object decl) {

		if (decl instanceof IFunctionArgumentDeclaration) {
			return null;

		} else if (decl instanceof IArgumentDeclaration) {
			Object obj = ((IArgumentDeclaration)decl).getObject();
			if (obj instanceof IDeclaration) {
				return obj;
			} else {
				return null;
			}

		} else if (decl instanceof ISpecificFunctionDeclarationWithArgDeclBinds) {
			return ((ISpecificFunctionDeclarationWithArgDeclBinds)decl)
				.getParametricFunctionCallDeclaration();

		} else if (decl instanceof IAtomicFunctionCallInternDeclaration) {
			return ((IAtomicFunctionCallInternDeclaration)decl).getArgumentDeclaration();

		} else if (decl instanceof IParametricFunctionCallDeclaration) {
			return ((IParametricFunctionCallDeclaration)decl).getFunctionCallDeclaration();

		} else if (decl instanceof IFunctionCallDeclaration) {
			return ((IFunctionCallDeclaration)decl).getFunctionCallInternDeclaration();

		} else if (decl instanceof IComposedFunctionCallInternDeclaration) {
			return ((IComposedFunctionCallInternDeclaration)decl).getSpecificFunctionDeclaration();

		} else if (decl instanceof IActivityAgentInstanceDeclaration) {
			return ((IActivityAgentInstanceDeclaration)decl).getActivityAgentClassDeclaration();

		} else if (decl instanceof IRessourceInstanceDeclaration) {
			return ((IRessourceInstanceDeclaration)decl).getObjectClassDeclaration();

		} else if (decl instanceof IBodyVariableInstanceDeclaration) {
			return ((IBodyVariableInstanceDeclaration)decl).getBodyVariableClassDeclaration();

		} else if (decl instanceof IBodyVariableClassDeclaration) {
			return ((IBodyVariableClassDeclaration)decl).getVariableClassDeclaration();

		} else {
			return null;
		}
	}

	private static boolean nodeContainsDeclaration(
		IDeclaration declarationToFind,
		Object nodeobj,
		Set visitedNodeObjSet) {

		if (nodeobj instanceof Iterator) {
			while (((Iterator)nodeobj).hasNext()) {
				Object element = ((Iterator)nodeobj).next();
				if (nodeContainsDeclaration(declarationToFind, element, visitedNodeObjSet))
					return true;
			}
			return false;

		} else if (visitedNodeObjSet.contains(nodeobj)) {
			return false;

		} else {

			visitedNodeObjSet.add(nodeobj);

			if (declarationToFind instanceof IUserTypeDeclaration) {
				ISpecificType typeToFind = ((IUserTypeDeclaration)declarationToFind).getSpecificType();

				if (nodeobj instanceof IComposedTypeAttribute)
					return (((IComposedTypeAttribute)nodeobj).getAttributeType() == typeToFind);

				if ((nodeobj instanceof IUserTypeDeclaration) && ((IUserTypeDeclaration)nodeobj).isComposedType()) 
					return (
						(((IUserTypeDeclaration)nodeobj).getSpecificComposedType() != null)
							&& ((IUserTypeDeclaration)nodeobj).getSpecificComposedType().getFather().equals(
								typeToFind));

				if (nodeobj instanceof IParametricFunctionCallDeclaration)
					return CollectionsHelper.contains(
						((IParametricFunctionCallDeclaration)nodeobj).getReferencedSpecificTypes(),
						typeToFind);

				if (nodeobj instanceof IFeatureVariableClassDeclaration)
					return ((IFeatureVariableClassDeclaration)nodeobj).getSpecificType().equals(typeToFind);

				// unwrap from IAtomicFunctionCallInternDeclaration
				if (unWrapArgumentDeclaration(nodeobj) instanceof IArgumentDeclaration)
					return ((IArgumentDeclaration)unWrapArgumentDeclaration(nodeobj))
						.getSpecificType()
						.equals(
						typeToFind);

				// unwrap from IBodyVariableClassDeclaration
				if (unWrapArgumentDeclaration(nodeobj) instanceof IVariableClassDeclaration) {
					return ((IVariableClassDeclaration)unWrapArgumentDeclaration(nodeobj))
						.getSpecificType()
						.equals(
						typeToFind);
				}

				if (nodeobj instanceof IComposedFunctionCallInternDeclaration) {
					return (
						((IComposedFunctionCallInternDeclaration)nodeobj).getSpecificOutputType()
							== typeToFind);
				}

				return false;
			}

			if (Utils.equals(declarationToFind, nodeobj))
				return true;

			Iterator it = _unwrappers.iterator();
			while (it.hasNext()) {
				IUnwrapper unwrapper = (IUnwrapper)it.next();
				Object unWrappedObj = unwrapper.unWrap(nodeobj);
				if (unWrappedObj != null) {
					if (Utils.equals(declarationToFind, unWrappedObj)) {
						return true;
					} else if (nodeContainsDeclaration(declarationToFind, unWrappedObj, visitedNodeObjSet)) {
						return true;
					}
				}
			}

			return false;
		}
	}
}
