package de.uniwue.ki.mas.basics.type;
import java.util.Iterator;
import java.util.Map;

import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.type.argument.IArgument;
import de.uniwue.ki.mas.basics.type.argument.IArgumentDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;

/**
 * A specific type is a 'real' type like number or boolean.
 * It knows which 'instances' of this type are ok and how to produce a name for them,
 * for example the {@link NumberType} knows that all 'instances' are {@link Number Numbers}.
 * Creation date: (29.08.2000 14:59:51)
 *
 * @author Christoph Oechslein
 */

public interface ISpecificType extends IType {

	/**
	 * This should return the generic type, of which this specific type is composed.
	 * Returns <code>null</code> if this specific type is not generated by a generic type.
	 */
	IGenericType getGenericType();

	/**
	 * Add a new subtype.
	 */
	void addSubType(ISpecificType newSubType);
	
	/**
	 * Get all subtypes of this type.
	 */
	Iterator getSubTypes();

	/** 
	 * See TypeManager#isMoreGeneral for details.
	 */
	boolean isMoreSpecific(ISpecificType type);

	/**
	 * @see IFunctionCallInternDeclaration#isConstant(ICreateFunctionArgs)
	 */
	boolean isConstant(ICreateFunctionArgs createFunctionArgs, Object anObject) throws ModelCompilingException;
	
	/**
	 * @see IFunctionCallInternDeclaration#isReadOnly(ICreateFunctionArgs)
	 */
	boolean isReadOnly(ICreateFunctionArgs createFunctionArgs, Object anObject) throws ModelCompilingException;
	
	//
	// for argument objects

	/**
	 * Generates an <code>IArgument</code> with the specified instance.
	 *
	 * @preconditions   isValueOk(value) must return true for the <code>value</code> argument.
	 * @see IFunction.createFunction(ICreateFunctionArgs, IFunctionList)
	 */
	IArgument createArgument(ICreateFunctionArgs createFunctionArgs, Object value)
		throws ModelCompilingException;

	/**
	  * Produces a string representation of a runtime instance of this type.
	  */
	void appendNameForArgumentObject(StringBuffer buff, Object anObject, int currDepth, int maxDepth);

	Object copyArgumentObject(Object anObject, Map objectMapping);

	void removeArgumentObject(Object anObject);

	//
	// for argument declarations

	/**
	 * Creates a default function call intern declaration for this type
	 * (Introduced because of VoidType which generates a composed function call with 'Noop').
	 */
	IFunctionCallInternDeclaration createDefaultFunctionCallInternDeclaration();

	/**
	 * Gets the default value of this type,
	 * for example for {@link NumberType} it's <code>0.0<c/ode>.
	 */
	Object getArgumentDeclarationDefaultValue();

	boolean isAtomicDefaultValue();

	/**
	  * Produces a string representation of an argument declaration object of this type.
	  */
	void appendNameForArgumentDeclarationObject(
		StringBuffer buff,
		Object anObject,
		int currDepth,
		int maxDepth);

	/**
	 * Returns true if the value is a correct instance of this type.
	 */
	boolean isArgumentDeclarationValueOk(Object value);

	IArgumentDeclaration copyArgumentDeclaration(IArgumentDeclaration argumentDeclaration, Map objectMapping);

	IArgumentDeclaration createArgumentDeclaration(Object value);

	/**
	 * Gets Called if an argument declaration with this type is removed.
	 */
	void removeArgumentDeclaration(IArgumentDeclaration argumentDeclaration);
}