package de.uniwue.ki.mas.basics.type;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.functionArgument.*;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionDeclaration.ISpecificFunctionDeclaration;

/**
 * Creation date: (17.09.2001 16:54:50)
 * @author Christoph Oechslein
 */

public final class SpecificFunctionType extends AbstractSpecificFunctionType {
	private final List _inputTypes;

	SpecificFunctionType(ISpecificType outputType, List inputTypes) {
		super(outputType);

		_inputTypes = inputTypes;
	}

	public boolean canCreate(ISpecificType type) {
		return equals(type);
	}

	/**
	* Compares two objects for equality. Returns a boolean that indicates
	* whether this object is equivalent to the specified object. This method
	* is used when an object is stored in a hashtable.
	* @param obj the Object to compare with
	* @return true if these Objects are equal; false otherwise.
	* @see java.util.Hashtable
	*/
	public boolean equals(Object obj) {
		// NOTE: obj might be an instance of any class
		if (!(obj instanceof ISpecificFunctionType)) {
			return false;
		}

		ISpecificFunctionType otherSpecificFunctionType = (ISpecificFunctionType)obj;
		if (!getSpecificOutputType().equals(otherSpecificFunctionType.getSpecificOutputType()))
			return false;

		if (getSpecificInputTypesLength() != otherSpecificFunctionType.getSpecificInputTypesLength())
			return false;

		Iterator it = getSpecificInputTypes();
		int index = 0;
		while (it.hasNext()) {
			ISpecificType currSpecificType = (ISpecificType)it.next();

			if (!currSpecificType.equals(otherSpecificFunctionType.getSpecificInputTypeAt(index)))
				return false;
			++index;
		}

		return true;
	}

	public boolean isMoreSpecific(ISpecificType type) {
		if (super.isMoreSpecific(type)) {
			return true;

		} else if (!(type instanceof ISpecificFunctionType)) {
			return false;

		} else {
			ISpecificFunctionType otherSpecificFunctionType = (ISpecificFunctionType)type;
			if (!getSpecificOutputType().isMoreSpecific(otherSpecificFunctionType.getSpecificOutputType()))
				return false;

			if (getSpecificInputTypesLength() != otherSpecificFunctionType.getSpecificInputTypesLength())
				return false;

			Iterator it = getSpecificInputTypes();
			int index = 0;
			while (it.hasNext()) {
				ISpecificType currSpecificType = (ISpecificType)it.next();

				if (!currSpecificType
					.isMoreSpecific(otherSpecificFunctionType.getSpecificInputTypeAt(index)))
					return false;
				++index;
			}

			return true;
		}
	}

	private SpecificFunctionDeclarationWithArgDeclBinds _argumentDeclarationDefaultValue = null;

	public Object getArgumentDeclarationDefaultValue() {
		if (_argumentDeclarationDefaultValue == null)
			_argumentDeclarationDefaultValue = getArgumentDeclarationDefaultValueIntern();

		//return _argumentDeclarationDefaultValue;
		//  
		//Default Value wird weiterverwendet und von Remove geloescht
		//-> Beim Zweiten mal steht ein unsinniger Wert im defaultValue
		return _argumentDeclarationDefaultValue.copy(new HashMap());
	}

	private SpecificFunctionDeclarationWithArgDeclBinds getArgumentDeclarationDefaultValueIntern() {
		SpecificFunctionDeclarationWithArgDeclBinds specificFunctionDeclarationWithArgDeclBinds =
			SpecificFunctionDeclarationWithArgDeclBindsFactory
				.createSpecificFunctionDeclarationWithArgDeclBinds();

		specificFunctionDeclarationWithArgDeclBinds.setSpecificOutputType(getSpecificOutputType());

		Iterator it = getSpecificInputTypes();
		while (it.hasNext()) {
			ISpecificType currSpecificType = (ISpecificType)it.next();

			specificFunctionDeclarationWithArgDeclBinds.getFunctionArgumentDeclarationList().add(
				FunctionArgumentDeclarationFactory.getInstance().createArgumentDeclaration(currSpecificType));
		}

		return specificFunctionDeclarationWithArgDeclBinds;
	}

	public String getName() {
		StringBuffer buff = new StringBuffer();
		buff.append("Fun<(");
		Iterator it = getSpecificInputTypes();
		while (it.hasNext()) {
			buff.append(((IType)it.next()).getName());
			if (it.hasNext()) {
				buff.append(", ");
			}
		}
		buff.append(") ");
		buff.append(getSpecificOutputType().getName());
		buff.append(">");
		return buff.toString();
	}

	public ISpecificType getSpecificInputTypeAt(int index) {
		return (ISpecificType)_inputTypes.get(index);
	}

	public final Iterator getSpecificInputTypes() {
		return _inputTypes.iterator();
	}

	public int getSpecificInputTypesLength() {
		return _inputTypes.size();
	}

	/**
	 * Generates a hash code for the receiver.
	 * This method is supported primarily for
	 * hash tables, such as those provided in java.util.
	 * @return an integer hash code for the receiver
	 * @see java.util.Hashtable
	 */
	public int hashCode() {
		// NOTE: if two objects are equal (equals(Object) returns true) they must have the same hash code
		// The new try...
		if (getSpecificInputTypes() != null && getSpecificInputTypes().hasNext()) {
			return 7 * getSpecificOutputType().hashCode()
				+ ((ISpecificType)getSpecificInputTypes().next()).hashCode();
		} else {
			return getSpecificOutputType().hashCode();
		}
	}

	public boolean hasSameSignature(ISpecificFunctionDeclaration specificFunctionDeclaration) {
		if (!TypeManager
			.getInstance()
			.isMoreGeneral(getSpecificOutputType(), specificFunctionDeclaration.getSpecificOutputType())) {
			return false;
		}

		if ((specificFunctionDeclaration.getSpecificInputMaxArgsLength()
			!= ISpecificFunctionDeclaration.UNLIMITED_ARGUMENT_COUNT)
			&& (getSpecificInputTypesLength() > specificFunctionDeclaration.getSpecificInputMaxArgsLength())) {
			return false;
		}

		if (getSpecificInputTypesLength() < specificFunctionDeclaration.getSpecificInputMinArgsLength()) {
			return false;
		}

		for (int i = 0; i < getSpecificInputTypesLength(); i++) {
			if (!TypeManager
				.getInstance()
				.isMoreGeneral(
					getSpecificInputTypeAt(i),
					specificFunctionDeclaration.getSpecificInputTypeAt(i))) {
				return false;
			}
		}
		return true;
	}
	
	public static void main(String[] arguments) {
		ISpecificType type1 = GenericFunctionType.getInstance().createDefaultSpecificType();
		ISpecificType type2 = GenericFunctionType.getInstance().createDefaultSpecificType();
		IFunctionCallInternDeclaration decl1 = type1.createDefaultFunctionCallInternDeclaration();
		IFunctionCallInternDeclaration decl2 = type1.createDefaultFunctionCallInternDeclaration();
		
		Assert.assertTrue(decl1.equals(decl2));
		Assert.assertTrue(type1.equals(type2));
	}

}