package de.uniwue.ki.mas.basics.type;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.type.composed.*;
import de.uniwue.ki.mas.basics.type.enum.GenericEnumType;
import de.uniwue.ki.mas.basics.type.enum.ISpecificEnumType;
import de.uniwue.ki.mas.basics.type.user.IUserTypeDeclaration;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.comparator.NamedComparator;
import de.uniwue.ki.util.functions.IMapFunction;
import de.uniwue.ki.util.functions.IPredicateFunction;

/**
 * Creation date: (30.07.2001 14:15:54)
 * @author Christoph Oechslein
 */

public final class TypeManager {
	private static final TypeManager _instance = new TypeManager();

	// List of all Types without Composed and Enum Types
	private final List _allTypes = new ArrayList();

	private final List _allTypesForVariables = new ArrayList();

	// Composed and Enum Types
	private final List _allAbstractTypes = new ArrayList();

	public Iterator getComplexTypes() {
		return _allAbstractTypes.iterator();
	}

	/**
	 * returns enumeration types defined by the system and plugins 
	 */
	public Iterator getSystemEnumTypes() {
		return getAllEnumTypes(null);
	}

	/**
	  * returns composed types defined by the system and plugins 
	  */
	public Iterator getSystemComposedTypes() {
		return getAllComposedTypes(null);
	}

	/**
	 * returns model defined composed types as well as system defined composed types 
	 */
	public Iterator getAllComposedTypes(ISeSAmDeclaration seSAmDeclaration) {
		return CollectionsHelper.select(getAllAbstractTypes(seSAmDeclaration), new IPredicateFunction() {
			public boolean isElementOk(Object obj) throws Exception {
				return (obj instanceof ISpecificComposedType);
			}

		});
	}

	/**
	 * returns model defined enumeration types as well as system defined enumeration types 
	 */
	public Iterator getAllEnumTypes(ISeSAmDeclaration seSAmDeclaration) {
		return CollectionsHelper.select(getAllAbstractTypes(seSAmDeclaration), new IPredicateFunction() {
			public boolean isElementOk(Object obj) throws Exception {
				return (obj instanceof ISpecificEnumType);
			}
		});
	}

	/**
	  * returns user defined types as well as system defined abstract types 
	  */
	public Iterator getAllAbstractTypes(ISeSAmDeclaration seSAmDeclaration) {
		Iterator modelDefinedAbstractTypes = CollectionsHelper.EMPTY_ITERATOR;
		if (seSAmDeclaration != null) {
			modelDefinedAbstractTypes =
				CollectionsHelper
					.map(
						seSAmDeclaration
							.getSeSAmModelDeclaration()
							.getSeSAmModelBasicsDeclaration()
							.getUserTypeDeclarationListAndFactory()
							.getUserTypeDeclarationList()
							.iterator(),
						new IMapFunction() {
				public Object map(Object obj) throws Exception {
					IUserTypeDeclaration typeDecl = ((IUserTypeDeclaration)obj);
					if (typeDecl.isComposedType())
						return typeDecl.getSpecificComposedType();
					else
						return typeDecl.getSpecificEnumType();
				}
			});
		}

		Iterator systemDefinedAbstractTypes = _allAbstractTypes.iterator();
		Iterator allTypesIterator =
			CollectionsHelper.concatenate(systemDefinedAbstractTypes, modelDefinedAbstractTypes);

		return allTypesIterator;
	}

	protected TypeManager() {
		super();
		initialize();
	}

	public final void addGenericType(IGenericType newType) {
		addType(newType);
	}

	public final void addSpecificType(ISpecificType newType) {
		addType(newType);
	}

	private void addType(IType newType) {

		if (!_allTypes.contains(newType)) {
			_allTypes.add(newType);

			// should be done explicitly
			//			if (newType instanceof ISpecificType) {
			//				ISpecificType newSpecificType = (ISpecificType) newType;
			//				Iterator it = newSpecificType.getSubTypes();
			//				while (it.hasNext()) {
			//					addType((ISpecificType) it.next());
			//				}
			//			}
		}

		Collections.sort(_allTypes, NamedComparator.getInstance());
	}

	public void removeType(IType oldType) {

		if (_allTypes.contains(oldType)) {
			_allTypes.remove(oldType);

			if (oldType instanceof ISpecificType) {
				ISpecificType oldSpecificType = (ISpecificType)oldType;
				Iterator it = oldSpecificType.getSubTypes();
				while (it.hasNext()) {
					removeType((ISpecificType)it.next());
				}
			}
		}

		Collections.sort(_allTypes, NamedComparator.getInstance());
	}

	/*
	 * To add a basic (non-model-dependent UserType like a ComposedType
	 * or an EnumType
	 */
	private final void addSystemAbstractTypeIntern(ISpecificType newType) {
		if (Assert.ASSERTION_ON) {
			Assert.assertTrue(
				"Type has to be of ISpecificEnumType or ISpecificComposedType",
				((newType instanceof ISpecificEnumType) || (newType instanceof ISpecificComposedType)));
		}

		if (!_allAbstractTypes.contains(newType)) {
			_allAbstractTypes.add(newType);
		}

		Collections.sort(_allAbstractTypes, NamedComparator.getInstance());
	}

	public final void addSystemAbstractType(ISpecificEnumType newType) {
		addSystemAbstractTypeIntern(newType);
	}

	public final void addSystemAbstractType(ISpecificComposedType newType) {
		addSystemAbstractTypeIntern(newType);
	}

	public final void addVariableGenericType(IGenericType newVariableType) {
		addVariableType(newVariableType);
	}

	public final void addVariableSpecificType(ISpecificType newVariableType) {

		// due to a Problem with SpecificTypePanel
		if (Assert.ASSERTION_ON)
			Assert.assertNull(newVariableType.getGenericType());

		addVariableType(newVariableType);
	}

	private final void addVariableType(IType newVariableType) {

		if (!_allTypesForVariables.contains(newVariableType)) {

			// due to a Problem with SpecificTypePanel
			Assert.assertTrue(
				(newVariableType instanceof IGenericType)
					|| ((newVariableType instanceof ISpecificType)
						&& (((ISpecificType)newVariableType).getGenericType() == null)));

			_allTypesForVariables.add(newVariableType);
		}

		Collections.sort(_allTypesForVariables, NamedComparator.getInstance());
	}

	public Iterator getAllSpecificTypes() {
		return CollectionsHelper.select(getAllTypes(), new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return obj instanceof ISpecificType;
			}
		});
	}

	public final Iterator getAllTypes() {
		return _allTypes.iterator();
	}

	public final Iterator getAllTypesForVariableClassDeclaration() {
		return _allTypesForVariables.iterator();
	}

	public final ISpecificType getDefaultType() {
		return DoubleType.getInstance();
	}

	public final ISpecificType getDefaultTypeForVariableClassDeclaration() {
		return DoubleType.getInstance();
	}

	public static final TypeManager getInstance() {
		return _instance;
	}

	public ISpecificType getSpecificTypeByName(String name) {

		Iterator it = getAllSpecificTypes();
		return getSpecificTypeByName(name, it);
	}

	public static ISpecificType getSpecificTypeByName(String name, Iterator it) {
		while (it.hasNext()) {
			ISpecificType currISpecificType = (ISpecificType)it.next();

			if (currISpecificType.getName().equalsIgnoreCase(name)) {
				return currISpecificType;
			}
		}

		return null;
	}

	public static List getSubTypes(ISpecificType type) {
		List newList = new LinkedList();
		if (type.isAtomicDefaultValue()) {
			newList.add(type);
		}

		getSubTypes(newList, type);
		return newList;
	}

	private static void getSubTypes(List newList, ISpecificType type) {
		Iterator it;
		it = type.getSubTypes();
		while (it.hasNext()) {
			ISpecificType currType = (ISpecificType)it.next();
			if (!newList.contains(currType)) {
				newList.add(currType);
				getSubTypes(newList, currType);
			}
		}
	}

	private void initialize() {
		addSpecificType(ObjectType.getInstance());
		addSpecificType(BooleanType.getInstance());
		//		addSpecificType(DoubleType.getInstance());
		//		addSpecificType(IntegerType.getInstance());
		addSpecificType(StringType.getInstance());
		addSpecificType(ActivityType.getInstance());
		addSpecificType(ObjectClassDeclarationType.getInstance());
		addSpecificType(ObjectInstanceDeclarationType.getInstance());
		//addSpecificType(SimObjectType.getInstance());

		addSpecificType(SimulationDeclarationType.getInstance());
		addSpecificType(SimulationType.getInstance());

		addGenericType(GenericNumberType.getInstance());
		addGenericType(GenericListType.getInstance());
		addGenericType(GenericIteratorType.getInstance());
		addGenericType(GenericVariableClassDeclarationType.getInstance());
		addGenericType(GenericFunctionType.getInstance());
		addGenericType(GenericHashtableType.getInstance());

		//		addGenericType(GenericEnumTypeItemType.getInstance());
		//		addGenericType(GenericComposedTypeAttributeType.getInstance());

		addGenericType(GenericSimObjectType.getInstance());
		addGenericType(GenericComposedType.getInstance());
		addGenericType(GenericEnumType.getInstance());

		addSpecificType(VoidType.getInstance());

		addVariableSpecificType(BooleanType.getInstance());
		//		addVariableSpecificType(DoubleType.getInstance());
		//		addVariableSpecificType(IntegerType.getInstance());
		addVariableSpecificType(StringType.getInstance());
		addVariableSpecificType(ActivityType.getInstance());
		// addVariableSpecificType(ObjectClassDeclarationType.getInstance());
		//addVariableSpecificType(SimObjectType.getInstance());
		addVariableGenericType(GenericSimObjectType.getInstance());

		addVariableGenericType(GenericNumberType.getInstance());
		//addVariableGenericType(GenericIteratorType.getInstance());
		addVariableGenericType(GenericListType.getInstance());
		addVariableGenericType(GenericHashtableType.getInstance());
		addVariableGenericType(GenericComposedType.getInstance());
		addVariableGenericType(GenericEnumType.getInstance());

		addSystemAbstractType(RootComposedType.getInstance());
	}

	/**
	 * Checks if type is directly in getAllTypes or indirectly by more general
	 */
	public final boolean isInAllTypes(IType type) {
		if (type instanceof ISpecificComposedType) {
			return true;
		} else {
			Iterator allTypes = getAllTypes();
			while (allTypes.hasNext()) {
				IType currType = (IType)allTypes.next();
				if (currType instanceof IGenericType) {
					if ((type instanceof ISpecificType)
						&& ((IGenericType)currType).canCreate((ISpecificType)type)) {
						return true;
					}

				} else if (((ISpecificType)currType).equals(type)) {
					return true;

				}
			}

			return false;
		}
	}

	/**
	 * Checks if type is directly in getAllTypesForVariableClassDeclaration or indirectly by inheritence
	 * E.g. object is not in AllTypesForVariableClassDeclaration but isInAllTypesForVariableClassDeclaration returns true
	 */
	public final boolean isInAllTypesForVariableClassDeclaration(ISpecificType type) {
		if (type instanceof ISpecificComposedType) {
			return true;
		} else {
			Iterator allTypes = getAllTypesForVariableClassDeclaration();

			while (allTypes.hasNext()) {
				IType currType = (IType)allTypes.next();

				if (currType instanceof IGenericType) {
					if (((IGenericType)currType).canCreate(type)) {
						return true;
					}

				} else if (isMoreGeneral(type, (ISpecificType)currType)) {
					return true;
				}
			}

			return false;
		}
	}

	/** 
	 * Returns true if specificType is more Specific than generalType.
	 * 
	 * examples:
	 * TypeFactory.getInstance().isMoreGeneral(
	 *      NumberType.getInstance(),
	 *      ObjectType.getInstance())
	 * // -> false
	 *
	 * TypeFactory.getInstance().isMoreGeneral(
	 *      ObjectType.getInstance(),
	 *      NumberType.getInstance())
	 * // -> true
	 *
	 * TypeFactory.getInstance().isMoreGeneral(
	 *      GenericSetType.getInstance().createSpecificSetType(
	 *           NumberType.getInstance()),
	 *      GenericSetType.getInstance())
	 * // -> false
	 *
	 * TypeFactory.getInstance().isMoreGeneral(
	 *      GenericSetType.getInstance(),
	 *      GenericSetType.getInstance().createSpecificSetType(
	 *           NumberType.getInstance()))
	 * // -> true
	 */
	public final boolean isMoreGeneral(ISpecificType generalType, ISpecificType specificType) {
		return generalType.isMoreSpecific(specificType);
	}

	public final void removeVariableType(IType newVariableType) {
		_allTypesForVariables.remove(newVariableType);
	}

	public final ISpecificType getMostGeneralType(Iterator specificTypes) {
		Assert.assertTrue(specificTypes.hasNext());

		ISpecificType specificType = (ISpecificType)specificTypes.next();

		while (specificTypes.hasNext() && (specificType != null)) {
			specificType =
				getInstance().getMostGeneralType(specificType, (ISpecificType)specificTypes.next());
		}

		return specificType;
	}

	public final ISpecificType getMostGeneralType(ISpecificType specificType1, ISpecificType specificType2) {
		if (Assert.ASSERTION_ON) {
			Assert.assertNotNull(specificType1);
			Assert.assertNotNull(specificType2);
		}

		if (specificType1.isMoreSpecific(specificType2)) {
			return specificType1;
		} else if (specificType2.isMoreSpecific(specificType1)) {
			return specificType2;

		} else {
			return null;
		}
	}
}