package de.uniwue.ki.mas.behaviour.activityGraph;
import junit.framework.Assert;
import de.uniwue.ki.mas.basics.action.FunctionActionFactory;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.IActivityGraphLeafDeclaration;
import de.uniwue.ki.mas.behaviour.rule.IActivityGraphRule;
import de.uniwue.ki.mas.entity.object.agent.IActivityAgent;
import de.uniwue.ki.mas.entity.object.agent.IAgentExecuteFunctionArgs;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * Creation date: (23.08.2001 19:41:37)
 * @author Christoph Oechslein
 */

abstract class AbstractActivityGraphLeaf extends AbstractActivityGraphNode implements IActivityGraphLeaf {
	private ICalculateNextActivityGraphLeafAble _calculateNextActivityGraphLeafAble;
	private IActivityGraphLeaf _otherwiseToAtomicActivity;
	private IActivityGraphRule[] _activityGraphRules;
	private IActivityGraphRule[] _emergencyRules;

	public AbstractActivityGraphLeaf(IActivityGraphLeafDeclaration activityDeclaration) {
		super(activityDeclaration);

		_activityGraphRules = null;
		_emergencyRules = null;
	}

	private void updateCalculateNextActivity() {
		setCalculateNextActivityGraphLeafAble(createCalculateNextActivityGraphLeafAble(this));
	}

	private static final ICalculateNextActivityGraphLeafAble createCalculateNextActivityGraphLeafAble(AbstractActivityGraphLeaf atomicActivity) {
		if ((atomicActivity.getChainRules() == null)
			&& (atomicActivity.getOtherwiseToAtomicActivity() != null)) {

			return new SequenceCalculateNextActivity(
				atomicActivity,
				atomicActivity.getOtherwiseToAtomicActivity());
		}

		if ((atomicActivity.getOtherwiseToAtomicActivity() == null)
			&& (atomicActivity.getChainRules() != null)
			&& (atomicActivity.getChainRules().length == 1)) {

			if (atomicActivity.getChainRules()[0].isAlwaysExecuteable()) {
				return new SequenceCalculateNextActivity(
					atomicActivity,
					atomicActivity.getChainRules()[0].getActivityGraphLeaf());
			}
		}

		return new DefaultCalculateNextActivityGraphLeaf(atomicActivity);
	}


	public void calculateResult(
		IAgentExecuteFunctionArgs objectExecuteFunctionArgs,
		IActivityGraphLeafReasonResult atomicActivityReasonResult)
		throws SimulationRuntimeException {

		IActivityGraphLeaf nextAtomicActivity = calculateNextActivityGraphLeaf(objectExecuteFunctionArgs);

		if (nextAtomicActivity == null) {
			atomicActivityReasonResult.setCurrentActivityGraphLeafHasEnded(false);
			atomicActivityReasonResult.setAction(getFunctionAction());
			atomicActivityReasonResult.setNextActivityGraphLeaf(null);

		} else {
			atomicActivityReasonResult.setCurrentActivityGraphLeafHasEnded(true);
			atomicActivityReasonResult.setAction(
				getExitFunctionAction(),
				nextAtomicActivity.getEntryAndNormalFunctionAction());
			atomicActivityReasonResult.setNextActivityGraphLeaf(nextAtomicActivity);
		}
	}

	private final IActivityGraphLeaf calculateNextActivityGraphLeaf(IAgentExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		return getCalculateNextActivityGraphLeafAble().calculateNextActivityGraphLeaf(objectExecuteFunctionArgs);
	}


	public IActivityGraphLeaf getActivityGraphLeaf(IActivityAgent agent) {
		return this;
	}

	protected final IActivityGraphLeafDeclaration getAtomicActivityDeclaration() {
		return (IActivityGraphLeafDeclaration) getActivityGraphNodeDeclaration();
	}


	public final IActivityGraphRule[] getChainRules() {
		return _activityGraphRules;
	}

	public final IActivityGraphRule[] getEmergencyRules() {
		return _emergencyRules;
	}

	/**
	 * Returns an action sequence of the entry and the normal function (@link #getFunctionAction()),
	 * gets called if this activity has just started.
	 *
	 * @return IFunction
	 * @see #getExitFunctionAction()
	 * @see #getFunctionAction()
	 */
	public IFunction getEntryAndNormalFunctionAction() {
		return getFunctionAction();
	}

	/**
	 * Returns the cleanup function action of this activity, gets called if a new activity was selected.
	 *
	 * @see #getFunctionAction()
	 * @see #getEntryAndNormalFunctionAction()
	 */
	protected IFunction getExitFunctionAction() {
		return FunctionActionFactory.getNoopFunction();
	}

	/**
	 * Normal Action, gets called if in this activity but not for the first time.
	 *
	 * @see #getExitFunctionAction()
	 * @see #getEntryAndNormalFunctionAction()
	 */
	protected IFunction getFunctionAction() {
		return FunctionActionFactory.getNoopFunction();
	}

	public boolean isInstantly() {
		return true;
	}

	public void setActivityGraphRules(IActivityGraphRule[] chainRules) {
		if (chainRules == null)
			_activityGraphRules = null;
		else if (chainRules.length == 0)
			_activityGraphRules = null;
		else
			_activityGraphRules = chainRules;

		updateCalculateNextActivity();
	}


	public void setEmergencyRules(IActivityGraphRule[] emergencyRules) {
		if (emergencyRules == null)
			_emergencyRules = null;
		else if (emergencyRules.length == 0)
			_emergencyRules = null;
		else
			_emergencyRules = emergencyRules;
	}

	public IActivityGraphLeaf getOtherwiseToAtomicActivity() {
		return _otherwiseToAtomicActivity;
	}

	public void setOtherwiseToAtomicActivity(IActivityGraphLeaf newOtherwiseToAtomicActivity) {
		if (Assert.ASSERTION_ON) {
			Assert.assertNull(_otherwiseToAtomicActivity);
			Assert.assertNotNull(newOtherwiseToAtomicActivity);
		}

		_otherwiseToAtomicActivity = newOtherwiseToAtomicActivity;

		updateCalculateNextActivity();
	}

	private ICalculateNextActivityGraphLeafAble getCalculateNextActivityGraphLeafAble() {
		return _calculateNextActivityGraphLeafAble;
	}

	private void setCalculateNextActivityGraphLeafAble(ICalculateNextActivityGraphLeafAble calculateNextActivityGraphLeafAble) {
		_calculateNextActivityGraphLeafAble = calculateNextActivityGraphLeafAble;
	}
}