package de.uniwue.ki.mas.behaviour.activityGraph;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.Assert;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.*;
import de.uniwue.ki.mas.behaviour.rule.*;
import de.uniwue.ki.mas.entity.object.IObjectCreateFunctionArgs;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclaration;
import de.uniwue.ki.mas.model.ModelCompilingException;

/**
 * Creation date: (14.09.2001 09:38:17)
 * @author Christoph Oechslein
 */

final class ActivityGraphCompiler {
	private static final Logger _logger = Logger.getLogger(ActivityGraphCompiler.class.getName());
	private final Map _activityMap = new HashMap();

	private final IObjectCreateFunctionArgs _objectCreateFunctionArgs;

	private final ActivityGraph _rootActivityGraph;

	public ActivityGraphCompiler(
		ActivityGraph rootActivityGraph,
		IObjectCreateFunctionArgs objectCreateFunctionArgs) {
		super();

		_rootActivityGraph = rootActivityGraph;
		_objectCreateFunctionArgs = objectCreateFunctionArgs;
	}

	private void addActivityGraphEdges(
		IActivityGraphNodeDeclaration activityDeclaration,
		AbstractActivityGraphLeaf atomicActivity)
		throws ModelCompilingException {

		List activityGraphEdges = new ArrayList();
		Iterator it = activityDeclaration.getActivityGraphEdgeDeclarations();
		while (it.hasNext()) {
			IActivityGraphEdgeDeclaration activityEdgeDeclaration =
				((IActivityGraphEdgeDeclaration)it.next());

			if (activityEdgeDeclaration instanceof IActivityGraphRuleDeclaration) {
				IRule rule =
					((IActivityGraphRuleDeclaration)activityEdgeDeclaration).createRule(
						_objectCreateFunctionArgs);

				if (rule != null) {
					if (!activityGraphEdges.contains(rule))
						activityGraphEdges.add(rule);

				} else {
					if (Assert.ASSERTION_ON && _logger.isLoggable(Level.INFO))
						_logger.info(
							"Removed ActivityGraphEdge '"
								+ activityEdgeDeclaration.getName()
								+ "' from '"
								+ activityEdgeDeclaration.getFromActivityGraphNodeDeclaration().getName()
								+ "' to '"
								+ activityEdgeDeclaration.getToActivityGraphNodeDeclaration().getName()
								+ "' in '"
								+ getObjectClassDeclaration().getName()
								+ "'");
				}

			} else if (activityEdgeDeclaration instanceof IOtherwiseActivityGraphEdgeDeclaration) {
				atomicActivity.setOtherwiseToAtomicActivity(
					getActivityGraphLeaf(activityEdgeDeclaration.getToActivityGraphNodeDeclaration()));
			}
		}

		atomicActivity.setActivityGraphRules(
			(IActivityGraphRule[])activityGraphEdges.toArray(
				new IActivityGraphRule[activityGraphEdges.size()]));
	}

	private IObjectClassDeclaration getObjectClassDeclaration() {
		return _objectCreateFunctionArgs
			.getObject()
			.getObjectInstanceDeclaration()
			.getObjectClassDeclaration();
	}

	private void addActivityGraphEdgesAndEmergencyRules(
		IActivityGraphNodeDeclaration activityDeclaration,
		AbstractActivityGraphLeaf atomicActivity,
		IActivityGraphRule[] emergencyRules)
		throws ModelCompilingException {

		if (atomicActivity == null) {
			return;
		}

		if ((emergencyRules != null) && (emergencyRules.length > 0)) {
			// System.err.println("Added emergency rules to " + atomicActivity.getName() + ": " + emergencyRules);
			if (Assert.ASSERTION_ON && _logger.isLoggable(Level.FINE)) {
				StringBuffer buff = new StringBuffer();

				buff.append("Added emergency rules to ");
				buff.append(atomicActivity.getName());
				buff.append(": [");

				for (int i = 0; i < emergencyRules.length; i++) {
					buff.append(emergencyRules[i].getName());

					if (i < emergencyRules.length - 1)
						buff.append(", ");
				}

				buff.append("]");

				_logger.fine(buff.toString());
			}
			atomicActivity.setEmergencyRules(emergencyRules);
		}

		addActivityGraphEdges(activityDeclaration, atomicActivity);
	}

	private void addCorrectDependencies(
		IActivityGraphNodeDeclaration activityDeclaration,
		IActivityGraphRule[] emergencyRules)
		throws ModelCompilingException {

		if (activityDeclaration instanceof IActivityGraphComposedNodeDeclaration) {
			addCorrectDependencies(
				((IActivityGraphComposedNodeDeclaration)activityDeclaration)
					.getInnerActivityGraphDeclaration(),
				emergencyRules);

		} else {
			addActivityGraphEdgesAndEmergencyRules(
				activityDeclaration,
				getActivityGraphLeaf(activityDeclaration),
				emergencyRules);
		}
	}

	private void addCorrectDependencies(
		IActivityGraphDeclaration activityGraphDeclaration,
		IActivityGraphRule[] emergencyRules)
		throws ModelCompilingException {

		addCorrectDependenciesToActivityGraphDeclaration(
			activityGraphDeclaration,
			emergencyRules);
	}

	private void addCorrectDependenciesToActivityGraphDeclaration(
		IActivityGraphDeclaration activityGraphDeclaration,
		IActivityGraphRule[] oldEmergencyRules)
		throws ModelCompilingException {

		if (!_activityMap.containsKey(activityGraphDeclaration)) {

			// Only old emergency rules should be in exit activities.
			// Otherwise endless loop will occur!! 
			if (activityGraphDeclaration.getParentActivityGraphComposedNodeDeclaration() != null)
				addActivityGraphEdgesAndEmergencyRules(
					activityGraphDeclaration.getParentActivityGraphComposedNodeDeclaration(),
					getActivityGraphLeaf(activityGraphDeclaration.getExitActivityGraphLeafDeclaration()),
					oldEmergencyRules);

			IActivityGraphRule[] newEmergencyRules =
				createEmergencyRules(activityGraphDeclaration, oldEmergencyRules);

			addActivityGraphEdgesAndEmergencyRules(
				activityGraphDeclaration.getEntryActivityGraphLeafDeclaration(),
				getActivityGraphLeaf(activityGraphDeclaration.getEntryActivityGraphLeafDeclaration()),
				newEmergencyRules);

			Iterator it = activityGraphDeclaration.getActivityGraphNodeDeclarationsWithoutEmergencyAtAll();
			while (it.hasNext()) {
				addCorrectDependencies((IActivityGraphNodeDeclaration)it.next(), newEmergencyRules);
			}
		}
	}

	private void allocateActivityGraphNode(IActivityGraphNodeDeclaration activityDeclaration)
		throws ModelCompilingException {
		if (!_activityMap.containsKey(activityDeclaration)) {
			if (activityDeclaration instanceof IActivityGraphLeafDeclaration) {
				allocateActivityGraphLeaf((IActivityGraphLeafDeclaration)activityDeclaration);

			} else {
				allocateActivityGraph(
					((IActivityGraphComposedNodeDeclaration)activityDeclaration)
						.getInnerActivityGraphDeclaration());
			}
		}
	}

	private void allocateActivityGraphLeaf(IActivityGraphLeafDeclaration activityDeclaration)
		throws ModelCompilingException {

		AbstractActivityGraphLeaf atomicActivity =
			(AbstractActivityGraphLeaf)activityDeclaration.createActivityGraphLeaf(_objectCreateFunctionArgs);

		if (atomicActivity != null) {
			if (Assert.ASSERTION_ON)
				Assert.assertEquals(activityDeclaration.isInstantly(), atomicActivity.isInstantly());

			putActivityGraphNode(activityDeclaration, atomicActivity);
		}
	}

	private void putActivityGraphNode(
		IActivityGraphNodeDeclaration activityDeclaration,
		AbstractActivityGraphLeaf atomicActivity) {
		_activityMap.put(activityDeclaration, atomicActivity);
	}

	private void allocateActivityGraph(IActivityGraphDeclaration activityGraphDeclaration)
		throws ModelCompilingException {

		allocateActivityGraphNode(activityGraphDeclaration.getEntryActivityGraphLeafDeclaration());
		allocateActivityGraphNode(activityGraphDeclaration.getExitActivityGraphLeafDeclaration());

		if (activityGraphDeclaration.getParentActivityGraphComposedNodeDeclaration() != null)
			putActivityGraphNode(
				activityGraphDeclaration.getParentActivityGraphComposedNodeDeclaration(),
				getActivityGraphLeaf(activityGraphDeclaration.getEntryActivityGraphLeafDeclaration()));

		Iterator it = activityGraphDeclaration.getActivityGraphNodeDeclarationsWithoutEmergencyAtAll();
		while (it.hasNext()) {
			allocateActivityGraphNode((IActivityGraphNodeDeclaration)it.next());
		}
	}

	private IActivityGraphRule[] createEmergencyRules(
		IActivityGraphDeclaration activityGraphDeclaration,
		IActivityGraphRule[] oldEmergencyRules)
		throws ModelCompilingException {

		ArrayList emergencyRules = new ArrayList(Arrays.asList(oldEmergencyRules));
		Iterator it = activityGraphDeclaration.getEmergencyRuleDeclarations();

		while (it.hasNext()) {
			IActivityGraphRuleDeclaration activityRuleDeclaration = (IActivityGraphRuleDeclaration)it.next();
			IRule currRule = activityRuleDeclaration.createRule(_objectCreateFunctionArgs);

			if (currRule != null) {
				if (!emergencyRules.contains(currRule))
					emergencyRules.add(currRule);

			} else {
				if (Assert.ASSERTION_ON && _logger.isLoggable(Level.INFO))
					_logger.info("Removed emergencyRule " + activityRuleDeclaration.getName());
			}
		}

		return (IActivityGraphRule[])emergencyRules.toArray(new IActivityGraphRule[emergencyRules.size()]);
	}

	private AbstractActivityGraphLeaf getActivityGraphLeaf(IActivityGraphNodeDeclaration activityDeclaration) {
		return (AbstractActivityGraphLeaf)_activityMap.get(activityDeclaration);
	}

	public void initActivityGraph() throws ModelCompilingException {

		IActivityGraphDeclaration rootActivityGraphDeclaration =
			_rootActivityGraph.getActivityGraphDeclaration();

		// create for all declarations runtime objects
		allocateActivityGraph(rootActivityGraphDeclaration);

		_rootActivityGraph.setEntryActivityGraphLeaf(
			getActivityGraphLeaf(rootActivityGraphDeclaration.getEntryActivityGraphLeafDeclaration()));
		_rootActivityGraph.setActivityGraphLeafDeclarationToActivityGraphLeafMap(_activityMap);

		addCorrectDependencies(rootActivityGraphDeclaration, new IActivityGraphRule[0]);

		boolean noopGraph =
			ActivityGraphNoopHelper.isNoop(
				_objectCreateFunctionArgs,
				_rootActivityGraph,
				rootActivityGraphDeclaration);

		if (noopGraph) {
			if (Assert.ASSERTION_ON && _logger.isLoggable(Level.INFO)) {
				_logger.info(
					"Activity Graph '"
						+ rootActivityGraphDeclaration.getName()
						+ "' is noop activity graph!");
			}

			_rootActivityGraph.setNoopActivityGraph(true);
		}
	}
}