package de.uniwue.ki.mas.behaviour.activityGraphDeclaration;
import java.util.*;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;
import mygale.mcge.net.*;
import de.uniwue.ki.mas.behaviour.activityGraph.ActivityFactory;
import de.uniwue.ki.mas.behaviour.activityGraph.IActivityGraph;
import de.uniwue.ki.mas.behaviour.rule.*;
import de.uniwue.ki.mas.entity.object.IObjectCreateFunctionArgs;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.AbstractDocumentableDeclaration;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.functions.IPredicateFunction;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (08.11.2000 19:46:47)
 * 
 * @author Christoph Oechslein
 */
final class ActivityGraphDeclaration
    extends AbstractDocumentableDeclaration
    implements IActivityGraphDeclaration {

    private final ActivityGraphNodeDeclarationList _activityDeclarationList =
        new ActivityGraphNodeDeclarationList();
    private final ActivityGraphEdgeDeclarationList _activityEdgeDeclarationList =
        new ActivityGraphEdgeDeclarationList();

    private final transient ChangeListener _changeListener = new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
            fireStateChanged();
        }
    };

    private final IEmergencyActivityGraphLeafDeclaration _emergencyActivityDeclaration;
    private final IEntryActivityGraphLeafDeclaration _entryAtomicActivityDeclaration;
    private final IExitActivityGraphLeafDeclaration _exitAtomicActivityDeclaration;

    private transient NetGraphModelChangedListener _netGraphModelChangedListener = null;

    private final IActivityGraphComposedNodeDeclaration _parentActivityGraphComposedNodeDeclaration;

    public ActivityGraphDeclaration() {
        this("Root Composed Activity", null);
    }

    public ActivityGraphDeclaration(IActivityGraphComposedNodeDeclaration parentActivityGraphComposedNodeDeclaration) {
        this("Untitled Composed Activity", parentActivityGraphComposedNodeDeclaration);
    }

    public ActivityGraphDeclaration(
        String name,
        IActivityGraphComposedNodeDeclaration parentActivityGraphComposedNodeDeclaration) {
        super(name);

        _parentActivityGraphComposedNodeDeclaration = parentActivityGraphComposedNodeDeclaration;

        getActivityGraphNodeDeclarationList().setComposition(false);
        getActivityGraphEdgeDeclarationList().setComposition(false);

        _emergencyActivityDeclaration = new EmergencyActivityGraphLeafDeclaration(this);
        _entryAtomicActivityDeclaration = new EntryActivityGraphLeafDeclaration(this);
        _exitAtomicActivityDeclaration = new ExitActivityGraphLeafDeclaration(this);

        initialize();
    }

    public void addActivityGraphEdgeDeclaration(IDeclarationNetEdge activityEdgeDeclaration) {
        if (Assert.ASSERTION_ON) {
            Assert.assertTrue(
                "Must contain " + activityEdgeDeclaration.getFromDeclarationNetNode().getName(),
                getActivityGraphNodeDeclarationList().contains(
                    activityEdgeDeclaration.getFromDeclarationNetNode()));
            Assert.assertTrue(
                "Must contain " + activityEdgeDeclaration.getToDeclarationNetNode().getName(),
                getActivityGraphNodeDeclarationList().contains(
                    activityEdgeDeclaration.getToDeclarationNetNode()));

            Assert.assertNotSame(
                activityEdgeDeclaration.getToDeclarationNetNode(),
                activityEdgeDeclaration.getFromDeclarationNetNode());
        }

        addActivityGraphEdgeDeclarationWithoutCheck(activityEdgeDeclaration);
    }

    public IActivityGraphEdgeDeclaration addActivityGraphEdgeDeclaration(
        IActivityGraphNodeDeclaration fromActivityDeclaration,
        IActivityGraphNodeDeclaration toActivityDeclaration) {

        IActivityGraphRuleDeclaration newActivityRuleDeclaration =
            createActivityGraphEdgeDeclaration(fromActivityDeclaration, toActivityDeclaration);

        addActivityGraphEdgeDeclaration(newActivityRuleDeclaration);
        return newActivityRuleDeclaration;

    }

    // nodes von nodes.
    protected void addActivityGraphEdgeDeclarationWithoutCheck(IDeclarationNetEdge activityEdgeDeclaration) {
        activityEdgeDeclaration.setBeginNode(activityEdgeDeclaration.getFromDeclarationNetNode());
        activityEdgeDeclaration.setEndNode(activityEdgeDeclaration.getToDeclarationNetNode());
        
        getActivityGraphEdgeDeclarationList().add(activityEdgeDeclaration);
        fireAddedEdge(activityEdgeDeclaration);
    }

    public void addActivityGraphNodeDeclaration(IActivityGraphNodeDeclaration newActivityDeclaration) {
        newActivityDeclaration.setActivityGraphDeclaration(this);

        newActivityDeclaration.addChangeListener(_changeListener);

        getActivityGraphNodeDeclarationList().add(newActivityDeclaration);

        fireAddedNode(newActivityDeclaration);
    }

    public void addEdge(INetEdge edge) {
        addActivityGraphEdgeDeclaration((IActivityGraphEdgeDeclaration) edge);
    }

    public void addEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration activityEdgeDeclaration) {
        Assert.assertSame(
            getEmergencyActivityGraphLeafDeclaration(),
            activityEdgeDeclaration.getFromActivityGraphNodeDeclaration());

        addActivityGraphEdgeDeclaration(activityEdgeDeclaration);
    }

    public IActivityGraphEdgeDeclaration addEmergencyRuleDeclaration(IActivityGraphNodeDeclaration toActivityDeclaration) {
        return addActivityGraphEdgeDeclaration(
            getEmergencyActivityGraphLeafDeclaration(),
            toActivityDeclaration);
    }

    public void addNetGraphModelChangedListener(NetGraphModelChangedListener newListener) {
        _netGraphModelChangedListener =
            NetGraphModelChangedEventMulticaster.add(_netGraphModelChangedListener, newListener);
    }

    public void addNode(INetNode node) {
        addActivityGraphNodeDeclaration((IActivityGraphNodeDeclaration) node);
    }

    public void copyInto(Object inToObject, Map objectMapping) {
        super.copyInto(inToObject, objectMapping);

        if (objectMapping == null) {
            objectMapping = new HashMap();
        }

        // exit, entry, emergency, statelike, activities

        IActivityGraphDeclaration dstActivityGraphDeclaration = (IActivityGraphDeclaration) inToObject;

        dstActivityGraphDeclaration.removeAllDeclarationNetNodes(true);

        getEntryActivityGraphLeafDeclaration().copyInto(
            dstActivityGraphDeclaration.getEntryActivityGraphLeafDeclaration(),
            objectMapping);
        getExitActivityGraphLeafDeclaration().copyInto(
            dstActivityGraphDeclaration.getExitActivityGraphLeafDeclaration(),
            objectMapping);
        getEmergencyActivityGraphLeafDeclaration().copyInto(
            dstActivityGraphDeclaration.getEmergencyActivityGraphLeafDeclaration(),
            objectMapping);

        copyIntoActivityGraphNodeDeclarations(dstActivityGraphDeclaration, objectMapping);
    }

    protected void copyIntoActivityGraphEdgeDeclarations(
        IActivityGraphDeclaration inComposedActivityDeclaration,
        Map activityDeclarationToCopiedActivityDeclarationMap) {

        Iterator it = getNodes();
        while (it.hasNext()) {
            Iterator rules = ((IActivityGraphNodeDeclaration) it.next()).getActivityGraphEdgeDeclarations();
            while (rules.hasNext()) {
                IActivityGraphEdgeDeclaration currActivityRuleDeclaration =
                    (IActivityGraphEdgeDeclaration) rules.next();
                IActivityGraphEdgeDeclaration copiedCurrActivityRuleDeclaration =
                    (IActivityGraphEdgeDeclaration) currActivityRuleDeclaration.copy(
                        activityDeclarationToCopiedActivityDeclarationMap);

                copiedCurrActivityRuleDeclaration.setFromActivityGraphNodeDeclaration(
                    (IActivityGraphNodeDeclaration) activityDeclarationToCopiedActivityDeclarationMap.get(
                        currActivityRuleDeclaration.getFromActivityGraphNodeDeclaration()));

                copiedCurrActivityRuleDeclaration.setToActivityGraphNodeDeclaration(
                    (IActivityGraphNodeDeclaration) activityDeclarationToCopiedActivityDeclarationMap.get(
                        currActivityRuleDeclaration.getToActivityGraphNodeDeclaration()));

                inComposedActivityDeclaration.addActivityGraphEdgeDeclaration(
                    copiedCurrActivityRuleDeclaration);

                Assert.assertTrue(
                    inComposedActivityDeclaration.getActivityGraphEdgeDeclarationList().contains(
                        copiedCurrActivityRuleDeclaration));
            }
        }

    }

    protected void copyIntoActivityGraphNodeDeclarations(
        IActivityGraphDeclaration inComposedActivityDeclaration,
        Map activityDeclarationToCopiedActivityDeclarationMap) {

        activityDeclarationToCopiedActivityDeclarationMap.put(
            getEntryActivityGraphLeafDeclaration(),
            inComposedActivityDeclaration.getEntryActivityGraphLeafDeclaration());
        activityDeclarationToCopiedActivityDeclarationMap.put(
            getExitActivityGraphLeafDeclaration(),
            inComposedActivityDeclaration.getExitActivityGraphLeafDeclaration());
        activityDeclarationToCopiedActivityDeclarationMap.put(
            getEmergencyActivityGraphLeafDeclaration(),
            inComposedActivityDeclaration.getEmergencyActivityGraphLeafDeclaration());

        Iterator it = getActivityGraphNodeDeclarationsWithoutEmergencyAtAll();
        while (it.hasNext()) {
            IActivityGraphNodeDeclaration currActivityDeclaration = (IActivityGraphNodeDeclaration) it.next();
            IActivityGraphNodeDeclaration copiedCurrActivityDeclaration =
                (IActivityGraphNodeDeclaration) currActivityDeclaration.copy(
                    activityDeclarationToCopiedActivityDeclarationMap);

            inComposedActivityDeclaration.addActivityGraphNodeDeclaration(copiedCurrActivityDeclaration);

            activityDeclarationToCopiedActivityDeclarationMap.put(
                currActivityDeclaration,
                copiedCurrActivityDeclaration);
        }

        copyIntoActivityGraphEdgeDeclarations(
            inComposedActivityDeclaration,
            activityDeclarationToCopiedActivityDeclarationMap);
    }

    public IActivityGraph createActivityGraph(IObjectCreateFunctionArgs objectCreateFunctionArgs)
        throws ModelCompilingException {
        if (objectCreateFunctionArgs.existsCompilationObject(this, false))
            return (IActivityGraph) objectCreateFunctionArgs.getCompilationObject(this, false);
        else {
            IActivityGraph newComposedActivity = ActivityFactory.createActivityGraph(this);
            objectCreateFunctionArgs.addCompilationObject(this, newComposedActivity, false);
            return newComposedActivity;
        }
    }

    protected IActivityGraphRuleDeclaration createActivityGraphEdgeDeclaration(
        IActivityGraphNodeDeclaration fromActivityDeclaration,
        IActivityGraphNodeDeclaration toActivityDeclaration) {

        return ActivityGraphRuleDeclarationFactory.createActivityGraphRuleDeclaration(
            fromActivityDeclaration,
            toActivityDeclaration);
    }

    protected ICopyable createInstance(Map objectMapping) {
        return ActivityGraphDeclarationFactory.createActivityGraphDeclaration();
    }

    protected void fireAddedEdge(INetEdge Edge) {
        if (_netGraphModelChangedListener != null)
            _netGraphModelChangedListener.addedEdge(new NetGraphModelChangedEvent(this, Edge));
        fireStateChanged();
    }

    protected void fireAddedNode(INetNode node) {
        if (_netGraphModelChangedListener != null)
            _netGraphModelChangedListener.addedNode(new NetGraphModelChangedEvent(this, node));
        fireStateChanged();
    }

    protected void fireRemovedEdge(INetEdge Edge) {
        if (_netGraphModelChangedListener != null)
            _netGraphModelChangedListener.removedEdge(new NetGraphModelChangedEvent(this, Edge));
        fireStateChanged();
    }

    protected void fireRemovedNode(INetNode node) {
        if (_netGraphModelChangedListener != null)
            _netGraphModelChangedListener.removedNode(new NetGraphModelChangedEvent(this, node));
        fireStateChanged();
    }

    protected void fireStateChanged() {
        super.fireStateChanged();
    }

    public final ActivityGraphEdgeDeclarationList getActivityGraphEdgeDeclarationList() {
        return _activityEdgeDeclarationList;
    }

    public final ActivityGraphNodeDeclarationList getActivityGraphNodeDeclarationList() {
        return _activityDeclarationList;
    }

    public Iterator getActivityGraphNodeDeclarationsWithoutEmergencyAtAll() {
        return CollectionsHelper.select(getNodes(), new IPredicateFunction() {
            public boolean isElementOk(Object element) {
                return (element != getExitActivityGraphLeafDeclaration())
                    && (element != getEntryActivityGraphLeafDeclaration())
                    && (element != getEmergencyActivityGraphLeafDeclaration());
            }
        });
    }

    protected String getDefaultName() {
        return "Composed Activity";
    }

    public Iterator getEdges() {
        return getActivityGraphEdgeDeclarationList().iterator();
    }

    public IEmergencyActivityGraphLeafDeclaration getEmergencyActivityGraphLeafDeclaration() {
        return _emergencyActivityDeclaration;
    }

    public Iterator getEmergencyRuleDeclarations() {
        return CollectionsHelper
            .select(
                getEmergencyActivityGraphLeafDeclaration().getActivityGraphEdgeDeclarations(),
                new IPredicateFunction() {
            public boolean isElementOk(Object obj) throws Exception {
                return obj instanceof IActivityGraphRuleDeclaration;
            }
        });
    }

    public IEntryActivityGraphLeafDeclaration getEntryActivityGraphLeafDeclaration() {
        return _entryAtomicActivityDeclaration;
    }

    public IExitActivityGraphLeafDeclaration getExitActivityGraphLeafDeclaration() {
        return _exitAtomicActivityDeclaration;
    }

    public Iterator getFromActivityGraphNodeDeclarations() {
        return CollectionsHelper.select(getNodes(), new IPredicateFunction() {
            public boolean isElementOk(Object element) {
                IActivityGraphNodeDeclaration fromActivityDeclaration =
                    (IActivityGraphNodeDeclaration) element;
                return fromActivityDeclaration != getExitActivityGraphLeafDeclaration();
            }
        });
    }

    public Iterator getNodes() {
        return getActivityGraphNodeDeclarationList().iterator();
    }

    public IActivityGraphComposedNodeDeclaration getParentActivityGraphComposedNodeDeclaration() {
        return _parentActivityGraphComposedNodeDeclaration;
    }

    public Iterator getToActivityGraphNodeDeclarations() {
        return CollectionsHelper.select(getNodes(), new IPredicateFunction() {
            public boolean isElementOk(Object element) {
                IActivityGraphNodeDeclaration toActivityDeclaration = (IActivityGraphNodeDeclaration) element;
                return (toActivityDeclaration != getEntryActivityGraphLeafDeclaration())
                    && (toActivityDeclaration != getEmergencyActivityGraphLeafDeclaration());
            }
        });
    }

    private void initialize() {
        addActivityGraphNodeDeclaration(_emergencyActivityDeclaration);
        addActivityGraphNodeDeclaration(_entryAtomicActivityDeclaration);
        addActivityGraphNodeDeclaration(_exitAtomicActivityDeclaration);
    }

    public boolean isInstantly() {
        Iterator allActivities = getNodes();

        while (allActivities.hasNext()) {
            IActivityGraphNodeDeclaration currIActivityDeclaration =
                (IActivityGraphNodeDeclaration) allActivities.next();

            if (currIActivityDeclaration != getEmergencyActivityGraphLeafDeclaration()) {
                if (!currIActivityDeclaration.isInstantly()) {
                    return false;
                }
            }
        }

        return true;
    }

    public boolean isomorphicEquals(IIsomorphicHashable anObject) {
        if (this == anObject)
            return true;

        if (!super.isomorphicEquals(anObject)) {
            return false;
        }

        if (!(anObject instanceof IActivityGraphDeclaration)) {
            return false;
        }

        IActivityGraphDeclaration composedActivityDeclaration = (IActivityGraphDeclaration) anObject;

        return getActivityGraphNodeDeclarationList().isomorphicEquals(
            composedActivityDeclaration.getActivityGraphNodeDeclarationList())
            && getActivityGraphEdgeDeclarationList().isomorphicEquals(
                composedActivityDeclaration.getActivityGraphEdgeDeclarationList());
    }

    public int isomorphicHashCode() {
        return 5 * super.isomorphicHashCode()
            + 3 * getActivityGraphNodeDeclarationList().isomorphicHashCode()
            + getActivityGraphEdgeDeclarationList().isomorphicHashCode();
    }

    public void remove() {

        removeAllDeclarationNetNodes(true);

        getExitActivityGraphLeafDeclaration().remove();
        getEntryActivityGraphLeafDeclaration().remove();
        getEmergencyActivityGraphLeafDeclaration().remove();

        super.remove();
    }

    public void removeDeclarationNetEdge(
            IDeclarationNetEdge activityEdgeDeclaration,
            boolean withCallingRemove) {

            Assert.assertTrue(CollectionsHelper.contains(getEdges(), activityEdgeDeclaration));

            activityEdgeDeclaration.removed();

            if (withCallingRemove)
                activityEdgeDeclaration.remove();

            getActivityGraphEdgeDeclarationList().remove(activityEdgeDeclaration);

            fireRemovedEdge(activityEdgeDeclaration);

            Assert.assertTrue(!CollectionsHelper.contains(getEdges(), activityEdgeDeclaration));
        }

    public void removeDeclarationNetNode(
        IDeclarationNetNode activityDeclaration,
        boolean withCallingRemove) {

        Assert.assertTrue(getActivityGraphNodeDeclarationList().contains(activityDeclaration));

        removeAllEdges(
            CollectionsHelper.asLinkedList(activityDeclaration.getNodeEdges()).iterator(),
            withCallingRemove);

        activityDeclaration.removeChangeListener(_changeListener);

        if (withCallingRemove)
            activityDeclaration.remove();

        getActivityGraphNodeDeclarationList().remove(activityDeclaration);

        Assert.assertTrue(!getActivityGraphNodeDeclarationList().contains(activityDeclaration));

        fireRemovedNode(activityDeclaration);
    }

    public void removeAllDeclarationNetEdges(boolean withCallingRemove) {
        Iterator it = getNodes();

        while (it.hasNext()) {
            removeAllDeclarationNetEdges(
                (IActivityGraphNodeDeclaration) it.next(),
                withCallingRemove);
        }
    }

    public void removeAllDeclarationNetEdges(
        IDeclarationNetNode activityDeclaration,
        boolean withCallingRemove) {
        removeAllEdges(
            CollectionsHelper.asLinkedList(activityDeclaration.getIncomingNodeEdges()).iterator(),
            withCallingRemove);
    }

    public void removeAllDeclarationNetNodes(boolean withCallingRemove) {
        Iterator it =
            CollectionsHelper
                .asLinkedList(getActivityGraphNodeDeclarationsWithoutEmergencyAtAll())
                .iterator();

        while (it.hasNext()) {
            IDeclarationNetNode currActivityDeclaration = (IDeclarationNetNode) it.next();
            removeDeclarationNetNode(currActivityDeclaration, withCallingRemove);
        }
    }

    private void removeAllEdges(Iterator rules, boolean withCallingRemove) {
        while (rules.hasNext()) {
            IDeclarationNetEdge activityEdgeDeclaration =
                (IDeclarationNetEdge) rules.next();
            removeDeclarationNetEdge(activityEdgeDeclaration, withCallingRemove);
        }
    }

    public void removeEdge(INetEdge edge) {
        removeDeclarationNetEdge(((IActivityGraphEdgeDeclaration) edge), true);
    }

    public final void removeEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration activityEdgeDeclaration) {
        removeDeclarationNetEdge(activityEdgeDeclaration, true);
    }

    public final void removeEmergencyRuleDeclaration(
        IActivityGraphEdgeDeclaration activityEdgeDeclaration,
        boolean withCallingRemove) {

        removeDeclarationNetEdge(activityEdgeDeclaration, withCallingRemove);
    }

    public void removeNetGraphModelChangedListener(NetGraphModelChangedListener newListener) {
        _netGraphModelChangedListener =
            NetGraphModelChangedEventMulticaster.remove(_netGraphModelChangedListener, newListener);
    }

    public void removeNode(INetNode node) {
        removeDeclarationNetNode(((IActivityGraphNodeDeclaration) node), true);
    }

    public void removeEdgesForNode(INetNode node) {
        Assert.fail("Should not get called in this implementation of the INetGraphModel-Interface...");
    }

    public boolean contains(INetNode obj) {
        // --steffen
        Assert.fail("should not be called");
        return false;
    }
}