package de.uniwue.ki.mas.behaviour.activityGraphDeclaration;
import java.util.Iterator;

import mygale.mcge.net.INetGraphModel;
import de.uniwue.ki.mas.behaviour.activityGraph.IActivityGraph;
import de.uniwue.ki.mas.behaviour.rule.ActivityGraphEdgeDeclarationList;
import de.uniwue.ki.mas.behaviour.rule.IActivityGraphEdgeDeclaration;
import de.uniwue.ki.mas.entity.object.IObjectCreateFunctionArgs;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.util.declaration.IDocumentableDeclaration;

/**
 * This class encodes a graph of {@link IActivityGraphNodeDeclaration IActivityDeclarations},
 * with follwing special activities (like a finite state machine):
 * <ul>
 *   <li> start and end activity,
 *   <li> emergency rule activitas (see {@link IActivityGraphEdgeDeclaration}
 *        and {@link IEmergencyAtomicActivity}).
 * </ul>
 * Creation date: (08.11.2000 19:37:05)
 *
 * @author Christoph Oechslein
 */
public interface IModifiableNetGraphModel extends INetGraphModel, IDocumentableDeclaration {

	IActivityGraphComposedNodeDeclaration getParentActivityGraphComposedNodeDeclaration();

	/**
	 * Adds a new node (activity) to this graph.
	 *
	 * @see #removeActivityGraphNodeDeclaration(IActivityGraphNodeDeclaration)
	 */
	void addActivityGraphNodeDeclaration(IActivityGraphNodeDeclaration newActivityDeclaration);

	/**
	 * Adds an directed edge rule from {@link IActivityGraphEdgeDeclaration#getFromActivityDeclaration()
	 * activityEdgeDeclaration.getFromActivityDeclaration()}
	 * to {@link IActivityGraphEdgeDeclaration#getToActivityDeclaration()
	 * activityEdgeDeclaration.getToActivityDeclaration()}.
	 * Both activities must exists in this graph (i.e. must be added by add(IActivityGraphNodeDeclaration)).
	 *
	 * @see #addActivityGraphEdgeDeclaration(IActivityGraphNodeDeclaration)
	 * @see #removeActivityGraphEdgeDeclaration(IActivityGraphEdgeDeclaration)
	 */
	void addActivityGraphEdgeDeclaration(IActivityGraphEdgeDeclaration activityEdgeDeclaration);

	/**
	 * Creates and adds a new directed edge rule to the graph.
	 *
	 * @see #addActivityGraphEdgeDeclaration(IActivityGraphEdgeDeclaration)
	 * @see #removeActivityGraphEdgeDeclaration(IActivityGraphEdgeDeclaration)
	 */
	IActivityGraphEdgeDeclaration addActivityGraphEdgeDeclaration(
		IActivityGraphNodeDeclaration fromActivityDeclaration,
		IActivityGraphNodeDeclaration toActivityDeclaration);

	/**
	 * Adds a new emergency rule declaration to the graph.
	 *
	 * @see #addEmergencyRuleDeclaration(IActivityGraphNodeDeclaration)
	 * @see #removeEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration)
	 */
	void addEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration newActivityEdgeDeclaration);

	/**
	 * Creates and adds a new emergency rule declaration to the graph.
	 *
	 * @see #addEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration)
	 * @see #removeEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration)
	 */
	IActivityGraphEdgeDeclaration addEmergencyRuleDeclaration(IActivityGraphNodeDeclaration toActivity);

	/**
	 * Compiles a runtime <code>IComposedActivity</code> object in the given
	 * compilation environment.
	 */

	IActivityGraph createActivityGraph(IObjectCreateFunctionArgs objectCreateFunctionArgs)
		throws ModelCompilingException;

	ActivityGraphNodeDeclarationList getActivityGraphNodeDeclarationList();

	/**
	 * Gets all {@link IActivityGraphNodeDeclaration IActivityDeclarations} in this graph
	 * excluding entry, exit and emergency activity.
	 *
	 * @return Iterator with {@link IActivityGraphNodeDeclaration IActivityDeclarations}
	 * @see #getAllActivityDeclarations()
	 * @see #getToActivityDeclarations()
	 */
	Iterator getActivityGraphNodeDeclarationsWithoutEmergencyAtAll();

	ActivityGraphEdgeDeclarationList getActivityGraphEdgeDeclarationList();

	/**
	 * Gets the emergency activity which knows all emergency rule declarations.
	 */
	IEmergencyActivityGraphLeafDeclaration getEmergencyActivityGraphLeafDeclaration();

	/**
	 * Returns the emergency activities.
	 *
	 * @return Iterator with {@link IEmergencyActivityGraphLeafDeclaration emergency
	 * activity declarations}
	 */

	Iterator getEmergencyRuleDeclarations();

	/**
	 * Returns entry activity declaration.
	 */
	IEntryActivityGraphLeafDeclaration getEntryActivityGraphLeafDeclaration();

	/**
	 * Returns exit activity declaration.
	 */
	IExitActivityGraphLeafDeclaration getExitActivityGraphLeafDeclaration();

	/**
	 * Gets all activityDeclarations in this graph including entry and emergency activity and excluding exit activity.
	 * Use this for the to part in edges
	 *
	 * @return Iterator with {@link IActivityGraphNodeDeclaration IActivityDeclarations}
	 * @see #getActivityDeclarations()
	 */
	Iterator getFromActivityGraphNodeDeclarations();

	/**
	 * Gets all activityDeclarations in this graph including exit activity and excluding entry and emergency activity.
	 * Use this for the to part in chainrules
	 *
	 * @return Iterator with {@link IActivityGraphNodeDeclaration IActivityDeclarations}
	 * @see #getActivityDeclarations()
	 */
	Iterator getToActivityGraphNodeDeclarations();

	/**
	 * Removes activity from Graph must be formerly add by <code>add</code> and
	 * must not be the start, end or the emergency activity.
	 * @see #addActivityGraphNodeDeclaration(IActivityGraphNodeDeclaration)
	 */
	void removeDeclarationNetNode(
		IDeclarationNetNode newActivityDeclaration,
		boolean withCallingRemove);

	/**
	  * Remove all activities besides entry, exit and emergency
	  */
	void removeAllDeclarationNetNodes(boolean withCallingRemove);

	void removeAllDeclarationNetEdges(
		IDeclarationNetNode activityDeclaration,
		boolean withCallingRemove);

	void removeAllActivityGraphEdgeDeclarations(boolean withCallingRemove);

	/**
	 * Remove the edge from the graph.
	 *
	 * @see #addActivityGraphEdgeDeclaration(IActivityGraphEdgeDeclaration)
	 * @see #addActivityGraphEdgeDeclaration(IActivityGraphNodeDeclaration)
	 */
	void removeActivityGraphEdgeDeclaration(
		IDeclarationNetEdge activityEdgeDeclaration,
		boolean withCallingRemove);

	/**
	 * Removes an emergency rule declaration from graph.
	 *
	 * @see #removeEmergencyRuleDeclaration(IActivityGraphEdgeDeclaration)
	 * @see #addEmergencyRuleDeclaration(IActivityGraphNodeDeclaration)
	 */
	void removeEmergencyRuleDeclaration(
		IActivityGraphEdgeDeclaration newActivityEdgeDeclaration,
		boolean withCallingRemove);
}