package de.uniwue.ki.mas.entity.body;

import java.util.Iterator;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.ICreateAndExecuteFunctionArgs;
import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.type.*;
import de.uniwue.ki.mas.entity.object.IObject;
import de.uniwue.ki.mas.entity.object.IObjectExecuteFunctionArgs;
import de.uniwue.ki.mas.entity.variable.*;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.name.AbstractNamed;

/**
 * Creation date: (23.08.2001 10:22:42)
 * @author Christoph Oechslein
 */

final class Body extends AbstractNamed implements IBody {
	private boolean[] _initializedVariables;

	private final int[] _dynamicVariablesIndexes;
	private final int[] _dynamicDoubleVariablesIndexes;
	private final int[] _dynamicIntVariablesIndexes;

	private IBodyVariable[] _variables;
	private IObject _object;
	private IBodyInstanceDeclaration _bodyInstanceDeclaration;
	private int runtimeNameBodyVariableIndex = -1; // no specified runtime name

	public Body(
		IBodyInstanceDeclaration bodyInstanceDeclaration,
		IObject object,
		IBodyVariable[] variables) {
		super();

		_object = object;
		_bodyInstanceDeclaration = bodyInstanceDeclaration;

		setVariables(variables);

		_dynamicVariablesIndexes = createDynamicIndexArray(variables, null, false);
		_dynamicDoubleVariablesIndexes = createDynamicIndexArray(variables, DoubleType.getInstance(), true);
		_dynamicIntVariablesIndexes = createDynamicIndexArray(variables, IntegerType.getInstance(), true);

		_initializedVariables = new boolean[getBodyVariablesLength()];
		
		//search the index of the name
		for (int i = 0;i<getBodyVariablesLength(); i++) {
		    if (getBodyVariableAt(i).isPartOfName()) runtimeNameBodyVariableIndex=i;
		}
	}
	
	public String getRuntimeName() {
	    if (runtimeNameBodyVariableIndex==-1) return null;
	    else return getBodyVariableAt(runtimeNameBodyVariableIndex).getName();
	}
	
	

	private final void uninitializeNonConstantVariables(ICreateFunctionArgs createFunctionArgs)
		throws ModelCompilingException {

		_initializedVariables = new boolean[getBodyVariablesLength()];

		// only reinit non constant default values, since the constant ones are inlined!

		BodyVariableInstanceDeclarationList bodyVariableInstanceDeclarationList =
			getBodyInstanceDeclaration().getBodyVariableInstanceDeclarationList();

		Iterator it = bodyVariableInstanceDeclarationList.iterator();

		for (int i = 0; i < _initializedVariables.length; i++) {
			IBodyVariableInstanceDeclaration bodyVariableInstanceDeclaration =
				(IBodyVariableInstanceDeclaration) it.next();

			if (!bodyVariableInstanceDeclaration.isConstant(createFunctionArgs))
				_initializedVariables[i] = false;
		}
	}

	private int[] createDynamicIndexArray(
		IBodyVariable[] variables,
		int index,
		int dynamicVariablesAmount,
		ISpecificType numberType,
		boolean onlyNumberType) {

		if (index >= variables.length) {
			if (dynamicVariablesAmount == 0)
				return null;
			else
				return new int[dynamicVariablesAmount];

		} else if (variables[index].isDynamicVariable()) {

			ISpecificType specificType =
				getBodyVariableInstanceDeclarationAt(index)
					.getBodyVariableClassDeclaration()
					.getVariableClassDeclaration()
					.getSpecificType();

			boolean hasNumberType;
			if (numberType == null) {
				hasNumberType =
					DoubleType.getInstance().equals(specificType)
						|| IntegerType.getInstance().equals(specificType);
			} else {
				hasNumberType = numberType.equals(specificType);
			}

			if ((onlyNumberType && hasNumberType) || (!onlyNumberType && !hasNumberType)) {
				int[] intArray =
					createDynamicIndexArray(
						variables,
						index + 1,
						dynamicVariablesAmount + 1,
						numberType,
						onlyNumberType);
				intArray[dynamicVariablesAmount] = index;
				return intArray;
			}
		}

		return createDynamicIndexArray(
			variables,
			index + 1,
			dynamicVariablesAmount,
			numberType,
			onlyNumberType);
	}

	private int[] createDynamicIndexArray(
		IBodyVariable[] variables,
		ISpecificType numberType,
		boolean onlyDoubleType) {

		return createDynamicIndexArray(variables, 0, 0, numberType, onlyDoubleType);
	}

	public final IBodyInstanceDeclaration getBodyInstanceDeclaration() {
		return _bodyInstanceDeclaration;
	}

	public final IBodyVariable getBodyVariableAt(int index) {
		return _variables[index];
	}

	private IBodyVariableInstanceDeclaration getBodyVariableInstanceDeclarationAt(int index) {
		return (IBodyVariableInstanceDeclaration) getBodyInstanceDeclaration()
			.getBodyVariableInstanceDeclarationList()
			.getElementAt(index);
	}

	public IBodyVariableInstanceDeclaration getBodyVariableInstanceDeclarationFor(IBodyVariable bodyVariable) {
		for (int i = 0; i < _variables.length; i++) {
			if (_variables[i] == bodyVariable) {
				return getBodyVariableInstanceDeclarationAt(i);
			}
		}

		return null;
	}

	public int getBodyVariablesLength() {
		return _variables.length;
	}

	public boolean isBodyVariableInitialized(int index) {
		return variableIsInitialized(index);
	}

	private final boolean variableIsInitialized(int index) {
		return (_initializedVariables == null) || _initializedVariables[index];
	}

	public final IObject getObject() {
		return _object;
	}

	private void initializeVariableFunctions(ICreateAndExecuteFunctionArgs createFunctionArgs)
		throws ModelCompilingException, SimulationRuntimeException {

		BodyVariableInstanceDeclarationList bodyVariableInstanceDeclarationList =
			getBodyInstanceDeclaration().getBodyVariableInstanceDeclarationList();

		Iterator it = bodyVariableInstanceDeclarationList.iterator();

		int index = 0;

		while (it.hasNext()) {
			if (!variableIsInitialized(index)) {
				IBodyVariableInstanceDeclaration bodyVariableInstanceDeclaration =
					(IBodyVariableInstanceDeclaration) it.next();

				_variables[index].initializeDefaultValue(bodyVariableInstanceDeclaration, createFunctionArgs);

				_initializedVariables[index] = true;
			}

			index++;
		}
	}

	private void initializeNextValueFunctions(ICreateFunctionArgs createFunctionArgs)
		throws ModelCompilingException {

		BodyVariableInstanceDeclarationList bodyVariableInstanceDeclarationList =
			getBodyInstanceDeclaration().getBodyVariableInstanceDeclarationList();

		Iterator it = bodyVariableInstanceDeclarationList.iterator();

		int index = 0;

		while (it.hasNext()) {
			IBodyVariableInstanceDeclaration bodyVariableInstanceDeclaration =
				(IBodyVariableInstanceDeclaration) it.next();

			_variables[index].initializeNextValueFunction(
				bodyVariableInstanceDeclaration,
				createFunctionArgs);

			index++;
		}
	}

	public void remove() {
		/*
		for (int i = 0; i < _variables.length; i++) {
		    _variables[i].remove();
		}
		*/

		// other methods produce errors, if e.g. body is in the gui
		// _variables = null;
	}

	public void reset(IObjectExecuteFunctionArgs executeFunctionArgs) throws SimulationRuntimeException {
		ICreateAndExecuteFunctionArgs createAndExecuteFunctionArgs =
			executeFunctionArgs.createCreateAndExecuteFunctionArgs();

		try {
			uninitializeNonConstantVariables(executeFunctionArgs.createCreateAndExecuteFunctionArgs());

			initializeVariableFunctions(createAndExecuteFunctionArgs);

			_initializedVariables = null;

			initializeNextValueFunctions(createAndExecuteFunctionArgs);

		} catch (ModelCompilingException exception) {
			throw new SimulationRuntimeException(exception);
		}
	}

	public void setObject(IObject newObject) {
		_object = newObject;
	}

	private void setVariables(IBodyVariable[] bodyVariables) {
		_variables = bodyVariables;
	}

	public void update(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		updateVariables(objectExecuteFunctionArgs);
		updateDoubleVariables(objectExecuteFunctionArgs);
		updateIntVariables(objectExecuteFunctionArgs);
	}

	private void updateDoubleVariables(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		if (_dynamicDoubleVariablesIndexes != null) {
			for (int i = 0; i < _dynamicDoubleVariablesIndexes.length; i++) {

				IBodyVariable currBodyVariable = getBodyVariableAt(_dynamicDoubleVariablesIndexes[i]);

				Assert.assertEquals(getObject(), objectExecuteFunctionArgs.getObject());

				if (objectExecuteFunctionArgs.hasObjectsVariableListener()) {
					double oldValue = currBodyVariable.updateDouble(objectExecuteFunctionArgs);

					double newValue = currBodyVariable.getDoubleValue();

					objectExecuteFunctionArgs.fireVariableValueChanged(
						getObject(),
						getObject(),
						currBodyVariable,
						new Double(oldValue),
						new Double(newValue));

				} else {
					currBodyVariable.updateDouble(objectExecuteFunctionArgs);
				}
			}
		}
	}

	private void updateIntVariables(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		if (_dynamicIntVariablesIndexes != null) {
			for (int i = 0; i < _dynamicIntVariablesIndexes.length; i++) {

				IBodyVariable currBodyVariable = getBodyVariableAt(_dynamicIntVariablesIndexes[i]);

				Assert.assertEquals(getObject(), objectExecuteFunctionArgs.getObject());

				if (objectExecuteFunctionArgs.hasObjectsVariableListener()) {
					int oldValue = currBodyVariable.updateInt(objectExecuteFunctionArgs);
					int newValue = currBodyVariable.getIntValue();

					objectExecuteFunctionArgs.fireVariableValueChanged(
						getObject(),
						getObject(),
						currBodyVariable,
						new Integer(oldValue),
						new Integer(newValue));

				} else {
					currBodyVariable.updateInt(objectExecuteFunctionArgs);
				}
			}
		}
	}

	private void updateVariables(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		if (_dynamicVariablesIndexes != null) {
			for (int i = 0; i < _dynamicVariablesIndexes.length; i++) {

				IBodyVariable currBodyVariable = getBodyVariableAt(_dynamicVariablesIndexes[i]);

				Assert.assertEquals(getObject(), objectExecuteFunctionArgs.getObject());

				if (objectExecuteFunctionArgs.hasObjectsVariableListener()) {
					Object oldValue = currBodyVariable.update(objectExecuteFunctionArgs);

					Object newValue = currBodyVariable.getValue();

					objectExecuteFunctionArgs.fireVariableValueChanged(
						getObject(),
						getObject(),
						currBodyVariable,
						oldValue,
						newValue);

				} else {
					currBodyVariable.update(objectExecuteFunctionArgs);
				}
			}
		}
	}

	public String getName() {
		return "Body";
		//		if (getObject() != null) {
		//			return getObject().getName();
		//
		//		} else {
		//			return "null";
		//		}
	}

	public String toString() {
		StringBuffer buff = new StringBuffer();
		buff.append(getClass().getName());
		buff.append('@');
		buff.append(Integer.toHexString(hashCode()));

		buff.append("[");
		//		buff.append(getName());
		//		//		buff.append("; ");
		//
		//		for (int i = 0; i < _variables.length; i++) {
		//			_variables[i].appendName(buff, 0, 10);
		//			if (i < _variables.length - 1)
		//				buff.append(", ");
		//		}
		//
		buff.append("]");

		return buff.toString();
	}
}