package de.uniwue.ki.mas.entity.object;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.feature.IObjectFeature;
import de.uniwue.ki.mas.basics.feature.IObjectFeatureInstanceDeclaration;
import de.uniwue.ki.mas.entity.body.IBody;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.name.AbstractNamed;

/**
 * Creation date: (13.09.2001 16:47:25)
 * @author Christoph Oechslein
 */

public abstract class AbstractObject extends AbstractNamed implements IObject {
	private final int _id;
	private boolean _dead = false;
	private IBody _body;
	private IObjectInstanceDeclaration _objectInstanceDeclaration;
	private IObjectFeature[] _features;

	public AbstractObject(
		IObjectInstanceDeclaration objectInstanceDeclaration,
		IObjectCreateFunctionArgs objectCreateFunctionArgs) {
		super();

		_objectInstanceDeclaration = objectInstanceDeclaration;
		_id = ObjectIdHelper.getNextId(objectInstanceDeclaration, objectCreateFunctionArgs);
	}

	public final IBody getBody() {
		return _body;
	}
	
	public IObjectFeature getObjectFeature(Class featureClass) {
		for (int i = 0; i < getObjectFeaturesCount(); i++) {
			IObjectFeature currIObjectFeature = getObjectFeatureAt(i);
			
					if (featureClass.isInstance(currIObjectFeature)) {
						return currIObjectFeature;
					}
				};
		return null;
	}

	public final String getName() {
	    //user specified name
	    if (getBody()!=null) {
	        String name = getBody().getRuntimeName();
	        if (name!=null) return name;
	    }
	    //generic name
		StringBuffer buff = new StringBuffer();
		String idStr = Integer.toHexString(getId());

		if (_objectInstanceDeclaration != null)
			buff.append(_objectInstanceDeclaration.getName());
		else
			buff.append("obj removed");
		buff.append("@");

		for (int i = 0; i < 3 - idStr.length(); i++) {
			buff.append('0');
		}

		buff.append(idStr);

		return buff.toString();
	}

	public final IObjectFeature getObjectFeatureAt(int index) {
		return _features[index];
	}

	public final int getObjectFeaturesCount() {
		if (_features == null)
			return 0;
		else
			return _features.length;
	}
	
	public Iterator getObjectFeatures() {
		if (_features == null) {
			return CollectionsHelper.EMPTY_ITERATOR;
		} else {
			return CollectionsHelper.arrayToIterator(_features);
		}
	}


	public final IObjectInstanceDeclaration getObjectInstanceDeclaration() {
		return _objectInstanceDeclaration;
	}

	public boolean isDead() {
		return _dead;
	}

	public void remove() {
		_body.remove();

		for (int i = 0; i < getObjectFeaturesCount(); i++) {
			getObjectFeatureAt(i).remove();
		}

		_body = null;
		_dead = true;
		_features = null;
		_objectInstanceDeclaration = null;
	}

	public void reset(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		setDead(false);

		if (Assert.ASSERTION_ON)
			Assert.assertNotNull("body is null", getBody());

		getBody().reset(objectExecuteFunctionArgs);

		for (int i = 0; i < getObjectFeaturesCount(); i++) {
			getObjectFeatureAt(i).reset(objectExecuteFunctionArgs);
		}
	}

	public final void setBody(IBody newBody) {
		Assert.assertNull(getBody());
		_body = newBody;
		
	}

	public void setDead(boolean newDead) {
		_dead = newDead;
	}

	public void setObjectFeatures(IObjectFeature[] objectFeatures) {
		if (Assert.ASSERTION_ON)
			Assert.assertNull("tries to set object feature a second time", _features);

		_features = objectFeatures;
	}

	public void update(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		updateObjectFeatures(objectExecuteFunctionArgs);
		updateBody(objectExecuteFunctionArgs);
	}

	private void updateBody(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		getBody().update(objectExecuteFunctionArgs);
	}

	private void updateObjectFeatures(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		for (int i = 0; i < getObjectFeaturesCount(); i++) {
			getObjectFeatureAt(i).update(objectExecuteFunctionArgs);
		}
	}

	public final int getId() {
		return _id;
	}

	public final void initialize(IObjectCreateFunctionArgs objectCreateFunctionArgs)
		throws ModelCompilingException {

		Assert.assertNull(getBody());

		IObject oldActiveObject = objectCreateFunctionArgs.getObject();

		objectCreateFunctionArgs.setObject(this);

		initializeIntern(objectCreateFunctionArgs);

		if (Assert.ASSERTION_ON)
			Assert.assertSame(this, objectCreateFunctionArgs.getObject());

		objectCreateFunctionArgs.setObject(oldActiveObject);
	}

	protected void initializeIntern(IObjectCreateFunctionArgs objectCreateFunctionArgs)
		throws ModelCompilingException {

		setBody(createBody(objectCreateFunctionArgs));
		setObjectFeatures(createObjectFeatureList(objectCreateFunctionArgs));

		try {
			reset((IObjectExecuteFunctionArgs) objectCreateFunctionArgs.createCreateAndExecuteFunctionArgs());

		} catch (SimulationRuntimeException exception) {
			throw new ModelCompilingException(exception);
		}
	}

	protected IBody createBody(IObjectCreateFunctionArgs objectCreateFunctionArgs)
		throws ModelCompilingException {
		return getObjectInstanceDeclaration().getBodyInstanceDeclaration().createBody(
			objectCreateFunctionArgs);
	}

	protected IObjectFeature[] createObjectFeatureList(IObjectCreateFunctionArgs objectCreateFunctionArgs) {
		List featureList = null;

		Iterator it = getObjectInstanceDeclaration().getObjectFeatureInstanceDeclarationList().iterator();

		while (it.hasNext()) {
			Object obj = it.next();
			if (obj != null) {

				IObjectFeatureInstanceDeclaration featureInstanceDeclaration =
					(IObjectFeatureInstanceDeclaration) obj;
				IObjectFeature feature =
					featureInstanceDeclaration.createObjectFeature(objectCreateFunctionArgs);

				if (feature != null) {
					if (featureList == null) {
						featureList = new LinkedList();
					}
					featureList.add(feature);
				}
			}
		}

		if (featureList == null)
			return null;
		else
			return (IObjectFeature[]) featureList.toArray(new IObjectFeature[featureList.size()]);
	}
}