package de.uniwue.ki.mas.entity.object;
import java.awt.event.ActionEvent;
import java.io.Serializable;
import java.util.Iterator;

import de.uniwue.ki.mas.basics.feature.IObjectFeature;
import de.uniwue.ki.mas.entity.body.IBody;
import de.uniwue.ki.mas.entity.object.agent.IAgent;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.IRemoveable;
import de.uniwue.ki.util.name.INamed;

/**
  * A runtime simulation object which is created by an {@link IObjectInstanceDeclaration}
  * and which can act (not reason and sense, see {@link IAgent}),
  * it also generates {@link ActionEvent action events}.
  * Creation date: (26.10.2000 11:16:38)
  *
  * @author Christoph Oechslein
  */

public interface IObject extends INamed, IRemoveable, Serializable {
	/**
	 * Returns the body of the agent, the body is the visible portion of the object.
	 *
	 * @see #setBody(IBody)
	 * @see #update(IObjectExecuteFunctionArgs)
	 */

	IBody getBody();
	
	IObjectFeature getObjectFeature(Class featureClass);

	/**
	 * Returns i-th feature runtime object.
	 *
	 * @see #setObjectFeatures(IObjectFeature[])
	 * @see #getObjectFeaturesCount()
	 * @see #update(IObjectExecuteFunctionArgs)
	 */
	IObjectFeature getObjectFeatureAt(int index);
	
	Iterator getObjectFeatures();

	/**
	 * Returns count of feature runtime objects.
	 *
	 * @see #setObjectFeatures(IObjectFeature[])
	 * @see #getObjectFeatureAt(int)
	 * @see #update(IObjectExecuteFunctionArgs)
	 */
	int getObjectFeaturesCount();

	/**
	 * Returns declaration of this runtime simulation object.
	 */
	IObjectInstanceDeclaration getObjectInstanceDeclaration();

	boolean isDead();

	/**
	 * Resets the object to default values.
	 */
	void reset(IObjectExecuteFunctionArgs objectExecuteFunctionArgs) throws SimulationRuntimeException;

	/**
	 * Returns the body of the agent, the body is the visible portion of the object.
	 *
	 * @see #update(IObjectExecuteFunctionArgs)
	 * @see #getBody()
	 */
	void setBody(IBody body);

	void setDead(boolean newDead);

	/**
	 * Sets the feature runtime objects for this agent, could be <code>null</code>.
	 *
	 * @see #update(IObjectExecuteFunctionArgs)
	 * @see #getObjectFeatures()
	 */
	void setObjectFeatures(IObjectFeature[] newFeatures);

	/**
	 * Update this agent under the given execution bindings (update features and the body).
	 */
	void update(IObjectExecuteFunctionArgs objectExecuteFunctionArgs) throws SimulationRuntimeException;

	int getId();

	/**
	 * Creates the Body and the features under the given compilation bindings.
	 * Body Variables are added and initialized with default values. 
	 * Agent Features are added.
	 * (For IAgent Reasoning engines are added)
	 */
	void initialize(IObjectCreateFunctionArgs objectCreateFunctionArgs) throws ModelCompilingException;
}