package de.uniwue.ki.mas.entity.object.agent;
import java.io.Serializable;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.basics.function.iterator.EmptyIteratorSimulationRuntimeException;
import de.uniwue.ki.mas.behaviour.activityGraph.IActivityGraphLeaf;
import de.uniwue.ki.mas.behaviour.reasoning.*;
import de.uniwue.ki.mas.entity.object.IObjectCreateFunctionArgs;
import de.uniwue.ki.mas.entity.object.IObjectExecuteFunctionArgs;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * <BR>Creation date: (29.03.00 13:49:53)
 * @author Christoph Oechslein
 */

public class ActivityAgent extends AbstractAgent implements IActivityAgent {
	private final IActivityGraphLeaf[] _currentActivityGraphLeafs;

	public ActivityAgent(
		IActivityAgentInstanceDeclaration declaration,
		IObjectCreateFunctionArgs objectCreateFunctionArgs) {

		super(declaration, objectCreateFunctionArgs);

		int reasoningEngineSize =
			declaration
				.getActivityAgentClassDeclaration()
				.getReasoningEngineDeclarationListAndFactory()
				.getDeclarationList()
				.size();
		_currentActivityGraphLeafs = new IActivityGraphLeaf[reasoningEngineSize];
	}

	public IActivityGraphLeaf[] getCurrentActivityGraphLeafs() {
		return _currentActivityGraphLeafs;
	}

	public void remove() {
		for (int i = 0; i < _currentActivityGraphLeafs.length; i++) {
			_currentActivityGraphLeafs[i] = null;
		}

		super.remove();
	}

	public void reset(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		super.reset(objectExecuteFunctionArgs);

		for (int i = 0; i < getReasoningEngineCount(); i++) {
			setCurrentAtomicActivityAt(
				i,
				getActivityReasoningEngineAt(i).getActivityGraph().getEntryActivityGraphLeaf(),
				(IActivityAgentExecuteFunctionArgs) objectExecuteFunctionArgs);
		}
	}

	private void setCurrentAtomicActivityAt(
		int index,
		IActivityGraphLeaf newCurrentAtomicActivity,
		IActivityAgentExecuteFunctionArgs functionArgs) {

		Assert.assertNotNull(newCurrentAtomicActivity);

		IActivityGraphLeaf oldCurrentAtomicActivity = _currentActivityGraphLeafs[index];
		_currentActivityGraphLeafs[index] = newCurrentAtomicActivity;

		functionArgs.fireCurrentActivityChanged(
			this,
			oldCurrentAtomicActivity,
			newCurrentAtomicActivity,
			index);
	}

	protected final boolean isReasoningFinished(int index) {
		return !getCurrentActivityGraphLeafs()[index].isInstantly();
	}

	protected static class ActivityReasoningEngineCalculationArgs
		implements IActivityReasoningEngineCalculationArgs, Serializable {

		private IActivityGraphLeaf _currentActivityGraphLeaf;
		private IAgentExecuteFunctionArgs _agentExecuteFunctionArgs;

		public IAgentExecuteFunctionArgs getAgentExecuteFunctionArgs() {
			return _agentExecuteFunctionArgs;
		}

		void setAgentExecuteFunctionArgs(IAgentExecuteFunctionArgs agentExecuteFunctionArgs) {
			_agentExecuteFunctionArgs = agentExecuteFunctionArgs;
		}

		public IActivityGraphLeaf getCurrentActivityGraphLeaf() {
			return _currentActivityGraphLeaf;
		}

		protected void setCurrentActivityGraphLeaf(IActivityGraphLeaf currentActivityGraphLeaf) {
			_currentActivityGraphLeaf = currentActivityGraphLeaf;
		}
	}

	private final ActivityReasoningEngineCalculationArgs _activityReasoningEngineCalculationArgs =
		new ActivityReasoningEngineCalculationArgs();

	protected void reasonAndAct(int index, IAgentExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		IActivityReasoningEngine activityReasoningEngine = getActivityReasoningEngineAt(index);

		if (!activityReasoningEngine.getActivityGraph().isNoopActivityGraph()) {
			super.reasonAndAct(index, objectExecuteFunctionArgs);
		}
	}

	protected final void act(int index, IAgentExecuteFunctionArgs agentExecuteFunctionArgs)
		throws SimulationRuntimeException {

		IActivityAgentExecuteFunctionArgs activityAgentExecuteFunctionArgs =
			(IActivityAgentExecuteFunctionArgs) agentExecuteFunctionArgs;

		activityAgentExecuteFunctionArgs.fireStartsActing(this, index);
		//Ensure invariant, that the Bindings are empty, Rainer 11.Mai 2004
		activityAgentExecuteFunctionArgs.clearArgumentBindings();

		try {
			getCurrentAction().act(agentExecuteFunctionArgs);

		} catch (EmptyIteratorSimulationRuntimeException exception) {
			// ignore exception
		}
		
		//Assert.assertTrue(!activityAgentExecuteFunctionArgs.hasArgumentBindings());
		
		activityAgentExecuteFunctionArgs.fireFinishedActing(this, index);
	}

	protected void reason(int index, IAgentExecuteFunctionArgs agentExecuteFunctionArgs)
		throws SimulationRuntimeException {

		IActivityAgentExecuteFunctionArgs activityAgentExecuteFunctionArgs =
			(IActivityAgentExecuteFunctionArgs) agentExecuteFunctionArgs;

		activityAgentExecuteFunctionArgs.fireStartsReasoning(this, index);

		_activityReasoningEngineCalculationArgs.setCurrentActivityGraphLeaf(
			getCurrentActivityGraphLeafs()[index]);
		_activityReasoningEngineCalculationArgs.setAgentExecuteFunctionArgs(agentExecuteFunctionArgs);

		IActivityReasoningEngine activityReasoningEngine = getActivityReasoningEngineAt(index);

		IActivityReasoningEngineResult activityReasoningEngineResult =
			(IActivityReasoningEngineResult) activityReasoningEngine.calculateReasoningEngineResult(
				_activityReasoningEngineCalculationArgs);

		if (activityReasoningEngineResult.hasCurrentActivityGraphLeafEnded()) {
			getCurrentActivityGraphLeafs()[index] = activityReasoningEngineResult.getNextActivityGraphLeaf();
		}

		IFunction nextAction = activityReasoningEngineResult.getAction();

		agentExecuteFunctionArgs.fireCurrentActionChanged(this, getCurrentAction(), nextAction);

		setCurrentAction(nextAction);

		activityAgentExecuteFunctionArgs.fireFinishedReasoning(this, index);
	}

	protected IActivityReasoningEngine getActivityReasoningEngineAt(int index) {
		return (IActivityReasoningEngine) getReasoningEngineAt(index);
	}
}