package de.uniwue.ki.mas.entity.object.world;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.entity.object.*;
import de.uniwue.ki.mas.entity.object.agent.ActivityAgent;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.math.RandomHelper;

/**
 * Creation date: (23.08.2001 18:21:14)
 * @author Christoph Oechslein
 */

class ActivityWorld extends ActivityAgent implements IActivityWorld {
	private final HashSet _removedObjects = new HashSet();
	private final LinkedList _addedObjects = new LinkedList();
	private final LinkedList _objectList = new LinkedList();

	private WorldListener _aWorldListener = null;

	public ActivityWorld(
		IActivityWorldInstanceDeclaration declaration,
		IObjectCreateFunctionArgs objectCreateFunctionArgs) {

		super(declaration, objectCreateFunctionArgs);
	}

	public final void addWorldListener(WorldListener newListener) {
		_aWorldListener = WorldEventMulticaster.add(_aWorldListener, newListener);
	}

	protected void fireAddedObject(IObject activeObject, IObject object) {
		if (_aWorldListener != null)
			_aWorldListener.addedObject(new WorldEvent(this, activeObject, object));
	}

	protected void fireRemovedObject(IObject activeObject, IObject object) {
		if (_aWorldListener != null)
			_aWorldListener.removedObject(new WorldEvent(this, activeObject, object));
	}

	public final void removeWorldListener(WorldListener newListener) {
		_aWorldListener = WorldEventMulticaster.remove(_aWorldListener, newListener);
	}

	public void addObjectInSimulation(IObject newObject, IWorldExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		_addedObjects.add(newObject);
		fireAddedObject(objectExecuteFunctionArgs.getObject(), newObject);
	}

	private void addStartObjects(IWorldExecuteFunctionArgs worldExecuteFunctionArgs)
		throws SimulationRuntimeException {
		try {

			Assert.assertTrue(_addedObjects.isEmpty());

			createStartObjects(worldExecuteFunctionArgs);
			initializeStartObjects(worldExecuteFunctionArgs);
			notifyListenerForStartObjects(worldExecuteFunctionArgs);

			updateObjectGroup();

		} catch (ModelCompilingException exception) {
			throw new SimulationRuntimeException(exception);
		}
	}

	//	private void addStartObjectsOld(IWorldExecuteFunctionArgs objectExecuteFunctionArgs)
	//		throws ModelCompilingException, SimulationRuntimeException {
	//		IWorldCreateFunctionArgs worldCreateFunctionArgs =
	//			(IWorldCreateFunctionArgs) objectExecuteFunctionArgs.createCreateAndExecuteFunctionArgs();
	//
	//		Iterator it =
	//			((IWorldInstanceDeclaration) getObjectInstanceDeclaration())
	//				.getObjectInstanceDeclarationList()
	//				.iterator();
	//
	//		while (it.hasNext()) {
	//			IObjectInstanceDeclaration objectInstanceDeclaration = (IObjectInstanceDeclaration) it.next();
	//
	//			IObject object = objectInstanceDeclaration.createObject(worldCreateFunctionArgs);
	//			object.initialize(worldCreateFunctionArgs);
	//
	//			addObjectInSimulation(object, objectExecuteFunctionArgs);
	//		}
	//	}

	private void createStartObjects(IWorldExecuteFunctionArgs worldExecuteFunctionArgs)
		throws ModelCompilingException {

		IWorldCreateFunctionArgs worldCreateFunctionArgs =
			(IWorldCreateFunctionArgs) worldExecuteFunctionArgs.createCreateAndExecuteFunctionArgs();

		Iterator it =
			((IWorldInstanceDeclaration) getObjectInstanceDeclaration())
				.getObjectInstanceDeclarationList()
				.iterator();

		while (it.hasNext()) {
			IObjectInstanceDeclaration objectInstanceDeclaration = (IObjectInstanceDeclaration) it.next();
			_addedObjects.add(objectInstanceDeclaration.createObject(worldCreateFunctionArgs));
		}
	}

	private void initializeStartObjects(IWorldExecuteFunctionArgs worldExecuteFunctionArgs)
		throws ModelCompilingException {

		IWorldCreateFunctionArgs worldCreateFunctionArgs =
			(IWorldCreateFunctionArgs) worldExecuteFunctionArgs.createCreateAndExecuteFunctionArgs();

		Iterator it = _addedObjects.iterator();
		while (it.hasNext()) {
			IObject obj = (IObject) it.next();
			obj.initialize(worldCreateFunctionArgs);
		}
	}

	private void notifyListenerForStartObjects(IWorldExecuteFunctionArgs worldExecuteFunctionArgs) {
		Iterator it = _addedObjects.iterator();
		while (it.hasNext()) {
			IObject obj = (IObject) it.next();
			fireAddedObject(worldExecuteFunctionArgs.getObject(), obj);
		}
	}

	public Iterator getAllObjects(boolean includingAdded, boolean excludingRemoved) {
		if (!includingAdded && !excludingRemoved) {
			return _objectList.iterator();

		} else if (_addedObjects.isEmpty() && _removedObjects.isEmpty()) {
			return _objectList.iterator();

		} else if (includingAdded && _removedObjects.isEmpty()) {
			return getCurrentObjectsWithAdded();

		} else if (excludingRemoved) {
			// object could be added and removed in the same tick
			// -> object both in added and removed list
			return CollectionsHelper.setDifference(getCurrentObjectsWithAdded(), _removedObjects);

		} else {
			return _objectList.iterator();
		}
	}

	private Iterator getCurrentObjectsWithAdded() {
		return CollectionsHelper.concatenate(_objectList.iterator(), _addedObjects.iterator());
	}

	public final int getObjectCount(boolean includingAdded, boolean excludingRemoved) {
		int count = _objectList.size();

		if (includingAdded && excludingRemoved) {
			// objects could be added and removed in the same tick! 

			if (_addedObjects.isEmpty() && _removedObjects.isEmpty()) {
				return _objectList.size();

			} else if (_removedObjects.isEmpty()) {
				// !_addedObjects.isEmpty()
				return _objectList.size() + _addedObjects.size();

			} else {
				return CollectionsHelper.size(getAllObjects(true, true));
			}

		} else {
			if (includingAdded) {
				// !excludingRemoved
				count += _addedObjects.size();
			} else {
				// !excludingRemoved && !includingAdded
				// do nothing
			}
		}

		return count;
	}

	public void remove() {
		_aWorldListener = null;

		removeAllObjects(null);

		super.remove();
	}

	private void removeAllObjects(IWorldExecuteFunctionArgs worldExecuteFunctionArgs) {
		updateObjectGroup();

		removeAllObjects(_objectList.iterator(), worldExecuteFunctionArgs);
		_objectList.clear();
	}

	private void removeAllObjects(Iterator it, IWorldExecuteFunctionArgs worldExecuteFunctionArgs) {
		while (it.hasNext()) {
			IObject currObject = (IObject) it.next();

			if (!currObject.isDead())
				removeObjectInSimulation(currObject, worldExecuteFunctionArgs);

			currObject.remove();
		}
	}

	public void removeObjectInSimulation(
		IObject anObject,
		IWorldExecuteFunctionArgs objectExecuteFunctionArgs) {

		if (Assert.ASSERTION_ON)
			Assert.assertEquals(false, anObject.isDead());

		// don't call remove of anObject it could be referenced by other agents

		if (anObject == this) {
			// stop the simulation by killing the world
			setDead(true);

		} else {

			// changed: set dead attribute in object to true -> no update, reason, act
			//          and no problem with map (due to fire removed object) 
			anObject.setDead(true);
			_removedObjects.add(anObject);

			if (objectExecuteFunctionArgs != null) {
				fireRemovedObject(objectExecuteFunctionArgs.getObject(), anObject);
			}
		}
	}

	public void reset(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		removeAllObjects((IWorldExecuteFunctionArgs) objectExecuteFunctionArgs);
		Assert.assertEquals(0, getObjectCount(true, false));

		super.reset(objectExecuteFunctionArgs);

		addStartObjects((IWorldExecuteFunctionArgs) objectExecuteFunctionArgs);
	}

	public final void shuffleObjects() {
		RandomHelper.shuffle(_objectList);
	}

	public void update(IObjectExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {
		updateObjectGroup();

		super.update(objectExecuteFunctionArgs);
	}

	private void updateObjectGroup() {
		// first add new objects then remove died ones
		// since an object can be added and removed in the same tick!!

		if (!_addedObjects.isEmpty()) {
			// already resetet in addObjInSim

			_objectList.addAll(_addedObjects);
			_addedObjects.clear();
		}

		if (!_removedObjects.isEmpty()) {

			int removedObjectCount = _removedObjects.size();
			Iterator it = _objectList.iterator();
			while (it.hasNext() && (removedObjectCount > 0)) {
				IObject currObject = (IObject) it.next();
				if (_removedObjects.contains(currObject)) {
					it.remove();

					// now set in removeObjInSim
					//currObject.setDead(true);

					removedObjectCount--;
				}
			}

			_removedObjects.clear();
		}
	}

	protected void initializeIntern(IObjectCreateFunctionArgs objectCreateFunctionArgs)
		throws ModelCompilingException {

		IWorldCreateFunctionArgs worldCreateFunctionArgs =
			(IWorldCreateFunctionArgs) objectCreateFunctionArgs;

		IWorld oldActiveWorld = worldCreateFunctionArgs.getWorld();

		worldCreateFunctionArgs.setWorld(this);

		super.initializeIntern(objectCreateFunctionArgs);

		if (Assert.ASSERTION_ON)
			Assert.assertSame(this, worldCreateFunctionArgs.getWorld());

		worldCreateFunctionArgs.setWorld(oldActiveWorld);
	}
}