package de.uniwue.ki.mas.entity.variableclass;
import java.util.*;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.functionCall.FunctionCallDeclarationFactory;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallDeclaration;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.basics.type.TypeManager;
import de.uniwue.ki.mas.util.declaration.AbstractDocumentableDeclaration;
import de.uniwue.ki.util.hash.IIsomorphicHashable;
/**
 * Creation date: (26.07.2001 12:34:33)
 * @author Christoph Oechslein
 */

public abstract class AbstractVariableClassDeclaration
	extends AbstractDocumentableDeclaration
	implements IVariableClassDeclaration, ChangeListener {

	private boolean _external = false;
	private boolean _writeable = false;
	private boolean _partOfName = false;
	
	private ISpecificType _specificType;
	private final IFunctionCallDeclaration _defaultValueFunctionCallDeclaration;

	private final List _bodyVariableClassDeclarationList = new LinkedList();

	public AbstractVariableClassDeclaration() {
		this(TypeManager.getInstance().getDefaultTypeForVariableClassDeclaration());
	}

	/**
	 * @param type ISpecificType
	 */
	public AbstractVariableClassDeclaration(ISpecificType type) {
		this(type, FunctionCallDeclarationFactory.getInstance().createFunctionCallDeclaration(type));
	}

	public AbstractVariableClassDeclaration(ISpecificType type, IFunctionCallDeclaration defaultValue) {
		super(true);

		setUsesShortDescription(true);
		setShortDescription("Untitled Variable");

		if (Assert.ASSERTION_ON)
			Assert.assertEquals(type, defaultValue.getSpecificOutputType());

		setSpecificType(type);

		_defaultValueFunctionCallDeclaration = defaultValue;

		_defaultValueFunctionCallDeclaration.addChangeListener(this);
	}

	/**
	 * @param newListener ChangeListener
	 * @see de.uniwue.ki.mas.util.declaration.IDeclaration#addChangeListener(ChangeListener)
	 */
	public void addChangeListener(ChangeListener newListener) {
		super.addChangeListener(newListener);
	}

	/**
	 * @param inToObject Object
	 * @param objectMapping java.util.Map
	 * @see de.uniwue.ki.util.ICopyable#copyInto(Object, Map)
	 */
	public void copyInto(Object inToObject, java.util.Map objectMapping) {
		super.copyInto(inToObject, objectMapping);

		IVariableClassDeclaration dstVariableClassDeclaration = (IVariableClassDeclaration) inToObject;

		dstVariableClassDeclaration.setWriteable(isWriteable());
		dstVariableClassDeclaration.setExternal(isExternal());
		dstVariableClassDeclaration.setPartOfName(isPartOfName());

		dstVariableClassDeclaration.setSpecificType(getSpecificType());

		getDefaultValueFunctionCallDeclaration().copyInto(
			dstVariableClassDeclaration.getDefaultValueFunctionCallDeclaration(),
			objectMapping);
	}

	/**
	 * @return IFunctionCallDeclaration
	 * @see de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration#getDefaultValueFunctionCallDeclaration()
	 */
	public IFunctionCallDeclaration getDefaultValueFunctionCallDeclaration() {
		return _defaultValueFunctionCallDeclaration;
	}

	/**
	 * @return ISpecificType
	 * @see de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration#getSpecificType()
	 */
	public ISpecificType getSpecificType() {
		return _specificType;
	}

	private void initializeDefaultValueFunctionCallDeclaration() {
		initializeFunctionCallDeclaration(getDefaultValueFunctionCallDeclaration());
	}

	protected void initializeFunctionCallDeclaration(IFunctionCallDeclaration functionCallDeclaration) {
		if ((getSpecificType() != null) && functionCallDeclaration != null) {
			functionCallDeclaration.setMostGeneralOutputType(getSpecificType());
		}
	}

	/**
	 * @return boolean
	 * @see de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration#isExternal()
	 */
	public boolean isExternal() {
		return _external;
	}

	/**
	 * @param anObject IIsomorphicHashable
	 * @return boolean
	 * @see de.uniwue.ki.util.hash.IIsomorphicHashable#isomorphicEquals(IIsomorphicHashable)
	 */
	public boolean isomorphicEquals(IIsomorphicHashable anObject) {
		return equals(anObject);

		// leads to error in the compiling
		/*
		if (anObject == this)
		    return true;
		
		if (!super.isomorphicEquals(anObject)) {
		    return false;
		}
		
		if (!(anObject instanceof IVariableClassDeclaration)) {
		    return false;
		}
		
		IVariableClassDeclaration variableClassDeclaration = (IVariableClassDeclaration) anObject;
		
		if (!isomorphicEquals(getDefaultValueFunctionCallDeclaration(), variableClassDeclaration.getDefaultValueFunctionCallDeclaration())) {
		    return false;
		}
		
		if (getHasNextValueFunctionCallDeclaration()) {
		    return isomorphicEquals(getNextValueFunctionCallDeclaration(), variableClassDeclaration.getNextValueFunctionCallDeclaration())
		        && variableClassDeclaration.getHasNextValueFunctionCallDeclaration() == true;
		} else {
		
		    return variableClassDeclaration.getHasNextValueFunctionCallDeclaration() == false;
		}
		*/
	}

	/**
	 * @return int
	 * @see de.uniwue.ki.util.hash.IIsomorphicHashable#isomorphicHashCode()
	 */
	public int isomorphicHashCode() {
		return hashCode();

		// leads to error in the compiling
		/*
		if (getHasNextValueFunctionCallDeclaration())
		    return 11 * super.isomorphicHashCode()
		        + 7 * isomorphicHashCode(getDefaultValueFunctionCallDeclaration())
		        + 3 * isomorphicHashCode(getNextValueFunctionCallDeclaration())
		        + 1;
		else
		    return 11 * super.isomorphicHashCode() + 7 * isomorphicHashCode(getDefaultValueFunctionCallDeclaration());
		*/
	}

	public boolean isWriteable() {
		return _writeable;
	}

	/**
	 * Remove Object (this).
	 * Gets called if Object will be deleted from modell
	 * @see de.uniwue.ki.util.IRemoveable#remove()
	 */
	public void remove() {
		getDefaultValueFunctionCallDeclaration().removeChangeListener(this);
		getDefaultValueFunctionCallDeclaration().remove();

		super.remove();
	}

	public void setExternal(boolean externalModifiable) {
		final boolean oldValue = _external;
		_external = externalModifiable;
		firePropertyChange("external", Boolean.valueOf(oldValue), Boolean.valueOf(externalModifiable));
		fireStateChanged();
	}

	public void setSpecificType(ISpecificType newType) {
		ISpecificType oldValue = _specificType;
		_specificType = newType;

		firePropertyChange("specificType", oldValue, newType);
		firePropertyChange("type", oldValue, newType);
		fireStateChanged();

		initializeDefaultValueFunctionCallDeclaration();
	}

	/**
	 * @param writeable boolean
	 * @see de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration#setWriteable(boolean)
	 */
	public void setWriteable(boolean writeable) {
		final boolean oldValue = _writeable;
		_writeable = writeable;
		firePropertyChange("writeable", Boolean.valueOf(oldValue), Boolean.valueOf(writeable));
		fireStateChanged();
	}

	/**
	 * @param e ChangeEvent
	 * @see javax.swing.event.ChangeListener#stateChanged(ChangeEvent)
	 */
	public void stateChanged(ChangeEvent e) {
		fireStateChanged();
	}

	public void addBodyVariableClassDeclaration(IBodyVariableClassDeclaration BodyVariableClassDeclaration) {
		Assert.assertFalse(_bodyVariableClassDeclarationList.contains(BodyVariableClassDeclaration));

		_bodyVariableClassDeclarationList.add(BodyVariableClassDeclaration);
	}

	public void removeBodyVariableClassDeclaration(IBodyVariableClassDeclaration BodyVariableClassDeclaration) {
		Assert.assertTrue(_bodyVariableClassDeclarationList.contains(BodyVariableClassDeclaration));

		_bodyVariableClassDeclarationList.remove(BodyVariableClassDeclaration);
	}

	public Iterator getBodyVariableClassDeclarations() {
		return _bodyVariableClassDeclarationList.iterator();
	}
    
	public boolean isPartOfName() {
        return _partOfName;
    }
    
	public void setPartOfName(boolean ofName) {
		final boolean oldValue = _partOfName;
		_partOfName = ofName;
		firePropertyChange("partOfName", Boolean.valueOf(oldValue), Boolean.valueOf(ofName));
		fireStateChanged();
    }
}