/**
 * @author glueckselig
 *
 * This is the mainClass for the HTMLWrite-module. Since this is the entry-point this is the only public
 * class in this package. All other classes are visible in the package only...
 * This class invokes all the other classes to get the htmlfile written.
 * 
 */

package de.uniwue.ki.mas.modelprint;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;

import de.uniwue.ki.mas.basics.feature.IObjectFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.feature.user.IUserFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.featureList.IFeatureClassDeclarationList;
import de.uniwue.ki.mas.basics.featureList.IObjectFeatureClassDeclarationList;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallDeclaration;
import de.uniwue.ki.mas.basics.functionUser.UserFunctionDeclarationList;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.*;
import de.uniwue.ki.mas.behaviour.reasoning.IActivityReasoningEngineDeclaration;
import de.uniwue.ki.mas.entity.object.IObjectInstanceDeclaration;
import de.uniwue.ki.mas.entity.objectclass.agentclass.AgentClassDeclarationList;
import de.uniwue.ki.mas.entity.objectclass.agentclass.IAgentClassDeclaration;
import de.uniwue.ki.mas.entity.variable.IBodyVariableInstanceDeclaration;
import de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration;
import de.uniwue.ki.mas.model.*;
import de.uniwue.ki.mas.modelprint.auxiliary.TagStack;
import de.uniwue.ki.mas.modelprint.auxiliary.UniqueAnchorGenerator;
import de.uniwue.ki.mas.sim.analysis.*;
import de.uniwue.ki.mas.sim.experiment.IExperimentDeclaration;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.sim.situation.ISituationDeclaration;
import de.uniwue.ki.mas.util.declaration.IDeclarationList;
import de.uniwue.ki.mas.xml.XMLReader;
import de.uniwue.ki.mas.xml.reader.XMLReadException;

public class HTMLModelWriter extends HTMLModelWriterBase {

	private ISeSAmDeclaration _seSAmDeclaration;
	private HTMLHeaderWriter _headerWriter;
	private HTMLVariableWriter _variableWriter;
	private HTMLInterlinkageWriter _interlinkageWriter;

	private static final String FILE_IN = "G:\\XML\\C4 mit FSG & VV & RM_15-07_2.xml";
	//ohne XSL Thresholdmodell2805_2.xml";
	//Original meinungsmodell_12 mit chainrule condition.xml";

	private HTMLModelWriter(File file, ISeSAmDeclaration seSAmDeclaration) {
		_idMap = new UniqueAnchorGenerator();
		_tags = new TagStack();

		_seSAmDeclaration = seSAmDeclaration;

		try {
			_htmlWriter = new BufferedHTMLWriter(file);
		} catch (IOException e) {
			System.out.println("Unable to open output-file for HTMLwriter.");
		}

		_headerWriter = new HTMLHeaderWriter(_seSAmDeclaration);
		_functionsWriter = new HTMLFunctionsWriter();
		_variableWriter = new HTMLVariableWriter();
		_interlinkageWriter = new HTMLInterlinkageWriter();
	}

	private HTMLModelWriter() {
		super();
	}

	/**
	 * Method writeHTMLFile. Invokes @see writeHTMLFile(File, ISeSAmDeclaration) 
	 * @param filename Name of the file to which the html shall be stored.
	 * @param seSAmDeclaration the declaration from which all the information gets retrieved.
	 */
	public static void writeHTMLFile(String filename, ISeSAmDeclaration seSAmDeclaration) {
		writeHTMLFile(new File(filename), seSAmDeclaration);
	}

	/**
	 * Method writeHTMLFile. Creates an object of HTMLModelWriter via its private constructor
	 * and initiates the html-processing by calling @see startHTMLWrite()
	 * @param file The file to which the html shall be stored.
	 * @param seSAmDeclaration
	 */
	public static void writeHTMLFile(File file, ISeSAmDeclaration seSAmDeclaration) {
		new HTMLModelWriter(file, seSAmDeclaration).startHTMLWrite();
	}

	/**
	 * Method startHTMLWrite. Initiates and distributes the different jobs to different classes and functions:
	 * writeDocumentHeader/Footer, Head, UserFunctions, UserFeatures, Classes, Situation, Analysis and Simulation.
	 */
	private void startHTMLWrite() {

		_htmlWriter.writeDocumentHeader();
		_htmlWriter.writeHead("0.1");

		//##########################################
		_headerWriter.writeOutContent();
		writeAnchor("topOfPage");
		writeUserFunctions(
			_seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSeSAmModelBasicsDeclaration()
				.getUserFunctionDeclarationListAndFactory()
				.getUserFunctionDeclarationList());

		writeUserFeatures(
			_seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSeSAmModelBasicsDeclaration()
				.getUserFeatureClassDeclarationListAndFactory()
				.getUserFeatureClassDeclarationList());

		writeAgentOrWorldClass(
			_seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getAgentClassDeclarationListAndFactory()
				.getAgentClassDeclarationList());

		writeAgentOrWorldClass(
			_seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getWorldClassDeclarationListAndFactory()
				.getWorldClassDeclarationList());

		writeSituationAndSimulation(
			_seSAmDeclaration.getSeSAmModelDeclaration().getSimulationElementsDeclaration());

		writeExperiments(
			_seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getExperimentDeclarationListAndFactory()
				.getExperimentDeclarationList()
				.iterator());

		//##########################################	
		_htmlWriter.writeDocumentFooter();
		_htmlWriter.closeWriter();

		System.out.println("HTMLModelWriter.startHTMLWrite(): All elements closed? " + _tags.empty());
	}

	private void writeAgentOrWorldClass(AgentClassDeclarationList agentClassDecl) {
		Iterator it = agentClassDecl.iterator();

		while (it.hasNext()) {
			IAgentClassDeclaration ACDecl = (IAgentClassDeclaration)it.next();

			String className = ACDecl.getClass().getName();

			write(
				createAnchor(ACDecl)
					+ _tags.openTag(CLASS_HEADING)
					+ "Name of "
					+ escape(
						className.substring(
							className.indexOf("Activity") + "Activity".length(),
							className.length() - "Declaration".length()))
					+ ": "
					+ escape(ACDecl.getName())
					+ _tags.closeTag());

			writeFeatures(ACDecl.getObjectFeatureClassDeclarationList());

			_variableWriter.writeBodyVariables(ACDecl.getBodyClassDeclaration());

			writeReasonings(ACDecl.getReasoningEngineDeclarationListAndFactory().getDeclarationList());

			write("<br><br><br><br>");
		}
	}

	private void writeSituationAndSimulation(ISimulationElementsDeclaration elementDecl) {
		writeSituation(
			elementDecl.getSituationDeclarationListAndFactory().getSituationDeclarationList().iterator());
		writeAnalysis(elementDecl.getAnalysisDeclarationListAndFactory().getDeclarationList().iterator());
		writeSimulation(
			elementDecl.getSimulationDeclarationListAndFactory().getSimulationDeclarationList().iterator());
	}

	private void writeSimulation(Iterator simIterator) {
		write("<br><br>" + _tags.openTag(SUB1_HEADING) + "Simulations" + _tags.closeTag());

		while (simIterator.hasNext()) {
			ISimulationDeclaration currentSim = (ISimulationDeclaration)simIterator.next();
			writeAnchor(currentSim);
			write(_tags.openTag(SUB2_HEADING) + escape(currentSim.getName()) + _tags.closeTag());

			write(
				"Situation: "
					+ createLink(
						currentSim.getSituationDeclaration(),
						currentSim.getSituationDeclaration().getName()));

			writeSimulation_Analysis(currentSim);
			writeSimulation_EndFunctions(currentSim);

			write("<br><br><br>");
		}
	}

	private void writeSimulation_EndFunctions(ISimulationDeclaration currentSim) {
		Iterator it = currentSim.getEndFunctionDeclarationListAndFactory().getDeclarationList().iterator();

		StringBuffer result = new StringBuffer();

		if (it.hasNext())
			result.append(_tags.openTag(SUB3_HEADING) + "EndFunctions" + _tags.closeTag());

		while (it.hasNext()) {
			IFunctionCallDeclaration element = (IFunctionCallDeclaration)it.next();
			result.append(HTMLFunctionsWriter.createFunctionCall(element));
		}

		write(result.toString());
	}

	private void writeSimulation_Analysis(ISimulationDeclaration currentSim) {
		Iterator analysisIt = currentSim.getAnalysisDeclarationList().iterator();

		StringBuffer result = new StringBuffer();

		if (analysisIt.hasNext()) {
			result.append(_tags.openTag(SUB3_HEADING) + "Analysis" + _tags.closeTag());
		}
		while (analysisIt.hasNext()) {
			IAnalysisDeclaration element = (IAnalysisDeclaration)analysisIt.next();
			result.append(createLink(element, element.getName()));
			if (analysisIt.hasNext())
				result.append("<br>");
		}

		write(result.toString());
	}
	private void writeAnalysis(Iterator analysisIterator) {
		if (analysisIterator.hasNext())
			write(_tags.openTag(SUB1_HEADING) + "Analysis'" + _tags.closeTag());

		StringBuffer buffer = new StringBuffer();
		while (analysisIterator.hasNext()) {
			buffer.append(createSingleAnalysis((IAnalysisDeclaration)analysisIterator.next()));
		}

		write(buffer.toString());
	}

	private String createSingleAnalysis(IAnalysisDeclaration element) {
		StringBuffer buffer = new StringBuffer();

		buffer.append(
			_tags.openTag(SUB2_HEADING)
				+ createAnchor(element)
				+ escape(element.getName())
				+ _tags.closeTag());

		if (element.getAnalysisTypeDeclaration() instanceof CSVAnalysisTypeDeclaration) {
			CSVAnalysisTypeDeclaration currentType =
				(CSVAnalysisTypeDeclaration)element.getAnalysisTypeDeclaration();

			buffer.append(
				_tags.openTag(CSV_BLOCK)
					+ "CSV-analysis: <u>"
					+ escape(element.getName())
					+ "</u> (Type: "
					+ escape(currentType.getName())
					+ ")");
			buffer.append("<br>CSV file: " + escape(currentType.getFileString()) + _tags.closeTag());
		}

		Iterator analysisItemIterator =
			element.getAnalysisItemDeclarationListAndFactory().getAnalysisItemDeclarationList().iterator();

		while (analysisItemIterator.hasNext()) {
			IAnalysisItemDeclaration currentItem = (IAnalysisItemDeclaration)analysisItemIterator.next();

			buffer.append(_tags.openTag(ANALYSIS_ITEM_MARGIN_TOP));

			buffer.append(escape(currentItem.getName()) + "<br>");
			buffer.append(
				HTMLFunctionsWriter.createFunctionCall(currentItem.getValueFunctionCallDeclaration()));

			buffer.append(_tags.closeTag());
		}

		buffer.append("<br><br><br>");

		return buffer.toString();
	}

	private void writeSituation(Iterator sitIterator) {
		write("<br><br>" + _tags.openTag(SUB1_HEADING) + "Situations" + _tags.closeTag());

		while (sitIterator.hasNext()) {
			ISituationDeclaration currentSit = (ISituationDeclaration)sitIterator.next();
			writeAnchor(currentSit);
			write(_tags.openTag(SUB2_HEADING) + escape(currentSit.getName()) + _tags.closeTag());

			writeObjectInstanceDeclaration(currentSit.getWorldInstanceDeclaration());

			Iterator it =
				currentSit.getWorldInstanceDeclaration().getObjectInstanceDeclarationList().iterator();

			while (it.hasNext()) {
				IObjectInstanceDeclaration element = (IObjectInstanceDeclaration)it.next();

				writeObjectInstanceDeclaration(element);
			}

			write("<br><br>");
		}
	}

	private void writeExperiments(Iterator it) {
		StringBuffer result = new StringBuffer();

		if (it.hasNext()) {
			result.append("<br><br>" + _tags.openTag(SUB1_HEADING) + "Experiments" + _tags.closeTag());
		}

		while (it.hasNext()) {
			IExperimentDeclaration element = (IExperimentDeclaration)it.next();
			result.append(
				createAnchor(element)
					+ _tags.openTag(SUB2_HEADING)
					+ escape(element.getName())
					+ _tags.closeTag());
			result.append("Configurationfile: " + escape(element.getConfigurationFileName()));
			result.append(
				HTMLFunctionsWriter.createFunctionCall(element.getSimulationActionFunctionCallDeclaration()));

			result.append("<br>");
		}

		write(result.toString());
	}

	private void writeObjectInstanceDeclaration(IObjectInstanceDeclaration element) {
		write(
			escape(element.getName())
				+ ": "
				+ createLink(
					element.getObjectClassDeclaration(),
					element.getObjectClassDeclaration().getName()));

		writeBodyVariableInstanceDeclarations(
			element.getBodyInstanceDeclaration().getBodyVariableInstanceDeclarationList().iterator(),
			"Variables with non-default values");
	}

	private void writeBodyVariableInstanceDeclarations(Iterator it, String title) {
		if (it.hasNext()) {
			write(_tags.openTag(SUB3_HEADING) + escape(title) + _tags.closeTag());
		}

		writeVariableInstanceDeclaration(it);
	}

	private void writeVariableInstanceDeclaration(Iterator it) {
		while (it.hasNext()) {
			IBodyVariableInstanceDeclaration element = (IBodyVariableInstanceDeclaration)it.next();
			if (element.getUsesOwnDefaulValueFunctionCallDeclaration()) {
				IFunctionCallDeclaration elemFunc = element.getDefaultValueFunctionCallDeclaration();

				write(escape(element.getName()));
				write(
					HTMLFunctionsWriter.createFunctionCallIntern(
						elemFunc.getFunctionCallInternDeclaration()));

				write("<br>");
			}
		}
	}

	private void writeUserFunctions(UserFunctionDeclarationList userList) {
		_functionsWriter.writeUserFunctionsMain(userList);
	}

	private void writeUserFeatures(IFeatureClassDeclarationList userFeatList) {
		write("<br><br>" + _tags.openTag(SUB1_HEADING) + "UserFeatures" + _tags.closeTag());

		Iterator iterator = userFeatList.iterator();
		while (iterator.hasNext()) {

			IUserFeatureClassDeclaration userFeat = ((IUserFeatureClassDeclaration)iterator.next());

			writeAnchor(userFeat);
			write("<br><br>" + _tags.openTag(SUB2_HEADING) + escape(userFeat.getName()) + _tags.closeTag());

			write("<br>");

			writeUserFeatureVariables(userFeat);

			_functionsWriter.writeUserFunctionsSubtitle(
				userFeat.getUserFunctionDeclarationListAndFactory().getUserFunctionDeclarationList());
		}
	}

	private void writeUserFeatureVariables(IUserFeatureClassDeclaration userFeat) {
		Iterator userFeatVariablesIt =
			userFeat
				.getFeatureVariableClassDeclarationListAndFactory()
				.getVariableClassDeclarationList()
				.iterator();
		write("<br><br>" + _tags.openTag(SUB3_HEADING) + "Variables" + _tags.closeTag());

		write("<table cellpadding=\"10\">");
		StringBuffer result = new StringBuffer();
		while (userFeatVariablesIt.hasNext()) {
			IVariableClassDeclaration element = (IVariableClassDeclaration)userFeatVariablesIt.next();
			result.append(_variableWriter.createVariableClassDeclaration(element));
		}

		write(result.toString());
		write("</table>");
	}

	private void writeReasoning(IActivityReasoningEngineDeclaration reasoning) {
		write("<br><br>" + _tags.openTag(SUB1_HEADING) + "Reasoning" + _tags.closeTag() + "<br><br>");
		IActivityGraphDeclaration composition = reasoning.getRootActivityGraphDeclaration();

		writeComposedActivity(composition);

		// the following writes out the chainRules or Interlinkage
		// since this is part of the reasoning it is here.
		write("<br><br><br><br>");
		_interlinkageWriter.writeReasoningInterlinkage(composition);
	}

	private void writeReasonings(IDeclarationList reasonings) {
		write("<br><br>" + _tags.openTag(SUB1_HEADING) + "Reasonings" + _tags.closeTag() + "<br><br>");

		Iterator reasIter = reasonings.iterator();
		while (reasIter.hasNext()) {
			IActivityReasoningEngineDeclaration reasoning =
				(IActivityReasoningEngineDeclaration)reasIter.next();

			writeReasoning(reasoning);
		}

		write("<br><br><br><br>");
	}

	private void writeComposedActivity(IActivityGraphDeclaration composition) {
		write("<br><br>" + _tags.openTag(SUB2_HEADING) + "Definitions of Activities" + _tags.closeTag());

		writeReasoningComposedActivityAllActivList(composition);
	}

	private void writeReasoningComposedActivityAllActivList(IActivityGraphDeclaration composedAct) {
		writeReasoningComposedActivityAllActivList(
			composedAct.getActivityGraphNodeDeclarationList().iterator());
	}

	// schreibt alle Funktionen einer Klasse
	private void writeReasoningComposedActivityAllActivList(Iterator activityIterator) {

		while (activityIterator.hasNext()) {
			IActivityGraphNodeDeclaration currentActivity =
				(IActivityGraphNodeDeclaration)activityIterator.next();

			write(_tags.openTag(ROOT_FUNCTION));
			writeAnchor(currentActivity);

			if (currentActivity instanceof IActivityGraphDeclaration) {
				IActivityGraphDeclaration composedAct = (IActivityGraphDeclaration)currentActivity;
				write("<em class=\"composedActivity\">" + escape(currentActivity.getName()) + "</em>");
				writeReasoningComposedActivityAllActivList(composedAct);

			} else if (currentActivity instanceof IActivityGraphLeafDeclaration) {
				IActivityGraphLeafDeclaration atomicAct = (IActivityGraphLeafDeclaration)currentActivity;
				write("<em>" + escape(currentActivity.getName()) + "</em>");
				if (atomicAct instanceof IActionActivityGraphLeafDeclaration) {

					write(
						HTMLFunctionsWriter.createFunctionCallIterator(
							((IActionActivityGraphLeafDeclaration)atomicAct)
								.getFunctionActionCallDeclarationListAndFactory()
								.getFunctionActionCallDeclarationList()
								.iterator()));
				}

				//				else if (atomicAct instanceof IDecisionAtomicActivityDeclaration) {
				//
				//				}
			}

			write(_tags.closeTag());
		}
	}

	private void writeFeatures(IObjectFeatureClassDeclarationList declList) {
		write(
			"<br><br>"
				+ _tags.openTag(SUB1_HEADING)
				+ escape(declList.getName())
				+ _tags.closeTag()
				+ "<br><br>");

		Iterator featuresIt = declList.iterator();
		while (featuresIt.hasNext()) {
			IObjectFeatureClassDeclaration currentDecl = ((IObjectFeatureClassDeclaration)featuresIt.next());
			String featureN = currentDecl.getName();
			write(createLink(currentDecl, featureN) + "<br>");
		}
	}

	public static void main(String[] args) {
		ISeSAmDeclaration seSAmDeclaration = SeSAmDeclarationFactory.createSeSAmDeclaration(false);
		File file = new File(FILE_IN);

		seSAmDeclaration.setFile(file);

		try {
			XMLReader.readModelInto(file, seSAmDeclaration, null, true);

		} catch (XMLReadException e) {
		} catch (IOException e) {
		}

		File htmlFile = new File(FILE_OUT);
		writeHTMLFile(htmlFile, seSAmDeclaration);

		System.out.println("fertig");
	}
}
