
package de.uniwue.ki.mas.refactoring;

import de.uniwue.ki.mas.basics.functionArgument.ISpecificFunctionDeclarationWithArgDeclBinds;
import de.uniwue.ki.mas.basics.functionCall.IAtomicFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;

/**
 * Refactoring "Extract Method" wird auf ein Objekt vom Typ
 * IAtomicFunctionCallInternDeclaration angewendet, das als Argument einer
 * Funktion hherer Ordnung vorkommt. Es ist als RefactoringSubject referenziert.
 * Es wird eine neue UserFunction erstellt, die eine Kopie des 
 * RefactoringSubjects als Funktions-Body enthlt. Das RefactoringSubject wird
 * gelscht und durch einen Aufruf der neuen UserFunction ersetzt.
 * 
 * @author Michael Scherer
 */
public class ExtractHigherOrderFunctionArgument extends ExtractMethod {

	public ExtractHigherOrderFunctionArgument(
		ISeSAmDeclaration seSAmDeclaration,
		Object _refactoringSubject) 
	{
		super(seSAmDeclaration, _refactoringSubject);
		this.name = "Extract Method (Function Argument)";
	}

	/**
	 * Fhrt das Refactoring "Extract Method" aus. Lediglich
	 * AtomicFunctionInternCallDeclarations, deren Argument auf eine (anonyme) 
	 * Funktion verweist, werden besonders behandelt (also Argumente von 
	 * Funktionen hherer Ordnung). Alle anderen Flle werden mit den Methoden 
	 * der Oberklasse ExtractMethod abgedeckt. 
	 * 
	 * Die Unterschiede zur Klasse <code>ExtractMethod</code>:
	 *    Der Wurzelknoten des Codebaumes, der als eine neue UserFunction extrahiert
	 *    werden soll, ist nicht durch das RefactoringSubject gegeben. Er muss mit
	 *    der Funktion getRootNode() ermittelt werden.
	 * 
	 * Ansonsten existieren keine Unterschiede zwischen ExtractMethod und diesem
	 * Refactoring.  
	 * 
	 * @param  ui  Userinterface das zur Interaktion mit dem Anwender benutzt werden soll
	 * @param  _defaultFunctionName  Standardwert fr den Namen der neu zu erstellenden UserFunction
	 * @return true, falls Refactoring erfolgreich, false anderenfalls
	 * @see de.uniwue.ki.mas.refactoring.AbstractRefactoring#execute(de.uniwue.ki.mas.refactoring.IRefactoringUI)
	 */
	public boolean execute(IRefactoringUI ui, String defaultName) 
	{
		return execute(ui, defaultName, getRootNode());
	}
	
	/**
	 * Ermittelt den Wurzelknoten des Codebaumes derjenigen Funktion, die als
	 * Argument einer Funktion hherer Ordnung verwendet wird (siehe auch
	 * isHigherOrderFunctionArgument()). 
	 * 
	 * @return Wurzelknoten des Codebaumes der Funktion
	 */
	private IFunctionCallInternDeclaration getRootNode() {
		try {
			return (
				((ISpecificFunctionDeclarationWithArgDeclBinds)
					((IAtomicFunctionCallInternDeclaration) getRefactoringSubject())
					.getArgumentDeclaration()
					.getObject()
				).getParametricFunctionCallDeclaration()
				.getFunctionCallDeclaration()
				.getFunctionCallInternDeclaration()
			);
		} catch (Exception ex) {
			return null;
		}
	}
	
	/**
	 * Entscheidet, ob das Refactoring auf das als RefactoringSubject 
	 * referenzierte Objekt angewendet werden kann.
	 * 
	 * @return true, wenn Anwendung mglich, false sonst
	 */
	public boolean isApplicable() {
		return (
			(getRefactoringSubject() instanceof IAtomicFunctionCallInternDeclaration) && 
			!isUserFunctionDeclaration(getRefactoringSubject()) &&
			isHigherOrderFunctionArgument()
		);
	}
}
