package de.uniwue.ki.mas.refactoring;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import de.uniwue.ki.mas.basics.feature.user.IUserFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IAtomicFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IComposedFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionDeclaration.IFunctionDeclaration;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.basics.functionUser.UserFunctionDeclarationList;
import de.uniwue.ki.mas.entity.variableclass.IFeatureVariableClassDeclaration;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.functions.IMapFunction;

/**
 * @author Michael Scherer
 */
public class MoveToUserFeature extends AbstractRefactoring {

	public MoveToUserFeature(ISeSAmDeclaration seSAmDeclaration,
			Object _refactoringSubject) {
		super(seSAmDeclaration, _refactoringSubject);
		this.name = "Move To User Feature";
	}
	
	public boolean isApplicable() {
		return (
			isUserFunctionWithoutInternalReferences(getRefactoringSubject()) &&
			getTargetFeatureClassesList((IUserFunctionDeclaration) getRefactoringSubject()).isEmpty() == false
		);
	}

	protected boolean isUserFunctionWithoutInternalReferences(Object function) {
		if (!(function instanceof IUserFunctionDeclaration))
			return false;

		IUserFunctionDeclaration subject = (IUserFunctionDeclaration) function;
		IFunctionCallInternDeclaration functionRoot = subject.getParametricFunctionCallDeclaration()
		.getFunctionCallDeclaration().getFunctionCallInternDeclaration();

		return (
			getInternalUserFeatureFunctionCalls(functionRoot).isEmpty() &&
			getInternalUserFeatureVariableReferences(functionRoot).isEmpty());
	}

	public boolean execute(IRefactoringUI ui) {
		LinkedList featureClassDeclarationList = CollectionsHelper.asLinkedList(
			getTargetFeatureClassesList((IUserFunctionDeclaration) getRefactoringSubject()).iterator());
		
		if (featureClassDeclarationList.size() == 1) return
			execute(((IUserFeatureClassDeclaration) featureClassDeclarationList.getFirst())
					.getUserFunctionDeclarationListAndFactory().getUserFunctionDeclarationList());

		// Get names of the target feature classes 
		ArrayList featureClassNames = CollectionsHelper.asArrayList(
			CollectionsHelper.map(
				featureClassDeclarationList.iterator(),
				new IMapFunction() {
					public Object map(Object obj) throws Exception {
						return ((IUserFeatureClassDeclaration) obj).getName();
					}					
				}
			));
		
		Object selection =  ui.getComboBoxSelection(
				"Choose Target", 
				"Please select an User Feature:", 
				featureClassNames, 
				null);
		
		if (selection == null)
			return false;
		
		IUserFeatureClassDeclaration selectedFeatureClass = 
			(IUserFeatureClassDeclaration) featureClassDeclarationList.get(
				featureClassNames.indexOf(selection));
		
		return execute(selectedFeatureClass.getUserFunctionDeclarationListAndFactory()
				.getUserFunctionDeclarationList());
	}
	
	public boolean execute(UserFunctionDeclarationList targetFunctionDeclarationList) {
		IUserFunctionDeclaration newDeclaration = 
				(IUserFunctionDeclaration) ((ICopyable) getRefactoringSubject()).copy(null);
		newDeclaration.setExternal(true);
		
		Morpher.replaceFunctionDeclaration(
				getUserFunctionDeclarationListFor((IUserFunctionDeclaration) getRefactoringSubject()),
				targetFunctionDeclarationList,
				(IFunctionDeclaration) getRefactoringSubject(),
				newDeclaration);

		updateUserFunctionCalls(getSeSAmDeclaration().getSeSAmDeclarationTreeModel().getRoot(),
				(IUserFunctionDeclaration) getRefactoringSubject(), newDeclaration);
		
		return true;
	}
	
	/*
	 * Returns all references to only internally visible feature varibles found within the FunctionCallInternDeclaration
	 * given as an argument. 
	 */
	protected Collection getInternalUserFeatureVariableReferences(IFunctionCallInternDeclaration declaration) {
		LinkedList items = new LinkedList();
		
		if (declaration instanceof IComposedFunctionCallInternDeclaration) {
			Iterator children = ((IComposedFunctionCallInternDeclaration) declaration).getChildren();
			while (children.hasNext())
				items.addAll(getInternalUserFeatureVariableReferences(
					(IFunctionCallInternDeclaration) children.next()));
		}
			
		if (!(declaration instanceof IAtomicFunctionCallInternDeclaration))
			return items;
		
		Object callReference = ((IAtomicFunctionCallInternDeclaration) declaration).getArgumentDeclaration()
			.getObject();
		
		if (!(callReference instanceof IFeatureVariableClassDeclaration))
			return items;
		
		if (((IFeatureVariableClassDeclaration) callReference).isExternal())
			return items;
		
		items.add(declaration);
		return items;
	}
	
	/*
	 * Returns all calls to functions marked as internal occuring within the FunctionCallInternDeclaration
	 * given as an argument. 
	 */
	protected Collection getInternalUserFeatureFunctionCalls(IFunctionCallInternDeclaration declaration) {
		LinkedList items = new LinkedList();
		
		if (!(declaration instanceof IComposedFunctionCallInternDeclaration)) 
			return items;
		
		IComposedFunctionCallInternDeclaration composedCallDeclaration = 
			(IComposedFunctionCallInternDeclaration) declaration;
		
		Iterator children = composedCallDeclaration.getChildren();
		while (children.hasNext())
			items.addAll(getInternalUserFeatureFunctionCalls((IFunctionCallInternDeclaration) children.next()));

		if (!(composedCallDeclaration.getSpecificFunctionDeclaration() instanceof IUserFunctionDeclaration))
			return items;
		
		IUserFunctionDeclaration userFunction = (IUserFunctionDeclaration) composedCallDeclaration
			.getSpecificFunctionDeclaration();
		
		if (userFunction.isExternal())
			return items;
		
		Iterator declarationLists = getUserFeatureFunctionDeclarationLists().iterator();
		while (declarationLists.hasNext())
			if (((AbstractList) declarationLists.next()).contains(userFunction))
					items.add(declaration);
		
		return items;
	}

	/*
	 * Retruns a list of all User Feature classes which don't contain the UserFunction given as
	 * an argument.
	 */
	private Collection getTargetFeatureClassesList(IUserFunctionDeclaration subject) {
		LinkedList featureClasses = new LinkedList(getSeSAmDeclaration()
			.getSeSAmModelDeclaration().getSeSAmModelBasicsDeclaration()
			.getUserFeatureClassDeclarationListAndFactory().getUserFeatureClassDeclarationList());
		
		Object featureClassOfSubject = null;
		Iterator iter = featureClasses.iterator();
		while(iter.hasNext()) {
			IUserFeatureClassDeclaration feature = (IUserFeatureClassDeclaration) iter.next();
			
			if (feature.getUserFunctionDeclarationListAndFactory().getDeclarationList().contains(subject))
				featureClassOfSubject = feature;
		}
		
		if (featureClassOfSubject != null)
			featureClasses.remove(featureClassOfSubject);
		
		return featureClasses;
	}

}
