package de.uniwue.ki.mas.refactoring;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import de.uniwue.ki.mas.basics.functionArgument.IFunctionArgumentDeclaration;
import de.uniwue.ki.mas.basics.functionCall.FunctionCallDeclarationFactory;
import de.uniwue.ki.mas.basics.functionCall.IComposedFunctionCallInternDeclaration;
import de.uniwue.ki.mas.basics.functionUser.IUserFunctionDeclaration;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.tree.AbstractObjectTreeNode;

/**
 * @author Michael Scherer
 * 
 * Removes all unused arguments from a UserFunction (referenced by RefactoringSubject)
 */
public class RemoveUnusedArguments extends AbstractRefactoring {

	protected RemoveUnusedArguments(ISeSAmDeclaration seSAmDeclaration, Object _refactoringSubject) {
		super(seSAmDeclaration, _refactoringSubject);
		this.name = "Remove Unused Arguments";
	}
	
	public boolean isApplicable() {
		if (! (this.getRefactoringSubject() instanceof IUserFunctionDeclaration))
			return false;
		
		return (getUnusedFunctionArguments().size() > 0); 
	}
	
	public boolean execute(IRefactoringUI ui) {
		return execute();
	}

	public boolean execute() {
		Collection unusedArguments = getUnusedFunctionArguments();
		
		IUserFunctionDeclaration newUserFunction = 
			(IUserFunctionDeclaration) ((ICopyable) getRefactoringSubject()).copy(null);
		
		removeArguments(newUserFunction, unusedArguments);
				
		updateFunctionCalls(newUserFunction, unusedArguments);
		
		replaceRefactoringSubject(newUserFunction);

		return true;
	}

	private Collection getUnusedFunctionArguments() {
		LinkedList unusedArgs = new LinkedList();
		
		Iterator arguments = ((IUserFunctionDeclaration) this.getRefactoringSubject())
			.getFunctionArgumentDeclarationList().iterator();
		while (arguments.hasNext()) {
			IFunctionArgumentDeclaration item = 
				(IFunctionArgumentDeclaration) arguments.next();
			if (item.getReferencedAtomicFunInternCallDeclCount() == 0)
				// Use 'addFirst' to let it's iterator go from last to first argument 
				// declaration. So deleting an function argument while going through 
				// doesn't change index of the remaining declarations.
				unusedArgs.addFirst(item);	
		}

		return unusedArgs;
	}
	
	void updateFunctionCalls(IUserFunctionDeclaration newUserFunction, 
			Collection unusedArguments) 
	{
		Iterator decls = collectFunctionCallDeclarations(
				(AbstractObjectTreeNode) getSeSAmDeclaration()
					.getSeSAmDeclarationTreeModel().getRoot()).iterator();
		
		while(decls.hasNext()) {
			IComposedFunctionCallInternDeclaration oldFunctionCall = 
				(IComposedFunctionCallInternDeclaration) decls.next();
			
			LinkedList newArgumentList = new LinkedList();

			// Copy all old arguments
			Iterator oldArguments = oldFunctionCall.getChildren();
			while (oldArguments.hasNext())
				newArgumentList.add(((ICopyable) oldArguments.next()).copy(null));
			
			// Remove unused arguments
			Iterator iter = unusedArguments.iterator();
			while (iter.hasNext()) {
				newArgumentList.remove(
					indexOfArgument((IFunctionArgumentDeclaration) iter.next())
				);
			}

			// Replace old with new function call
			IComposedFunctionCallInternDeclaration newFunctionCall =
				FunctionCallDeclarationFactory.getInstance().
				createComposedFunctionCallInternDeclaration(
					null,
					newUserFunction,
					newArgumentList
				);			
			Morpher.replaceInternCall(oldFunctionCall, newFunctionCall);
		}
	}
	
	private int indexOfArgument(IFunctionArgumentDeclaration argument) {
		return ((IUserFunctionDeclaration) getRefactoringSubject()).
			getFunctionArgumentDeclarationList().lastIndexOf(argument);
	}
	
	private void removeArguments(IUserFunctionDeclaration declaration, 
			Collection unusedArguments) 
	{
		Iterator iter = unusedArguments.iterator();

		while (iter.hasNext()) {
			int argumentIndex = indexOfArgument((IFunctionArgumentDeclaration) iter.next());
			declaration.getFunctionArgumentDeclarationList().remove(argumentIndex);
		}
	}

	private void replaceRefactoringSubject(IUserFunctionDeclaration newUserFunction) {
		Morpher.replaceFunctionDeclaration(
				getUserFunctionDeclarationListFor((IUserFunctionDeclaration) getRefactoringSubject()),
				(IUserFunctionDeclaration) getRefactoringSubject(),
				newUserFunction
			);
		}
}
