package de.uniwue.ki.mas.sim.analysis;
import java.util.*;

import de.uniwue.ki.mas.basics.functionCall.FunctionCallDeclarationFactory;
import de.uniwue.ki.mas.basics.functionCall.IFunctionCallDeclaration;
import de.uniwue.ki.mas.basics.type.BooleanType;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.sim.simulation.ISimulationCreateFunctionArgs;
import de.uniwue.ki.mas.util.declaration.AbstractDocumentableDeclaration;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (15.10.2001 16:55:56)
 * @author Christoph Oechslein
 */
final class AnalysisDeclaration
	extends AbstractDocumentableDeclaration
	implements IAnalysisDeclaration, IAnalysisDeltaSettings {

	private IAnalysisTypeDeclaration _analysisTypeDeclaration = new CSVAnalysisTypeDeclaration();
	private final AnalysisItemDeclarationList _analysisItemDeclarationList =
		new AnalysisItemDeclarationList();
	private final AnalysisItemDeclarationListAndFactory _analysisItemDeclarationListAndFactory =
		new AnalysisItemDeclarationListAndFactory(this);

	private int _deltaTicks = 1;
	private int _startTick = 0;

	private boolean _defaultAnalysis = false;

	private final IFunctionCallDeclaration _decisionFunctionCallDeclaration =
		FunctionCallDeclarationFactory.getInstance().createFunctionCallDeclaration(BooleanType.getInstance());
	private boolean _useDecisionFunctionCallDeclaration = false;

	public AnalysisDeclaration() {
		super("Untitled Analysis");
	}

	public void copyInto(Object intoObject, Map objectMapping) {
		super.copyInto(intoObject, objectMapping);

		AnalysisDeclaration intoAnalysisDeclaration = (AnalysisDeclaration) intoObject;

		intoAnalysisDeclaration.setDeltaTicks(getDeltaTicks());
		intoAnalysisDeclaration.setStartTick(getStartTick());

		intoAnalysisDeclaration.setUsingDecisionFunctionCallDeclaration(
			isUsingDecisionFunctionCallDeclaration());
		getDecisionFunctionCallDeclaration().copyInto(
			intoAnalysisDeclaration.getDecisionFunctionCallDeclaration(),
			objectMapping);

		intoAnalysisDeclaration.setAnalysisTypeDeclaration(
			(IAnalysisTypeDeclaration) getAnalysisTypeDeclaration().copy(objectMapping));

		getAnalysisItemDeclarationList().copyInto(
			intoAnalysisDeclaration.getAnalysisItemDeclarationList(),
			objectMapping);
	}

	public IAnalysis createAnalysis(ISimulationCreateFunctionArgs createFunctionArgs)
		throws ModelCompilingException {

		return getAnalysisTypeDeclaration().createAnalysis(createFunctionArgs, this);
	}

	protected ICopyable createInstance(Map objectMapping) {
		return new AnalysisDeclaration();
	}

	final AnalysisItemDeclarationList getAnalysisItemDeclarationList() {
		return _analysisItemDeclarationList;
	}

	public final AnalysisItemDeclarationListAndFactory getAnalysisItemDeclarationListAndFactory() {
		return _analysisItemDeclarationListAndFactory;
	}

	public IAnalysisTypeDeclaration getAnalysisTypeDeclaration() {
		return _analysisTypeDeclaration;
	}

	public final IFunctionCallDeclaration getDecisionFunctionCallDeclaration() {
		return _decisionFunctionCallDeclaration;
	}

	public final int getDeltaTicks() {
		return _deltaTicks;
	}

	public final int getStartTick() {
		return _startTick;
	}

	public boolean isomorphicEquals(IIsomorphicHashable anObject) {
		return equals(anObject);
	}

	public int isomorphicHashCode() {
		return hashCode();
	}

	public boolean isUsingDecisionFunctionCallDeclaration() {
		return _useDecisionFunctionCallDeclaration;
	}

	public void remove() {
		_analysisItemDeclarationList.remove();

		super.remove();

	}

	public void setAnalysisTypeDeclaration(IAnalysisTypeDeclaration newAnalysisTypeDeclaration) {
		IAnalysisTypeDeclaration oldValue = getAnalysisTypeDeclaration();

		_analysisTypeDeclaration = newAnalysisTypeDeclaration;

		IAnalysisItemDeclarationFactory analysisItemDeclarationFactory =
			getAnalysisTypeDeclaration().getAnalysisItemDeclarationFactory();

		List newItems = new ArrayList(getAnalysisItemDeclarationList().size());
		Iterator it = getAnalysisItemDeclarationList().iterator();
		
		HashMap objectMapping =  new HashMap();

		while (it.hasNext()) {
			IAnalysisItemDeclaration oldAnalysisItemDeclaration = (IAnalysisItemDeclaration) it.next();
			IAnalysisItemDeclaration newAnalysisItemDeclaration =
				analysisItemDeclarationFactory.createAnalysisItemDeclaration();

			oldAnalysisItemDeclaration.copyInto(newAnalysisItemDeclaration,objectMapping);

			newItems.add(newAnalysisItemDeclaration);
		}

		getAnalysisItemDeclarationList().clear();

		getAnalysisItemDeclarationList().addAll(newItems);

		firePropertyChange("analysisTypeDeclaration", oldValue, newAnalysisTypeDeclaration);
	}

	public void setDeltaTicks(int newDeltaTicks) {
		newDeltaTicks = Math.max(1, newDeltaTicks);

		int oldValue = _deltaTicks;

		_deltaTicks = newDeltaTicks;

		firePropertyChange("deltaTicks", oldValue, newDeltaTicks);
	}

	public void setStartTick(int newStartTick) {
		_startTick = newStartTick;
	}

	public void setUsingDecisionFunctionCallDeclaration(boolean newUseDecisionFunctionCallDeclaration) {
		_useDecisionFunctionCallDeclaration = newUseDecisionFunctionCallDeclaration;
	}

	public boolean isDefaultAnalysis() {
		return _defaultAnalysis;
	}

	public void setDefaultAnalysis(boolean defaultAnalysis) {
		boolean oldValue = isDefaultAnalysis();

		_defaultAnalysis = defaultAnalysis;

		firePropertyChange("defaultAnalysis", oldValue, defaultAnalysis);
		fireStateChanged();
	}

	public IAnalysisDeltaSettings getAnalysisDeltaSettings() {
		return this;
	}
}