package de.uniwue.ki.mas.sim.analysis;
import java.io.*;
import java.util.List;

import ViolinStrings.Strings;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.basics.type.ISpecificType;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.mas.sim.simulation.*;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.file.FileUtilities;
import de.uniwue.ki.util.file.TokenReplacer;
/**
 * Creation date: (15.10.2001 17:08:42)
 * @author Christoph Oechslein
 */
public final class CSVAnalysis extends AbstractValueAnalysis {

	private final String _timeColumnName;
	private final String _separatorString;

	public static final TokenReplacer TOKEN_REPLACER =
		TokenReplacer.createTokenReplacer(createTokenReplaceConfigList(null));

	private final File _file;
	private File _replacedFile;
	protected transient Writer _fileWriter;

	private final ISimulation _simulation;

	private final EventListener _eventListener = new EventListener();

	private final class EventListener implements SimulationListener, StartPauseFinishListener, Serializable {
		public void reseted(SimulationEvent event) {
			try {
				addToAllFinishedList();
				resetFileWriter();

			} catch (IOException exception) {
				exception.printStackTrace(System.err);
			}
		}

		public void finished(StartPauseFinishEvent event) {
			try {
				((ISimulation) event.getSource()).removeStartPauseFinishListener(this);
				addToAllFinishedList();
				resetFileWriter();

			} catch (IOException exception) {
				exception.printStackTrace(System.err);
			}
		}

		public void paused(StartPauseFinishEvent event) {
			if (_fileWriter != null) {
				try {
					_fileWriter.flush();

				} catch (IOException exception) {
					exception.printStackTrace(System.err);
				}
			}
		}

		public void started(StartPauseFinishEvent event) {
			addToAllStartedList();
		}
	}

	public CSVAnalysis(
		IAnalysisDeclaration analysisDeclaration,
		ISimulationCreateFunctionArgs createFunctionArgs,
		String[] nameList,
		IFunction[] valueFunctionList,
		ISpecificType[] analyseValueTypeArray)
		throws ModelCompilingException {

		super(analysisDeclaration, createFunctionArgs, nameList, valueFunctionList, analyseValueTypeArray);

		CSVAnalysisTypeDeclaration csvAnalysisTypeDeclaration =
			(CSVAnalysisTypeDeclaration) analysisDeclaration.getAnalysisTypeDeclaration();

		_file = new File(csvAnalysisTypeDeclaration.getFileString());
		_separatorString = csvAnalysisTypeDeclaration.getSeparatorString().replaceAll("\\\\t", "\t");
		_timeColumnName = csvAnalysisTypeDeclaration.getTimeColumnName();

		_simulation = createFunctionArgs.getSimulation();

		_simulation.addStartPauseFinishListener(_eventListener);
		_simulation.addSimulationListener(_eventListener);
	}

	public void analyse(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {
		if (_fileWriter == null) {
			initFileWriter(simulationExecuteFunctionArgs);
		}

		super.analyse(simulationExecuteFunctionArgs);

	}

	protected void analyseIntern(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {
		super.analyseIntern(simulationExecuteFunctionArgs);

		int size = getCurrentValuesCount();

		try {
			_fileWriter.write('\n');
			_fileWriter.write(
				Integer.toString(
					simulationExecuteFunctionArgs.getSimulation().getSituation().getTimeDurationInTicks()));

			for (int i = 0; i < size; i++) {
				_fileWriter.write(_separatorString);
				_fileWriter.write(replaceSeperatorString(getNameForEntry(i)));
			}

		} catch (IOException exception) {
			throw new SimulationRuntimeException(exception);
		}
	}

	private String replaceSeperatorString(String line) {
		return replaceSeperatorStringWith(line, new String[] { "#", "_", "?", ",", ";" });
	}

	private String replaceSeperatorStringWith(String line, String[] sepChars) {
		for (int i = 0; i < sepChars.length; i++) {
			if (!Strings.contains(_separatorString, sepChars[i])) {
				return Strings.change(line, _separatorString, sepChars[i]);
			}
		}

		return line;
	}

	private Writer createFileWriter(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException, IOException {

		return new BufferedWriter(
			new FileWriter(getReplacedFile(simulationExecuteFunctionArgs.getSimulation())));
	}

	private static List createTokenReplaceConfigList(final ISimulation sim) {
		return CollectionsHelper.createList(new TokenReplacer.TokenReplaceConfig('E', "Simulation") {
			public String getReplaceString() {
				if (sim == null)
					return "Untitled Simulation";
				else
					return sim.getName();
			}
		}, new TokenReplacer.TokenReplaceConfig('e', "Situation") {
			public String getReplaceString() {
				if (sim == null)
					return "Untitled Situation";
				else
					return sim.getSimulationDeclaration().getSituationDeclaration().getName();
			}
		});
	}

	protected synchronized void resetFileWriter() throws IOException {
		if (_fileWriter != null) {
			_fileWriter.flush();
			_fileWriter.close();
			_fileWriter = null;
		}

		_replacedFile = null;
	}

	public void remove() {
		try {
			resetFileWriter();

		} catch (IOException e) {
			e.printStackTrace(System.err);
		}

		_simulation.removeStartPauseFinishListener(_eventListener);
		_simulation.removeSimulationListener(_eventListener);

		super.remove();
	}

	private File getAbsoluteFileWithReplacedTokens(ISimulation sim) {
		return new File(
			TokenReplacer.createTokenReplacer(createTokenReplaceConfigList(sim)).replaceTokens(
				_file.toString(),
				sim.getStartDate()));
	}

	private File getReplacedFile(ISimulation sim) throws IOException {
		if (_replacedFile == null) {
			File absoluteFileWithReplacedTokens = getAbsoluteFileWithReplacedTokens(sim);

			File parentFile = absoluteFileWithReplacedTokens.getParentFile();
			if (parentFile == null) {
				_replacedFile = absoluteFileWithReplacedTokens;

			} else {
				CSVAnalysisSaveAllFinished.mkdirs(parentFile);

				_replacedFile =
					FileUtilities.getWriteableNonExistingFile(
						absoluteFileWithReplacedTokens,
						"protokoll",
						".csv");
			}
		}

		return _replacedFile;
	}

	private synchronized void initFileWriter(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {
		try {
			resetFileWriter();

			_fileWriter = createFileWriter(simulationExecuteFunctionArgs);

			_fileWriter.write(replaceSeperatorString(_timeColumnName));

			int size = getNameArray().length;

			for (int i = 0; i < size; i++) {
				String name = getNameArray()[i];

				_fileWriter.write(_separatorString);
				_fileWriter.write(replaceSeperatorString(name));
			}

		} catch (IOException exception) {
			throw new SimulationRuntimeException(exception);
		}
	}

	protected void finalize() throws Throwable {
		resetFileWriter();

		super.finalize();
	}

	protected void addToAllFinishedList() {
		if (_replacedFile != null) {
			CSVAnalysisSaveAllFinished.addToAllFinishedList(_replacedFile);
		}
	}

	protected void addToAllStartedList() {
		if (_replacedFile != null) {
			CSVAnalysisSaveAllFinished.addToAllStartedList(_replacedFile);
		}
	}
}