package de.uniwue.ki.mas.sim.analysis;
import java.util.*;

import org.jfree.data.*;

import de.uniwue.ki.mas.basics.function.ICreateFunctionArgs;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.mas.sim.simulation.ISimulationExecuteFunctionArgs;

/**
 * Creation date: (15.10.2001 17:08:42)
 * @author Christoph Oechslein
 */
public class SeriesChartAnalysis extends AbstractAnalysis implements XYDataset {
	protected List _listeners;

	private DatasetGroup _group;
	private final int _windowSize;
	private final LinkedList _dataList = new LinkedList();
	private final LinkedList _timeList;
	private int _currentTimeTick = 0;
	private String[] _legendItemLabels;

	public SeriesChartAnalysis(
		IAnalysisDeclaration analysisDeclaration,
		ICreateFunctionArgs createFunctionArgs,
		int windowSize,
		String[] nameList,
		IFunction[] valueFunctionList)
		throws ModelCompilingException {

		super(analysisDeclaration, createFunctionArgs, nameList, valueFunctionList);

		_windowSize = windowSize;

		if (isUsingDecisionFunction())
			_timeList = new LinkedList();
		else
			_timeList = null;
	}

	public void addChangeListener(DatasetChangeListener listener) {
		if (_listeners == null) {
			_listeners = new LinkedList();
		}

		_listeners.add(listener);
	}

	protected void analyseIntern(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {

		int currentTimeTick =
			simulationExecuteFunctionArgs.getSimulation().getSituation().getTimeDurationInTicks();

		_currentTimeTick = currentTimeTick;

		if (isUsingDecisionFunction())
			_timeList.addFirst(new Integer(currentTimeTick));

		double[] newArray;

		if ((_windowSize <= 0) || (_dataList.size() < _windowSize)) {
			newArray = new double[getValueFunctionArray().length];

		} else {
			newArray = (double[]) _dataList.removeLast();
			if (isUsingDecisionFunction())
				_timeList.removeLast();
		}

		for (int i = 0; i < getValueFunctionArray().length; i++) {
			IFunction function = getValueFunctionArray()[i];

			double result = function.executeForDouble(simulationExecuteFunctionArgs);

			newArray[i] = result;
		}

		_dataList.addFirst(newArray);

		fireDatasetChanged();
	}

	protected void fireDatasetChanged() {
		notifyListeners(new DatasetChangeEvent(this, this));
	}

	/**
	 * @param series int
	 * @return int
	 * @see org.jfree.data.XYDataset#getItemCount(int)
	 */
	public int getItemCount(int series) {
		return _dataList.size();
	}

	public int getLegendItemCount() {
		return getSeriesCount();
	}

	/**
	 * @return String[]
	 */
	public String[] getLegendItemLabels() {
		if (_legendItemLabels == null) {
			int seriesCount = getSeriesCount();
			String[] labels = new String[seriesCount];
			for (int i = 0; i < seriesCount; i++) {
				labels[i] = getSeriesName(i);
			}
			_legendItemLabels = labels;
		}

		return _legendItemLabels;
	}

	public int getSeriesCount() {
		return getNameArray().length;
	}

	public String getSeriesName(int series) {
		return getNameArray()[series];
	}

	public Number getXValue(int series, int item) {
		if (isUsingDecisionFunction())
			return (Integer) _timeList.get(item);
		else
			return new Integer(_currentTimeTick - item * getDeltaTicks());
	}

	public Number getYValue(int series, int item) {
		double value = ((double[]) _dataList.get(item))[series];

		if (Double.isInfinite(value) || Double.isNaN(value))
			return new Double(0.0);
		else
			return new Double(value);
	}

	protected void notifyListeners(DatasetChangeEvent event) {
		if (_listeners != null) {
			Iterator iterator = _listeners.iterator();
			while (iterator.hasNext()) {
				DatasetChangeListener listener = (DatasetChangeListener) iterator.next();
				listener.datasetChanged(event);
			}
		}
	}

	/**
	 * @param listener DatasetChangeListener
	 * @see org.jfree.data.Dataset#removeChangeListener(DatasetChangeListener)
	 */
	public void removeChangeListener(DatasetChangeListener listener) {
		_listeners.remove(listener);
	}

	/**
	 * @param simulationExecuteFunctionArgs ISimulationExecuteFunctionArgs
	 * @throws SimulationRuntimeException
	 * @see de.uniwue.ki.mas.sim.analysis.IAnalysis#init(ISimulationExecuteFunctionArgs)
	 */
	public void init(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {
	}

	public void setGroup(DatasetGroup group) {
		_group = group;
	}

	public DatasetGroup getGroup() {
		return _group;
	}
}