package de.uniwue.ki.mas.sim.experiment;
import java.io.File;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.model.*;
import de.uniwue.ki.mas.sim.simulation.*;
/**
 * Creation date: (29.01.2002 15:45:22)
 * @author Christoph Oechslein
 */
final class ThreadedExperiment extends AbstractExperiment implements IExperiment {
	private static final int EXPERIMENT_INITIALIZED = 0;
	private static final int EXPERIMENT_RUNNING = 1;
	private static final int EXPERIMENT_PAUSED = 2;
	private static final int EXPERIMENT_FINISHED = 3;
	private static final int EXPERIMENT_FINISHED_WITH_ERRORS = 4;

	private List _simulationsToRemove = new LinkedList();

	private int _experimentStatus = EXPERIMENT_INITIALIZED;

	private final IExperimentDeclaration _experimentDeclaration;
	private final IFunction _experimentFunction;
	private final ExperimentFunctionArgs _experimentFunctionArgs;
	private transient Throwable _exception;

	private transient final StartPauseFinishListener _startPauseFinishListener =
		new StartPauseFinishListener() {
		public void paused(StartPauseFinishEvent event) {
			if (!isPaused())
				firePaused();
		}

		public void started(StartPauseFinishEvent event) {
			if (!isRunning())
				fireStarted();
		}

		public void finished(StartPauseFinishEvent event) {
			// if sim is finished due to an error 
			// then in startSim an Exception was thrown 
			// which is caught in createThread which calls fireFinished!
		}
	};

	public ThreadedExperiment(
		ISeSAmDeclaration seSAmDeclaration,
		IExperimentDeclaration experimentDeclaration)
		throws ModelCompilingException {
		super(experimentDeclaration.getName(), new File(experimentDeclaration.getConfigurationFileName()));

		_experimentFunctionArgs = new ExperimentFunctionArgs(seSAmDeclaration, this);
		_experimentDeclaration = experimentDeclaration;
		_experimentFunction =
			_experimentDeclaration.getSimulationActionFunctionCallDeclaration().createFunction(
				getExperimentFunctionArgs());
	}

	private void addListener(ISimulation simulation) {
		simulation.addStartPauseFinishListener(_startPauseFinishListener);
	}

	public ISimulation createSimulation(ISimulationDeclaration simulationDeclaration)
		throws ModelCompilingException {

		try {
			// need its own ExperimentFunctionArgs !!
			ExperimentFunctionArgs ownExperimentFunctionArgs =
				new ExperimentFunctionArgs(getExperimentFunctionArgs());

			IThreadedSimulation sim =
				SimulationFactory.createThreadedSimulation(simulationDeclaration, ownExperimentFunctionArgs);

			ownExperimentFunctionArgs.setSimulation(sim);

			Assert.assertFalse(_simulationsToRemove.contains(sim));

			_simulationsToRemove.add(sim);

			return sim;

		} catch (SimulationRuntimeException e) {
			throw new ModelCompilingException(e);
		}
	}

	private void createThreadAndStartExperiment() {

		Thread thread = new Thread(new Runnable() {
			public void run() {

				try {
					actExperimentFunction();

					fireFinished();

				} catch (Throwable exception) {
					fireFinished(exception);

				} finally {
					synchronized (ThreadedExperiment.this) {
						ThreadedExperiment.this.notifyAll();
					}
				}
			}
		}, getName());

		fireStarted();

		thread.start();
	}

	protected final void actExperimentFunction() throws SimulationRuntimeException {
		_experimentFunction.act(getExperimentFunctionArgs());
	}

	public void fireCurrentSimulationChanged(ISimulation oldSimulation, ISimulation newSimulation) {
		if (oldSimulation != null) {
			removeListener(oldSimulation);
		}

		if (newSimulation != null) {
			addListener(newSimulation);
		}

		super.fireCurrentSimulationChanged(oldSimulation, newSimulation);
	}

	protected void fireFinished() {
		setException(null);

		setExperimentStatus(EXPERIMENT_FINISHED);

		super.fireFinished();
	}

	protected void fireFinished(Throwable exception) {
		setException(exception);

		setExperimentStatus(EXPERIMENT_FINISHED_WITH_ERRORS);

		super.fireFinished(exception);
	}

	protected void firePaused() {
		setExperimentStatus(EXPERIMENT_PAUSED);

		super.firePaused();
	}

	protected void fireStarted() {
		setExperimentStatus(EXPERIMENT_RUNNING);

		super.fireStarted();
	}

	public synchronized ISimulation getCurrentSimulation() {
		return getExperimentFunctionArgs().getSimulation();
	}

	private synchronized IThreadedSimulation getCurrentThreadedSimulation() {
		return (IThreadedSimulation)getCurrentSimulation();
	}

	public final IExperimentDeclaration getExperimentDeclaration() {
		return _experimentDeclaration;
	}

	protected final ExperimentFunctionArgs getExperimentFunctionArgs() {
		return _experimentFunctionArgs;
	}

	public synchronized boolean isFinished() {
		return _experimentStatus == EXPERIMENT_FINISHED;
	}

	public synchronized boolean isFinishedWithErrors() {
		return _experimentStatus == EXPERIMENT_FINISHED_WITH_ERRORS;
	}

	public synchronized boolean isPaused() {
		return _experimentStatus == EXPERIMENT_PAUSED;
	}

	public synchronized boolean isRunning() {
		return _experimentStatus == EXPERIMENT_RUNNING;
	}

	public synchronized void pauseInThread() {
		if (!isRunning()) {
			throw new IllegalStateException("Pause only possible in running state");
		}

		firePaused();

		IThreadedSimulation simulation = getCurrentThreadedSimulation();

		if (simulation != null) {
			simulation.pauseInThread();
		}
	}

	public synchronized void terminateInThread() {
		fireFinished();

		IThreadedSimulation simulation = getCurrentThreadedSimulation();

		if (simulation != null) {
			simulation.terminateInThread();
		}
	}

	public void remove() {
		removeCurrentSimulation();

		if (_simulationsToRemove != null) {
			Iterator it = _simulationsToRemove.iterator();

			while (it.hasNext()) {
				ISimulation element = (ISimulation)it.next();

				if (element instanceof IThreadedSimulation)
					 ((IThreadedSimulation)element).pauseInThread();
				element.remove();
			}

			_simulationsToRemove = null;
		}

		if (getExperimentFunctionArgs() != null)
			getExperimentFunctionArgs().remove();

		super.remove();
	}
	
	public void removeSimulation(ISimulation sim) {
		if (sim!=null) {
		if (sim instanceof IThreadedSimulation)
			((IThreadedSimulation)sim).pauseInThread();
		sim.remove();
		if (_simulationsToRemove.contains(sim)) _simulationsToRemove.remove(sim);
		}
	}

	private void removeCurrentSimulation() {
		terminateInThread();
		removeSimulation(getCurrentSimulation());
	}

	private void removeListener(ISimulation simulation) {
		simulation.removeStartPauseFinishListener(_startPauseFinishListener);
	}

	private synchronized void setExperimentStatus(int newExperimentStatus) {
		_experimentStatus = newExperimentStatus;
	}

	public synchronized void startInThread() {
		if (isRunning()) {
			throw new IllegalStateException("Already in running state");
		}

		if (isFinished() || isFinishedWithErrors()) {
			throw new IllegalStateException("Already finished");
		}

		// is done in the thread
		// setExperimentStatus(EXPERIMENT_RUNNING);
		// fireStarted();

		if (isPaused()) {
			unPauseCurrentSimulation();

		} else {
			createThreadAndStartExperiment();
		}
	}

	public void startSimulation(ISimulation simulation) throws SimulationRuntimeException {
		getExperimentFunctionArgs().setSimulation(simulation);

		IThreadedSimulation threadedSimulation = (IThreadedSimulation)simulation;

		threadedSimulation.startInThread();

		// wait for simulation to finish
		synchronized (this) {
			while (!threadedSimulation.isFinished()
				&& !threadedSimulation.isFinishedWithErrors()
				&& !isFinished()
				&& !isFinishedWithErrors()) {

				if (isPaused() && !threadedSimulation.isPaused()) {
					threadedSimulation.pauseInThread();
				}

				try {
					wait(100);
				} catch (InterruptedException exception) {
				}
			}

			// now wait for thread finishing		
			Thread simulationThread = threadedSimulation.getSimulationThread();

			if (simulationThread != null) {
				while (simulationThread.isAlive()) {
					try {
						wait(100);
					} catch (InterruptedException exception) {
					}
				}
			}

			if (threadedSimulation.isFinishedWithErrors()) {
				throw new ExperimentRuntimeException(threadedSimulation.getException());
			}
		}	
	}

	private void unPauseCurrentSimulation() {
		fireStarted();

		IThreadedSimulation simulation = getCurrentThreadedSimulation();

		if (simulation != null)
			simulation.startInThread();
	}

	public Throwable getException() {
		return _exception;
	}

	private void setException(Throwable exception) {
		_exception = exception;
	}
}