package de.uniwue.ki.mas.sim.logging;

import java.util.*;

import de.uniwue.ki.mas.behaviour.activityGraph.IActivityGraphLeaf;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.IActivityGraphNodeDeclaration;
import de.uniwue.ki.mas.entity.object.agent.IActivityAgent;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclaration;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclarationList;
import de.uniwue.ki.mas.entity.objectclass.agentclass.IActivityAgentClassDeclaration;
import de.uniwue.ki.mas.entity.objectclass.worldclass.WorldClassDeclarationList;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.sim.analysis.IAnalysis;
import de.uniwue.ki.mas.sim.experiment.IExperiment;
import de.uniwue.ki.mas.sim.listener.*;
import de.uniwue.ki.mas.sim.simulation.AnalysisEvent;
import de.uniwue.ki.mas.sim.simulation.ISimulation;
import de.uniwue.ki.mas.sim.situation.SituationEvent;
import de.uniwue.ki.mas.util.declaration.TwoUnmodifiableDeclarationList;
import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.math.MathHelper;

/**
 * Creation date: 20.09.2002, 11:26:44
 * @author oechslein
 */
public class SimulationStatisticLogger extends AbstractLogger {
	private long _sumOfObjects = 0;
	private long _maxObjectCount = 0;
	private ISeSAmDeclaration _seSAmDeclaration;

	private static class TwoUnmodifiableObjectClassDeclarationList
		extends TwoUnmodifiableDeclarationList
		implements IObjectClassDeclarationList {

		TwoUnmodifiableObjectClassDeclarationList(
			IObjectClassDeclarationList agents,
			WorldClassDeclarationList world) {
			super(world, agents, false);
		}
	}

	public SimulationStatisticLogger() {
	}

	public IObjectClassDeclarationList getObjectClassDeclarationList() {
		if (getSeSAmDeclaration() == null)
			return null;
		else
			return new TwoUnmodifiableObjectClassDeclarationList(
				getSeSAmDeclaration().getSeSAmModelDeclaration().getAvailibleObjectClassDeclarationList(),
				getSeSAmDeclaration()
					.getSeSAmModelDeclaration()
					.getWorldClassDeclarationListAndFactory()
					.getWorldClassDeclarationList());
	}

	public List getAnalysisList() {
		if (getSimulation() == null)
			return Collections.EMPTY_LIST;
		else
			return getSimulation().getAnalysisList();
	}

	private Map _agentClassInfoMap = new HashMap();
	private Map _analysisInfoMap = new HashMap();

	public static class TimeInfo {
		private long _time = 0;
		private int _amount = 0;

		public double getMeanTime() {
			return (double) getTime() / (double) _amount;
		}

		public long getTime() {
			return _time;
		}

		public void updateTime(long time) {
			_time += time;
			_amount++;
		}
	}

	public final static class ActivityInfo extends TimeInfo {
		private final IActivityGraphLeaf _activity;

		public ActivityInfo(IActivityGraphLeaf activity) {
			_activity = activity;
		}

		protected IActivityGraphLeaf getActivity() {
			return _activity;
		}
	}

	public final static class AnalysisInfo extends TimeInfo {
		//private final IAnalysis _analysis;

		public AnalysisInfo(IAnalysis analysis) {
			//_analysis = analysis;
		}
	}

	public final static class AgentClassInfo {

		private final IObjectClassDeclaration _objectClassDeclaration;

		private long _updateTime = 0;
		private long _reasoningTime = 0;
		private long _actionTime = 0;
		private final Map _activityMap = new HashMap();

		private int _updateAmount = 0;
		private int _reasoningAmount = 0;
		private int _actionAmount = 0;

		public AgentClassInfo(IObjectClassDeclaration objectClassDeclaration) {
			super();

			_objectClassDeclaration = objectClassDeclaration;

		}

		protected void updateActivityInfo(IActivityGraphLeaf atomicActivity, long consumedTime) {
			ActivityInfo activityInfo;
			if (!_activityMap.containsKey(atomicActivity)) {
				activityInfo = new ActivityInfo(atomicActivity);
				_activityMap.put(atomicActivity, activityInfo);
			} else {
				activityInfo = (ActivityInfo) _activityMap.get(atomicActivity);
			}

			activityInfo.updateTime(consumedTime);
		}

		protected void updateUpdateTime(long consumedTime) {
			_updateTime += consumedTime;
			_updateAmount++;
		}

		protected void updateActionTime(long consumedTime) {
			_actionTime += consumedTime;
			_actionAmount++;
		}

		protected void updateReasoningTime(long consumedTime) {
			_reasoningTime += consumedTime;
			_reasoningAmount++;
		}

		public Iterator getActivityInfos() {
			return _activityMap.values().iterator();
		}

		public Iterator getSortedActivityInfos() {
			return CollectionsHelper
				.asSortedArrayList(
					getActivityInfos(),
					MathHelper.getDoubleComparator(new MathHelper.IMapDouble() {
				public double mapDouble(Object o) {
					return ((ActivityInfo) o).getTime();
				}
			})).iterator();
		}

		public long getTotalTime() {
			return getActionTime() + getReasoningTime() + getUpdateTime();
		}

		public int getTotalAmount() {
			return getActionAmount() + getReasoningAmount() + getUpdateAmount();
		}

		public long getTotalMeanTime() {
			if (getTotalAmount() == 0)
				return 0;
			else
				return getTotalTime() / getTotalAmount();
		}

		public long getReasoningTime() {
			return _reasoningTime;
		}

		public long getActionTime() {
			return _actionTime;
		}

		public int getReasoningAmount() {
			return _reasoningAmount;
		}

		public int getActionAmount() {
			return _actionAmount;
		}

		public long getUpdateTime() {
			return _updateTime;
		}

		public int getUpdateAmount() {
			return _updateAmount;
		}

		public IObjectClassDeclaration getObjectClassDeclaration() {
			return _objectClassDeclaration;
		}
	}

	public double getObjectsPerTick() {
		if (getSituation() == null)
			return 0;
		else
			return (double) _sumOfObjects / (double) getSituation().getTimeDurationInTicks();
	}

	public double getMillisPerTick() {
		if (getSituation() == null)
			return 0;
		else
			return (double) getTotalTime() / (double) getSituation().getTimeDurationInTicks();
	}

	public double getMillisPerTickAndPerObject() {
		if (_sumOfObjects == 0)
			return 0;
		else
			return getMillisPerTick() / _sumOfObjects;
	}

	public double getMillisPerObject() {
		if (_sumOfObjects == 0)
			return 0;
		else
			return (double) getTotalTime() / _sumOfObjects;
	}

	public long getTotalTime() {
		synchronized (this) {
			if (_stopTime != -1) {
				return _stopTime - _startTime - _sumOfStopTime;

			} else {
				return System.currentTimeMillis() - _startTime - _sumOfStopTime;
			}
		}
	}

	public void nextTimeStep(SituationEvent event) {
		super.nextTimeStep(event);

		int objectCount = event.getSituation().getWorld().getObjectCount(false, false);

		if (objectCount > _maxObjectCount)
			_maxObjectCount = objectCount;

		_sumOfObjects += objectCount;
	}

	public long getMaxObjectCount() {
		return _maxObjectCount;
	}

	private long _stopTime = -1;
	private long _startTime = -1;
	private long _sumOfStopTime;

	protected void startedExperiment(IExperiment exp) {
		super.startedExperiment(exp);

		synchronized (this) {
			_stopTime = -1;
			_startTime = -1;
			_sumOfStopTime = 0;
			_sumOfObjects = 0;
			_maxObjectCount = 0;
			_agentClassInfoMap.clear();
			_analysisInfoMap.clear();
		}
	}

	protected void startedSimulation(ISimulation simulation) {
		synchronized (this) {
			if (_stopTime != -1) {
				_sumOfStopTime += (System.currentTimeMillis() - _stopTime);
				_stopTime = -1;

			} else {
				_startTime = simulation.getStartDate().getTime();
				_stopTime = -1;
				_sumOfStopTime = 0;
				_sumOfObjects = 0;
				_maxObjectCount = 0;
				_agentClassInfoMap.clear();
				_analysisInfoMap.clear();
			}
		}

		super.startedSimulation(simulation);
	}

	protected void pausedSimulation(ISimulation sim) {
		super.pausedSimulation(sim);

		synchronized (this) {
			_stopTime = System.currentTimeMillis();
		}
	}

	protected void finishedSimulation(ISimulation sim) {
		super.finishedSimulation(sim);

		synchronized (this) {
			_stopTime = System.currentTimeMillis();
		}
	}

	public AgentClassInfo getAgentClassInfo(IObjectClassDeclaration objectClassDeclaration) {
		if (_agentClassInfoMap.containsKey(objectClassDeclaration)) {
			return (AgentClassInfo) _agentClassInfoMap.get(objectClassDeclaration);

		} else {
			AgentClassInfo newAgentClassInfo = new AgentClassInfo(objectClassDeclaration);
			_agentClassInfoMap.put(objectClassDeclaration, newAgentClassInfo);
			return newAgentClassInfo;
		}
	}

	public AnalysisInfo getAnalysisInfo(IAnalysis analysis) {
		if (_analysisInfoMap.containsKey(analysis)) {
			return (AnalysisInfo) _analysisInfoMap.get(analysis);

		} else {
			AnalysisInfo newAnalysisInfo = new AnalysisInfo(analysis);
			_analysisInfoMap.put(analysis, newAnalysisInfo);
			return newAnalysisInfo;
		}
	}

	public ActivityInfo getActivityInfo(
		IActivityAgentClassDeclaration activityAgentClassDeclaration,
		IActivityGraphNodeDeclaration activityDeclaration) {
		AgentClassInfo agentClassInfo = getAgentClassInfo(activityAgentClassDeclaration);

		Iterator it = agentClassInfo.getActivityInfos();
		while (it.hasNext()) {
			ActivityInfo activityInfo = (ActivityInfo) it.next();
			if (activityInfo.getActivity().getActivityGraphNodeDeclaration() == activityDeclaration)
				return activityInfo;
		}

		return null;
	}

	private final AgentClassInfo getAgentClassInfo(AgentEvent event) {
		return getAgentClassInfo(event.getAgent().getObjectInstanceDeclaration().getObjectClassDeclaration());
	}

	private long _updateStartTime = 0;

	public void updateStarts(AgentEvent event) {
		super.updateStarts(event);

		_updateStartTime = System.currentTimeMillis();
	}

	public void updateFinished(AgentEvent event) {
		super.updateFinished(event);

		getAgentClassInfo(event).updateUpdateTime(System.currentTimeMillis() - _updateStartTime);
		_updateStartTime = 0;
	}

	public void actingStarts(ActivityAgentEvent event) {
		super.actingStarts(event);

		_updateStartTime = System.currentTimeMillis();
	}

	public void actingFinished(ActivityAgentEvent event) {
		super.actingFinished(event);

		AgentClassInfo agentClassInfo = getAgentClassInfo(event);

		long consumedTime = System.currentTimeMillis() - _updateStartTime;

		getAgentClassInfo(event).updateActionTime(consumedTime);
		_updateStartTime = 0;

		IActivityAgent activityAgent = event.getActivityAgent();
		agentClassInfo.updateActivityInfo(
			activityAgent.getCurrentActivityGraphLeafs()[event.getCurrActivityReasoningEngineIndex()],
			consumedTime);
	}

	public void reasoningStarts(ActivityAgentEvent event) {
		super.reasoningStarts(event);

		_updateStartTime = System.currentTimeMillis();
	}

	public void reasoningFinished(ActivityAgentEvent event) {
		super.reasoningFinished(event);

		getAgentClassInfo(event).updateReasoningTime(System.currentTimeMillis() - _updateStartTime);
		_updateStartTime = 0;
	}

	public void analysisStarted(AnalysisEvent analysisEvent) {
		super.analysisStarted(analysisEvent);

		_updateStartTime = System.currentTimeMillis();
	}

	public void analysisFinished(AnalysisEvent analysisEvent) {
		super.analysisFinished(analysisEvent);

		getAnalysisInfo(analysisEvent.getAnalysis()).updateTime(
			System.currentTimeMillis() - _updateStartTime);
		_updateStartTime = 0;
	}

	public ArrayList getSortedAnalysisInfo(Comparator comparator) {
		return CollectionsHelper.asSortedArrayList(_analysisInfoMap.values().iterator(), comparator);
	}

	public ArrayList getSortedAgentClassInfo(Comparator comparator) {
		return CollectionsHelper.asSortedArrayList(_agentClassInfoMap.values().iterator(), comparator);
	}

	private interface IntMapFunction {
		double mapDouble(AgentClassInfo value);
	}

	private ArrayList getSortedAgentClassInfo(final IntMapFunction intMapFunction) {
		return getSortedAgentClassInfo(new Comparator() {
			public int compare(Object o1, Object o2) {
				return MathHelper.compare(
					intMapFunction.mapDouble((AgentClassInfo) o1),
					intMapFunction.mapDouble((AgentClassInfo) o2));
			}
		});
	}

	public ArrayList getSortedByTotalTimeAgentClassInfo() {
		return getSortedAgentClassInfo(new IntMapFunction() {
			public double mapDouble(AgentClassInfo agentClassInfo) {
				return agentClassInfo.getTotalTime();
			}
		});
	}

	public ArrayList getSortedByTotalMeanTimeAgentClassInfo() {
		return getSortedAgentClassInfo(new IntMapFunction() {
			public double mapDouble(AgentClassInfo agentClassInfo) {
				return agentClassInfo.getTotalMeanTime();
			}
		});
	}

	public ArrayList getSortedByTotalAmountAgentClassInfo() {
		return getSortedAgentClassInfo(new IntMapFunction() {
			public double mapDouble(AgentClassInfo agentClassInfo) {
				return agentClassInfo.getTotalAmount();
			}
		});
	}

	public void currentActivityGraphLeafChanged(CurrentActivityGraphLeafChangedEvent event) {
		super.currentActivityGraphLeafChanged(event);

	}

	public void setSeSAmDeclaration(ISeSAmDeclaration seSAmDeclaration) {
		_seSAmDeclaration = seSAmDeclaration;
	}

	public ISeSAmDeclaration getSeSAmDeclaration() {
		return _seSAmDeclaration;
	}
}