package de.uniwue.ki.mas.sim.logging;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Iterator;

import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.*;
import de.uniwue.ki.mas.behaviour.reasoning.IActivityReasoningEngineDeclaration;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclaration;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclarationList;
import de.uniwue.ki.mas.entity.objectclass.agentclass.IActivityAgentClassDeclaration;
import de.uniwue.ki.mas.sim.analysis.IAnalysis;
import de.uniwue.ki.util.math.MathHelper;
import de.uniwue.ki.util.name.INamed;
import de.uniwue.ki.util.tree.AbstractTreeTableModel;
import de.uniwue.ki.util.tree.TreeTableModel;

/**
 * Creation date: 25.09.2002, 10:24:42
 * @author oechslein
 */
public final class StatisticTreeTableModel extends AbstractTreeTableModel {
	protected static final NumberFormat _percentFormat = NumberFormat.getPercentInstance();
	protected static final DecimalFormat _decimalFormat = new DecimalFormat("0.###E0");

	static {
		_percentFormat.setMaximumFractionDigits(2);
	}

	private static final String[] COLUMN_NAMES =
		{ "Name", "UpdateTime", "Reasoning Time", "Action Time", "Total Time" };
	private static final Class[] COLUMN_CLASSES =
		{ TreeTableModel.class, String.class, String.class, String.class, String.class };

	private final String _analysis_Root = "Analysis";

	public StatisticTreeTableModel(SimulationStatisticLogger simulationStatisticLogger) {
		super(simulationStatisticLogger);
	}

	protected SimulationStatisticLogger getSimulationStatisticLogger() {
		return (SimulationStatisticLogger)getRoot();
	}

	protected IActivityAgentClassDeclaration getActivityAgentClassDeclaration(IActivityGraphNodeDeclaration activityDeclaration) {
		Iterator it = getSimulationStatisticLogger().getObjectClassDeclarationList().iterator();

		while (it.hasNext()) {
			IObjectClassDeclaration objectClassDeclaration = (IObjectClassDeclaration)it.next();
			if (objectClassDeclaration instanceof IActivityAgentClassDeclaration) {
				IActivityAgentClassDeclaration activityAgentClassDeclaration =
					(IActivityAgentClassDeclaration)objectClassDeclaration;

				if (hasActivityDeclaration(activityAgentClassDeclaration
					.getReasoningEngineDeclarationListAndFactory()
					.getDeclarationList()
					.iterator(),
					activityDeclaration))
					return activityAgentClassDeclaration;
			}
		}

		return null;
	}

	private boolean hasActivityDeclaration(
		Iterator reasoningEngines,
		IActivityGraphNodeDeclaration activityDeclaration) {
		while (reasoningEngines.hasNext()) {
			IActivityReasoningEngineDeclaration element =
				(IActivityReasoningEngineDeclaration)reasoningEngines.next();
			if (hasActivityDeclaration(element.getRootActivityGraphDeclaration(), activityDeclaration)) {
				return true;
			}
		}

		return false;
	}

	private boolean hasActivityDeclaration(
		IActivityGraphNodeDeclaration currActivityDeclaration,
		IActivityGraphNodeDeclaration activityDeclaration) {

		if (currActivityDeclaration == activityDeclaration) {
			return true;

		} else if (currActivityDeclaration instanceof IActivityGraphComposedNodeDeclaration) {
			IActivityGraphComposedNodeDeclaration currActivityGraphComposedNodeDeclaration =
				(IActivityGraphComposedNodeDeclaration)currActivityDeclaration;

			return hasActivityDeclaration(
				currActivityGraphComposedNodeDeclaration.getInnerActivityGraphDeclaration(),
				activityDeclaration);
		}

		return false;
	}

	private boolean hasActivityDeclaration(
		IActivityGraphDeclaration currActivityGraphDeclaration,
		IActivityGraphNodeDeclaration activityDeclaration) {

		Iterator it = currActivityGraphDeclaration.getActivityGraphNodeDeclarationList().iterator();

		while (it.hasNext()) {
			IActivityGraphNodeDeclaration childActivityDeclaration = (IActivityGraphNodeDeclaration)it.next();
			if (hasActivityDeclaration(childActivityDeclaration, activityDeclaration))
				return true;
		}

		return false;
	}

	public int getColumnCount() {
		return COLUMN_NAMES.length;
	}

	public String getColumnName(int column) {
		return COLUMN_NAMES[column];
	}

	public Class getColumnClass(int column) {
		return COLUMN_CLASSES[column];
	}

	public Object getValueAt(Object node, int column) {
		switch (column) {
			case 0 :
				if (node instanceof INamed)
					return ((INamed)node).getName();
				else
					return node + "";

			case 1 :
				return (String)getUpdateCommand().getValue(node);

			case 2 :
				return (String)getReasoningCommand().getValue(node);

			case 3 :
				return (String)getActionCommand().getValue(node);

			case 4 :
				return (String)getTotalCommand().getValue(node);

			default :
				return null;
		}
	}

	abstract class AbstractCommand {
		protected abstract double getValue(SimulationStatisticLogger.AgentClassInfo agentClassInfo);
		protected double getValue(
			SimulationStatisticLogger.AgentClassInfo agentClassInfo,
			SimulationStatisticLogger.ActivityInfo activityInfo) {
			return 0;
		}
		protected double getValue(SimulationStatisticLogger.AnalysisInfo analysisInfo) {
			return 0;
		}

		private double getValue(
			IActivityAgentClassDeclaration activityAgentClassDeclaration,
			IActivityGraphNodeDeclaration activityDeclaration) {

			if (activityDeclaration instanceof IActivityGraphDeclaration) {
				IActivityGraphDeclaration composedActivityDeclaration =
					(IActivityGraphDeclaration)activityDeclaration;
				Iterator it = composedActivityDeclaration.getActivityGraphNodeDeclarationList().iterator();

				double sum = 0;
				while (it.hasNext()) {
					sum += getValue(activityAgentClassDeclaration, (IActivityGraphNodeDeclaration)it.next());
				}

				return sum;

			} else {

				SimulationStatisticLogger.AgentClassInfo agentClassInfo =
					getSimulationStatisticLogger().getAgentClassInfo(activityAgentClassDeclaration);
				SimulationStatisticLogger.ActivityInfo activityInfo =
					getSimulationStatisticLogger().getActivityInfo(
						activityAgentClassDeclaration,
						activityDeclaration);

				if (activityInfo == null)
					return 0;
				else
					return getValue(agentClassInfo, activityInfo);
			}
		}

		public Object getValue(Object node) {
			if (node instanceof SimulationStatisticLogger) {
				//SimulationStatisticLogger simulationStatisticLogger = (SimulationStatisticLogger) node;
				return "";

			} else if (node instanceof IObjectClassDeclaration) {
				return transformToPercentAndAbsolute(
					getValue(
						getSimulationStatisticLogger().getAgentClassInfo((IObjectClassDeclaration)node)));

			} else if (node == _analysis_Root) {
				double sum = 0;
				Iterator it = getSimulationStatisticLogger().getAnalysisList().iterator();
				while (it.hasNext()) {
					IAnalysis analysis = (IAnalysis)it.next();
					sum += getValue(getSimulationStatisticLogger().getAnalysisInfo(analysis));
				}

				return transformToPercentAndAbsolute(sum);

			} else if (node instanceof IActivityGraphNodeDeclaration) {
				IActivityGraphNodeDeclaration activityDeclaration = (IActivityGraphNodeDeclaration)node;
				IActivityAgentClassDeclaration activityAgentClassDeclaration =
					getActivityAgentClassDeclaration(activityDeclaration);

				return transformToPercentAndAbsolute(
					getValue(activityAgentClassDeclaration, activityDeclaration));

			} else if (node instanceof IAnalysis) {
				IAnalysis analysis = (IAnalysis)node;

				return transformToPercentAndAbsolute(
					getValue(getSimulationStatisticLogger().getAnalysisInfo(analysis)));

			} else {
				return "";
			}
		}

		private String transformToPercentAndAbsolute(double value) {
			if (MathHelper.doubleEqual(value, 0))
				return "0% (0 ms)";
			else
				return _percentFormat.format(value / getSimulationStatisticLogger().getTotalTime())
					+ " ("
					+ _decimalFormat.format(value)
					+ " ms)";
		}
	}

	private AbstractCommand getUpdateCommand() {
		return new AbstractCommand() {
			protected double getValue(SimulationStatisticLogger.AgentClassInfo agentClassInfo) {
				return agentClassInfo.getUpdateTime();
			}
			public Object getValue(Object node) {
				if ((node instanceof IObjectClassDeclaration))
					return super.getValue(node);
				else
					return "";
			}
		};
	}

	private AbstractCommand getReasoningCommand() {
		return new AbstractCommand() {
			protected double getValue(SimulationStatisticLogger.AgentClassInfo agentClassInfo) {
				return agentClassInfo.getReasoningTime();
			}
			public Object getValue(Object node) {
				if ((node instanceof IActivityAgentClassDeclaration)
					|| (node instanceof IActivityGraphNodeDeclaration))
					return super.getValue(node);
				else
					return "";
			}
		};
	}

	private AbstractCommand getActionCommand() {
		return new AbstractCommand() {
			protected double getValue(SimulationStatisticLogger.AgentClassInfo agentClassInfo) {
				return agentClassInfo.getActionTime();
			}
			protected double getValue(
				SimulationStatisticLogger.AgentClassInfo agentClassInfo,
				SimulationStatisticLogger.ActivityInfo activityInfo) {
				return activityInfo.getTime();
			}
			protected double getValue(SimulationStatisticLogger.AnalysisInfo analysisInfo) {
				return analysisInfo.getTime();
			}
		};
	}

	private AbstractCommand getTotalCommand() {
		return new AbstractCommand() {
			protected double getValue(SimulationStatisticLogger.AgentClassInfo agentClassInfo) {
				return agentClassInfo.getTotalTime();
			}
			protected double getValue(
				SimulationStatisticLogger.AgentClassInfo agentClassInfo,
				SimulationStatisticLogger.ActivityInfo activityInfo) {
				return activityInfo.getTime();
			}
			protected double getValue(SimulationStatisticLogger.AnalysisInfo analysisInfo) {
				return analysisInfo.getTime();
			}
		};
	}

	public Object getChild(Object parent, int index) {
		if (parent instanceof SimulationStatisticLogger) {
			SimulationStatisticLogger simulationStatisticLogger = (SimulationStatisticLogger)parent;

			IObjectClassDeclarationList objectClassDeclarationList =
				simulationStatisticLogger.getObjectClassDeclarationList();

			if ((objectClassDeclarationList != null) && (index < objectClassDeclarationList.size()))
				return (IObjectClassDeclaration)objectClassDeclarationList.get(index);
			else
				return _analysis_Root;

		} else if (parent == _analysis_Root) {
			return (IAnalysis)getSimulationStatisticLogger().getAnalysisList().get(index);

		} else if (parent instanceof IActivityAgentClassDeclaration) {
			IActivityAgentClassDeclaration activityAgentClassDeclaration =
				(IActivityAgentClassDeclaration)parent;

			return (
				(IActivityReasoningEngineDeclaration)activityAgentClassDeclaration
					.getReasoningEngineDeclarationListAndFactory()
					.getDeclarationList()
					.get(index))
				.getRootActivityGraphDeclaration();

		} else if (parent instanceof IActivityGraphDeclaration) {
			IActivityGraphDeclaration composedActivityDeclaration = (IActivityGraphDeclaration)parent;

			return getActivityDeclaration(
				composedActivityDeclaration.getActivityGraphNodeDeclarationList(),
				index);
		}

		return null;
	}

	private IActivityGraphNodeDeclaration getActivityDeclaration(
		ActivityGraphNodeDeclarationList activityDeclarationList,
		int index) {
		Iterator it = activityDeclarationList.iterator();

		while (it.hasNext()) {
			IActivityGraphNodeDeclaration activityDeclaration = (IActivityGraphNodeDeclaration)it.next();
			if (activityDeclaration instanceof IEmergencyActivityGraphLeafDeclaration) {
			} else if (index == 0) {
				return activityDeclaration;
			} else {
				index--;
			}
		}

		return null;
	}

	public int getChildCount(Object parent) {
		if (parent instanceof SimulationStatisticLogger) {
			SimulationStatisticLogger simulationStatisticLogger = (SimulationStatisticLogger)parent;

			IObjectClassDeclarationList objectClassDeclarationList =
				simulationStatisticLogger.getObjectClassDeclarationList();

			if (objectClassDeclarationList != null)
				return objectClassDeclarationList.size() + 1;
			else
				return 1;

		} else if (parent == _analysis_Root) {
			return getSimulationStatisticLogger().getAnalysisList().size();

		} else if (parent instanceof IActivityAgentClassDeclaration) {
			IActivityAgentClassDeclaration activityAgentClassDeclaration =
				(IActivityAgentClassDeclaration)parent;

			return activityAgentClassDeclaration
				.getReasoningEngineDeclarationListAndFactory()
				.getDeclarationList()
				.size();

		} else if (parent instanceof IActivityGraphDeclaration) {
			IActivityGraphDeclaration composedActivityDeclaration = (IActivityGraphDeclaration)parent;

			return composedActivityDeclaration.getActivityGraphNodeDeclarationList().size() - 1;
		}

		return 0;
	}

	public boolean isLeaf(Object node) {
		if (node instanceof SimulationStatisticLogger) {
			return false;

		} else if (node == _analysis_Root) {
			return false;

		} else if (node instanceof IActivityAgentClassDeclaration) {
			return false;

		} else if (node instanceof IActivityGraphDeclaration) {
			return false;
		}

		return true;
	}
}
