package de.uniwue.ki.mas.sim.simulation;
import java.util.Date;
import java.util.Iterator;

import junit.framework.Assert;
import de.uniwue.ki.mas.model.SimulationRuntimeException;
import de.uniwue.ki.mas.sim.analysis.IAnalysis;
import de.uniwue.ki.util.name.AbstractNamed;

/**
 * <BR>Creation date: (26.04.00 14:37:11)
 * @author Christoph Oechslein
 */

abstract class AbstractSimulation extends AbstractNamed implements ISimulation {
	private SimulationListener _simulationListener = null;

	private StartPauseFinishListener _startPauseFinishListener = null;

	protected static final int SIMULATION_UNINITIALIZED = -1;
	protected static final int SIMULATION_INITIALIZED = 0;
	protected static final int SIMULATION_RUNNING = 1;
	protected static final int SIMULATION_PAUSED = 2;
	protected static final int SIMULATION_FINISHED = 3;
	protected static final int SIMULATION_FINISHED_WITH_ERRORS = 4;

	private int _simulationStatus = SIMULATION_UNINITIALIZED;
	private Throwable _exception = null;
	private Date _startDate;

	private SimulationFunctionArgs _simulationFunctionArgs;

	public AbstractSimulation(SimulationFunctionArgs simulationFunctionArgs) {
		Assert.assertNotNull(simulationFunctionArgs);
		_simulationFunctionArgs = simulationFunctionArgs;
	}

	protected void initSimulation(boolean withReset) throws SimulationRuntimeException {
		_startDate = new Date();
		setSimulationStatus(SIMULATION_INITIALIZED);
	}

	public void resetSimulation() {

		if (!isRunning()) {

			if (getSimulationStatus() != SIMULATION_INITIALIZED) {
				setSimulationStatus(SIMULATION_UNINITIALIZED);

				try {
					initSimulation(true);

					fireSimulationReseted();

				} catch (SimulationRuntimeException exception) {
					fireSimulationFinished(exception);
				}
			}

		} else {
			throw new IllegalStateException("Restarting not possible during simulation");
		}

	}

	public void addSimulationListener(SimulationListener newListener) {
		_simulationListener = SimulationEventMulticaster.add(_simulationListener, newListener);

	}

	public void addStartPauseFinishListener(StartPauseFinishListener newListener) {
		_startPauseFinishListener =
			StartPauseFinishEventMulticaster.add(_startPauseFinishListener, newListener);

	}

	protected void analyseAnalysisList() throws SimulationRuntimeException {

		Iterator it = getAnalysisList().iterator();
		if (getSimulationFunctionArgs().hasAnalysisListeners()) {
			while (it.hasNext()) {
				IAnalysis analysis = (IAnalysis) it.next();

				getSimulationFunctionArgs().fireStartedAnalysis(analysis, this);
				analysis.analyse(getSimulationFunctionArgs());
				getSimulationFunctionArgs().fireFinishedAnalysis(analysis, this);
			}

		} else {
			while (it.hasNext()) {
				((IAnalysis) it.next()).analyse(getSimulationFunctionArgs());
			}
		}
	}

	protected void fireSimulationFinished(Throwable exception) {

		setException(exception);

		setSimulationStatus((exception == null) ? SIMULATION_FINISHED : SIMULATION_FINISHED_WITH_ERRORS);

		if (_startPauseFinishListener != null)
			_startPauseFinishListener.finished(new SimulationEvent(this, exception, getSimulationFunctionArgs()));

		if ((exception != null) && (_startPauseFinishListener != null) && (_simulationListener != null))
			exception.printStackTrace(System.err);
	}

	protected void fireSimulationPaused() {
		setSimulationStatus(SIMULATION_PAUSED);

		if (_startPauseFinishListener != null)
			_startPauseFinishListener.paused(new SimulationEvent(this, getSimulationFunctionArgs()));
	}

	protected void fireSimulationReseted() {
		if (_simulationListener != null)
			_simulationListener.reseted(new SimulationEvent(this, getSimulationFunctionArgs()));
	}

	protected void fireSimulationStarted() {
		setSimulationStatus(SIMULATION_RUNNING);

		if (_startPauseFinishListener != null)
			_startPauseFinishListener.started(new SimulationEvent(this, getSimulationFunctionArgs()));
	}

	protected void initAnalysisList() throws SimulationRuntimeException {
		getSimulationFunctionArgs().setObject(getSimulationFunctionArgs().getWorld());

		Iterator it = getAnalysisList().iterator();
		while (it.hasNext()) {
			((IAnalysis) it.next()).init(getSimulationFunctionArgs());
		}
	}

	public void remove() {
		_simulationStatus = SIMULATION_UNINITIALIZED;

		_exception = null;
		_simulationListener = null;
		_startPauseFinishListener = null;
		_startDate = null;

		Iterator it = getAnalysisList().iterator();
		while (it.hasNext()) {
			((IAnalysis) it.next()).remove();
			it.remove();
		}

		Assert.assertEquals(0, getAnalysisList().size());

		getSituation().remove();
	}

	public void removeSimulationListener(SimulationListener newListener) {
		_simulationListener = SimulationEventMulticaster.remove(_simulationListener, newListener);

	}

	public void removeStartPauseFinishListener(StartPauseFinishListener newListener) {
		_startPauseFinishListener =
			StartPauseFinishEventMulticaster.remove(_startPauseFinishListener, newListener);
	}

	protected void setSimulationStatus(int simulationStatus) {
		_simulationStatus = simulationStatus;
	}

	protected int getSimulationStatus() {
		return _simulationStatus;
	}

	public Throwable getException() {
		return _exception;
	}

	public void setException(Throwable exception) {
		_exception = exception;
	}

	public boolean isFinished() {
		return getSimulationStatus() == SIMULATION_FINISHED;
	}

	public boolean isFinishedWithErrors() {
		return getSimulationStatus() == SIMULATION_FINISHED_WITH_ERRORS;
	}

	public boolean isPaused() {
		return getSimulationStatus() == SIMULATION_PAUSED;
	}

	public boolean isRunning() {
		return getSimulationStatus() == SIMULATION_RUNNING;
	}

	public Date getStartDate() {
		return _startDate;
	}

	public boolean isShouldTerminate() {
		return false;
	}

	public SimulationFunctionArgs getSimulationFunctionArgs() {
		return _simulationFunctionArgs;
	}
}