package de.uniwue.ki.mas.sim.simulation;
import java.util.Iterator;
import java.util.Map;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.functionCall.NamedFunctionCallDeclaration;
import de.uniwue.ki.mas.basics.functionCall.NamedFunctionCallDeclarationFactory;
import de.uniwue.ki.mas.basics.type.BooleanType;
import de.uniwue.ki.mas.sim.analysis.AnalysisDeclarationList;
import de.uniwue.ki.mas.sim.analysis.IAnalysisDeclaration;
import de.uniwue.ki.mas.sim.situation.ISituationDeclaration;
import de.uniwue.ki.mas.util.declaration.*;
import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (30.07.2001 18:37:16)
 * @author Christoph Oechslein
 */

class SimulationDeclaration extends AbstractDocumentableDeclaration implements ISimulationDeclaration {
	private final transient ChangeListener _changeListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			fireStateChanged();
		}
	};

	private final DeclarationListListener _declarationListListener = new DeclarationListListener() {
		public void declarationAdded(DeclarationListEvent event) {
			IAnalysisDeclaration analysisDeclaration = (IAnalysisDeclaration)event.getDeclaration();
			analysisDeclaration.addRemoveVetoListener(_neverRemoveVetoListener);

		}
		public void declarationRemoved(DeclarationListEvent event) {
			IAnalysisDeclaration analysisDeclaration = (IAnalysisDeclaration)event.getDeclaration();
			analysisDeclaration.removeRemoveVetoListener(_neverRemoveVetoListener);
		}
	};

	private final DefaultDeclarationListAndFactory _endFunctionDeclarationListAndFactory =
		new DefaultDeclarationListAndFactory(new DefaultDeclarationList() {
		public boolean accept(Object anObject) {
			return super.accept(anObject)
				&& (anObject instanceof NamedFunctionCallDeclaration)
				&& Utils.equals(
					((NamedFunctionCallDeclaration)anObject).getMostGeneralOutputType(),
					BooleanType.getInstance());
		}
		public String getName() {
			return "End Function List";
		}
	}, new NamedFunctionCallDeclarationFactory(BooleanType.getInstance(), Boolean.FALSE));

	protected final RemoveVetoListener _neverRemoveVetoListener = new NeverRemoveVetoListener(this);
	private ISituationDeclaration _situationDeclaration;
	private final AnalysisDeclarationList _analysisDeclarationList;

	private boolean _testSimulationDeclaration;
	private final IDeclarationList _afterRunTestActionDefinitionList = new DefaultDeclarationList();
	private final IDeclarationList _duringRunTestActionDefinitionList = new DefaultDeclarationList();

	protected SimulationDeclaration() {
		super();

		_analysisDeclarationList = new AnalysisDeclarationList();
		_analysisDeclarationList.setComposition(false);
		_analysisDeclarationList.addDeclarationListListener(_declarationListListener);

		getAnalysisDeclarationList().addChangeListener(_changeListener);
		getEndFunctionDeclarationListAndFactory().getDeclarationList().addChangeListener(_changeListener);
	}

	public SimulationDeclaration(ISituationDeclaration situationDeclaration) {
		this();

		setSituationDeclaration(situationDeclaration);
	}

	public void copyInto(Object intoObject, Map objectMapping) {
		super.copyInto(intoObject, objectMapping);

		ISimulationDeclaration intoSimulationDeclaration = (ISimulationDeclaration)intoObject;

		intoSimulationDeclaration.setTestSimulationDeclaration(isTestSimulationDeclaration());
		getAfterRunTestActionDefinitionList().copyInto(
			intoSimulationDeclaration.getAfterRunTestActionDefinitionList(),
			objectMapping);
		getDuringRunTestActionDefinitionList().copyInto(
			intoSimulationDeclaration.getDuringRunTestActionDefinitionList(),
			objectMapping);

		if (intoSimulationDeclaration.getSituationDeclaration() == null)
			intoSimulationDeclaration.setSituationDeclaration(getSituationDeclaration());

		// don't copy the individual analysis declarations
		//getAnalysisDeclarationList().copyInto(intoSimulationDeclaration.getAnalysisDeclarationList());

		Assert.assertFalse(intoSimulationDeclaration.getAnalysisDeclarationList().isComposition());

		intoSimulationDeclaration.getAnalysisDeclarationList().clear();

		Iterator it = getAnalysisDeclarationList().iterator();
		while (it.hasNext()) {
			IAnalysisDeclaration analysisDeclaration = (IAnalysisDeclaration)it.next();

			IAnalysisDeclaration copyAnalysisDeclaration;

			if ((objectMapping != null) && (objectMapping.containsKey(analysisDeclaration))) {
				copyAnalysisDeclaration = (IAnalysisDeclaration)objectMapping.get(analysisDeclaration);

			} else {
				copyAnalysisDeclaration = analysisDeclaration;
			}

			if (!intoSimulationDeclaration.getAnalysisDeclarationList().contains(copyAnalysisDeclaration)) {
				intoSimulationDeclaration.getAnalysisDeclarationList().add(copyAnalysisDeclaration);
			}
		}

		getEndFunctionDeclarationListAndFactory().copyInto(
			intoSimulationDeclaration.getEndFunctionDeclarationListAndFactory(),
			objectMapping);
	}

	protected ICopyable createInstance(Map objectMapping) {
		return new SimulationDeclaration();
	}

	protected void fireStateChanged() {
		super.fireStateChanged();
	}

	public IDeclarationList getAfterRunTestActionDefinitionList() {
		return _afterRunTestActionDefinitionList;
	}

	public final AnalysisDeclarationList getAnalysisDeclarationList() {
		return _analysisDeclarationList;
	}

	protected String getDefaultName() {
		if (getSituationDeclaration() == null)
			return "Sim [no situation given]";
		else
			return "Sim [" + getSituationDeclaration().getName() + "]";
	}

	public IDeclarationList getDuringRunTestActionDefinitionList() {
		return _duringRunTestActionDefinitionList;
	}

	public final IDeclarationListAndFactory getEndFunctionDeclarationListAndFactory() {
		return _endFunctionDeclarationListAndFactory;
	}

	public ISituationDeclaration getSituationDeclaration() {
		return _situationDeclaration;
	}

	public boolean isomorphicEquals(IIsomorphicHashable anObject) {
		if (anObject == this)
			return true;

		if (!super.isomorphicEquals(anObject)) {
			return false;
		}

		if (!(anObject instanceof ISimulationDeclaration)) {
			return false;
		}

		ISimulationDeclaration simulationDeclaration = (ISimulationDeclaration)anObject;

		return isTestSimulationDeclaration() == simulationDeclaration.isTestSimulationDeclaration()
			&& getAfterRunTestActionDefinitionList().isomorphicEquals(
				simulationDeclaration.getAfterRunTestActionDefinitionList())
			&& getDuringRunTestActionDefinitionList().isomorphicEquals(
				simulationDeclaration.getDuringRunTestActionDefinitionList())
			&& getAnalysisDeclarationList().isomorphicEquals(simulationDeclaration.getAnalysisDeclarationList())
			&& isomorphicEquals(getSituationDeclaration(), simulationDeclaration.getSituationDeclaration())
			&& getEndFunctionDeclarationListAndFactory().isomorphicEquals(
				simulationDeclaration.getEndFunctionDeclarationListAndFactory());
	}

	public int isomorphicHashCode() {
		return 13 * getDuringRunTestActionDefinitionList().isomorphicHashCode()
			+ 11 * getAfterRunTestActionDefinitionList().isomorphicHashCode()
			+ 7 * getSituationDeclaration().isomorphicHashCode()
			+ 5 * getAnalysisDeclarationList().isomorphicHashCode()
			+ 3 * getEndFunctionDeclarationListAndFactory().isomorphicHashCode()
			+ isomorphicHashCode(isTestSimulationDeclaration());
	}

	public boolean isTestSimulationDeclaration() {
		return _testSimulationDeclaration;
	}

	public void remove() {
		getAnalysisDeclarationList().removeChangeListener(_changeListener);
		getAnalysisDeclarationList().clear();
		getEndFunctionDeclarationListAndFactory().getDeclarationList().removeChangeListener(_changeListener);

		getAfterRunTestActionDefinitionList().remove();
		getDuringRunTestActionDefinitionList().remove();

		setSituationDeclaration(null);

		super.remove();
	}

	public void setSituationDeclaration(ISituationDeclaration situationDeclaration) {
		ISituationDeclaration oldValue = _situationDeclaration;

		_situationDeclaration = situationDeclaration;

		if (oldValue != null) {
			oldValue.removeRemoveVetoListener(_neverRemoveVetoListener);
			oldValue.removeChangeListener(_changeListener);
		}

		if (situationDeclaration != null) {
			situationDeclaration.addRemoveVetoListener(_neverRemoveVetoListener);
			situationDeclaration.addChangeListener(_changeListener);
		}

		firePropertyChange("situationDeclaration", oldValue, situationDeclaration);
		fireStateChanged();
	}

	public void setTestSimulationDeclaration(boolean testSimulationDeclaration) {
		_testSimulationDeclaration = testSimulationDeclaration;

		fireStateChanged();
	}
}