package de.uniwue.ki.mas.sim.simulation;
import java.util.Iterator;

import junit.framework.Assert;
import de.uniwue.ki.mas.entity.object.IObject;
import de.uniwue.ki.mas.entity.object.IObjectExecuteFunctionArgs;
import de.uniwue.ki.mas.entity.object.agent.IAgent;
import de.uniwue.ki.mas.entity.object.world.IWorld;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * The world act, then all agents sense, reason and act (update),
 * the order of the agents is shuffled each time and one agent is
 * updated until it finished.
 * Creation date: (26.10.2000 11:24:12)
 *
 * @author Christoph Oechslein
 */

final class SimulationStrategieSerial implements ISimulationStrategie {

	/**
	 * First advance the time, then update the world, then update
	 * (see {@link #simulateOneStepAgent(IAgent, IObjectExecuteFunctionArgs) simulateOneStepAgent}
	 * and {@link #simulateOneStepObject(IObject, IObjectExecuteFunctionArgs) simulateOneStepAgent})
	 * shuffled agents.
	 */
	public void simulateOneStep(ISimulationExecuteFunctionArgs simulationExecuteFunctionArgs)
		throws SimulationRuntimeException {

		IWorld world = simulationExecuteFunctionArgs.getWorld();

		ISimulation simulation = simulationExecuteFunctionArgs.getSimulation();

		simulation.getSituation().nextTimeStep();

		boolean updateWorldBeforeAllObjects = simulation.isUpdateWorldBeforeAllObjects();

		if (updateWorldBeforeAllObjects) {
			Assert.assertFalse(world.isDead());

			simulateOneStepObject(world, simulationExecuteFunctionArgs);
		}

		// now new objects has been added to the world
		// create a random order
		world.shuffleObjects();

		Iterator objects = world.getAllObjects(false, false);

		while (objects.hasNext()) {
			IObject obj = (IObject) objects.next();

			if (!obj.isDead()) {
				simulateOneStepObject(obj, simulationExecuteFunctionArgs);
			}
		}

		if (!updateWorldBeforeAllObjects) {
			Assert.assertFalse(world.isDead());

			simulateOneStepObject(world, simulationExecuteFunctionArgs);
		}
	}

	/**
	 * Simulate one agent by first setting the object in the executing binding and doing
	 * <OL>
	 *   <LI> {@link IObject#update() update features},
	 *   <LI> {@link IAgent#reason() reason} and
	 *   <LI> {@link for one time#act() act}
	 * </OL>
	 * until the agent has {@link IAgent#isFinishedOneTick() finished}.
	 */
	private void simulateOneStepObject(
		IObject obj,
		ISimulationExecuteFunctionArgs objectExecuteFunctionArgs)
		throws SimulationRuntimeException {

		if (objectExecuteFunctionArgs.getSimulation().isShouldTerminate()) {
			return;
		}

		Assert.assertFalse(obj.isDead());

		IObject currentActiveObject = objectExecuteFunctionArgs.getObject();

		objectExecuteFunctionArgs.setObject(obj);
		obj.update(objectExecuteFunctionArgs);

		if (Assert.ASSERTION_ON) {
			Assert.assertSame(obj, objectExecuteFunctionArgs.getObject());
		}

		objectExecuteFunctionArgs.setObject(currentActiveObject);
	}
}