package de.uniwue.ki.mas.sim.simulation;
import junit.framework.Assert;
import de.uniwue.ki.mas.basics.function.IFunction;
import de.uniwue.ki.mas.model.ModelCompilingException;
import de.uniwue.ki.mas.model.SimulationRuntimeException;

/**
 * Creation date: (12.09.2001 12:54:28)
 * @author Christoph Oechslein
 */

final class ThreadedSimulation extends Simulation implements IThreadedSimulation {

	private boolean _simulationShouldPause = false;
	private boolean _simulationShouldTerminate = false;
	private transient Thread _simulationThread;

	public ThreadedSimulation(
		IFunction nameFunction,
		ISimulationDeclaration simulationDeclaration,
		ISimulationStrategie simulationsStrategie,
		SimulationFunctionArgs simulationFunctionArgs)
		throws ModelCompilingException {

		super(nameFunction, simulationDeclaration, simulationsStrategie, simulationFunctionArgs);
	}

	public ThreadedSimulation(
		IFunction nameFunction,
		ISimulationDeclaration simulationDeclaration,
		SimulationFunctionArgs simulationFunctionArgs)
		throws ModelCompilingException {

		super(nameFunction, simulationDeclaration, simulationFunctionArgs);
	}

	protected void cleanup(Throwable exception) {

		boolean currentIsSimulationShouldPause;

		synchronized (ThreadedSimulation.this) {
			ThreadedSimulation.this.notifyAll();
			currentIsSimulationShouldPause = isShouldPause();
		}

		if (exception != null) {
			// an exception occurred

			fireSimulationFinished(exception);

		} else if (currentIsSimulationShouldPause) {
			synchronized (ThreadedSimulation.this) {
				setSimulationShouldPause(false);
			}

			fireSimulationPaused();

		} else {

			fireSimulationFinished(null);
		}
	}

	public Thread getSimulationThread() {
		return _simulationThread;
	}

	private synchronized boolean isShouldPause() {
		return _simulationShouldPause;
	}

	public synchronized boolean isShouldTerminate() {
		return _simulationShouldTerminate;
	}

	public void remove() {
		terminateInThread();

		if (getSimulationFunctionArgs() != null)
			getSimulationFunctionArgs().remove();

		if (_simulationThread != null)
			_simulationThread = null;

		super.remove();
	}

	private synchronized void setSimulationShouldPause(boolean newSimulationShouldPause) {
		_simulationShouldPause = newSimulationShouldPause;
	}

	private synchronized void setSimulationShouldTerminate() {
		_simulationShouldTerminate = true;
	}

	private void setSimulationThread(Thread newSimulationThread) {
		Assert.assertTrue((_simulationThread == null) || (!_simulationThread.isAlive()));

		_simulationThread = newSimulationThread;
	}

	protected synchronized boolean simulationNextStep() throws SimulationRuntimeException {
		return super.simulationNextStep() && !isShouldPause() && !isShouldTerminate();
	}

	public void startSimulation() {
		Assert.assertNotNull(getSimulationFunctionArgs());

		startSimulation(false);
	}

	private void startSimulation(boolean asynchron) {
		Assert.assertNotNull(getSimulationFunctionArgs());

		if (isRunning()) {
			throw new IllegalStateException("Starting not possible during simulation");

		} else {
			try {

				synchronized (this) {
					waitForOldThread();

					if (getSimulationStatus() == SIMULATION_UNINITIALIZED) {
						initSimulation(true);
					}

					setSimulationShouldPause(false);
				}

				// was here but after that isRunning is true but thread is not alive see isSimRunning...
				//fireSimulationStarted(simulationExecuteFunctionArgs);

				if (asynchron) {

					Thread thread;
					synchronized (this) {
						thread = new Thread(new Runnable() {
							public void run() {
								try {
									fireSimulationStarted();
									simulate();

								} finally {
									synchronized (ThreadedSimulation.this) {
										ThreadedSimulation.this.notifyAll();
									}
								}
							}
						}, getName());

						if (Assert.ASSERTION_ON)
							thread.setPriority(Thread.MIN_PRIORITY);
						else
							thread.setPriority(Thread.currentThread().getPriority() - 1);
						setSimulationThread(thread);
					}

					thread.start();

				} else {
					fireSimulationStarted();
					simulate();
				}

			} catch (Exception exception) {

				setSimulationShouldPause(true);

				fireSimulationFinished(exception);
			}
		}
	}

	private synchronized void waitForOldThread() {
		Assert.assertTrue(!isRunning());

		while ((getSimulationThread() != null) && (getSimulationThread().isAlive())) {
			try {
				wait(10);
			} catch (InterruptedException exception) {
			}
		}
	}

	public void pauseInThread() {
		if (!isSimThreadRunning()) {
			pauseOrTerminateError();

		} else {
			setSimulationShouldPause(true);
		}
	}

	public void terminateInThread() {
		if (!isSimThreadRunning()) {
			pauseOrTerminateError();
		}

		setSimulationShouldTerminate();

		synchronized (this) {
			int counter = 0;
			while (isRunning() && (counter++ < 100)) {
				try {
					wait(10);
				} catch (InterruptedException e) {
				}
			}
		}

		if (getSimulationStatus() != SIMULATION_FINISHED_WITH_ERRORS)
			setSimulationStatus(SIMULATION_FINISHED);
	}

	private synchronized boolean isSimThreadRunning() {
		return (getSimulationThread() != null) && !(isRunning() && !getSimulationThread().isAlive());
	}

	private void pauseOrTerminateError() {
		if (getSimulationThread() == null) {
			setException(new SimulationRuntimeException("getSimulationThread() is null"));
		} else {
			setException(new SimulationRuntimeException("Is in Running state, but sim thread is not alive"));
		}

		setSimulationStatus(SIMULATION_FINISHED_WITH_ERRORS);

		//setSimulationThread(null);
	}

	protected void simulate() {

		Throwable exception = null;

		try {
			while (simulationNextStep()) {
				simulateOneStep();
				Thread.yield();
			}

			if (!isShouldPause())
				executeAfterTestAction();

		} catch (SimulationRuntimeException occuredException) {
			exception = occuredException;

		} catch (Throwable occuredException) {
			exception = occuredException;
		}

		cleanup(exception);
	}

	public void startInThread() {
		startSimulation(true);
	}
}