package de.uniwue.ki.mas.test;

import java.io.*;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.swingui.TestRunner;
import de.uniwue.ki.mas.application.SeSAmCmd;
import de.uniwue.ki.mas.basics.functionDeclaration.FunctionDeclarationCompiler;
import de.uniwue.ki.mas.model.ISeSAmDeclaration;
import de.uniwue.ki.mas.model.SeSAmDeclarationFactory;
import de.uniwue.ki.mas.sim.experiment.*;
import de.uniwue.ki.mas.sim.simulation.*;
import de.uniwue.ki.mas.sim.situation.ISituationDeclaration;
import de.uniwue.ki.mas.xml.*;
import de.uniwue.ki.mas.xml.reader.XMLReadException;
import de.uniwue.ki.mas.xml.writer.XMLWriteException;

/**
 * Creation date: (31.10.2000 13:40:53)
 * @author Christoph Oechslein
 */

public final class TestXMLTestModels extends TestCase {
	private static final String TO_SAVE_EXTENSION = "-saved.xml";

	public static final String LOAD_MODEL_DIRECTORY = "XmlModels/ToLoad/";
	public static final String SIMULATE_MODEL_DIRECTORY = "XmlModels/ToSimulate/";

	private final File _simulationFile;

	static {
		try {
			SeSAmCmd.loadPluginIniFiles();
		} catch (XMLIniFileException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	public TestXMLTestModels(String name, File simulationFile) {
		super(name);

		_simulationFile = simulationFile;
	}

	public static void main(String[] args) {
		TestRunner.run(TestXMLTestModels.class);
	}

	private static File getDirectory(String name) {
		return new File(TestXMLTestModels.class.getPackage().getName().replace('.', '/') + "/" + name);
	}

	public static TestSuite suite() {
		TestSuite anSuite = new TestSuite("XMLModels");

		System.out.println("compile is " + FunctionDeclarationCompiler.isDefaultCompileOn());
		System.out.println("inline is " + FunctionDeclarationCompiler.isDefaultInlineOn());

		File[] files;
		files = getDirectory(LOAD_MODEL_DIRECTORY).listFiles();

		if (files != null) {
			for (int i = 0; i < files.length; i++) {
				if (files[i].getName().endsWith(".xml")) {
					anSuite.addTest(new TestXMLTestModels("testLoadXmlModel", files[i]));
				}
			}
		}

		files = getDirectory(SIMULATE_MODEL_DIRECTORY).listFiles();

		if (files != null) {
			for (int i = 0; i < files.length; i++) {
				if (files[i].getName().endsWith(".xml"))
					anSuite.addTest(new TestXMLTestModels("testSimulateXmlModel", files[i]));
			}
		}

		return anSuite;
	}

	public void testLoadXmlModel() throws Throwable {
		testLoadXmlModel(_simulationFile);
	}

	private void testLoadXmlModel(File file) throws Throwable {

		ISeSAmDeclaration seSAmDeclaration = readModelFileAndCreateSimulationsAndExperiments(file);

		testSerialization(file.getName(), seSAmDeclaration);

		File newModelFile = writeModelFile(file.getName(), seSAmDeclaration);

		ISeSAmDeclaration newSeSAmDeclaration = readModelFileAndCreateSimulationsAndExperiments(newModelFile);

		System.out.print("\n\n");

		newSeSAmDeclaration.remove();
		seSAmDeclaration.remove();
	}

	private void testSerialization(String fileName, ISeSAmDeclaration seSAmDeclaration)
		throws IOException, ClassNotFoundException, XMLWriteException, Throwable {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		ObjectOutputStream oos = new ObjectOutputStream(baos);

		try {
			oos.writeObject(seSAmDeclaration);

		} catch (NotSerializableException e) {
			e.printStackTrace();
		}

		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(baos.toByteArray()));

		//ISeSAmDeclaration newSeSAmDeclaration2 = (ISeSAmDeclaration) 
		ois.readObject();

		//		File newModelFile2 = writeModelFile(fileName, newSeSAmDeclaration2);
		//
		//		newModelFile2.deleteOnExit();
		//
		//		ISeSAmDeclaration newSeSAmDeclaration =
		//			readModelFileAndCreateSimulationsAndExperiments(newModelFile2);
		//
		//		newSeSAmDeclaration.remove();

		// leed to problems
		//newSeSAmDeclaration2.remove();

		oos.close();
		ois.close();
	}

	//	private void testSerialization(Object obj)
	//		throws IOException, ClassNotFoundException, XMLWriteException {
	//
	//		ByteArrayOutputStream baos = new ByteArrayOutputStream();
	//
	//		ObjectOutputStream oos = new ObjectOutputStream(baos);
	//
	//		System.out.println("Write obj " + obj);
	//		try {
	//			oos.writeObject(obj);
	//
	//		} catch (NotSerializableException e) {
	//			e.printStackTrace();
	//		}
	//
	//		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(baos.toByteArray()));
	//
	//		System.out.println("Read obj " + obj);
	//		Object obj2 = ois.readObject();
	//
	//		Assert.assertSame(obj.getClass(), obj2.getClass());
	//	}

	private File writeModelFile(String fileName, ISeSAmDeclaration seSAmDeclaration)
		throws IOException, XMLWriteException {
		File newModelFile = File.createTempFile("testXMLModelFiles", TO_SAVE_EXTENSION);

		System.out.println("Write file' " + fileName + "' to file '" + newModelFile.getName() + "'");

		XMLWriter.writeModel(newModelFile, seSAmDeclaration, null);

		newModelFile.deleteOnExit();

		return newModelFile;
	}

	private ISeSAmDeclaration readModelFileAndCreateSimulationsAndExperiments(File file)
		throws XMLReadException, IOException, Throwable {

		ISeSAmDeclaration seSAmDeclaration = SeSAmDeclarationFactory.createSeSAmDeclaration(false);

		System.out.println("Load file '" + file.getName() + "'");

		XMLReader.readModelInto(file, seSAmDeclaration, null, true);

		createSimulationsAndExperiments(seSAmDeclaration, false);

		return seSAmDeclaration;
	}

	private void createSimulationsAndExperiments(ISeSAmDeclaration seSAmDeclaration, boolean runSimulations)
		throws Throwable {

		Level oldLevel = Logger.getLogger("").getLevel();
		Logger.getLogger("").setLevel(Level.WARNING);

		int experimentCount = createExperiments(seSAmDeclaration, runSimulations);
		int simulationCount = createSimulations(seSAmDeclaration, runSimulations);

		int situationCount = 0;

		if (experimentCount == 0 || simulationCount == 0) {
			situationCount = createSituations(seSAmDeclaration, runSimulations);
			if (situationCount == 1) {
				createSituations(seSAmDeclaration, runSimulations);
			}
		}

		if (situationCount + simulationCount + experimentCount == 1) {
			createExperiments(seSAmDeclaration, runSimulations);
			createSimulations(seSAmDeclaration, runSimulations);
			createSituations(seSAmDeclaration, runSimulations);
		}

		Logger.getLogger("").setLevel(oldLevel);
	}

	private int createSituations(ISeSAmDeclaration seSAmDeclaration, boolean runSimulations)
		throws Throwable {
		int count = 0;

		Iterator it =
			seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getSituationDeclarationListAndFactory()
				.getSituationDeclarationList()
				.iterator();

		SimulationDeclarationFactory simulationDeclarationFactory =
			seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getSimulationDeclarationListAndFactory()
				.getSimulationDeclarationFactory();

		while (it.hasNext()) {
			ISituationDeclaration situationDeclaration = (ISituationDeclaration)it.next();

			System.out.println("Create Situation '" + situationDeclaration.getName() + "'");

			ISimulationDeclaration simulationDeclaration =
				simulationDeclarationFactory.createSimulationDeclaration(situationDeclaration);

			IExperiment exp = ExperimentFactory.createExperiment(seSAmDeclaration, simulationDeclaration);

			if (runSimulations) {
				System.out.println("Simulate Situation '" + exp.getName() + "'");

				simulateExperiment(exp);
			}

			exp.remove();
			count++;
		}

		return count;
	}

	private int createSimulations(ISeSAmDeclaration seSAmDeclaration, boolean runSimulations)
		throws Throwable {

		int count = 0;

		Iterator it =
			seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getSimulationDeclarationListAndFactory()
				.getSimulationDeclarationList()
				.iterator();

		while (it.hasNext()) {
			ISimulationDeclaration simulationDeclaration = (ISimulationDeclaration)it.next();

			System.out.println("Create Simulation '" + simulationDeclaration.getName() + "'");

			IExperiment exp = ExperimentFactory.createExperiment(seSAmDeclaration, simulationDeclaration);

			if (runSimulations) {
				System.out.println("Simulate Simulation '" + exp.getName() + "'");

				simulateExperiment(exp);
			}

			exp.remove();
			count++;
		}

		return count;
	}

	private int createExperiments(ISeSAmDeclaration seSAmDeclaration, boolean runSimulations)
		throws Throwable {

		int count = 0;

		Iterator it =
			seSAmDeclaration
				.getSeSAmModelDeclaration()
				.getSimulationElementsDeclaration()
				.getExperimentDeclarationListAndFactory()
				.getExperimentDeclarationList()
				.iterator();

		while (it.hasNext()) {
			IExperimentDeclaration experimentDeclaration = (IExperimentDeclaration)it.next();

			System.out.println("Create Experiment '" + experimentDeclaration.getName() + "'");

			IExperiment exp = ExperimentFactory.createExperiment(seSAmDeclaration, experimentDeclaration);

			if (runSimulations) {
				System.out.println("Simulate Experiment '" + exp.getName() + "'");

				simulateExperiment(exp);
			}

			exp.remove();
			count++;
		}

		return count;
	}

	private void simulateExperiment(IExperiment exp) throws Throwable {
		long startTime = System.currentTimeMillis();
		exp.startInThread();

		do {
			synchronized (this) {
				try {
					wait(1 * 1000);
				} catch (InterruptedException exception) {
				}
			}

			if (System.currentTimeMillis() - startTime > 1 * 60 * 1000) {
				System.out.print("Terminated experiment: " + exp.getName());
				ISimulation currentSimulation = exp.getCurrentSimulation();
				if (currentSimulation != null)
					System.out.println(
						" at tick " + currentSimulation.getSituation().getTimeDurationInTicks() + ".");
				else
					System.out.println(" curr sim == null!");
				exp.terminateInThread();
				break;
			}

		} while (exp.isRunning());

		if (exp.getException() != null)
			throw exp.getException();
	}

	public void testSimulateXmlModel() throws Throwable {
		testSimulateXmlModel(_simulationFile);
	}

	private void testSimulateXmlModel(File file) throws Throwable {
		ISeSAmDeclaration seSAmDeclaration = SeSAmDeclarationFactory.createSeSAmDeclaration(false);

		System.out.println("Load file '" + file.getName() + "'");

		XMLReader.readModelInto(file, seSAmDeclaration, null, true);

		createSimulationsAndExperiments(seSAmDeclaration, true);

		System.out.println();

		seSAmDeclaration.remove();
	}

	public String getName() {
		return super.getName() + " (" + _simulationFile.getName() + ")";
	}
}