package de.uniwue.ki.mas.util.declaration;
import java.beans.*;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;
import de.uniwue.ki.util.copy.AbstractCopyable;
import de.uniwue.ki.util.event.StateChangeMulticaster;
import de.uniwue.ki.util.hash.IIsomorphicHashable;

/**
 * Creation date: (25.07.2001 14:11:45)
 * 
 * @author Christoph Oechslein
 */

public abstract class AbstractDeclaration extends AbstractCopyable implements IDeclaration {
	private transient ChangeListener _changeListener = null;
	private RemoveVetoListener _removeVetoListener = null;
	private transient PropertyChangeSupport _propertyChange;
	private boolean TODOflag = false;

	public AbstractDeclaration() {
		super();
	}

	private RemoveVetoListener vetoListener = new RemoveVetoListener() {
		public boolean canBeRemoved() {
			return false;
		}
	};

	private LinkedList dependentObjects = new LinkedList();

	public boolean hasTODOFlag() {
		return TODOflag;
	}

	public void setTODOFlag(boolean flag) {
		TODOflag = flag;
		fireStateChanged();
	}

	protected void addRemoveVetoListernerTo(IDeclaration decl) {
		if ((decl != null) && (!dependentObjects.contains(decl))) {
			dependentObjects.add(decl);
			decl.addRemoveVetoListener(vetoListener);
		}
	}

	protected void removeRemoveVetoListenerFrom(IDeclaration decl) {
		if ((decl != null) && (dependentObjects.contains(decl))) {
			dependentObjects.remove(decl);
			decl.removeRemoveVetoListener(vetoListener);
		}
	}

	public void addChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.add(_changeListener, newListener);
	}

	public synchronized void addPropertyChangeListener(PropertyChangeListener listener) {
		getPropertyChange().addPropertyChangeListener(listener);
	}

	public synchronized void addPropertyChangeListener(
		String propertyName,
		PropertyChangeListener listener) {
		getPropertyChange().addPropertyChangeListener(propertyName, listener);
	}

	public void addRemoveVetoListener(RemoveVetoListener newListener) {
		if (Assert.ASSERTION_ON)
			Assert.assertNotNull(newListener);

		_removeVetoListener = RemoveVetoMulticaster.add(_removeVetoListener, newListener);
	}

	public void removeAllRemoveVetoListeners() {
		_removeVetoListener = null;
	}

	/**
	 * Returns <code>true</code> if <code>recursive</code> is <code>true</code>
	 * or asks removeVetoListeners.
	 */
	public boolean canBeRemoved() {
		boolean result = removingOkForListeners();
		//		if (!result)
		//		   System.err.println(this +" can not be removed because of " +
		// _removeVetoListener);
		return result;
	}

	/**
	 * Returns <code>true</code> if <code>declaration</code> is <code>null</code>
	 * or asks <code>declaration</code> via its
	 * {@link IDeclaration#canBeRemoved(boolean) non static method}.
	 */
	protected static final boolean canBeRemoved(IDeclaration declaration) {
		return declaration == null || declaration.canBeRemoved();
	}

	protected void firePropertyChange(PropertyChangeEvent evt) {
		getPropertyChange().firePropertyChange(evt);
	}

	protected void firePropertyChange(String propertyName, int oldValue, int newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	protected void firePropertyChange(String propertyName, boolean oldValue, boolean newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	protected void fireStateChanged() {
		if (_changeListener == null) {
			return;

		} else {
			_changeListener.stateChanged(new ChangeEvent(this));
		}
	}

	protected static final String getName(IDeclaration declaration) {
		if (declaration != null)
			return declaration.getName();
		else
			return "<null>";
	}

	protected PropertyChangeSupport getPropertyChange() {
		if (_propertyChange == null) {
			_propertyChange = new PropertyChangeSupport(this);
		}
		return _propertyChange;
	}

	protected RemoveVetoListener getRemovingNotOkListener() {
		if (_removeVetoListener == null) {
			return null;

		} else {
			return RemoveVetoMulticaster.getRemovingNotOkListener(_removeVetoListener);
		}
	}

	public final synchronized boolean hasListeners(String propertyName) {
		return getPropertyChange().hasListeners(propertyName);
	}

	public boolean isomorphicEquals(IIsomorphicHashable anObject) {
		if (!(anObject instanceof IDeclaration)) return false;
		
		if (!((IDeclaration)anObject).getName().equals(getName())) return false;
		
		return true;
	}

	public static final boolean isomorphicEquals(
		IIsomorphicHashable anObject1,
		IIsomorphicHashable anObject2) {
		if ((anObject1 == null) || (anObject2 == null)) {
			return anObject1 == anObject2;

		} else {
			boolean result = anObject1.isomorphicEquals(anObject2);
			Assert.assertEquals(result, anObject2.isomorphicEquals(anObject1));
			return result;
		}
	}

	public static final boolean isomorphicEquals(
		Iterator declarationIterator1,
		Iterator declarationIterator2) {

		if ((declarationIterator1 == null) || (declarationIterator2 == null)) {
			return declarationIterator1 == declarationIterator2;
		} else {
			while (declarationIterator1.hasNext()) {
				IDeclaration currDecl1 = (IDeclaration)declarationIterator1.next();
				if (!declarationIterator2.hasNext()) {
					return false;
				}
				IDeclaration currDecl2 = (IDeclaration)declarationIterator2.next();
				if (!currDecl1.isomorphicEquals(currDecl2)) {
					return false;
				}
			}

			if (declarationIterator2.hasNext()) {
				return false;
			} else {
				return true;
			}
		}
	}

	public int isomorphicHashCode() {
		// by default everything is isomorphic equal
		// should be 42
		return 43;
	}

	public static final int isomorphicHashCode(IIsomorphicHashable anObject) {
		if (anObject == null) {
			return 0;
		} else {
			return anObject.isomorphicHashCode();
		}
	}

	public static final int isomorphicHashCode(boolean anBoolean) {
		return anBoolean ? 1 : 0;
	}

	/**
	 * @see de.uniwue.ki.util.IRemoveable#remove()
	 */
	public void remove() {
		Iterator it = dependentObjects.iterator();
		while (it.hasNext()) {
			IDeclaration element = (IDeclaration)it.next();
			element.removeRemoveVetoListener(vetoListener);
		}

		_changeListener = null;
		_propertyChange = null;
		_removeVetoListener = null;

		/*
		 * Assert.assertTrue( "Removing was not okay for listeners, try calling
		 * #canBeRemoved() before calling #remove()\n" + "Listener: " +
		 * getRemovingNotOkListener(), removingOkForListeners());
		 */
	}

	/**
	 * Calls {@@link IDeclaration#remove()}if <code>declaration</code> is
	 * not <code>null</code>.
	 */
	protected static final void remove(IDeclaration declaration) {
		if (declaration != null)
			declaration.remove();
	}

	/**
	 * Calls {@@link IDeclaration#remove()}if <code>declarationList</code>
	 * is not <code>null</code>.
	 */
	protected static final void remove(IDeclarationList declarationList) {
		if (declarationList != null)
			declarationList.remove();
	}

	public void removeChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.remove(_changeListener, newListener);
	}

	public synchronized void removePropertyChangeListener(PropertyChangeListener listener) {
		getPropertyChange().removePropertyChangeListener(listener);
	}

	public synchronized void removePropertyChangeListener(
		String propertyName,
		PropertyChangeListener listener) {
		getPropertyChange().removePropertyChangeListener(propertyName, listener);
	}

	public void removeRemoveVetoListener(RemoveVetoListener listener) {
		Assert.assertNotNull(listener);

		_removeVetoListener = RemoveVetoMulticaster.remove(_removeVetoListener, listener);
	}

	protected final boolean removingOkForListeners() {
		return (_removeVetoListener == null) || (_removeVetoListener.canBeRemoved());
	}

	/**
	 * @param e
	 *            ChangeEvent
	 */
	public void stateChanged(ChangeEvent e) {
		fireStateChanged();
	}
}