package de.uniwue.ki.mas.util.declaration;
import java.util.Comparator;
import java.util.Iterator;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;
import de.uniwue.ki.util.event.StateChangeMulticaster;
import de.uniwue.ki.util.hash.IIsomorphicHashable;
import de.uniwue.ki.util.list.AbstractMutableListModel;
import de.uniwue.ki.util.list.IMutableListModel;

/**
* Adds default implementations for most methods and has listener support methods.
* Creation date: (03.11.2000 10:22:31)
*
* @author Christoph Oechslein
*/

public abstract class AbstractDeclarationList extends AbstractMutableListModel implements IDeclarationList {
	private transient ChangeListener _changeListener = null;
	private transient DeclarationListListener _declarationListListener = null;

	public boolean accept(Object anObject) {
		return anObject instanceof IDeclaration;
	}

	public void addChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.add(_changeListener, newListener);
	}

	public void addDeclarationListListener(DeclarationListListener newListener) {
		_declarationListListener = DeclarationListEventMulticaster.add(_declarationListListener, newListener);
	}

	public boolean canBeRemoved(IDeclaration declaration) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(contains(declaration));

		if (!isComposition()) {
			return true;
		} else {
			return declaration.canBeRemoved();
		}
	}

	public void clear(boolean withCallingRemove) {
		removeRange(0, size(), withCallingRemove);
	}

	public boolean containsDeclaration(String name) {
		Iterator it = iterator();
		while (it.hasNext()) {
			IDeclaration currentDeclaration = (IDeclaration)it.next();
			if (name.equals(currentDeclaration.getName()))
				return true;
		}
		return false;
	}

	protected void fireDeclarationAdded(IDeclaration declaration, int index) {
		if ((_throwNoEvents == true) || (_declarationListListener == null)) {
			return;
		}

		_declarationListListener.declarationAdded(
			DeclarationListEvent.createAddedDeclarationListEvent(this, declaration, index));
	}

	protected void fireDeclarationRemoved(IDeclaration declaration, int index) {
		if ((_throwNoEvents == true) || (_declarationListListener == null)) {
			return;
		}

		_declarationListListener.declarationRemoved(
			DeclarationListEvent.createRemovedDeclarationListEvent(this, declaration, index));
	}

	protected void fireStateChanged() {
		if ((_throwNoEvents == true) || (_changeListener == null)) {
			return;
		}

		_changeListener.stateChanged(new ChangeEvent(this));
	}

	public void appendName(StringBuffer buff, int currDepth, int maxDepth) {
		appendAllNames(buff, this, currDepth, maxDepth, -1);
	}

	public static void appendAllNames(
		StringBuffer buff,
		IDeclarationList declarationList,
		int currDepth,
		int maxDepth,
		int maxLength) {

		if (currDepth >= maxDepth) {
			buff.append("[...]");

		} else {
			buff.append("[");

			int i = 0;
			Iterator it = declarationList.iterator();
			while (it.hasNext()) {
				IDeclaration currIDeclaration = (IDeclaration)it.next();

				currIDeclaration.appendName(buff, currDepth + 1, maxDepth);
				i++;

				if ((maxLength != -1) && (i >= maxLength)) {
					buff.append(" ...");
					break;
				} else if (it.hasNext()) {
					buff.append(", ");
				}
			}

			buff.append("]");
		}
	}

	public IDeclaration getDeclarationByName(String name) {
		Iterator it = iterator();
		while (it.hasNext()) {
			IDeclaration currentDeclaration = (IDeclaration)it.next();
			if (currentDeclaration.getName().equals(name))
				return currentDeclaration;
		}
		return null;
	}

	public String getName() {
		return "Declaration List";
	}

	public boolean isomorphicEquals(IIsomorphicHashable anObject) {
		if (!(anObject instanceof IDeclarationList)) {
			return false;

		} else {

			IDeclarationList declarationList = (IDeclarationList)anObject;

			if (declarationList.size() != size()) {
				return false;
			} else {
				Iterator it1 = iterator();
				Iterator it2 = declarationList.iterator();

				while (it1.hasNext()) {
					IDeclaration currDeclaration1 = (IDeclaration)it1.next();
					IDeclaration currDeclaration2 = (IDeclaration)it2.next();

					boolean result = currDeclaration1.isomorphicEquals(currDeclaration2);
					Assert.assertEquals(result, currDeclaration2.isomorphicEquals(currDeclaration1));

					if (!result) {
						return false;
					}
				}
				return true;
			}
		}
	}

	public int isomorphicHashCode() {
		int hashCode = 1;
		Iterator i = iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			hashCode =
				31 * hashCode
					+ (obj == null
						? 0
						: ((obj instanceof IIsomorphicHashable)
							? ((IIsomorphicHashable)obj).isomorphicHashCode()
							: obj.hashCode()));
		}
		return hashCode;
	}

	public void moveTo(Object transferedObject, IMutableListModel targetMutableListModel, int newIndex) {
		if (this == targetMutableListModel) {
			changeIndex(transferedObject, newIndex);

		} else {
			int currSize = size();

			int oldIndex = indexOf(transferedObject);
			Assert.assertTrue(oldIndex != -1);

			Assert.assertTrue(newIndex != -1);

			newIndex = Math.max(0, Math.min(currSize, newIndex));
			// should be true
			Assert.assertTrue((0 <= newIndex) && (newIndex <= currSize));

			remove((IDeclaration)transferedObject, false);

			if (newIndex == currSize) {
				targetMutableListModel.add(transferedObject);
			} else {
				targetMutableListModel.add(newIndex, transferedObject);
			}

			fireContentsChanged(this, Math.min(oldIndex, newIndex), Math.max(oldIndex, newIndex));
		}
	}

	public void remove() {
	}

	public Object remove(int index, boolean withCallingRemove) {
		throw new UnsupportedOperationException();
	}

	public void remove(IDeclaration declaration, boolean withCallingRemove) {
		remove(indexOf(declaration), withCallingRemove);
	}

	public void removeChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.remove(_changeListener, newListener);
	}

	public void removeDeclarationListListener(DeclarationListListener newListener) {
		_declarationListListener =
			DeclarationListEventMulticaster.remove(_declarationListListener, newListener);
		return;
	}

	protected void removeRange(int fromIndex, int toIndex, boolean withCallingRemove) {
		int oldSize = size();

		for (int i = 0; i < toIndex - fromIndex; i++) {
			remove(0, withCallingRemove);
		}

		Assert.assertEquals(oldSize - (toIndex - fromIndex), size());
	}

	public void sort(Comparator comparator) {
		boolean oldCompositionSetting = isComposition();
		try {
			setComposition(false);

			super.sort(comparator);

		} finally {
			setComposition(oldCompositionSetting);
		}

		fireStateChanged();
	}
}