package de.uniwue.ki.mas.util.declaration;
import javax.swing.event.*;

import junit.framework.Assert;

/**
 * Delegates most methods to another {@link DeclarationListEvent declaration list}
 * and forward {@link DeclarationListEvent declaration events}.
 * Creation date: (02.11.2000 10:57:02)
 *
 * @author Christoph Oechslein
 */

public abstract class AbstractDelegateDeclarationList
	extends AbstractDeclarationList
	implements IDeclarationList {
	protected final IDeclarationList _delegateDeclarationList;

	/**
	 * For receiving added and removed {@link DeclarationListListener declaration events}.
	 */
	protected final DelegateListener delegateListener = new DelegateListener();

	private class DelegateListener implements ListDataListener, ChangeListener, DeclarationListListener {
		/**
		 * Gets called if the given {@link IDeclaration declaration}
		 * is was removed from a {@link IDeclarationList declaration list}.
		 *
		 * @see #declarationAdded(DeclarationListEvent)
		 */
		public void declarationRemoved(DeclarationListEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.declarationRemoved(e.getDeclaration(), e.getIndex());
			}
		}

		/**
		 * Gets called if the given {@link IDeclaration declaration}
		 * was added to a {@link IDeclarationList declaration list}.
		 *
		 * @param e DeclarationListEvent
		 * @see #declarationRemoved(DeclarationListEvent)
		 */
		public void declarationAdded(DeclarationListEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.declarationAdded(e.getDeclaration(), e.getIndex());
			}
		}

		/**
				 * Invoked when the target of the listener has changed its state.
				 *
				 * @param e  a ChangeEvent object
				 * @see javax.swing.event.ChangeListener#stateChanged(ChangeEvent)
		 */
		public void stateChanged(ChangeEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.fireStateChanged();
			}
		}

		/**
		 * Sent when the contents of the list has changed in a way
		 * that's too complex to characterize with the previous
		 * methods.  Index0 and index1 bracket the change.
		 *
		 * @param e  a ListDataEvent encapuslating the event information
		 * @see javax.swing.event.ListDataListener#contentsChanged(ListDataEvent)
		 */
		public void contentsChanged(ListDataEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.contentsChanged(e.getIndex0(), e.getIndex1());
			}
		}

		/**
		 * Sent after the indices in the index0,index1
		 * interval have been inserted in the data model.
		 * The new interval includes both index0 and index1.
		 *
		 * @param e  a ListDataEvent encapuslating the event information
		 * @see javax.swing.event.ListDataListener#intervalAdded(ListDataEvent)
		 */
		public void intervalAdded(ListDataEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.intervalAdded(e.getIndex0(), e.getIndex1());
			}
		}

		/**
		 * Sent after the indices in the index0,index1 interval
		 * have been removed from the data model.  The interval
		 * includes both index0 and index1.
		 *
		 * @param e  a ListDataEvent encapuslating the event information
		 * @see javax.swing.event.ListDataListener#intervalRemoved(ListDataEvent)
		 */
		public void intervalRemoved(ListDataEvent e) {
			if (e.getSource() == _delegateDeclarationList) {
				AbstractDelegateDeclarationList.this.intervalRemoved(e.getIndex0(), e.getIndex1());
			}
		}
	}

	/**
	 * @param _delegateDeclarationList IDeclarationList
	 */
	public AbstractDelegateDeclarationList(IDeclarationList _delegateDeclarationList) {
		super();

		if (Assert.ASSERTION_ON)
			Assert.assertNotNull(_delegateDeclarationList);

		this._delegateDeclarationList = _delegateDeclarationList;

		_delegateDeclarationList.addListDataListener(delegateListener);
		_delegateDeclarationList.addChangeListener(delegateListener);
		_delegateDeclarationList.addDeclarationListListener(delegateListener);
	}

	/**
	 * @param anObject Object
	 * @return boolean
	 * @see de.uniwue.ki.util.IMutableListModel#accept(Object)
	 */
	public boolean accept(Object anObject) {
		return getDelegateDeclarationList().accept(anObject);
	}

	protected void contentsChanged(int index0, int index1) {
		fireContentsChanged(this, index0, index1);
	}

	protected void declarationAdded(IDeclaration declaration, int index) {
		fireDeclarationAdded(declaration, index);
	}

	protected void declarationRemoved(IDeclaration declaration, int index) {
		fireDeclarationRemoved(declaration, index);
	}

	/**
	 * @return IDeclarationList
	 */
	public final IDeclarationList getDelegateDeclarationList() {
		return _delegateDeclarationList;
	}

	protected void intervalAdded(int index0, int index1) {
		fireIntervalAdded(this, index0, index1);
	}

	protected void intervalRemoved(int index0, int index1) {
		fireIntervalRemoved(this, index0, index1);
	}

	/**
	 * @see de.uniwue.ki.util.IRemoveable#remove()
	 */
	public void remove() {
		if (getDelegateDeclarationList() != null) {
			getDelegateDeclarationList().removeListDataListener(delegateListener);
			getDelegateDeclarationList().removeChangeListener(delegateListener);
		}
		super.remove();
	}
}