package de.uniwue.ki.mas.util.declaration;
import java.util.ArrayList;
import java.util.RandomAccess;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import junit.framework.Assert;

/**
 * Default implementation for {@link IDeclarationList}
 * saves declarations in an {@link ArrayList}.
 * Creation date: (18.05.00 11:23:42)

 * @author Christoph Oechslein
 */

public class DefaultDeclarationList
	extends AbstractDeclarationList
	implements IDeclarationList, ChangeListener, RandomAccess {

	private boolean _compositionLink = true;
	private final ArrayList _delegate;
	private NeverRemoveVetoListener _neverRemoveVetoListener;

	public DefaultDeclarationList() {
		this(new ArrayList());
	}

	public DefaultDeclarationList(int size) {
		this(new ArrayList(size));
	}

	public DefaultDeclarationList(IDeclarationList declarationList) {
		this();

		addAll(declarationList);
	}

	private DefaultDeclarationList(ArrayList delegate) {
		super();

		_delegate = delegate;
	}

	public void add(int index, Object anObject) {
		IDeclaration declaration = (IDeclaration)anObject;

		Assert.assertTrue("Could not accept " + declaration, accept(declaration));

		_delegate.add(index, declaration);
		addListener(declaration);
		fireIntervalAdded(this, index, index);
		fireDeclarationAdded(declaration, index);
		fireStateChanged();
	}

	private void addListener(IDeclaration declaration) {
		declaration.addChangeListener(this);

		// make no sense since because of IDeclaration#canBeRemove return false
		/*
		if (isComposition()) {
		   declaration.addRemoveVetoListener(getNeverRemoveVetoListener());
		}
		*/
	}

	public boolean canBeRemoved() {
		return true;
	}

	public int changeIndex(Object anObject, int newIndex) {
		int oldIndex = indexOf(anObject);

		if (oldIndex == newIndex)
			return oldIndex;

		if (Assert.ASSERTION_ON) {
			Assert.assertTrue(oldIndex != -1);
			Assert.assertTrue((0 <= newIndex) && (newIndex <= size()));
		}

		int currSize = size();

		_delegate.remove(oldIndex);

		if (newIndex == currSize) {
			_delegate.add(anObject);

		} else {
			_delegate.add(newIndex, anObject);
		}

		fireContentsChanged(this, Math.min(oldIndex, newIndex), Math.max(oldIndex, newIndex));
		fireStateChanged();

		return oldIndex;
	}

	public Object get(int index) {
		return _delegate.get(index);
	}

	protected synchronized NeverRemoveVetoListener getNeverRemoveVetoListener() {
		if (_neverRemoveVetoListener == null) {
			_neverRemoveVetoListener = new NeverRemoveVetoListener(this);
		}

		return _neverRemoveVetoListener;
	}

	public boolean isComposition() {
		return _compositionLink;
	}

	public void remove() {
		// going to delete elements
		int amount = size();
		for (int index = 0; index < amount; index++) {
			// always remove the first one
			removeIntern(0, true);
		}

		if (Assert.ASSERTION_ON) {
			Assert.assertEquals(0, size());
		}
	}

	public Object remove(int index) {
		return removeIntern(index, true);
	}

	public final Object remove(int index, boolean withCallingRemove) {
		return removeIntern(index, withCallingRemove);
	}

	private Object removeIntern(int index, boolean withCallingRemove) {
		IDeclaration oldDeclaration = (IDeclaration)get(index);

		_delegate.remove(index);
		removeListener(oldDeclaration);

		fireIntervalRemoved(this, index, index);
		fireDeclarationRemoved(oldDeclaration, index);
		fireStateChanged();

		if (withCallingRemove && isComposition())
			oldDeclaration.remove();

		return oldDeclaration;
	}

	private void removeListener(IDeclaration declaration) {
		declaration.removeChangeListener(this);

		// make no sense since because of IDeclaration#canBeRemove return false
		/*
		if (isComposition()) {
		   declaration.removeRemoveVetoListener(getNeverRemoveVetoListener());
		}
		*/
	}

	public Object set(int index, Object anObject) {
		Assert.assertTrue("Could not accept " + anObject, accept(anObject));

		IDeclaration newDeclaration = (IDeclaration)anObject;
		IDeclaration oldDeclaration = (IDeclaration)get(index);

		if (oldDeclaration == newDeclaration) {
			changeIndex(oldDeclaration, index);
			return oldDeclaration;

		} else {
			removeListener(oldDeclaration);
			_delegate.set(index, newDeclaration);

			if (isComposition())
				oldDeclaration.remove();

			addListener(newDeclaration);

			fireContentsChanged(this, index, index);
			fireDeclarationRemoved(oldDeclaration, index);
			fireDeclarationAdded(newDeclaration, index);
			fireStateChanged();

			return oldDeclaration;
		}
	}

	public void setComposition(boolean newIsComposition) {
		_compositionLink = newIsComposition;
	}

	public int size() {
		return _delegate.size();
	}

	public void stateChanged(ChangeEvent evt) {
		fireStateChanged();
	}
}