package de.uniwue.ki.mas.util.declaration;
import java.util.Iterator;

import javax.swing.event.*;

import junit.framework.Assert;
/**
* View two declaration lists as one list (read only).
* Creation date: (18.12.2000 11:44:09)
*
* @author Christoph Oechslein
*/
public class TwoUnmodifiableDeclarationList extends AbstractDeclarationList {
	private IDeclarationList _declarationListA;
	private IDeclarationList _declarationListB;

	private boolean _composition = false;

	private final EventListener _eventListener = new EventListener();

	private class EventListener implements ChangeListener, ListDataListener, DeclarationListListener {
		public void stateChanged(ChangeEvent evt) {
			if (evt.getSource() == getDeclarationListA() || evt.getSource() == getDeclarationListB()) {
				fireStateChanged();
			}
		}

		public void declarationAdded(DeclarationListEvent evt) {
			if (evt.getSource() == getDeclarationListA()) {
				fireDeclarationAdded(evt.getDeclaration(), evt.getIndex());
			} else if (evt.getSource() == getDeclarationListB()) {
				fireDeclarationAdded(evt.getDeclaration(), evt.getIndex() + getDeclarationListA().size());
			}
		}

		public void declarationRemoved(DeclarationListEvent evt) {
			if (evt.getSource() == getDeclarationListA()) {
				fireDeclarationRemoved(evt.getDeclaration(), evt.getIndex());
			} else if (evt.getSource() == getDeclarationListB()) {
				fireDeclarationRemoved(evt.getDeclaration(), evt.getIndex() + getDeclarationListA().size());
			}
		}

		public void intervalAdded(ListDataEvent evt) {
			if (evt.getSource() == getDeclarationListA()) {
				fireIntervalAdded(TwoUnmodifiableDeclarationList.this, evt.getIndex0(), evt.getIndex1());
			} else if (evt.getSource() == getDeclarationListB()) {
				fireIntervalAdded(
					TwoUnmodifiableDeclarationList.this,
					evt.getIndex0() + getDeclarationListA().size(),
					evt.getIndex1() + getDeclarationListA().size());
			}
		}

		public void intervalRemoved(ListDataEvent evt) {
			if (evt.getSource() == getDeclarationListA()) {
				fireIntervalRemoved(TwoUnmodifiableDeclarationList.this, evt.getIndex0(), evt.getIndex1());
			} else if (evt.getSource() == getDeclarationListB()) {
				fireIntervalRemoved(
					TwoUnmodifiableDeclarationList.this,
					evt.getIndex0() + getDeclarationListA().size(),
					evt.getIndex1() + getDeclarationListA().size());
			}
		}

		public void contentsChanged(ListDataEvent evt) {
			if (evt.getSource() == getDeclarationListA()) {
				fireContentsChanged(TwoUnmodifiableDeclarationList.this, evt.getIndex0(), evt.getIndex1());
			} else if (evt.getSource() == getDeclarationListB()) {
				fireContentsChanged(
					TwoUnmodifiableDeclarationList.this,
					evt.getIndex0() + getDeclarationListA().size(),
					evt.getIndex1() + getDeclarationListA().size());
			}
		}
	}

	public TwoUnmodifiableDeclarationList(
		IDeclarationList declarationListA,
		IDeclarationList declarationListB,
		boolean composition) {
		super();

		Assert.assertNotNull(declarationListA);
		Assert.assertNotNull(declarationListB);

		setDeclarationListA(declarationListA);
		setDeclarationListB(declarationListB);

		setComposition(composition);
	}

	public boolean isComposition() {
		return _composition;
	}

	public void setComposition(boolean composition) {
		_composition = composition;
	}

	private void addListener(IDeclarationList declarationList) {
		if (declarationList != null) {
			declarationList.addChangeListener(_eventListener);
			declarationList.addDeclarationListListener(_eventListener);
			declarationList.addListDataListener(_eventListener);
			Iterator it = declarationList.iterator();
			int index = 0;
			while (it.hasNext()) {
				fireDeclarationAdded((IDeclaration) it.next(), index++);
			}
		}
	}

	protected void fireContentsChanged(Object source, int index0, int index1) {
		super.fireContentsChanged(source, index0, index1);
	}

	protected void fireDeclarationAdded(IDeclaration declaration, int index) {
		super.fireDeclarationAdded(declaration, index);
	}

	protected void fireDeclarationRemoved(IDeclaration declaration, int index) {
		super.fireDeclarationRemoved(declaration, index);
	}

	protected void fireIntervalAdded(Object source, int index0, int index1) {
		super.fireIntervalAdded(source, index0, index1);
	}

	protected void fireIntervalRemoved(Object source, int index0, int index1) {
		super.fireIntervalRemoved(source, index0, index1);
	}

	protected void fireStateChanged() {
		super.fireStateChanged();
	}

	public Object get(int index) {
		Assert.assertNotNull(getDeclarationListA());
		Assert.assertNotNull(getDeclarationListB());

		if (getDeclarationListA() != null && index < getDeclarationListA().size())
			return getDeclarationListA().get(index);

		else if (getDeclarationListB() != null)
			return getDeclarationListB().get(index - getDeclarationListA().size());

		else
			return null;
	}

	public IDeclarationList getDeclarationListA() {
		return _declarationListA;
	}

	public IDeclarationList getDeclarationListB() {
		return _declarationListB;
	}

	public void remove() {
		setDeclarationListA(null);
		setDeclarationListB(null);

		super.remove();
	}

	private void removeListener(IDeclarationList declarationList) {
		if (declarationList != null) {
			declarationList.removeChangeListener(_eventListener);
			declarationList.removeDeclarationListListener(_eventListener);
			declarationList.removeListDataListener(_eventListener);
			Iterator it = declarationList.iterator();
			int index = 0;
			while (it.hasNext()) {
				fireDeclarationRemoved((IDeclaration) it.next(), index++);
			}

			if (isComposition()) {
				declarationList.remove();
			}
		}
	}

	public void setDeclarationListA(IDeclarationList newDeclarationListA) {
		IDeclarationList oldDeclarationListA = getDeclarationListA();

		if (oldDeclarationListA != newDeclarationListA) {
			removeListener(oldDeclarationListA);

			_declarationListA = newDeclarationListA;

			if (newDeclarationListA != null) {
				addListener(newDeclarationListA);

				fireContentsChanged(this, 0, getSize());
				fireStateChanged();
			}
		}
	}

	public void setDeclarationListB(IDeclarationList newDeclarationListB) {
		IDeclarationList oldDeclarationListB = getDeclarationListB();

		if (oldDeclarationListB != newDeclarationListB) {
			removeListener(oldDeclarationListB);

			_declarationListB = newDeclarationListB;

			if (newDeclarationListB != null) {
				addListener(newDeclarationListB);

				fireContentsChanged(this, 0, getSize());
				fireStateChanged();
			}
		}
	}

	public int size() {
		int sizeA = 0;
		if (getDeclarationListA() != null)
			sizeA = getDeclarationListA().size();

		int sizeB = 0;
		if (getDeclarationListB() != null)
			sizeB = getDeclarationListB().size();

		return sizeA + sizeB;
	}
}