package de.uniwue.ki.mas.xml.reader;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;

import junit.framework.Assert;
import de.uniwue.ki.mas.basics.feature.IFeatureClassDeclaration;
import de.uniwue.ki.mas.basics.feature.spatial.*;
import de.uniwue.ki.mas.basics.feature.spatial.position.Position2DType;
import de.uniwue.ki.mas.basics.function.ExecuteIteratorFactory;
import de.uniwue.ki.mas.basics.functionArgument.*;
import de.uniwue.ki.mas.basics.functionDeclaration.list.IFunctionDeclarationListAble;
import de.uniwue.ki.mas.basics.type.*;
import de.uniwue.ki.mas.basics.type.argument.ArgumentDeclarationFactory;
import de.uniwue.ki.mas.basics.type.argument.IArgumentDeclaration;
import de.uniwue.ki.mas.basics.type.composed.*;
import de.uniwue.ki.mas.basics.type.enum.ISpecificEnumType;
import de.uniwue.ki.mas.basics.type.enum.SpecificEnumType;
import de.uniwue.ki.mas.basics.type.user.IUserTypeDeclaration;
import de.uniwue.ki.mas.behaviour.activityGraphDeclaration.IActivityGraphNodeDeclaration;
import de.uniwue.ki.mas.entity.objectclass.IObjectClassDeclaration;
import de.uniwue.ki.mas.entity.variableclass.IBodyVariableClassDeclaration;
import de.uniwue.ki.mas.entity.variableclass.IVariableClassDeclaration;
import de.uniwue.ki.mas.sim.simulation.ISimulationDeclaration;
import de.uniwue.ki.mas.xml.castor.XMLRootObject;
import de.uniwue.ki.mas.xml.castor.v01.model.*;
import de.uniwue.ki.util.Position2D;
import de.uniwue.ki.util.Utils;

/**
  * Creation date: (07.02.2001 09:54:13)
  * @author Christoph Oechslein
  */
public class XMLModelTypeReader {

	private final transient XMLModelReader _xmlModelReader;
	private final IFunctionDeclarationListAble _functionDeclarationList;

	private static final List _xmlTypeReaders = new LinkedList();

	public static void addTypeReader(IXMLTypeReader typeReader) {
		_xmlTypeReaders.add(typeReader);
	}

	public XMLModelTypeReader(
		XMLModelReader xmlModelReader,
		IFunctionDeclarationListAble functionDeclarationList) {
		super();
		_xmlModelReader = xmlModelReader;
		_functionDeclarationList = functionDeclarationList;
	}

	File createRealAbsolutePath(String fileString) {
		if (fileString == null)
			return null;

		File file = extractFile(fileString);

		if (file.isAbsolute()) {
			return file;

		} else {
			File rootFile = _xmlModelReader.getSeSAmDeclaration().getFile();

			if (rootFile == null) {
				return file.getAbsoluteFile();

			} else {
				File absolutFile = new File(rootFile.getParentFile(), fileString).getAbsoluteFile();
				Assert.assertTrue(absolutFile.isAbsolute());

				return absolutFile;
			}
		}
	}

	private File extractFile(String fileString) {
		try {
			return new File(new URI(fileString));

		} catch (URISyntaxException e) {
			// upward compatibiliy
			return new File(fileString);

		} catch (IllegalArgumentException e) {
			// upward compatibiliy
			return new File(fileString);
		}
	}

	public IArgumentDeclaration createRealArgumentDeclaration(XMLTypeValueGroup xmlTypeValueGroup)
		throws XMLReadException {

		if (xmlTypeValueGroup.getParameterID() != null) {

			XMLFunctionParameterID xmlFunctionParameterID = xmlTypeValueGroup.getParameterID();
			XMLFunctionParameter xmlFunctionParameter = (XMLFunctionParameter) xmlFunctionParameterID.getId();

			return (IFunctionArgumentDeclaration) _xmlModelReader.getDeclaration(xmlFunctionParameter);

		} else {
			ISpecificType type = createRealType(xmlTypeValueGroup);
			Object value = createRealTypeValue(xmlTypeValueGroup);

			// due to backward compatibility
			if ((type instanceof SpecificVariableClassDeclarationType)
				&& (_xmlModelReader.getVersion() <= 0.1)) {

				SpecificVariableClassDeclarationType variableClassDeclarationType =
					(SpecificVariableClassDeclarationType) type;

				if (Utils
					.equals(
						variableClassDeclarationType.getVariableClassDeclarationType().getGenericType(),
						GenericIteratorType.getInstance())) {

					SpecificIteratorType specificIteratorType =
						(SpecificIteratorType) variableClassDeclarationType.getVariableClassDeclarationType();
					type =
						GenericVariableClassDeclarationType
							.getInstance(
								variableClassDeclarationType.isExternal(),
								variableClassDeclarationType.isWriteable())
							.createSpecificVariableType(
								GenericListType.getInstance().createSpecificListType(
									specificIteratorType.getIteratorSpecificType()));
				}
			}

			// due to backward compatibility
			if ((_xmlModelReader.getVersion() <= 0.1)
				&& DoubleType.getInstance().equals(type)
				&& (((Number) value).intValue() == ((Number) value).doubleValue())) {
				type = IntegerType.getInstance();
				value = new Integer(((Number) value).intValue());
			}

			Assert.assertNotNull(type);
			return ArgumentDeclarationFactory.getInstance().createArgumentDeclaration(type, value);
		}
	}

	private IFunctionArgumentDeclaration createRealFunctionArgumentDeclaration(XMLFunctionParameter xmlFunctionParameter)
		throws XMLReadException {

		ISpecificType argumentType = createRealType(xmlFunctionParameter.getXMLTypeGroup());
		String name = (xmlFunctionParameter.getName() == null) ? "Untitled" : xmlFunctionParameter.getName();

		//		if ((_xmlModelReader.getVersion() <= 0.1) && (SimulationType.getInstance().equals(argumentType))) {
		//			return FunctionArgumentDeclarationFactory.getInstance().createArgumentDeclaration(
		//				name,
		//				SimulationDeclarationType.getInstance());
		//		} else {
		return FunctionArgumentDeclarationFactory.getInstance().createArgumentDeclaration(name, argumentType);
		//		}
	}

	public CircularSpatialInfo createRealSpatialInfo(XMLSpatialInfo xmlSpatialInfo) throws XMLReadException {

		CircularSpatialInfo spatialInfo = new CircularSpatialInfo(null);
		spatialInfo.getPosition().x = xmlSpatialInfo.getX();
		spatialInfo.getPosition().y = xmlSpatialInfo.getY();
		spatialInfo.setDirection(xmlSpatialInfo.getDirection());
		spatialInfo.setSpeed(xmlSpatialInfo.getSpeed());
		spatialInfo.setRadius(xmlSpatialInfo.getRadius());
		spatialInfo.setDrawPriority(xmlSpatialInfo.getDrawPriority());

		if (xmlSpatialInfo.getImagePath() != null) {
			spatialInfo.setImagePath(createRealAbsolutePath(xmlSpatialInfo.getImagePath()));
		}

		return spatialInfo;
	}

	public ISpecificType createRealType(XMLTypeGroup xmlTypeGroup) throws XMLReadException {

		if (xmlTypeGroup.getVoidType() != null) {

			return VoidType.getInstance();
		}

		if (xmlTypeGroup.getBooleanType() != null) {

			return BooleanType.getInstance();
		}

		if (xmlTypeGroup.getObjectType() != null) {

			return ObjectType.getInstance();
		}

		if (xmlTypeGroup.getNumberType() != null) {

			return DoubleType.getInstance();
		}

		if (xmlTypeGroup.getDoubleType() != null) {

			return DoubleType.getInstance();
		}

		if (xmlTypeGroup.getIntegerType() != null) {

			return IntegerType.getInstance();
		}

		if (xmlTypeGroup.getSpatialInfoType() != null) {

			return SpatialInfoType.getInstance();
		}

		if (xmlTypeGroup.getPosition2DType() != null) {

			return Position2DType.getInstance();
		}

		if (xmlTypeGroup.getStringType() != null) {

			return StringType.getInstance();
		}

		if (xmlTypeGroup.getImageType() != null) {

			return ImageType.getInstance();
		}

		if (xmlTypeGroup.getActivityType() != null) {

			return ActivityType.getInstance();
		}

		if (xmlTypeGroup.getObjectClassType() != null) {

			return ObjectClassDeclarationType.getInstance();
		}

		if (xmlTypeGroup.getFeatureClassType() != null) {

			return FeatureClassDeclarationType.getInstance();
		}

		if (xmlTypeGroup.getObjectInstanceType() != null) {
			return ObjectInstanceDeclarationType.getInstance();
		}

		// backward compatibility
		if (xmlTypeGroup.getSetType() != null) {

			return GenericIteratorType.getInstance().createSpecificIteratorType(
				createRealType(xmlTypeGroup.getSetType().getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getIteratorType() != null) {

			return GenericIteratorType.getInstance().createSpecificIteratorType(
				createRealType(xmlTypeGroup.getIteratorType().getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getListType() != null) {

			return GenericListType.getInstance().createSpecificListType(
				createRealType(xmlTypeGroup.getListType().getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getHashtableType() != null) {

			return GenericHashtableType.getInstance().createSpecificHashtableType(
				createRealType(xmlTypeGroup.getHashtableType().getKeyType().getXMLTypeGroup()),
				createRealType(xmlTypeGroup.getHashtableType().getValueType().getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getVariableType() != null) {

			XMLVariableType xmlVariableType = xmlTypeGroup.getVariableType();

			return GenericVariableClassDeclarationType
				.getInstance(xmlVariableType.getExternal(), xmlVariableType.getWriteable())
				.createSpecificVariableType(createRealType(xmlVariableType.getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getFunctionArgRestType() != null) {

			XMLFunctionArgRestType xmlFunctionArgRestType = xmlTypeGroup.getFunctionArgRestType();

			return GenericFunctionType.getInstance().createSpecificFunctionType(
				createRealType(xmlFunctionArgRestType.getXMLTypeGroup()),
				createRealType(xmlFunctionArgRestType.getInputType().getXMLTypeGroup()));
		}

		if (xmlTypeGroup.getFunctionArgListType() != null) {

			XMLFunctionArgListType xmlFunctionArgListType = xmlTypeGroup.getFunctionArgListType();
			XMLFunctionArgListInputTypes xmlFunctionArgListInputTypes =
				xmlFunctionArgListType.getInputTypes();
			int size = xmlFunctionArgListInputTypes.getXMLTypeGroupCount();
			List inputTypes = new ArrayList(size);

			for (int i = 0; i < size; i++) {
				inputTypes.add(createRealType(xmlFunctionArgListInputTypes.getXMLTypeGroup(i)));
			}

			return GenericFunctionType.getInstance().createSpecificFunctionType(
				createRealType(xmlFunctionArgListType.getXMLTypeGroup()),
				inputTypes);
		}

		// backward compatibility
		if (xmlTypeGroup.getBodyType() != null) {
			return GenericSimObjectType.getGeneralSimObjectType();
		}

		if (xmlTypeGroup.getSimObjectType() != null) {
			return getSimObjectType((XMLRootObject) xmlTypeGroup.getSimObjectType().getId());
		}

		if (xmlTypeGroup.getSimulationType() != null) {
			//			if (_xmlModelReader.getVersion() <= 0.1)
			//				return SimulationType.getInstance();
			//			else
			return SimulationDeclarationType.getInstance();
		}

		if (xmlTypeGroup.getSimulationRunType() != null) {
			return SimulationType.getInstance();
		}

		if (xmlTypeGroup.getUserType() != null) {
			XMLUserTypeID userTypeID = xmlTypeGroup.getUserType();
		
			return _xmlModelReader.getSpecificTypeByID(userTypeID);
		}

		// backward compatibility
		if (xmlTypeGroup.getComposedType() != null) {
			IUserTypeDeclaration userTypeDeclaration =
				_xmlModelReader.getRealUserTypeDeclaration(
					(XMLUserType) xmlTypeGroup.getComposedType().getId());

			return userTypeDeclaration.getSpecificType();
		}

		if (xmlTypeGroup.getComposedTypeAttributeType() != null) {
			XMLComposedTypeAttributeType xmlComposedAttrType = xmlTypeGroup.getComposedTypeAttributeType();

			SpecificComposedType composedType;

			if ((XMLUserType) xmlComposedAttrType.getComposedTypeID() != null)
				composedType =
					_xmlModelReader
						.getRealUserTypeDeclaration((XMLUserType) xmlComposedAttrType.getComposedTypeID())
						.getSpecificComposedType();
			else
				return (ISpecificComposedType) TypeManager.getSpecificTypeByName(
					xmlComposedAttrType.getComposedTypeName(),
					TypeManager.getInstance().getSystemComposedTypes());

			GenericComposedTypeAttributeType genericComposedTypeAttributeType;

			if (!xmlComposedAttrType.hasWriteable() || !xmlComposedAttrType.getWriteable())
				genericComposedTypeAttributeType = GenericComposedTypeAttributeType.getInstance();
			else
				genericComposedTypeAttributeType = GenericComposedTypeAttributeType.getInstanceWriteable();

			return genericComposedTypeAttributeType.createSpecificComposedTypeAttributeType(
				composedType.getAttributeAt(xmlComposedAttrType.getIndex()),
				composedType);
		}

		if (xmlTypeGroup.getUnknownType() != null) {

			// xmlTypeGroup.getUnknownType().getSettings();
			String typeName = xmlTypeGroup.getUnknownType().getName();
			ISpecificType unknownType = TypeManager.getInstance().getSpecificTypeByName(typeName);

			if (unknownType == null)
				throw new XMLReadException("Not able to find unknown type with name " + typeName);

			return unknownType;
		}

		Assert.fail("Unknown xmlTypeGroup: " + xmlTypeGroup);

		return null;
	}

	public ISpecificType createRealType(XMLTypeValueGroup xmlTypeValueGroup) throws XMLReadException {

		if (xmlTypeValueGroup.getVoid() != null) {

			return VoidType.getInstance();
		}

		if (xmlTypeValueGroup.getBoolean() != null) {

			return BooleanType.getInstance();
		}

		if (xmlTypeValueGroup.getNumber() != null) {

			return DoubleType.getInstance();
		}

		if (xmlTypeValueGroup.getDouble() != null) {

			return DoubleType.getInstance();
		}

		if (xmlTypeValueGroup.getInteger() != null) {

			return IntegerType.getInstance();
		}

		if (xmlTypeValueGroup.getSpatialInfo() != null) {

			return SpatialInfoType.getInstance();
		}

		if (xmlTypeValueGroup.getPosition2D() != null) {

			return Position2DType.getInstance();
		}

		if (xmlTypeValueGroup.getString() != null) {

			return StringType.getInstance();
		}

		if (xmlTypeValueGroup.getImage() != null) {

			return ImageType.getInstance();
		}

		if ((xmlTypeValueGroup.getObjectID() != null) || (xmlTypeValueGroup.getObject() != null)) {

			return ObjectType.getInstance();
		}

		if (xmlTypeValueGroup.getActivityID() != null) {

			return ActivityType.getInstance();
		}

		if ((xmlTypeValueGroup.getSimulationID() != null) || (xmlTypeValueGroup.getSimulation() != null)) {
			//			if (_xmlModelReader.getVersion() <= 0.1)
			//				return SimulationType.getInstance();
			//			else
			return SimulationDeclarationType.getInstance();
		}

		if (xmlTypeValueGroup.getSimulationRun() != null) {
			return SimulationType.getInstance();
		}

		if (xmlTypeValueGroup.getObjectClassID() != null) {
			return ObjectClassDeclarationType.getInstance();
		}

		if (xmlTypeValueGroup.getFeatureClassID() != null) {
			return FeatureClassDeclarationType.getInstance();
		}

		if (xmlTypeValueGroup.getObjectInstance() != null) {
			return ObjectInstanceDeclarationType.getInstance();
		}

		if (xmlTypeValueGroup.getObjectInstanceID() != null) {
			return ObjectInstanceDeclarationType.getInstance();
		}

		// backward compatibility
		if (xmlTypeValueGroup.getBody() != null) {
			return GenericSimObjectType.getGeneralSimObjectType();
		}

		if (xmlTypeValueGroup.getSimObject() != null) {
			return getSimObjectType((XMLRootObject) xmlTypeValueGroup.getSimObject().getId());
		}

		// backward compatibility
		if (xmlTypeValueGroup.getSet() != null) {
			return GenericIteratorType.getInstance().createSpecificIteratorType(
				createRealType(xmlTypeValueGroup.getSet().getXMLTypeGroup()));
		}

		if (xmlTypeValueGroup.getIterator() != null) {
			return GenericIteratorType.getInstance().createSpecificIteratorType(
				createRealType(xmlTypeValueGroup.getIterator().getXMLTypeGroup()));
		}

		if (xmlTypeValueGroup.getList() != null) {

			return GenericListType.getInstance().createSpecificListType(
				createRealType(xmlTypeValueGroup.getList().getXMLTypeGroup()));
		}

		if (xmlTypeValueGroup.getHashtable() != null) {

			return GenericHashtableType.getInstance().createSpecificHashtableType(
				createRealType(xmlTypeValueGroup.getHashtable().getKeyType().getXMLTypeGroup()),
				createRealType(xmlTypeValueGroup.getHashtable().getValueType().getXMLTypeGroup()));
		}

		if (xmlTypeValueGroup.getVariableID() != null) {

			return createSpecificVariableType(xmlTypeValueGroup.getVariableID());
		}

		if (xmlTypeValueGroup.getParametricFunction() != null) {

			XMLUnnamedParametricFunction xmlUnnamedParametricFunction =
				xmlTypeValueGroup.getParametricFunction();
			ISpecificType outputType =
				_xmlModelReader.createRealType(
					xmlUnnamedParametricFunction.getFunctionCall(),
					_functionDeclarationList);
			List inputsArgs = new ArrayList();

			for (int i = 0; i < xmlUnnamedParametricFunction.getParameterCount(); i++) {

				XMLFunctionParameter xmlFunctionParameter = xmlUnnamedParametricFunction.getParameter(i);
				inputsArgs.add(createRealType(xmlFunctionParameter.getXMLTypeGroup()));
			}

			return GenericFunctionType.getInstance().createSpecificFunctionType(outputType, inputsArgs);
		}

		if (xmlTypeValueGroup.getFunction() != null) {

			XMLFunction xmlFunction = xmlTypeValueGroup.getFunction();

			return createRealType(xmlFunction.getXMLTypeGroup());
		}

		if (xmlTypeValueGroup.getParameterID() != null) {

			XMLFunctionParameter xmlFunctionParameter =
				(XMLFunctionParameter) xmlTypeValueGroup.getParameterID().getId();

			return getFunctionArgumentDeclaration(xmlFunctionParameter).getSpecificType();
		}

		if (xmlTypeValueGroup.getComposed() != null) {
			if ((XMLUserType) xmlTypeValueGroup.getComposed().getComposedTypeID() != null)
				return _xmlModelReader
					.getRealUserTypeDeclaration(
						(XMLUserType) xmlTypeValueGroup.getComposed().getComposedTypeID())
					.getSpecificComposedType();
			else
				return (ISpecificComposedType) TypeManager.getSpecificTypeByName(
					xmlTypeValueGroup.getComposed().getComposedTypeName(),
					TypeManager.getInstance().getSystemComposedTypes());

		}

		if (xmlTypeValueGroup.getComposedAttribute() != null) {
			XMLComposedTypeAttributeValue xmlComposedAttr = xmlTypeValueGroup.getComposedAttribute();

			return createSpecificComposedTypeAttributeType(xmlComposedAttr);
		}

		// backward compatibility
		if (xmlTypeValueGroup.getComposedTypeAttribute() != null) {
			XMLComposedTypeAttributeValue xmlComposedAttr = xmlTypeValueGroup.getComposedTypeAttribute();

			return createSpecificComposedTypeAttributeType(xmlComposedAttr);
		}

		if (xmlTypeValueGroup.getEnum() != null) {
			//look if this is a system defined enumeration
			if (xmlTypeValueGroup.getEnum().getEnumTypeName() != null) {
				return TypeManager.getSpecificTypeByName(
					xmlTypeValueGroup.getEnum().getEnumTypeName(),
					TypeManager.getInstance().getComplexTypes());
			}

			//model defined enumeration
			SpecificEnumType specificEnumType =
				_xmlModelReader
					.getRealUserTypeDeclaration((XMLUserType) xmlTypeValueGroup.getEnum().getEnumTypeID())
					.getSpecificEnumType();

			return specificEnumType;
		}

		Iterator it = _xmlTypeReaders.iterator();
		while (it.hasNext()) {
			IXMLTypeReader typeReader = (IXMLTypeReader) it.next();
			if (typeReader.isResponsibleForValue(xmlTypeValueGroup)) {
				return typeReader.createSpecificTypeFromValueGroup(xmlTypeValueGroup);
			}
		}

		if (xmlTypeValueGroup.getUnknown() != null) {

			// xmlTypeValueGroup.getUnknown().getSettings();
			// xmlTypeValueGroup.getUnknown().getUnknownType().getSettings();
			return TypeManager.getInstance().getSpecificTypeByName(
				xmlTypeValueGroup.getUnknown().getUnknownType().getName());
		}

		Assert.fail("Unknown xmlTypeValueGroup: " + xmlTypeValueGroup);

		return null;
	}

	private ISpecificType getSimObjectType(XMLRootObject id) {
		if (id == null) {
			return GenericSimObjectType.getGeneralSimObjectType();

		} else {
			return new DelayedSimObjectType(id, _xmlModelReader);
		}
	}

	private ISpecificType createSpecificComposedTypeAttributeType(XMLComposedTypeAttributeValue xmlComposedAttr)
		throws XMLReadException {
		ISpecificComposedType composedType;

		if ((XMLUserType) xmlComposedAttr.getComposedTypeID() != null)
			composedType =
				_xmlModelReader
					.getRealUserTypeDeclaration((XMLUserType) xmlComposedAttr.getComposedTypeID())
					.getSpecificComposedType();
		else
			composedType =
				(ISpecificComposedType) TypeManager.getSpecificTypeByName(
					xmlComposedAttr.getComposedTypeName(),
					TypeManager.getInstance().getSystemComposedTypes());

		GenericComposedTypeAttributeType genericComposedTypeAttributeType;

		if (!xmlComposedAttr.hasWriteable() || !xmlComposedAttr.getWriteable())
			genericComposedTypeAttributeType = GenericComposedTypeAttributeType.getInstance();
		else
			genericComposedTypeAttributeType = GenericComposedTypeAttributeType.getInstanceWriteable();

		return genericComposedTypeAttributeType.createSpecificComposedTypeAttributeType(
			composedType.getAttributeAt(xmlComposedAttr.getIndex()),
			composedType);
	}

	private Object createRealTypeValue(XMLTypeValueGroup xmlTypeValueGroup) throws XMLReadException {

		if (xmlTypeValueGroup.getBoolean() != null) {

			return Boolean.valueOf(xmlTypeValueGroup.getBoolean().getValue());
		}

		if (xmlTypeValueGroup.getNumber() != null) {

			return new Double(xmlTypeValueGroup.getNumber().getValue());
		}

		if (xmlTypeValueGroup.getDouble() != null) {

			return new Double(xmlTypeValueGroup.getDouble().getValue());
		}

		if (xmlTypeValueGroup.getInteger() != null) {

			return new Integer(xmlTypeValueGroup.getInteger().getValue());
		}

		if (xmlTypeValueGroup.getString() != null) {

			return xmlTypeValueGroup.getString().getValue();
		}

		if (xmlTypeValueGroup.getFile() != null) {

			String filePath = xmlTypeValueGroup.getFile().getValue();

			if (filePath == null) {
				return null;

			} else {
				return createRealAbsolutePath(filePath);
			}
		}

		if (xmlTypeValueGroup.getImage() != null) {

			return createRealAbsolutePath(xmlTypeValueGroup.getImage().getValue());
		}

		if (xmlTypeValueGroup.getSpatialInfo() != null) {

			return createRealSpatialInfo(xmlTypeValueGroup.getSpatialInfo());
		}

		if (xmlTypeValueGroup.getPosition2D() != null) {

			XMLPosition2D xmlPosition2D = xmlTypeValueGroup.getPosition2D();

			return new Position2D(xmlPosition2D.getX(), xmlPosition2D.getY());
		}

		if (xmlTypeValueGroup.getObject() != null) {

			// can only save null values
			return null;
		}

		if (xmlTypeValueGroup.getObjectID() != null) {

			return _xmlModelReader.getDeclaration((XMLRootObject) xmlTypeValueGroup.getObjectID().getId());
		}

		if (xmlTypeValueGroup.getActivityID() != null) {
			if (xmlTypeValueGroup.getActivityID().getId() == null) {
				return null;

			} else {
				return (IActivityGraphNodeDeclaration) _xmlModelReader.getDeclaration(
					(XMLRootObject) xmlTypeValueGroup.getActivityID().getId());
			}
		}

		if (xmlTypeValueGroup.getSimulationID() != null) {
			if (xmlTypeValueGroup.getSimulationID().getId() == null) {
				return null;

			} else {
				return (ISimulationDeclaration) _xmlModelReader.getDeclaration(
					(XMLRootObject) xmlTypeValueGroup.getSimulationID().getId());
			}
		}

		if (xmlTypeValueGroup.getSimulation() != null) {
			return _xmlModelReader.createRealSimulationWithNameFunctionCallDeclDeclaration(
				xmlTypeValueGroup.getSimulation());
		}

		if (xmlTypeValueGroup.getSimulationRun() != null) {
			return null;
		}

		if (xmlTypeValueGroup.getObjectClassID() != null) {

			return (IObjectClassDeclaration) _xmlModelReader.getDeclaration(
				(XMLRootObject) xmlTypeValueGroup.getObjectClassID().getId());
		}

		if (xmlTypeValueGroup.getFeatureClassID() != null) {

			if (xmlTypeValueGroup.getFeatureClassID().getId() != null) {

				return (IFeatureClassDeclaration) _xmlModelReader.getDeclaration(
					(XMLRootObject) xmlTypeValueGroup.getFeatureClassID().getId());
			} else {

				return _xmlModelReader.getRealBuildinFeatureClassDeclaration(
					xmlTypeValueGroup.getFeatureClassID().getBuildin(),
					null);
			}
		}

		if (xmlTypeValueGroup.getObjectInstance() != null) {
			return _xmlModelReader.getRealObjectInstanceDeclaration(xmlTypeValueGroup.getObjectInstance());
		}

		if (xmlTypeValueGroup.getObjectInstanceID() != null) {
			return _xmlModelReader.getRealObjectInstanceDeclaration(
				(XMLObjectInstance) xmlTypeValueGroup.getObjectInstanceID().getId());
		}

		if (xmlTypeValueGroup.getVariableID() != null) {
			XMLVariableID xmlVariableID = xmlTypeValueGroup.getVariableID();

			if (xmlVariableID.getId() == null) {

				return null;
			} else {

				Object obj = _xmlModelReader.getDeclaration((XMLRootObject) xmlVariableID.getId());
				Assert.assertNotNull(obj);
				Assert.assertTrue(
					(obj instanceof IBodyVariableClassDeclaration)
						|| (obj instanceof IVariableClassDeclaration));

				return obj;
			}
		}

		// backward compatibility
		if (xmlTypeValueGroup.getBody() != null) {
			// can only save null values
			return null;
		}

		if (xmlTypeValueGroup.getSimObject() != null) {
			// can only save null values
			return null;
		}

		// backward compatibility
		if (xmlTypeValueGroup.getSet() != null) {

			XMLSet xmlSet = xmlTypeValueGroup.getSet();
			int entrySize = xmlSet.getXMLTypeValueGroupCount();
			ArrayList set = new ArrayList(entrySize);

			for (int i = 0; i < entrySize; i++) {
				set.add(createRealTypeValue(xmlSet.getXMLTypeValueGroup(i)));
			}

			return set;
		}

		if (xmlTypeValueGroup.getIterator() != null) {

			return ExecuteIteratorFactory.EMPTY_EXECUTE_ITERATOR;
		}

		if (xmlTypeValueGroup.getList() != null) {

			XMLList xmlList = xmlTypeValueGroup.getList();
			int entrySize = xmlList.getXMLTypeValueGroupCount();
			ArrayList set = new ArrayList(entrySize);

			for (int i = 0; i < entrySize; i++) {
				set.add(createRealTypeValue(xmlList.getXMLTypeValueGroup(i)));
			}

			return set;
		}

		if (xmlTypeValueGroup.getHashtable() != null) {

			XMLHashTable xmlHashTable = xmlTypeValueGroup.getHashtable();
			int entrySize = xmlHashTable.getEntryCount();

			//if (entrySize == 0)
			//return null;
			Map map = new HashMap(entrySize);

			for (int i = 0; i < entrySize; i++) {
				map.put(
					createRealTypeValue(xmlHashTable.getEntry(i).getKey().getXMLTypeValueGroup()),
					createRealTypeValue(xmlHashTable.getEntry(i).getValue().getXMLTypeValueGroup()));
			}

			return map;
		}

		if (xmlTypeValueGroup.getVoid() != null) {

			XMLVoid xmlVoid = xmlTypeValueGroup.getVoid();

			return xmlVoid.getValue();
		}

		if (xmlTypeValueGroup.getParametricFunction() != null) {

			return createSpecificFunctionDeclarationWithArgDeclBinds(
				xmlTypeValueGroup.getParametricFunction());
		}

		if (xmlTypeValueGroup.getFunction() != null) {

			XMLFunction xmlFunction = xmlTypeValueGroup.getFunction();
			ISpecificFunctionType specificFunctionType =
				(ISpecificFunctionType) createRealType(xmlFunction.getXMLTypeGroup());

			return _xmlModelReader.getRealSpecificFunctionDeclaration(
				xmlFunction.getName(),
				specificFunctionType,
				_functionDeclarationList);
		}

		if (xmlTypeValueGroup.getParameterID() != null) {

			XMLFunctionParameterID xmlFunctionParameterID = xmlTypeValueGroup.getParameterID();
			XMLFunctionParameter xmlFunctionParameter = (XMLFunctionParameter) xmlFunctionParameterID.getId();
			IFunctionArgumentDeclaration functionArgumentDeclaration =
				getFunctionArgumentDeclaration(xmlFunctionParameter);

			return functionArgumentDeclaration;
		}

		if (xmlTypeValueGroup.getComposed() != null) {
			Assert.assertEquals(
				"Cannot read composedType Values",
				0,
				xmlTypeValueGroup.getComposed().getXMLTypeValueGroupCount());

			return null;
		}

		if (xmlTypeValueGroup.getComposedAttribute() != null) {
			// just return null, info is in the type
			return null;
		}

		// backward compatibility
		if (xmlTypeValueGroup.getComposedTypeAttribute() != null) {
			// just return null, info is in the type
			return null;
		}

		if (xmlTypeValueGroup.getEnum() != null) {
			ISpecificEnumType specificEnumType = null;

			if (xmlTypeValueGroup.getEnum().getEnumTypeName() != null) {
				specificEnumType =
					(ISpecificEnumType) (TypeManager
						.getSpecificTypeByName(
							xmlTypeValueGroup.getEnum().getEnumTypeName(),
							TypeManager.getInstance().getSystemEnumTypes()));
			} else {
				specificEnumType =
					_xmlModelReader
						.getRealUserTypeDeclaration((XMLUserType) xmlTypeValueGroup.getEnum().getEnumTypeID())
						.getSpecificEnumType();
			}

			return specificEnumType.getItemList().get(xmlTypeValueGroup.getEnum().getIndex());
		}

		Iterator it = _xmlTypeReaders.iterator();
		while (it.hasNext()) {
			IXMLTypeReader typeReader = (IXMLTypeReader) it.next();
			if (typeReader.isResponsibleForValue(xmlTypeValueGroup)) {
				return typeReader.createSpecificTypeValue(xmlTypeValueGroup);
			}
		}

		if (xmlTypeValueGroup.getUnknown() != null) {

			// xmlTypeValueGroup.getUnknown().getSettings();
			// xmlTypeValueGroup.getUnknown().getUnknownType().getSettings();
			return null;
		}

		Assert.fail("Unknown xmlTypeValueGroup: " + xmlTypeValueGroup);

		return null;
	}

	private SpecificFunctionDeclarationWithArgDeclBinds createSpecificFunctionDeclarationWithArgDeclBinds(XMLUnnamedParametricFunction xmlUnnamedParametricFunction)
		throws XMLReadException {

		SpecificFunctionDeclarationWithArgDeclBinds specificFunctionDeclarationWithArgDeclBinds =
			SpecificFunctionDeclarationWithArgDeclBindsFactory
				.createSpecificFunctionDeclarationWithArgDeclBinds();

		_xmlModelReader.setParametricFunctionCallDeclaration(
			specificFunctionDeclarationWithArgDeclBinds.getParametricFunctionCallDeclaration(),
			xmlUnnamedParametricFunction,
			_functionDeclarationList);

		return specificFunctionDeclarationWithArgDeclBinds;
	}

	public SpecificVariableClassDeclarationType createSpecificVariableType(XMLVariableID xmlVariableID)
		throws XMLReadException {

		boolean writeable = xmlVariableID.getWriteable();
		boolean hasWriteable = xmlVariableID.hasWriteable();
		boolean external = xmlVariableID.getExternal();
		boolean hasExternal = xmlVariableID.hasExternal();
		XMLTypeGroup xmlTypeGroup = xmlVariableID.getXMLTypeGroup();
		XMLVariableClass xmlVariableClass = null;

		if (xmlVariableID.getId() instanceof XMLVariableClass) {
			xmlVariableClass = (XMLVariableClass) xmlVariableID.getId();
		} else if (xmlVariableID.getId() instanceof XMLBodyVariableClass) {

			XMLBodyVariableClass xmlBodyVariableClass = (XMLBodyVariableClass) xmlVariableID.getId();
			xmlVariableClass =
				(xmlBodyVariableClass.getVariableClass() != null)
					? xmlBodyVariableClass.getVariableClass()
					: (XMLVariableClass) xmlBodyVariableClass.getVariableClassID();
		}

		if (xmlVariableClass != null) {

			if (!hasWriteable) {
				writeable = xmlVariableClass.getWriteable();
			}

			if (!hasExternal) {
				external = xmlVariableClass.getExternal();
			}

			if (xmlTypeGroup == null) {
				xmlTypeGroup = xmlVariableClass.getXMLTypeGroup();
			}

			//			if (DoubleType.getInstance().equals(createRealType(xmlTypeGroup))
			//				&& IntegerType.getInstance().equals(createRealType(xmlVariableClass.getXMLTypeGroup()))) {
			//				xmlTypeGroup = xmlVariableClass.getXMLTypeGroup();
			//			}
		}

		Assert.assertNotNull(xmlTypeGroup);

		return (SpecificVariableClassDeclarationType) GenericVariableClassDeclarationType
			.getInstance(external, writeable)
			.createSpecificVariableType(createRealType(xmlTypeGroup));
	}

	public IFunctionArgumentDeclaration getFunctionArgumentDeclaration(XMLFunctionParameter xmlFunctionParameter)
		throws XMLReadException {

		if (!_xmlModelReader.containsXMLObject(xmlFunctionParameter)) {
			_xmlModelReader.setDeclaration(
				xmlFunctionParameter,
				createRealFunctionArgumentDeclaration(xmlFunctionParameter));
		}

		return (IFunctionArgumentDeclaration) _xmlModelReader.getDeclaration(xmlFunctionParameter);
	}
}