/*
 * J.A.D.E. Java(TM) Addition to Default Environment. Latest release available at http://jade.dautelle.com/
 * This class is public domain (not copyrighted).
 */
package de.uniwue.ki.util;
import java.io.InvalidObjectException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;

/**
 * <p>
 * This abstract class represents the base class for all enumerates. Enumerates defined using this class have
 * the following advantages over the usual <code>"static int"</code> declaration:
 * <ul>
 * <li>They are strongly typed.</li>
 * <li>There is no assignment error (e.g. duplication).</li>
 * <li>They are <code>Comparable</code> and can be used as collection's keys (e.g. TreeMap).</li>
 * <li>They can be associated to specific integer values for use in <code>switch</code> statements,
 * bit-wise sets ("int" values power of 2) or look-up tables.</li>
 * <li>They are <code>Serializable</code>.</li>
 * <li>The enumeration is extendable (if the constructor is not private).</li>
 * </ul>
 * </p>
 * 
 * <p>
 * Here is an example how this class can be used to define IDs for mouse events. There is no need to reserve
 * predefined ranges of ID values for each event sub-class (Ref. current AWT implementation).
 * 
 * <pre>
 *  public class Event extends java.util.EventObject { public static class ID extends Enum {} } public class MouseEvent extends Event { public static final ID MOUSE_CLICKED = new ID(); public static final ID MOUSE_PRESSED = new ID(); public static final ID MOUSE_RELEASED = new ID(); ... } public class KeyEvent extends Event { public static final ID KEY_PRESSED = new ID(); public static final ID KEY_RELEASED = new ID(); ... }
 *     </pre>
 * 
 * </p>
 * 
 * <p>
 * Here is another example where enums with predefined values are used in a <code>switch</code> statement.
 * Name or value duplication would result in exception being thrown at initialization time.
 * 
 * <pre>
 *  public class Graphic { public static class Color extends Enum { public Color(int value, String name) { super(value, name); } } static final int RED_VALUE = 0xFF0000; static final Color RED = new Color(RED_VALUE, "red");
 * 
 *  static final int GREEN_VALUE = 0x00FF00; static final Color GREEN = new Color(GREEN_VALUE, "green");
 * 
 *  static final int BLUE_VALUE = 0x0000FF; static final Color BLUE = new Color(BLUE_VALUE, "blue"); } ... Color color; switch (color.intValue()) { case RED_VALUE: case GREEN_VALUE: case BLUE_VALUE: }
 *    </pre>
 * 
 * </p>
 * 
 * <p>
 * Note: This class is an improved version of the <a href=
 * "http://developer.java.sun.com/developer/Books/shiftintojava/page1.html"> Typesafe Enum Pattern</a>.
 * </p>
 * 
 * <p>
 * <i>This class is <b>public domain</b> (not copyrighted).</i>
 * </p>
 * 
 * @author <a href="mailto:jean-marie@dautelle.com">Jean-Marie Dautelle</a>
 * @version 3.0, September 10, 2002
 */
public abstract class Enum extends Number implements Serializable, Comparable {

    /**
     * Holds the enumerations per class.
     */
    private static final HashMap ENUMS_PER_CLASS = new HashMap();

    /**
     * Holds the value of this enum.
     */
    private final long _value;

    /**
     * Holds the name of this enum (or <code>null</code> if anonymous)
     */
    private final String _name;

    /**
     * Default constructor. The value of this enum is one more than the previous enum (starting at <code>0</code>).
     * This constructor is used for anonymous extendable enumerations (with value automatically assigned).
     * 
     * @throws IllegalStateException
     *             if an enum with same value already exists.
     */
    protected Enum() {
        this(null);
    }

    /**
     * Creates an enumerate of specified name. The value of this enum is one more than the previous enum
     * (starting at <code>0</code>). This constructor is used for extendable named enumerations (with
     * value automatically assigned).
     * 
     * @param name
     *            the name for this enum.
     * @throws IllegalStateException
     *             if an enum with same name or same value already exists.
     */
    protected Enum(String name) {
        synchronized (ENUMS_PER_CLASS) {
            ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(this.getClass());
            if (enums == null) {
                enums = new ArrayList();
                ENUMS_PER_CLASS.put(this.getClass(), enums);
            }
            _value = enums.size();
            _name = name;
            checkDuplicate(_name, _value, enums);
            enums.add(this);
        }
    }

    /**
     * Creates an enumerate of specified value. This constructor is used for non-extendable anonymous
     * enumerations (e.g. masks with value power of 2).
     * 
     * @param value
     *            the value for this enum.
     * @throws IllegalStateException
     *             if an enum with same value already exists.
     */
    protected Enum(long value) {
        this(value, null);
    }

    /**
     * Creates an enumerate of specified value and specified name. This constructor is used for
     * non-extendable named enumerations (e.g. state machine, <code>switch</code> statements).
     * 
     * @param value
     *            the value for this enum.
     * @param name
     *            the name for this enum.
     * @throws IllegalStateException
     *             if an enum with same name or same value already exists.
     */
    protected Enum(long value, String name) {
        synchronized (ENUMS_PER_CLASS) {
            ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(this.getClass());
            if (enums == null) {
                enums = new ArrayList();
                ENUMS_PER_CLASS.put(this.getClass(), enums);
            }
            _value = value;
            _name = name;
            checkDuplicate(_name, _value, enums);
            enums.add(this);
        }
    }

    /**
     * Compares this enum with the specified object for order. Returns a negative integer, zero, or a
     * positive integer as this object is less than, equal to, or greater than the specified object.
     * 
     * @param o
     *            the object to be compared.
     * @return a negative integer, zero, or a positive integer as this object is less than, equal to, or
     *         greater than the specified object.
     * @throws ClassCastException
     *             if the specified object's type prevents it from being compared to this object.
     */
    public int compareTo(Object o) {
        if (this.getClass() == o.getClass()) {
            return (((Enum) o)._value == _value) ? 0 : (((Enum) o)._value < _value) ? -1 : 1;
        } else {
            throw new ClassCastException("Cannot compare " + this.getClass() + " with " + o.getClass());
        }
    }

    /**
     * Returns a read-only list of the specified enumeration. The collection returned is backed by the actual
     * collection of enums -- so it grows as more enums are defined. For example:
     * 
     * <pre>
     *  public static class Color extends Enum { ... public static Collection getInstances() { return getInstances(Color.class); } }
     * </pre>
     * 
     * @param enumClass
     *            the enum class.
     * @return an unmodifiable view of the enum collection.
     */
    protected static Collection getInstances(Class enumClass) {
        synchronized (ENUMS_PER_CLASS) {
            ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(enumClass);
            if (enums == null) {
                enums = new ArrayList();
                ENUMS_PER_CLASS.put(enumClass, enums);
            }
            return Collections.unmodifiableCollection(enums);
        }
    }

    /**
     * Returns the enum with the specified name. For example:
     * 
     * <pre>
     *  public static class Color extends Enum { ... public static Color getInstance(String name) { return (Color) getInstance(Color.class, name); } }
     * </pre>
     * 
     * @param name
     *            the name of the enum to search for.
     * @return the enum with the specified name or <code>null</code> if not found (e.g. class not
     *         initialized).
     */
    protected static Enum getInstance(Class enumClass, String name) {
        synchronized (ENUMS_PER_CLASS) {
            ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(enumClass);
            if (enums != null) {
                Enum enum = search(name, enums);
                return enum;
            }
            return null;
        }
    }

    /**
     * Returns the enum with the specified value. For example:
     * 
     * <pre>
     *  public static class Color extends Enum { ... public static Color getInstance(int value) { return (Color) getInstance(Color.class, value); } }
     * </pre>
     * 
     * @param value
     *            the value of the enum to search for.
     * @return the enum with the specified value or <code>null</code> if not found (e.g. class not
     *         initialized).
     */
    protected static Enum getInstance(Class enumClass, long value) {
        synchronized (ENUMS_PER_CLASS) {
            ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(enumClass);
            if (enums != null) {
                Enum enum = search(value, enums);
                return enum;
            }
            return null;
        }
    }

    /////////////////////////////
    // Parent abstract methods //
    /////////////////////////////

    /**
     * Returns this enum's value as a <code>int</code>. This may involve rounding or truncation.
     * 
     * @return the numeric value represented by this enum after conversion to type <code>int</code>.
     */
    public final int intValue() {
        return (int) _value;
    }

    /**
     * Returns this enum's value as a <code>long</code>.
     * 
     * @return the numeric value represented by this enum after conversion to type <code>long</code>.
     */
    public final long longValue() {
        return _value;
    }

    /**
     * Returns this enum's value as a <code>float</code>. This may involve rounding.
     * 
     * @return the numeric value represented by this enum after conversion to type <code>float</code>.
     */
    public final float floatValue() {
        return _value;
    }

    /**
     * Returns this enum's value as a <code>double</code>. This may involve rounding.
     * 
     * @return the numeric value represented by this enum after conversion to type <code>double</code>.
     */
    public final double doubleValue() {
        return (double) _value;
    }

    /**
     * Returns the name of this enum.
     * 
     * @return the name of this enum or <code>null</code> if this enum is anonymous.
     */
    public final String getName() {
        return _name;
    }

    /**
     * Returns a string representation of this enum.
     * 
     * @return this enum's name if not anonymous; otherwise returns the decimal representation of its integer
     *         value.
     */
    public String toString() {
        return (_name != null) ? _name : String.valueOf(_value);
    }

    /**
     * Overrides <code>readResolve()</code> to support <code>Serialization</code>.
     * 
     * @return the enum from the corresponding enumeration.
     * @throws InvalidObjectException
     *             if the enum does not exist (e.g. class not initialized).
     */
    private Object readResolve() throws ObjectStreamException {
        ArrayList enums = (ArrayList) ENUMS_PER_CLASS.get(this.getClass());
        if (enums != null) {
            if (_name != null) { // Search by name.
                Enum enum = search(_name, enums);
                if (enum != null) {
                    return enum;
                }
            } else { // Search by value.
                Enum enum = search(_value, enums);
                if (enum != null) {
                    return enum;
                }
            }
        }
        throw new InvalidObjectException("Enum not found, class " + this.getClass() + " not initialized?");
    }

    /**
     * Searches for the enum with the specified value in the specified list.
     * 
     * @param value
     *            the value of the enum to search for.
     * @param enums
     *            the list of enums to search into.
     * @return the enum with the specified value or <code>null</code> if not found.
     */
    private static Enum search(long value, ArrayList enums) {
        for (int i = 0; i < enums.size(); i++) {
            Enum enum = (Enum) enums.get(i);
            if (value == enum._value) {
                return enum;
            }
        }
        return null;
    }

    /**
     * Searches for the enum with the specified name in the specified list.
     * 
     * @param name
     *            the name of the enum to search for.
     * @param enums
     *            the list of enums to search into.
     * @return the enum with the specified name or <code>null</code> if not found.
     */
    private static Enum search(String name, ArrayList enums) {
        for (int i = 0; i < enums.size(); i++) {
            Enum enum = (Enum) enums.get(i);
            if (name.equals(enum._name)) {
                return enum;
            }
        }
        return null;
    }

    /**
     * Checks if an enum with the specified name or value already exists in the specified list.
     * 
     * @param name
     *            the name to search for or <code>null</code> for none.
     * @param value
     *            the value to search for.
     * @param enums
     *            the list of enums to search into.
     * @throws IllegalStateException
     *             if an enum with same name or value exists.
     */
    private static void checkDuplicate(String name, long value, ArrayList enums) {
        if ((name != null) && (search(name, enums) != null)) {
            throw new IllegalStateException("Enum with name \"" + name + "\" already exists.");
        }
        if (search(value, enums) != null) {
            throw new IllegalStateException("Enum with value " + value + " already exists.");
        }
    }
}