package de.uniwue.ki.util.collection;
import java.util.*;

import junit.framework.Assert;

import de.uniwue.ki.util.Utils;
import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.functions.*;
import de.uniwue.ki.util.math.MathHelper;
/**
 * @author Christoph Oechslein
 */
public final class CollectionsHelper {

	private static final boolean DEBUG_COLLECTION_HELPER = false;
	public static final boolean DEFERRED_PROCESSING = true;

	public static final List EMPTY_LIST = Collections.EMPTY_LIST;
	public static final Iterator EMPTY_ITERATOR = new EmptyIterator();
	public static final Enumeration EMPTY_ENUMERATION = new EmptyEnumeration();
	public static final IteratorWithException EMPTY_ITERATOR_WITH_EXCEPTION =
		new EmptyIteratorWithException();
	public static final ListIterator EMPTY_LISTITERATOR = new EmptyListIterator();

	private static class EmptyIterator implements Iterator {
		public final boolean hasNext() {
			return false;
		}
		public final Object next() throws NoSuchElementException {
			throw new NoSuchElementException();
		}
		public final void remove() {
			throw new UnsupportedOperationException();
		}
	}
	private static class EmptyEnumeration implements Enumeration {
		public final boolean hasMoreElements() {
			return false;
		}
		public final Object nextElement() {
			throw new UnsupportedOperationException();
		}
	}
	private static class EmptyIteratorWithException extends EmptyIterator implements IteratorWithException {
		public final Object nextWithException() throws Exception {
			throw new UnsupportedOperationException();
		}
	}
	private static class EmptyListIterator extends EmptyIterator implements ListIterator {
		public final Object previous() {
			throw new UnsupportedOperationException();
		}
		public final void set(Object anObject) {
			throw new UnsupportedOperationException();
		}
		public final void add(Object anObject) {
			throw new UnsupportedOperationException();
		}
		public final int nextIndex() {
			throw new UnsupportedOperationException();
		}
		public final int previousIndex() {
			throw new UnsupportedOperationException();
		}
		public final boolean hasPrevious() {
			return false;
		}
	}
	public static final void addAllWithException(Collection dstCol, IteratorWithException srcIt)
		throws Exception {
		while (srcIt.hasNext()) {
			dstCol.add(srcIt.nextWithException());
		}
	}

	public static final void addAll(Collection dstCol, Iterator srcIt) {
		while (srcIt.hasNext()) {
			dstCol.add(srcIt.next());
		}
	}

	public static final Object aggregateIteratorWithException(
		IteratorWithException it,
		IAggregateFunction func)
		throws Exception {
		while (it.hasNext()) {
			func.add(it.nextWithException());
		}
		return func.getResult();
	}

	public static final Object aggregateIterator(Iterator it, IAggregateFunction func) throws Exception {
		while (it.hasNext()) {
			func.add(it.next());
		}
		return func.getResult();
	}

	public static final Iterator arrayToIterator(Object[] array) {
		return Arrays.asList(array).iterator();
	}

	public static final Object[] asArray(Iterator it) {
		return asArrayList(it).toArray();
	}

	public static final Enumeration asEnumeration(final Iterator it) {
		return new Enumeration() {
			public boolean hasMoreElements() {
				return it.hasNext();
			}
			public Object nextElement() {
				return it.next();
			}
		};
	}

	public static final Iterator asIterator(final Enumeration en) {
		return new Iterator() {

			public boolean hasNext() {
				return en.hasMoreElements();
			}

			public Object next() {
				return en.nextElement();
			}

			public void remove() {
				// not supported by Enumeration
			}

		};
	}

	public static final ArrayList asArrayList(Iterator it) {
		ArrayList newArrayList = new ArrayList();

		addAll(newArrayList, it);

		return newArrayList;
	}

	public static final ArrayList asArrayListWithException(IteratorWithException it) throws Exception {
		ArrayList newArrayList = new ArrayList();

		addAllWithException(newArrayList, it);

		return newArrayList;
	}

	public static final ArrayList asArrayList(Iterator it, boolean trimToSize) {
		ArrayList arrayList = asArrayList(it);

		if (trimToSize)
			arrayList.trimToSize();

		return arrayList;
	}

	public static final ArrayList asArrayListWithException(IteratorWithException it, boolean trimToSize)
		throws Exception {
		ArrayList arrayList = asArrayListWithException(it);

		if (trimToSize)
			arrayList.trimToSize();

		return arrayList;
	}

	public static final LinkedList asLinkedList(Iterator it) {
		LinkedList newLinkedList = new LinkedList();

		addAll(newLinkedList, it);

		return newLinkedList;
	}

	public static final LinkedList asLinkedListWithException(IteratorWithException it) throws Exception {
		LinkedList newLinkedList = new LinkedList();

		addAllWithException(newLinkedList, it);

		return newLinkedList;
	}

	public static final HashSet asHashSet(Iterator it) {
		HashSet newHashSet = new HashSet();

		addAll(newHashSet, it);

		return newHashSet;
	}

	public static final HashSet asHashSetWithException(IteratorWithException it) throws Exception {
		HashSet newHashSet = new HashSet();

		addAllWithException(newHashSet, it);

		return newHashSet;
	}

	public static final void clear(Iterator it) {
		while (it.hasNext()) {
			it.next();
			it.remove();
		}
	}

	public static final IteratorWithException concatenateWithException(IteratorWithException iterator)
		throws Exception {
		if (isEmpty(iterator)) {
			return EMPTY_ITERATOR_WITH_EXCEPTION;

		} else if (DEFERRED_PROCESSING) {
			IteratorWithException result = (IteratorWithException) iterator.nextWithException();

			while (iterator.hasNext()) {
				IteratorWithException iteratorA = (IteratorWithException) iterator.nextWithException();

				result = concatenateWithException(result, iteratorA);
			}

			return debugIterator(result);

		} else {
			List result = asLinkedListWithException((IteratorWithException) iterator.nextWithException());

			while (iterator.hasNext()) {
				addAllWithException(result, (IteratorWithException) iterator.nextWithException());
			}
			return asIteratorWithException(result);
		}
	}

	public static final IteratorWithException concatenateWithException(
		IteratorWithException iteratorA,
		IteratorWithException iteratorB)
		throws Exception {

		if (!iteratorA.hasNext())
			return debugIterator(iteratorB);
		else if (!iteratorB.hasNext())
			return debugIterator(iteratorA);
		else if (DEFERRED_PROCESSING)
			return debugIterator(new ConcatenateTwoIteratorWithException(iteratorA, iteratorB));
		else {
			List result = new LinkedList();
			addAllWithException(result, iteratorA);
			addAllWithException(result, iteratorB);
			return asIteratorWithException(result);
		}
	}

	public static final IteratorWithException concatenateWithException(
		IteratorWithException iteratorA,
		Iterator iteratorB)
		throws Exception {

		if (!iteratorA.hasNext())
			return debugIterator(asIteratorWithException(iteratorB));
		else if (!iteratorB.hasNext())
			return debugIterator(iteratorA);
		else if (DEFERRED_PROCESSING)
			return concatenateWithException(iteratorA, asIteratorWithException(iteratorB));
		else {
			List result = new LinkedList();
			addAllWithException(result, iteratorA);
			addAll(result, iteratorB);
			return asIteratorWithException(result);
		}
	}

	public static final IteratorWithException concatenateWithException(
		Iterator iteratorA,
		IteratorWithException iteratorB)
		throws Exception {

		if (!iteratorA.hasNext())
			return debugIterator(iteratorB);
		else if (!iteratorB.hasNext())
			return debugIterator(asIteratorWithException(iteratorA));
		else if (DEFERRED_PROCESSING)
			return concatenateWithException(asIteratorWithException(iteratorA), iteratorB);
		else {
			List result = new LinkedList();
			addAll(result, iteratorA);
			addAllWithException(result, iteratorB);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator concatenate(Iterator iteratorA, Iterator iteratorB) {
		if (!iteratorA.hasNext())
			return debugIterator(iteratorB);
		else if (!iteratorB.hasNext())
			return debugIterator(iteratorA);
		else if (DEFERRED_PROCESSING)
			return debugIterator(new ConcatenateTwoIterator(iteratorA, iteratorB));
		else {
			List result = new LinkedList();
			addAll(result, iteratorA);
			addAll(result, iteratorB);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator concatenate(Iterator iterator1, Iterator iterator2, Iterator iterator3) {
		return concatenate(iterator1, concatenate(iterator2, iterator3));
	}

	public static final boolean containsWithException(IteratorWithException iterator, Object obj)
		throws Exception {
		while (iterator.hasNext()) {
			if (Utils.equals(obj, iterator.nextWithException()))
				return true;
		}

		return false;
	}

	public static final boolean contains(Iterator iterator, Object obj) {
		while (iterator.hasNext()) {
			if (Utils.equals(obj, iterator.next()))
				return true;
		}

		return false;
	}

	public static final boolean containsWithException(
		IteratorWithException iterator,
		IPredicateFunction predicateFunction)
		throws Exception {
		while (iterator.hasNext()) {
			if (predicateFunction.isElementOk(iterator.nextWithException()))
				return true;
		}

		return false;
	}

	public static final boolean contains(Iterator iterator, IPredicateFunction predicateFunction) {
		while (iterator.hasNext()) {
			Object obj = iterator.next();
			try {
				if (predicateFunction.isElementOk(obj))
					return true;
			} catch (Exception e) {
				throw new RuntimeException();
			}
		}

		return false;
	}

	public static Collection copyCollection(Collection srcCollection) {
		try {
			Collection dstCollection = (Collection) srcCollection.getClass().newInstance();
			dstCollection.addAll(srcCollection);
			return dstCollection;

		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}
	
	public static Collection deepCopyCollection(Collection srcCollection) {
		try {
			Collection dstCollection = (Collection) srcCollection.getClass().newInstance();
			Iterator it = srcCollection.iterator();
			while (it.hasNext()) {
                Object element = (Object) it.next();
                if (element instanceof Collection) 
                    dstCollection.add(deepCopyCollection((Collection) element));
                else if (element instanceof ICopyable)
                    dstCollection.add(((ICopyable) element).copy(null));
                else 
                    dstCollection.add(element);
            }
			return dstCollection;
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}	

	public static final Iterator createIterator(Object first) {
		return new SingletonIterator(first);
	}

	public static final Iterator createIterator(Object anObject1, Object anObject2) {
		return createConsIterator(anObject1, createIterator(anObject2));
	}

	public static final Iterator createIterator(Object anObject1, Object anObject2, Object anObject3) {
		return createConsIterator(anObject1, createConsIterator(anObject2, createIterator(anObject3)));
	}

	public static final Iterator createIterator(
		Object anObject1,
		Object anObject2,
		Object anObject3,
		Object anObject4) {
		return createConsIterator(anObject1, anObject2, createIterator(anObject3, anObject4));
	}

	public static final Iterator createConsIterator(
		Object anObject1,
		Object anObject2,
		Object anObject3,
		Iterator it) {
		return createConsIterator(
			anObject1,
			createConsIterator(anObject2, createConsIterator(anObject3, it)));
	}

	public static final Iterator createConsIterator(Object anObject1, Object anObject2, Iterator it) {
		return createConsIterator(anObject1, createConsIterator(anObject2, it));
	}

	public static final Iterator createConsIterator(Object first, Iterator rest) {
		return new ConsIterator(first, rest);
	}

	public static final IteratorWithException asIteratorWithException(Iterator it) {
		return IteratorWithExceptionAdapter.createIteratorWithException(it);
	}

	public static final IteratorWithException asIteratorWithException(List list) {
		return DelegateListIteratorWithException.createIteratorWithException(list);
	}

	public static final IteratorWithException createIteratorWithException(Object obj) {
		return asIteratorWithException(createIterator(obj));
	}

	public static final IteratorWithException createIteratorWithException(Object obj1, Object obj2) {
		return asIteratorWithException(createIterator(obj1, obj2));
	}

	public static final IteratorWithException createIteratorWithException(
		Object obj1,
		Object obj2,
		Object obj3) {
		return asIteratorWithException(createIterator(obj1, obj2, obj3));
	}

	public static final IteratorWithException createIteratorWithException(
		Object obj1,
		Object obj2,
		Object obj3,
		Object obj4) {
		return asIteratorWithException(createIterator(obj1, obj2, obj3, obj4));
	}

	public static final IteratorWithException createIteratorWithException(
		Object obj1,
		Object obj2,
		Object obj3,
		Object obj4,
		Object obj5) {
		return asIteratorWithException(createConsIterator(obj1, createIterator(obj2, obj3, obj4, obj5)));
	}

	public static final List createList(Object obj) {
		List newList = new ArrayList(1);

		newList.add(obj);

		return newList;
	}

	public static final List createList(Object obj1, Object obj2) {
		List newList = new ArrayList(2);

		newList.add(obj1);
		newList.add(obj2);

		return newList;
	}

	public static final List createList(Object obj1, Object obj2, Object obj3) {
		List newList = new ArrayList(3);

		newList.add(obj1);
		newList.add(obj2);
		newList.add(obj3);

		return newList;
	}

	public static final List createList(Object obj1, Object obj2, Object obj3, Object obj4) {
		List newList = new ArrayList(4);

		newList.add(obj1);
		newList.add(obj2);
		newList.add(obj3);
		newList.add(obj4);

		return newList;
	}

	public static final List createList(Object obj1, Object obj2, Object obj3, Object obj4, Object obj5) {
		List newList = new ArrayList(5);

		newList.add(obj1);
		newList.add(obj2);
		newList.add(obj3);
		newList.add(obj4);
		newList.add(obj5);

		return newList;
	}

	public static final List createList(
		Object obj1,
		Object obj2,
		Object obj3,
		Object obj4,
		Object obj5,
		Object obj6) {
		List newList = new ArrayList(6);

		newList.add(obj1);
		newList.add(obj2);
		newList.add(obj3);
		newList.add(obj4);
		newList.add(obj5);
		newList.add(obj6);

		return newList;
	}

	public static final Iterator enumerationToIterator(final Enumeration e) {
		return debugIterator(new Iterator() {
			public final boolean hasNext() {
				return e.hasMoreElements();
			}
			public final Object next() {
				return e.nextElement();
			}
			public final void remove() {
				throw new UnsupportedOperationException();
			}
		});
	}

	public static final IteratorWithException selectWithException(
		IteratorWithException it,
		IPredicateFunction filterFunction)
		throws Exception {

		if (isEmpty(it))
			return EMPTY_ITERATOR_WITH_EXCEPTION;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new SelectIteratorWithException(it, filterFunction));
		else {
			List result = new LinkedList();
			selectWithException(it, filterFunction, result);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator select(Iterator it, IPredicateFunction filterFunction) {
		if (isEmpty(it))
			return EMPTY_ITERATOR;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new SelectIterator(it, filterFunction));
		else {
			List result = new LinkedList();
			try {
				select(it, filterFunction, result);
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
			return result.iterator();
		}
	}

	// changed from private to public, cause I needed it --steffen
	public static void select(Iterator it, IPredicateFunction filterFunction, List result)
		throws Exception {
		while (it.hasNext()) {
			Object currObj = it.next();
			if (filterFunction.isElementOk(currObj)) {
				result.add(currObj);
			}
		}
	}

	private static void selectWithException(
		IteratorWithException it,
		IPredicateFunction filterFunction,
		List result)
		throws Exception {
		while (it.hasNext()) {
			Object currObj = it.nextWithException();
			if (filterFunction.isElementOk(currObj)) {
				result.add(currObj);
			}
		}
	}

	public static Object findWithException(IteratorWithException it, IPredicateFunction predicateFunction)
		throws Exception {
		while (it.hasNext()) {
			Object curr = it.nextWithException();
			if (predicateFunction.isElementOk(curr))
				return curr;
		}
		return null;
	}

	public static Object find(Iterator it, IPredicateFunction predicateFunction) {
		while (it.hasNext()) {
			Object curr = it.next();
			try {
				if (predicateFunction.isElementOk(curr))
					return curr;
			} catch (Exception exception) {
				throw new RuntimeException(exception);
			}
		}
		return null;
	}

	public static final Iterator flatten(Iterator iterator) {
		if (iterator.hasNext()) {
			Iterator iteratorA = (Iterator) iterator.next();
			if (!iteratorA.hasNext()) {
				return flatten(iterator);
			} else if (iterator.hasNext()) {
				return debugIterator(concatenate(iteratorA, flatten(iterator)));
			} else {
				return debugIterator(iteratorA);
			}
		} else {
			return EMPTY_ITERATOR;
		}
	}

	public static final Iterator flatten(Iterator iterator, IMapFunction mapFunction) {
		return flatten(map(iterator, mapFunction));
	}

	public static final Iterator flatten(
		Iterator iterator,
		IPredicateFunction predicateFunction,
		IMapFunction mapFunction) {

		return flatten(map(select(iterator, predicateFunction), mapFunction));
	}

	/**
	 * Returns the element at the specified position in this iterator. Rainer
	 * Herrler Creation date: (05.09.2000 19:27:38)
	 * 
	 * @param index
	 *            index of element to return.
	 * @param it
	 *            Iterator.
	 * @return the element at the specified position in this list.
	 * @throws IndexOutOfBoundsException
	 *             if the index is out of range (index &lt; 0 || index &gt;=
	 *             iteratorToList(i).size()).
	 */
	public static final Object getWithException(IteratorWithException it, int index) throws Exception {
		if (index < 0)
			throw new IndexOutOfBoundsException("index < 0");
		while (it.hasNext()) {
			Object localObject = it.nextWithException();
			if (index == 0)
				return localObject;
			index -= 1;
		}
		throw new IndexOutOfBoundsException();
	}

	/**
	 * Returns the element at the specified position in this iterator. Rainer
	 * Herrler Creation date: (05.09.2000 19:27:38)
	 * 
	 * @param index
	 *            index of element to return.
	 * @param it
	 *            Iterator.
	 * @return the element at the specified position in this list.
	 * @throws IndexOutOfBoundsException
	 *             if the index is out of range (index &lt; 0 || index &gt;=
	 *             iteratorToList(i).size()).
	 */
	public static final Object get(Iterator it, int index) {
		if (index < 0)
			throw new IndexOutOfBoundsException("index < 0");
		while (it.hasNext()) {
			Object localObject = it.next();
			if (index == 0)
				return localObject;
			index -= 1;
		}
		throw new IndexOutOfBoundsException();
	}

	public static final int indexOf(Iterator it, Object obj) {
		int index = 0;

		while (it.hasNext()) {
			if (Utils.equals(obj, it.next()))
				return index;

			index++;
		}

		return -1;
	}

	public static final int indexOfWithException(IteratorWithException it, Object obj) throws Exception {
		int index = 0;

		while (it.hasNext()) {
			if (Utils.equals(obj, it.nextWithException()))
				return index;

			index++;
		}

		return -1;
	}

	public static boolean hasDuplicateEntries(Iterator it) {
		return hasDuplicateEntries(asArrayList(it));
	}

	public static boolean hasDuplicateEntries(List list) {
		if (list instanceof RandomAccess) {
			for (int i = 0; i < list.size(); i++) {
				Object currObj = list.get(i);

				for (int j = 0; j < list.size(); j++) {
					if ((j != i) && (currObj.equals(list.get(j)))) {
						return true;
					}
				}
			}

		} else {
			Iterator it = list.iterator();
			int i = 0;
			while (it.hasNext()) {
				Object currObj = it.next();

				Iterator it2 = list.iterator();
				int j = 0;
				while (it2.hasNext()) {
					Object currObj2 = it2.next();
					if ((j != i) && (Utils.equals(currObj, currObj2))) {
						return true;
					}
					j++;
				}
				i++;
			}
		}

		return false;
	}

	public static List getDuplicateEntries(Iterator it) {
		return getDuplicateEntries(asArrayList(it));
	}

	public static List getDuplicateEntries(List list) {
		List duplicateList = new LinkedList();

		if (list instanceof RandomAccess) {
			for (int i = 0; i < list.size(); i++) {
				Object currObj = list.get(i);

				for (int j = 0; j < list.size(); j++) {
					if ((j != i) && (currObj.equals(list.get(j)))) {
						duplicateList.add(currObj);
						break;
					}
				}
			}

		} else {
			int i = 0;
			Iterator it = list.iterator();
			while (it.hasNext()) {
				Object currObj = it.next();

				Iterator it2 = list.iterator();
				int j = 0;
				while (it2.hasNext()) {
					Object currObj2 = it2.next();
					if ((j != i) && (Utils.equals(currObj, currObj2))) {
						duplicateList.add(currObj);
						break;
					}
					j++;
				}
				i++;
			}
		}

		return duplicateList;
	}

	public static final IteratorWithException intersectionWithException(IteratorWithException iterator)
		throws Exception {
		if (isEmpty(iterator)) {
			return EMPTY_ITERATOR_WITH_EXCEPTION;

		} else if (DEFERRED_PROCESSING) {
			IteratorWithException result = (IteratorWithException) iterator.nextWithException();

			while (iterator.hasNext() && !isEmpty(result)) {
				IteratorWithException iteratorA = (IteratorWithException) iterator.nextWithException();

				result = intersectionWithException(result, iteratorA);
			}

			return debugIterator(result);

		} else {
			List result = asLinkedListWithException((IteratorWithException) iterator.nextWithException());

			while (iterator.hasNext()) {
				intersection(
					result,
					asLinkedListWithException((IteratorWithException) iterator.nextWithException()));
			}

			return asIteratorWithException(result);
		}
	}

	public static final IteratorWithException intersectionWithException(
		IteratorWithException iteratorA,
		IteratorWithException iteratorB)
		throws Exception {

		if (isEmpty(iteratorA) || isEmpty(iteratorB)) {
			return EMPTY_ITERATOR_WITH_EXCEPTION;

		} else if (DEFERRED_PROCESSING) {
			final Collection elementsOfIteratorB = asLinkedListWithException(iteratorB);
			IPredicateFunction inListBFunction = new IPredicateFunction() {
				public boolean isElementOk(Object obj) {
					return elementsOfIteratorB.contains(obj);
				}
			};
			return debugIterator(selectWithException(iteratorA, inListBFunction));

		} else {
			List result = new LinkedList();
			intersectionWithException(asLinkedListWithException(iteratorA), iteratorB, result);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator intersection(Iterator iteratorA, Iterator iteratorB) {
		if (isEmpty(iteratorA) || isEmpty(iteratorB)) {
			return EMPTY_ITERATOR;

		} else if (DEFERRED_PROCESSING) {
			try {
				return intersectionWithException(
					asIteratorWithException(iteratorA),
					asIteratorWithException(iteratorB));
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}

		} else {
			List result = new LinkedList();
			intersection(asLinkedList(iteratorA), iteratorB, result);
			return result.iterator();
		}
	}

	private static void intersection(Collection listA, Iterator iteratorB, List result) {
		while (iteratorB.hasNext()) {
			Object currObjB = iteratorB.next();
			if (listA.contains(currObjB)) {
				result.add(currObjB);
			}
		}
	}

	private static void intersectionWithException(
		Collection listA,
		IteratorWithException iteratorB,
		List result)
		throws Exception {
		while (iteratorB.hasNext()) {
			Object currObjB = iteratorB.nextWithException();
			if (listA.contains(currObjB)) {
				result.add(currObjB);
			}
		}
	}

	private static void intersection(List result, List listB) {
		Iterator resultIt = result.iterator();

		while (resultIt.hasNext()) {
			if (!listB.contains(resultIt.next())) {
				resultIt.remove();
			}
		}
	}

	public static final boolean isEmpty(Iterator it) {
		return (it == EMPTY_ITERATOR) || !it.hasNext();
	}

	public static final boolean isEmpty(List list) {
		return (list == Collections.EMPTY_LIST) || list.isEmpty();
	}

	public static boolean isEqual(Iterator it1, Iterator it2, boolean testNumbers) {
		if (it1 == it2)
			return true;

		while (it1.hasNext() && it2.hasNext()) {
			Object obj1 = it1.next();
			Object obj2 = it2.next();

			if (testNumbers && (obj1 instanceof Number)) {
				if (obj2 instanceof Number) {
					if (!MathHelper
						.doubleEqual(((Number) obj1).doubleValue(), ((Number) obj2).doubleValue())) {
						return false;
					}
				} else {
					return false;

				}
			} else if (!Utils.equals(obj1, obj2))
				return false;
		}

		return !(it2.hasNext() || it2.hasNext());
	}

	public static final boolean isEqual(Iterator it1, List list2, boolean testNumbers) {
		return isEqual(it1, list2.iterator(), testNumbers);
	}

	public static final boolean isEqual(List list1, Iterator it2, boolean testNumbers) {
		return isEqual(list1.iterator(), it2, testNumbers);
	}

	public static final boolean isEqual(List list1, List list2, boolean testNumbers) {
		return isEqual(list1.iterator(), list2.iterator(), testNumbers);
	}

	public static final boolean isSetEqual(Iterator param1, Iterator param2) {
		return isSetEqual(asLinkedList(param1), asLinkedList(param2));
	}

	public static boolean isSetEqual(List param1, List param2) {
		return isSubSet(param1, param2) && isSubSet(param2, param1);
	}

	public static final boolean isSubSet(Iterator biggerSet, Iterator subSet) {
		return isSubSet(asLinkedList(biggerSet), asLinkedList(subSet));
	}

	public static boolean isSubSet(List biggerSet, List subSet) {
		Iterator it = biggerSet.iterator();
		while (it.hasNext()) {
			if (!subSet.contains(it.next())) {
				return false;
			}
		}

		return true;
	}

	public static final Object[] iteratorToArray(Iterator it, Object[] dstArray) {
		return asLinkedList(it).toArray(dstArray);
	}

	public static final void iteratorToListWithException(IteratorWithException srcIt, List dstList)
		throws Exception {
		dstList.clear();

		addAllWithException(dstList, srcIt);
	}

	public static final void iteratorToList(Iterator srcIt, List dstList) {
		dstList.clear();

		addAll(dstList, srcIt);
	}

	public static final IteratorWithException mapWithException(
		IteratorWithException it,
		IMapFunction mapFunction)
		throws Exception {

		if (isEmpty(it))
			return EMPTY_ITERATOR_WITH_EXCEPTION;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new MapIteratorWithException(it, mapFunction));
		else {
			List result = new LinkedList();
			mapWithException(it, mapFunction, result);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator map(Iterator it, IMapFunction mapFunction) {
		if (isEmpty(it))
			return EMPTY_ITERATOR;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new MapIterator(it, mapFunction));
		else {
			List result = new LinkedList();
			try {
				map(it, mapFunction, result);
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
			return asIteratorWithException(result);
		}
	}

	private static void map(Iterator it, IMapFunction mapFunction, List result) throws Exception {
		while (it.hasNext()) {
			result.add(mapFunction.map(it.next()));
		}
	}

	private static void mapWithException(IteratorWithException it, IMapFunction mapFunction, List result)
		throws Exception {
		while (it.hasNext()) {
			result.add(mapFunction.map(it.nextWithException()));
		}
	}

	public static final IteratorWithException removeWithException(
		IteratorWithException it,
		Object removeObject)
		throws Exception {

		if (isEmpty(it))
			return EMPTY_ITERATOR_WITH_EXCEPTION;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new RemoveObjectIteratorWithException(it, removeObject));
		else {
			List result = new LinkedList();
			removeWithException(it, removeObject, result);
			return asIteratorWithException(result);
		}
	}

	public static final Iterator remove(Iterator it, Object removeObject) {
		if (isEmpty(it))
			return EMPTY_ITERATOR_WITH_EXCEPTION;
		else if (DEFERRED_PROCESSING)
			return debugIterator(new RemoveObjectIterator(it, removeObject));
		else {
			List result = new LinkedList();
			try {
				remove(it, removeObject, result);
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
			return result.iterator();
		}
	}

	private static void remove(Iterator it, Object removeObject, List result) {
		while (it.hasNext()) {
			Object currObj = it.next();
			if (!Utils.equals(removeObject, currObj)) {
				result.add(currObj);
			}
		}
	}

	private static void removeWithException(IteratorWithException it, Object removeObject, List result)
		throws Exception {
		while (it.hasNext()) {
			Object currObj = it.nextWithException();
			if (!Utils.equals(removeObject, currObj)) {
				result.add(currObj);
			}
		}
	}

	public static final void removeDuplicates(List list) {

		Iterator it = list.iterator();

		int index = 0;
		while (it.hasNext()) {
			Object element = it.next();

			Iterator it2 = list.iterator();

			boolean shouldRemove = false;
			int index2 = 0;
			while (it2.hasNext()) {
				Object element2 = it2.next();
				if ((index2 > index) && (Utils.equals(element2, element))) {
					shouldRemove = true;
					break;
				}

				index2++;
			}

			if (shouldRemove) {
				it.remove();
			} else {
				index++;
			}
		}
	}

	public static final Iterator removeDuplicates(Iterator iterator) {

		Set set = new HashSet();

		while (iterator.hasNext()) {
			set.add(iterator.next());
		}

		return set.iterator();
	}

	public static final Iterator removeDuplicates(Iterator iterator, Comparator comparator) {

		Set set = new TreeSet(comparator);

		while (iterator.hasNext()) {
			set.add(iterator.next());
		}

		return set.iterator();
	}

	public static final IteratorWithException reverseWithException(IteratorWithException it)
		throws Exception {
		List dummy = asArrayListWithException(it);
		Collections.reverse(dummy);
		return asIteratorWithException(dummy);
	}

	public static final Iterator reverse(Iterator it) {
		ArrayList dummy = asArrayList(it);
		Collections.reverse(dummy);
		return dummy.iterator();
	}

	public static final IteratorWithException setDifferenceWithException(
		IteratorWithException iteratorA,
		IteratorWithException iteratorB)
		throws Exception {

		return setDifferenceWithException(iteratorA, asLinkedListWithException(iteratorB));
	}

	public static final Iterator setDifference(Iterator iteratorA, Iterator iteratorB) {
		try {
			return setDifferenceWithException(asIteratorWithException(iteratorA), asLinkedList(iteratorB));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public static final IteratorWithException setDifferenceWithException(
		IteratorWithException iteratorA,
		Collection collB)
		throws Exception {
		if (DEFERRED_PROCESSING) {
			return debugIterator(selectWithException(iteratorA, createNotInListBFunction(collB)));
		} else {
			List result = new LinkedList();
			while (iteratorA.hasNext()) {
				Object currObj = iteratorA.nextWithException();
				if (!collB.contains(currObj)) {
					result.add(currObj);
				}
			}
			return asIteratorWithException(result);
		}
	}

	private static IPredicateFunction createNotInListBFunction(final Collection collB) {
		return new IPredicateFunction() {
			public boolean isElementOk(Object obj) {
				return !collB.contains(obj);
			}
		};
	}

	public static final Iterator setDifference(Iterator iteratorA, Collection collB) {
		if (DEFERRED_PROCESSING) {
			return debugIterator(select(iteratorA, createNotInListBFunction(collB)));
		} else {
			List result = new LinkedList();
			while (iteratorA.hasNext()) {
				Object currObj = iteratorA.next();
				if (!collB.contains(currObj)) {
					result.add(currObj);
				}
			}
			return result.iterator();
		}
	}

	public static int setSize(List list, int newSize, Object initialElement) {
		int oldSize = list.size();

		if (newSize < oldSize) {
			for (int i = newSize; i < oldSize; i++) {
				list.remove(i);
			}
		} else if (newSize > oldSize) {
			for (int i = oldSize; i < newSize; i++) {
				list.add(initialElement);
			}
		}

		if (Assert.ASSERTION_ON)
			Assert.assertEquals(list.size(), newSize);

		return oldSize;
	}

	public static final int sizeWithException(IteratorWithException it) throws Exception {
		int count = 0;
		while (it.hasNext()) {
			it.nextWithException();
			count++;
		}
		return count;
	}

	public static final int size(Iterator it) {
		int count = 0;
		while (it.hasNext()) {
			it.next();
			count++;
		}
		return count;
	}

	public static final IteratorWithException unionWithException(
		IteratorWithException iteratorA,
		IteratorWithException iteratorB)
		throws Exception {

		if (isEmpty(iteratorA)) {
			return debugIterator(iteratorB);

		} else if (isEmpty(iteratorB)) {
			return debugIterator(iteratorA);

		} else if (DEFERRED_PROCESSING) {
			final Collection elementsOfIteratorB = asArrayListWithException(iteratorB);
			IPredicateFunction notInListBFunction = new IPredicateFunction() {
				public boolean isElementOk(Object obj) {
					return !elementsOfIteratorB.contains(obj);
				}
			};

			return debugIterator(
				concatenateWithException(
					selectWithException(iteratorA, notInListBFunction),
					elementsOfIteratorB.iterator()));

		} else {
			List listA = asLinkedListWithException(iteratorA);
			unionWithException(iteratorB, listA);
			return asIteratorWithException(listA);
		}
	}

	public static final IteratorWithException unionWithException(IteratorWithException iterator)
		throws Exception {
		if (isEmpty(iterator)) {
			return EMPTY_ITERATOR_WITH_EXCEPTION;

		} else if (DEFERRED_PROCESSING) {
			IteratorWithException result = (IteratorWithException) iterator.nextWithException();

			while (iterator.hasNext()) {
				IteratorWithException iteratorA = (IteratorWithException) iterator.nextWithException();

				result = unionWithException(result, iteratorA);
			}

			return debugIterator(result);

		} else {
			List listA = asLinkedListWithException((IteratorWithException) iterator.nextWithException());

			while (iterator.hasNext()) {
				unionWithException((IteratorWithException) iterator.nextWithException(), listA);
			}
			return asIteratorWithException(listA);
		}
	}

	public static final Iterator union(Iterator iteratorA, Iterator iteratorB) {
		try {
			return unionWithException(asIteratorWithException(iteratorA), asIteratorWithException(iteratorB));
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private static void unionWithException(IteratorWithException iteratorB, List listA) throws Exception {
		while (iteratorB.hasNext()) {
			Object currObjB = iteratorB.nextWithException();
			if (!listA.contains(currObjB)) {
				listA.add(currObjB);
			}
		}
	}

	public static final ArrayList asSortedArrayList(Iterator it, Comparator comparator) {
		ArrayList list = asArrayList(it);
		Collections.sort(list, comparator);
		return list;
	}

	public static final IteratorWithException cartesianProductWithException(
		ITwoArgResultFunction twoArgResultFunction,
		IteratorWithException iterator1,
		List list2)
		throws Exception {

		if (list2.isEmpty() || isEmpty(iterator1)) {
			return EMPTY_ITERATOR_WITH_EXCEPTION;

		} else if (DEFERRED_PROCESSING) {
			return debugIterator(
				new FunctionCartesianProductIteratorWithException(twoArgResultFunction, iterator1, list2));

		} else {
			List result = new LinkedList();
			while (iterator1.hasNext()) {
				Object currObj1 = iterator1.nextWithException();

				Iterator iterator2 = list2.iterator();
				while (iterator2.hasNext()) {
					result.add(twoArgResultFunction.execute(currObj1, iterator2.next()));
				}
			}
			return asIteratorWithException(result);

		}
	}

	public static final Iterator cartesianProduct(
		ITwoArgResultFunction twoArgResultFunction,
		Iterator iterator1,
		List list2) {
		if (list2.isEmpty() || isEmpty(iterator1)) {
			return EMPTY_ITERATOR;
		} else {
			try {
				return debugIterator(
					new FunctionCartesianProductIteratorWithException(
						twoArgResultFunction,
						asIteratorWithException(iterator1),
						list2));
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	private static final IteratorWithException debugIterator(IteratorWithException it) throws Exception {
		if (DEBUG_COLLECTION_HELPER)
			return asIteratorWithException(it);
		else
			return it;
	}

	private static final Iterator debugIterator(Iterator it) {
		if (DEBUG_COLLECTION_HELPER)
			return asLinkedList(it).iterator();
		else
			return it;
	}

	public static List copyList(List list) {
        return (List)copyCollection(list);
    }
	
	public static List deepCopyList(List list) {
        return (List)deepCopyCollection(list);
    }
}