package de.uniwue.ki.util.file;

import java.io.*;
import java.util.StringTokenizer;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import junit.framework.Assert;

/**
 * Creation date: (30.01.2002 15:13:21)
 * @author Christoph Oechslein
 */
public final class FileUtilities {

	private static File createNotExistingFile(File file, int depth, String tempPreffix, String tempSuffix)
		throws IOException {

		File newFile = new File(file.getParent() + File.separatorChar + depth + "_" + file.getName());

		if (!newFile.exists()) {
			return newFile;

		} else if (depth > 1000) {
			return File.createTempFile(tempPreffix, tempSuffix, file.getParentFile());

		} else {
			return createNotExistingFile(file, depth + 1, tempPreffix, tempSuffix);
		}
	}

	public static File changeFileStringSuffix(File file, String suffix) {
		int index = file.getName().indexOf('.');
		if (index == -1)
			return new File(file.getParent(), file.getName() + '.' + suffix);
		else
			return new File(file.getParent(), file.getName().substring(0, index + 1) + suffix);
	}

	public static String getFileStringWithSuffix(File file, String suffix) {
		return getFileWithSuffix(file, suffix).toString();
	}

	public static String getFileStringWithSuffix(String fileName, String suffix) {
		return getFileWithSuffix(fileName, suffix).toString();
	}

	public static File getFileWithSuffix(File file, String suffix) {
		String fileName = file.getName();

		if (fileName.indexOf('.') != -1)
			return file;
		else
			return new File(file.toString() + "." + suffix);
	}

	public static File getFileWithSuffix(String fileName, String suffix) {
		return getFileWithSuffix(new File(fileName), suffix);
	}

	public static File getWriteableNonExistingFile(File file, String tempPreffix, String tempSuffix)
		throws IOException {

		Assert.assertNotNull(file);

		File parentFile = file.getParentFile();
		if (parentFile != null) {

			parentFile.mkdirs();

			if (file.exists()) {
				file = createNotExistingFile(file, 1, tempPreffix, tempSuffix);
			}
		}

		return file;
	}

	public static File getWriteableNonExistingFile(
		File file,
		String tempPreffix,
		String suffix,
		boolean forceSuffix)
		throws IOException {
		if (forceSuffix)
			return getFileWithSuffix(getWriteableNonExistingFile(file, tempPreffix, suffix), suffix);
		else
			return getWriteableNonExistingFile(file, tempPreffix, suffix);
	}

	public static File getWriteableNonExistingFile(String fileString, String tempPreffix, String tempSuffix)
		throws IOException {
		return getWriteableNonExistingFile(new File(fileString), tempPreffix, tempSuffix);
	}

	public static File getWriteableNonExistingFile(
		String fileString,
		String tempPreffix,
		String tempSuffix,
		boolean forceSuffix)
		throws IOException {
		return getWriteableNonExistingFile(new File(fileString), tempPreffix, tempSuffix, forceSuffix);
	}

	public static String createFileNameWithoutRoot(File inputFile) throws IOException {

		String root = getRoot(inputFile);

		if (root == null)
			return inputFile.toString();
		else
			return inputFile.toString().substring(root.length());
	}

	private static String repairBadFileString(String badFileString) {
		return badFileString.replace('?', '_').replace(':', '_');
	}

	public static File repairBadFile(File badFile) throws IOException {
		return new File(getRoot(badFile) + repairBadFileString(createFileNameWithoutRoot(badFile)));
	}

	public static String getRoot(File inputFile) throws IOException {

		String inputFileName;
		try {
			inputFileName = inputFile.getCanonicalPath();

		} catch (IOException e) {
			inputFileName = inputFile.getAbsolutePath();
		}

		File[] roots = File.listRoots();
		for (int i = 0; i < roots.length; i++) {
			String rootFileName = roots[i].getPath();
			if (inputFileName.startsWith(rootFileName))
				return rootFileName;
		}

		return null;
	}

	public static void copyFile(File inFile, File outFile) throws FileNotFoundException, IOException {
		copyInputStreamToOutputStream(new FileInputStream(inFile), new FileOutputStream(outFile), true, 1024);
	}

	public static void copyInputStreamToFile(InputStream is, File outFile)
		throws FileNotFoundException, IOException {
		copyInputStreamToOutputStream(is, new FileOutputStream(outFile), true, 1024);
	}

	public static void copyFileToOutputStream(File inFile, OutputStream os, boolean closeOutputStream)
		throws FileNotFoundException, IOException {
		copyInputStreamToOutputStream(new FileInputStream(inFile), os, closeOutputStream, 1024);
	}

	public static void copyInputStreamToOutputStream(InputStream is, OutputStream os)
		throws FileNotFoundException, IOException {

		copyInputStreamToOutputStream(is, os, true);
	}

	public static void copyInputStreamToOutputStream(
		InputStream is,
		OutputStream os,
		boolean closeOutputStream)
		throws FileNotFoundException, IOException {

		copyInputStreamToOutputStream(is, os, closeOutputStream, 1024);
	}

	public static void copyInputStreamToOutputStream(
		InputStream is,
		OutputStream os,
		boolean closeOutputStream,
		int bufferSize)
		throws FileNotFoundException, IOException {

		BufferedInputStream bis =
			(is instanceof BufferedInputStream)
				? (BufferedInputStream) is
				: new BufferedInputStream(is, bufferSize);

		byte[] data = new byte[bufferSize];
		int byteCount;

		while ((byteCount = bis.read(data, 0, bufferSize)) > -1) {
			os.write(data, 0, byteCount);
		}

		os.flush();
		if (closeOutputStream)
			os.close();
	}

	public static void copyReaderToWriter(Reader reader, Writer writer)
		throws FileNotFoundException, IOException {

		copyReaderToWriter(reader, writer, true);
	}

	public static void copyReaderToWriter(Reader reader, Writer writer, boolean closeOutputStream)
		throws FileNotFoundException, IOException {

		copyReaderToWriter(reader, writer, closeOutputStream, 1024);
	}

	public static void copyReaderToWriter(
		Reader reader,
		Writer writer,
		boolean closeOutputStream,
		int bufferSize)
		throws FileNotFoundException, IOException {

		BufferedReader bufferedReader =
			(reader instanceof BufferedReader)
				? (BufferedReader) reader
				: new BufferedReader(reader, bufferSize);

		char[] data = new char[bufferSize];
		int charCount;

		while ((charCount = bufferedReader.read(data, 0, bufferSize)) > -1) {
			writer.write(data, 0, charCount);
		}

		writer.flush();
		if (closeOutputStream)
			writer.close();
	}

	public static String getContentsAsString(Reader reader) throws IOException {
		StringWriter writer = new StringWriter();

		copyReaderToWriter(reader, writer);

		reader.close();

		return writer.toString();
	}

	public static byte[] encodeAsGZip(String str) {
		try {
			ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
			GZIPOutputStream gzipOutputStream = new GZIPOutputStream(byteArrayOutputStream);

			gzipOutputStream.write(str.getBytes());

			gzipOutputStream.close();
			byteArrayOutputStream.close();

			return byteArrayOutputStream.toByteArray();

		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static String decodeFromGZip(byte[] data) {
		try {
			return getContentsAsString(
				new InputStreamReader(new GZIPInputStream(new ByteArrayInputStream(data))));

		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static String createRelativeFileStringIfPossible(File file, File rootFile) {
		if (Assert.ASSERTION_ON)
			Assert.assertNotNull(rootFile);

		String fileString = file.getAbsoluteFile().getPath();
		String rootFileParentString = rootFile.getAbsoluteFile().getParent();

		if (fileString.startsWith(rootFileParentString)) {
			String relativeFileString = fileString.substring(rootFileParentString.length() + 1);

			Assert.assertTrue(fileString.equals(rootFileParentString + File.separator + relativeFileString));

			return relativeFileString.replace(File.separatorChar, '/');

		} else {
			Assert.assertTrue(new File(fileString).isAbsolute());

			return fileString.replace(File.separatorChar, '/');
		}
	}

	public static void replaceJarFromClassPath(File newJarFile, String oldJarName)
		throws FileNotFoundException, IOException {
		File oldJarFile = getJarFromClassPath(oldJarName);
		if (oldJarFile != null) {
			replaceJar(newJarFile, oldJarFile);
		}
	}

	private static void replaceJar(File newJarFile, File oldJarFile)
		throws FileNotFoundException, IOException {
		File oldBackupJarFile =
			new File(
				oldJarFile.toString().substring(0, oldJarFile.toString().length() - ".jar".length())
					+ "-old.jar");
		if (oldBackupJarFile.exists()) {
			oldBackupJarFile.delete();
		}

		FileUtilities.copyFile(oldJarFile, oldBackupJarFile);
		if (!oldJarFile.delete() || !newJarFile.renameTo(oldJarFile)) {
			// that is the normal behaviour since the jar should be locked
			// don't know what to do if file is readonly 
			FileUtilities.copyFile(newJarFile, oldJarFile);
		}
	}

	public static File getJarFromClassPath(String jarName) {
		StringTokenizer stringTokenizer =
			new StringTokenizer(System.getProperty("java.class.path"), File.pathSeparator, false);

		while (stringTokenizer.hasMoreTokens()) {
			String entry = stringTokenizer.nextToken();

			if (entry.endsWith(jarName)) {
				return new File(entry);
			}
		}

		return null;
	}
}