package de.uniwue.ki.util.jarclassloader;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.zip.*;

import junit.framework.Assert;

import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.file.FileUtilities;

/**
 ** JarClassLoader provides a minimalistic ClassLoader which shows how to
 ** instantiate a class which resides in a .jar file.
<br><br>
 **
 ** @author	John D. Mitchell, Non, Inc., Mar  3, 1999
 **
 ** @version 0.5
 *
 * http://www.javaworld.com/javaworld/javatips/jw-javatip70.html
 **
 **/

/*
 * 
 * @author glueckselig
 *
 * Diese Klasse ist die Hauptklasse des 'Klasse aus Jar-Datei laden'-Moduls.
 * Ursprnglich war nur der Konstruktor und die Methode byte[] loadClassBytes(String className) gegeben.
 * Die Klasse wurde folgendermassen erweitert:
 *    - es knnen mehrere JARs als Quellen registriert werden.
 * 
 * Sie kann sowohl fr das dynamische Laden von LookAndFeels verwendet werden, als auch fr das Laden von
 * Plugins.
 * 
 * Dieser ClassLoader verwendet das Delegations-Modell, das seit JavaVersion 1.2 Verwendung findet.
 */

public class JarClassLoader extends ClassLoader {

	private ArrayList _jarResourcesList;
	private ArrayList _registeredResourcesNameList;

	private static JarClassLoader _jarClassLoader;

	private JarClassLoader() {

		super();
		// setzt den System-ClassLoader als parent und delegiert loadClass-Aufrufe erstmal dorthin	

		_jarResourcesList = new ArrayList();
		_registeredResourcesNameList = new ArrayList();

		initPaths();
	}

	public static synchronized JarClassLoader getInstance() {
		if (_jarClassLoader == null) {
			_jarClassLoader = new JarClassLoader();
		}

		return _jarClassLoader;
	}

	private void initPaths() {
		includePluginDirectory();
	}

	private void includePluginDirectory() {
		// hnlich wie LookAndFeelManagerAuxiliary.registerLookAndFeels()
		File lFdirectory = getPluginDirectory();

		if ((lFdirectory != null) && lFdirectory.isDirectory()) {
			File[] landfs = lFdirectory.listFiles(new FileFilter() {
				public boolean accept(File pathname) {
					return pathname.getName().endsWith(".zip") || pathname.getName().endsWith(".jar");
				}
			});

			for (int index = 0; index < landfs.length; index++) {
				String landfileString = landfs[index].getAbsolutePath();

				registerArchive(landfileString, true);
			}
		}
	}

	private static File _pluginDirectory;

	public synchronized static File getPluginDirectory() {
		if (_pluginDirectory == null) {
			String pluginDirName = System.getProperty(JarClassLoader.class.getName() + ".pluginDir", null);

			if (pluginDirName == null) {
				String[] pluginDirNames =
					{
						System.getProperty("user.dir") + "/plugins",
						System.getProperty("user.dir") + "/../plugins",
						"plugins",
						"../plugins" };

				for (int i = 0; i < pluginDirNames.length; i++) {
					pluginDirName = pluginDirNames[i];
					File pluginDir = new File(pluginDirName);

					if (pluginDir.isDirectory()) {
						_pluginDirectory = pluginDir;
						break;
					}
				}

			} else {
				_pluginDirectory = new File(pluginDirName);
			}
		}
		return _pluginDirectory;
	}

	public Class findClass(String name) throws ClassNotFoundException {

		//System.out.println("jcl - finding class: " + name);

		byte[] classBytes = loadClassBytes(name);
		if (classBytes == null) {
			throw new ClassNotFoundException(name);
		}

		Class result = defineClass(name, classBytes, 0, classBytes.length);
		// in defineClass wird ggf.(wann?) nochmal findClass aufgerufen!
		if (result == null) {
			throw new ClassFormatError();
		}

		return result;
	}

	private byte[] loadClassBytes(String className) {

		// Support the MultiClassLoader's class name munging facility.
		className = formatClassName(className);

		// Attempt to get the class data from one of the JarResources in the list of resources.
		ArrayList classesFound = new ArrayList();
		Iterator iterator = _jarResourcesList.iterator();
		while (iterator.hasNext()) {
			JarResources resources = (JarResources) iterator.next();

			byte[] resource = resources.getResource(className);
			if (resource != null) {

				classesFound.add(resource);
			}

		}

		//		Assert.assertEquals(1, classesFound.size());
		if (classesFound.size() != 1) {
			System.out.println("ERROR: jcl.loadClassBytes: " + classesFound.size() + " : " + className);
		}

		return classesFound.size() != 1 ? null : (byte[]) classesFound.get(0);

	}

	/**
	 * Calls registerArchive(String, boolean) setting deflation by default to true.
	 */
	public boolean loadArchive(String jarFilename) {
		return registerArchive(jarFilename, true);
	}

	public boolean registerArchive(String jarFilename, boolean deflateArchive) {

		if (!containedInList(jarFilename)) {

			_registeredResourcesNameList.add(jarFilename);
			if (deflateArchive) {

				_jarResourcesList.add(new JarResources(jarFilename));

				// jarFilename could be only in entry but not in getLFDir...
				try {
					ClassLoaderExtender.addToClassPath(new URL[] { new File(jarFilename).toURL()});
				} catch (MalformedURLException e) {
					e.printStackTrace();
				}
			}

			return true;
		}

		return false;
	}

	private boolean containedInList(String jarName) {

		boolean found = false;

		Iterator iter = _registeredResourcesNameList.iterator();
		while (!found && iter.hasNext()) {

			String resourceName = (String) iter.next();

			found |= resourceName.equals(jarName);
		}

		return found;
	}

	public List getInstancesOfClassname(String classname) throws ClassNotFoundException {

		try {

			Class targetClass = loadClass(classname); //Class.forName(classname);
			if (targetClass != null) {

				ArrayList result = new ArrayList();

				Iterator iterator = _jarResourcesList.iterator();
				while (iterator.hasNext()) {

					JarResources resource = (JarResources) iterator.next();
					Iterator resourceIt = resource.getAllRescourceNames();
					while (resourceIt.hasNext()) {

						String resName = (String) resourceIt.next();

						if (isClass(resName)) {

							String resNameWithoutDotClass = detail(resName).replace('/', '.');
							Class cls = loadClass(resNameWithoutDotClass, true);
							// wenn defineClass ohne Klassenname als erstes Argument aufgerufen wird, 
							// muss die Klasse vor ihrem Gebrauch erst noch resolved werden!
							//resolveClass(cls);

							if (isInstanceof(cls, classname)) {

								try {
									result.add(cls.newInstance());
								} catch (InstantiationException e) {
									e.printStackTrace();
								} catch (IllegalAccessException e) {
									e.printStackTrace();
								} catch (Throwable e) {

									e.printStackTrace();
								}
							}
						}
					}

				}

				return result;
			}
		} catch (ClassFormatError e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}

		throw new ClassNotFoundException("Class " + classname + " not found.");

	}

	private boolean isClass(String name) {

		return name.indexOf(".class") >= 0;
	}

	private boolean isInstanceof(Class cls, String classname) {

		Class superclass = cls;
		while (superclass != null) {

			if (superclass.getName().equals(classname)) {

				return true;
			}

			superclass = superclass.getSuperclass();

		}

		return false;
	}

	private String detail(String className) {

		int index = className.lastIndexOf(".");
		if (index >= 0) {

			return className.substring(0, index);
		} else {

			return className;
		}
	}

	private static String formatClassName(String className) {

		return className.replace('.', '/') + ".class";
	}

	// wird von java.util.ResourceBundle.findBundle() verwendet
	public InputStream getResourceAsStream(String name) {

		URL url = getResource(name);
		try {

			return url != null ? url.openStream() : null;
		} catch (IOException e) {
			return null;
		}
	}

	protected URL findResource(String name) {

		// TODO muss umgeschrieben werden -> aus der URL wird ein Stream gezogen
		URL url = null;
		try {

			JarResources resources = getJarResourceContainingClass(name);
			String string = resources.getJarFilename();

			String location = "jar:file:" + string.replaceAll("\\\\", "/") + "!/" + name;

			url = null;
			try {

				url = new URL(location);
			} catch (MalformedURLException e) {

				e.printStackTrace();
			}
		} catch (NoSuchElementException e) {

			url = getResourceURLByName(name);
		}

		return url;
	}

	private URL getResourceURLByName(String name) {

		URL url = null;

		Iterator iterator = _registeredResourcesNameList.iterator();
		while (iterator.hasNext() && url == null) {

			String element = (String) iterator.next();

			if (element.endsWith(name)) {

				String resLocation = "file:" + element;
				try {

					url = new URL(resLocation);
				} catch (MalformedURLException e) {

					e.printStackTrace();
				}
			}
		}

		return url;
	}

	private JarResources getJarResourceContainingClass(String name) throws NoSuchElementException {

		ArrayList resultList = new ArrayList();

		Iterator iterator = _jarResourcesList.iterator();
		while (iterator.hasNext()) {

			JarResources res = (JarResources) iterator.next();

			byte[] resource = res.getResource(name);
			if (resource != null) {

				resultList.add(res);
			}
		}

		if (resultList.size() == 1) {

			return (JarResources) resultList.get(0);
		} else {

			throw new NoSuchElementException();
		}
	}

	private JarResources getJarResourceByLocation(String location) {

		for (Iterator iter = _jarResourcesList.iterator(); iter.hasNext();) {
			JarResources element = (JarResources) iter.next();

			if (element.getJarFilename().equals(location)) {

				return element;
			}
		}

		return null;
	}

	public List getClassnamesOfDescendantsFromArchiveGivenByLocation(String classname, String archive) {

		JarResources resource = getJarResourceByLocation(archive);
		if (resource == null) {
			return null;
		}

		try {
			return findDescandantsInResource(classname, resource);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return null;
		}

	}

	//	###############################

	public List getClassnamesOfDescendantsOf(String classname) throws Exception {

		try {

			Class targetClass = loadClass(classname); //Class.forName
			if (targetClass != null) {

				ArrayList result = new ArrayList();

				Iterator iterator = _jarResourcesList.iterator();
				while (iterator.hasNext()) {

					JarResources resource = (JarResources) iterator.next();
					result.addAll(findDescandantsInResource(classname, resource));

				}

				return result;
			}
		} catch (ClassFormatError e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}

		throw new Exception(
			"Problems with class "
				+ classname
				+ " in LookAndFeelManagerAuxiliary.getClassnamesOfDescendantsOf(String).");
	}

	private ArrayList findDescandantsInResource(String classname, JarResources resource)
		throws ClassNotFoundException {

		ArrayList result = new ArrayList();

		Iterator resourceIt = resource.getAllRescourceNames();
		while (resourceIt.hasNext()) {

			String resName = (String) resourceIt.next();

			if (isClass(resName)) {

				String resNameWithoutDotClass = detail(resName).replace('/', '.');
				Class cls = loadClass(resNameWithoutDotClass); //loadClass
				Assert.assertNotNull("Class not found!", cls);
				// wenn defineClass ohne Klassenname als erstes Argument aufgerufen wird, 
				// muss die Klasse vor ihrem Gebrauch erst noch resolved werden!
				//resolveClass(cls);
				if (isInstanceof(cls, classname)) {
					result.add(resNameWithoutDotClass);
				}
			}
		}

		return result;
	}

	public Iterator getIniFilesForPlugins() {

		File pluginDir = getPluginDirectory();

		if (pluginDir == null)
			return CollectionsHelper.EMPTY_ITERATOR;

		//TODO remove
		System.out.println("Loading plugins from " + pluginDir.getAbsolutePath());

		List readerList = new LinkedList();

		if (pluginDir.isDirectory()) {
			addIniFilesForPluginsFromJars(pluginDir, readerList);
			addIniFilesForPlugins(pluginDir, readerList);
		} else {
			Assert.fail("Not a directory or does not exist: " + getPluginDirectory());
		}

		return readerList.iterator();
	}

	private void addIniFilesForPlugins(File directory, List readerList) {
		File[] pluginsF = directory.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				return pathname.getName().endsWith(".ini");
			}
		});

		for (int index = 0; index < pluginsF.length; index++) {
			//TODO remove
			System.out.println(
				"Loading plugin " + pluginsF[index].getAbsolutePath() + " (must be in class path).");

			try {
				readerList.add(new BufferedReader(new FileReader(pluginsF[index])));
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
		}
	}

	private void addIniFilesForPluginsFromJars(File directory, List readerList) {
		File[] pluginsF = directory.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				return pathname.getName().endsWith(".zip") || pathname.getName().endsWith(".jar");
			}
		});

		for (int index = 0; index < pluginsF.length; index++) {
			try {
				//TODO remove
				System.out.println("Loading plugin " + pluginsF[index].getCanonicalFile());

				ZipFile zipFile = new ZipFile(pluginsF[index].getCanonicalFile());
				ZipEntry entry = zipFile.getEntry("plugin.ini");

				if (entry != null) {
					InputStream zipEntryInputStream = zipFile.getInputStream(entry);
					ByteArrayOutputStream arrayOutputStream =
						new ByteArrayOutputStream((int) entry.getSize());
					FileUtilities.copyInputStreamToOutputStream(zipEntryInputStream, arrayOutputStream);
					readerList.add(
						new InputStreamReader(new ByteArrayInputStream(arrayOutputStream.toByteArray())));
				}

				zipFile.close();
			} catch (ZipException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} finally {
			}
		}
	}
}
