package de.uniwue.ki.util.jarclassloader;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import de.uniwue.ki.util.collection.CollectionsHelper;

/**
 * JarResources: JarResources maps all resources included in a
 * Zip or Jar file. Additionaly, it provides a method to extract one
 * as a blob.
 */
public final class JarResources {
	// jar resource mapping tables
	private Hashtable _htSizes = new Hashtable();
	private Hashtable _htJarContents = new Hashtable();

	// deferred initializing
	private boolean _initialized = false;

	// a jar file
	private String _jarFileName;

	/**
	  * creates a JarResources. It extracts all resources from a Jar
	  * into an internal hashtable, keyed by resource names.
	  * @param jarFileName a jar or zip file
	  */
	public JarResources(String jarFileName) {
		_jarFileName = jarFileName;
	}

	/**
	  * Extracts a jar resource as a blob.
	  * @param name a resource name.
	  */
	public byte[] getResource(String name) {
		if (!_initialized)
			init();
		return (byte[]) _htJarContents.get(name);
	}

	public String getJarFilename() {
		return _jarFileName;
	}

	public Iterator getAllRescourceNames() {
		if (!_initialized)
			init();
		return CollectionsHelper.asIterator(_htJarContents.keys());
	}

	/** initializes internal hash tables with Jar file resources.  */
	private void init() {
		try {

			// extracts just sizes only. 
			ZipFile zf = new ZipFile(_jarFileName);

			try {
				Enumeration e = zf.entries();
				while (e.hasMoreElements()) {

					ZipEntry ze = (ZipEntry) e.nextElement();

					_htSizes.put(ze.getName(), new Long(ze.getSize()));
				}
			} finally {
				zf.close();
			}

			// extract resources and put them into the hashtable.
			FileInputStream fis = new FileInputStream(_jarFileName);
			BufferedInputStream bis = new BufferedInputStream(fis);
			ZipInputStream zis = new ZipInputStream(bis);

			try {
				ZipEntry ze = null;
				while ((ze = zis.getNextEntry()) != null) {

					if (ze.isDirectory()) {
						continue;
					}

					//					System.out.println("ze.getName()=" + ze.getName() + "," + "getSize()=" + ze.getSize());

					long lsize = ze.getSize();
					// -1 means unknown size.
					if (lsize == -1) {
						lsize = ((Number) _htSizes.get(ze.getName())).longValue();
					}

					if (lsize >= Integer.MAX_VALUE) {
						throw new RuntimeException(
							"Jar entry '"
								+ ze.getName()
								+ " in jar file '"
								+ _jarFileName
								+ "' too large "
								+ lsize);
					}

					int size = (int) lsize;

					byte[] b = new byte[size];
					int rb = 0;
					int chunk = 0;
					while ((size - rb) > 0) {
						chunk = zis.read(b, rb, size - rb);
						if (chunk == -1) {
							break;
						}

						rb += chunk;
					}

					// add to internal resource hashtable
					_htJarContents.put(ze.getName(), b);

					//					System.out.println(
					//						ze.getName() + "  rb=" + rb + ",size=" + size + ",csize=" + ze.getCompressedSize());
				}
			} finally {
				zis.close();
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			_initialized = true;
		}
	}
}
