package de.uniwue.ki.util.list;

import java.util.*;

import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

import junit.framework.Assert;

import de.uniwue.ki.util.copy.ICopyable;
import de.uniwue.ki.util.event.ListDataMulticaster;

/**
 * Adds default implementations for most methods and has listener support methods.
 * Creation date: (03.11.2000 10:22:31)
 *
 * @author Christoph Oechslein
 */

public abstract class AbstractMutableListModel extends AbstractList implements IMutableListModel {
	protected transient boolean _throwNoEvents = false;

	private transient ListDataListener _listDataListener = null;

	public boolean accept(Object anObject) {
		return true;
	}

	public void addListDataListener(ListDataListener l) {
		_listDataListener = ListDataMulticaster.add(_listDataListener, l);
	}

	public int changeIndex(Object anObject, int newIndex) {
		int oldIndex = indexOf(anObject);

		if (oldIndex == newIndex)
			return oldIndex;

		if (Assert.ASSERTION_ON) {
			Assert.assertTrue(oldIndex != -1);
			Assert.assertTrue((0 <= newIndex) && (newIndex <= size()));
		}

		int currSize = size();
		remove(oldIndex);

		if (newIndex == currSize) {
			add(anObject);
		} else {
			add(newIndex, anObject);
		}

		fireContentsChanged(this, Math.min(oldIndex, newIndex), Math.max(oldIndex, newIndex));

		return oldIndex;
	}

	public ICopyable copy(Map objectMapping) {
		if ((objectMapping != null) && (objectMapping.containsKey(this))) {
			return (ICopyable) objectMapping.get(this);
		} else {
			ICopyable newObject = createInstance(objectMapping);

			if (newObject == this)
				return this;
			else {

				copyInto(newObject, objectMapping);

				return newObject;
			}
		}
	}

	public void copyInto(Object intoObject, Map objectMapping) {
		IMutableListModel dstMutableListModel = (IMutableListModel) intoObject;

		dstMutableListModel.clear();

		Iterator it = iterator();

		if (objectMapping == null) {
			while (it.hasNext()) {
				ICopyable copyable = (ICopyable) it.next();
				dstMutableListModel.add(copyable.copy(objectMapping));
			}

		} else {
			while (it.hasNext()) {
				ICopyable copyable = (ICopyable) it.next();
				if (objectMapping.containsKey(copyable)) {
					dstMutableListModel.add(objectMapping.get(copyable));

				} else {
					Object copyOfCopyable = copyable.copy(objectMapping);
					objectMapping.put(copyable, copyOfCopyable);
					dstMutableListModel.add(copyOfCopyable);
				}
			}
		}
	}

	protected ICopyable createInstance(Map objectMapping) {
		try {
			return (ICopyable) getClass().newInstance();

		} catch (InstantiationException exception) {
			throw new RuntimeException(exception);

		} catch (IllegalAccessException exception) {
			throw new RuntimeException(exception);
		}
	}

	// Comparison and hashing

	/**
	 * Compares the specified object with this list for equality.  Returns
	 * <tt>true</tt> if and only if the specified object is also a list, both
	 * lists have the same size, and all corresponding pairs of elements in
	 * the two lists are <i>equal</i>.  (Two elements <tt>e1</tt> and
	 * <tt>e2</tt> are <i>equal</i> if <tt>(e1==null ? e2==null :
	 * e1.equals(e2))</tt>.)  In other words, two lists are defined to be
	 * equal if they contain the same elements in the same order.<p>
	 *
	 * This implementation first checks if the specified object is this
	 * list. If so, it returns <tt>true</tt>; if not, it checks if the
	 * specified object is a list. If not, it returns <tt>false</tt>; if so,
	 * it iterates over both lists, comparing corresponding pairs of elements.
	 * If any comparison returns <tt>false</tt>, this method returns
	 * <tt>false</tt>.  If either iterator runs out of elements before the
	 * other it returns <tt>false</tt> (as the lists are of unequal length);
	 * otherwise it returns <tt>true</tt> when the iterations complete.
	 *
	 * @param o the object to be compared for equality with this list.
	 * 
	 * @return <tt>true</tt> if the specified object is equal to this list.
	 * @see List#equals(Object)
	 */
	public boolean equals(Object o) {
		return this == o;
	}

	protected void fireContentsChanged(Object source, int index0, int index1) {
		if (!_throwNoEvents && _listDataListener != null) {
			_listDataListener.contentsChanged(
				new ListDataEvent(source, ListDataEvent.CONTENTS_CHANGED, index0, index1));
		}
	}

	protected void fireIntervalAdded(Object source, int index0, int index1) {
		if (!_throwNoEvents && _listDataListener != null) {
			_listDataListener.intervalAdded(
				new ListDataEvent(source, ListDataEvent.INTERVAL_ADDED, index0, index1));
		}
	}

	protected void fireIntervalRemoved(Object source, int index0, int index1) {
		if (!_throwNoEvents && _listDataListener != null) {
			_listDataListener.intervalRemoved(
				new ListDataEvent(source, ListDataEvent.INTERVAL_REMOVED, index0, index1));
		}
	}

	public Object getElementAt(int index) {
		return get(index);
	}

	public String getName() {
		return "Mutable List Model";
	}

	public void appendName(StringBuffer buff, int currDepth, int maxDepth) {
		if (currDepth >= maxDepth) {
			buff.append("...");
		} else {
			buff.append(getName());
		}
	}

	public int getSize() {
		return size();
	}

	/**
	 * Returns the hash code value for this list. <p>
	 *
	 * This implementation uses exactly the code that is used to define the
	 * list hash function in the documentation for the <tt>List.hashCode</tt>
	 * method.
	 *
	 * @return the hash code value for this list.
	 * @see List#hashCode()
	 */
	public int hashCode() {
		return System.identityHashCode(this);
	}

	public void moveTo(Object transferedObject, IMutableListModel targetMutableListModel, int newIndex) {
		if (targetMutableListModel == this) {
			changeIndex(transferedObject, newIndex);

		} else {
			int currSize = size();

			int oldIndex = indexOf(transferedObject);
			Assert.assertTrue(oldIndex != -1);
			Assert.assertTrue((0 <= newIndex) && (newIndex <= currSize));

			remove(transferedObject);

			if (newIndex == currSize) {
				targetMutableListModel.add(transferedObject);
			} else {
				targetMutableListModel.add(newIndex, transferedObject);
			}

			fireContentsChanged(this, Math.min(oldIndex, newIndex), Math.max(oldIndex, newIndex));
		}
	}

	public void removeListDataListener(ListDataListener l) {
		_listDataListener = ListDataMulticaster.remove(_listDataListener, l);
	}

	public void sort(Comparator comparator) {
		try {
			_throwNoEvents = true;

			Collections.sort(this, comparator);

		} finally {
			_throwNoEvents = false;
		}

		fireContentsChanged(this, 0, size());
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		Iterator e = iterator();
		buf.append(getClass().getName());
		buf.append("@");
		buf.append(Integer.toHexString(hashCode()));
		buf.append("[");
		final int maxIndex = 5;
		final int size = size();
		for (int i = 0; i < size; i++) {
			buf.append(String.valueOf(e.next()));
			if (i > maxIndex) {
				buf.append(" ...");
				break;
			}
			if (i < size - 1)
				buf.append(", ");
		}
		buf.append("]");
		return buf.toString();
	}
}