package de.uniwue.ki.util.math;
import java.awt.*;
import java.awt.geom.Point2D;
import java.util.Comparator;

import junit.framework.Assert;
/**
 * Add a few commonly used mathematic functions.
 * <BR>Creation date: (23.03.00 16:50:36)
 *
 * @author Christoph Oechslein
 */

public class MathHelper {

	public interface IMapDouble {
		double mapDouble(Object o);
	}

	public static final Comparator getDoubleComparator(final IMapDouble mapDouble) {
		return new Comparator() {
			public int compare(Object o1, Object o2) {
				return MathHelper.compare(mapDouble.mapDouble(o1), mapDouble.mapDouble(o2));
			}
		};
	}

	public static Point intersectWith(Polygon polygon, int ptX, int ptY, int dx, int dy) {
		double middleX = getMiddleX(polygon) + dx;
		double middleY = getMiddleY(polygon) + dy;

		Rectangle polygonBounds = polygon.getBounds();
		
		polygonBounds.width += 1;
		polygonBounds.height += 1;

		polygonBounds.translate(dx, dy);

		int length = polygon.npoints;
		int[] xpoints = polygon.xpoints;
		int[] ypoints = polygon.ypoints;

		Point minDistancePoint = new Point((int)middleX, (int)middleY);
		Point point = new Point();
		double minDistanceSq = Double.MAX_VALUE;
		for (int i = 0; i < length; i++) {
			double lineBx1 = xpoints[i] + dx;
			double lineBy1 = ypoints[i] + dy;

			double lineBx2 = xpoints[(i + 1) % length] + dx;
			double lineBy2 = ypoints[(i + 1) % length] + dy;

			if (MathHelper
				.calculateIntersection(point, ptX, ptY, middleX, middleY, lineBx1, lineBy1, lineBx2, lineBy2)
				&& polygonBounds.contains(point.x, point.y)
				&& (point.distanceSq(ptX, ptY) < minDistanceSq)) {

				minDistanceSq = point.distanceSq(ptX, ptY);
				minDistancePoint.setLocation(point);
			}
		}

		return minDistancePoint;
	}

	private static int getMiddle(int[] points, int length) {
		int sum = 0;

		for (int i = 0; i < length; i++) {
			sum += points[i];
		}

		return sum / length;
	}

	public static int getMiddleX(Polygon polygon) {
		return getMiddle(polygon.xpoints, polygon.npoints);
	}

	public static int getMiddleY(Polygon polygon) {
		return getMiddle(polygon.ypoints, polygon.npoints);
	}

	public static Point intersectWith(Rectangle rectangle, int ptX, int ptY, int dx, int dy) {
		Polygon polygon =
			new Polygon(
				new int[] {
					rectangle.x,
					rectangle.x + rectangle.width,
					rectangle.x + rectangle.width,
					rectangle.x },
				new int[] {
					rectangle.y,
					rectangle.y,
					rectangle.y + rectangle.height,
					rectangle.y + rectangle.height },
				4);
		return intersectWith(polygon, ptX, ptY, dx, dy);
	}

	public static boolean calculateIntersection(
		Point2D srcPoint,
		double lineAx1,
		double lineAy1,
		double lineAx2,
		double lineAy2,
		double lineBx1,
		double lineBy1,
		double lineBx2,
		double lineBy2) {

		double lineADeltaX = lineAx2 - lineAx1;
		double lineADeltaY = lineAy2 - lineAy1;

		double lineBDeltaX = lineBx2 - lineBx1;
		double lineBDeltaY = lineBy2 - lineBy1;

		double pointDenominator = lineBDeltaY * lineADeltaX - lineADeltaY * lineBDeltaX;

		if (pointDenominator == 0) {
			// parallel lines
			srcPoint.setLocation(Double.MAX_VALUE, Double.MAX_VALUE);
			return false;
		} else {
			double pointNumerator =
				lineADeltaX * (lineBx1 * lineBDeltaY - lineBy1 * lineBDeltaX)
					+ lineBDeltaX * (lineAy1 * lineADeltaX - lineAx1 * lineADeltaY);

			double x = pointNumerator / pointDenominator;
			double y = 0.0;

			if (lineADeltaX != 0) {
				y = (x - lineAx1) * lineADeltaY / lineADeltaX + lineAy1;
				if (lineBDeltaX != 0) {
					if (Assert.ASSERTION_ON)
						Assert.assertEquals(y, (x - lineBx1) * lineBDeltaY / lineBDeltaX + lineBy1, 1e-3);
				}
			} else {
				Assert.assertTrue(lineBDeltaX != 0);
				y = (x - lineBx1) * lineBDeltaY / lineBDeltaX + lineBy1;
			}

			srcPoint.setLocation(x, y);
			return true;
		}
	}

	public static boolean isParallel(
		double lineAx1,
		double lineAy1,
		double lineAx2,
		double lineAy2,
		double lineBx1,
		double lineBy1,
		double lineBx2,
		double lineBy2) {

		double lineADeltaX = lineAx2 - lineAx1;
		double lineADeltaY = lineAy2 - lineAy1;

		double lineBDeltaX = lineBx2 - lineBx1;
		double lineBDeltaY = lineBy2 - lineBy1;

		// double lineADirection = lineADeltaY / lineADeltaX;
		// double lineBDirection = lineBDeltaY / lineBDeltaX;

		double pointDenominator = lineBDeltaY * lineADeltaX - lineADeltaY * lineBDeltaX;

		if (pointDenominator == 0) {
			return true;
		} else {
			return false;
		}
	}

	public static final double max(double x, double y) {
		return Math.max(x, y);
	}

	public static final double max(double x, double y, double z) {
		return max(x, max(y, z));
	}

	public static final double max(double a1, double a2, double a3, double a4) {
		return max(max(a1, a2), max(a3, a4));
	}

	public static final long max(long x, long y) {
		return Math.max(x, y);
	}

	public static final long max(long x, long y, long z) {
		return max(x, max(y, z));
	}

	public static final long max(long a1, long a2, long a3, long a4) {
		return max(max(a1, a2), max(a3, a4));
	}

	public static final double min(double x, double y) {
		return Math.min(x, y);
	}

	public static final double min(double x, double y, double z) {
		return min(x, min(y, z));
	}

	public static final double min(double a1, double a2, double a3, double a4) {
		return min(min(a1, a2), min(a3, a4));
	}

	public static final long min(long x, long y) {
		return Math.min(x, y);
	}

	public static final long min(long x, long y, long z) {
		return min(x, min(y, z));
	}

	public static final long min(long a1, long a2, long a3, long a4) {
		return min(min(a1, a2), min(a3, a4));
	}

	public static final int sign(long x) {
		if (x >= 0)
			return +1;
		else
			return -1;
	}

	public static final int compare(double a, double b) {
		if (doubleEqual(a, b))
			return 0;

		else if (doubleGreater(a, b))
			return -1;

		else
			return 1;
	}

	public static final boolean doubleEqual(double a, double b) {
		Assert.assertFalse("Left side of double comparison was not a number. Maybe you dived by zero before.",Double.isNaN(a));
		Assert.assertFalse("Right side of double comparison was not a number. Maybe you dived by zero before.",Double.isNaN(b));

		return DhbMath.equals(a, b);
	}

	public static final boolean doubleLower(double a, double b) {
		// a<b == !(a>=b)
		return !doubleGreaterEqual(a, b);
	}

	public static final boolean doubleGreater(double a, double b) {
		// a>b == !(a<=b)
		return !doubleLowerEqual(a, b);
	}

	public static final boolean doubleLowerEqual(double a, double b) {
		return doubleEqual(a, b) || (a < b);
	}

	public static final boolean doubleGreaterEqual(double a, double b) {
		return doubleEqual(a, b) || (a > b);
	}

	/**
	 * This class determines the parameters of the floating point
	 * representation
	 * HK: I have refactored it somewhat to make it thread-safe and
	 * to make it easier to understand and to fit into my newsletter.
	 * The algorithms are the same as in the book.
	 *
	 * @author Didier H. Besset
	 */
	private static final class DhbMath {
		/** Radix used by floating-point numbers. */
		private static final int RADIX = computeRadix();
		/** Largest positive value which, when added to 1.0, yields 0 */
		private static final double MACHINE_PRECISION = computeMachinePrecision();
		/** Typical meaningful precision for numerical calculations. */
		private static final double DEFAULT_NUMERICAL_PRECISION = Math.sqrt(MACHINE_PRECISION);

		//		static {
		//			System.out.println("Floating-point machine parameters");
		//			System.out.println("---------------------------------");
		//			System.out.println("radix = " + DhbMath.getRadix());
		//			System.out.println("Machine precision = " + DhbMath.getMachinePrecision());
		//			System.out.println("Default numerical precision = " + DhbMath.defaultNumericalPrecision());
		//			System.out.println("Double.MIN_VALUE = " + Double.MIN_VALUE);
		//			
		//			System.out.println(DhbMath.equals(2.71828182845905, (2.71828182845904 + 0.00000000000001)));
		//			System.out.println(DhbMath.equals(2.71828182845905, 2.71828182845904));
		//			System.out.println(DhbMath.equals(2.718281828454, 2.718281828455));
		//			System.out.println(DhbMath.equals(2.7182814, 2.7182815));
		//
		//			// returns all false !!!
		//			System.out.println(2.71828182845905 == (2.71828182845904 + 0.00000000000001));
		//			System.out.println(2.71828182845905 == 2.71828182845904);
		//			System.out.println(2.718281828454 == 2.718281828455);
		//			System.out.println(2.7182814 == 2.7182815);
		//		}

		private static int computeRadix() {
			int radix = 0;
			double a = 1.0d;
			double tmp1, tmp2;
			do {
				a += a;
				tmp1 = a + 1.0d;
				tmp2 = tmp1 - a;
			} while (tmp2 - 1.0d != 0.0d);
			double b = 1.0d;
			while (radix == 0) {
				b += b;
				tmp1 = a + b;
				radix = (int) (tmp1 - a);
			}
			return radix;
		}

		private static double computeMachinePrecision() {
			double floatingRadix = getRadix();
			double inverseRadix = 1.0d / floatingRadix;
			double machinePrecision = 1.0d;
			double tmp = 1.0d + machinePrecision;
			while (tmp - 1.0d != 0.0d) {
				machinePrecision *= inverseRadix;
				tmp = 1.0d + machinePrecision;
			}
			return machinePrecision;
		}

		/**
		 * @return int
		 */
		public static int getRadix() {
			return RADIX;
		}

		/**
		 * @return double
		 */
		public static double getMachinePrecision() {
			return MACHINE_PRECISION;
		}

		/**
		 * @return double
		 */
		public static double defaultNumericalPrecision() {
			return DEFAULT_NUMERICAL_PRECISION;
		}

		/**
		 * @param a double
		 * @param b double
		 * @return true if the difference between a and b is less than
		 * the default numerical precision
		 */
		public static boolean equals(double a, double b) {
			return equals(a, b, defaultNumericalPrecision());
		}

		/**
		 * @param a double
		 * @param b double
		 * @param precision double
		 * @return true if the relative difference between a and b is
		 * less than precision
		 */
		public static boolean equals(double a, double b, double precision) {
			double norm = Math.max(Math.abs(a), Math.abs(b));
			return norm < precision || Math.abs(a - b) < precision * norm;
		}
	}

}