package de.uniwue.ki.util.math;
import java.util.Random;

public final class RandomDistributionHelper implements IRandomWrapper {
	private final Random _random;

	public RandomDistributionHelper(Random random) {
		super();

		_random = random;
	}

	// generate integers between 0 and 2^32
	public synchronized int nextInt() {
		return _random.nextInt();
	}

	// generate integers between 0 and n-1 (inclusive)
	public synchronized int nextInt(int n) {
		return _random.nextInt(n);
	}

	// generate random boolean variables
	public synchronized boolean nextBoolean() {
		return _random.nextBoolean();
	}

	// generates true with probability p
	public synchronized boolean nextBoolean(double p) {
		return (nextDouble() < p);
	}

	// generate U(0,1)
	// E(X)=1/2 ; Var(X)=1/12
	public synchronized double nextDouble() {
		return _random.nextDouble();
	}

	// generate U(0,a)
	// E(X)=a/2 ; Var(X)=a^2/12
	public synchronized double nextDouble(double a) {
		return nextDouble(0, a);
	}

	// generate U(a,b)
	// E(X)=(b-a)/2 ; Var(X)=(b-a)^2/12
	public synchronized double nextDouble(double a, double b) {
		return a + (b - a) * nextDouble();
	}

	// generate N(0,1)
	// E(X)=1 ; Var(X)=1
	public synchronized double nextGaussian() {
		return _random.nextGaussian();
	}

	// generate N(m,s2)
	// E(X)=m ; Var(X)=s2
	public synchronized double nextGaussian(double m, double s2) {
		return nextGaussian() * Math.sqrt(s2) + m;
	}

	// generate Poisson(1)
	// E(X)=1 ; Var(X)=1
	public synchronized int nextPoisson() {
		return nextPoisson(1);
	}

	// generate Poisson(lambda)
	// E(X)=lambda ; Var(X)=lambda
	public synchronized int nextPoisson(double lambda) {
		int v = -1;
		double l = Math.exp(-lambda), p;
		p = 1.0;
		while (p >= l) {
			p *= nextDouble();
			v++;
		}
		return v;
	}

	// generate Beta(alpha,beta)
	// E(X)=a/(a+b) ; Var(X)=ab/[(a+b+1)(a+b)^2]
	public synchronized double nextBeta(double alpha, double beta) {
		if (alpha <= 0 || beta <= 0)
			throw new IllegalArgumentException("alpha and beta must be strictly positive.");
		if (alpha == 1 && beta == 1)
			return nextDouble();
		else if (alpha >= 1 && beta >= 1) {
			double A = alpha - 1,
				B = beta - 1,
				C = A + B,
				L = C * Math.log(C),
				mu = A / C,
				sigma = 0.5 / Math.sqrt(C);
			double y = nextGaussian(), x = sigma * y + mu;
			while (x < 0 || x > 1) {
				y = nextGaussian();
				x = sigma * y + mu;
			}
			double u = nextDouble();
			while (Math.log(u) >= A * Math.log(x / A) + B * Math.log((1 - x) / B) + L + 0.5 * y * y) {
				y = nextGaussian();
				x = sigma * y + mu;
				while (x < 0 || x > 1) {
					y = nextGaussian();
					x = sigma * y + mu;
				}
				u = nextDouble();
			}
			return x;
		} else {
			double v1 = Math.pow(nextDouble(), 1 / alpha), v2 = Math.pow(nextDouble(), 1 / beta);
			while (v1 + v2 > 1) {
				v1 = Math.pow(nextDouble(), 1 / alpha);
				v2 = Math.pow(nextDouble(), 1 / beta);
			}
			return v1 / (v1 + v2);
		}
	}

	// generate ChiSq(1)
	// E(X)=1 ; Var(X)=2
	public synchronized double nextChiSq() {
		return nextGamma(0.5, 2, 0);
	}

	// generate ChiSq(df)
	// E(X)=df ; Var(X)=2*df
	public synchronized double nextChiSq(int df) {
		return nextGamma(0.5 * (double) df, 2, 0);
	}

	// generate shifted-ChiSq(df)
	// E(X)=df+lambda ; Var(X)=2*df
	public synchronized double nextChiSq(int df, double lambda) {
		return nextGamma(0.5 * (double) df, 2, lambda);
	}

	// generate Exp(1)
	// E(X)=1 ; Var(X)=1
	public synchronized double nextExp() {
		return nextGamma(1, 1, 0);
	}

	// generate Exp(beta)
	// E(X)=beta ; Var(X)=beta^2
	public synchronized double nextExp(double beta) {
		return nextGamma(1, beta, 0);
	}

	// generate shifted-Exp(beta)
	// E(X)=beta+lambda ; Var(X)=beta^2
	public synchronized double nextExp(double beta, double lambda) {
		return nextGamma(1, beta, lambda);
	}

	// generate Gamma(1,1)
	// E(X)=1 ; Var(X)=1
	public synchronized double nextGamma() {
		return nextGamma(1, 1, 0);
	}

	// generate Gamma(alpha,beta)
	// E(X)=alpha*beta ; Var(X)=alpha*beta^2
	public synchronized double nextGamma(double alpha, double beta) {
		return nextGamma(alpha, beta, 0);
	}

	// generate shifted-Gamma(alpha,beta)
	// E(X)=alpha*beta+lambda ; Var(X)=alpha*beta^2
	public synchronized double nextGamma(double alpha, double beta, double lambda) {
		double gamma = 0;
		if (alpha <= 0 || beta <= 0)
			throw new IllegalArgumentException("alpha and beta must be strictly positive.");
		if (alpha < 1) {
			double b, p;
			boolean flag = false;
			b = 1 + alpha * Math.exp(-1);
			while (!flag) {
				p = b * nextDouble();
				if (p > 1) {
					gamma = -Math.log((b - p) / alpha);
					if (nextDouble() <= Math.pow(gamma, alpha - 1))
						flag = true;
				} else {
					gamma = Math.pow(p, 1 / alpha);
					if (nextDouble() <= Math.exp(-gamma))
						flag = true;
				}
			}
		} else if (alpha == 1)
			gamma = -Math.log(nextDouble());
		else {
			double y = -Math.log(nextDouble());
			while (nextDouble() > Math.pow(y * Math.exp(1 - y), alpha - 1))
				y = -Math.log(nextDouble());
			gamma = alpha * y;
		}
		return beta * gamma + lambda;
	}

}