package de.uniwue.ki.util.math;

import java.security.SecureRandom;
import java.util.*;
import java.util.List;
import java.util.Random;

import junit.framework.Assert;
/**
 * <BR>Creation date: (29.03.00 19:33:07)
 * @author Christoph Oechslein
 */
public final class RandomHelper {
	private static final IRandomWrapper _randomWrapper;

	static {
		String wrapper = System.getProperty("de.uniwue.ki.util.math.RandomWrapper", "standard");

		IRandomWrapper randomWrapper = null;

		if ("standard".equals(wrapper)) {
			randomWrapper = new RandomNumberGenerator();

		} else if ("secureRandom".equals(wrapper)) {
			randomWrapper = new RandomDistributionHelper(new SecureRandom());

		} else if ("javaRandom".equals(wrapper)) {
			randomWrapper = new RandomDistributionHelper(new Random());

		} else if ("ranecu".equals(wrapper)) {
			randomWrapper = RandomElementHelper.RANECU;

		} else if ("ranlux".equals(wrapper)) {
			randomWrapper = RandomElementHelper.RANLUX_STD;

		} else if ("ranlux high".equals(wrapper)) {
			randomWrapper = RandomElementHelper.RANLUX_HIGHEST;

		} else if ("ranmar".equals(wrapper)) {
			randomWrapper = RandomElementHelper.RANMAR;

		} else {
			System.out.println("Error: Random number generator with name " + wrapper + " not know!");
			wrapper = "standard";
			randomWrapper = new RandomNumberGenerator();
		}

		if (!"standard".equals(wrapper)) {
			System.out.println("Use '" + wrapper + "' as random number generator.");
		}
		_randomWrapper = randomWrapper;
	}

	private RandomHelper() {
		super();
	}

	public static Object chooseObject(Object anObject, List objectList) {
		int randomPos = getRandomInt(objectList.size() + 1);

		if (randomPos == 0) {
			return anObject;
		} else {
			return objectList.get(randomPos - 1);
		}
	}

	public static Object chooseObject(List objectList) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(!objectList.isEmpty());
		return objectList.get(getRandomInt(objectList.size()));
	}

	public static void shuffle(List list) {
		if (_randomWrapper instanceof Random) {
			Collections.shuffle(list, (Random) _randomWrapper);

		} else {
			int size = list.size();
			Object arr[] = list.toArray();

			// Shuffle array
			for (int i = size; i > 1; i--)
				swap(arr, i - 1, _randomWrapper.nextInt(i));

			// Dump array back into list
			ListIterator it = list.listIterator();
			for (int i = 0; i < arr.length; i++) {
				it.next();
				it.set(arr[i]);
			}
		}
	}

	private static void swap(Object[] arr, int i, int j) {
		Object tmp = arr[i];
		arr[i] = arr[j];
		arr[j] = tmp;
	}

	/**
	* Generates true with probability 0.5
	*/
	public static final boolean getRandomBoolean() {
		return _randomWrapper.nextBoolean();
	}

	/**
	 * Generates true with probability p
	 */
	public static final boolean getRandomBoolean(double p) {
		return _randomWrapper.nextBoolean(p);
	}

	/**
	 * Returns random double in the interval [0, to], i.e. excluding to, including 0.
	
	 */
	public static final double getRandomDouble(double to) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(to > 0);
		return _randomWrapper.nextDouble(to);
	}

	/**
	 * Returns random double in the interval [from, to], i.e. excluding to, including from.
	 */
	public static final double getRandomDouble(double from, double to) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(to >= from);
		return _randomWrapper.nextDouble(from, to);
	}

	/**
	 * Returns random int in the interval [0, to), i.e. excluding to, including 0.
	 */
	public static final int getRandomInt(int to) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(to > 0);
		return _randomWrapper.nextInt(to);
	}

	/**
	 * Returns random int in the interval [from, to), i.e. excluding to, including from.
	 */
	public static final int getRandomInt(int from, int to) {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(to > from);
		return _randomWrapper.nextInt(to - from) + from;
	}

	/**
	 * Returns random double in the normal distribution N(0,1)
	 * E(X)=0 ; Var(X)=1
	 */
	public static final double getRandomNormalDouble() {
		return _randomWrapper.nextGaussian();
	}

	/**
	 * Returns random double in the normal distribution N(m,s2)
	 * E(X)=m ; Var(X)=s2
	 */
	public static final double getRandomNormalDouble(double m, double s2) {
		if (MathHelper.doubleEqual(0, s2))
			return m;
		else
			return _randomWrapper.nextGaussian(m, s2);
	}

	/**
	 * Returns random double in the exponentional distribution Exp(lambda)
	 * E(X)=1/lambda ; Var(X)=1/lambda^2
	 */
	public static final double getRandomExp(double lambda) {
		return _randomWrapper.nextExp(1 / lambda);
	}

	/**
	 * Returns random double in the Gamma distribution Gamma(alpha, lambda) 
	 * E(X)=alpha/lambda ; Var(X)=alpha/lambda^2
	 */
	public static final double getRandomGamma(double alpha, double lambda) {
		return _randomWrapper.nextGamma(alpha, 1 / lambda);
	}

	/**
	 * Returns random double in the Beta distribution Beta(alpha, beta) 
	 * E(X)=a/(a+b) ; Var(X)=ab/[(a+b+1) (a+b) ^2]
	 */
	public static final double getRandomBeta(double alpha, double beta) {
		return _randomWrapper.nextBeta(alpha, beta);
	}

	/**
	 * Returns random double in the Poisson distribution Poisson(lambda) 
	 * E(X)=lambda ; Var(X)=lambda
	 */
	public static final int getRandomPoisson(double lambda) {
		return _randomWrapper.nextPoisson(lambda);
	}
}