package de.uniwue.ki.util.rectangular;
import java.beans.*;
import java.util.Iterator;
import java.util.NoSuchElementException;

import de.uniwue.ki.util.rectangular.event.*;
/**
 * Creation date: (15.11.2000 13:40:57)
 * @author Christoph Oechslein
 */
public abstract class AbstractRectangularMapModel implements IRectangularMapModel {

	private transient RectangularMapObjectAddedListener _rectangularMapObjectAddedListener = null;
	private transient RectangularMapObjectRemovedListener _rectangularMapObjectRemovedListener = null;
	private transient RectangularMapChangedListener _rectangularMapChangedListener = null;
	private transient PropertyChangeSupport _propertyChange;

	private class MapIterator implements Iterator {

		private int _currX;
		private int _currY;

		private int _currentObjectIndex;

		public MapIterator() {
			this(0, 0, true);
		}

		protected MapIterator(int currX, int currY, boolean shouldCallSearchNextObject) {
			super();

			_currX = currX;
			_currY = currY;

			_currentObjectIndex = -1;

			if (shouldCallSearchNextObject)
				searchNextObject();
		}

		protected int getStartX() {
			return 0;
		}

		protected int getStartY() {
			return 0;
		}

		protected int getEndX() {
			return getColsCount();
		}

		protected int getEndY() {
			return getRowsCount();
		}

		protected void searchNextObject() {
			_currentObjectIndex++;

			boolean found = false;
			do {

				if (getObjectCount(_currX, _currY) > _currentObjectIndex) {
					found = true;

				} else if (++_currX < getEndX()) {
					_currentObjectIndex = 0;

				} else if (++_currY < getEndY()) {
					_currentObjectIndex = 0;
					_currX = getStartX();

				} else {
					_currentObjectIndex = -1;
					found = true;
				}

			} while (!found);
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}

		public Object next() throws NoSuchElementException {
			Object value = getObjectAt(_currX, _currY, _currentObjectIndex);

			searchNextObject();

			return value;
		}

		public boolean hasNext() {
			return _currentObjectIndex != -1;
		}
	}

	private final class PartMapIterator extends MapIterator {

		private int _startX;
		private int _startY;

		private int _endX;
		private int _endY;

		public PartMapIterator(int startX, int startY, int endX, int endY) {
			super(startX, startY, false);

			_startX = startX;
			_startY = startY;

			_endX = endX;
			_endY = endY;

			searchNextObject();
		}

		protected int getStartX() {
			return _startX;
		}

		protected int getStartY() {
			return _startY;
		}

		protected int getEndX() {
			return _endX;
		}

		protected int getEndY() {
			return _endY;
		}
	}

	public AbstractRectangularMapModel() {
		super();
	}

	public void addPropertyChangeListener(PropertyChangeListener listener) {
		getPropertyChange().addPropertyChangeListener(listener);
	}

	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPropertyChange().addPropertyChangeListener(propertyName, listener);
	}

	public void addRectangularMapChangedListener(RectangularMapChangedListener newListener) {
		_rectangularMapChangedListener =
			RectangularMapChangedEventMulticaster.add(_rectangularMapChangedListener, newListener);
	}

	public void addRectangularMapObjectAddedListener(RectangularMapObjectAddedListener newListener) {
		_rectangularMapObjectAddedListener =
			RectangularMapObjectAddedEventMulticaster.add(_rectangularMapObjectAddedListener, newListener);
	}

	public void addRectangularMapObjectRemovedListener(RectangularMapObjectRemovedListener newListener) {
		_rectangularMapObjectRemovedListener =
			RectangularMapObjectRemovedEventMulticaster.add(
				_rectangularMapObjectRemovedListener,
				newListener);
	}

	public void fireMapChanged() {
		if (_rectangularMapChangedListener != null)
			_rectangularMapChangedListener.mapChanged(new RectangularMapChangedEvent(this));
	}

	public void fireObjectAdded(Object addedObject, int col, int row) {
		if (_rectangularMapObjectAddedListener != null)
			_rectangularMapObjectAddedListener.objectAdded(
				new RectangularMapObjectAddedEvent(this, addedObject, col, row));
	}

	public void fireObjectRemoved(Object removedObject, int col, int row) {
		if (_rectangularMapObjectRemovedListener != null)
			_rectangularMapObjectRemovedListener.objectRemoved(
				new RectangularMapObjectRemovedEvent(this, removedObject, col, row));
	}

	public void firePositionChanged(int col, int row) {
		if (_rectangularMapChangedListener != null)
			_rectangularMapChangedListener.positionChanged(new RectangularMapChangedEvent(this, col, row));
	}

	public void firePropertyChange(PropertyChangeEvent evt) {
		getPropertyChange().firePropertyChange(evt);
	}

	public void firePropertyChange(String propertyName, int oldValue, int newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	public void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	public void firePropertyChange(String propertyName, boolean oldValue, boolean newValue) {
		getPropertyChange().firePropertyChange(propertyName, oldValue, newValue);
	}

	public Iterator getAllObjects() {
		return new MapIterator();
	}

	public Iterator getAllObjects(int startCol, int startRow, int endCol, int endRow) {
		startCol = getCorrectedCol(startCol);
		startRow = getCorrectedRow(startRow);

		endCol = getCorrectedCol(endCol - 1) + 1; // excluding endPoint
		endRow = getCorrectedRow(endRow - 1) + 1; // excluding endPoint

		return new PartMapIterator(startCol, startRow, endCol, endRow);
	}

	protected int getCorrectedCol(int col) {
		if (col >= getColsCount()) {
			if (isTorus()) {
				return col % getColsCount();

			} else {
				return getColsCount() - 1;
			}

		} else if (col < 0) {
			if (isTorus()) {
				// attention -1 % 4 = -1 (not 3)
				//return col % getColsCount();
				return (col % getColsCount()) + getColsCount();

			} else {
				return 0;
			}

		} else {
			return col;
		}
	}

	protected int getCorrectedRow(int row) {
		if (row >= getRowsCount()) {
			if (isTorus()) {
				return row % getRowsCount();

			} else {
				return getRowsCount() - 1;
			}

		} else if (row < 0) {
			if (isTorus()) {
				// attention -1 % 4 = -1 (not 3)
				//return row % getRowsCount();
				return (row % getRowsCount()) + getRowsCount();

			} else {
				return 0;
			}

		} else {
			return row;
		}
	}

	protected PropertyChangeSupport getPropertyChange() {
		if (_propertyChange == null) {
			_propertyChange = new PropertyChangeSupport(this);
		}

		return _propertyChange;
	}

	public void removePropertyChangeListener(PropertyChangeListener listener) {
		getPropertyChange().removePropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPropertyChange().removePropertyChangeListener(propertyName, listener);
	}

	public void removeRectangularMapChangedListener(RectangularMapChangedListener newListener) {
		_rectangularMapChangedListener =
			RectangularMapChangedEventMulticaster.remove(_rectangularMapChangedListener, newListener);
		return;
	}

	public void removeRectangularMapObjectAddedListener(RectangularMapObjectAddedListener newListener) {
		_rectangularMapObjectAddedListener =
			RectangularMapObjectAddedEventMulticaster.remove(_rectangularMapObjectAddedListener, newListener);
		return;
	}

	public void removeRectangularMapObjectRemovedListener(RectangularMapObjectRemovedListener newListener) {
		_rectangularMapObjectRemovedListener =
			RectangularMapObjectRemovedEventMulticaster.remove(
				_rectangularMapObjectRemovedListener,
				newListener);
		return;
	}

	public void remove() {
		_rectangularMapObjectAddedListener = null;
		_rectangularMapObjectRemovedListener = null;
		_propertyChange = null;
		_rectangularMapChangedListener = null;
	}
}