package de.uniwue.ki.util.rectangular;
import java.io.Serializable;
import java.util.Iterator;

import junit.framework.Assert;

import de.uniwue.ki.util.rectangular.event.*;
/**
 * Creation date: (01.12.2000 16:02:40)
 * @author Christoph Oechslein
 */
public class PartRectangularMapModel extends AbstractRectangularMapModel {
	private int startRow = 0;
	private int endRow = 0;
	private int startCol = 0;
	private int endCol = 0;

	private final IvjEventHandler ivjEventHandler = new IvjEventHandler();
	private IRectangularMapModel ivjDelegate = null;

	private final class IvjEventHandler
		implements
			RectangularMapChangedListener,
			RectangularMapObjectAddedListener,
			RectangularMapObjectRemovedListener,
			java.beans.PropertyChangeListener,
			Serializable {

		public void mapChanged(RectangularMapChangedEvent event) {
			if (event.getSource() == PartRectangularMapModel.this.getDelegate())
				delegateMapChanged(event);
		}

		public void objectAdded(RectangularMapObjectAddedEvent event) {
			if (event.getSource() == PartRectangularMapModel.this.getDelegate())
				delegateObjectAdded(event);
		}

		public void objectRemoved(RectangularMapObjectRemovedEvent event) {
			if (event.getSource() == PartRectangularMapModel.this.getDelegate())
				delegateObjectRemoved(event);
		}

		public void positionChanged(RectangularMapChangedEvent event) {
			if (event.getSource() == PartRectangularMapModel.this.getDelegate())
				delegatePositionChanged(event);
		}

		public void propertyChange(java.beans.PropertyChangeEvent evt) {
			if (evt.getSource() == PartRectangularMapModel.this.getDelegate()
				&& (evt.getPropertyName().equals("rowsCount"))) {
				delegateRowsCount(getDelegate().getRowsCount());
			}
			if (evt.getSource() == PartRectangularMapModel.this.getDelegate()
				&& (evt.getPropertyName().equals("colsCount")))
				delegateColsCount(getDelegate().getColsCount());
		}
	}

	public PartRectangularMapModel() {
		super();
		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public PartRectangularMapModel(IRectangularMapModel rectangularMapModel) {
		super();

		setDelegate(rectangularMapModel);
		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void addObject(int col, int row, java.lang.Object anObject) {
		checkBounds(col, row);
		getDelegate().addObject(convertColToDelegate(col), convertRowToDelegate(row), anObject);
	}

	public void changePosition(int oldCol, int oldRow, int newCol, int newRow, java.lang.Object anObject) {
		checkBounds(oldCol, oldRow);
		checkBounds(newCol, newRow);

		getDelegate().changePosition(
			convertColToDelegate(oldCol),
			convertRowToDelegate(oldRow),
			convertColToDelegate(newCol),
			convertRowToDelegate(newRow),
			anObject);
	}

	private void checkBounds(int col, int row) {
		if (!isColAndRowInInterval(convertColToDelegate(col), convertRowToDelegate(row)))
			throw new RectangularMapIndexOutOfBoundException(col, row);
	}

	private void checkInvariants() {
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(getStartCol() <= getEndCol());
		Assert.assertTrue(getStartRow() <= getEndRow());

		Assert.assertTrue(getColsCount() <= getDelegate().getColsCount());
		Assert.assertTrue(getRowsCount() <= getDelegate().getRowsCount());

		Assert.assertTrue(getEndCol() < getDelegate().getColsCount());
		Assert.assertTrue(getEndRow() < getDelegate().getRowsCount());

		Assert.assertTrue(getStartCol() >= 0);
		Assert.assertTrue(getStartRow() >= 0);
	}

	protected int convertColFromDelegate(int delegateCol) {
		return delegateCol - getStartCol();
	}

	protected int convertColToDelegate(int col) {
		return getStartCol() + col;
	}

	protected int convertRowFromDelegate(int delegateRow) {
		return delegateRow - getStartRow();
	}

	protected int convertRowToDelegate(int row) {
		return getStartRow() + row;
	}

	protected void delegateColsCount(int newColsCount) {
		if (newColsCount <= getEndCol())
			setEndCol(newColsCount - 1);
	}

	protected void delegateMapChanged(de.uniwue.ki.util.rectangular.event.RectangularMapChangedEvent event) {
		fireMapChanged();
	}

	protected void delegateObjectAdded(
		de.uniwue.ki.util.rectangular.event.RectangularMapObjectAddedEvent event) {
		int delegateCol = convertColFromDelegate(event.getCol());
		int delegateRow = convertRowFromDelegate(event.getRow());

		if (isColAndRowInInterval(delegateCol, delegateRow))
			fireObjectAdded(event.getAddedObject(), delegateCol, delegateRow);
	}

	protected void delegateObjectRemoved(
		de.uniwue.ki.util.rectangular.event.RectangularMapObjectRemovedEvent event) {
		int delegateCol = convertColFromDelegate(event.getCol());
		int delegateRow = convertRowFromDelegate(event.getRow());

		if (isColAndRowInInterval(delegateCol, delegateRow))
			fireObjectRemoved(event.getRemovedObject(), delegateCol, delegateRow);
	}

	protected void delegatePositionChanged(
		de.uniwue.ki.util.rectangular.event.RectangularMapChangedEvent event) {
		int delegateCol = convertColFromDelegate(event.getCol());
		int delegateRow = convertRowFromDelegate(event.getRow());

		if (isColAndRowInInterval(delegateCol, delegateRow))
			firePositionChanged(delegateCol, delegateRow);
	}

	protected void delegateRowsCount(int newRowsCount) {
		if (newRowsCount <= getEndRow())
			setEndRow(newRowsCount - 1);
	}

	/**
	 * @return Iterator
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel#getAllObjects()
	 */
	public Iterator getAllObjects() {
		return getAllObjects(0, 0, getColsCount(), getRowsCount());
	}

	/**
	 * @param startX int
	 * @param startY int
	 * @param endX int
	 * @param endY int
	 * @return java.util.Iterator
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel#getAllObjects(int, int, int, int)
	 */
	public java.util.Iterator getAllObjects(int startX, int startY, int endX, int endY) {
		return getDelegate().getAllObjects(
			convertColToDelegate(startX),
			convertRowToDelegate(startY),
			convertColToDelegate(endX),
			convertRowToDelegate(endY));
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel#getColsCount()
	 */
	public int getColsCount() {
		return getEndCol() - getStartCol() + 1;
	}

	/**
	 * Return the delegate property value.
	 * @return de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */

	protected IRectangularMapModel getDelegate() {

		return ivjDelegate;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 */
	public int getEndCol() {
		return endCol;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 */
	public int getEndRow() {
		return endRow;
	}

	/**
	 * @param col int
	 * @param row int
	 * @param index int
	 * @return java.lang.Object
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */
	public java.lang.Object getObjectAt(int col, int row, int index) {
		checkBounds(col, row);

		return getDelegate().getObjectAt(convertColToDelegate(col), convertRowToDelegate(row), index);
	}

	/**
	 * @param col int
	 * @param row int
	 * @return int
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */
	public int getObjectCount(int col, int row) {
		checkBounds(col, row);

		return getDelegate().getObjectCount(convertColToDelegate(col), convertRowToDelegate(row));
	}

	/**
	 * @param col int
	 * @param row int
	 * @return java.util.Iterator
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */
	public java.util.Iterator getObjectsAt(int col, int row) {
		checkBounds(col, row);

		return getDelegate().getObjectsAt(convertColToDelegate(col), convertRowToDelegate(row));
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel#getRowsCount()
	 */
	public int getRowsCount() {
		return getEndRow() - getStartRow() + 1;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 */
	public int getStartCol() {
		return startCol;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (01.12.2000 16:17:24)
	 * @author Christoph Oechslein
	 * @return int
	 */
	public int getStartRow() {
		return startRow;
	}

	/**
	 * @return boolean
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */
	public boolean isAdjustable() {
		return false;
	}

	protected boolean isColAndRowInInterval(int delegateCol, int delegateRow) {
		return isColInInterval(delegateCol) && isRowInInterval(delegateRow);
	}

	protected boolean isColInInterval(int delegateCol) {
		return getStartCol() <= delegateCol && delegateCol <= getEndCol();
	}

	protected boolean isRowInInterval(int delegateRow) {
		return getStartRow() <= delegateRow && delegateRow <= getEndRow();
	}

	/**
	 * @return boolean
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel#isTorus()
	 */
	public boolean isTorus() {
		return getDelegate().isTorus();
	}

	/**
	 * @param col int
	 * @param row int
	 * @param anObject java.lang.Object
	 * @see de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */
	public void removeObject(int col, int row, java.lang.Object anObject) {
		getDelegate().removeObject(convertColToDelegate(col), convertRowToDelegate(row), anObject);
	}

	public void removeObjectAt(int col, int row, int index) {
		getDelegate().removeObjectAt(convertColToDelegate(col), convertRowToDelegate(row), index);
	}
	public void setColsCount(int newColsCount) {
		throw new UnsupportedOperationException("RectangularMapModel is not adjustable");
	}

	/**
	 * Set the delegate to a new value.
	 * @param newValue de.uniwue.ki.util.rectangular.IRectangularMapModel
	 */

	private void setDelegate(IRectangularMapModel newValue) {
		if (ivjDelegate != newValue) {
			if (ivjDelegate != null) {
				ivjDelegate.removeRectangularMapObjectAddedListener(ivjEventHandler);
				ivjDelegate.removeRectangularMapObjectRemovedListener(ivjEventHandler);
				ivjDelegate.removeRectangularMapChangedListener(ivjEventHandler);
				ivjDelegate.removePropertyChangeListener(ivjEventHandler);
			}
			ivjDelegate = newValue;

			/* Listen for events from the new object */
			if (ivjDelegate != null) {
				ivjDelegate.addRectangularMapObjectAddedListener(ivjEventHandler);
				ivjDelegate.addRectangularMapObjectRemovedListener(ivjEventHandler);
				ivjDelegate.addRectangularMapChangedListener(ivjEventHandler);
				ivjDelegate.addPropertyChangeListener(ivjEventHandler);
			}

			setStartCol(0);
			setStartRow(0);
			setEndCol(newValue.getColsCount() - 1);
			setEndRow(newValue.getRowsCount() - 1);
		}
	}

	public void setEndCol(int newEndCol) {
		int oldValue = getEndCol();
		endCol = Math.min(newEndCol, getDelegate().getColsCount() - 1);

		firePropertyChange("endCol", oldValue, getEndCol());

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void setEndRow(int newEndRow) {
		int oldValue = getEndRow();
		endRow = Math.min(newEndRow, getDelegate().getRowsCount() - 1);

		firePropertyChange("endRow", oldValue, getEndRow());

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void setRowsCount(int newRowsCount) {
		throw new UnsupportedOperationException("RectangularMapModel is not adjustable");
	}

	public void setStartCol(int newStartCol) {
		int oldValue = getStartCol();
		if (newStartCol > getEndCol())
			startCol = getEndCol();
		else
			startCol = newStartCol;

		firePropertyChange("startCol", oldValue, getStartCol());

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void setStartRow(int newStartRow) {
		int oldValue = getStartRow();
		if (newStartRow > getEndRow())
			startRow = getEndRow();
		else
			startRow = newStartRow;

		firePropertyChange("startRow", oldValue, getStartRow());

		if (Assert.ASSERTION_ON)
			checkInvariants();
	}

	public void setTorus(boolean newSetting) {
		getDelegate().setTorus(newSetting);
	}

	public void remove() {
		setDelegate(null);

		super.remove();
	}
}