package de.uniwue.ki.util.tree;
import java.util.*;

import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.tree.TreeNode;

import junit.framework.Assert;

import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.list.IMutableListModel;
import de.uniwue.ki.util.name.INamed;

/**
 * Creation date: (07.02.2002 13:48:41)
 * @author Christoph Oechslein
 */
public abstract class AbstractMutableListTreeNode
	extends AbstractObjectTreeNode
	implements IMutableListTreeNode, INamed {

	private final IObjectTreeNodeFactory _treeNodeFactory;
	private List _childrenNodes = null;

	private final ListDataListener _listDataListener = new ListDataListener() {
		public void intervalAdded(ListDataEvent e) {
			AbstractMutableListTreeNode.this.intervalAdded(e);
		}

		public void intervalRemoved(ListDataEvent e) {
			AbstractMutableListTreeNode.this.intervalRemoved(e);
		}

		public void contentsChanged(ListDataEvent e) {
			AbstractMutableListTreeNode.this.contentsChanged(e);
		}
	};

	public AbstractMutableListTreeNode(
		IObjectTreeNode parent,
		Object userObject,
		IObjectTreeNodeFactory treeNodeFactory) {
		super(parent, userObject);

		_treeNodeFactory = treeNodeFactory;
	}

	protected void addListener(IMutableListModel mutableListModel) {
		mutableListModel.addListDataListener(_listDataListener);
	}

	public Enumeration children() {
		return CollectionsHelper.asEnumeration(iterator());
	}

	public Iterator iterator() {
		return getChildrenNodes().iterator();
	}

	protected void contentsChanged(ListDataEvent e) {
		if (isChildrenNodesCreated()) {
			removeChildren();
			_childrenNodes = createChildrenNodeList(_childrenNodes);
		}
	}

	private List createChildrenNodeList() {
		return createChildrenNodeList(new ArrayList(getMutableListModel().size()));
	}

	private List createChildrenNodeList(List childrenNodeList) {
		Iterator it = getMutableListModel().iterator();
		while (it.hasNext()) {
			childrenNodeList.add(_treeNodeFactory.createObjectTreeNode(this, it.next()));
		}

		return childrenNodeList;
	}

	public boolean getAllowsChildren() {
		return true;
	}

	public TreeNode getChildAt(int childIndex) {
		if (Assert.ASSERTION_ON) 
			Assert.assertTrue("Index "+childIndex+" not available in "+getName(),childIndex<getChildCount());
		return (TreeNode) getChildrenNodes().get(childIndex);
	}

	public int getChildCount() {
		return getChildrenNodes().size();
	}

	protected final List getChildrenNodes() {
		if (!isChildrenNodesCreated()) {
			synchronized (this) {
				if (!isChildrenNodesCreated()) {
					_childrenNodes = createChildrenNodeList();
				}
			}
		}

		return _childrenNodes;
	}

	public int getIndex(TreeNode node) {
		return getChildrenNodes().indexOf(node);
	}

	protected void intervalAdded(ListDataEvent e) {
		if (!isChildrenNodesCreated())
			return;

		int index0 = e.getIndex0();
		int index1 = e.getIndex1();

		for (int i = index0; i <= index1; i++) {
			getChildrenNodes().add(
				i,
				_treeNodeFactory.createObjectTreeNode(this, getMutableListModel().get(i)));
		}
	}

	protected boolean isChildrenNodesCreated() {
		return _childrenNodes != null;
	}

	protected void intervalRemoved(ListDataEvent e) {
		if (!isChildrenNodesCreated())
			return;

		int index0 = e.getIndex0();
		int index1 = e.getIndex1();

		for (int i = 0; i <= index1 - index0; i++) {
			IObjectTreeNode objectTreeNode = (IObjectTreeNode) getChildrenNodes().remove(index0);
			objectTreeNode.remove();
		}
	}

	public boolean isLeaf() {
		return false;
	}

	public void remove() {
		if (getMutableListModel() != null)
			removeListener(getMutableListModel());

		removeChildren();
	}

	private void removeChildren() {
		if (_childrenNodes != null) {
			Iterator it = _childrenNodes.iterator();
			while (it.hasNext()) {
				IObjectTreeNode currIObjectTreeNode = (IObjectTreeNode) it.next();

				currIObjectTreeNode.remove();
				it.remove();
			}
		}
	}

	protected void removeListener(IMutableListModel mutableListModel) {
		mutableListModel.removeListDataListener(_listDataListener);
	}
}