package mygale.mcge.cmd;
import java.awt.event.ActionEvent;

import javax.swing.*;

import de.uniwue.ki.gui.util.SwingMoreUtilities;

/**
 * Abstract class for all editor commands. The editor serves as a command shell for executing actions in much
 * the same way that a DOS or UNIX commmand command shell executes programs. Once an instance of a Cmd is
 * made, it can be sent the doIt() message to perform that action.
 * <p>
 * Since this is subclassed from class AbstractAction in the Swing user interface library, Cmd objects can be
 * easily added to menus and toolbars.
 * <p>
 * <center><img src="doc-files/Cmd_uml.gif" width="245" height="146" border="0" usemap="#umlclick">
 * <map name="umlclick">
 * <area shape="Rect" coords="68,0,149,27"  href="http://java.sun.com/products/jdk/1.2/docs/api/javax/swing/AbstractButton.html">
 * <area shape="Rect" coords="125,117,244,145"  href="CmdUndoRedoListener.html">
 * <area shape="Rect" coords="0,117,90,145"  href="CmdGraphEditor.html"></map></center>
 * </p>
 * 
 * @see mygale.mcge.core.JGraph
 */
public abstract class Cmd extends AbstractAction {

    ////
    // constants

    /**
     * used to say that this command have an associated icon.
     */
    public static final boolean HAS_ICON = true;
    /**
     * used to say that this command have not an associated icon.
     */
    public static final boolean NO_ICON = false;

    /**
     * Construct a new Cmd.
     * 
     * @param name
     *            name of this command.
     */
    public Cmd(String name) {
        super(name);

        updateEnabled();
    }

    /**
     * Construct a new Cmd.
     * 
     * @param name
     *            name of this command.
     * @param iconSize
     *            int
     * @param iconName
     *            path to the asscoiated icon.
     */
    public Cmd(String name, int iconSize, String iconName) {
        super(name);

        if (iconName != null) {
            addIcon(SwingMoreUtilities.getJavaSunIcon(iconSize, iconName));
        }

        updateEnabled();
    }

    /**
     * This method was called each time this command needs to be ran.
     * 
     * @param ae
     *            the description of the event. Not used.
     * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
     */
    public void actionPerformed(ActionEvent ae) {
        doIt();
    }

    private void addIcon(Icon icon) {
        putValue(Action.SMALL_ICON, icon);
    }

    /**
     * Perform whatever Cmd this Cmd is meant to do. Subclasses should override this to do whatever is
     * intended.
     */
    public abstract void doIt();

    /**
     * Return a name for this Cmd suitable for display to the user.
     * 
     * @return a string that is the name of this command.
     */
    public String getName() {
        return (String) getValue(NAME);
    }

    /**
     * Return the tooltip text associated to the command.
     * 
     * @return a string that is the tooltip of this command.
     */
    public abstract String getToolTipText();

    /**
     * Set a name for this Cmd suitable for display to the user.
     * 
     * @param name
     *            a string that is the name of this command.
     */
    public void setName(String name) {
        putValue(NAME, name);
    }

    /**
     * Return <code>true</code> if this action should be available to the user. Sublass implementations of
     * this method should always call <code>super.shouldBeEnabled()</code> first.
     * 
     * @return <code>true</code> if this command is enabled, otherwise <code>false</code>.
     */
    public boolean shouldBeEnabled() {
        return true;
    }

    /**
     * Determine if this Cmd should be shown as grayed out in menus and toolbars.
     */
    public void updateEnabled() {
        setEnabled(shouldBeEnabled());
    }
}