package mygale.mcge.cmd;


import java.awt.*;
import java.util.*;

import junit.framework.Assert;

import mygale.mcge.core.*;
import mygale.mcge.fig.*;
import mygale.mcge.net.*;
import mygale.mcge.sel.Selection;
import mygale.mcge.undoredo.*;

/** Abstract class that permits to align objects.
 *  <p><center>
 *    <img src="doc-files/CmdAlign_uml.gif" width="93" height="83" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,0,92,28"  href="CmdUndoable.html">
 *    </map>
 *  </center></p>
 *
 */
public abstract class CmdAlign extends CmdUndoable {

	/** This is the list of changes.
	 */
	protected Vector _changes = new Vector();

	/** This class permits to describe a change.
	 *
		 */
	protected static class AlignChange {

	/** A reference to the changed figure.
	 */
	public final IFig figure;

	/** Old location of the changed figure.
	 */
	public final Point oldLocation;

	/** New location of the changed figure.
	 */
	public final Point newLocation;

	/** Construct a new AlignChange.
	 *
	 * @param fig a reference to the changed figure.
	 * @param op old location of the given figure.
	 * @param np new location of the given figure.
	 */
	public AlignChange(IFig fig, Point op, Point np) {
		figure = fig;
		oldLocation = new Point( op );
		newLocation = new Point( np );
	}

	}

	/** This classe permits to undo/redo a instance of this command
	 *
		 */
	class UndoAlign
	implements UndoableCmd, RedoableCmd {

	/** This is changes.
	 */
	protected Vector _changes;

	/** Construct a new UndoAlign.
	 *
	 * @param changes the changes.
	 */
	public UndoAlign(Vector changes) {
		_changes = changes;
	}

/** This method was called by the undo command.
 *  It must undo something.
 *
 * throws mygale.mcge.throwable.MCGEInternalError Something wrong !
 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
 */
public void undoIt() {
   JGraph graph = getGraph();
   Rectangle rFigs = null;
   Enumeration enum = _changes.elements();
   while (enum.hasMoreElements()) {
      AlignChange chgt = (AlignChange) enum.nextElement();
      if (rFigs == null)
         rFigs = new Rectangle(chgt.figure.getBounds());
      else
         rFigs.add(chgt.figure.getBounds());
      chgt.figure.setLocation(chgt.oldLocation.x, chgt.oldLocation.y);
      rFigs.add(chgt.figure.getBounds());
   }
   graph.damaged(rFigs);
}

/** This method was called by the redo command.
 *  It must redo something.
 *
 * throws mygale.mcge.throwable.MCGEInternalError Something wrong !
 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
 */
public void redoIt() {
   JGraph graph = getGraph();
   Rectangle rFigs = null;
   Enumeration enum = _changes.elements();
   while (enum.hasMoreElements()) {
      AlignChange chgt = (AlignChange) enum.nextElement();
      if (rFigs == null)
         rFigs = new Rectangle(chgt.figure.getBounds());
      else
         rFigs.add(chgt.figure.getBounds());
      chgt.figure.setLocation(chgt.newLocation.x, chgt.newLocation.y);
      rFigs.add(chgt.figure.getBounds());
   }
   graph.damaged(rFigs);
}

	} /* class UndoAlign */


    /** Construct a new CmdAlign.
     *
     * @param editor a reference to the current editor.
     * @param name String
 * @param iconSize int
 * @param iconName the path to the associated icon.
     */
    public CmdAlign(JGraphEditor editor, String name, int iconSize, String iconName) {
    super(editor, name, iconSize, iconName );
    }


/** Add this command to the undo list.
     */
    protected void addUndoCmd() {
    if ( _changes.size() > 0 ) {
        addUndoCmd( new UndoAlign( _changes ) );
    }
    }


/** Destroy the nudo/redo changes.
     */
    protected void clearUndoChanges() {
    _changes = new Vector();
    }


/** This method is called each time the command wans to run itself.
     *
     * throws mygale.mcge.throwable.MCGEInternalError Something wrong !
     */
public void doIt() {
   JGraph graph = getGraph();
   Selection sel = graph.getSelection();
   if (sel.unlockedSize() > 1) {
      Rectangle rSelect = sel.getBounds();

      Iterator it = sel.getUnlockedFigs();
      clearUndoChanges();
      while (it.hasNext()) {
         IFig fig = (IFig) it.next();
         rSelect.add(fig.getBounds());

         Point oldLocation = fig.getLocation();
         Point newLocation = getNewLocation(fig, oldLocation);

         fig.setLocation(newLocation.x, newLocation.y);

         if (fig instanceof IFigNode) {
            updateFigNode((IFigNode) fig, newLocation, rSelect);
         }

         rSelect.add(fig.getBounds());
         saveChange(fig, oldLocation, newLocation);
      }

      int handleSize = getGraph().getPrefs().getHandleSize();

      rSelect.grow(handleSize * 2, handleSize * 2);

      graph.damaged(rSelect);
      addUndoCmd();
   }
}


protected abstract Point getNewLocation(IFig fig, Point oldLocation);


    // Undo/redo API

    /** This method permit to register (from doIt() for example)
     *  a change.
     *
     * @param fig a reference to the changed figure.
     * @param oldLocation old location of the given figure.
     * @param newLocation new location of the given figure.
     */
    protected void saveChange(IFig fig, Point oldLocation, Point newLocation ) {
    _changes.addElement( new AlignChange( fig, oldLocation, newLocation ) );
    }


protected void updateFigNode(IFigNode figNode, Point newLocation, Rectangle rSelect) {
   int new_x = newLocation.x;
   int new_y = newLocation.y;

   INetNode netNode = (INetNode) figNode.getNetObject();
   Iterator it2 = netNode.getNodeEdges();
   while (it2.hasNext()) {
      INetEdge netEdge = (INetEdge) it2.next();
      IFigEdge figEdge = (IFigEdge) getGraph().getFigMapper().getFigFor(netEdge);

      rSelect.add(figEdge.getBounds());

      Point newPoint = new Point(new_x, new_y);

      if (netEdge.getBeginNode() == netNode) {
         figEdge.setCtrlPointAt(0, newPoint);

      } else {
         if (Assert.ASSERTION_ON) Assert.assertSame(netNode, netEdge.getEndNode());
         figEdge.setCtrlPointAt(figEdge.getCtrlPointSize() - 1, newPoint);
      }

      figEdge.updateEndsLocation();
      rSelect.add(figEdge.getBounds());
   }
}
}