package mygale.mcge.cmd;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.Vector;

import mygale.mcge.core.JGraph;
import mygale.mcge.core.JGraphEditor;
import mygale.mcge.fig.*;
import mygale.mcge.net.INetEdge;
import mygale.mcge.net.INetNode;
import mygale.mcge.sel.Selection;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.undoredo.*;
import de.uniwue.ki.util.collection.CollectionsHelper;

/**
 * Cmd to delete Figs from view.
 * <p>
 * This command scan all selected figures and deletes, in the prior order, edges, nodes, other figs. This
 * algorithm permits to correctly managed unselect edges that are attached to a selected node.
 * 
 *  
 */
public class CmdDelete extends CmdUndoable {

    /**
     * This the list of figures that are prior for deletion e.g., edges.
     */
    protected Vector _priorFigs;
    /**
     * This the list of figures that aren't prior for deletion e.g., nodes.
     */
    protected Vector _normalFigs;

    /**
     * This classe permits to undo/redo the deletion of edges
     *  
     */
    class UndoDeleteEdge implements UndoableCmd, RedoableCmd {

        /**
         * A reference to the main deleted figure.
         */
        protected final IFigEdge _figure;

        /**
         * The depth level of the deleted figure.
         */
        protected final int _depthLevel;

        /**
         * The port attached to the begin of the edge.
         */
        protected final INetNode _beginPort;

        /**
         * The port attached to the end of the edge.
         */
        protected final INetNode _endPort;

        /**
         * The net-object attached to the figure.
         */
        protected final INetEdge _net;

        /**
         * Construct a new UndoDeleteEdge.
         * 
         * @param changes
         *            the changes.
         */
        public UndoDeleteEdge(IFigEdge figure) {
            _figure = figure;
            _depthLevel = getGraph().getLayerManager().getDepthLevel(figure);
            _net = (INetEdge) figure.getNetObject();
            _endPort = _net.getBeginNode();
            _beginPort = _net.getEndNode();
        }

        /**
         * This method was called by the undo command. It must undo something.
         * 
         * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
         */
        public void undoIt() {
            JGraph graph = getGraph();
            _figure.setNetObject(_net);
            graph.getFigMapper().setFigFor(_net, _figure);

            _net.setBeginNode(_beginPort);
            _net.setEndNode(_endPort);
            Rectangle damagedBounds = _figure.getBounds();
            graph.add(_figure);
            _figure.updateEndsLocation();
            damagedBounds.add(_figure.getBounds());
            graph.damaged(damagedBounds);
        }

        /**
         * This method was called by the redo command. It must redo something.
         * 
         * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
         */
        public void redoIt() {
            JGraph graph = getGraph();
            Rectangle damagedBounds = _figure.getDamagedBounds(graph.getPrefs());
            graph.remove(_figure);
            graph.damaged(damagedBounds);
        }

    } /* class UndoDeleteEdge */

    /**
     * This classe permits to undo/redo the deletion of figures that are not edges.
     *  
     */
    class UndoDeleteFig implements UndoableCmd, RedoableCmd {

        /**
         * A reference to the main deleted figure.
         */
        protected final IFigLayered _figure;

        /**
         * The depth level of the deleted figure.
         */
        protected final int _depthLevel;

        /**
         * Construct a new UndoDeleteEdge.
         * 
         * @param changes
         *            the changes.
         */
        public UndoDeleteFig(IFigLayered figure) {
            _figure = figure;
            _depthLevel = getGraph().getLayerManager().getDepthLevel(figure);
        }

        /**
         * This method was called by the undo command. It must undo something.
         * 
         * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
         */
        public void undoIt() {
            JGraph graph = getGraph();
            graph.add(_figure);
            graph.damaged(_figure.getBounds());
        }

        /**
         * This method was called by the redo command. It must redo something.
         * 
         * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
         */
        public void redoIt() {
            JGraph graph = getGraph();
            graph.remove(_figure);
            graph.damaged(_figure.getBounds());
        }

    } /* class UndoDeleteFig */

    /**
     * Construct a new CmdDelete.
     * 
     * @param editor
     *            JGraphEditor
     * @param iconSize
     *            int
     */
    public CmdDelete(JGraphEditor editor, int iconSize) {
        super(editor, "Delete", iconSize, "general/Delete");
    }

    protected boolean canSomeUnlockedFigsBeDeleted(Selection sel) {
        Iterator it = sel.getUnlockedFigs();
        while (it.hasNext()) {
            IFigObject currFigObject = (IFigObject) it.next();
            if (currFigObject.getNetObject().canBeRemoved())
                return true;
        }

        return false;
    }

    /**
     * Delete the given edge. This method must update <var>_priorFigs</var> and <var>_normalFigs</var>.
     * 
     * @param edge
     *            a reference to the edge to delete.
     */
    protected void deleteEdge(IFigEdge edge) {
        if (!_priorFigs.contains(edge)) {
            _priorFigs.addElement(new UndoDeleteEdge(edge));
        }
    }

    /**
     * Delete the given figure. This method must update <var>_priorFigs</var> and <var>_normalFigs</var>.
     * 
     * @param figure
     *            a reference to the figure to delete.
     */
    protected boolean deleteFig(IFigLayered fig) {
        
        if (fig instanceof IFigObject) {
            if (!((IFigObject) fig).getNetObject().canBeRemoved()) {
                return false;
            } else if (fig instanceof IFigNode) {
                deleteNode((IFigNode) fig);
            } else if (fig instanceof IFigEdge) {
                deleteEdge((IFigEdge) fig);
            }
            return true;

        } else {
            deleteFigIntern(fig);
            return true;
        }
    }

    private void deleteFigIntern(IFigLayered fig) {
        if (!_normalFigs.contains(fig)) {
            _normalFigs.addElement(new UndoDeleteFig(fig));
        }
    }

    /**
     * Delete the given node. This method must update <var>_priorFigs</var> and <var>_normalFigs</var>.
     * 
     * @param node
     *            a reference to the node to delete.
     */
    protected void deleteNode(IFigNode node) {
        
        IFigMapper figMapper = getGraph().getFigMapper();
       
        Iterator it = ((INetNode) node.getNetObject()).getNodeEdges();
        while (it.hasNext()) {
            INetEdge netEdge = (INetEdge) it.next();
            deleteEdge((IFigEdge) figMapper.getFigFor(netEdge));
        }
        
        deleteFigIntern(node);
    }

    /**
     * Perform whatever Cmd this Cmd is meant to do. Subclasses should override this to do whatever is
     * intended.
     * 
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     */
    public void doIt() {
        
        JGraph graph = getGraph();
        Selection sel = graph.getSelection();
        
        if (sel.unlockedSize() <= 0) {
            throw new MCGEInternalError("try to delete an empty selection.");
        } else {
            boolean oldInform = sel.setCmdInform(false);
            Iterator it = CollectionsHelper.asLinkedList(sel.getUnlockedFigs()).iterator();
            _priorFigs = new Vector();
            _normalFigs = new Vector();

            while (it.hasNext()) {
                IFigLayered fig = (IFigLayered) it.next();
                // Prepare for the undoing
                if (deleteFig(fig)) {
                    // do the deletion
                    graph.remove(fig);
                    sel.deselect(fig);
                }
            }

            Rectangle r = sel.getBounds();
            graph.damaged(r);
            sel.setCmdInform(oldInform);
            sel.informCmd();
            // Save undoing commands
            CommandSet undoCmds = new CommandSet();
            undoCmds.add(_priorFigs);
            undoCmds.add(_normalFigs);
            addUndoCmd(undoCmds);
        }
    }

    /**
     * Return the tooltip text associated to the command.
     * 
     * @return a string that is the tooltip of this command.
     */
    public String getToolTipText() {
        return "delete the selection";
    }

    /**
     * Return <code>true</code> if this action should be available to the user. Sublass implementations of
     * this method should always call <code>super.shouldBeEnabled()</code> first.
     * 
     * @return <code>true</code> if this command is enabled, otherwise <code>false</code>.
     */
    public boolean shouldBeEnabled() {
        JGraph graph = getGraph();

        // at initializing
        if (getGraph() == null)
            return false;
        else {
            return canSomeUnlockedFigsBeDeleted(graph.getSelection());
        }
    }
}