package mygale.mcge.cmd;

import java.awt.Rectangle;
import java.util.*;

import mygale.mcge.core.*;
import mygale.mcge.fig.*;
import mygale.mcge.sel.Selection;
import mygale.mcge.undoredo.*;

/** Abstract Cmd to move to the back a set of
 *  {@link mygale.mcge.fig.Fig figures}.
 *  <p><center>
 *    <img src="doc-files/CmdMoveFigToBack_uml.gif" width="111" height="83" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="10,0,100,28"  href="CmdUndoable.html">
 *    </map>
 *  </center></p>
 *
 */
public abstract class CmdMoveFigToBack extends CmdUndoable {

	/** This is the list of changes.
	 */
	protected Vector _changes = new Vector();

	/** This class permits to describe a change.
	 *
		 */
	protected static class DepthChange {

		/** A reference to the changed figure.
		 */
		public final IFigLayered figure;

		/** Old depth of the changed figure.
		 */
		public final int oldDepth;

		/** New depth of the changed figure.
		 */
		public final int newDepth;

		/** Construct a new DepthChange.
		 *
		 * @param fig a reference to the changed figure.
		 * @param od old depth of the given figure.
		 * @param nd new depth of the given figure.
		 */
		public DepthChange(IFigLayered fig, int od, int nd) {
			figure = fig;
			oldDepth = od;
			newDepth = nd;
		}

	}

	/** This class permits to directly managed undo/redo commands.
	 *
		 */
	protected class UndoMove implements UndoableCmd, RedoableCmd {

		/** This is changes.
		 */
		protected Vector _changes;

		/** Construct a new UndoMove.
		 *
		 * @param changes the changes.
		 */
		public UndoMove(Vector changes) {
			_changes = changes;
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 *
		 * throws mygale.mcge.throwable.MCGEInternalError Something wrong !
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			JGraph graph = getGraph();

			Rectangle rFigs = null;
			Enumeration enum = _changes.elements();
			while (enum.hasMoreElements()) {
				DepthChange chgt = (DepthChange)enum.nextElement();
				if (rFigs == null)
					rFigs = new Rectangle(chgt.figure.getBounds());
				else
					rFigs.add(chgt.figure.getBounds());
				graph.getLayerManager().setDepthLevel(chgt.figure, chgt.oldDepth);
			}
			graph.damaged(rFigs);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 *
		 * throws mygale.mcge.throwable.MCGEInternalError Something wrong !
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			JGraph graph = getGraph();

			Rectangle rFigs = null;
			Enumeration enum = _changes.elements();
			while (enum.hasMoreElements()) {
				DepthChange chgt = (DepthChange)enum.nextElement();
				if (rFigs == null)
					rFigs = new Rectangle(chgt.figure.getBounds());
				else
					rFigs.add(chgt.figure.getBounds());
				graph.getLayerManager().setDepthLevel(chgt.figure, chgt.newDepth);
			}
			graph.damaged(rFigs);
		}

	} /* class UndoMove */

	/** Return the tooltip text associated to the command.
	     *
	     * @return a string that is the tooltip of this command.
	     */
	public String getToolTipText() {
		return "set the figure to back";
	}

	/** Return <code>true</code> if this action should be available
	     *  to the user. Sublass implementations of this method should
	     *  always call <code>super.shouldBeEnabled()</code> first.
	     *
	     * @return <code>true</code> if this command is enabled,
	     *         otherwise <code>false</code>.
	     */
	public boolean shouldBeEnabled() {
		JGraph graph = getGraph();

		// at initializing
		if (getGraph() == null)
			return false;
		else {
			enabled = true;
			Selection sel = graph.getSelection();
			if ((!isCorrectSize(sel.unlockedSize())) || (isBottomWinget(graph, sel))) {
				enabled = false;
			}
			return enabled && super.shouldBeEnabled();
		}
	}

	/** Add this command to the undo list.
	     */
	protected void addUndoCmd() {
		if (_changes.size() > 0) {
			super.addUndoCmd(new UndoMove(_changes));
		}
	}

	/** Destroy the nudo/redo changes.
	     */
	protected void clearUndoChanges() {
		_changes = new Vector();
	}

	/** Reply if the selection contains the bottom winget.
	     *  The bottom winget is the
	     *  {@link mygale.mcge.fig.Fig figure} that has the
	     *  higher depth.
	     *
	     * @param graph a reference to the graph editor.
	     * @param sel a reference to the current selection.
	     * @return <code>true</code> if the given selection
	     *         contains a figure that is on the background,
	     *         otherwise <code>false</code>.
	     */
	protected boolean isBottomWinget(JGraph graph, Selection sel) {
		if (sel.size() == 1) {
			IFig selFig = (IFig)sel.getAllFigs().next();
			Vector figs = graph.getLayerManager().getContents();
			IFig topFig = (IFig)figs.elementAt(0);
			return topFig == selFig;
		}
		return false;
	}

	/** Reply if the given size corresponds to
	     *  an valid selection size.
	     *
	     * @param size the size to test.
	     * @return <code>true</code> if the given size is
	     *         a valid size for the current command.
	     */
	protected abstract boolean isCorrectSize(int size);

	/** This method permit to register (from doIt() for example)
	 *  a change.
	 *
	 * @param fig a reference to the changed figure.
	 * @param oldDepth old depth of the given figure.
	 * @param newDepth new depth of the given figure.
	 */
	protected void saveChange(IFigLayered fig, int oldDepth, int newDepth) {
		_changes.addElement(new DepthChange(fig, oldDepth, newDepth));
	}

	///

	/** Construct a new CmdMoveFigToBack.
	 *
	 * @param editor a reference to the current editor.
	 * @param name String
	* @param iconSize int
	* @param iconName the path to the associated icon.
	 */
	public CmdMoveFigToBack(JGraphEditor editor, String name, int iconSize, String iconName) {
		super(editor, name, iconSize, iconName);
	}
}
