package mygale.mcge.core;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.*;

import junit.framework.Assert;
import mygale.mcge.cmd.*;
import mygale.mcge.event.mouse.MCGEMouseManager;
import mygale.mcge.fig.*;
import mygale.mcge.layer.*;
import mygale.mcge.mode.*;
import mygale.mcge.net.*;
import mygale.mcge.sel.Selection;
import mygale.mcge.sel.SelectionImpl;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.util.prefs.*;
import de.uniwue.ki.gui.util.IDisposeable;
import de.uniwue.ki.gui.util.JDisposeablePanel;

/**
 * This class provides an editor for manipulating graphical network The editor (JGraph) is the central class
 * of the graph editing framework. All the net-level models, graphical objects, layers, editor modes, editor
 * commands, and supporting dialogs and frames are implemented in their own classes.
 * <p>
 * An {@link mygale.mcge.layer.LayerManager Editor's LayerManager}has a stack of
 * {@link mygale.mcge.layer.Layer Layer's}. Normally Layers contain {@link Fig Figs}. Some Figs are linked
 * to {@link NetObject NetObjects}. When Figs are selected, they are stored in a
 * {@link mygale.mcge.sel.Selection selection manager}. The behavior of the editor is determined by its
 * current {@link mygale.mcge.mode.Mode Mode}. The {@link mygale.mcge.mode.ModeManager editor's ModeManager}
 * keeps track of all the active Modes. Modes interpert user input events and decide how to change the state
 * of the diagram. The Editor acts as a shell for executing {@link mygale.mcge.cmd.Cmd Commands}that modify
 * the document or the Editor itself.
 * <p>
 * When Figs change visible state (e.g., color, size, or postition) they tell their
 * {@link mygale.mcge.layer.Layer Layer}that they are damaged and need to be repainted. The Layer tells to
 * the editors that are editing that Layer. Each editor has a RedrawManager that keeps track of what parts of
 * the screen need to be redrawn, and does the redraws asynchronously.
 * <p>
 * <center><img src="doc-files/JGraph_uml.gif" width="453" height="250" border="0" usemap="#umlclick">
 * <map name="umlclick">
 * <area shape="Rect" coords="73,0,172,28"  href="../util/prefs/PrefChangedListener.html">
 * <area shape="Rect" coords="185,0,257,28"  href="http://java.sun.com/products/jdk/1.2/docs/api/javax/swing/JPanel.html">
 * <area shape="Rect" coords="334,0,406,28"  href="JGraphEditor.html">
 * <area shape="Rect" coords="0,82,73,109"  href="../sel/Selection.html">
 * <area shape="Rect" coords="334,41,406,68"  href="../layer/LayerManager.html">
 * <area shape="Rect" coords="0,41,73,68"  href="JGraphToolBar.html">
 * <area shape="Rect" coords="334,91,406,118"  href="../util/prefs/Prefs.html">
 * <area shape="Rect" coords="0,141,100,167"  href="../net/NetGraphModel.html">
 * <area shape="Rect" coords="104,190,176,218"  href="../mode/ModeManager.html">
 * <area shape="Rect" coords="334,141,452,167"  href="../event/mouse/MCGEMouseManager.html">
 * <area shape="Rect" coords="185,221,257,249"  href="../net/NetObject.html">
 * <area shape="Rect" coords="271,221,343,249"  href="../cmd/Cmd.html">
 * <area shape="Rect" coords="353,185,424,213"  href="../fig/Fig.html"></map></center>
 * </p>
 *  
 */
public class JGraph extends JDisposeablePanel implements PrefChangedListener, IDisposeable {

    private boolean _inNetGraphEventHandling = false;

    private final NetGraphModelChangedListener _netGraphModelChangedListener =
        new NetGraphModelChangedListener() {

        public void addedNode(NetGraphModelChangedEvent event) {
            if (!_inNetGraphEventHandling && (event.getSource() == JGraph.this.getNetGraphModel())) {
                try {
                    _inNetGraphEventHandling = true;

                    addFigForNode(event.getNetNode());
                } finally {
                    _inNetGraphEventHandling = false;
                }
            }
        }

        public void removedNode(NetGraphModelChangedEvent event) {
            if (!_inNetGraphEventHandling && (event.getSource() == JGraph.this.getNetGraphModel())) {
                try {
                    _inNetGraphEventHandling = true;

                    removeFigForNode(event.getNetNode());
                } finally {
                    _inNetGraphEventHandling = false;
                }
            }
        }

        public void addedEdge(NetGraphModelChangedEvent event) {
            if (!_inNetGraphEventHandling && (event.getSource() == JGraph.this.getNetGraphModel())) {
                try {
                    _inNetGraphEventHandling = true;

                    addFigForEdge(event.getNetEdge());
                } finally {
                    _inNetGraphEventHandling = false;
                }
            }
        }
        public void removedEdge(NetGraphModelChangedEvent event) {
            if (!_inNetGraphEventHandling && (event.getSource() == JGraph.this.getNetGraphModel())) {
                try {
                    _inNetGraphEventHandling = true;

                    removeFigForEdge(event.getNetEdge());
                } finally {
                    _inNetGraphEventHandling = false;
                }
            }
        }

    };

    /**
     * The size that will be added to the workspace each time it mustbe resized.
     */
    public static final int WORKSPACE_NOMANLAND = 10;

    ////
    // instance variables

    /**
     * The scale at which to draw the diagram.
     */
    private double _scale = 1.0;

    /**
     * The toolbars for this JGraph.
     */
    private Component _toolbar;

    /**
     * The {@link JGraphEditor JGraphEditor}that contains this object.
     */
    private JGraphEditor _graphEditor;

    /**
     * The user interface mode that the JGraph is currently in. Generally {@link mygale.mcge.mode.Mode Modes}
     * that the user has to think about are a bad idea. But even in a very easy to use editor there are
     * plenty of "spring-loaded" modes that change the way the system interperts input. For example, when
     * placing a new node, the editor is in {@link mygale.mcge.mode.ModeCreateNode ModeCreateNode}, and when
     * dragging a handle of an node the JGraph is in {@link mygale.mcge.mode.ModeReshape ModeReshape}. In
     * each case moving or dragging the mouse has a different effect.
     * 
     * @see mygale.mcge.mode.Mode
     */
    private ModeManager _modeManager;

    /**
     * The {@link mygale.mcge.mode.ModeManager ModeManager}needs to managed specific
     * {@link mygale.mcge.event.mouse.MCGEMouseEvent mouse events}. This JGraph redirect all Swing mouse
     * events to a specific {@link mygale.mcge.event.mouse.MCGEMouseManager mouse manager}.
     */
    private MCGEMouseManager _mouseManager;

    /**
     * The {@link mygale.mcge.layer.LayerManager LayerManager}for this editor.
     */
    private LayerManager _layerManager;

    /**
     * The {@link mygale.mcge.sel.Selection selection manager}for this editor.
     */
    private Selection _selection;

    /**
     * The {@link mygale.mcge.util.prefs.Prefs preference manager}for this editor.
     */
    private Prefs _prefsManager;

    /**
     * The origin of the workspace area. This point permits to manager "infinite" workspace.
     */
    private Point _originPoint = new Point(0, 0);

    /**
     * This is the instance of a {@link NetGraphModel graph diagram}.
     */
    private INetGraphModel _netGraphModel = null;

    /**
     * Factory for creating new FigNode instances.
     */
    protected final IFigNodeFactory _figNodeFactory;

    /**
     * Factory for creating new FigEdge instances.
     */
    protected final IFigEdgeFactory _figEdgeFactory;

    /**
     * Maps Nodes to Figs
     */
    protected IFigMapper _figMapper; // better throw a nullpointerexception = new DefaultFigMapper();

    /// Constructor

    /**
     * Construct a new JGraph.
     * 
     * @link JGraphEditor graph editor}.
     * 
     * @param graphEditor
     *            a reference to the enclosing { *
     * @param figNodeFactory
     *            IFigNodeFactory
     * @param figPortFactory
     *            IFigPortFactory
     * @param figEdgeFactory
     *            IFigEdgeFactory
     */
    public JGraph(JGraphEditor graphEditor, IFigNodeFactory figNodeFactory, IFigEdgeFactory figEdgeFactory) {

        super();

        _graphEditor = graphEditor;
        _figNodeFactory = figNodeFactory;
        _figEdgeFactory = figEdgeFactory;

        initialize();
    }

    /**
     * Add a {@link Fig Fig}into the diagram being edited.
     * <p>
     * We assume that the given figure are completly and correctly connected to a
     * {@link mygale.mcge.net.NetObject NetObject}if it is as subclass of
     * {@link FigModelElement FigModelElement}.
     * 
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     * @param fig
     *            a reference to the new figure.
     */
    public void add(IFigLayered fig) {
        if (!(fig instanceof IFigObject)) {
            _layerManager.add(fig);
        } else {
            INetObject netobject = ((IFigObject) fig).getNetObject();
            getFigMapper().setFigFor(netobject, (IFigObject) fig);
            if (netobject instanceof INetNode) {
                add((INetNode) netobject);
            } else if (netobject instanceof INetEdge) {
                add((INetEdge) netobject);
            } else {
                throw new MCGEInternalError("Cannot add a NetObject of type " + netobject.getClass());
            }
        }
    }

    /**
     * Add a {@link mygale.mcge.net.NetEdge NetEdge}into the diagram being edited.
     * <p>
     * We assume that the given net-object are completly and correctly connected to a
     * {@link mygale.mcge.figEdge.FigObject FigObject}.
     * 
     * @param netnode
     *            a reference to the new net-object.
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     */
    protected void add(INetEdge netEdge) {
        try {
            _inNetGraphEventHandling = true;

            getNetGraphModel().addEdge(netEdge);
        } finally {
            _inNetGraphEventHandling = false;
        }

        addFigForEdge(netEdge);
    }

    /**
     * Add a {@link mygale.mcge.net.NetNode NetNode}into the diagram being edited.
     * <p>
     * We assume that the given net-object are completly and correctly connected to a
     * {@link mygale.mcge.figNode.FigObject FigObject}.
     * 
     * @param netnode
     *            a reference to the new net-object.
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     */
    protected void add(INetNode netnode) {
        try {
            _inNetGraphEventHandling = true;

            getNetGraphModel().addNode(netnode);
        } finally {
            _inNetGraphEventHandling = false;
        }

        addFigForNode(netnode);
    }

    /**
     * Add a {@link mygale.mcge.net.NetEdge NetEdge}into the diagram being edited.
     * <p>
     * We assume that the given net-object are completly and correctly connected to a
     * {@link mygale.mcge.figEdge.FigObject FigObject}.
     * 
     * @param netnode
     *            a reference to the new net-object.
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     */
    protected void addFigForEdge(INetEdge netedge) {
        IFigEdge figEdge = createFigEdge(netedge);

        getLayerManager().add(figEdge);
    }

    /**
     * Add a {@link mygale.mcge.net.NetNode NetNode}into the diagram being edited.
     * <p>
     * We assume that the given net-object are completly and correctly connected to a
     * {@link mygale.mcge.figNode.FigObject FigObject}.
     * 
     * @param netnode
     *            a reference to the new net-object.
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     */
    protected void addFigForNode(INetNode netnode) {
        addFigForNode(netnode, true);
    }

    protected void addFigForNode(INetNode netnode, boolean withAddingEdges) {
        IFigNode figNode = createFigNode(netnode);

        if (withAddingEdges) {
            // add all attached edges
            Iterator it = netnode.getNodeEdges();
            while (it.hasNext()) {
                addFigForEdge((INetEdge) it.next());
            }
        }

        getLayerManager().add(figNode);
    }

    /**
     * Set the current user interface mode to the given {@link mygale.mcge.mode.Mode Mode instance}.
     * 
     * @param m
     *            the mode to begin.
     */
    public void beginMode(Mode m) {
        _modeManager.beginMode(m);
    }

    /**
     * Change the workspace area if necessary.
     * <p>
     * This method is called each time the workspace could be extended in one of the four cardinal
     * directions. The size of the workspace and/or the origin point are updated if the given Rectangle is
     * outside the current workspace.
     * 
     * @param rectOrigin
     *            Rectangle
     */
    public void changeWorkspaceSize(Rectangle rectOrigin) {

        synchronized (_originPoint) {
            Rectangle workSpaceRect = getWorkspaceRect();

            Rectangle rect = new Rectangle(rectOrigin);
            rect.grow(5, 5);

            if (!workSpaceRect.contains(rect)) {
                workSpaceRect.add(rect);

                setWorkspaceRect(workSpaceRect);
            }
        }
    }

    /**
     * This method permits to clear the content of a status bar This bar is not directly managed by the MCGE.
     */
    public void clearStatus() {
        if (getGraphEditor() != null)
            getGraphEditor().clearStatus();
    }

    public IFigEdge createFigEdge(INetEdge netEdge) {
        return createFigEdge(netEdge, getFigEdgeFactory());
    }

    public IFigEdge createFigEdge(INetEdge netEdge, IFigEdgeFactory figEdgeFactory) {
        return FigHelper.createFigEdge(getFigMapper(), netEdge, figEdgeFactory);
    }

    public IFigNode createFigNode(INetNode netNode) {
        return createFigNode(netNode, getFigNodeFactory());
    }

    public IFigNode createFigNode(INetNode netNode, IFigNodeFactory figNodeFactory) {
        return FigHelper.createFigNode(getFigMapper(), netNode, figNodeFactory);
    }

    ////
    // recording damage to the display for later repair

    /**
     * Calling any one of the following damaged() methods adds a damaged region (rectangle) to this Editor's
     * RedrawManager. If the given rectangle is <code>null</code> the entire drawing area was damaged.
     * 
     * @param rect
     *            Rectangle
     */
    public void damaged(Rectangle rect) {
        if (rect == null) {
            damagedAll();

        } else {
            repaint(
                transformFromWorkSpacePointX(rect.x) - 1,
                transformFromWorkSpacePointY(rect.y) - 1,
                transformFromWorkSpacePointX(rect.x + rect.width) + 1,
                transformFromWorkSpacePointY(rect.y + rect.height) + 1);
        }
    }

    /**
     * Calling any one of the following damaged() methods adds a damaged region (the bounds of the given
     * {@link Fig figure}) to this Editor's RedrawManager. If the given figure is <code>null</code> the
     * entire drawing area was damaged.
     * 
     * @param fig
     *            a figure that must be refreshed
     */
    public void damaged(IFig fig) {
        if (fig != null) {
            damaged(fig.getDamagedBounds(getPrefs()));

        } else {
            damagedAll();
        }
    }

    /**
     * Calling any one of the following damaged() methods adds a damaged region (the bounds of the given
     * {@link mygale.mcge.sel.Selection selection}) to this Editor's RedrawManager.
     * <p>
     * The bounds of a selection is calculate by the union of all selected {@link Fig figures}.
     * 
     * @param sel
     *            a selection that must be refreshed
     */
    public void damaged(Selection sel) {
        damaged(sel.getBounds());
    }

    /**
     * Mark the entire visible area of this Editor as damaged. <p Currently called when a LayerGrid is
     * adjusted. This will be useful for ActionRefresh if I get around to it. Also some Actions may perfer to
     * do this instead of keeping track of all modified objects, but only in cases where most of the visible
     * area is expected to change anyway.
     */
    public final void damagedAll() {
        repaint();
    }

    ///
    // Debugging API

    /**
     * Print the contents of the graph model.
     */
    public void debugPrintNetDiagram() {
        try {
            // Display graph structure
            System.out.println("Contents of the graph diagram");
            Iterator itNodes = getNetGraphModel().getNodes();
            Vector displayedEdges = new Vector();
            while (itNodes.hasNext()) {
                INetNode node = (INetNode) itNodes.next();
                System.out.println(
                    "{ NetNode at " + ((IFigLayered) getFigMapper().getFigFor(node)).getBounds());
                Iterator itEdges = node.getNodeEdges();
                while (itEdges.hasNext()) {
                    INetEdge edge = (INetEdge) itEdges.next();
                    System.out.println("\t edge: " + edge.toString());
                    displayedEdges.addElement(edge);
                }
                System.out.println("}");
            }
            // Display undisplay edges
            Iterator itEdges2 = getNetGraphModel().getEdges();
            while (itEdges2.hasNext()) {
                INetEdge edge = (INetEdge) itEdges2.next();
                if (!displayedEdges.contains(edge)) {
                    System.out.println("undisplayed edge: " + edge.toString());
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * @see de.uniwue.ki.gui.util.IDisposeable#dispose()
     */
    public void dispose() {

        setNetGraphModel(null);

        _prefsManager.removePrefChangedListener(this);

        super.dispose();

        // after super.dispose since:
        Assert.assertTrue(isDisposed());
        if ((getGraphEditor() != null) && (!getGraphEditor().isDisposed())) {
            getGraphEditor().dispose();
        }
    }

    /**
     * Returns a collection of all {@link Fig Figs}in the {@link mygale.mcge.layer.Layer layer}currently
     * being edited.
     * 
     * @return an Iterator of all figures.
     */
    public Iterator getAllFigs() {
        return _layerManager.getAllFigs();
    }

    public final IFigEdgeFactory getFigEdgeFactory() {
        return _figEdgeFactory;
    }

    protected IFigObject getFigFor(INetObject anNetObject) {
        return getFigMapper().getFigFor(anNetObject);
    }

    public IFigMapper getFigMapper() {
        return _figMapper;
    }

    public final IFigNodeFactory getFigNodeFactory() {
        return _figNodeFactory;
    }

    ////
    /// methods related to editor state: graphical attributes, modes, view

    /**
     * Return the {@link JGraphEditor JGraphEditor}that encloses this JGraph.
     * 
     * @return JGraphEditor
     */
    public JGraphEditor getGraphEditor() {
        return _graphEditor;
    }

    /**
     * Return the {@link mygale.mcge.layer.LayerManager LayerManager}of the JGraph.
     * 
     * @return a reference to the layer manager.
     */
    public LayerManager getLayerManager() {
        return _layerManager;
    }

    /**
     * Return the {@link mygale.mcge.mode.ModeManager ModeManger}of the JGraph.
     * 
     * @return a reference to the mode manager.
     */
    public ModeManager getModeManager() {
        return _modeManager;
    }

    /**
     * Return the {@link mygale.mcge.net.NetGraphModel NetGraphModel}.
     * 
     * @return a reference to the graph model.
     */
    public INetGraphModel getNetGraphModel() {
        return _netGraphModel;
    }

    ///
    // PopupMenu

    /**
     * This method return the popup menu of the editor.
     * 
     * @return a reference to a popup menu.
     */
    public JPopupMenu getPopupMenu() {
        JPopupMenu popupMenu = new JPopupMenu();
        popupMenu.add(new CmdSelectAll(getGraphEditor()));
        popupMenu.add(new CmdDeselectAll(getGraphEditor()));
        return popupMenu;
    }

    /**
     * This method return the preference manager.
     * 
     * @return the current preference manager.
     */
    public Prefs getPrefs() {
        return _prefsManager;
    }

    /*********************************************************************************************************
     * Return this Editor's current display scale.
     */
    public final double getScale() {
        return _scale;
    }

    /**
     * Return the {@link mygale.mcge.sel.Selection Selection}of the JGraph.
     * 
     * @return a reference to the selection manager.
     */
    public Selection getSelection() {
        return _selection;
    }

    /**
     * Reply the rectangle that represents the workspace.
     * 
     * @preturn a Rectangle that correspond to the drawing workspace size.
     * 
     * @return Rectangle
     */
    public Rectangle getWorkspaceRect() {
        Dimension preferredSize = super.getPreferredSize();
        Dimension size = super.getSize();

        return new Rectangle(
            transformToWorkSpacePointX(0),
            transformToWorkSpacePointY(0),
            transformToWorkSpacePointX((int) (Math.rint(Math.max(size.width, preferredSize.width)))),
            transformToWorkSpacePointY((int) (Math.rint(Math.max(size.height, preferredSize.height)))));
    }

    /**
     * Reply the top Fig in the current layer that contains the given point. This is used in determining what
     * the user clicked on, among other uses.
     * 
     * @param x
     *            x coordinate.
     * @param y
     *            y coordinate.
     * @return the topless figure under the position (<var>x</var>,<var>y</var>).
     */
    public final IFigLayered hit(int x, int y) {
        return hit(new Point(x, y));
    }

    /**
     * Reply the top Fig in the current layer that contains the given point. This is used in determining what
     * the user clicked on, among other uses.
     * 
     * @param p
     *            a position.
     * @return the topless figure under the position <var>p</var>.
     */
    public IFigLayered hit(Point p) {
        return _layerManager.hit(p);
    }

    private void initialize() {
        // Initiate JGraph components
        makePrefsManager();
        makeLayerManager();
        makeGraphDiagram();
        makeSelection();
        makeModeManager();
        makeMouseManager();
        makeToolBar();
        // Set the Layout and event managing
        // setLayout(null);
        addMouseListener(_mouseManager);
        addMouseMotionListener(_mouseManager);
        addKeyListener(_modeManager);
        // init components
        _mouseManager.beginTrackFigure();
        // finalize the construction
        setAutoscrolls(true);
        setFocusCycleRoot(true);
        validate();
    }

    private void insertEdgeFigsInLayer(INetGraphModel netGraphModel) {
        java.util.Iterator it = netGraphModel.getEdges();
        while (it.hasNext()) {
            addFigForEdge((INetEdge) it.next());
        }
    }

    protected void insertFigsInLayer(INetGraphModel netGraphModel) {
        insertNodeFigsInLayer(netGraphModel);
        insertEdgeFigsInLayer(netGraphModel);
    }

    private void insertNodeFigsInLayer(INetGraphModel netGraphModel) {
        Iterator it = netGraphModel.getNodes();

        while (it.hasNext()) {
            addFigForNode((INetNode) it.next(), false);
        }
    }

    /**
     * Overwrite this method to customize the default toolbar "Alignment".
     * <p>
     * We assume that <a href="#_tabbedToolBar">_tabbedToolBar</a> is already set before the call to rthis
     * method.
     */
    protected void makeAlignmentToolBar() {
        //Create commands
        JToolBar toolbar = createAlignmentToolBar();
        // Add the toolbar to this editor
         ((JTabbedPane) getToolbar()).add("Alignment", toolbar);
    }

    protected JToolBar createAlignmentToolBar() {
        CmdSelect cmdSelect = new CmdSelect(getGraphEditor(), 24);
        CmdAlignLeft cmdAlignLeft = new CmdAlignLeft(getGraphEditor(), 24);
        CmdAlignHCenter cmdAlignHCenter = new CmdAlignHCenter(getGraphEditor(), 24);
        CmdAlignRight cmdAlignRight = new CmdAlignRight(getGraphEditor(), 24);
        CmdAlignTop cmdAlignTop = new CmdAlignTop(getGraphEditor(), 24);
        CmdAlignVCenter cmdAlignVCenter = new CmdAlignVCenter(getGraphEditor(), 24);
        CmdAlignBottom cmdAlignBottom = new CmdAlignBottom(getGraphEditor(), 24);
        // Register command in the selection manager.
        // Each time the selection changed, these commands
        // was updated.
        _selection.addCmdToInform(cmdSelect);
        _selection.addCmdToInform(cmdAlignLeft);
        _selection.addCmdToInform(cmdAlignHCenter);
        _selection.addCmdToInform(cmdAlignRight);
        _selection.addCmdToInform(cmdAlignTop);
        _selection.addCmdToInform(cmdAlignVCenter);
        _selection.addCmdToInform(cmdAlignBottom);
        //Create the toolbar
        JToolBar toolbar = new JGraphToolBar();
        toolbar.setFloatable(false);
        // Add commands
        toolbar.add(cmdSelect);
        toolbar.addSeparator();
        toolbar.add(cmdAlignLeft);
        toolbar.add(cmdAlignHCenter);
        toolbar.add(cmdAlignRight);
        toolbar.addSeparator();
        toolbar.add(cmdAlignTop);
        toolbar.add(cmdAlignVCenter);
        toolbar.add(cmdAlignBottom);
        return toolbar;
    }

    ////
    // methods related to adding, removing, and accessing Figs
    // shown in the editor

    /**
     * Create an instance of the {@link mygale.mcge.net.NetGraphModel graph diagram}.
     * <p>
     * We assume that <a href="#_graphdiagram">_graphdiagram</a> is set in this method.
     * <p>
     * By default, this method create an {@link mygale.mcge.net.NetGraphModelImpl NetGraphModelImpl}.
     */
    protected void makeGraphDiagram() {
        setNetGraphModel(new DefaultNetGraphModel());
    }

    /**
     * Overwrite this method to customize the {@link mygale.mcge.layer.LayerManager JGraph LayerManager}.
     * This method perform also an initialization of the LayerManger.
     * <p>
     * We assume that <a href="#_layerManager">_layerManager</a> is set in this method.
     * <p>
     * Default layers are {@link mygale.mcge.layer.LayerGrid LayerGrid}and
     * {@link mygale.mcge.layer.LayerDiagram LayerDiagram}.
     */
    protected void makeLayerManager() {
        _layerManager = new LayerManager(this);
        _layerManager.addLayer(new LayerGrid(this));
        _layerManager.addLayer(new LayerDiagram(this));
    }

    /**
     * Overwrite this method to customize the {@link mygale.mcge.mode.ModeManager JGraph ModeManager}. This
     * method perform also an initialization of the ModeManger.
     * <p>
     * We assume that <a href="#_modeManager">_modeManager</a> is set in this method.
     * <p>
     * Default modes are {@link mygale.mcge.mode.ModeSelect ModeSelect},
     * {@link mygale.mcge.mode.ModeReshape ModeReshape},{@link mygale.mcge.mode.ModePopup ModePopup},
     * {@link mygale.mcge.mode.ModeEditProperties ModeEditProperties}and
     * {@link mygale.mcge.mode.ModeScrollDrawingArea ModeScrollDrawingArea}.
     */
    protected void makeModeManager() {
        _modeManager = new ModeManager(this);
        _modeManager.beginMode(new ModeSelect(_modeManager));
        _modeManager.beginMode(new ModeReshape(_modeManager));
        _modeManager.beginMode(new ModePopup(_modeManager));
        _modeManager.beginMode(new ModeEditProperties(_modeManager));
        _modeManager.beginMode(new ModeScrollDrawingArea(_modeManager));
        _modeManager.beginMode(new ModeDelete(_modeManager));
    }

    /**
     * Overwrite this method to customize the
     * {@link mygale.mcge.event.mouse.MCGEMouseManager JGraph MouseManager}. This method perform also an
     * initialization of the MouseManger.
     * <p>
     * We assume that <a href="#_mouseManager">_mouseManager</a> is set in this method.
     */
    protected void makeMouseManager() {
        _mouseManager = new MCGEMouseManager(this, _modeManager);
    }

    /**
     * Overwrite this method to customize the default toolbar "Personalized".
     * <p>
     * We assume that <a href="#_tabbedToolBar>_tabbedToolBar</a> is already set before the call to rthis
     * method.
     * <p>
     * By default this method do nothing.
     */
    protected void makePersonalizedToolBar() {
    }

    /// intialisation of the differents JGraph components

    /**
     * Overwrite this method to customize the preference manager.
     * <p>
     * We assume that {@link #_prefsManager _prefManager}is set in this method.
     * <p>
     * The default preference manager is a {@link mygale.mcge.util.prefs.Prefs Prefs}.
     */
    protected void makePrefsManager() {
        _prefsManager = new Prefs();

        _prefsManager.addPrefChangedListener(this);
    }

    /**
     * Overwrite this method to customize the {@link mygale.mcge.sel.Selection JGraph Selection}. This
     * method perform also an initialization of the Selection.
     * <p>
     * We assume that <a href="#_selection">_selection</a> is set in this method.
     */
    protected void makeSelection() {
        _selection = new SelectionImpl(this);
    }

    /**
     * Overwrite this method to customize the default toolbar "Standard".
     * <p>
     * We assume that <a href="#_tabbedToolbar">_tabbedToolBar</a> is already set before the call to rthis
     * method.
     */
    protected void makeStandardToolBar() {
        // Create commands
        CmdLock cmdLock = new CmdLock(getGraphEditor(), 24);
        CmdUnlock cmdUnlock = new CmdUnlock(getGraphEditor(), 24);
        CmdDelete cmdDelete = new CmdDelete(getGraphEditor(), 24);
        CmdBringToFront cmdBringToFront = new CmdBringToFront(getGraphEditor(), 24);
        CmdSendToBack cmdSendToBack = new CmdSendToBack(getGraphEditor(), 24);
        CmdBringForward cmdBringForward = new CmdBringForward(getGraphEditor(), 24);
        CmdSendBackward cmdSendBackward = new CmdSendBackward(getGraphEditor(), 24);
        CmdUndo cmdUndo = new CmdUndo(getGraphEditor().getUndoRedoManager(), 24);
        CmdRedo cmdRedo = new CmdRedo(getGraphEditor().getUndoRedoManager(), 24);
        // Register command in the selection manager.
        // Each time the selection changed, these commands
        // was updated.
        _selection.addCmdToInform(cmdBringToFront);
        _selection.addCmdToInform(cmdSendToBack);
        _selection.addCmdToInform(cmdBringForward);
        _selection.addCmdToInform(cmdSendBackward);
        _selection.addCmdToInform(cmdLock);
        _selection.addCmdToInform(cmdUnlock);
        _selection.addCmdToInform(cmdDelete);
        // Create the toolbar
        JToolBar toolbar = new JGraphToolBar();
        toolbar.setFloatable(false);
        // Add commands
        toolbar.add(new CmdSelect(getGraphEditor(), 24));
        //toolbar.addSeparator() ;
        toolbar.add(cmdDelete);
        toolbar.addSeparator();
        toolbar.add(cmdLock);
        toolbar.add(cmdUnlock);
        toolbar.addSeparator();
        toolbar.add(cmdBringToFront);
        toolbar.add(cmdSendToBack);
        toolbar.add(cmdBringForward);
        toolbar.add(cmdSendBackward);
        toolbar.addSeparator();
        toolbar.add(cmdUndo);
        toolbar.add(cmdRedo);
        // Add the toolbar into this editor
         ((JTabbedPane) getToolbar()).add("Standard", toolbar);
    }

    /**
     * Overwrite this method to customize the JGraph ToolBar. This method perform also an initialization of
     * the Toolbar.
     * <p>
     * We assume that {@link #_tabbedToolbar _tabbedToolBar}is set in this method.
     * <p>
     * Default toolbars are "Standard", "Alignement" and "Personalized".
     * 
     * @see #makeStandardToolBar()
     * @see #makeAlignmentToolBar()
     * @see #makePersonalizedToolBar()
     */
    protected void makeToolBar() {
        setToolbar(new JTabbedPane(SwingConstants.BOTTOM));

        makePersonalizedToolBar();
        makeStandardToolBar();
        makeAlignmentToolBar();
    }

    ////
    // display methods

    /**
     * Paints the graphs nodes by calling paint() on {@link mygale.mcge.layer.LayerManager layers}and
     * {@link mygale.mcge.mode.ModeManager modes}.
     * <p>
     * Whenever their is a change to the screen, this method will eventually be called from the
     * RedrawManager.
     * 
     * @param g
     *            the graphical context.
     */
    public void paint(Graphics g) {
        Graphics2D g2d = (Graphics2D) g;

        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_ALPHA_INTERPOLATION,
            java.awt.RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);

        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_TEXT_ANTIALIASING,
            java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_ANTIALIASING,
            java.awt.RenderingHints.VALUE_ANTIALIAS_ON);

        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_INTERPOLATION,
            java.awt.RenderingHints.VALUE_INTERPOLATION_BILINEAR);

        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_COLOR_RENDERING,
            java.awt.RenderingHints.VALUE_COLOR_RENDER_QUALITY);
        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_DITHERING,
            java.awt.RenderingHints.VALUE_DITHER_ENABLE);
        g2d.setRenderingHint(
            java.awt.RenderingHints.KEY_RENDERING,
            java.awt.RenderingHints.VALUE_RENDER_QUALITY);

        g2d.translate(-_originPoint.x, -_originPoint.y);

        if (getScale() != 1.0) {
            g2d.scale(getScale(), getScale());
        }

        _layerManager.paint(g2d);
        _modeManager.paint(g2d);
    }

    //
    // Preferences

    /**
     * This method was called each time a preference was changed.
     * 
     * @param event
     *            the description of the event.
     * @see mygale.mcge.util.prefs.PrefChangedListener#prefChanged(PrefChangedEvent)
     */
    public void prefChanged(PrefChangedEvent event) {
        if (event.prefType == PrefChangedEvent.GUI) {
            damagedAll();
        }
    }

    /**
     * Remove the given {@link Fig Fig}from the diagram being edited.
     * <p>
     * We assume that the attached {@link mygale.mcge.net.NetObject net-object}are removed from the graph
     * diagram. if it is as subclass of {@link FigModelElement FigModelElement}.
     * 
     * @throws mygale.mcge.throwable.MCGEInternalError
     *             something wrong !
     * @param fig
     *            a reference to figure that will be removed.
     */
    public void remove(IFigLayered fig) {
        if (!(fig instanceof IFigObject)) {
            _layerManager.remove(fig);
        } else {
            INetObject netobject = ((IFigObject) fig).getNetObject();
            if (netobject instanceof INetNode) {
                remove((INetNode) netobject);
            } else if (netobject instanceof INetEdge) {
                remove((INetEdge) netobject);
            } else {
                throw new MCGEInternalError(
                    "Cannot remove a FigModelElement "
                        + "that was not attached to a "
                        + "NetNode or a NetEdge");
            }
        }
    }

    /**
     * Remove the given {@link mygale.mcge.net.NetEdge edge}from the diagram being edited.
     * <p>
     * We assume that the attached {@link mygale.mcge.figEdge.FigObject figure}are removed from the graph
     * diagram.
     * 
     * @param netedge
     *            a reference to edge that will be removed.
     */
    protected void remove(INetEdge netedge) {
        removeFigForEdge(netedge);

        try {
            _inNetGraphEventHandling = true;

            getNetGraphModel().removeEdge(netedge);
        } finally {
            _inNetGraphEventHandling = false;
        }
    }

    /**
     * Remove the given {@link mygale.mcge.net.NetNode Node}from the diagram being edited.
     * <p>
     * We assume that the attached {@link mygale.mcge.figNode.FigObject figure}are removed from the graph
     * diagram.
     * 
     * @param netNode
     *            a reference to Node that will be removed.
     */
    protected void remove(INetNode netNode) {
        removeFigForNode(netNode);

        try {
            _inNetGraphEventHandling = true;

            getNetGraphModel().removeNode(netNode);

        } finally {
            _inNetGraphEventHandling = false;
        }
    }

    /**
     * This method was called each time a {@link Fig figure}must be removed.
     * 
     * @param fig
     *            a reference to the figure to remove.
     */
    public void removed(IFigLayered fig) {
        remove(fig);
    }

    protected void removeFigFor(IFigObject figObject) {
        getFigMapper().removeFigFor(figObject.getNetObject());
    }

    protected void removeFigForEdge(INetEdge netEdge) {
        IFigEdge figEdge = (IFigEdge) getFigFor(netEdge);

        if (figEdge != null) {
            if (Assert.ASSERTION_ON) {
                FigHelper.checkFigEdge(_figMapper, netEdge, figEdge);
            }

            _layerManager.remove(figEdge);
        }

        getFigMapper().removeFigFor(netEdge);
    }

    protected void removeFigForNode(INetNode netNode) {
        IFigNode figNode = (IFigNode) getFigFor(netNode);

        // remove all attached fig edges
        Iterator it = netNode.getNodeEdges();
        while (it.hasNext()) {
            removeFigForEdge((INetEdge) it.next());
        }

        if (figNode != null) {

            // remove the fig node
            _layerManager.remove(figNode);
        }

        getFigMapper().removeFigFor(netNode);
    }

    protected void resizeWorkSpace(double oldScale, double newScale) {
        Rectangle workspaceRect = getWorkspaceRect();

        workspaceRect.width = (int) Math.rint(workspaceRect.width * newScale / oldScale);
        workspaceRect.height = (int) Math.rint(workspaceRect.height * newScale / oldScale);

        setWorkspaceRect(workspaceRect);
    }

    /**
     * Scroll the JGraph so that the given point is visible. This is used when the user wants to drag an
     * object a long distance. If the given point are outside the workspace area, the scroll is made until
     * the workspace bounds are reached.
     * 
     * @param x
     *            x coordinate of the point that must be visible.
     * @param y
     *            y coordinate of the point that must be visible.
     */
    public void scrollToShow(int x, int y) {
        x -= _originPoint.x;
        y -= _originPoint.y;
        double speed = 0.8;
        Component c = getParent();
        if (c != null && c instanceof JViewport) {
            JViewport view = (JViewport) c;

            Dimension d = getSize();
            if (x > d.height - 10)
                x = d.height - 10;
            else if (x < 10)
                x = 10;
            if (y > d.width - 10)
                y = d.width - 10;
            else if (y < 10)
                y = 10;

            Rectangle viewRect = view.getViewRect();

            JScrollPane scroll = getGraphEditor().getScrollPane();
            JScrollBar hscroll = scroll.getHorizontalScrollBar();
            JScrollBar vscroll = scroll.getVerticalScrollBar();

            int vpos = vscroll.getValue();
            int vposMin = vscroll.getMinimum();
            int vposMax = vscroll.getMaximum();

            int hpos = hscroll.getValue();
            int hposMin = hscroll.getMinimum();
            int hposMax = hscroll.getMaximum();

            if (x > viewRect.x + viewRect.width) {
                if (x > hposMax - 20) {
                    hscroll.setValue(hposMax);
                } else {
                    int dx = Math.abs(x - (viewRect.x + viewRect.width));
                    int newHpos = hpos + ((int) (speed * dx));
                    if (newHpos > hposMax) {
                        newHpos = hposMax;
                    }
                    hscroll.setValue(newHpos);
                }
            } else if (x <= viewRect.x) {
                if (viewRect.x < 20) {
                    hscroll.setValue(hposMin);
                } else {
                    int dx = Math.abs(viewRect.x - x);
                    int newHpos = hpos - ((int) (speed * dx));
                    if (newHpos < hposMin) {
                        newHpos = hposMin;
                    }
                    hscroll.setValue(newHpos);
                }
            }

            if (y > viewRect.y + viewRect.height) {
                if (y > vposMax - 20) {
                    vscroll.setValue(vposMax);
                } else {
                    int dy = Math.abs(y - (viewRect.y + viewRect.height));
                    int newVpos = vpos + ((int) (speed * dy));
                    if (newVpos > vposMax) {
                        newVpos = vposMax;
                    }
                    vscroll.setValue(newVpos);
                }

            } else if (y < viewRect.y) {
                if (viewRect.y < 20) {
                    vscroll.setValue(vposMin);
                } else {
                    int dy = Math.abs(viewRect.y - y);
                    int newVpos = vpos - ((int) (speed * dy));
                    if (newVpos < vposMin) {
                        newVpos = vposMin;
                    }
                    vscroll.setValue(newVpos);
                }
            }
        }
    }

    public void setFigMapper(IFigMapper newFigMapper) {
        _figMapper = newFigMapper;
    }

    public void setNetGraphModel(INetGraphModel newNetGraphModel) {
        INetGraphModel oldNetGraphModel = getNetGraphModel();

        if (newNetGraphModel != oldNetGraphModel) {

            if (oldNetGraphModel != null) {
                getLayerManager().removeAll();
                oldNetGraphModel.removeNetGraphModelChangedListener(_netGraphModelChangedListener);
            }

            _netGraphModel = newNetGraphModel;

            firePropertyChange("netGraphModel", oldNetGraphModel, newNetGraphModel);

            if (newNetGraphModel != null) {
                insertFigsInLayer(newNetGraphModel);
                newNetGraphModel.addNetGraphModelChangedListener(_netGraphModelChangedListener);
            }
        }
    }

    /**
     * Set this Editor's drawing scale. A value of 1.0 draws at 1 to 1. A value greater than 1 draws larger,
     * less than 1 draws smaller.
     * 
     * @param newScale
     *            double
     */
    public void setScale(double newScale) {
        double oldScale = getScale();
        if ((newScale > 0) && (newScale != oldScale)) {

            _scale = newScale;
            resizeWorkSpace(oldScale, newScale);

            damagedAll();
        }
    }

    //
    // Statusbar

    /**
     * This method permits to draw a text to a status bar. This bar is not directly managed by the MCGE.
     * 
     * @param text
     *            the text to display in the status bar.
     */
    public void setStatusMsg(String text) {
        if (getGraphEditor() != null)
            getGraphEditor().setStatusMsg(text);
    }

    private void setWorkspaceRect(Rectangle newWorkspaceRect) {

        _originPoint = new Point(newWorkspaceRect.x, newWorkspaceRect.y);

        super.setPreferredSize(
            new Dimension(
                transformFromWorkSpacePointX(newWorkspaceRect.width),
                transformFromWorkSpacePointY(newWorkspaceRect.height)));

        revalidate();
    }

    private int transformFromWorkSpacePointX(int workspacePointX) {
        return (int) Math.rint(workspacePointX * getScale() - _originPoint.x);
    }

    private int transformFromWorkSpacePointY(int workspacePointY) {
        return (int) Math.rint(workspacePointY * getScale() - _originPoint.y);
    }

    private int transformToWorkSpacePointX(int intX) {
        return (int) Math.rint((intX + _originPoint.x) / getScale());
    }

    private int transformToWorkSpacePointY(int intY) {
        return (int) Math.rint((intY + _originPoint.y) / getScale());
    }

    /**
     * Scales the mouse coordinates (which match the drawing scale) back to the model scale. *
     * 
     * @param me
     *            MouseEvent
     */
    public void translateMouseEvent(MouseEvent me) {
        Point point = me.getPoint();

        /*
         * Assert.assertEquals(point.x, transformToWorkSpacePoint(translatedPoint).x, 1);
         * Assert.assertEquals(point.y, transformToWorkSpacePoint(translatedPoint).y, 1);
         *  
         */

        me.translatePoint(
            transformToWorkSpacePointX(point.x) - point.x,
            transformToWorkSpacePointY(point.y) - point.y);

    }

    /**
     * Return the ToolBars of the JGraph.
     * 
     * @return a reference to the toolbar contener.
     */
    public Component getToolBar() {
        return getToolbar();
    }

    protected void setToolbar(Component toolbar) {
        _toolbar = toolbar;
    }

    protected Component getToolbar() {
        return _toolbar;
    }
}