package mygale.mcge.core;

import java.awt.*;

import javax.swing.*;

import junit.framework.Assert;
import mygale.mcge.undoredo.UndoRedoManager;
import mygale.mcge.undoredo.UndoRedoManagerImpl;
import de.uniwue.ki.gui.util.JDisposeablePanel;
/** A panel that display a toolbar, a graph editing pane, and optionally a status bar.
 *  <p><center>
 *    <img src="doc-files/JGraphEditor_uml.gif" width="344" height="155" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="117,123,189,149"  href="JGraph.html">
 *      <area shape="Rect" coords="0,123,72,149"  href="JGraphToolBar.html">
 *      <area shape="Rect" coords="243,55,343,82"  href="../undoredo/UndoRedoManager.html">
 *      <area shape="Rect" coords="117,0,189,28"  href="http://java.sun.com/products/jdk/1.2/docs/api/javax/swing/JPanel.html">
 *      <area shape="Rect" coords="243,123,315,149"  href="http://java.sun.com/products/jdk/1.2/docs/api/javax/swing/JLabel.html">
 *    </map>
 *  </center></p>
 *
 * @auhtor Mahdi Hannoun
 */
public abstract class JGraphEditor extends JDisposeablePanel {

	/** The graph pane (shown in the middle of window).
	 */
	protected JGraph _graph;

	/** The ScrollPane that contains the Graph Area.
	 */
	protected JScrollPane _scroll;

	/** This is a reference to a external status bar.
	 */
	protected JLabel _statusbar = null;

	/** This is a reference to a external Undo/Redo manager.
	 */
	protected UndoRedoManager _undoredoManager = null;

	/** Construct a new JGraphEditor with
	     *  an internal undo/redo manager.
	     */
	public JGraphEditor() {
		this(null);
	}

	///
	// Constructors

	/** Construct a new JGraphEditor with the given title and given
	 *  {@link mygale.mcge.undoredo.UndoRedoManager UndoRedoManager}.
	 *
	 * @param undoredo a reference to an undo/redo manager.
	 */
	public JGraphEditor(UndoRedoManager undoredo) {
		super();

		// Create undo/redo manager
		_undoredoManager = undoredo;

		initialize();

	}

	/** This method permits to clear the content of a status bar
	     *  This bar is not directly managed by the MCGE.
	     *  We assume that a status bar is a JLabel object.
	     */
	public void clearStatus() {
		setStatusMsg("");
	}

	/**
	 * @see de.uniwue.ki.gui.util.IDisposeable#dispose()
	 */
	public void dispose() {

		super.dispose();

		// after super.dispose since:
		if (Assert.ASSERTION_ON)
			Assert.assertTrue(isDisposed());
		if ((getGraph() != null) && (!getGraph().isDisposed())) {
			getGraph().dispose();
		}

	}

	/** Return the {@link JGraph graph editing area}.
	     *
	    @link JGraph JGraph}.
	     * @return a reference to a { */
	public JGraph getGraph() {
		return _graph;
	}

	// Accessors

	/** Return the scroll pane that contains the
	 *  {@link JGraph graph editing area}.
	 *
	 * @return a reference to the scroll pane.
	 */
	public JScrollPane getScrollPane() {
		return _scroll;
	}

	/** Return the current undo/redo manager.
	     *
	    @link mygale.mcge.undoredo.UndoRedoManager UndoRedoManager}.
	     * @return a reference to a
	     *         { */
	public UndoRedoManager getUndoRedoManager() {
		return _undoredoManager;
	}

	private void initialize() {
		makeUndoRedoManager();

		// Create the layout
		setLayout(new GridBagLayout());

		// Create Scroller
		_scroll =
			new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		_scroll.setBorder(null);
		_scroll.getHorizontalScrollBar().setBlockIncrement(25);
		_scroll.getVerticalScrollBar().setBlockIncrement(25);
		// _scroll.setPreferredSize(new Dimension(500, 500));
		//_scroll.setMinimumSize(new Dimension(640, 480)); // removed this because scrollers were not visible in small windows - rh 12. Juni 2004

		// Create the graph editing area
		_graph = makeGraph();
		// _graph.setDrawingAreaSize(500, 500);

		// Create toolbars
		Component toolbar = _graph.getToolBar();

		JViewport port = _scroll.getViewport();
		port.add(_graph);

		GridBagConstraints tabbedToobarConstraints = new GridBagConstraints();
		tabbedToobarConstraints.gridx = 0;
		tabbedToobarConstraints.gridy = 0;
		tabbedToobarConstraints.fill = GridBagConstraints.HORIZONTAL;
		tabbedToobarConstraints.weightx = 0.0;
		tabbedToobarConstraints.weighty = 0.0;

		GridBagConstraints scrollerConstraints = new GridBagConstraints();
		scrollerConstraints.gridx = 0;
		scrollerConstraints.gridy = 1;
		scrollerConstraints.fill = GridBagConstraints.BOTH;
		scrollerConstraints.weightx = 1.0;
		scrollerConstraints.weighty = 1.0;

		add(toolbar, tabbedToobarConstraints);
		add(_scroll, scrollerConstraints);

		// pack() ;
		// setSize(600, 450);
	}

	/** Create and return the default instance of the
	     *  {@link JGraph graph editing area}.
	     *
	     * @return a reference to a {@link JGraph JGraph}.
	     */
	protected abstract JGraph makeGraph();

	//
	// Undo/redo API

	/** This method create an instance of an undo/redo
	 *  manager if it was not given to the constructor. */
	public void makeUndoRedoManager() {
		if (_undoredoManager == null) {
			_undoredoManager = new UndoRedoManagerImpl();
		}
	}

	//
	// Statusbar API

	/** This method permits to set the reference to a status bar.
	 *  This bar is not directly managed by the MCGE.
	 *  We assume that a status bar is a JLabel object.
	 *
	 * @param bar the status bar.
	 */
	public void setStatusBar(JLabel bar) {
		_statusbar = bar;
	}

	/** This method permits to draw a text to a status bar.
	     *  This bar is not directly managed by the MCGE.
	     *  We assume that a status bar is a JLabel object.
	     *
	     * @param text the text to display in the status bar.
	     */
	public void setStatusMsg(String text) {
		if (_statusbar != null) {
			if ("".equals(text)) {
				_statusbar.setText(" ");
			} else {
				_statusbar.setText(text);
			}
		}
	}
}