package mygale.mcge.event.mouse;

import java.awt.Event;
import java.awt.Point;
import java.awt.event.InputEvent;

import mygale.mcge.fig.IFigLayered;

/** This class permits to store informations about mouse events.
 *  <p><center>
 *    <img src="doc-files/MCGEMouseEvent_uml.gif" width="254" height="137" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="9,0,126,27"  href="MCGEMouseManager.html">
 *      <area shape="Rect" coords="180,54,252,81"  href="../../fig/Fig.html">
 *      <area shape="Rect" coords="0,108,135,135"  href="MCGEMouseEventListener.html">
 *    </map>
 *  </center></p>

 */
public class MCGEMouseEvent {

	/** This field contains the event's event id.
	 *
	 * @see #getID()
	 */
	protected int _id;

	/** This field controls whether or not the event is sent back
	 *  down to the peer once the source has processed it.
	 *  <code>false</code> means it's sent to the peer,
	 *  <code>true</code> means it's not.
	 *  Semantic events always have a <code>true</code> value
	 *  since they were generated by the peer in response to
	 *  a low-level event.
	 *
	 * @see #consume()
	 * @see #isConsumed()
	 */
	protected boolean _consumed = false;

	/**store the flag modifiers (key pressed) for this event.
	 *
	 * @see #isShiftDown()
	 * @see #isControlDown()
	 * @see #isAltDown()
	 * @see #isAltGraphDown()
	 * @see #isMetaDown()
	 * @see #getModifiers()
	 */
	protected int _modifiers;

	/** The first number in the range of ids used for mouse events.
	 */
	public static final int MCGE_MOUSE_FIRST = 600;

	/** The last number in the range of ids used for mouse events.
	 */
	public static final int MCGE_MOUSE_LAST = 607;

	/** The "mouse clicked" event. This MCGEMouseEvent occurs when a mouse
	 *  button is pressed and released quickly on the graph editor's area.
	 */
	public static final int MCGE_MOUSE_CLICKED = MCGE_MOUSE_FIRST;

	/** The "mouse double clicked" event. This MCGEMouseEvent occurs when a mouse
	 *  button is pressed and released twice quickly on the graph editor's area.
	 */
	public static final int MCGE_MOUSE_DOUBLE_CLICKED = MCGE_MOUSE_FIRST + 1;

	/** The "mouse pressed" event. This MCGEMouseEvent occurs when a mouse
	 *  button is pushed down on the graph editor's area. Note that this event
	 *  is exclusif with the "clicked" and "doucle clicked" event.
	 */
	public static final int MCGE_MOUSE_PRESSED = 2 + MCGE_MOUSE_FIRST;

	/** The "mouse released" event. This MCGEMouseEvent occurs when a mouse
	 *  button is let up.
	 */
	public static final int MCGE_MOUSE_RELEASED = 3 + MCGE_MOUSE_FIRST;

	/** The "mouse moved" event. This MCGEMouseEvent occurs when the mouse
	 *  position changes.
	 */
	public static final int MCGE_MOUSE_MOVED = 4 + MCGE_MOUSE_FIRST;

	/** The "mouse dragged" event. This MCGEMouseEvent occurs when the mouse
	 *  position changes while the "drag" modifier is active (for example, the
	 *  shift key).
	 */
	public static final int MCGE_MOUSE_DRAGGED = 5 + MCGE_MOUSE_FIRST;

	/** The "mouse entered" event. This MCGEMouseEvent occurs when the mouse
	 *  cursor enters a figure on the graph editor area.
	 */
	public static final int MCGE_MOUSE_ENTERED = 6 + MCGE_MOUSE_FIRST;

	/** The "mouse exited" event. This MCGEMouseEvent occurs when the mouse
	 *  cursor leaves a figure on the graph editor area.
	 */
	public static final int MCGE_MOUSE_EXITED = 7 + MCGE_MOUSE_FIRST;

	/** The mouse events x coordinate.
	 *  The <var>x</var> value is relative to the component
	 *  that fired the event and relative to the
	 *  current workspace.
	 *
	 * @see #getX()
	 */
	protected int _x;

	/** The mouse events y coordinate.
	 *  The <var>y</var> value is relative to the component
	 *  that fired the event and relative to the
	 *  current owrkspace.
	 *
	 * @see #getY()
	 */
	protected int _y;

	/** The real mouse events x coordinate.
	 *  The <var>r_x</var> value is relative to the component
	 *  that fired the event.
	 *
	 * @see #getRealX()
	 */
	protected int _realX;

	/** The mouse events y coordinate.
	 *  The <var>r_y</var> value is relative to the component
	 *  that fired the event.
	 *
	 * @see #getRealY()
	 */
	protected int _realY;

	/** The {@link mygale.mcge.fig.FigLayered figure} on wich
	 *  the mouse is in.
	 */
	protected IFigLayered _fig;

	/////

	/** Construct a new MCGEMouseEvent.
	@link mygale.mcge.fig.Fig figure} where the
	 *            mouse are.
	 * @param x the x coordinate of the mouse in the workspace.
	 * @param y the y coordinate of the mouse in the workspace.
	 * @param rx the real x coordinate of the mouse.
	 * @param ry the real y coordinate of the mouse.
	 * @param modifiers values of modifiers when the event occurs.
	 * @param id id of this event.
	 * @param fig the { */
	public MCGEMouseEvent(int x, int y, int rx, int ry, int modifiers, int id, IFigLayered fig) {
		this._x = x;
		this._y = y;
		this._realX = rx;
		this._realY = ry;
		this._modifiers = modifiers;
		this._id = id;
		this._fig = fig;
		this._consumed = false;
	}

	/** Returns <code>true</code> if the event has been consumed.
	     *  otherwise returns <code>false</code>.
	     *
	     * @return boolean
	 * @see #consume()
	     */
	public boolean isConsumed() {
		return _consumed;
	}

	/** Returns the modifiers flag for this event.
	 *
	 * @return an interger that represents the modifiers' flags.
	 */
	public int getModifiers() {
		return _modifiers;
	}

	/** Returns the horizontal <var>x</var> position of the
	     *  event relative to the source component.
	     *
	     * @return an integer indicating horizontal position relative to
	     *         the component.
	     */
	public int getX() {
		return _x;
	}

	/** Returns the vertical <var>y</var> position of the
	     *  event relative to the source component.
	     *
	     * @return an integer indicating vertical position relative to
	     *         the component.
	     */
	public int getY() {
		return _y;
	}

	/** Consume the event.
	     *
	     * @see #isConsumed()
	     */
	public void consume() {
		_consumed = true;
	}

	/** Return the {@link mygale.mcge.fig.Fig figure}
	     *  on which the mouse is.
	     *
	    @link mygale.mcge.fig.Fig figure}
	     *         under the mouse.
	     * @return the { */
	public IFigLayered getFigure() {
		return _fig;
	}

	/** Returns the event type.
	 *
	 * @return the id of this event.
	 */
	public int getID() {
		return _id;
	}

	/** Returns the <var>x</var>,<var>y</var> position of
	     *  the event relative to the source component.
	     *
	     * @return a Point object containing the <var>x</var> and
	     *         <var>y</var> coordinates relative to the
	     *         source component
	     */
	public Point getPoint() {
		int x;
		int y;
		synchronized (this) {
			x = this._x;
			y = this._y;
		}
		return new Point(x, y);
	}

	/** Returns the horizontal <var>x</var> real position of
	     *  the event relative to the source component.
	     *
	     * @return an integer indicating horizontal position relative to
	     *         the component
	     */
	public int getRealX() {
		return _realX;
	}

	/** Returns the vertical <var>y</var> real position of
	     *  the event relative to the source component.
	     *
	     * @return an integer indicating vertical position relative to
	     *         the component
	     */
	public int getRealY() {
		return _realY;
	}

	/** Returns whether or not the Alt modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the alt key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isAltDown() {
		return (_modifiers & Event.ALT_MASK) != 0;
	}

	/** Returns whether or not the Alt-Graph modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the altgr key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isAltGraphDown() {
		return (_modifiers & InputEvent.ALT_GRAPH_MASK) != 0;
	}

	/** Returns whether or not the Control modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the control key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isControlDown() {
		return (_modifiers & Event.CTRL_MASK) != 0;
	}

	/** Returns whether or not the left mouse button
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a left mouse button
	     *         was down when the event occurs,
	     *         otherwise <code>false</code>.
	     */
	public boolean isLeftButtonDown() {
		return (_modifiers & InputEvent.BUTTON1_MASK) != 0;
	}

	/** Returns whether or not the Meta modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the meta key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isMetaDown() {
		return (_modifiers & Event.META_MASK) != 0;
	}

	/** Returns whether or not the middle mouse button
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a middle mouse button
	     *         was down when the event occurs,
	     *         otherwise <code>false</code>.
	     */
	public boolean isMiddleButtonDown() {
		return (_modifiers & InputEvent.BUTTON2_MASK) != 0;
	}

	/** Returns whether or not the "key" modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isModifierKeyDown() {
		return (
			_modifiers
				& Event.SHIFT_MASK
				& Event.CTRL_MASK
				& Event.META_MASK
				& Event.ALT_MASK
				& InputEvent.ALT_GRAPH_MASK)
			!= 0;
	}

	/** Returns whether or not the left mouse button
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a left mouse button
	     *         and only the left button,
	     *         was down when the event occurs,
	     *         otherwise <code>false</code>.
	     */
	public boolean isOnlyLeftButtonDown() {
		return (_modifiers & InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK;
	}

	/** Returns whether or not the middle mouse button
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a middle mouse button
	     *         and only the middle button
	     *         was down when the event occurs,
	     *         otherwise <code>false</code>.
	     */
	public boolean isOnlyMiddleButtonDown() {
		return (_modifiers & InputEvent.BUTTON2_MASK) == InputEvent.BUTTON2_MASK;
	}

	/** Returns whether or not the right mouse button
	     *  is down on this event.
	     *
	     */
	/**
	 * Returns whether or not the first button is down on this event.
	 * @return <code>true</code> if the a right mouse button
	 *         and only the right button
	 *         was down when the event occurs,
	 *         otherwise <code>false</code>.
	 */
	public boolean isOnlyRightButtonDown() {
		return (_modifiers & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK;
	}

	/** Returns whether or not the right mouse button
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the a right mouse button
	     *         was down when the event occurs,
	     *         otherwise <code>false</code>.
	     */
	public boolean isRightButtonDown() {
		return (_modifiers & InputEvent.BUTTON3_MASK) != 0;
	}

	/** Returns whether or not the Shift modifier
	     *  is down on this event.
	     *
	     * @return <code>true</code> if the shift key was down when
	     *         the event occurs, otherwise <code>false</code>.
	     */
	public boolean isShiftDown() {
		return (_modifiers & Event.SHIFT_MASK) != 0;
	}

	/** Returns a parameter string identifying this event.
	     *  This method is useful for event-logging and for debugging.
	     *
	     * @return a string identifying the event and its attributes.
	     */
	public String paramString() {
		String typeStr;
		switch (_id) {
			case MCGE_MOUSE_PRESSED :
				typeStr = "MCGE_MOUSE_PRESSED";
				break;
			case MCGE_MOUSE_RELEASED :
				typeStr = "MCGE_MOUSE_RELEASED";
				break;
			case MCGE_MOUSE_CLICKED :
				typeStr = "MCGE_MOUSE_CLICKED";
				break;
			case MCGE_MOUSE_DOUBLE_CLICKED :
				typeStr = "MCGE_DOUBLE_MOUSE_CLICKED";
				break;
			case MCGE_MOUSE_MOVED :
				typeStr = "MCGE_MOUSE_MOVED";
				break;
			case MCGE_MOUSE_DRAGGED :
				typeStr = "MCGE_MOUSE_DRAGGED";
				break;
			case MCGE_MOUSE_ENTERED :
				typeStr = "MCGE_MOUSE_ENTERED";
				break;
			case MCGE_MOUSE_EXITED :
				typeStr = "MCGE_MOUSE_EXITED";
				break;
			default :
				typeStr = "unknown type";
		}
		return typeStr
			+ ",("
			+ _x
			+ ","
			+ _y
			+ ")"
			+ ",B=("
			+ (isLeftButtonDown() ? "L" : "l")
			+ ","
			+ (isMiddleButtonDown() ? "M" : "m")
			+ ","
			+ (isRightButtonDown() ? "R" : "r")
			+ "),M="
			+ getModifiers();
	}
}
