package mygale.mcge.event.mouse;

import java.awt.event.*;

import javax.swing.Timer;

import mygale.mcge.core.JGraph;
import mygale.mcge.fig.IFigLayered;

/** This class is the manager, which translate real mouse events
 *  into MCGE mouse events.
 *  <p>
 *  The difference between a real mouse event and a MCGE mouse event
 *  is that each MCGE event are no ambigus i.e., when a MCGE "click"
 *  event occurs previous real events "mouse pressed" and
 *  "mouse released" are ignored.
 *  <p><center>
 *    <img src="doc-files/MCGEMouseManager_uml.gif" width="394" height="169" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="163,0,235,28"  href="../../core/JGraph.html">
 *      <area shape="Rect" coords="0,77,100,104"  href="MCGEMouseEvent.html">
 *      <area shape="Rect" coords="320,77,392,104"  href="../../fig/Fig.html">
 *      <area shape="Rect" coords="131,141,266,167"  href="MCGEMouseEventListener.html">
 *    </map>
 *  </center></p>

 */
public class MCGEMouseManager implements MouseListener, MouseMotionListener, ActionListener {

	/** This is the editor that intersepts all events and redirect
		*  mouse event to this MCGEMouseManager.
		*/
	private JGraph _graph;

	/** This is the mouse event listener.
		*  Only one listener are needed because,
		*  only the {@link mygale.mcge.mode.ModeManager ModeManager}
		*  use muose events.
		*/
	private MCGEMouseEventListener _mcgeMouseListener;

	/** This is the <var>x</var>,<var>y</var> coordinates
		*  of the mouse in the workspace.
		*/
	protected int _x, _y;

	/** This is the real <var>x</var>,<var>y</var> coordinates
		*  of the mouse in the workspace.
		*/
	protected int _realX, _realY;

	/** This is the modifiers.
		*/
	protected int _modifiers;

	/** This timer permits to generated MCGE click and
		*  MCGE doubleclick events. */
	protected Timer _timer;

	/** <code>true</code> if this manager must track
		*  {@link mygale.mcge.fig.Fig figures},
		*  otherwise <code>false</code>.
		*
		* @see #_figure
		*/
	protected boolean _trackFigure = false;

	/** This is the {@link mygale.mcge.fig.FigLayered figures}
		*  under the mouse cursor.
		*
		* @see #_trackFigure
		*/
	protected IFigLayered _figure = null;

	/* The internal state of the MCGE Mouse Manager. Used to keep track
		* of the mouse event to detect "click" and "doucble click".
		*/
	protected int _state;

	/** Internal state of this Manager : DEFAULT STATE (see finite-state automate).
		*/
	protected static final int INITIAL_STATE = 0;
	/** Internal state of this Manager (see finite-state automate).
		*/
	protected static final int WAITING_FIRST_RELEASE = 1;
	/** Internal state of this Manager (see finite-state automate).
		*/
	protected static final int WAITING_SECOND_CLICK = 2;
	/** Internal state of this Manager (see finite-state automate).
		*/
	protected static final int WAITING_SECOND_RELEASE = 3;

	/** This the delay under a mouse-pressed event is translate
		*  into a MCGE-click event.
		*/
	protected static final int CLICK_DELAY = 200;

	/** This the delay under a mouse-pressed event is translate
		*  into a MCGE-doubleclick event.
		*/
	protected static final int DOUBLECLICK_DELAY = 100;

	/** Construct a new MCGEMouseManager.
	 *
	@link mygale.mcge.mode.ModeManager ModeManager}).
	 * @param graph the editor that send mouse events to this manager.
	 * @param mcgeMouseListener the listener that received MCGe events
	 *                          (usually a
	 *                          { */
	public MCGEMouseManager(JGraph graph, MCGEMouseEventListener mcgeMouseListener) {
		_graph = graph;
		_mcgeMouseListener = mcgeMouseListener;
		_timer = new Timer(CLICK_DELAY, this);
		_timer.setRepeats(false);
		_state = INITIAL_STATE;

		_trackFigure = false;
		_figure = null;
	}

	/** Invoked when the timout is finished.
	 *
	 * @throws mygale.mcge.throwable.MCGEInternalError something wrong appends !
	 * @param e not used.
	 * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		switch (_state) {
			case INITIAL_STATE :
				// not a double click, send a mouse pressed event
				//throw new MCGEInternalError( "timeout in the INITIAL STATE" ) ;
				System.out.println("MCGEMouseManager.actionPerformed: timeout in the INITIAL STATE");
			case WAITING_FIRST_RELEASE :
				// not a click, send a mouse pressed event
				_state = INITIAL_STATE;
				trackFigure(_x, _y, _realX, _realY, _modifiers);
				_mcgeMouseListener.mcgeMousePressed(
					new MCGEMouseEvent(
						_x,
						_y,
						_realX,
						_realY,
						_modifiers,
						MCGEMouseEvent.MCGE_MOUSE_PRESSED,
						_figure));
				break;
			case WAITING_SECOND_CLICK :
				// not a double click, send a mouse click event
				_state = INITIAL_STATE;
				trackFigure(_x, _y, _realX, _realY, _modifiers);
				_mcgeMouseListener.mcgeMouseClicked(
					new MCGEMouseEvent(
						_x,
						_y,
						_realX,
						_realY,
						_modifiers,
						MCGEMouseEvent.MCGE_MOUSE_CLICKED,
						_figure));
				break;
			case WAITING_SECOND_RELEASE :
				// not a double click, send a mouse click, and then a mouse pressed
				_state = INITIAL_STATE;
				trackFigure(_x, _y, _realX, _realY, _modifiers);
				_mcgeMouseListener.mcgeMouseClicked(
					new MCGEMouseEvent(
						_x,
						_y,
						_realX,
						_realY,
						_modifiers,
						MCGEMouseEvent.MCGE_MOUSE_CLICKED,
						_figure));
				_mcgeMouseListener.mcgeMousePressed(
					new MCGEMouseEvent(
						_x,
						_y,
						_realX,
						_realY,
						_modifiers,
						MCGEMouseEvent.MCGE_MOUSE_PRESSED,
						_figure));
		}
	}

	////
	// accessor to keep track of figure

	/** Initiate the {@link mygale.mcge.fig.Fig figure} tracking.
	 */
	public void beginTrackFigure() {
		_trackFigure = true;
		_figure = null;
	}

	/** Stop the {@link mygale.mcge.fig.Fig figure} tracking.
	     */
	public void endTrackFigure() {
		_trackFigure = false;
		_figure = null;
	}

	/** Reply if {@link mygale.mcge.fig.Fig figure} tracking is
	     *  used.
	     *
	     * @return <code>true</code> if the tracking is enabled,
	     *         otherwise <code>false</code>.
	     */
	public boolean isFigureTracked() {
		return _trackFigure;
	}

	/** the real "mouse clicked" event was ignored by this manager.
	     *
	     * @param me the real mouse event.
	     * @see java.awt.event.MouseListener#mouseClicked(MouseEvent)
	 */
	public void mouseClicked(MouseEvent me) {
		// do nothing
	}

	/** Pass the "mouse dragged" events to
	     *  {@link mygale.mcge.event.mouse.MCGEMouseEventListener MCGEMouseEventListener}
	     *  only if the internal state was correct.
	     *
	     * @param me the real mouse event.
	     * @see java.awt.event.MouseMotionListener#mouseDragged(MouseEvent)
	 */
	public void mouseDragged(MouseEvent me) {
		if (_state == INITIAL_STATE) {
			int realX = me.getX();
			int realY = me.getY();
			_graph.translateMouseEvent(me);
			trackFigure(me.getX(), me.getY(), realX, realY, me.getModifiers());
			_mcgeMouseListener.mcgeMouseDragged(
				new MCGEMouseEvent(
					me.getX(),
					me.getY(),
					realX,
					realY,
					me.getModifiers(),
					MCGEMouseEvent.MCGE_MOUSE_DRAGGED,
					_figure));
		}
		me.consume();
	}

	/** the real "mouse entered" event was ignored by this manager.
	     *
	     * @param me the real mouse event.
	     * @see java.awt.event.MouseListener#mouseEntered(MouseEvent)
	 */
	public void mouseEntered(MouseEvent me) {
		// do nothing
	}

	/** the real "mouse exited" event was ignored by this manager.
	     *
	     * @param me the real mouse event.
	     */
	public void mouseExited(MouseEvent me) {
		// do noting
	}

	//
	// mouse even handling

	/** Pass the "mouse moved" events to
	 *  {@link mygale.mcge.event.mouse.MCGEMouseEventListener MCGEMouseEventListener}
	 *  only if the internal state was correct.
	 *
	 * @param me the real mouse event.
	 * @see java.awt.event.MouseMotionListener#mouseMoved(MouseEvent)
	 */
	public void mouseMoved(MouseEvent me) {
		if (_state == INITIAL_STATE) {
			int realX = me.getX();
			int realY = me.getY();
			_graph.translateMouseEvent(me);
			trackFigure(me.getX(), me.getY(), realX, realY, me.getModifiers());
			_mcgeMouseListener.mcgeMouseMoved(
				new MCGEMouseEvent(
					me.getX(),
					me.getY(),
					realX,
					realY,
					me.getModifiers(),
					MCGEMouseEvent.MCGE_MOUSE_MOVED,
					_figure));
		}
		me.consume();
	}

	/** Initiate the waiting for a MCGE-click or MCGE-doubleclick
	     *  events that will be passed to
	     *  {@link mygale.mcge.event.mouse.MCGEMouseEventListener MCGEMouseEventListener}
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong appends !
	     * @param me the real mouse event.
	     * @see java.awt.event.MouseListener#mousePressed(MouseEvent)
	 */
	public void mousePressed(MouseEvent me) {
		if (isValidButton(me)) {
			switch (_state) {
				case INITIAL_STATE :
					_state = WAITING_FIRST_RELEASE;
					_realX = me.getX();
					_realY = me.getY();
					_graph.translateMouseEvent(me);
					_x = me.getX();
					_y = me.getY();
					_modifiers = me.getModifiers();
					me.consume();
					_timer.setDelay(CLICK_DELAY);
					_timer.start();
					break;
				case WAITING_FIRST_RELEASE :
				case WAITING_SECOND_RELEASE :
					//throw new MCGEInternalError( "detected a mouse pressed, but we"+
					//                 " are in the state : " + _state ) ;
					System.out.println(
						"MCGEMouseManager.mousePressed: "
							+ "detected a mouse pressed, but we"
							+ " are in the state : "
							+ _state);
				case WAITING_SECOND_CLICK :
					_timer.stop();
					_state = WAITING_SECOND_RELEASE;
					_realX = me.getX();
					_realY = me.getY();
					_graph.translateMouseEvent(me);
					_x = me.getX();
					_y = me.getY();
					_modifiers = me.getModifiers();
					me.consume();
					_timer.setDelay(CLICK_DELAY);
					_timer.start();
					break;
				default :
					//throw new MCGEInternalError( "unknow state" ) ;
					System.out.println("MCGEMouseManager.mousePressed: unknow state");
			}
		} else {
			me.consume();
		}
	}

	/** Pass the "mouse released" or "moure doubleclick" events to
	     *  {@link mygale.mcge.event.mouse.MCGEMouseEventListener MCGEMouseEventListener}
	     *  only if the internal state was correct.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong appends !
	     * @param me the real mouse event.
	     * @see java.awt.event.MouseListener#mouseReleased(MouseEvent)
	 */
	public void mouseReleased(MouseEvent me) {
		if (isValidButton(me)) {
			int realX = me.getX();
			int realY = me.getY();
			_graph.translateMouseEvent(me);
			switch (_state) {
				case INITIAL_STATE :
					trackFigure(me.getX(), me.getY(), realX, realY, me.getModifiers());
					_mcgeMouseListener.mcgeMouseReleased(
						new MCGEMouseEvent(
							me.getX(),
							me.getY(),
							realX,
							realY,
							me.getModifiers(),
							MCGEMouseEvent.MCGE_MOUSE_RELEASED,
							_figure));
					me.consume();
					break;
				case WAITING_FIRST_RELEASE :
					_timer.stop();
					_state = WAITING_SECOND_CLICK;
					_realX = realX;
					_realY = realY;
					_x = me.getX();
					_y = me.getY();
					_modifiers = me.getModifiers();
					me.consume();
					_timer.setDelay(DOUBLECLICK_DELAY);
					_timer.start();
					break;

				case WAITING_SECOND_CLICK :
					//throw new MCGEInternalError( "detected a mouse released, but we"+
					//                 " are in the state : "+_state ) ;
					System.out.println(
						"MCGEMouseManager.mouseReleased: "
							+ "detected a mouse released, but we"
							+ " are in the state : "
							+ _state);

				case WAITING_SECOND_RELEASE :
					_timer.stop();
					_state = INITIAL_STATE;
					trackFigure(me.getX(), me.getY(), realX, realY, me.getModifiers());
					_mcgeMouseListener.mcgeMouseDoubleClicked(
						new MCGEMouseEvent(
							me.getX(),
							me.getY(),
							realX,
							realY,
							me.getModifiers(),
							MCGEMouseEvent.MCGE_MOUSE_DOUBLE_CLICKED,
							_figure));
					me.consume();
					break;
				default :
					//throw new MCGEInternalError( "unknow state" ) ;
					System.out.println("MCGEMouseManager.mouseReleased: " + "unknow state");
			}
		} else {
			me.consume();
		}
	}

	/** Reply is the buttons from an event are recognized.
	     *  That permits to ignore the wheel that was not
	     *  really supported by the AWT library.
	     *
	     * @return <code>true</code> if the button into the given
	     *         event was recognized by this manager,
	     *         otherwise <code>false</code>.
	     */
	protected static boolean isValidButton(MouseEvent me) {
		int modifiers = me.getModifiers();
		return (
			((modifiers & InputEvent.BUTTON1_MASK) != 0)
				|| ((modifiers & InputEvent.BUTTON2_MASK) != 0)
				|| ((modifiers & InputEvent.BUTTON3_MASK) != 0));
	}

	/** This methid was called each time a event occurs.
	     *  It permits to track figures.
	     *
	     * @param x x coordinate of the mouse in the workspace.
	     * @param y y coordinate of the mouse in the workspace.
	     * @param rx real x coordinate of the mouse.
	     * @param ry real y coordinate of the mouse.
	     * @param modifiers value of the modifiers when the event occurs.
	     */
	protected void trackFigure(int x, int y, int realX, int realY, int modifiers) {
		if (_trackFigure == true) {
			IFigLayered fig = _graph.hit(x, y);
			if (fig == _figure)
				return;
			if ((fig != null) && (_figure == null)) {
				_figure = fig;
				_mcgeMouseListener.mcgeMouseEntered(
					new MCGEMouseEvent(
						x,
						y,
						realX,
						realY,
						modifiers,
						MCGEMouseEvent.MCGE_MOUSE_ENTERED,
						_figure));
			} else if ((_figure != null) && (fig == null)) {
				_figure = fig;
				_mcgeMouseListener.mcgeMouseExited(
					new MCGEMouseEvent(
						x,
						y,
						realX,
						realY,
						modifiers,
						MCGEMouseEvent.MCGE_MOUSE_EXITED,
						_figure));
			} else {
				_mcgeMouseListener.mcgeMouseExited(
					new MCGEMouseEvent(
						x,
						y,
						realX,
						realY,
						modifiers,
						MCGEMouseEvent.MCGE_MOUSE_EXITED,
						_figure));
				_figure = fig;
				_mcgeMouseListener.mcgeMouseEntered(
					new MCGEMouseEvent(
						x,
						y,
						realX,
						realY,
						modifiers,
						MCGEMouseEvent.MCGE_MOUSE_ENTERED,
						_figure));
			}
		}
	}
}
