package mygale.mcge.fig;

import java.awt.*;

import mygale.mcge.util.prefs.Prefs;

/** This class is the base class for model element that
 *  they must not be managed directly by the editor.
 *  <p><center>
 *    <img src="doc-files/FigInternal_uml.gif" width="205" height="137" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="10,0,82,28"  href="FigObject.html">
 *      <area shape="Rect" coords="0,108,92,136"  href="FigPort.html">
 *      <area shape="Rect" coords="131,0,204,28"  href="../util/prefs/Prefs.html">
 *      <area shape="Rect" coords="131,54,204,82"  href="../net/NetObject.html">
 *    </map>
 *  </center></p>
 *
 */
public abstract class AbstractFig implements IFig {

	/** The position and the dimension of this Fig.
	 */
	protected Rectangle _rect = new Rectangle(0, 0, 0, 0);

	/** Construct a new FigInternal and set the default position to (0,0)
	     *  and the default size to {0,0).
	     */
	public AbstractFig() {
		setBounds(0, 0, 0, 0);
	}

	/** Construct a new Fig.
	 *
	 * @param x horizontal position of this Fig.
	 * @param y vertical position of this Fig.
	 * @param width width of this Fig.
	 * @param height height of this Fig.
	 */
	public AbstractFig(int x, int y, int width, int height) {
		setBounds(x, y, width, height);
	}

	public void add(Point pointToAdd) {
		_rect.add(pointToAdd);
	}

	/**
	 * @param boundsToAdd Rectangle
	 */
	public void add(Rectangle boundsToAdd) {
		_rect.add(boundsToAdd);
	}

	/** Clean this Fig.
	     * @see mygale.mcge.fig.IFig#cleanUp()
	 */
	public void cleanUp() {
	}

	/** Reply <code>true</code> if given coords are in this Fig.
	     *
	     * @param x horizontal coord.
	     * @param y vertical coord.
	     * @return <code>true</code> if the point
	     *         (<var>x</var>,<var>y</var>) is in this Fig.
	     */
	public boolean contains(int x, int y) {
		return _rect.contains(x, y);
	}

	/** Reply the position and the dimension of this Fig.
	     *
	     * @return position and dimension.
	     * @see mygale.mcge.fig.IFig#getBounds()
	 */
	public Rectangle getBounds() {
		return _rect;
	}

	/** Reply the height of this Fig.
	     *
	     * @return height.
	     * @see mygale.mcge.fig.IFig#getHeight()
	 */
	public int getHeight() {
		return _rect.height;
	}

	/** Reply the position of this Fig.
	     *
	     * @return the position of this Fig.
	     * @see mygale.mcge.fig.IFig#getLocation()
	 */
	public Point getLocation() {
		return _rect.getLocation();
	}

	/** Reply the size of this Fig.
	     *
	     * @return the size of this Fig.
	     */
	public Dimension getSize() {
		return _rect.getSize();
	}

	/** Reply the width of this Fig.
	     *
	     * @return width.
	     * @see mygale.mcge.fig.IFig#getWidth()
	 */
	public int getWidth() {
		return _rect.width;
	}

	/** Reply the horizontal position of this Fig.
	 *
	 * @return horizontal position.
	 * @see mygale.mcge.fig.IFig#getX()
	*/
	public int getX() {
		return _rect.x;
	}

	/** Reply the vertical position of this Fig.
	     *
	     * @return vertical position.
	     */
	public int getY() {
		return _rect.y;
	}

	/** Change the position and the dimension of this Fig.
	     *
	     * @param x horizontal position of this Fig.
	     * @param y vertical position of this Fig.
	     * @param width width of this Fig.
	     * @param height height of this Fig.
	     * @see mygale.mcge.fig.IFig#setBounds(int, int, int, int)
	 */
	public void setBounds(int x, int y, int width, int height) {		
		_rect.setBounds(x, y, width, height);
	}

	/** Set the height of this Fig.
	     *
	     * @param height height.
	     * @see mygale.mcge.fig.IFig#setHeight(int)
	 */
	public void setHeight(int height) {
		_rect.height = height;
	}

	/** Change the position of this Fig.
	     *
	     * @param x the new horizontal position of this Fig.
	     * @param y the new vertical position of this Fig.
	     * @see mygale.mcge.fig.IFig#setLocation(int, int)
	 */
	public void setLocation(int x, int y) {
		_rect.x = x;
		_rect.y = y;
	}

	/** Set the width of this Fig.
	     *
	     * @param width width.
	     * @see mygale.mcge.fig.IFig#setWidth(int)
	 */
	public void setWidth(int width) {
		_rect.width = width;
	}

	/** Move this Fig.
	     *
	     * @param dx horizontal movement dimension.
	     * @param dy vertical movement dimension.
	     * @see mygale.mcge.fig.IFig#translate(int, int)
	 */
	public void translate(int dx, int dy) {
		_rect.translate(dx, dy);
	}

	/** Reply <code>true</code> if the entire figure is contained within
	     *  the given Rectangle. This can be used by ModeSelect to select
	     *  Figs that are totally within the selection rectangle.
	     *
	     *  @param r a rectangle
	     *  @return <code>true</code> if the entire figure is contained within
	     *          the rectangle parameter. otherwise <code>false</code>
	     * @see mygale.mcge.fig.IFig#within(Rectangle)
	 */
	public boolean within(Rectangle r) {
		/*
		return r.contains(_rect.x, _rect.y) && r.contains(_rect.x+_rect.width,
		                       _rect.y + _rect.height) ;
		                       */
		return r.contains(_rect);
	}

	/** Reply the position and the dimension of the area that
	     *  will be refresh when the Fig was damaged.
	     *
	     * @param prefs Prefs
	 * @return the damaged area.
	     * @see mygale.mcge.fig.IFig#getDamagedBounds(Prefs)
	 */
	public Rectangle getDamagedBounds(Prefs prefs) {
		Rectangle r = new Rectangle(getBounds());
		int selectionSize = prefs.getHandleSize();
		r.x -= selectionSize + 5;
		r.y -= selectionSize + 5;
		r.width += (selectionSize + 5) * 2;
		r.height += (selectionSize + 5) * 2;
		return r;
	}

	/** Reply <code>true</code> if the given Point
	     *  is in this Fig. This is used to determine  if the user
	     *  is trying to select this Fig.
	     *
	     * @param p the Pointd} that may contained in this Fig.
	     * @param prefs Prefs
	 * @return <code>true</code> if <var>p</var> is in this
	     *         Fig, <code>false</code> otherwise.
	     * @see mygale.mcge.fig.IFig#hit(Point, Prefs)
	 */
	public boolean hit(Point p, Prefs prefs) {
		return _rect.contains(p);
	}

	/** Reply <code>true</code> if the given Rectangle
	 *  contains some pixels of the Fig. This is used to determine
	 *  if the user is trying to select this Fig.
	 *
	 * @param r the Rectangle that may contains a part of this Fig.
	 * @param prefs Prefs
	 * @return <code>true</code> if <var>r</var> contains a
	 *         part of this Fig, <code>false</code> otherwise.
	 */
	public boolean hit(Rectangle r, Prefs prefs) {
		return _rect.intersects(r);
	}

	/** Reply <code>true</code> if the object intersects the given rectangle. Used
	    *  for selective redrawing and by ModeSelect to select all Figs
	    *  that are partly within the selection rectangle.
	    *
	    *  @param r a rectangle
	    *   * @param prefs Prefs
	@return <code>true</code> if the figure intersect the given
	    *          rectangled,. otherwise <code>false</code>
	    * @see mygale.mcge.fig.IFig#intersects(Rectangle, Prefs)
	 */
	public boolean intersects(Rectangle r, Prefs prefs) {
		/*
		return !((r.x + r.width <= _rect.x) ||
		      (r.y + r.height <= _rect.y) ||
		      (r.x >= _rect.x + _rect.width) ||
		      (r.y >= _rect.y + _rect.height));
		      */
		return r.intersects(_rect);

	}
}
