package mygale.mcge.fig;
// Copyright (C) 2000  Mahdi Hannoun, Stphane Galland
//
// This program is free software; you can redistribute it and/or modify

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import junit.framework.Assert;

import mygale.mcge.fig.arrow.Arrow;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.util.prefs.Prefs;

import de.uniwue.ki.util.Utils;

/** Abstract {@link Fig Fig} classe for representing edges between
 *  {@link FigPort ports}.
 *  <p><center>
 *    <img src="doc-files/FigEdge_uml.gif" width="313" height="200" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="118,0,208,28"  href="FigModelElement.html">
 *      <area shape="Rect" coords="239,0,311,28"  href="Mode.html">
 *      <area shape="Rect" coords="0,54,73,81"  href="arrow/Arrow.html">
 *      <area shape="Rect" coords="239,54,311,81"  href="FigPort.html">
 *      <area shape="Rect" coords="59,117,149,144"  href="FigEdgeLinear.html">
 *      <area shape="Rect" coords="181,117,271,144"  href="FigEdgePoly.html">
 *      <area shape="Rect" coords="118,172,208,199"  href="FigEdgeSpline.html">
 *    </map>
 *  </center></p>

 */
public abstract class AbstractFigEdge extends AbstractFigLayeredObject implements IFigEdge {

	protected final class ShadowHelper {
		public Point _first, _second, _secondLast, _last;

		ShadowHelper() {
			super();
		}

		public void calculatePoints(int dx, int dy, boolean shadowModeOn) {

			int size = getCtrlPointSize();

			if (!shadowModeOn) {

				_first = getCtrlPointAt(0);
				_second = getCtrlPointAt(1);
				_secondLast = getCtrlPointAt(size - 2);
				_last = getCtrlPointAt(size - 1);

			} else {

				_first = new Point(getCtrlPointAt(0));
				_last = new Point(getCtrlPointAt(size - 1));

				calculateSourceAndDestEdgePoint(_first, _last, dx, dy);

				if (size == 2) {
					_second = _last;
					_secondLast = _first;

				} else {
					_second = getCtrlPointAt(1);
					_secondLast = getCtrlPointAt(size - 2);
				}

			}
		}
	}

	protected static final void drawLine(Graphics2D g2d, Point begin, Point end) {
		g2d.drawLine(begin.x, begin.y, end.x, end.y);
	}

	protected static final void drawLine(Graphics2D g2d, Point begin, Point end, int dx, int dy) {
		g2d.drawLine(begin.x + dx, begin.y + dy, end.x + dx, end.y + dy);
	}

	static double getAngle(int x1, int y1, int x2, int y2) {
		double xnbegin = x1 - x2;
		double ynbegin = y1 - y2;
		double norme = Math.sqrt(xnbegin * xnbegin + ynbegin * ynbegin);
		double angle;
		if (norme != 0.0) {
			ynbegin = ynbegin / norme;
			angle = Math.asin(ynbegin);
			if (x1 < x2)
				angle = Math.PI - angle;
		} else {
			angle = Double.NaN;
		}
		return angle;
	}

	/** Try to remove control points that are in the same location.
	     *
	     * @param edge IFigEdge
	 * @param ctrlPoint Point
	 * @param list java.util.List
	 * @return boolean
	 */
	public static boolean tryToRemoveTheCurrentCtrlPoint(
		IFigEdge edge,
		Point ctrlPoint,
		java.util.List list) {

		int size = list.size();

		if (size == 2) {
			return false;
		}

		int ctrlPointIndex = list.indexOf(ctrlPoint);

		// try to remove the current ctrl point if this point
		// is over an over control point

		for (int i = 0; i < size; i++) {
			if (i != ctrlPointIndex) {
				Point pt2 = (Point) list.get(i);
				if (Point2D.distance(ctrlPoint.x, ctrlPoint.y, pt2.x, pt2.y) < 6.0) {
					list.remove(ctrlPointIndex);
					return true;
				}
			}
		}

		// try to remove the current ctrl point if the
		// two line connected to this point has the same
		// angle.

		Point pt1 = (Point) list.get(ctrlPointIndex - 1);
		Point pt2 = (Point) list.get(ctrlPointIndex + 1);

		if (Math
			.abs(
				getAngle(pt1.x, pt1.y, ctrlPoint.x, ctrlPoint.y)
					- getAngle(ctrlPoint.x, ctrlPoint.y, pt2.x, pt2.y))
			< Math.PI * 5 / 180) {
			// 10 degres
			list.remove(ctrlPointIndex);
			return true;
		}

		return false;

	}

	private IFigNode _beginFigNode;

	/** This is the set of control points that composed
	 *  this edge.
	 */
	protected final ArrayList _controlPoints = new ArrayList(2);

	/** This is the arrow connected to the target port.
	 */
	protected Arrow _endArrow = null;
	private IFigNode _endFigNode;

	protected final ShadowHelper _shadowHelper = new ShadowHelper();

	/** <code>true</code> if the edge is currently
	 *  display as an shadow (during moving).
	 */
	protected int _shadowMoveMode = IFigEdge.NO_SHADOW_MOVE;

	/** This is the arrow connected to the source port.
	 */
	protected Arrow _startArrow = null;

	/** Construct a new FigEdge.
	     */
	public AbstractFigEdge() {
		this(0, 0, 40, 40);
	}

	////
	// Constructors

	/** Contruct a new FigEdge.
	 *
	 * @param xfrom horizontal position of the begin point for this FigEdge.
	 * @param yfrom vertical position of the begin point for this FigEdge.
	 * @param xto horizontal position of the end point for this FigEdge.
	 * @param yto vertical position of the end point for this FigEdge.
	 */
	public AbstractFigEdge(int xfrom, int yfrom, int xto, int yto) {
		super();

		_controlPoints.add(new Point(xfrom, yfrom));
		_controlPoints.add(new Point(xto, yto));
		calcBounds();
	}

	/** Initiate the shadowing drawing.
	     */
	public void beginShadowMoveMode(int shadowMoveMode) {
		_shadowMoveMode = shadowMoveMode;
	}

	/** Calculate the bounds of arrows.
	     *  This method call the method
	     *  {@link #calcArrowBounds(Arrow,Point,Point) calcArrowBounds(Arrow,Point,Point)}
	     *  for each arrow.
	     *  We assume that this method updates
	     *  <var>_rect</var>.
	     */
	protected void calcArrowBounds(Rectangle dstRect, int dx, int dy) {

		if (_startArrow != null)
			dstRect.add(dx + _startArrow.getWidth() + _rect.x, dy + _startArrow.getHeight() + _rect.y);

		if (_endArrow != null)
			dstRect.add(dx + _endArrow.getWidth() + _rect.x, dy + _endArrow.getHeight() + _rect.y);
	}

	/** Calculate the bounds of the fig.
	     */
	public void calcBounds() {
		calcBounds(_rect, 0, 0);
	}

	/** Calculate the bounds of the fig.
	     */
	private void calcBounds(Rectangle dstRect, int dx, int dy) {

		switch (_shadowMoveMode) {

			case IFigEdge.NO_SHADOW_MOVE :
			case IFigEdge.BOTH_PORTS_SHADOW_MOVE :
				{
					Point firstPoint = getCtrlPointAt(0);

					dstRect.setBounds(firstPoint.x, firstPoint.y, dx, dy);

					for (int i = 1; i < getCtrlPointSize(); i++) {
						Point pt = (Point) _controlPoints.get(i);
						dstRect.add(pt.x + dx, pt.y + dy);

					}

					calcArrowBounds(dstRect, dx, dy);
					return;
				}
			case IFigEdge.SRC_PORTS_SHADOW_MOVE :
				{
					Point firstPoint = getCtrlPointAt(0);

					dstRect.setBounds(firstPoint.x, firstPoint.y, dx, dy);

					for (int i = 1; i < getCtrlPointSize(); i++) {
						dstRect.add((Point) _controlPoints.get(i));
					}

					calcArrowBounds(dstRect, dx, dy);
					return;
				}

			case IFigEdge.DST_PORTS_SHADOW_MOVE :
				{
					int size = getCtrlPointSize();
					Point lastPoint = getCtrlPointAt(size - 1);

					dstRect.setBounds(lastPoint.x, lastPoint.y, dx, dy);

					for (int i = 0; i < size - 1; i++) {
						dstRect.add((Point) _controlPoints.get(i));
					}

					calcArrowBounds(dstRect, dx, dy);
					return;
				}
		}

		Assert.fail();

	}

	protected void calculateSourceAndDestEdgePoint(Point sourceEdgePoint, Point destEdgePoint) {
		calculateSourceAndDestEdgePoint(sourceEdgePoint, destEdgePoint, 0, 0);
	}

	protected void calculateSourceAndDestEdgePoint(
		Point sourceEdgePoint,
		Point destEdgePoint,
		int dx,
		int dy) {

		if ((_beginFigNode == null) || (_endFigNode == null)) {
			return;
		}

		int size = getCtrlPointSize();

		if (size == 2) {
			calculateSourceAndDestEdgePointTwoControlPoints(sourceEdgePoint, destEdgePoint, dx, dy);

		} else if (size > 2) {
			calculateSourceAndDestEdgePointMoreThanTwoControlPoints(sourceEdgePoint, destEdgePoint, dx, dy);

		} else {
			throw new MCGEInternalError("control points size < 2");
		}
	}

	private void calculateSourceAndDestEdgePointMoreThanTwoControlPoints(
		Point sourceEdgePoint,
		Point destEdgePoint,
		int dx,
		int dy) {

		int size = getCtrlPointSize();

		switch (_shadowMoveMode) {

			case IFigEdge.NO_SHADOW_MOVE :
			case IFigEdge.BOTH_PORTS_SHADOW_MOVE :

				/*
				   sourceEdgePoint.translate(dx, dy);
				   destEdgePoint.translate(dx, dy);
				   */
				sourceEdgePoint.setLocation(_beginFigNode.intersectWith(getCtrlPointAt(1), dx, dy));
				destEdgePoint.setLocation(_endFigNode.intersectWith(getCtrlPointAt(size - 2), dx, dy));
				return;

			case IFigEdge.SRC_PORTS_SHADOW_MOVE :
				sourceEdgePoint.setLocation(_beginFigNode.intersectWith(getCtrlPointAt(1), dx, dy));
				destEdgePoint.setLocation(_endFigNode.intersectWith(getCtrlPointAt(size - 2), 0, 0));
				return;

			case IFigEdge.DST_PORTS_SHADOW_MOVE :
				sourceEdgePoint.setLocation(_beginFigNode.intersectWith(getCtrlPointAt(1), 0, 0));
				destEdgePoint.setLocation(_endFigNode.intersectWith(getCtrlPointAt(size - 2), dx, dy));
				return;

		}

		Assert.fail();

	}

	private void calculateSourceAndDestEdgePointTwoControlPoints(
		Point sourceEdgePoint,
		Point destEdgePoint,
		int dx,
		int dy) {

		Point sourcePortCenter =
			new Point(
				_beginFigNode.getX() + _beginFigNode.getX() + _beginFigNode.getWidth() / 2,
				_beginFigNode.getY() + _beginFigNode.getY() + _beginFigNode.getHeight() / 2);

		Point destPortCenter =
			new Point(
				_endFigNode.getX() + _endFigNode.getX() + _endFigNode.getWidth() / 2,
				_endFigNode.getY() + _endFigNode.getY() + _endFigNode.getHeight() / 2);

		int sourceDx = 0;
		int sourceDy = 0;

		int destDx = 0;
		int destDy = 0;

		switch (_shadowMoveMode) {

			case IFigEdge.BOTH_PORTS_SHADOW_MOVE :
			case IFigEdge.NO_SHADOW_MOVE :

				sourceDx = dx;
				sourceDy = dy;
				sourcePortCenter.translate(dx, dy);
				destDx = dx;
				destDy = dy;
				destPortCenter.translate(dx, dy);
				break;

			case IFigEdge.SRC_PORTS_SHADOW_MOVE :
				sourceDx = dx;
				sourceDy = dy;
				sourcePortCenter.translate(dx, dy);
				break;

			case IFigEdge.DST_PORTS_SHADOW_MOVE :
				destDx = dx;
				destDy = dy;
				destPortCenter.translate(dx, dy);
				break;

			default :
				Assert.fail();
		}

		// compute the source point of the edge
		sourceEdgePoint.setLocation(_beginFigNode.intersectWith(destPortCenter, sourceDx, sourceDy));
		// compute the destination point of the edge
		destEdgePoint.setLocation(_endFigNode.intersectWith(sourceEdgePoint, destDx, destDy));

		// recompute the sourceEdgePoint and the destination edge
		sourceEdgePoint.setLocation(_beginFigNode.intersectWith(destEdgePoint, sourceDx, sourceDy));
		destEdgePoint.setLocation(_endFigNode.intersectWith(sourceEdgePoint, destDx, destDy));
	}

	/** Stop the shadowing drawing.
	     * @param transX int
	 * @param transY int
	 * @see mygale.mcge.fig.IFigEdge#endShadowMoveMode(int, int)
	 */
	public void endShadowMoveMode(int transX, int transY) {

		if (_shadowMoveMode == IFigEdge.BOTH_PORTS_SHADOW_MOVE) {
			translate(transX, transY);

		} else {
			updateEndsLocation();
		}

		_shadowMoveMode = IFigEdge.NO_SHADOW_MOVE;

		/*   if (mergeCtrlPointsIfPossible()) {
		      calcBounds();
		   }
		   */

	}

	/**
	 * Insert the method's description here.
	 * Creation date: (20.04.2001 14:27:43)
	 * @return mygale.mcge.fig.FigPort
	 * @see mygale.mcge.fig.IFigEdge#getBeginFigNode()
	 */
	public IFigNode getBeginFigNode() {
		return _beginFigNode;
	}

	/** return the point of the path point edge defined
	     *  by the parameter.
	     *
	     * @param index the <var>index</var> of the point from
	     *              the path point edge to get.
	     *              0 corresponds to the sourcePortPoint
	     *              and size-1 (where "size" denote the
	     *              size of the path point list)
	     *              the destination port point.
	     * @return coordinates of the control point.
	     * @see mygale.mcge.fig.IFigEdge#getCtrlPointAt(int)
	 */
	public  Point getCtrlPointAt(int index) {
		return (Point) _controlPoints.get(index);
	}

	/** Get the edge control points.
	     *
	     * @return control points of this edge.
	     * @see mygale.mcge.fig.IFigEdge#getCtrlPoints()
	 */
	public Iterator getCtrlPoints() {
		return _controlPoints.iterator();
	}
	
	public void clearCtrlPoints() {
		
		_controlPoints.clear();
	}

	/** Return the number of point in the path of the edge.
	     *  this number est equal to 2 when the edge is a
	     *  {@link FigEdgeLinear FigEdgeLinear} :
	     *  only the port source point and the destination port point.
	     *
	     * @return an integer that is the size of the path.
	     * @see mygale.mcge.fig.IFigEdge#getCtrlPointSize()
	 */
	public final int getCtrlPointSize() {
		return _controlPoints.size();
	}

	/** Reply the arrow connected to the target port.
	     *
	     * @return the end arrow or <code>null</code> if none.
	     */
	public Arrow getEndArrow() {
		return _endArrow;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (20.04.2001 14:27:43)
	 * @return mygale.mcge.fig.FigPort
	 * @see mygale.mcge.fig.IFigEdge#getEndFigNode()
	 */
	public IFigNode getEndFigNode() {
		return _endFigNode;
	}

	/** Initiate the shadowing drawing.
	     * @param dx int
	 * @param dy int
	 * @return Rectangle
	 * @see mygale.mcge.fig.IFigEdge#getShadowBounds(int, int)
	 */
	public Rectangle getShadowBounds(int dx, int dy) {
		Rectangle rect = new Rectangle();
		calcBounds(rect, dx, dy);
		return rect;
	}

	/////
	// Arrow API

	/** Reply the arrow connected to the source port.
	 *
	 * @return the start arrow or <code>null</code> if none.
	 */
	public Arrow getStartArrow() {
		return _startArrow;
	}

	/** Reply <code>true</code> if the given Point
	 *  is in this Fig. This is used to determine
	 *  if the user is trying to select this Fig.
	 *
	 * @param p the Point that may contained in this Fig.
	 * @param prefs Prefs
	 * @return <code>true</code> if <var>p</var> is
	 *         in this Fig, <code>false</code> otherwise.
	 * @see mygale.mcge.fig.IFig#hit(Point, Prefs)
	 */
	public boolean hit(Point p, Prefs prefs) {
		if (getDamagedBounds(prefs).contains(p.x, p.y)) {
			int size = _controlPoints.size();
			for (int i = 0; i < size - 1; i++) {
				Point p1 = getCtrlPointAt(i);
				Point p2 = getCtrlPointAt(i + 1);
				if (Line2D.ptSegDist(p1.x, p1.y, p2.x, p2.y, p.x, p.y) < 5.0) {
					return true;
				}
			}
		}

		return false;
	}

	/** Insert a point into the path defined by the given index.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @param index the index of the control point in the path.
	     * @param pt coordinates of the control point.
	     * @see mygale.mcge.fig.IFigEdge#insertCtrlPointAt(int, Point)
	 */
	public void insertCtrlPointAt(int index, Point pt) {
		if ((index < 0) || (index > _controlPoints.size())) {
			throw new MCGEInternalError("try to insert point in the path " + "out of the bounds");
		}
		_controlPoints.add(index, pt);
		calcBounds();
	}

	/** Reply always <code>false</code> because
	     *  an edge cant be moved directly.
	     *  You should move the connected figure or
	     *  one of the edge control point.
	     *
	     * @return always <code>false</code>.
	     * @see mygale.mcge.fig.IFigLayered#isMovable()
	 */
	public boolean isMovable() {
		return false;
	}

	/** Reply <code>true</code> if this Fig can be
	     *  reshaped by the user.
	     *  Reply allways <code>false</code> because an
	     *  edge cant be reshape directly. You should
	     *  move the connected figures or one of the
	     *  edge control point.
	     *
	     * @return always <code>false</code>.
	     * @see mygale.mcge.fig.IFigLayered#isReshapable()
	 */
	public boolean isReshapable() {
		return false;
	}

	protected boolean mergeCtrlPointsIfPossible() {
		for (int i = 0; i < getCtrlPointSize() - 2; i++) {
			if (mergeCtrlPointsIfPossible(getCtrlPointAt(i), getCtrlPointAt(i + 1), getCtrlPointAt(i + 2))) {
				mergeCtrlPointsIfPossible();
				return true;
			}
		}

		return false;
	}

	protected boolean mergeCtrlPointsIfPossible(Point pt1, Point pt2, Point pt3) {
		if (mergeOfCtrlPointsIsPossible(pt1, pt2, pt3)) {
			_controlPoints.remove(pt2);
			return true;
		} else {
			return false;
		}
	}

	protected boolean mergeOfCtrlPointsIsPossible(Point pt1, Point pt2, Point pt3) {
		return de.uniwue.ki.util.math.MathHelper.isParallel(
			pt1.x,
			pt1.y,
			pt2.x,
			pt2.y,
			pt2.x,
			pt2.y,
			pt3.x,
			pt3.y);
	}

	protected void paintArrow(
		Graphics2D g2d,
		Prefs prefs,
		Arrow arrow,
		int from_x,
		int from_y,
		int to_x,
		int to_y,
		int dx,
		int dy) {

		if (arrow != null) {
			arrow.paint(g2d, prefs, from_x + dx, from_y + dy, to_x + dx, to_y + dy);
		}
	}

	protected void paintArrow(
		Graphics2D g2d,
		Prefs prefs,
		Arrow arrow,
		Point first,
		Point second,
		int dx,
		int dy) {
		paintArrow(g2d, prefs, arrow, first.x, first.y, second.x, second.y, dx, dy);
	}

	protected void paintArrows(Graphics2D g2d, Prefs prefs, int dx, int dy, boolean shadowModeOn) {

		// _shadowHelper.calculatePoints(dx, dy, shadowModeOn);

		paintArrows(
			g2d,
			prefs,
			_shadowHelper._first,
			_shadowHelper._second,
			_shadowHelper._secondLast,
			_shadowHelper._last,
			dx,
			dy);
	}

	/** Method to paint arrows of this Fig.
	     *
	     * @param g2d the graphic context.
	     */
	protected void paintArrows(
		Graphics2D g2d,
		Prefs prefs,
		Point firstPoint,
		Point secondPoint,
		Point secondLastPoint,
		Point lastPoint,
		int dx,
		int dy) {

		paintArrow(g2d, prefs, _startArrow, firstPoint, secondPoint, dx, dy);
		paintArrow(g2d, prefs, _endArrow, secondLastPoint, lastPoint, dx, dy);
	}

	public void paintContent(Graphics2D g2d, Prefs prefs, int dx, int dy, boolean shadowModeOn) {

		g2d.setColor(prefs.getFigEdgeColor());

		int size = getCtrlPointSize();

		if (_shadowMoveMode == IFigEdge.BOTH_PORTS_SHADOW_MOVE) {

			for (int i = 0; i < (size - 1); i++) {
				drawLine(g2d, getCtrlPointAt(i), getCtrlPointAt(i + 1), dx, dy);
			}

			drawLine(g2d, _shadowHelper._secondLast, _shadowHelper._last);

			paintArrows(g2d, prefs, dx, dy, shadowModeOn);

		} else {
			_shadowHelper.calculatePoints(dx, dy, shadowModeOn);

			if (size == 2) {
				drawLine(g2d, _shadowHelper._first, _shadowHelper._last);

			} else if (shadowModeOn) {

				if (_shadowMoveMode == IFigEdge.SRC_PORTS_SHADOW_MOVE) {
					drawLine(g2d, _shadowHelper._first, _shadowHelper._second);

				} else {
					drawLine(g2d, _shadowHelper._secondLast, _shadowHelper._last);
				}

			} else {

				drawLine(g2d, _shadowHelper._first, _shadowHelper._second);

				for (int i = 1; i < (size - 2); i++) {
					drawLine(g2d, getCtrlPointAt(i), getCtrlPointAt(i + 1));
				}

				drawLine(g2d, _shadowHelper._secondLast, _shadowHelper._last);
			}

			paintArrows(g2d, prefs, 0, 0, shadowModeOn);
		}
	}

	/** Method to paint the selection of this FigEdge.
	     *
	     * @param g2d the graphic context.
	     * @param prefs Prefs
	 */
	public void paintSelection(Graphics2D g2d, Prefs prefs) {
		int _SELECTION_SIZE = prefs.getHandleSize();
		if (_locked) {
			g2d.setColor(prefs.getLockedHandleColor());
		} else {
			g2d.setColor(prefs.getHandleColor());
		}
		int size = _controlPoints.size();
		for (int i = 0; i < size; i++) {
			Point p = getCtrlPointAt(i);
			paintHandle(
				g2d,
				p.x - _SELECTION_SIZE / 2,
				p.y - _SELECTION_SIZE / 2,
				_SELECTION_SIZE,
				_SELECTION_SIZE,
				prefs);
		}
	}

	/** Remove a point at the given index from the path.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @param index the index of the control point in the path.
	     * @see mygale.mcge.fig.IFigEdge#removeCtrlPointAt(int)
	 */
	public void removeCtrlPointAt(int index) {
		if ((index < 0) || (index >= (_controlPoints.size() - 1))) {
			throw new MCGEInternalError("try to remove point in the path " + "out of the bounds");
		}
		_controlPoints.remove(index);
		calcBounds();
	}

	/**
	 * @param newBeginFigPort IFigPort
	 * @see mygale.mcge.fig.IFigEdge#setBeginFigNode(IFigPort)
	 */
	public void setBeginFigNode(IFigNode newBeginFigNode) {
		if (Assert.ASSERTION_ON)
			Assert.assertNotNull(newBeginFigNode);

		if (!Utils.equals(newBeginFigNode, _beginFigNode)) {
			Assert.assertNull(_beginFigNode);

			_beginFigNode = newBeginFigNode;

			updateEndsLocation();
		}
	}

	// Accessors

	/** Change the position and the dimension of this FigEdge.
	 *  This method only change the location.
	 *
	 * @param x horizontal position of this FigEdge.
	 * @param y vertical position of this FigEdge.
	 * @param width width of this FigEdge (ignore).
	 * @param height height of this FigEdge (ignore).
	 * @see mygale.mcge.fig.IFig#setBounds(int, int, int, int)
	 */
	public void setBounds(int x, int y, int width, int height) {
		setLocation(x, y);
	}

	/** Set the point of the path defined by the given index.
	     *
	     * @param index the index of the control point in the path.
	     * @param pt new coordinates of the control point.
	     * @see mygale.mcge.fig.IFigEdge#setCtrlPointAt(int, Point)
	 */
	public void setCtrlPointAt(int index, Point pt) {
		if (index >= getCtrlPointSize()) {
			for (int i = getCtrlPointSize(); i <= index; i++) {
				_controlPoints.add(null);
			}
		}
		_controlPoints.set(index, pt);
		calcBounds();
	}

	public void setCtrlPoints(Iterator it) {
		_controlPoints.clear();
		while (it.hasNext()) {
			_controlPoints.add(it.next());
		}
		calcBounds();
	}

	/** Change the arrow connected to the target port.
	     *
	     * @param arrow the new arrow.
	     */
	public void setEndArrow(Arrow arrow) {
		_endArrow = arrow;
		if (_endArrow != null) {
			_endArrow.setEdge(this);
			/*
			int size = getCtrlPointSize();
			if (size > 1) {
			   Point last = getCtrlPointAt(size - 1);
			   Point previous = getCtrlPointAt(size - 2);
			}
			*/
		}
	}

	public void setEndFigPort(IFigNode newEndFigPort) {
		Assert.assertNotNull(newEndFigPort);

		if (!Utils.equals(newEndFigPort, _endFigNode)) {

			Assert.assertNull(_endFigNode);

			_endFigNode = newEndFigPort;

			updateEndsLocation();
		}
	}

	/** Change the position of this FigEdge.
	     *
	     * @param x the new horizontal position of this FigEdge.
	     * @param y the new vertical position of this FigEdge.
	     * @see mygale.mcge.fig.IFig#setLocation(int, int)
	 */
	public void setLocation(int x, int y) {
		if (_controlPoints != null) {

			for (int i = 0; i < _controlPoints.size(); i++) {
				getCtrlPointAt(i).setLocation(x, y);
			}

			updateEndsLocation();

		} else {
			super.setLocation(x, y);
		}
	}

	/** Change the arrow connected to the source port.
	     *
	     * @param arrow the new arrow.
	     */
	public void setStartArrow(Arrow arrow) {
		_startArrow = arrow;
		if (_startArrow != null) {
			_startArrow.setEdge(this);

			/*
			if (getCtrlPointSize() > 1) {
			   Point first = getCtrlPointAt(0);
			   Point second = getCtrlPointAt(1);
			}
			*/
		}
	}

	/** Move this FigEdge.
	     *
	     * @param dx horizontal movement dimension.
	     * @param dy vertical movement dimension.
	     * @see mygale.mcge.fig.IFig#translate(int, int)
	 */
	public void translate(int dx, int dy) {
		for (int i = 0; i < _controlPoints.size(); i++) {
			getCtrlPointAt(i).translate(dx, dy);
		}

		updateEndsLocation();
	}

	/** Try to remove control points that are in the same location.
	     *
	     * @param ctrlPoint Point
	 * @return boolean
	 * @see mygale.mcge.fig.IFigEdge#tryToRemoveTheCurrentCtrlPoint(Point)
	 */
	public boolean tryToRemoveTheCurrentCtrlPoint(Point ctrlPoint) {
		return tryToRemoveTheCurrentCtrlPoint(this, ctrlPoint, _controlPoints);
	}

	/** This method update the location of ends.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @see mygale.mcge.fig.IFigEdge#updateEndsLocation()
	 */
	public void updateEndsLocation() {
		calculateSourceAndDestEdgePoint(getCtrlPointAt(0), getCtrlPointAt(getCtrlPointSize() - 1));
		calcBounds();
	}
}