package mygale.mcge.fig;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import mygale.mcge.cmd.*;
import mygale.mcge.core.JGraph;
import mygale.mcge.core.JGraphEditor;
import mygale.mcge.util.prefs.Prefs;

import de.uniwue.ki.gui.util.actions.AbstractBoundAction;
import de.uniwue.ki.gui.util.actions.InspectAction;

import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.event.StateChangeMulticaster;

/** This class is the base class for basic drawing objects such as
 *  rectangles, lines, text, circles, etc.
 *  <p><center>
 *    <img src="doc-files/Fig_uml.gif" width="309" height="142" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="235,0,308,28"  href="../cmd/Cmd.html">
 *      <area shape="Rect" coords="118,0,190,28"  href="../mode/Mode.html">
 *      <area shape="Rect" coords="235,54,307,82"  href="../layer/LayerContener.html">
 *      <area shape="Rect" coords="0,54,73,82"  href="../util/prefs/Prefs.html">
 *      <area shape="Rect" coords="14,113,113,139"  href="FigModelElement.html">
 *      <area shape="Rect" coords="131,113,203,139"  href="FigText.html">
 *      <area shape="Rect" coords="221,113,293,139"  href="FigEllipse.html">
 *      <area shape="Rect" coords="0,0,73,28"  href="../core/JGraphEditor.html">
 *    </map>
 *  </center></p>
 *
 */
public abstract class AbstractFigLayered extends AbstractFig implements IFigLayered {

	private transient ChangeListener _changeListener = null;
	/** <code>true</code> if this object was locked
	 *  and cannot be moved/changed by the user. */
	protected boolean _locked = false;

	/** <code>true</code> if this object was selected.
	 */
	protected boolean _selected = false;

	/** Construct a new Fig. The default location is set to (0,0) and
	     *  the default size to
	     *  ({@link #getMinimalWidth() getMinimalWidth},
	     *  {@link #getMinimalHeight() getMinimanHeight}).
	     */
	public AbstractFigLayered() {
		super();

		setBounds(0, 0, getMinimalWidth(), getMinimalHeight());
	}

	private void addActions(JPopupMenu popupMenu, Iterator it) {
		while (it.hasNext()) {
			Object currAction = it.next();

			if (currAction instanceof Action) {
				popupMenu.add(AbstractBoundAction.createJMenuItem((Action) currAction, false));
			} else if (currAction == null) {
				popupMenu.addSeparator();
			} else {
				popupMenu.add((JMenuItem) currAction);
			}
		}
	}

	/**
	 * @param newListener ChangeListener
	 * @see mygale.mcge.fig.IFigLayered#addChangeListener(ChangeListener)
	 */
	public void addChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.add(_changeListener, newListener);
	}

	/** This method paint a string into a clipping rect.
	     *
	     * @param g2d the graphic context.
	     * @param text the text to draw.
	     * @param x x coordinate where the text must be draw.
	     * @param y y coordinate where the text must be draw.
	     */
	protected void drawClippedString(Graphics2D g2d, String text, int x, int y) {
		Shape oldClipArea = g2d.getClip();
		g2d.setClip(getX() + 1, getY() + 1, getWidth() - 2, getHeight() - 2);
		g2d.drawString(text, x, y);
		g2d.setClip(oldClipArea);
	}

	/** This method permits to edit the properties of this object and if it is
	     *  a {@link FigObject FigObject}, the properties of the attached
	     *  {@link mygale.mcge.net.NetObject NetObject}.
	     *  This method must be otherwrite by subclasses.
	     * @param parent Container
	 * @see mygale.mcge.fig.IFigLayered#editProperties(Container)
	 */
	public void editProperties(Container parent) {
	}

	/**
	 * Method to support listener events.
	 */
	public void fireStateChanged() {
		if (_changeListener != null) {
			_changeListener.stateChanged(new ChangeEvent(this));
		}
	}

	/** Reply the minimal height of the figure.
	     * @return int
	 * @see mygale.mcge.fig.IFigLayered#getMinimalHeight()
	 */
	public int getMinimalHeight() {
		return 40;
	}

	/** Reply the minimal width of the figure .
	     * @return int
	 * @see mygale.mcge.fig.IFigLayered#getMinimalWidth()
	 */
	public int getMinimalWidth() {
		return 40;
	}

	/** This method return the popup for this Fig.
	     *
	     * @param parent Container
	 * @param forOnlyOneFig <code>true</code> if caller of this
	     *                      method need a popupmenu for one
	     *                      Fig, <code>false</code> if
	     *                      the caller need a popup menu for
	     *                      more than one Fig.
	     * @return a popup menu.
	     * @see mygale.mcge.fig.IFigLayered#getPopupMenu(Container, boolean)
	 */
	public final JPopupMenu getPopupMenu(Container parent, boolean forOnlyOneFig) {
		JGraph graph = (JGraph) parent;
		JGraphEditor editor = graph.getGraphEditor();

		JPopupMenu popupMenu = new JPopupMenu();

		popupMenu.add(AbstractBoundAction.createJMenuItem(new CmdProperties(editor, 16), true));

		addActions(popupMenu, getPopupMenuImportantActions(graph, forOnlyOneFig));

		popupMenu.addSeparator();

		addActions(popupMenu, getPopupMenuMoreActions(graph, forOnlyOneFig));

		return popupMenu;
	}

	protected Iterator getPopupMenuImportantActions(JGraph graph, boolean forOnlyOneFig) {
		return CollectionsHelper.EMPTY_ITERATOR;
	}

	protected Iterator getPopupMenuMoreActions(JGraph graph, boolean forOnlyOneFig) {

		JGraphEditor editor = graph.getGraphEditor();

		LinkedList list = new LinkedList();

		list.add(new CmdDelete(editor, 16));

		// Locking
		if (forOnlyOneFig) {
			if (isLocked()) {
				list.add(new CmdUnlock(editor, 16));
			} else {
				list.add(new CmdLock(editor, 16));
			}

		} else {
			JMenu menuLocking = new JMenu("Locking");

			AbstractBoundAction.setFont(menuLocking, false);
			menuLocking.add(AbstractBoundAction.createJMenuItem(new CmdLock(editor, 16)));
			menuLocking.add(AbstractBoundAction.createJMenuItem(new CmdUnlock(editor, 16)));

			list.add(menuLocking);
		}

		list.add(null);

		// Create submenu "Ordering"
		JMenu menuOrdering = new JMenu("Ordering");
		AbstractBoundAction.setFont(menuOrdering, false);
		menuOrdering.add(AbstractBoundAction.createJMenuItem(new CmdBringToFront(editor, 16)));
		menuOrdering.add(AbstractBoundAction.createJMenuItem(new CmdSendToBack(editor, 16)));

		if (forOnlyOneFig) {
			menuOrdering.add(AbstractBoundAction.createJMenuItem(new CmdBringForward(editor, 16)));
			menuOrdering.add(AbstractBoundAction.createJMenuItem(new CmdSendBackward(editor, 16)));
		}

		list.add(menuOrdering);
		// Creat submenu "Alignment"
		if (!forOnlyOneFig) {
			JMenu menuAlignment = new JMenu("Alignment");

			AbstractBoundAction.setFont(menuAlignment, false);
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignLeft(editor, 16)));
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignHCenter(editor, 16)));
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignRight(editor, 16)));
			menuAlignment.addSeparator();
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignTop(editor, 16)));
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignVCenter(editor, 16)));
			menuAlignment.add(AbstractBoundAction.createJMenuItem(new CmdAlignBottom(editor, 16)));

			list.add(menuAlignment);
		}

		if (forOnlyOneFig) {
			list.add(
				AbstractBoundAction.createJMenuItem(
					new InspectAction(graph.getSelection().getAllFigs().next()),
					false));
		}

		//colorChooser
		//		list.add(null);
		//		list.add(
		//			AbstractBoundAction.createJMenuItem(
		//				new ColorChooseAction(graph.getSelection().getAllFigs()),
		//				false));

		return list.iterator();
	}

	/** Reply the direction of the reshaping if the mouse is on a reshape area.
	     *
	    @link #RESHAPE_NONE #RESHAPE_NONE}.
	     * @param prefs Prefs
	 * @param x the horizontal position of the mouse.
	     * @param y the vertical position of the mouse.
	     * @return the reshape direction or { * @see mygale.mcge.fig.IFigLayered#getReshapeDirection(Prefs, int, int)
	 */
	public int getReshapeDirection(Prefs prefs, int x, int y) {
		int handleSize = prefs.getHandleSize();
		int x1 = getX() - handleSize / 2;
		int y1 = getY() - handleSize / 2;
		int xc = getX() + (getWidth() - handleSize) / 2;
		int yc = getY() + (getHeight() - handleSize) / 2;
		int x2 = getX() + getWidth() - handleSize / 2;
		int y2 = getY() + getHeight() - handleSize / 2;
		int pts[] = { x1, y1, xc, y1, x2, y1, x1, yc, x2, yc, x1, y2, xc, y2, x2, y2 };
		for (int i = 0; i < 16; i += 2) {
			if (x >= pts[i]
				&& x <= (pts[i] + handleSize)
				&& y >= pts[i + 1]
				&& y <= (pts[i + 1] + handleSize)) {
				return (i / 2) + 1;
			}
		}
		return IFigLayered.RESHAPE_NONE;
	}

	/** Reply <code>true</code> if the object intersects the given rectangle. Used
	     *  for selective redrawing and by ModeSelect to select all Figs
	     *  that are partly within the selection rectangle.
	     *
	     *  @param r a rectangle
	     *   * @param prefs Prefs
	@return <code>true</code> if the figure intersects
	     *          the given rectangle. otherwise <code>false</code>
	     * @see mygale.mcge.fig.IFig#intersects(Rectangle, Prefs)
	 */
	public boolean intersects(Rectangle r, Prefs prefs) {
		/*
		Rectangle rDamaged = getDamagedBounds();
		return !((r.x + r.width < rDamaged.x) ||
		 (r.y + r.height < rDamaged.y) ||
		 (r.x > rDamaged.x + rDamaged.width) ||
		 (r.y > rDamaged.y + rDamaged.height));
		 */
		return r.intersects(getDamagedBounds(prefs));
	}

	/** Reply the lock-state of this Fig. When a Fig is locked,
	     *  it should not be moved by the user.
	     *
	     * @return <code>true</code> if this Fig is locked.
	     * @see mygale.mcge.fig.IFigLayered#isLocked()
	 */
	public boolean isLocked() {
		return _locked;
	}

	/** Reply <code>true</code> if this Fig can be
	     *  moved around by the user
	     *
	     * @return <code>true</code> if this Fig can be moved,
	     *         <code>false</code> otherwise.
	     * @see mygale.mcge.fig.IFigLayered#isMovable()
	 */
	public boolean isMovable() {
		return true;
	}

	/** Reply <code>true</code> if this Fig can be
	     *  reshaped by the user
	     *
	     * @return <code>true</code> if this Fig can be reshaped,
	     *         <code>false</code> otherwise.
	     * @see mygale.mcge.fig.IFigLayered#isReshapable()
	 */
	public boolean isReshapable() {
		return true;
	}

	/** Reply the select-state of this Fig.
	     *
	     * @return <code>true</code> if this Fig is locked.
	     * @see mygale.mcge.fig.IFigLayered#isSelected()
	 */
	public boolean isSelected() {
		return _selected;
	}

	////
	// Painting methods

	/** Method to paint this Fig. This method call
	 *  {@link #paintContent(Graphics) paintContent()} and
	 *  {@link #paintSelection(Graphics) paintSelection()}.
	 *
	 * @param g2d Graphics2D
	 * @param prefs Prefs
	 * @see mygale.mcge.fig.IFigLayered#paint(Graphics2D, Prefs)
	 */
	public void paint(Graphics2D g2d, Prefs prefs) {
		paintContent(g2d, prefs, 0, 0, false);
		if (isSelected())
			paintSelection(g2d, prefs);
	}

	/** Method to paint the content of this Fig. Subclasses must
	     *  overwrite this method.
	     *
	     * @param g2d Graphics2D
	 * @param prefs Prefs
	 */
	public void paintContent(Graphics2D g2d, Prefs prefs) {
	}

	protected void paintContent(Graphics2D g2d, Prefs prefs, int dx, int dy, boolean shadowModeOn) {

		AffineTransform oldTransform = new AffineTransform(g2d.getTransform());

		g2d.translate(dx, dy);

		paintContent(g2d, prefs, shadowModeOn);

		g2d.setTransform(oldTransform);
	}

	protected void paintContent(Graphics2D g2d, Prefs prefs, boolean shadowModeOn) {

		paintContent(g2d, prefs);
	}

	/** Method to paint the shadow of this Fig. By default
	     *  this method draw a rectangle corresponding to the bounds.
	     *
	     * @param g the graphic context.
	     * @param dx horizontal translation of this Fig.
	     * @param dy vertical translation of this Fig.
	     */
	protected void paintContentShadowed(Graphics2D g2d, int dx, int dy, Prefs prefs) {

		Composite oldComposite = g2d.getComposite();

		g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f)); // macht's durchsichtig

		paintContent(g2d, prefs, dx, dy, true);

		g2d.setComposite(oldComposite);
	}

	/** Method to paint an handle for this Fig
	     *  By default, this method draw rectangle.
	     *
	     * @param g2d Graphics2D
	 * @param x where to put the handle along the x axis.
	     * @param y where to put the handle along the y axis.
	     * @param width width of the handle
	     * @param height height of the handle.
	     * @param prefs Prefs
	 */
	public void paintHandle(Graphics2D g2d, int x, int y, int width, int height, Prefs prefs) {
		g2d.fillRect(x, y, width, height);
	}

	/** Method to paint the selection of this Fig.
	     *  By default, this method draw 8 rectangulars.
	     *
	     * @param g2d the graphics context.
	     * @param prefs Prefs
	 */
	public void paintSelection(Graphics2D g2d, Prefs prefs) {
		int selectionSize = prefs.getHandleSize();
		int x1 = getX() - selectionSize / 2;
		int y1 = getY() - selectionSize / 2;
		int xc = getX() + (getWidth() - selectionSize) / 2;
		int yc = getY() + (getHeight() - selectionSize) / 2;
		int x2 = getX() + getWidth() - selectionSize / 2;
		int y2 = getY() + getHeight() - selectionSize / 2;
		int pts[] = { x1, y1, xc, y1, x2, y1, x1, yc, x2, yc, x1, y2, xc, y2, x2, y2 };
		if (_locked) {
			g2d.setColor(prefs.getLockedHandleColor());
		} else {
			g2d.setColor(prefs.getHandleColor());
		}
		for (int i = 0; i < 16; i += 2) {
			paintHandle(g2d, pts[i], pts[i + 1], selectionSize, selectionSize, prefs);
		}
	}

	/** Method to paint the shadow of this Fig. By default
	     *  this method draw a rectangle corresponding to the bounds.
	     *
	     * @param g2d the graphic context.
	     * @param dx horizontal translation of this Fig.
	     * @param dy vertical translation of this Fig.
	     */
	public void paintShadow(Graphics2D g2d, int dx, int dy, Prefs prefs) {
		paintContentShadowed(g2d, dx, dy, prefs);
	}

	public void removeChangeListener(ChangeListener newListener) {
		_changeListener = StateChangeMulticaster.remove(_changeListener, newListener);
	}

	/** Set the lock-state of this Fig. When a Fig is locked,
	     *  it should not be moved by the user.
	     *
	     * @param lock <code>true</code> if this Fig is locked.
	     * @see mygale.mcge.fig.IFigLayered#setLocked(boolean)
	 */
	public void setLocked(boolean lock) {
		_locked = lock;
		fireStateChanged();
	}

	/** Set the select-state of this Fig.
	     *
	     * @param select <code>true</code> if this Fig was selected.
	     * @see mygale.mcge.fig.IFigLayered#setSelected(boolean)
	 */
	public void setSelected(boolean select) {
		if (isSelected() != select) {
			_selected = select;
			fireStateChanged();
		}
	}
}