package mygale.mcge.fig;

import java.awt.*;

import mygale.mcge.net.INetObject;
import mygale.mcge.util.prefs.Prefs;

/** Abstract class to present a node in a diagram.
 *  <p><center>
 *    <img src="doc-files/FigNode_uml.gif" width="299" height="146" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="104,0,194,28"  href="FigModelElement.html">
 *      <area shape="Rect" coords="0,54,71,81"  href="FigEdge.html">
 *      <area shape="Rect" coords="225,54,298,81"  href="FigPort.html">
 *      <area shape="Rect" coords="45,117,135,144"  href="FigNodeEllipse.html">
 *      <area shape="Rect" coords="167,117,257,144"  href="FigNodeRect.html">
 *    </map>
 *  </center></p>

 */
public abstract class AbstractFigNode extends AbstractFigLayeredObject implements IFigNode {


	private Color _backgroundColorValue = null;

	public AbstractFigNode() {
		super();

		updateDimensions();
	}

	public Color getBackgroundColor() {
		return _backgroundColorValue;
	}

	public Color getBackgroundColor(Prefs prefs) {
		return (_backgroundColorValue != null) ? _backgroundColorValue : prefs.getFigBackgroundColor();
	}

	public void setBackgroundColor(Color newColor) {
		_backgroundColorValue = newColor;

		fireStateChanged();
	}

	public void add(Point pointToAdd) {
		super.add(pointToAdd);

		updateDimensions();
	}

	public void add(Rectangle boundsToAdd) {
		super.add(boundsToAdd);

		updateDimensions();
	}

	/** Reply the position and the dimension of the area that
	     *  will be refresh when the Fig was damaged.
	     *
	     * @param prefs Prefs
	 * @return the damaged area.
	     * @see mygale.mcge.fig.IFig#getDamagedBounds(Prefs)
	 */
	public Rectangle getDamagedBounds(Prefs prefs) {
		Rectangle r = super.getDamagedBounds(prefs);
		return r;
	}

	/////
	//  paint methods

	/** Method to paint this FigNode. This method call
	 *  {@link #paintContent(Graphics) paintContent()},
	 *  {@link #paintPorts(Graphics) paintPorts()} and
	 *  {@link #paintSelection(Graphics) paintSelection()}.
	 *
	 * @param g2d the graphic context.
	 * @param prefs Prefs
	 * @see mygale.mcge.fig.IFigLayered#paint(Graphics2D, Prefs)
	 */
	public void paint(Graphics2D g2d, Prefs prefs) {
		Rectangle oldClip = g2d.getClipBounds();
		Rectangle bounds = getBounds();

		g2d.clipRect(bounds.x, bounds.y, bounds.width + 1, bounds.height + 1);
		paintContent(g2d, prefs);
		paintPorts(g2d, prefs);
		g2d.setClip(oldClip);

		if (isSelected())
			paintSelection(g2d, prefs);

	}

	/** Method to paint this
	     *  {@link FigPort FigPorts} owned bt thisFigNode.
	     *  This method call
	     *  {@link FigPort#paint(Graphics,int,int) paint()}
	     *  of each FigPort.
	     *
	     * @param g2d the graphic context.
	     * @param prefs Prefs
	 */
	public void paintPorts(Graphics2D g2d, Prefs prefs) {
		if (_active)
			paintActivatedPort(g2d, prefs);
	}

	/** Change the position and the dimension of this Fig.
	     *
	     * @param x horizontal position of this Fig.
	     * @param y vertical position of this Fig.
	     * @param width width of this Fig.
	     * @param height height of this Fig.
	     * @see mygale.mcge.fig.IFig#setBounds(int, int, int, int)
	 */
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);

		updateDimensions();
	}

	/** Set the Height of this Fig.
	     *
	     * @param Height Height.
	     * @see mygale.mcge.fig.IFig#setHeight(int)
	 */
	public void setHeight(int height) {
		super.setHeight(height);

		updateDimensions();
	}

	/** Change the position of this Fig.
	     *
	     * @param x the new horizontal position of this Fig.
	     * @param y the new vertical position of this Fig.
	     * @see mygale.mcge.fig.IFig#setLocation(int, int)
	 */
	public void setLocation(int x, int y) {
		Point oldloc = getLocation();
		translate(x - oldloc.x, y - oldloc.y);
	}

	// Accessors

	/** Set the {@link mygale.mcge.net.NetObject NetObject}
	 *  that owns this Fig.
	 *  This method set the owner and compute the size of
	 *  this node bounds and of the ports' bounds.
	 *
	 * @param netobject the {@link mygale.mcge.net.NetObject NetObject}.
	 */
	public void setNetObject(INetObject netobject) {
		super.setNetObject(netobject);

		updateDimensions();
	}

	/** Set the width of this Fig.
	     *
	     * @param width width.
	     * @see mygale.mcge.fig.IFig#setWidth(int)
	 */
	public void setWidth(int width) {
		super.setWidth(width);

		updateDimensions();
	}

	/** Move this Fig.
	     *
	     * @param dx horizontal movement dimension.
	     * @param dy vertical movement dimension.
	     * @see mygale.mcge.fig.IFig#translate(int, int)
	 */
	public void translate(int dx, int dy) {
		super.translate(dx, dy);

		updateDimensions();
	}

	protected void updateDimensions() {

		// weil port == node => keine Neuberechnungen mehr ntig
		//				if (getNetObject() != null) {
		//					calcFigPortBounds();
		//				}
	}

	// ++++++++++++++++++++++++++++++++++++++++++++
	// ++++++++++++++++++++++++++++++++++++++++++++
	// ++++++++++++++++++++++++++++++++++++++++++++

	/** <code>true</code> if the port is actived.
	   */
	protected boolean _active = false;

	//  various accessors

	

	/** Get the select-state of this fig.
		   *
		   * @return <code>true</code> if this Fig was selected.
		   * @see mygale.mcge.fig.IFigPort#getActiveState()
	 */
	public boolean getActiveState() {
		return _active;
	}

	////
	// Painting methods

	/** Method to paint this port as an activated port.
		   *  You must use the given <var>figOwnerX</var> and
		   *  <var>figOwnerY</var> parameters to draw the port.
		   *  These coordinates correspond to the upper left
		   *  corner of the Fig owner  (usually a
		   *  {@link FigNode FigNode})
		   *  By defaut this method do nothing.
		   *
		   * @param g2d the graphic context.
		   * @param figOwnerX the x coordinate of the Fig owner.
		   * @param figOwnerY the y coordinate of the Fig owner.
		   * @param prefs Prefs
	 */
	private static final BasicStroke _activeHandleStroke = new BasicStroke(2f);
	public void paintActivatedPort(Graphics2D g2d, Prefs prefs) {
		Stroke oldStroke = g2d.getStroke();
		g2d.setStroke(_activeHandleStroke);
		g2d.setColor(getActivePortColor(prefs));

		g2d.draw(getShape());

		g2d.setStroke(oldStroke);
	}

	public abstract Shape getShape();

	/** Method to paint this port as an inactivated port.
		   *  You must use the given <var>figOwnerX</var> and
		   *  <var>figOwnerY</var> parameters to draw the port.
		   *  These coordinates correspond to the upper left
		   *  corner of the Fig owner  (usually a
		   *  {@link FigNode FigNode})
		   *  By defaut this method do nothing.
		   *
		   * @param g2d the graphic context.
		   * @param figOwnerX the x coordinate of the Fig owner.
		   * @param figOwnerY the y coordinate of the Fig owner.
		   * @param prefs Prefs
	 */

	/** Set the select-state of this fig.
		   *
		   * @param active boolean
	 * @see mygale.mcge.fig.IFigPort#setActiveState(boolean)
	 */
	public void setActiveState(boolean active) {
		if (active != _active) {
			_active = active;
			fireStateChanged();
		}
	}

	/** Return the color of active port handles.
		   *
		   * @return the color of active handles.
		   */
	protected Color getActivePortColor(Prefs prefs) {
		return prefs.getActivePortColor();
	}
}