package mygale.mcge.fig;

import java.awt.*;
import java.awt.geom.GeneralPath;
import java.util.Vector;

import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.util.MCGEGlobals;
import mygale.mcge.util.prefs.Prefs;

/** FigEdgePoly classe for representing edges between
 *  {@link FigPort ports}. It is draw with a spline.
 *  <p>
 *  A spline is a curved line that connects each edge's
 *  control points (stored in <var>_controlPoints</var>).
 *  <p><center>
 *    <img src="doc-files/FigEdgeSpline_uml.gif" width="93" height="83" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,0,92,28"  href="FigEdge.html">
 *    </map>
 *  </center></p>
 *
 */
//TODO knnte gelscht werden - wird in SeSAm nicht bentigt
public class FigEdgeSpline extends AbstractFigEdge {

	/** This is the minimal distance between an edge's control
	 *  point and a spline's control point.
	 */
	public static final int _CTRLPTS_DIST = 10;

	/** This is the control points used to managed the spline.
	 *  It additional control points, which are in both side
	 *  of the inherited control points.
	 */
	protected Vector _splineCtrlPoints = new Vector();

	/** Construct a new FigEdgeSpline.
	     */
	public FigEdgeSpline() {
		super();
	}

	////
	// Constructors

	/** Contruct a new FigEdgeSpline.
	 *
	 * @param xfrom horizontal position of the begin point for this FigEdge.
	 * @param yfrom vertical position of the begin point for this FigEdge.
	 * @param xto horizontal position of the end point for this FigEdge.
	 * @param yto vertical position of the end point for this FigEdge.
	 */
	public FigEdgeSpline(int xfrom, int yfrom, int xto, int yto) {
		super(xfrom, yfrom, xto, yto);
	}

	/** Calculate the bounds of the fig.
	     *  This method calls
	     *  {@link #calcSplineCtrlPoints() calcSplineCtrlPoints()}
	     *  and {@link #calcSplineBounds() calcSplineBounds()}.
	     */
	public void calcBounds() {
		calcSplineCtrlPoints();
		super.calcBounds();
		calcSplineBounds();
	}

	/** Calculate the bounds of this spline.
	     *  We assume that this method updates
	     *  <var>_rect</var>.
	     *  We assume too that all spline's control
	     *  points are calculated when this method
	     *  was called.
	     *
	     * @see #calcSplineCtrlPoints()
	     */
	public void calcSplineBounds() {
		GeneralPath splinePath = getSplinePath(0, 0);
		add(splinePath.getBounds());
	}

	////
	// Accessors

	/** return the max number of control point for the edge :
	 *  <code>Interger.MAX_VALUE</code> for a poly edge.
	 *
	 * @return always <code>Interger.MAX_VALUE</code>.
	 */
	public int getMaxAllowedCtrlPoints() {
		return Integer.MAX_VALUE;
	}

	/** Calculate one spline's control point from the
	     *  given edge's control points.
	     *  This method calculate the point, which are one
	     *  the line (<var>firstPoint</var>)-(<var>secondPoint</var>)
	     *  and are at <var>_CTRLPTS_DIST</var> points
	     *  from the <var>firstPoint</var>.
	     *
	     * @param firstPoint first edge's control point.
	     * @param secondPoint second edge's control point.
	     * @param a reference to the spline's control point.
	     */
	protected Point calcSplineCtrlPoint(Point firstPoint, Point secondPoint) {
		Point ctrlPoint = new Point(firstPoint.x, firstPoint.y);
		double vector_x = secondPoint.x - firstPoint.x;
		double vector_y = secondPoint.y - firstPoint.y;
		double norme = Math.sqrt(vector_x * vector_x + vector_y * vector_y);
		if (norme != 0.0) {
			ctrlPoint.x = firstPoint.x + (int) ((vector_x *  _CTRLPTS_DIST) / norme);
			ctrlPoint.y = firstPoint.y + (int) ((vector_y *  _CTRLPTS_DIST) / norme);
		}
		return ctrlPoint;
	}

	/** Calculate the spline's control points.
	     *  <p>
	     *  We assume that the edge's control points are set.
	     */
	protected void calcSplineCtrlPoints() {
		int count = _controlPoints.size();
		_splineCtrlPoints = new Vector();
		if (count > 1) {
			Point p1 = null;
			Point p2 = getCtrlPointAt(0);
			Point p3 = getCtrlPointAt(1);
			// Calculate the first spline's control point
			_splineCtrlPoints.addElement(calcSplineCtrlPoint(p2, p3));
			// Calcultate intermediary spline's control points
			for (int i = 1; i < (count - 1); i++) {
				p1 = p2;
				p2 = p3;
				p3 = getCtrlPointAt(i + 1);
				Point ctrlP1 = new Point();
				Point ctrlP2 = new Point();
				calcSplineCtrlPoints(p1, p2, p3, ctrlP1, ctrlP2);
				_splineCtrlPoints.addElement(ctrlP1);
				_splineCtrlPoints.addElement(ctrlP2);
			}
			// Calculate the last spline's control point
			_splineCtrlPoints.addElement(calcSplineCtrlPoint(p3, p2));
		}
	}

	/** Calculate one spline's control point from the
	     *  given edge's control points.
	     *  This method change the values of parameters
	     *  <var>firstResult</var>and
	     *  <var>secondResult</var>.
	     *
	     * @param firstPoint first edge's control point.
	     * @param secondPoint second edge's control point.
	     * @param thirdPoint third edge's control point.
	     * @param firstResult a reference to the first spline's control point.
	     * @param secondResult a reference to the second spline's control point.
	     */
	protected void calcSplineCtrlPoints(
		Point firstPoint,
		Point secondPoint,
		Point thirdPoint,
		Point firstResult,
		Point secondResult) {
		double alpha = MCGEGlobals.getAngleFromVector(secondPoint, firstPoint);
		double beta = MCGEGlobals.getAngleFromVector(secondPoint, thirdPoint);
		if ((alpha != Double.NaN) && (beta != Double.NaN)) {
			double diff = MCGEGlobals.normalizeAngle(alpha - beta);
			double gamma = MCGEGlobals.normalizeAngle((diff - Math.PI) / 2.0);
			if (diff < 180.0) {
				alpha = MCGEGlobals.normalizeAngle(alpha - gamma);
			} else {
				alpha = MCGEGlobals.normalizeAngle(alpha - gamma + Math.PI);
			}
			beta = MCGEGlobals.normalizeAngle(alpha + Math.PI);
			double firstDist = getCtrlPointDist(firstPoint, secondPoint, thirdPoint);
			double secondDist = getCtrlPointDist(thirdPoint, secondPoint, firstPoint);
			firstResult.x = secondPoint.x + (int) (Math.cos(alpha) * firstDist);
			firstResult.y = secondPoint.y + (int) (Math.sin(alpha) * firstDist);
			secondResult.x = secondPoint.x + (int) (Math.cos(beta) * secondDist);
			secondResult.y = secondPoint.y + (int) (Math.sin(beta) * secondDist);
		}
	}

	/** Compute the distance between a edge's control point and
	     *  a spline's control point.
	     *  If <var>delta1</var> is the distance between
	     *  <var>sourcePoint</var> and <var>centralPoint</var> and
	     *  <var>delta2</var> is the distance between
	     *  <var>centralPoint</var> and <var>targetPoint</var>, then
	     *  the distance between the <var>centralPoint</var> and
	     *  a spline's control point, which is between
	     *  <var>sourcePoint</var> and <var>centralPoint</var>;
	     *  is calculate with the folowing formula :
	     *  <p>
	     *  <code>dist = ( delta1 * delta1 ) / ( delta1 + delta2 )</code>
	     *
	     *  @param sourcePoint the first edge's control point.
	     *  @param centralPoint the second edge's control point.
	     *  @param targetPoint the third edged's control point.
	     *  @return the calculed distance, or <var>_CTRLPTS_DIST</var>
	     *          if the distance is less than <var>_CTRLPTS_DIST</var>.
	     */
	protected double getCtrlPointDist(Point sourcePoint, Point centralPoint, Point targetPoint) {
		double dist = 0.0;
		double delta1 = MCGEGlobals.distance(sourcePoint.x, sourcePoint.y, centralPoint.x, centralPoint.y);
		double delta2 = MCGEGlobals.distance(centralPoint.x, centralPoint.y, targetPoint.x, targetPoint.y);
		double denom = delta1 + delta2;
		if (denom != 0.0) {
			dist = (delta1 * delta1) / denom;
		}
		if (dist < _CTRLPTS_DIST)
			dist =  _CTRLPTS_DIST;
		return dist;
	}

	/** Reply a graphic path that correspond to the spline.
	     *  <p>
	     *  We assume that this method was called when this
	     *  spline have at least 2 control points.
	     *
	     * @param dx translation along the x axis.
	     * @param dy translation along the y axis.
	     * @return the spline Graph2D object.
	     */
	protected GeneralPath getSplinePath(int dx, int dy) {
		int i, j;
		GeneralPath curve = new GeneralPath();
		Point p1 = getCtrlPointAt(0);
		curve.moveTo(p1.x + dx, p1.y + dy);
		// Draw the spline
		for (i = 0, j = 1;(j < _controlPoints.size()) && (i < _splineCtrlPoints.size()); i += 2, j++) {
			Point p2 = (Point) _splineCtrlPoints.elementAt(i);
			Point p3 = (Point) _splineCtrlPoints.elementAt(i + 1);
			Point p4 = getCtrlPointAt(j);
			curve.curveTo(p2.x + dx, p2.y + dy, p3.x + dx, p3.y + dy, p4.x + dx, p4.y + dy);
		}
		return curve;
	}

	/** Reply if the given point composed the first segment of the spline.
	     *
	     * @param p a point that must be in <var>_controlPoints</var>.
	     * @return <code>true</code> if the given point if the second
	     *         point of the edge's path, otherwise <code>false</code>.
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     */
	protected boolean isFirstSegment(int p_x, int p_y) {
		if (_controlPoints.size() < 2) {
			throw new MCGEInternalError("not enought points in" + "_controlPoints.");
		}
		return (getCtrlPointAt(1).x == p_x) && (getCtrlPointAt(1).y == p_y);
	}

	/** Method to paint this Fig.
	 *
	 * @param g2d the graphic context.
	 * @param prefs Prefs
	 * @param dx int
	 * @param dy int
	 * @param shadowModeOn boolean
	 */
	public void paintContent(Graphics2D g2d, Prefs prefs, int dx, int dy, boolean shadowModeOn) {
		paintSpline(g2d, prefs, dx, dy);
		paintArrows(g2d, prefs, dx, dy, shadowModeOn);
	}

	/** This method paint a spline.
	     *  <p>
	     *  We assume that this method was called when this
	     *  spline have at least 4 control points.
	     *
	     * @param g2d the graphic context.
	     */
	protected void paintSpline(Graphics2D g2d, Prefs prefs, int dx, int dy) {
		g2d.setColor(prefs.getFigEdgeColor());
		// Draw the path
		g2d.draw(getSplinePath(dx, dy));
	}
}