package mygale.mcge.fig;

import java.awt.*;
import java.awt.geom.*;
import java.awt.geom.Ellipse2D.Double;

import mygale.mcge.util.MCGEGlobals;
import mygale.mcge.util.prefs.Prefs;

/** Class to represent a node in a diagram. It is paint as an ellipse.
 *  <p><center>
 *    <img src="doc-files/FigNodeEllipse_uml.gif" width="93" height="83" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="10,0,82,28"  href="FigNode.html">
 *    </map>
 *  </center></p>

 */
public class FigNodeEllipse extends AbstractFigNode {

	public FigNodeEllipse() {
		super();
	}

	/** Method to paint this Fig.
	 *
	 * @param g2d the graphic context.
	 */
	public void paintContent(Graphics2D g2, Prefs prefs) {

		Ellipse2D.Double ellipse = getEllipse();

		g2.setColor(getBackgroundColor(prefs));
		g2.fill(ellipse);

		g2.setColor(prefs.getFigForegroundColor());
		g2.draw(ellipse);
	}

	private Double getEllipse() {
		return new Ellipse2D.Double(getX(), getY(), getWidth(), getHeight());
	}

	/** Method to compute the intersection of a line drawed from
		 *  a given point with the active handle of the port.
		 *  This method compute the intersection from the line
		 *  (<var>pt</var>)-(center of this port) with the frame
		 *  of this port.
		 *  The frame of this port is "translated" by the
		 *  <var>dx</var> and <var>dy</var> quantities.
		 *
		 * @param pt the point that permits to define a line.
		 * @param dx translation step along the x axis.
		 * @param dy translation step along the x axis.
		 * @return the intersection point.
		 * @see mygale.mcge.fig.IFigPort#intersectWith(Point, int, int)
	 */
	public Point intersectWith(Point pt, int dx, int dy) {
		Rectangle bounds = (Rectangle) getBounds().clone();
		bounds.translate(dx, dy);

		Ellipse2D.Double ellipse = new Ellipse2D.Double(bounds.x, bounds.y, bounds.width, bounds.height);

		if (ellipse.contains(pt)) {
			return intersectWithInternalPoint(pt, bounds);
		} else {
			return intersectWithExternalPoint(pt, bounds);
		}
	}

	public Shape getShape() {
		return getEllipse();
	}

	/** Method to compute the intersection of a line drawed from
		 *  a given point with the active handle of the port.
		 *  <p>
		 *  Wee supopse that the given point is outside
		 *  the port bounds.
		 *
		 * @param pt the point that permits to define a line.
		 * @param absRect the absolute bounds of this port.
		 * @return the intersection point.
		 */
	private Point intersectWithExternalPoint(Point pt, Rectangle absRect) {
		Point center = new Point(absRect.x + (absRect.width / 2), absRect.y + (absRect.height / 2));

		double angle = getAngle(center.x, center.y, pt.x, pt.y);

		Arc2D.Double arc =
			new Arc2D.Double(absRect.x, absRect.y, absRect.width, absRect.height, -angle, 1, Arc2D.OPEN);

		Point2D point = arc.getStartPoint();
		return new Point((int) point.getX(), (int) point.getY());
	}

	/** Method to compute the intersection of a line drawed from
		 *  a given point with the active handle of the port.
		 *  <p>
		 *  Wee supopse that the given point is inside
		 *  the port bounds. Then this method returns
		 *  the center point of this port's bounds.
		 *
		 * @param pt the point that permits to define a line.
		 * @param absRect the absolute bounds of this port.
		 * @return the intersection point.
		 */
	private Point intersectWithInternalPoint(Point pt, Rectangle absRect) {
		return new Point(absRect.x + (absRect.width / 2), absRect.y + (absRect.height / 2));
	}

	/** Reply the angle between the
		 *  (<var>x1</var>,<var>y1</var>)-(+INF,<var>y1</var>)
		 *  and (<var>x1</var>,<var>y1</var>)-(<var>x2</var>,<var>y2</var>).
		 *  The resulting angle is in degrees.
		 *
		 * @param x1 x coordinate of the "origin".
		 * @param y1 y coordinate of the "origin".
		 * @param x2 x coordinate of a point.
		 * @param y2 y coordinate of a point.
		 * @return an angle in degrees.
		 */
	private double getAngle(int x1, int y1, int x2, int y2) {
		double angle = MCGEGlobals.getAngleFromVector(x1, y1, x2, y2);
		return Math.toDegrees(angle);
	}
}