package mygale.mcge.fig.arrow;


import java.awt.*;
import java.awt.geom.AffineTransform;

import junit.framework.Assert;

import mygale.mcge.fig.IFigEdge;
import mygale.mcge.util.MCGEGlobals;
import mygale.mcge.util.prefs.Prefs;

/** This class is the base class for the arrows that will
 *  be paint in the ends of edges.
 *  <p><center>
 *    <img src="doc-files/package_uml.gif" width="286" height="88" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="95,0,167,27"  href="Arrow.html">
 *      <area shape="Rect" coords="212,0,285,27"  href="../FigEdge.html">
 *      <area shape="Rect" coords="0,59,99,86"  href="ArrowSimpleArrow.html">
 *      <area shape="Rect" coords="207,59,279,86"  href="ArrowCircle.html">
 *      <area shape="Rect" coords="117,59,189,86"  href="ArrowDiamond.html">
 *    </map>
 *  </center></p>
 *
 */
public abstract class Arrow {

   /** The instance of the
	*  {@link mygale.mcge.fig.Fig Fig}
	*	that contains this arrow.
	*/
   protected IFigEdge _edge = null;

   /** The bounds.
	*/
   protected Rectangle _rect = new Rectangle(0, 0, 0, 0);

/** Construct a new Arrow.
     */
public Arrow() {
}




/** Construct a new Arrow.
 *
 * @param edge the edge that contains this arrow.
 */
public Arrow(IFigEdge edge) {
   setEdge(edge);
}


/**
 * @param x_begin int
 * @param y_begin int
 * @param x_end int
 * @param y_end int
 * @return double
 */
public static double getAngle(int x_begin, int y_begin, int x_end, int y_end) {
   double angle = 0;

   if ((x_begin != x_end) || (y_begin != y_end)) {
      angle = MCGEGlobals.getAngleFromVector(x_end, y_end, x_begin, y_begin);
      if (Assert.ASSERTION_ON) Assert.assertTrue(!Double.isNaN(angle));
   }

   return angle;
}


/** Set the edge that contains this arrow.
     *
     * @return the edge that contains this arrow.
     */
public IFigEdge getEdge() {
   return _edge;
}


/** Reply the color of filling area.
     *  By default this method take the color
     *  from the preference manager.
     *
     * @return the color of the arrow.
     */
protected Color getFillColor(Prefs prefs) {
   return prefs.getFigBackgroundColor();
}


/** Reply the height of the arrow, before the rotation.
     *
     * @return an integer that is the height of this arrow.
     */
public abstract int getHeight();


/** Reply the color of shape.
     *  By default this method take the color
     *  from the preference manager.
     *
     * @return the color of the arrow.
     */
protected Color getShapeColor(Prefs prefs) {
   return prefs.getFigForegroundColor();
}


/** Reply the width of the arrow, before the rotation.
     *
     * @return an integer that is the width of this arrow.
     */
public abstract int getWidth();




/** Paint the arrow.
 *
 * @param g the graphical context in which
 *          the arrow must be paint.
 */
protected void paint(Graphics2D g2d, Prefs prefs, double angle, int x, int y) {
   if (_edge != null) {
      AffineTransform at = new AffineTransform(g2d.getTransform());
      g2d.rotate(angle, x, y);
      paintContent(g2d, x, y, prefs);
      g2d.setTransform(at);
   }
}


/**
 * @param g2d Graphics2D
 * @param prefs Prefs
 * @param x_begin int
 * @param y_begin int
 * @param x_end int
 * @param y_end int
 */
public void paint(Graphics2D g2d, Prefs prefs, int x_begin, int y_begin, int x_end, int y_end) {
   paint(g2d, prefs, getAngle(x_begin, y_begin, x_end, y_end), x_end, y_end);
}


/** Paint the arrow.
     *  the point (<var>x</var>,<var>y</var>) is the point where
     *  the arrow will be connect to a port.
     *  You must paint the arrow on the right of this
     *  point to be outside the node
     *  and to the left to be inside the node.
     *
     * @param g the graphical context in which
     *          the arrow must be paint.
     * @param x x coordinate of the arrow.
     * @param y y coordinate of the arrow.
     */
protected abstract void paintContent(Graphics2D g, int x, int y, Prefs prefs);




/** Set the edge that contains this arrow.
 *
 * @param edge the edge that contains this arrow.
 */
public void setEdge(IFigEdge edge) {
   _edge = edge;
}
}