package mygale.mcge.layer;


import java.awt.Graphics2D;

import mygale.mcge.core.JGraph;
import mygale.mcge.util.prefs.Prefs;

/** A Layer is like a drawing layer in high-end drawing applications
 *  (e.g2d., MacDraw Pro).  A Layer is like a sheet of clear plastic that
 *  can contain part of the picture being drawn and multiple layers
 *  are put on top of each other to make the overall
 *  picture.  Different layers can be hidden independently.
 *  In mcge the Layer class is more abstract than
 *  described above.  Generalizing the concept of a
 *  layer to handle grids and other computed display features gives
 *  more power and allows the framework to be extended in building
 *  various applications. For example an application that needs polar
 *  coordinates might use LayerPolar, and an application that used a
 *  world map might implement LayerMap. But since layers can be
 *  composed, the user could put a grid in front of or behind the
 *  map.
 *  <p><center>
 *    <img src="doc-files/Layer_uml.gif" width="299" height="151" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,64,72,91"  href="LayerManager.html">
 *      <area shape="Rect" coords="117,0,189,27"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="225,64,297,91"  href="../util/prefs/Prefs.html">
 *      <area shape="Rect" coords="54,122,126,149"  href="LayerGrid.html">
 *      <area shape="Rect" coords="176,122,249,149"  href="LayerDiagram.html">
 *    </map>
 *  </center></p>

 */
public abstract class Layer {

   /** Does the user not want to see this Layer right new ? */
   protected boolean _hidden = false;

   /** This is the {@link mygale.mcge.core.JGraph JGraphs},
    *  which displays this Layer
    */
   protected JGraph _graph = null;


   /** Construct a new Layer. This abstract class really does nothing
    *  in its constructor, but subclasses may have meaningful
    *  constructors.
    *
    * @param graph the graph that displays this Layer.
    */
   public Layer(JGraph graph) {
      _graph = graph;
   }

   /** Get editor which display this Layer.
        *
       @link mygale.mcge.core.JGraph JGraphs},
        *         which display this Layer.
        * @return A { * @see mygale.mcge.core.JGraph
        */
   public JGraph getGraph() {
      return _graph;
   }

   /** Set editor which display this Layer.
        *
       @link mygale.mcge.core.JGraph JGraphs},
        *              which display this Layer.
        * @param graph A { * @see JGraph
        */
   public void setGraph(JGraph graph) {
      _graph = graph;
   }

   /** Get the hidden flag.
        *
        * @return <code>true</code> if this Layer must not be
        *         displayed, <code>false</code> otherwise.
        */
   public boolean getHidden() {
      return _hidden;
   }



   /** Set the hidden flag.
    *
    * @param hidden <code>true</code> if this Layer must
    *        not be displayed, <code>false</code> otherwise.
    */
   public void setHidden(boolean hidden) {
      _hidden = hidden;
   }



   /** This method return the preference manager.
    *
    * @return the current preference manager.
    */
   public Prefs getPrefs() {
      return _graph.getPrefs();
   }


   /** Paint this Layer on the given Graphics. Subclasses should
    *  overide
    *  {@link mygale.mcge.layer.Layer#paintContents(Graphics) paintContents()},
    *  which is called from here if the Layer is not hidden.
    *
   @link java.awt.Graphics graphic context} in which
    *          the Layer must be drawed.
    * @param g2d { * @see java.awt.Graphics
    */
   public final void paint(Graphics2D g2d) {
      if (!_hidden) {
         paintContents(g2d);
      }
   }

   /** Method to paint the contents of this Layer.
        *
       @link java.awt.Graphics graphic context} in which
        *          the Layer must be drawed.
        * @param g2d { * @see java.awt.Graphics
        */
   public abstract void paintContents(Graphics2D g2d);

   /** Ask all {@link mygale.mcge.core.JGraph JGraph} to completly
        *  redraw their display.
        */
   public void refreshGraph() {
      _graph.damagedAll();
   }
}