package mygale.mcge.layer;


import java.awt.Point;
import java.util.*;

import mygale.mcge.core.JGraph;
import mygale.mcge.fig.IFigLayered;
import mygale.mcge.util.prefs.Prefs;

/** This interface represents the layers that contain
 *  {@link mygale.mcge.fig.FigLayered FigLayered objets}.
 *  <p><center>
 *    <img src="doc-files/LayerContener_uml.gif" width="308" height="88" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,0,72,27"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="234,0,306,27"  href="../fig/FigLayered.html">
 *      <area shape="Rect" coords="117,58,189,87"  href="LayerDiagram.html">
 *    </map>
 *  </center></p>
 *
 * @see mygale.mcge.layer.LayerDiagram
 */
public interface LayerContainer {

   /** Add a {@link mygale.mcge.fig.FigLayered FigLayered} to this Layer.
        *
        * @param figure the new {@link mygale.mcge.fig.FigLayered FigLayered}
        *               to add to this Layer.
        */
   void add(IFigLayered figure);

   /** Move a {@link mygale.mcge.fig.FigLayered FigLayered} by one step to the front of
        *  this Layer.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}, which must be
        *               in the forward.
        */
   void bringForward(IFigLayered figure);

   /** Put a list of {@link mygale.mcge.fig.FigLayered Figs} in the front of this
        *  Layer. Figs will appear before all other Figs (but in the same order).
        *
        * @param figures the {@link mygale.mcge.fig.FigLayered Figs}, which must be in
        *                the front. */
   void bringToFront(Iterator figures);

   /** Put a {@link mygale.mcge.fig.FigLayered} in the front of this Layer.
        *  FigLayered will appears before all other Figs.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}, which must
        *               be in the front.
        */
   void bringToFront(IFigLayered figure);

   /** This method inform {@link mygale.mcge.core.JGraph JGraphs} that a
        *  {@link mygale.mcge.fig.FigLayered FigLayered} area must be refresh.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered} that it must be
        *               refresh.
        */
   void damaged(IFigLayered figure);

   /** A {@link mygale.mcge.fig.FigLayered FigLayered} in this Layer has been deleted.
        *  Notify all Editors so that they can deselect the FigLayered.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}, which is deleted.
        */
   void deleted(IFigLayered figure);

   /** Get {@link mygale.mcge.fig.FigLayered Figs} of this Layer.
        *
        * @return The list of elements in this Layer.
        */
   Iterator getAllFigs();

   /** Get Layer content.
        *
        * @return The content of this Layer. */
   Vector getContents();

   /** Reply the depth level of the given {@link mygale.mcge.fig.FigLayered FigLayered}.
        *
        * @param fig a reference to the figure.
        * @return the depth level (0 for the Front, -1 for "not found" )
        */
   int getDepthLevel(IFigLayered fig);

   /** Get editor which display this Layer.
        *
        * @return A {@link mygale.mcge.core.JGraph JGraphs}, which
        *         display this Layer.
        * @see mygale.mcge.core.JGraph
        */
   JGraph getGraph();

   /** This method return the preference manager.
        *
        * @return the current preference manager.
        */
   Prefs getPrefs();

   /** Return the {@link mygale.mcge.fig.FigLayered FigLayered} under the mouse.
        *
        * @param r {@link java.awt.Rectangle Rectangle} that contains a
        *          {@link mygale.mcge.fig.FigLayered FigLayered}.
        * @return {@link mygale.mcge.fig.FigLayered FigLayered} in <var>r</var> or
        *         <code>null<code> if not found.
        */
   IFigLayered hit(Point r);

   /** Remove a {@link mygale.mcge.fig.FigLayered FigLayered} from this Layer.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered} that may
        *               be removed from this Layer.
        */
   void remove(IFigLayered figure);

   /** Remove all {@link mygale.mcge.fig.FigLayered FigLayered} from this Layer. */
   void removeAll();

   /** Move a {@link mygale.mcge.fig.FigLayered FigLayered} by one step to the back of
        *  this Layer.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}, which must be
        *               move backward.
        */
   void sendBackward(IFigLayered figure);

   /** Put a list of {@link mygale.mcge.fig.FigLayered Figs} to the
        *  back of this Layer. Figs will appear backward all other
        *  Figs (but in the same order).
        *
        * @param figures the {@link mygale.mcge.fig.FigLayered Figs}, which must
        *                be in the back.
        */
   void sendToBack(Iterator figures);

   /** Put a {@link mygale.mcge.fig.FigLayered FigLayered} to the back of this Layer.
        *  FigLayered will appears backward all other Figs.
        *
        * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}, which must be
        *               n the back.
        */
   void sendToBack(IFigLayered figure);

   /** Change the depth level of this FigLayered.
        *  This level can be changed by move to the
        *  back or to the front the figure.
        *
        * @param fig a reference to the figure.
        * @param depth the depth level (0 for the Front, 1 ... )
        */
   void setDepthLevel(IFigLayered fig, int depth);
}