package mygale.mcge.layer;


import java.awt.*;

import mygale.mcge.core.JGraph;

/** Paint a background drawing guide consisting of horizontal and
 *  vertical lines in a neutral color. This feature is common to
 *  many drawing applications (e.g2d., xfig). LayerGrid is in concept
 *  a {@link mygale.mcge.layer.Layer Layer}, just like any other so
 *  it can be composed, hidden, and reordered.
 *  <p><center>
 *    <img src="doc-files/LayerGrid_uml.gif" width="187" height="84" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="113,0,185,27"  href="Layer.html">
 *      <area shape="Rect" coords="0,56,73,83"  href="../core/JGraph.html">
 *    </map>
 *  </center></p>
 *
 * @see mygale.mcge.layer.Layer
 */
public class LayerGrid extends Layer {

   /** spacing between the lines.
    */
   protected int _spacing = 16;

   /** <code>true</code> means paint grid lines, <code>false</code> means
    *  paint only dots where the lines would intersect. Painting dots is
    *  about as useful as painting lines and it looks less clustered.
    * But lines are more familiar to some people.
    */
   protected boolean _paintlines = false;

   /** The size of the dots. Dots are actually small rectangles.
    */
   protected int _dotsize = 2;

   /** <code>true</code> means to fill in the image stamp or drawing area with
    *  the background color. <code>false</code> meas to just paint the lines or dots.
    */
   protected boolean _paintbackground = true;

   /** The image stamp is used to paint the grid quickly.
    *  Basically I make a pretty big off screen image and paint my grid on it,
    *  then I repeatly bitblt that image to the screen (or some other off
    *  screen image), More pixels are being drawn this way, but since
    *  bitblt is a fast operation it works much faster.
    */
   private transient Image _stamp = null;

   /** The size of the iamge stamp.
    */
   private Dimension _stampsize = new Dimension(128, 128);

   /** As an example of different grid styles 5 are defined.
    *
    * @see #adjust()
    */
   private int _style = 2;
   private final int NUM_STYLES = 5;

   ////
   // Constructors

   /** Construct a new LayerGrid
    *
    * @param graph this is the editor.
    */
   public LayerGrid(JGraph graph) {
      super(graph);
   }

   /** Construct a new LayerGrid
        *
        * @param graph this is the editor.
        * @param spacing space between the lines.
        * @param lines <code>true</code> if lines must be drawed,
        *              <code>false</code> otherwise.
        */
   public LayerGrid(JGraph graph, int spacing, boolean lines) {
      super(graph);
      _spacing = spacing;
      _paintlines = lines;
   }

   
   //user interface

   /** Eventually this will open a dialog box to let the user adjust
    *  the grid line spacing, colors, and whether links or dots are
    *  shown, For now it just cycles among 5 predefined styles.
    */
   public void adjust() {
      _style = (_style + 1) % NUM_STYLES;
      _stamp = null;
      switch (_style) {
         case 0 :
            setHidden(false);
            _paintlines = true;
            _spacing = 16;
            break;
         case 1 :
            setHidden(false);
            _paintlines = true;
            _spacing = 8;
            break;
         case 2 :
            setHidden(false);
            _paintlines = false;
            _spacing = 16;
            break;
         case 3 :
            setHidden(false);
            _paintlines = false;
            _spacing = 32;
            break;
         default :
            setHidden(true);
      }
      refreshGraph();
   }

   ////
   // Painting methods

   /** Method to paint the contents of this Layer.
    *
   @link java.awt.Graphics graphic context} in which
    *          the Layer must be drawed.
    * @param g2d { */
   public void paintContents(Graphics2D g2d) {
      // Create the double-buffer image
      if (_stamp == null) {
         if (_spacing > _stampsize.height)
            _stampsize.height = _stampsize.width = _spacing;
         if (_graph == null)
            return;
         _stamp = _graph.createImage(_stampsize.width, _stampsize.height);
         if (_stamp == null)
            return;
      }

      // Paint the grid into the buffer
      if (_paintlines)
         paintLines();
      else
         paintDots();

      // Put the buffer into the graphic context
      Rectangle clip = g2d.getClipBounds();
      int bot = clip.y + clip.height;
      int right = clip.x + clip.width;
      int x, y;
      if (clip.x >= 0)
         x = clip.x / _spacing * _spacing;
      else
         x = - (-clip.x / _spacing + 1) * _spacing;
      while (x <= right) {
         if (clip.y >= 0)
            y = clip.y / _spacing * _spacing;
         else
            y = - (-clip.y / _spacing + 1) * _spacing;
         while (y <= bot) {
            g2d.drawImage(_stamp, x, y, null);
            y += _stampsize.height;
         }
         x += _stampsize.width;
      }
   }

/** Print dots on the given stamp image.
     */
private void paintDots() {
   Graphics2D g2d = (Graphics2D) _stamp.getGraphics();
   g2d.clipRect(0, 0, _stamp.getWidth(null), _stamp.getHeight(null));
   Rectangle clip = g2d.getClipBounds();

   // Fill background if necessary
   if (_paintbackground) {
      g2d.setColor(getPrefs().getBackgroundColor());
      g2d.fillRect(clip.x, clip.y, clip.width, clip.height);
   }

   int x = clip.x / _spacing * _spacing - _spacing;
   int y = clip.y / _spacing * _spacing - _spacing;
   int bot = clip.y + clip.height;
   int right = clip.x + clip.width;

   // Draw dots
   g2d.setColor(getPrefs().getForegroundColor());
   while (x <= right) {
      y = 0;
      while (y <= bot) {
         g2d.fillRect(x, y, _dotsize, _dotsize);
         y += _spacing;
      }
      x += _spacing;
   }
}

/** Print lines on the given stamp image.
     */
private void paintLines() {
   Graphics2D g2d = (Graphics2D) _stamp.getGraphics();
   g2d.clipRect(0, 0, _stamp.getWidth(null), _stamp.getHeight(null));
   Rectangle clip = g2d.getClipBounds();

   // Fill the background if necessary
   if (_paintbackground) {
      g2d.setColor(getPrefs().getBackgroundColor());
      g2d.fillRect(clip.x, clip.y, clip.width, clip.height);
   }

   int x = clip.x / _spacing * _spacing - _spacing;
   int y = clip.y / _spacing * _spacing - _spacing;
   int stepsX = clip.width / _spacing + 2;
   int stepsY = clip.height / _spacing + 2;
   int bot = clip.y + clip.height;
   int right = clip.x + clip.width;

   // Draw lines
   g2d.setColor(getPrefs().getForegroundColor());
   while (stepsX > 0) {
      g2d.drawLine(x, 0, x, bot);
      x += _spacing;
      stepsX--;
   }
   while (stepsY > 0) {
      g2d.drawLine(0, y, right, 0);
      y += _spacing;
      stepsY--;
   }
}
}