package mygale.mcge.layer;

import java.awt.*;
import java.util.*;

import mygale.mcge.core.JGraph;
import mygale.mcge.fig.IFigLayered;

/**
 * This class implements a king of {@link mygale.mcge.layer.Layer Layer}that contains other Layers.
 * <p>
 * <center><img src="doc-files/LayerManager_uml.gif" width="342" height="128" border="0" usemap="#umlclick">
 * <map name="umlclick"><area shape="Rect" coords="122,0,194,27"  href="../core/JGraph.html">
 * <area shape="Rect" coords="0,68,72,95"  href="../fig/FigLayered.html">
 * <area shape="Rect" coords="239,68,311,95"  href="Layer.html"></map></center>
 * </p>
 *  
 */
public class LayerManager {

    /**
     * The layers contained within this LayerManager.
     */
    private final ArrayList _layers = new ArrayList(2);

    /**
     * In Most editors one {@link mygale.mcge.layer.Layer Layer}is the active layer and all mouse clicks go
     * to the contents of that layer.
     */
    private Layer _activelayer;

    /**
     * This is the {@link mygale.mcge.core.JGraph graph editor}instance.
     */
    private JGraph _graph = null;

    /**
     * Construct a new Layer. This abstract class really does nothing in its constructor, but subclasses may
     * have meaningful constructors.
     * 
     * @link mygale.mcge.core.JGraph JGraph} instance that contains this layer manager.
     * 
     * @param editor {
     */
    public LayerManager(JGraph editor) {
        _graph = editor;
    }

    /**
     * Add a {@link mygale.mcge.fig.FigLayered FigLayered}to this Layer.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered} to add to this Layer.
     * 
     * @param figure
     *            the new {
     */
    public void add(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).add(figure);
        }
    }

    /**
     * Add a sublayer to this layer.
     * 
     * @link mygale.mcge.layer.Layer Layer} that must be add into this layer.
     * 
     * @param lay
     *            the new {
     */
    public void addLayer(Layer lay) {
        if (_layers.contains(lay))
            return;

        _layers.add(lay);

        lay.setGraph(_graph);
        setActiveLayer(lay);
    }

    /**
     * Move a {@link mygale.mcge.fig.FigLayered FigLayered}by one step to the front of this Layer.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered}, which must be in the forward.
     * 
     * @param figure
     *            the {
     */
    public void bringForward(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).bringForward(figure);
        }
    }

    /**
     * Put a list of {@link mygale.mcge.fig.FigLayered Figs}in the front of this Layer. Figs will appear
     * before all other Figs (but in the same order).
     * 
     * @link mygale.mcge.fig.FigLayered Figs}, which must be in the front. *
     * 
     * @param figures
     *            the {
     */
    public void bringToFront(Iterator figures) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).bringToFront(figures);
        }
    }

    /**
     * Put a {@link mygale.mcge.fig.FigLayered}in the front of this Layer. FigLayered will appears before
     * all other Figs.
     * 
     * @param figure
     *            the {@link mygale.mcge.fig.FigLayered FigLayered}, which must be in the front.
     */
    public void bringToFront(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).bringToFront(figure);
        }
    }

    /**
     * This method inform {@link mygale.mcge.core.JGraph JGraphs}that a
     * {@link mygale.mcge.fig.FigLayered FigLayered}area must be refresh.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered} that it must be refresh.
     * 
     * @param figure
     *            the {
     */
    public void damaged(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).damaged(figure);
        }
    }

    /**
     * A {@link mygale.mcge.fig.FigLayered FigLayered}in this Layer has been deleted. Notify all Editors so
     * that they can deselect the FigLayered.
     * 
     * @param figure
     *            the {@link mygale.mcge.fig.FigLayered FigLayered}, which is deleted.
     */
    public void deleted(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).deleted(figure);
        }
    }

    /**
     * Reply which layer is the active one.
     * 
     * @link mygale.mcge.layer.Layer Layer} in this layer.
     * 
     * @return the active {
     */
    public Layer getActiveLayer() {
        return _activelayer;
    }

    /**
     * Get {@link mygale.mcge.fig.FigLayered Figs}of this Layer.
     * 
     * @return The list of elements in this Layer.
     */
    public Iterator getAllFigs() {
        return (_activelayer != null && _activelayer instanceof LayerContainer)
            ? ((LayerContainer) _activelayer).getAllFigs()
            : null;
    }

    /**
     * When an editor or some tool wants to look at all the Figs thar are contained in this layer, reply the
     * contents of my active layer. Maybe this should really reply all the contents of all layers.
     * 
     * @link mygale.mcge.layer.Layer Layer} or <code>null</code> if no active layer.
     * 
     * @return the content of the active {
     */
    public Vector getContents() {
        return (_activelayer != null && _activelayer instanceof LayerContainer)
            ? ((LayerContainer) _activelayer).getContents()
            : null;
    }

    /**
     * Reply the depth level of the given {@link mygale.mcge.fig.FigLayered FigLayered}.
     * 
     * @param fig
     *            a reference to the figure.
     * @return the depth level (0 for the Front, -1 for "not found" )
     */
    public int getDepthLevel(IFigLayered fig) {
        if (_activelayer != null && _activelayer instanceof LayerContainer)
            return ((LayerContainer) _activelayer).getDepthLevel(fig);
        else
            return -1;
    }

    /**
     * Get the {@link mygale.mcge.core.JGraph JGraph}editor for this LayerManager.
     * 
     * @link mygale.mcge.core.JGraph JGraph} instance attached to this LayerManager.
     * 
     * @return the {
     */
    public JGraph getGraph() {
        return _graph;
    }

    /**
     * Return the {@link mygale.mcge.fig.FigLayered FigLayered}under the mouse.
     * 
     * @link java.awt.Rectangle Rectangle} that contains a {@link mygale.mcge.fig.FigLayered FigLayered}.
     * @link mygale.mcge.fig.FigLayered FigLayered} in <var>r</var> or <code>null<code> if not found.
         * @param p Point
     * @return {
     */
    public IFigLayered hit(Point p) {
        return (_activelayer != null && _activelayer instanceof LayerContainer)
            ? ((LayerContainer) _activelayer).hit(p)
            : null;
    }

    /**
     * Paint the contents of the LayerManager.
     * 
     * @link java.awt.Graphics graphic context} where layers must be drawn.
     * 
     * @param g2d {
     */
    public void paint(Graphics2D g2d) {
        Iterator it = _layers.iterator();

        while (it.hasNext()) {
            ((Layer) it.next()).paint(g2d);
        }
    }

    /**
     * Remove a {@link mygale.mcge.fig.FigLayered FigLayered}from this Layer.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered} that may be removed from this Layer.
     * 
     * @param figure
     *            the {
     */
    public void remove(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).remove(figure);
        }
    }

    /**
     * Remove all {@link mygale.mcge.fig.FigLayered FigLayered}from this Layer.
     */
    public void removeAll() {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).removeAll();
        }
    }

    /**
     * Remove all sublayers from this layer.
     */
    public void removeAllLayer() {
        _layers.clear();
        _activelayer = null;
    }

    /**
     * Remove a layer from this layer.
     * 
     * @link mygale.mcge.layer.Layer Layer} that must be removed from this layer.
     * 
     * @param lay
     *            the {
     */
    public void removeLayer(Layer lay) {
        if (!_layers.contains(lay))
            return;
        _layers.remove(lay);
        if (_activelayer == lay) {
            if (_layers.size() >= 1)
                _activelayer = (Layer) _layers.get(0);
            else
                _activelayer = null;
        }
    }

    /**
     * Move a {@link mygale.mcge.fig.FigLayered FigLayered}by one step to the back of this Layer.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered}, which must be move backward.
     * 
     * @param figure
     *            the {
     */
    public void sendBackward(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).sendBackward(figure);
        }
    }

    /**
     * Put a list of {@link mygale.mcge.fig.FigLayered Figs}to the back of this Layer. Figs will appear
     * backward all other Figs (but in the same order).
     * 
     * @param figures
     *            the {@link mygale.mcge.fig.FigLayered Figs}, which must be in the back.
     */
    public void sendToBack(Iterator figures) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).sendToBack(figures);
        }
    }

    /**
     * Put a {@link mygale.mcge.fig.FigLayered FigLayered}to the back of this Layer. FigLayered will appears
     * backward all other Figs.
     * 
     * @link mygale.mcge.fig.FigLayered FigLayered}, which must be n the back.
     * 
     * @param figure
     *            the {
     */
    public void sendToBack(IFigLayered figure) {
        if (_activelayer != null && _activelayer instanceof LayerContainer) {
            ((LayerContainer) _activelayer).sendToBack(figure);
        }
    }

    /**
     * Change the active sub layer.
     * 
     * @param lay
     *            the new active {@link mygale.mcge.layer.Layer Layer}.
     */
    public void setActiveLayer(Layer lay) {
        if (_layers.contains(lay))
            _activelayer = lay;
        else
            System.err.println("mygale.mcge.layer.LayerManager : That layer is not one of my layers");
    }

    /**
     * Change the depth level of this FigLayered. This level can be changed by move to the back or to the
     * front the figure.
     * 
     * @param fig
     *            a reference to the figure.
     * @param depth
     *            the depth level (0 for the Front, 1 ... )
     */
    public void setDepthLevel(IFigLayered fig, int depth) {
        if (_activelayer != null && _activelayer instanceof LayerContainer)
             ((LayerContainer) _activelayer).setDepthLevel(fig, depth);
    }

    /**
     * Set the {@link mygale.mcge.core.JGraph JGraph}editor for this LayerManager.
     * 
     * @link mygale.mcge.core.JGraph JGraph} instance attached to this LayerManager.
     * 
     * @param editor
     *            the {
     */
    public void setGraph(JGraph editor) {
        _graph = editor;

        Iterator it = _layers.iterator();

        while (it.hasNext()) {
            ((Layer) it.next()).setGraph(_graph);
        }
    }
}