package mygale.mcge.mode;

import java.awt.Cursor;
import java.awt.Graphics2D;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.event.mouse.MCGEMouseEventListener;
import mygale.mcge.util.prefs.Prefs;

/** This is the abstract superclass of all editor modes.  A Mode is
 *  responsible for handling most of the events that come to the
 *  Editor.  A Mode defines a context for interperting those events.
 *  For example, a mouse drag in {@link ModeSelect ModeSelect}
 *  will define a selection rectangle, while a mouse drag in
 *  {@link ModeCreateEdge ModeCreateEdge} will drag out a
 *  rubberband line.  Placing the logic for most event handing in
 *  Modes is key to keeping the Editor source code small and
 *  manageable, and also key to allowing addition of new kinds of user
 *  interactions without always modifying Editor and having to
 *  integrate ones modifications with other contributors
 *  modifications.  Modes should interpert user input and ask the
 *  Editor to execute {@link mygale.mcge.cmd.Cmd Cmds}.
 *  Placing the logic to manipulate the document into Cmds
 *  helps keep Mode's small and promotes sharing of Cmd code.
 *  <p><center>
 *    <img src="doc-files/Mode_uml.gif" width="470" height="358" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="252,275,351,302"  href="ModeCreateEdge.html">
 *      <area shape="Rect" coords="0,275,99,302"  href="ModeCreateFig.html">
 *      <area shape="Rect" coords="180,162,279,189"  href="ModeEditProperties.html">
 *      <area shape="Rect" coords="167,329,302,357"  href="ModeModifiyEdgeCtrlPoint.html">
 *      <area shape="Rect" coords="117,275,216,302"  href="ModeMove.html">
 *      <area shape="Rect" coords="14,162,140,189"  href="ModeScrollDrawingArea.html">
 *      <area shape="Rect" coords="243,162,392,189"  href="ModeSelect.html">
 *      <area shape="Rect" coords="45,216,144,243"  href="ModePopup.html">
 *      <area shape="Rect" coords="369,275,468,302"  href="ModeReshape.html">
 *      <area shape="Rect" coords="185,216,284,243"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="117,0,260,28"  href="../event/mouse/MCGEMouseEventListener.html">
 *      <area shape="Rect" coords="248,104,347,131"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="14,104,86,131"  href="ModeManager.html">
 *      <area shape="Rect" coords="104,36,176,63"  href="http://java.sun.com/products/jdk/1.2/docs/api/java/awt/event/KeyListener.html">
 *    </map>
 *  </center></p>

 */
public abstract class Mode implements KeyListener, MCGEMouseEventListener {

	/** The {@link ModeManager ModeManger} that contains
	 *  this Mode.
	 */
	protected ModeManager _modeManager;

	/** Construct a new Mode.
	     */
	public Mode() {
		_modeManager = null;
	}

	/** Construct a new Mode.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public Mode(ModeManager modeManager) {
		_modeManager = modeManager;
	}

	/** Reply the parent ModeManager of this Mode.
	     * @return a reference to the ModeManager that
	     *         contains this Mode.
	     */
	public ModeManager getModeManager() {
		return _modeManager;
	}

	/** Set the parent ModeManager of this Mode.
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public void setModeManager(ModeManager modeManager) {
		_modeManager = modeManager;
		setCursor(getInitialCursor());
	}

	/** Some Mode's should never be exited, but by default any Mode can
	     *  exit. Mode's which return <code>false</code> for canExit()
	     *  will not be popped from a {@link ModeManager ModeManager}.
	     *
	     * @return <code>true</code> if this mode can be stopped and popped
	     *         from the ModeManager's stack.
	     */
	public boolean canExit() {
		return true;
	}

	/** Ask to this mode to return to a stable state.
	     *  This method was called each time this Mode
	     *  need to be in a stable state e.g., initial
	     *  mode state.
	     */
	public void cleanMode() {
	}

	/** Clean and pop this mode from the
	 *  {@link ModeManager ModeManager} stack.
	 *
	 * @see #setExclusive(boolean)
	 * @see ModeManager#endMode()
	 * @see #cleanMode()
	 */
	public void done() {
		setCursor(Cursor.getDefaultCursor());
		setExclusive(false);
		cleanMode();
		_modeManager.endMode();
	}

	/** Returns the cursor that should be shown when this Mode starts.
	 *
	 * @return a reference to a cursor.
	 */
	public Cursor getInitialCursor() {
		return Cursor.getDefaultCursor();
	}

	/** Get the preferences manager.
	 *
	 * @return a reference to the preference manager.
	 */
	public Prefs getPrefs() {
		return _modeManager.getPrefs();
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		return "Type <escape> to cancel this mode";
	}

	/** Reply is this mode is exclusive.
	     *
	    @link ModeManager ModeManager},
	     *         otherwise <code>false</code>.
	     * @return <code>true</code> if this Mode is the exclusive
	     *         mode of the enclosing { * @see ModeManager#isExclusiveMode(Mode)
	     */
	public boolean isExclusive() {
		return _modeManager.isExclusiveMode(this);
	}

	/** Invoked when a key-pressed event was received.
	 *  <p>
	 *  If the key 'ESCAPE' was pressed, the mode is destroy.
	 *
	 * @param ke description of the key-event.
	 * @see java.awt.event.KeyEvent
	 */
	public void keyPressed(KeyEvent ke) {
		if (ke.getKeyCode() == KeyEvent.VK_ESCAPE) {
			done();
			ke.consume();
		}
	}

	/** Invoked when a key-released event was received.
	     *
	     * @param ke description of the key-event.
	     * @see java.awt.event.KeyEvent
	     */
	public void keyReleased(KeyEvent ke) {
	}

	/** Invoked when a key-typed event was received.
	     *
	     * @param ke description of the key-event.
	     * @see java.awt.event.KeyEvent
	     */
	public void keyTyped(KeyEvent ke) {
	}

	/** Invoked when a mouse-clicked event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDoubleClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseDoubleClicked(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-dragged event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-entered event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-exited event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-moved event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-pressed event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
	}

	/** Invoked when a mouse-released event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
	}

	/** Request the key focus for the JGraph parent.
	     *  <p>
	     *  This method add to the focus manager to give
	     *  keyboard events to the
	     *  {@link mygale.mcge.core.JGraph JGraph parent}
	     */
	public void requestFocus() {
		if (_modeManager != null)
			_modeManager.requestFocus();
	}

	/** Set the mouse cursor to some appropriate for this mode.
	     *
	     * @param c a reference to the new mouse cursor.
	     */
	public void setCursor(Cursor c) {
		_modeManager.getGraph().setCursor(c);
	}

	/** Set this mode to exclusive state or not.
	     *
	    @link ModeManager ModeManager},
	     *              otherwise <code>false</code>.
	     * @param exclu <code>true</code> if this Mode must be the exclusive
	     *              mode of the enclosing { * @see ModeManager#setExclusiveMode(Mode,boolean)
	     */
	public void setExclusive(boolean exclu) {
		_modeManager.setExclusiveMode(this, exclu);
	}

	/** Set the mouse cursor to some appropriate for this mode.
	     *
	     * @param nCursor the id of a predefined cursor.
	     * @see java.awt.Cursor
	     */
	public void setPredefinedCursor(int nCursor) {
		_modeManager.getGraph().setCursor(Cursor.getPredefinedCursor(nCursor));
	}

	/**
	 * @return JGraph
	 */
	public JGraph getGraph() {
		return _modeManager.getGraph();
	}

	/** Modes can paint themselves to give the user feedback. For
	 *  example, {@link ModeSelect ModeSelect} paints the
	 *  a selection rectangle that permits to select objects.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
	}
}
