package mygale.mcge.mode;

import java.awt.*;
import java.util.Vector;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.*;
import mygale.mcge.net.INetEdge;
import mygale.mcge.net.INetNode;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.undoredo.RedoableCmd;
import mygale.mcge.undoredo.UndoableCmd;

/** This class implement a special mode that manage the
 *  creation of edge between two node of the graph
 *  by the mouse.
 *  <p>
 *  This mode is normaly called by the
 *  {@link mygale.mcge.cmd.CmdCreateEdge CmdCreateEdge}
 *  which pass to the mode an instance of
 *  {@link mygale.mcge.fig.FigEdge FigEdge} and
 *  their attached
 *  {@link mygale.mcge.net.NetEdge NetEdge} to be
 *  added in the graph.
 *  <p>
 *  The creation of edge can be made in two manner : Drag
 *  the mouse from a port and release the the mouse on a port
 *  to create a edge with two control points only. Or click
 *  on a port, the click the mouse to put new control point
 *  until you click on a port for the final control point.
 *  <p>
 *  This mode manage also the highlight of the
 *  {@link mygale.mcge.fig.FigPort ports} on which
 *  the mouse enter or exit.
 *  <p>
 *  <font color="#FF0000"><b>Todo:</b> finish to comment this class.</font>
 *  <p><center>
 *    <img src="doc-files/ModeCreateEdge_uml.gif" width="408" height="200" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/FigPort.html">
 *      <area shape="Rect" coords="14,117,113,144"  href="../fig/FigEdge.html">
 *      <area shape="Rect" coords="271,117,407,144"  href="../fig/FigNode.html">
 *      <area shape="Rect" coords="77,171,176,199"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="212,171,311,199"  href="../sel/Selection.html">
 *    </map>
 *  </center></p>

 * @see ModeCreateFig
 * @see ModeCreateEdge
 */
public class ModeCreateEdge extends ModeUndoable {

	/** temporary reference to the port on wich the current (not the last) mouse
		* event has is. used to detect the enter and the exit of a port.
		*/
	//	private IFigPort _curFigPort = null;

	/** Temporary reference to the fig that own the FigPort on wich
		*  the current (not the last) mouse event is.
		*/
	private IFigNode _curFigNode = null;

	////
	// the current FigEdge and NetEdge to be added

	/** The FigEdge instance to be added.
		*/
	private IFigEdge _figEdge = null;
	/** The NetEdge instance to be added.
		*/
	private INetEdge _netEdge = null;

	////
	// the current FigNode and FigPort for the connection

	/** The FigNode wich own the port of the beginning connection.
		*/
	private IFigNode _sourceFigNode = null;
	/** The FigPort that is the beginning connection.
		*/
	//	private IFigPort _sourceFigPort = null;

	/** The FigNode wich own the port of the connection end.
		*/
	private IFigNode _destinationFigNode = null;

	/** The FigPort this is the end of the connection.
		*/
	//	private IFigPort _destinationFigPort = null;

	////
	// constants

	protected static final int CONNECT_FROM_PORT = 1;
	protected static final int CONNECT_TO_PORT = 2;

	protected static final int SOURCE_CONNECTION_PORT = 1;
	protected static final int DESTINATION_CONNECTION_PORT = 2;

	private static final int INITIAL_STATE = 1;
	private static final int DRAG_LINE_EDGE_STATE = 2;
	private static final int DRAG_POLY_EDGE_STATE = 3;

	////
	//

	protected Vector _points = null;
	protected int _state = INITIAL_STATE;
	protected Rectangle _rectCtrlPoints = new Rectangle(0, 0, 0, 0);

	/** This class is the undo/redo command for the creation
		*  of figures.
		*
		
		*/
	protected class UndoCreate implements UndoableCmd, RedoableCmd {

		/** A reference to the added figure.
		 */
		protected final IFigEdge _figure;

		/** A reference to the added figure.
		 */
		protected final INetEdge _net;

		/** Start port.
		 */
		protected final INetNode _beginPort;

		/** End port.
		 */
		protected final INetNode _endPort;

		/** Construct a new UndoCreate.
		 *
		 * @param figure the added figure.
		 */
		public UndoCreate(IFigEdge figure) {
			_figure = figure;
			_net = (INetEdge)figure.getNetObject();
			_endPort = _net.getBeginNode();
			_beginPort = _net.getEndNode();
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			JGraph graph = _modeManager.getGraph();
			Rectangle damagedBounds = _figure.getDamagedBounds(getPrefs());
			graph.remove(_figure);
			graph.damaged(damagedBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			JGraph graph = _modeManager.getGraph();
			_figure.setNetObject(_net);
			graph.getFigMapper().setFigFor(_net, _figure);
			_net.setBeginNode(_beginPort);
			_net.setEndNode(_endPort);
			Rectangle damagedBounds = _figure.getBounds();
			graph.add(_figure);
			graph.damaged(damagedBounds);
		}

	} /* class UndoCreate */

	/** Construct a new ModeMove. The ModeManager owner, the FigEdge and
	     *  the NetEdge should be set before using this object.
	     */
	public ModeCreateEdge() {
		super();
		setEdge(null, null);
		_state = INITIAL_STATE;
	}

	/** Construct a new ModeMode with the given parent, instance of FigEdge to be
	 * added and the attached instance of NetEdge.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 * @param figEdge a reference to the figure, which will be
	 *                added.
	 * @param netEdge a reference to the NetNode, which will be
	 *                added.
	 */
	public ModeCreateEdge(ModeManager modeManager, IFigEdge figEdge, INetEdge netEdge) {
		super(modeManager);
		setEdge(figEdge, netEdge);
		_state = INITIAL_STATE;
	}

	protected void addCtrlPoint(Point pt) {
		_points.add(pt);
	}

	protected void calcCtrlPointBounds() {
		Point pt = (Point)_points.elementAt(0);
		int size = _points.size();
		_rectCtrlPoints.setBounds(pt.x, pt.y, 0, 0);
		for (int i = 1; i < size; i++) {
			_rectCtrlPoints.add((Point)_points.elementAt(i));
		}
	}

	/** clean the current create edge mode.
	     */
	public void cleanMode() {
		setActiveState(_sourceFigNode, false);
		setActiveState(_destinationFigNode, false);
		_points = null;
		_state = INITIAL_STATE;
	}

	/** Search the port under the mouse.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @param me
	     * @param connectType int
	 */
	public void computeCurrentPort(MCGEMouseEvent me, int connectType) {
		IFig fig = me.getFigure();
		if (fig != null) {
			if (fig instanceof IFigNode) {

				INetNode netNode = (INetNode) ((IFigNode)fig).getNetObject();

				if (connectType == CONNECT_FROM_PORT) {
					if (netNode.canConnect(_netEdge)) {
						_curFigNode = (IFigNode)fig;
						return;
					}
				} else if (connectType == CONNECT_TO_PORT) {
					if (netNode.canConnect(_netEdge, (INetNode)_sourceFigNode.getNetObject())) {
						_curFigNode = (IFigNode)fig;
						return;
					} else {
						// cant connect to
					}
				} else {
					throw new MCGEInternalError("unknow connection port !!");
				}
				//				}
			}
		}
		_curFigNode = null;
	}

	protected void damaged() {
		_modeManager.getGraph().damaged(_rectCtrlPoints);
	}

	protected Point getCtrlPointAt(int index) {
		return new Point((Point)_points.elementAt(index));
	}

	protected Point getFirstCtrlPoint() {
		return new Point((Point)_points.elementAt(0));
	}

	protected Point getLastCtrlPoint() {
		return new Point((Point)_points.elementAt(_points.size() - 1));
	}

	protected int getNbCtrlPoints() {
		return _points.size();
	}

	/** Insert the edge into the graph model.
	     */
	protected void insertFigEdge() {
		if ((_sourceFigNode != null) && (_destinationFigNode != null)) {
			for (int i = 0; i < _points.size(); i++) {
				_figEdge.setCtrlPointAt(i, (Point)_points.get(i));
			}
			_netEdge.setBeginNode((INetNode)_sourceFigNode.getNetObject());
			_netEdge.setEndNode((INetNode)_destinationFigNode.getNetObject());
			// Save the undo command
			addUndoCmd(new UndoCreate(_figEdge));
			// Add into the graph model
			_modeManager.getGraph().add(_figEdge);
		}
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		return "Drag from a port to create an edge. Type <escape> to cancel.";
	}

	/** Invoked when a mouse-clicked event was received.
	     *  <p>
	     *  Handle the click mouse event. If we are in the intial state and
	     *  the mouse is on a port, begin to create a multipoint edge from this port
	     *  by adding the control point of this port to the edge and a control
	     *  point that reflect the position of the mouse.
	     *  If we are in the multi point edge creation state, and the mouse isn't on a port
	     *  add the current position of the mouse as a new ctrl point to the edge
	     *  and redraw it.
	     *  If we are in the multi-point edge creation state, and the mouse is on a port,
	     *  add the current position as the last control poit of the edge and add
	     *  this edge to the graph, then qui the mode.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		if (_state == INITIAL_STATE) {

			updateSourcePort(me);

			// now set the two first point of the poly edge
			if (_sourceFigNode != null) {
				_points = new Vector();

				Point pt2 = new Point(me.getX(), me.getY());
				Point pt1 = _sourceFigNode.intersectWith(pt2, 0, 0);
				addCtrlPoint(pt1);
				addCtrlPoint(pt2);

				calcCtrlPointBounds();
				_state = DRAG_POLY_EDGE_STATE;

				damaged();
				me.consume();
				return;
			}
		} else if (_state == DRAG_POLY_EDGE_STATE) {
			int curSize = getNbCtrlPoints();
			int maxSize = _figEdge.getMaxAllowedCtrlPoints();
			if (curSize <= maxSize) {

				// compute the last port
				updateDestinationPort(me);

				if (_destinationFigNode != null) {
					// cas final
					calcCtrlPointBounds();

					Point firstPt = getCtrlPointAt(curSize - 2);
					Point lastPt = getCtrlPointAt(curSize - 1);

					lastPt.x = me.getX();
					lastPt.y = me.getY();

					lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
					if (curSize == 2) {
						firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
						lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
						// recompute to improve the precision
						firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
						lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
					}

					setCtrlPointAt(curSize - 2, firstPt);
					setCtrlPointAt(curSize - 1, lastPt);

					updateCtrlPointBounds();

					// insrer la figEdge
					insertFigEdge();

					me.consume();
					done();
					return;
				} else {
					// cas gnral
					if (curSize == maxSize) {
						done();
						return;
					}

					calcCtrlPointBounds();

					Point firstPt = getCtrlPointAt(curSize - 2);
					Point lastPt = getCtrlPointAt(curSize - 1);

					lastPt.x = me.getX();
					lastPt.y = me.getY();

					if (curSize == 2) {
						firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
					}

					setCtrlPointAt(curSize - 2, firstPt);
					setCtrlPointAt(curSize - 1, lastPt);
					addCtrlPoint(me.getPoint());

					updateCtrlPointBounds();
					damaged();
				}

				me.consume();
				return;
			}
		}
		done();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  On double click mouse event, quit the current more.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDoubleClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseDoubleClicked(MCGEMouseEvent me) {
		done();
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  On mouse dragging, stretch the connection line.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if (!me.isOnlyLeftButtonDown() || (!isExclusive())) {
			done();
			return;
		}

		if (_state != DRAG_LINE_EDGE_STATE) {
			done();
		} else {
			updateDestinationPort(me);
			calcCtrlPointBounds();

			Point firstPt = getFirstCtrlPoint();
			Point lastPt = me.getPoint();

			if (_destinationFigNode == null) {
				firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
			} else {
				lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);

				// recompute to improve precision
				firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
				lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
			}

			setFirstCtrlPoint(firstPt);
			setLastCtrlPoint(lastPt);

			updateCtrlPointBounds();

			damaged();
		}

		me.consume();
	}

	/** Invoked when a mouse-entered event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-entered event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-moved event was received.
	     *  <p>
	     *  handle the move mouse event. if we are in the intial state, just
	     *  highlight the port on which the mouse is. if we are in the create
	     *  multi point edge, update the last ctrl point of the edge and redraw
	     *  it.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
		if (_state == INITIAL_STATE) {
			updateSourcePort(me);
		} else if (_state == DRAG_POLY_EDGE_STATE) {
			updateDestinationPort(me);

			int curSize = getNbCtrlPoints();
			if (curSize == 2) {
				calcCtrlPointBounds();

				Point firstPt = getFirstCtrlPoint();
				Point lastPt = getLastCtrlPoint();

				lastPt.x = me.getX();
				lastPt.y = me.getY();
				firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
				if (_destinationFigNode != null) {
					lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);

					// recompute to improve precision.
					firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
					lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
				}

				setFirstCtrlPoint(firstPt);
				setLastCtrlPoint(lastPt);

				updateCtrlPointBounds();
				_modeManager.getGraph().damaged(_rectCtrlPoints);
			} else {
				calcCtrlPointBounds();

				Point firstPt = getCtrlPointAt(curSize - 2);
				Point lastPt = getCtrlPointAt(curSize - 1);

				lastPt.x = me.getX();
				lastPt.y = me.getY();
				if (_destinationFigNode != null) {
					lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);
				} else {
					lastPt.x = me.getX();
					lastPt.y = me.getY();
				}

				setLastCtrlPoint(lastPt);

				updateCtrlPointBounds();
				damaged();
			}
		} else {
			throw new MCGEInternalError("unkonw state in MouseMoved method" + " of the ModeCreateEdge class");
		}
		me.consume();
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  Handle mouse down events by preparing for a drag. If the mouse
	 *  down event happens on a port, begin to create the edge connection.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	*/
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if (!me.isOnlyLeftButtonDown() || (isExclusive())) {
			done();
			return;
		}

		updateSourcePort(me);

		if (_sourceFigNode == null) {
			done();
		} else {
			_state = DRAG_LINE_EDGE_STATE;
			setExclusive(true);

			Point pt2 = me.getPoint();
			Point pt1 = _sourceFigNode.intersectWith(pt2, 0, 0);

			_points = new Vector();
			addCtrlPoint(pt1);
			addCtrlPoint(pt2);
			calcCtrlPointBounds();

			damaged();
		}

		me.consume();
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  Handle mouse release event. If the mouse down event
	     *  happens on a port, add the created edge.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if (!me.isOnlyLeftButtonDown() || (!isExclusive())) {
			return;
		}

		if (_state != DRAG_LINE_EDGE_STATE) {
			return;
		} else {
			updateDestinationPort(me);

			if (_destinationFigNode == null) {
				return;
			}

			calcCtrlPointBounds();
			setActiveState(_sourceFigNode, false);
			setActiveState(_destinationFigNode, false);

			Point firstPt = getFirstCtrlPoint();
			Point lastPt = me.getPoint();
			lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);

			// recompute to improve precision
			firstPt = _sourceFigNode.intersectWith(lastPt, 0, 0);
			lastPt = _destinationFigNode.intersectWith(firstPt, 0, 0);

			setFirstCtrlPoint(firstPt);
			setLastCtrlPoint(lastPt);

			_state = INITIAL_STATE;
			insertFigEdge();
		}
		me.consume();
		done();
	}

	private void setActiveState(IFigNode figNode, boolean activate) {
		if (figNode != null)
			figNode.setActiveState(activate);
	}

	/** Paint the line beetwen the control point of
	 *  the edge to create.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
		if ((_state == DRAG_POLY_EDGE_STATE) || (_state == DRAG_LINE_EDGE_STATE)) {
			int curSize = getNbCtrlPoints();
			g2d.setColor(getPrefs().getShadowColor());
			for (int i = 0; i < curSize - 1; i++) {
				Point firstPt = getCtrlPointAt(i);
				Point lastPt = getCtrlPointAt(i + 1);
				g2d.drawLine(firstPt.x, firstPt.y, lastPt.x, lastPt.y);
			}
		}
	}

	protected void setCtrlPointAt(int index, Point pt) {
		_points.setElementAt(pt, index);
	}

	/** Set the FigEdge with their NetEdge to be added. The NetEdge should be
	 *  attached to the figEdge by the caller of this method.
	 *
	 * @param figEdge a reference to the figure, which will be
	 *                added.
	 * @param netEdge a reference to the NetNode, which will be
	 *                added.
	 */
	public void setEdge(IFigEdge figEdge, INetEdge netEdge) {
		_figEdge = figEdge;
		_netEdge = netEdge;
	}

	protected void setFirstCtrlPoint(Point pt) {
		_points.setElementAt(pt, 0);
	}

	protected void setLastCtrlPoint(Point pt) {
		_points.setElementAt(pt, _points.size() - 1);
	}

	protected void updateCtrlPointBounds() {
		int size = _points.size();
		for (int i = 0; i < size; i++) {
			_rectCtrlPoints.add((Point)_points.elementAt(i));
		}
	}

	protected void updateDestinationPort(MCGEMouseEvent me) {
		// compute the first port
		computeCurrentPort(me, CONNECT_TO_PORT);
		if (_curFigNode != null) {
			if (_curFigNode != _destinationFigNode) {
				if (_sourceFigNode != _curFigNode) {
					setActiveState(_destinationFigNode, false);
				}
				_destinationFigNode = _curFigNode;
				setActiveState(_destinationFigNode, true);
			}
		} else {
			if (_sourceFigNode != _destinationFigNode) {
				setActiveState(_destinationFigNode, false);
			}

			_destinationFigNode = null;
		}
	}

	// Utility methods

	protected void updateSourcePort(MCGEMouseEvent me) {
		// compute the first port
		computeCurrentPort(me, CONNECT_FROM_PORT);
		if (_curFigNode != null) {
			if (_curFigNode != _sourceFigNode) {
				setActiveState(_sourceFigNode, false);
				_sourceFigNode = _curFigNode;
				setActiveState(_sourceFigNode, true);
			}
		} else {
			setActiveState(_sourceFigNode, false);
			_sourceFigNode = null;
		}
	}
}