package mygale.mcge.mode;

import java.awt.*;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.IFigLayered;
import mygale.mcge.sel.Selection;
import mygale.mcge.undoredo.RedoableCmd;
import mygale.mcge.undoredo.UndoableCmd;

/** This class implements a Mode that permits to create a Fig
 *  (except {@link mygale.mcge.fig.FigEdge FigEdges}).
 *  <p>
 *  This Mode permits to put a new
 *  {@link mygale.mcge.fig.Fig figure} into the
 *  {@link mygale.mcge.core.JGraph graph's editor}.
 *  There have two way to pout a new figure : the
 *  user can simply click or he can use a sizing rectangle,
 *  which can be dragged to choose the size of the figure.
 *  The new figure will be put at the position choosed by the
 *  used and have the dimensions choosed by the user
 *  (if the sizing rectangle was used) or the default ones.
 *  <p>
 *  We assume that the new Fig is the figure given to the
 *  {@link #ModeCreateFig(ModeManager,Fig) constructor}.
 *  We assume too that this figure are completly linked
 *  to a {@link mygale.mcge.net.NetObject NetObject} iff
 *  the figure was an instance of
 *  {@link mygale.mcge.fig.FigModelElement FigModelElement}.
 *  <p><center>
 *    <img src="doc-files/ModeCreateFig_uml.gif" width="363" height="142" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/Fig.html">
 *      <area shape="Rect" coords="28,114,127,141"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="262,114,361,141"  href="../sel/Selection.html">
 *      <area shape="Rect" coords="145,114,244,141"  href="ModeCreateNode.html">
 *    </map>
 *  </center></p>
 *
 * @see ModeCreateEdge
 * @see ModeCreateNode
 */
public class ModeCreateFig extends ModeUndoable {

	/** This is the Fig that must be added to the diagram.
	 */
	protected IFigLayered _figToAdd;

	/** This is the x coordonate of the initial point of the sizing rectangle.
	 */
	protected int _xAnchor;
	/** This is the y coordonate of the initial point of the sizing rectangle.
	 */
	protected int _yAnchor;

	/** This is the sizing rectangle.
	 */
	protected Rectangle _sizingRect = new Rectangle(0, 0, 0, 0);

	/** <code>true</code> when the sizing rectangle should be painted.
	 */
	protected boolean _showSizingRect = false;

	/** This class is the undo/redo command for the creation
	 *  of figures.
	 *
		 */
	protected class UndoCreate implements UndoableCmd, RedoableCmd {

		/** A reference to the added figure.
		 */
		protected final IFigLayered _figure;

		/** The bounds of the added figure.
		 */
		protected final Rectangle _bounds;

		/** Construct a new UndoCreate.
		 *
		 * @param figure the added figure.
		 */
		public UndoCreate(IFigLayered figure) {
			_figure = figure;
			_bounds = new Rectangle(figure.getBounds());
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			JGraph graph = _modeManager.getGraph();
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			graph.remove(_figure);
			graph.damaged(damagedBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			JGraph graph = _modeManager.getGraph();
			_figure.setBounds(_bounds.x, _bounds.y, _bounds.width, _bounds.height);
			graph.add(_figure);
			graph.damaged(_figure.getBounds());
		}

	} /* class UndoCreate */

	/** Construct a new ModeCreateFig with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 * @param figInstance a reference to the figure, which will be
	 *                    added.
	 */
	public ModeCreateFig(ModeManager modeManager, IFigLayered figInstance) {
		super(modeManager);
		_figToAdd = figInstance;
	}

	// Creation API

	/** Add the {@link mygale.mcge.fig.Fig Fig} into the diagram.
	 *  <p>
	 *  The new figure will be put at the position
	 *  (<var>x</var>,<var>y</var>) and have the dimensions
	 *  (<var>width</var>,<var>height</var>).
	 *  <p>
	 *  We assume that the new Fig is the figure given to the
	 *  {@link #ModeCreateFig(ModeManager,Fig) constructor}.
	 *  We assume too that this figure are completly linked
	 *  to a {@link mygale.mcge.net.NetObject NetObject} iff
	 *  the figure was an instance of
	 *  {@link mygale.mcge.fig.FigModelElement FigModelElement}.
	 *
	 * @param x the position of the figure in the workspace.
	 * @param y the position of the figure in the workspace.
	 * @param width width of the figure.
	 * @param height height of the figure.
	 */
	public void addFig(int x, int y, int width, int height) {
		_figToAdd.setBounds(
			x,
			y,
			Math.max(_figToAdd.getMinimalWidth(), width),
			Math.max(_figToAdd.getMinimalHeight(), height));
		JGraph graph = _modeManager.getGraph();
		createFig(graph);
		Selection sel = graph.getSelection();
		boolean oldInform = sel.setCmdInform(false);
		sel.deselectAll();
		sel.select(_figToAdd);
		sel.setCmdInform(oldInform);
		sel.informCmd();
		graph.damaged(_figToAdd);
	}

	/** This method put this mode to a stable state.
	 */
	public void cleanMode() {
		setCursor(Cursor.getDefaultCursor());
	}

	/** Returns the cursor that should be shown when this Mode starts.
	     *  For ModeCreateFig, the default cursor is a crosshair.
	     *
	     * @return a reference to a cursor.
	     */
	public Cursor getInitialCursor() {
		return Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR);
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		if (isExclusive())
			return "Drag the rectangle to set the figure's size. Type <escape> to cancel.";
		else
			return "Click to set the start point of a figure. Type <escape> to cancel.";
	}

	/** Invoked when a mouse-clicked event was received.
	     *  <p>
	     *  If the user simple click, this mode create a new
	     *  {@link mygale.mcge.fig.Fig Fig} with the minimal sizes.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		if (me.isOnlyLeftButtonDown() && !isExclusive()) {
			addFig(me.getX(), me.getY(), _figToAdd.getMinimalWidth(), _figToAdd.getMinimalHeight());
			me.consume();
		}
		if (!isExclusive())
			done();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  Ignore the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDoubleClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseDoubleClicked(MCGEMouseEvent me) {
		if (!isExclusive())
			done();
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  The user moves the mouse while a sizing rectangle
	     *  is display. This method update this one.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}

		int x = me.getX(), y = me.getY();
		JGraph graph = _modeManager.getGraph();

		graph.damaged(_sizingRect);
		updateSizingRect(x, y);
		graph.damaged(_sizingRect);
		graph.scrollToShow(x, y);
		me.consume();
	}

	/** Invoked when a mouse-entered event was received.
	     *  <p>
	     *  consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *  <p>
	     *  consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-moved event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  The user try to create a fig with a sizing rectangle.
	 *  Initiate the sizing rectangle.
	 *
	 * @param me description of the mouse-event.
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive())) {
			return;
		}

		int x = me.getX(), y = me.getY();

		_sizingRect.x = x;
		_sizingRect.y = y;
		_sizingRect.width = 0;
		_sizingRect.height = 0;
		setPredefinedCursor(Cursor.HAND_CURSOR);
		setExclusive(true);
		startSizingRect(x, y);
		me.consume();
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  The user releases the mouse while a sizing rectangle
	     *  is display. This method create a new figure
	     *  with the sizes given by the sizing rectangle; and
	     *  clear the sizing rectangle.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}

		endSizingRect(me.getX(), me.getY());
		addFig(_sizingRect.x, _sizingRect.y, _sizingRect.width, _sizingRect.height);
		_modeManager.getGraph().damaged(_sizingRect);
		done();
		me.consume();
	}

	/** Add the {@link mygale.mcge.fig.Fig Fig} into the
	     *  {@link mygale.mcge.core.JGraph editor}.
	     *  <p>
	     *  This method was called by
	     *  {@link #addFig(int,int,int,int) addFig()}.
	     *  This method simply add the figure to the graph editor.
	     *  You must override this methode to change the
	     *  creation of figures.
	     *  <p>
	     *  We assume that the new figure is stored in
	     *  <var>_figToAdd</var> attribute.
	     *  <p>
	     *  This method are responsible to update the undo list.
	     *
	     * @param graph the graph in which the figure will be added.
	     */
	protected void createFig(JGraph graph) {
		graph.add(_figToAdd);
		// Save the undoing command
		addUndoCmd(new UndoCreate(_figToAdd));
	}

	/** Clear the sizing rectangle.
	     *
	     * @param x new position for the rectangle.
	     * @param y new position for the rectangle.
	     */
	void endSizingRect(int x, int y) {
		updateSizingRect(x, y);
		_showSizingRect = false;
	}

	/** Initiate the sizing rectangle with the given point.
	 *
	 * @param x initial position of the rectangle.
	 * @param y initial position of the rectangle.
	 */
	void startSizingRect(int x, int y) {
		_showSizingRect = false;
		_xAnchor = x;
		_yAnchor = y;
	}

	/** Update the sizing rectangle with the given point.
	     *  Change the second point of the sizing rectangle.
	     *  Keep the initial point unchanged.
	     *
	     * @param x new position for the rectangle.
	     * @param y new position for the rectangle.
	     */
	void updateSizingRect(int x, int y) {
		Rectangle r = _modeManager.getGraph().getWorkspaceRect();
		if (x < r.x)
			x = r.x;
		else if (x > (r.x + r.width))
			x = r.x + r.width;

		if (y < r.y)
			y = r.y;
		else if (y > (r.y + r.height))
			y = r.y + r.height;

		int bound_w = Math.abs(_xAnchor - x);
		int bound_h = Math.abs(_yAnchor - y);

		int bound_x = Math.min(_xAnchor, x);
		int bound_y = Math.min(_yAnchor, y);

		_sizingRect.x = bound_x;
		_sizingRect.y = bound_y;
		_sizingRect.width = bound_w;
		_sizingRect.height = bound_h;
		_showSizingRect = true;
	}

	/** Paint the sizing rectangle if appropriate.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
		if (_showSizingRect) {
			Color rectColor = getPrefs().getShadowColor();
			g2d.setColor(rectColor);
			g2d.drawRect(_sizingRect.x, _sizingRect.y, _sizingRect.width, _sizingRect.height);
		}
	}
}
