package mygale.mcge.mode;

import mygale.mcge.core.JGraph;
import mygale.mcge.fig.IFigNode;
import mygale.mcge.net.INetNode;

/** This class implements a Mode that permits to create a Node
 *  <p>
 *  This Mode permits to put a new
 *  {@link mygale.mcge.fig.FigNode FigNode} into the
 *  {@link mygale.mcge.core.JGraph graph's editor}.
 *  There have two way to pout a new figure : the
 *  user can simply click or he can use a sizing rectangle,
 *  which can be dragged to choose the size of the figure.
 *  The new figure will be put at the position choosed by the
 *  used and have the dimensions choosed by the user
 *  (if the sizing rectangle was used) or the default ones.
 *  <p>
 *  We assume that the new Fig is the figure given to the
 *  {@link #ModeCreateNode(ModeManager,FigNode,NetNode) constructor}.
 *  We assume too that this figure are not linked
 *  to the given {@link mygale.mcge.net.NetObject NetObject}.
 *  The linkage will be make by this Mode during the call to
 *  {@link #createFig(JGraph) createFig()}.
 *  <p><center>
 *    <img src="doc-files/ModeCreateNode_uml.gif" width="363" height="142" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="145,0,244,28"  href="ModeCreateFig.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/FigNode.html">
 *      <area shape="Rect" coords="28,114,127,141"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="262,114,361,141"  href="../sel/Selection.html">
 *      <area shape="Rect" coords="145,114,244,141"  href="../net/NetNode.html">
 *    </map>
 *  </center></p>
 *
 * @see ModeCreateFig
 * @see ModeCreateEdge
 */
public class ModeCreateNode extends ModeCreateFig {

	/** This is the Net that must be added to the diagram.
	 */
	protected INetNode _netToAdd;

	/** Construct a new ModeSelect with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 * @param figInstance a reference to the figure, which will be
	 *                    added.
	 * @param netInstance a reference to the NetNode, which will be
	 *                    added.
	 */
	public ModeCreateNode(ModeManager modeManager, IFigNode figInstance, INetNode netInstance) {
		super(modeManager, figInstance);
		_netToAdd = netInstance;
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		if (isExclusive())
			return "Drag the rectangle to set the node's size. Type <escape> to cancel.";
		else
			return "Click to set the start point of a node. Type <escape> to cancel.";
	}

	/** Add the {@link mygale.mcge.fig.Fig Fig} into the
	 *  {@link mygale.mcge.core.JGraph editor}.
	 *  <p>
	 *  This method was called by
	 *  {@link mygale.mcge.mode.ModeCreateFig#addFig(int,int,int,int) addFig()}.
	 *  This method simply add the figure to the graph editor.
	 *  You must override this methode to change the
	 *  creation of figures.
	 *  <p>
	 *  We assume that the new figure is stored in
	 *  <var>_figToAdd</var> attribute.
	 *  This method are responsible to update the undo list.
	 *  <p>
	 *  We override this method because, we need to
	 *  correctly link the
	 *  {@link mygale.mcge.fig.FigNode FigNode} and the
	 *  {@link mygale.mcge.net.NetNode NetNode}, and
	 *  we preferes to add the NetNode in the model
	 *  insteed of the FigNode.
	 *
	 * @param graph the graph in which the figure will be added.
	 */
	protected void createFig(JGraph graph) {
		// Make links between FigNode and NetNode
		 ((IFigNode)_figToAdd).setNetObject(_netToAdd);
		graph.getFigMapper().setFigFor(_netToAdd, (IFigNode)_figToAdd);
		// Add objects to graph diagram
		graph.add(_figToAdd);
		// Save the undoing command
		addUndoCmd(new UndoCreate(_figToAdd));
	}
}
