package mygale.mcge.mode;

import java.awt.Graphics2D;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.event.mouse.MCGEMouseEventListener;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.util.prefs.Prefs;

/** ModeManager keeps track of all the
 *  {@link Mode Modes} for a given
 *  {@link mygale.mcge.core.JGraph JGraph}.
 *  Events are passed to the Modes for handling.  The submodes are
 *  prioritized according to their order on a stack, i.e., the last
 *  Mode added gets the first chance to handle an Event.
 *  <p>
 *  This ModeManager takes into account an exclusive Mode. An exclusive
 *  mode is a {@link Mode Mode} was received all events before the
 *  stacked modes. If an event was not eated by the exclusive mode,
 *  the exclusive mode was destroy and the event was ignored.
 *  <p><center>
 *    <img src="doc-files/ModeManager_uml.gif" width="336" height="200" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,104,73,132"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="104,0,248,28"  href="../event/mouse/MCGEMouseEventListener.html">
 *      <area shape="Rect" coords="235,104,334,132"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="118,172,190,198"  href="Mode.html">
 *      <area shape="Rect" coords="91,36,163,63"  href="http://java.sun.com/products/jdk/1.2/docs/api/java/awt/event/KeyListener.html">
 *    </map>
 *  </center></p>

 */
public class ModeManager implements MCGEMouseEventListener, KeyListener {

	private interface KeyEventModeInvoker {
		void invoke(Mode mode, KeyEvent event);
	}

	private static final KeyEventModeInvoker _keyPressedModeInvoker = new KeyEventModeInvoker() {
		public void invoke(Mode mode, KeyEvent event) {
			mode.keyPressed(event);
		}
	};

	private static final KeyEventModeInvoker _keyTypedModeInvoker = new KeyEventModeInvoker() {
		public void invoke(Mode mode, KeyEvent event) {
			mode.keyTyped(event);
		}
	};

	private interface MCGEMouseEventModeInvoker {
		void invoke(Mode mode, MCGEMouseEvent event);
	}

	private static final MCGEMouseEventModeInvoker _mcgeMouseClickedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseClicked(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseDoubleClickedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseDoubleClicked(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseDraggedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseDragged(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseEnteredModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseEntered(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseExitedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseExited(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseMovedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseMoved(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMousePressedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMousePressed(event);
		}
	};

	private static final MCGEMouseEventModeInvoker _mcgeMouseReleasedModeInvoker =
		new MCGEMouseEventModeInvoker() {
		public void invoke(Mode mode, MCGEMouseEvent event) {
			mode.mcgeMouseReleased(event);
		}
	};

	/** The stack of {@link Mode Mode's} that are all
	 *  active simultainously, the order of Mode's on
	 *  the stack is their priority, i.e., the topmost
	 *  Mode gets the first chance to handle an incoming
	 *  Event.
	 */
	protected java.util.List _stackMode = new ArrayList();

	/** The Editor that owns this ModeManager.
	 */
	protected JGraph _graph;

	/** This is the exclusive Mode.
	 */
	protected Mode _exclusiveMode = null;

	/**  Construct a ModeManager with no modes.
	 *
	 * @param graph a reference to the editor that contains this ModeManager.
	 */
	public ModeManager(JGraph graph) {
		_graph = graph;
	}

	/** Set the graph editor's current mode to the given Mode instance.
	     *
	     * @param mode the mode to push to the stack.
	     */
	public void beginMode(Mode mode) {
		mode.setModeManager(this);
		pushMode(mode);
		if (mode.canExit())
			_graph.setStatusMsg(mode.instructions());
	}

	/** Remove the current mode, and set the graph editor's mode to
	      * the most topless mode.
	      */
	public void endMode() {
		Mode mode = popMode();
		if (isExclusiveMode(mode)) {
			unsetExclusiveMode();
		}
		if (mode.canExit())
			_graph.clearStatus();
	}

	/** Forward a key event to modes.
	 *  <p>
	 *  The given event is send to the exclusive mode if it exists.
	 *  while the event was not eated by a mode, it is send to
	 *  the next mode in the stack.
	 *  In no mode has eated the event, the default method was called.
	 *
	 * @param event a reference to the event to forward.
	 * @param strEventHandler the name of the method that must be called
	 *                        for each Mode.
	 * @param strDefaultEventHandler the name of the default method that
	 *                               will be used if no event eats the event.
	 * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	 */
	protected void forwardToMode(KeyEvent event, KeyEventModeInvoker modeInvoker) {
		// Send the event to the exclusive mode
		if (haveExclusiveMode()) {
			modeInvoker.invoke(_exclusiveMode, event);
			if (!event.isConsumed()) {
				// The exclusive mode had not eaten the event
				// It still not be the exclusibe event.
				Mode oldExclusiveMode = _exclusiveMode;
				unsetExclusiveMode();
				oldExclusiveMode.cleanMode();
				if (oldExclusiveMode.canExit()) {
					_graph.clearStatus();
					if (getTopMode() != oldExclusiveMode) {
						throw new MCGEInternalError(
							"an exclusive mode must be " + "on the top of the stack.");
					} else {
						popMode();
					}
				}
				event.consume();
			}
		} else {
			for (int i = _stackMode.size() - 1; i >= 0; --i) {
				modeInvoker.invoke((Mode)_stackMode.get(i), event);

				if (event.isConsumed()) {
					break;
				}
			}
		}
	}

	/** Forward a mouse  event to modes.
	     *  <p>
	     *  The given event is send ti the exclusive mode if it exists.
	     *  while the event was not eated by a mode, it is send to
	     *  the next mode in the stack.
	     *  In no mode has eated the event, the default method was called.
	     *
	     * @param event a reference to the event to forward.
	     * @param strEventHandler the name of the method that must be called
	     *                        for each Mode.
	     * @param strDefaultEventHandler the name of the default method that
	     *                               will be used if no event eats the event.
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     */
	protected void forwardToMode(MCGEMouseEvent event, MCGEMouseEventModeInvoker modeInvoker) {
		// Send the event to the exclusive mode
		if (haveExclusiveMode()) {
			modeInvoker.invoke(_exclusiveMode, event);
			if (!event.isConsumed()) {
				// The exclusive mode had not eaten the event
				// It still not be the exclusive event.
				Mode oldExclusiveMode = _exclusiveMode;
				unsetExclusiveMode();
				oldExclusiveMode.cleanMode();
				if (oldExclusiveMode.canExit()) {
					_graph.clearStatus();
					if (getTopMode() != oldExclusiveMode) {
						throw new MCGEInternalError(
							"an exclusive mode must be on " + "the top of the stack.");
					} else {
						popMode();
					}
				}
				event.consume();
			}
		} else {
			for (int i = _stackMode.size() - 1; i >= 0; --i) {
				modeInvoker.invoke((Mode)_stackMode.get(i), event);

				if (event.isConsumed())
					break;
			}
		}
	}

	/** Reply the graph that contains this ModeManager.
	 *
	 * @return a reference to the editor that contains this ModeManager.
	 */
	public JGraph getGraph() {
		return _graph;
	}

	/** Reply the {@link mygale.mcge.util.prefs.Prefs preference manager}
	     *  of the graph that contains this ModeManager.
	     *
	     * @return a reference to the preference manager.
	     */
	public Prefs getPrefs() {
		return _graph.getPrefs();
	}

	/** Reply the top (first) Mode. return null if the stack is empty.
	     *
	     * @return a reference to the first Mode.
	     */
	protected Mode getTopMode() {
		if (_stackMode.size() > 0)
			return (Mode)_stackMode.get(_stackMode.size() - 1);
		return null;
	}

	/** Reply if this ModeManager have an exclusive Mode.
	     *
	     * @return <code>true</code> if this ModeManager have an
	     *         exclusive mode, otherwise <code>false</code>.
	     */
	public boolean haveExclusiveMode() {
		return _exclusiveMode != null;
	}

	/** Reply if the given {@link Mode Mode} was the exclusive one.
	 *
	 *
	 * @param mode Mode
	 * @return <code>true</code> if <var>mode</var> was the exclusive
	 *         mode, otherwise <code>false</code>.
	 */
	public boolean isExclusiveMode(Mode mode) {
		return _exclusiveMode != null && _exclusiveMode == mode;
	}

	/** Pass events to all modes in order, until one consumes it.
	     *
	     * @param event KeyEvent
	 * @see java.awt.event.KeyListener#keyPressed(KeyEvent)
	 */
	public void keyPressed(KeyEvent event) {
		forwardToMode(event, _keyPressedModeInvoker);

		if (!event.isConsumed()) {
			event.consume();
		}
	}

	/** Do nothing, this event are not passed to the Modes.
	     *
	     * @param event KeyEvent
	 * @see java.awt.event.KeyListener#keyReleased(KeyEvent)
	 */
	public void keyReleased(KeyEvent event) {
		event.consume();
	}

	/** Pass events to all modes in order, until one consumes it.
	 *
	 * @param event KeyEvent
	 * @see java.awt.event.KeyListener#keyTyped(KeyEvent)
	 */
	public void keyTyped(KeyEvent event) {
		forwardToMode(event, _keyTypedModeInvoker);

		if (!event.isConsumed()) {
			event.consume();
		}

	}

	/** Invoked when the mouse has been clicked on a component.
	 *
	 * @param me a reference to the mouse-event description.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		requestFocus();
		forwardToMode(me, _mcgeMouseClickedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when the mouse has been double clicked on a component.
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDoubleClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseDoubleClicked(MCGEMouseEvent me) {
		requestFocus();
		forwardToMode(me, _mcgeMouseDoubleClickedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when a mouse button is pressed on a component and then
	     *  dragged.  Mouse drag events will continue to be delivered to
	     *  the component where the first originated until the mouse button is
	     *  released (regardless of whether the mouse position is within the
	     *  bounds of the component).
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		requestFocus();
		forwardToMode(me, _mcgeMouseDraggedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when the mouse enters a figure.
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		forwardToMode(me, _mcgeMouseEnteredModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when the mouse exits a figure.
	     *
	     * @param me a reference to the mouse-event description.
	     */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		forwardToMode(me, _mcgeMouseExitedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when the mouse button has been moved on a component
	     * (with no buttons no down).
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
		forwardToMode(me, _mcgeMouseMovedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when a mouse button has been pressed on a component.
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
		requestFocus();
		forwardToMode(me, _mcgeMousePressedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Invoked when a mouse button has been released on a component.
	     *
	     * @param me a reference to the mouse-event description.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		requestFocus();
		forwardToMode(me, _mcgeMouseReleasedModeInvoker);

		if (!me.isConsumed()) {
			me.consume();
		}
	}

	/** Paint each mode in the stack: bottom to top.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {

		for (int i = 0; i < _stackMode.size(); i++) {
			((Mode)_stackMode.get(i)).paint(g2d);
		}
	}

	/** Return the topmost Mode and remove it iff it can exit. If the stack is
	     * empty, return null.
	     *
	     * @return a reference to the (old) first mode.
	     */
	private Mode popMode() {
		int size = _stackMode.size();

		if (size == 0)
			return null;

		Mode tmpMode = (Mode)_stackMode.get(size - 1);

		if (tmpMode.canExit() == true) {
			_stackMode.remove(size - 1);
		}

		return tmpMode;
	}

	/** Add the given Mode to the stack iff another instance
	     *  of the same class is not already on the stack.
	     *  if the stack array has reached the maximum allowed elements,
	     *  grow the size stack array by "ARRAY_DELTA".
	     *
	     * @param mode the Mode to push into the stack.
	     */
	private void pushMode(Mode mode) {
		_stackMode.add(mode);
	}

	/** Request the key focus for the JGraph parent.
	     *  <p>
	     *  This method add to the focus manager to give
	     *  keyboard events to the
	     *  {@link mygale.mcge.core.JGraph JGraph parent}
	     */
	public void requestFocus() {
		if (_graph != null)
			_graph.requestFocus();
	}

	/** This method permits to reset this manager by remove temp modes
	     *  and clean permanent modes.
	     */
	public void resetModes() {
		_graph.clearStatus();
		unsetExclusiveMode();
		// Clean modes
		for (int i = 0; i < _stackMode.size(); i++) {
			((Mode)_stackMode.get(i)).cleanMode();
		}

		// Delete temp modes
		Mode mode = getTopMode();
		while (mode != null && mode.canExit()) {
			popMode();
			mode = getTopMode();
		}
	}

	/** Set or unset exclusive Mode for the given Mode.
	     *  <p>
	     *  If <var>exclu</var> was <code>true</code> then
	     *  <var>mode</var> is the new exclusive mode <b>only if</b>
	     *  it is already the exclusive mode or there are no
	     *  exclusive mode.
	     *  <p>
	     *  If <var>exclu</var> was <code>false</code> then
	     *  the current exclusive mode was unset <b>only if</b>
	     *  <var>mode</var> was the current exclusive mode.
	     *
	     * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	     * @param mode a reference to the new/old exclusive mode.
	     * @param exclu <code>true</code> if you would like to set
	     *              <var>mode</var> as an exclusive mode,
	     *              otherwise <code>false</code>.
	     */
	public void setExclusiveMode(Mode mode, boolean exclu) {
		if (exclu) {
			if (_exclusiveMode != null) {
				if (mode != _exclusiveMode) {
					throw new MCGEInternalError(
						"try to set an exclusive " + "mode when ModeManager " + "has already one.");
				}
			} else {
				_exclusiveMode = mode;
				_graph.setStatusMsg(mode.instructions());
			}

		} else if (_exclusiveMode == mode) {
			unsetExclusiveMode();
		}
	}

	/** Unset exclusive Mode. */
	public void unsetExclusiveMode() {
		if (_exclusiveMode != null)
			_graph.clearStatus();
		_exclusiveMode = null;
	}
}