package mygale.mcge.mode;
// Copyright (C) 2000  Mahdi Hannoun, Stphane Galland
//
// This program is free software; you can redistribute it and/or modify

import java.awt.*;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.*;
import mygale.mcge.net.INetEdge;
import mygale.mcge.net.INetNode;
import mygale.mcge.undoredo.*;

/** This class implement a special mode that manage the
 *  creation, the suppressio and the move of edge control
 *  point. This mode is normaly set by the ModeSelect we
 *  it detected a drag on an edge.
 *  <p>
 *  <font color="#FF0000"><b>Todo:</b> finish to comment this class.</font>
 *  <p><center>
 *    <img src="doc-files/ModeModifyEdgeCtrlPoint_uml.gif" width="389" height="88" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="171,0,243,28"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="315,59,388,86"  href="../fig/FigEdge.html">
 *    </map>
 *  </center></p>

 */
public class ModeModifyEdgeCtrlPoint extends ModeUndoable {

	/** the modified list of edge control point.
	 */
	protected ArrayList _points;
	/** the index of the current edge control point.
	 */
	protected int _curCtrlPoint;

	/** the figEdge on wich we are modifing control point.
	 */
	protected IFigEdge _figEdge;
	protected IFigNode _sourcePort;
	protected Point _sourcePortPos;
	protected IFigNode _destPort;
	protected Point _destPortPos;

	protected boolean _beginDrag = false;

	protected Rectangle _damagedRect = new Rectangle(0, 0, 0, 0);
	protected Rectangle _previousDamagedRect = new Rectangle(0, 0, 0, 0);

	/** This class is the undo/redo command for the modifcation.
	 *  This class permits to move the point location
	 *
	    * @see mygale.mcge.undoredo.CommandSet
	 */
	protected class UndoModify implements UndoableCmd, RedoableCmd {

		/** It is a new control point in the edge ?
		 */
		protected final boolean _newCtrlPoint;

		/** A reference to the updated figure.
		 */
		protected final IFigEdge _figure;

		/** This is the index of the control point.
		 */
		protected final int _index;

		/** This is the old location of the control point.
		 */
		protected final Point _oldLocation;

		/** This is the new location of the control point.
		 */
		protected final Point _newLocation;

		/** Construct a new UndoModify.
		 *
		 * @param figure a reference to the changed figure.
		 * @param index the index of the control point in the edge.
		 * @param nx new location of the control point.
		 * @param ny new location of the control point.
		 * @param newPts <code>true</code> if the given <var>index</var>
		 *               corresponds to a control point that are not
		 *               in the original edge, otherwise <code>false</code>.
		 */
		public UndoModify(IFigEdge figure, int index, int nx, int ny, boolean newPts) {
			_figure = figure;
			_index = index;
			_oldLocation = new Point(_figure.getCtrlPointAt(index));
			_newLocation = new Point(nx, ny);
			_newCtrlPoint = newPts;
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			if (_newCtrlPoint) {
				_figure.removeCtrlPointAt(_index);
			} else {
				_figure.setCtrlPointAt(_index, _oldLocation);
			}
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			if (_newCtrlPoint) {
				_figure.insertCtrlPointAt(_index, _newLocation);
			} else {
				_figure.setCtrlPointAt(_index, _newLocation);
			}
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

	} /* class UndoModify */

	/** This class is the undo/redo command for the removing
	 *  of control points.
	 *
	    * @see mygale.mcge.undoredo.CommandSet
	 */
	protected class UndoRemove implements UndoableCmd, RedoableCmd {

		/** A reference to the updated figure.
		 */
		protected final IFigEdge _figure;

		/** This is the index of the control point.
		 */
		protected final int _index;

		/** This is the old location of the control point.
		 */
		protected final Point _oldLocation;

		/** Construct a new UndoRemove.
		 *
		 * @param figure a reference to the changed figure.
		 * @param index the index of the control point in the edge.
		 * @param ox old location of the control point.
		 * @param oy old location of the control point.
		 */
		public UndoRemove(IFigEdge figure, int index, int ox, int oy) {
			_figure = figure;
			_index = index;
			_oldLocation = new Point(ox, oy);
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			_figure.insertCtrlPointAt(_index, _oldLocation);
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			_figure.removeCtrlPointAt(_index);
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

	} /* class UndoRemove */

	/** Construct a new ModeModifyEdgeCtrlPoint. The
	     *  {@link ModeManager ModeManager} owner
	     *  should be set before using this object.
	     */
	public ModeModifyEdgeCtrlPoint() {
		super();
		_beginDrag = false;
	}

	////
	// constructors 

	/** Construct a new ModeModifyEdgeCtrlPoint with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModeModifyEdgeCtrlPoint(ModeManager modeManager) {
		super(modeManager);
		_beginDrag = false;
	}

	protected void calcCtrlPointBounds() {
		_previousDamagedRect.setBounds(_damagedRect);
		int size = _points.size();
		Point pt = (Point) _points.get(0);
		_damagedRect.setBounds(pt.x, pt.y, 0, 0);
		for (int i = 1; i < size; i++) {
			pt = (Point) _points.get(i);
			_damagedRect.add(pt);
		}
	}

	protected void calcFinalCtrlPointBounds() {
		calcCtrlPointBounds();
		_damagedRect = _damagedRect.union(_figEdge.getDamagedBounds(getPrefs()));
	}

	/** Change the size of the current workspace. This permits to
	     *  have the new control points into the workspace.
	     */
	protected void changeWorkspaceSize() {
		_modeManager.getGraph().changeWorkspaceSize(_damagedRect.union(_previousDamagedRect));
	}

	/** Clean the current modify edge control point mode.
	     */
	public void cleanMode() {
		_beginDrag = false;
		_damagedRect = _previousDamagedRect.union(_damagedRect);
		if (_figEdge != null) {
			_damagedRect = _damagedRect.union(_figEdge.getDamagedBounds(getPrefs()));
		}
		damaged();
	}

	protected void computeCtrlPoint(MCGEMouseEvent me) {
		if (_figEdge.getMaxAllowedCtrlPoints() == 2) {
			return;
		} else if (_figEdge.isLocked() == true) {
			return;
		} else {
			// compute the port of the fig
			JGraph graph = getModeManager().getGraph();
			INetNode beginPort = ((INetEdge) _figEdge.getNetObject()).getBeginNode();
			INetNode beginNetNode = beginPort;
			_sourcePort = (IFigNode) graph.getFigMapper().getFigFor(beginPort);
			_sourcePortPos = graph.getFigMapper().getFigFor(beginNetNode).getLocation();
			// get the destination port of the edge
			INetNode destPort = ((INetEdge) _figEdge.getNetObject()).getEndNode();
			INetNode destNetNode = destPort;
			_destPort = (IFigNode) getModeManager().getGraph().getFigMapper().getFigFor(destPort);
			_destPortPos = graph.getFigMapper().getFigFor(destNetNode).getLocation();

			// Create a copy of the control points
			_points = new ArrayList();
			Iterator it = _figEdge.getCtrlPoints();
			while (it.hasNext()) {
				_points.add(new Point((Point) it.next()));
			}

			int size = _points.size();
			int x = me.getX();
			int y = me.getY();
			// Am I on a existing control point ?
			for (int i = 1; i < size - 1; i++) {
				Point pt = (Point) _points.get(i);
				if (Point2D.distance(pt.x, pt.y, x, y) < 6.0) {
					// Yes ! Begin to move the control point
					_curCtrlPoint = i;
					_beginDrag = true;
					break;
				}
			}
			if (_beginDrag == false) {
				// Am I trying to create an control point ?
				if (size < _figEdge.getMaxAllowedCtrlPoints()) {
					for (int i = 0; i < size - 1; i++) {
						Point pt1 = (Point) _points.get(i);
						Point pt2 = (Point) _points.get(i + 1);
						if (Line2D
							.ptLineDist(
								pt1.x,
								pt1.y,
								pt2.x,
								pt2.y,
								x,
								y)
							< 6.0) {
							// Yes, create the control point
							_points.add(i + 1, new Point(x, y));
							_curCtrlPoint = i + 1;
							_beginDrag = true;
							updateFirstAndLastCtrlPoint();
							break;
						}
					}
				}
			}
		}
	}

	////
	// utility methods

	/** This method call the damaged method of the graph that own
	 *  the mode manager of this mode to redraw the modified
	 *  area.
	 */
	protected void damaged() {
		_modeManager.getGraph().damaged(_damagedRect.union(_previousDamagedRect));
	}

	////
	// user feedback methods

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		return "Drag from a edge to move/create a control point. Type <escape> to cancel.";
	}

	/** Invoked when a mouse-clicked event was received.
	     *  <p>
	     *  Close this mode.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		if (isExclusive())
			return;
		done();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  Close this mode.
	     *
	     * @param me description of the mouse-event.
	     */
	public void mcgeMouseDoubleClick(MCGEMouseEvent me) {
		if (isExclusive())
			return;
		done();
	}

	/** Invoked when a mouse-dragged event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive()) || (!_beginDrag)) {
			return;
		}

		int x = me.getX();
		int y = me.getY();

		Point pt = (Point) _points.get(_curCtrlPoint);
		pt.setLocation(x, y);
		updateFirstAndLastCtrlPoint();
		calcCtrlPointBounds();
		damaged();
		me.consume();
	}

	/** Invoked when a mouse-entered event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		me.consume();
	}

	/** Invoked when a mouse-moved event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
		me.consume();
	}

	////
	// event handlers

	/** Invoked when a mouse-pressed event was received.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive())) {
			return;
		}

		if (me.getFigure() instanceof IFigEdge) {
			_figEdge = (IFigEdge) me.getFigure();
			computeCtrlPoint(me);
			if (_beginDrag == true) {
				calcCtrlPointBounds();
				setExclusive(true);
				damaged();
				return;
			}
		}
		done();
	}

	/** Invoked when a mouse-released event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive()) || (!_beginDrag)) {
			return;
		}

		int x = me.getX();
		int y = me.getY();

		Point pt = (Point) _points.get(_curCtrlPoint);
		// Prepare the undoing command
		CommandSet undoCmds = new CommandSet();
		undoCmds.add(
			new UndoModify(_figEdge, _curCtrlPoint, x, y, (_points.size() > _figEdge.getCtrlPointSize())));
		// Do the move
		pt.setLocation(x, y);
		tryToRemoveTheCurrentCtrlPoint(undoCmds);
		updateFirstAndLastCtrlPoint();
		calcFinalCtrlPointBounds();

		_figEdge.setCtrlPoints(_points.iterator());

		_beginDrag = false;
		setExclusive(false);
		damaged();
		changeWorkspaceSize();
		done();
		me.consume();
		// Add the undoing command
		addUndoCmd(undoCmds);
	}

	////
	// painting methods

	/** Paint the line beetwen the control points of the edge.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
		if (_beginDrag == true) {
			g2d.setColor(getPrefs().getShadowColor());
			int size = _points.size();
			for (int i = 0; i < (size - 1); i++) {
				Point begin = (Point) _points.get(i);
				Point end = (Point) _points.get(i + 1);
				g2d.drawLine(begin.x, begin.y, end.x, end.y);
			}
		}
	}

	/** Try to remove control points that are in the same location.
	     *
	     * @param undoCmds a reference to a undoing commands in which
	     *                 this method store the undo actions.
	     */
	protected void tryToRemoveTheCurrentCtrlPoint(CommandSet undoCmds) {

		Point curPt = (Point) _points.get(_curCtrlPoint);
		if (AbstractFigEdge.tryToRemoveTheCurrentCtrlPoint(_figEdge, curPt, _points)) {
			// Save this action into the undo list
			undoCmds.add(new UndoRemove(_figEdge, _curCtrlPoint, curPt.x, curPt.y));
		}
	}

	// internal methods

	protected void updateFirstAndLastCtrlPoint() {
		int size = _points.size();
		Point pt1 = (Point) _points.get(0);
		Point pt2 = (Point) _points.get(1);
		Point pt2bis = new Point(pt2);
		pt2bis.translate(-_sourcePortPos.x, -_sourcePortPos.y);
		if (!_sourcePort.hit(pt2bis, getPrefs())) {
			pt1.setLocation(_sourcePort.intersectWith(pt2, 0, 0));
		}

		Point pt3 = (Point) _points.get(size - 1);
		Point pt4 = (Point) _points.get(size - 2);
		Point pt4bis = new Point(pt4);
		pt4bis.translate(-_destPortPos.x, -_destPortPos.y);
		if (!_destPort.hit(pt4bis, getPrefs())) {
			pt3.setLocation(_destPort.intersectWith(pt4, 0, 0));
		}
	}
}