package mygale.mcge.mode;

import java.awt.*;
import java.util.*;

import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.*;
import mygale.mcge.net.*;
import mygale.mcge.sel.Selection;
import mygale.mcge.throwable.MCGEInternalError;
import mygale.mcge.undoredo.*;

/** This class implements a move mode, that intercept the
 *  mouse event to drag figures on the drawing area.
 *  <p><center>
 *    <img src="doc-files/ModeMove_uml.gif" width="363" height="143" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/Fig.html">
 *      <area shape="Rect" coords="77,114,176,141"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="212,114,311,141"  href="../sel/Selection.html">
 *    </map>
 *  </center></p>
 */
public class ModeMove extends ModeUndoable {

	/** The x coordinate of the beginning point when moving figs.
	 */
	protected int _xBegin;
	/** The y coordinate of the beginning point when moving figs.
	 */
	protected int _yBegin;

	/** The last x coordinate of the last move mouse event.
	 */
	protected int _lastX;
	/** The last y coordinate of the last move mouse event.
	 */
	protected int _lastY;

	/** The amount of x move sine the last move mouse event.
	 */
	protected int _transX = 0;
	/** The amount of y move sine the last move mouse event.
	 */
	protected int _transY = 0;

	/** The area to be redrawed during the move.
	 */
	protected Rectangle _rectMovingArea = new Rectangle(0, 0, 0, 0);

	/** The area of the edges to be redrawed during the move.
	 */
	protected Rectangle _rectEdgeMovingArea = new Rectangle(0, 0, 0, 0);

	/** The area of the initial moving area.
	 */
	protected Rectangle _rectInitialMovingArea = new Rectangle(0, 0, 0, 0);

	/** The area of the initial edge moving area.
	 */
	protected Rectangle _rectEdgeInitialMovingArea = new Rectangle(0, 0, 0, 0);

	/** A tempororay rectangle to compute the current drawing area.
	 */
	protected Rectangle _rectDrawingArea = new Rectangle(0, 0, 0, 0);

	/** A temporary rectangle to compute the current edges drawing area.
	 */
	protected Rectangle _rectEdgeDrawingArea = new Rectangle(0, 0, 0, 0);

	/** <code>true</code> if the moving rectangle should be drawed.
	 */
	protected boolean _drawMovingSelection = false;

	/** A the list of edge to be translated directly.
	 */
	protected Vector _edgeToTranslate = null;
	protected Vector _edgeWithSourceMoved = null;
	protected Vector _edgeWithDestMoved = null;
	
	
	protected IFigEdge _implicitlySelectedEdge = null;

	/** This class is the undo/redo command for the moving.
	 *  This class permits to move a figure.
	 *
	    * @see mygale.mcge.undoredo.CommandSet
	 */
	protected class UndoSubMove implements UndoableCmd, RedoableCmd {

		/** A reference to the moved figure.
		 */
		protected IFig _figure;

		/** Old location of the moved figure.
		 */
		protected Point _oldLocation;

		/** New location of the moved figure.
		 */
		protected Point _newLocation;

		/** Construct a new UndoSubMove.
		 *
		 * @param figure a reference to the moved figure.
		 * @param transX the translation along the x axis.
		 * @param transY the translation along the y axis.
		 */
		public UndoSubMove(IFig figure, int transX, int transY) {
			_figure = figure;
			_oldLocation = new Point(figure.getLocation());
			_newLocation = new Point(_oldLocation.x + transX, _oldLocation.y + transY);
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			_figure.setLocation(_oldLocation.x, _oldLocation.y);
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			Rectangle damagedBounds = new Rectangle(_figure.getBounds());
			_figure.setLocation(_newLocation.x, _newLocation.y);
			damagedBounds.add(_figure.getBounds());
			_modeManager.getGraph().damaged(damagedBounds);
		}

	} /* class UndoSubMove */

	/** This class is the undo/redo command for the moving.
	 *  This class permits to update edge locations.
	 *
	    * @see mygale.mcge.undoredo.CommandSet
	 */
	protected class UndoSubMoveEdges implements UndoableCmd, RedoableCmd {

		/** A set of edges to update.
		 */
		protected Vector _edges = new Vector();

		/** <code>true</code> if this command
		 *  will be used for redoing only,
		 *  <code>false</code> if this command
		 *  will be used for undoing only.
		 */
		protected boolean redoOnly;

		/** Construct a new UndoSubMoveEdges.
		 *
		 * @param edges1 a set of vectors.
		 * @param edges2 a set of vectors.
		 * @param edges3 a set of vectors.
		 * @param redoable <code>true</code> if this command
		 *                 will be used for redoing only,
		 *                 <code>false</code> if this command
		 *                 will be used for undoing only.
		 */
		public UndoSubMoveEdges(Vector edges1, Vector edges2, Vector edges3, boolean redoable) {
			_edges.addAll(edges1);
			_edges.addAll(edges2);
			_edges.addAll(edges3);
			redoOnly = redoable;
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 * @see mygale.mcge.undoredo.UndoableCmd#undoIt()
		 */
		public void undoIt() {
			if (!redoOnly) {
				updateEdges();
			}
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 */
		public void redoIt() {
			if (redoOnly) {
				updateEdges();
			}
		}

		/** Update egdes.
		 *
		 * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
		 */
		protected void updateEdges() {
			Enumeration enum = _edges.elements();
			Rectangle damagedBounds = null;
			while (enum.hasMoreElements()) {
				IFig fig = (IFig) enum.nextElement();
				if (fig instanceof IFigEdge) {
					if (damagedBounds == null) {
						damagedBounds = new Rectangle(fig.getBounds());
					} else {
						damagedBounds.add(fig.getBounds());
					}
					((IFigEdge) fig).updateEndsLocation();
					damagedBounds.add(fig.getBounds());
				} else {
					throw new MCGEInternalError(
						"Try to update an edge, but " + "the given Object was not an edge !");
				}
			}
			_modeManager.getGraph().damaged(damagedBounds);
		}

	} /* class UndoSubMoveEdges */

	/** Construct a new ModeMove. The
	     *  {@link ModeManager ModeManager} parent should be
	     *  set before using this object.
	     */
	public ModeMove() {
	}

	////
	// constructors 

	/** Construct a new ModeMode with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModeMove(ModeManager modeManager) {
		super(modeManager);
	}

	/** Ask to this mode to return into a stable state.
	     */
	public void cleanMode() {

		_drawMovingSelection = false;
		_rectDrawingArea.setBounds(_rectMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectInitialMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectEdgeMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectEdgeInitialMovingArea);
		_modeManager.getGraph().damaged(_rectDrawingArea);
	}

	////
	// user feedback methods

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		return "Drag an object (or selection) to move it. Type <escape> to cancel.";
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  This method move the shadows.
	     *
	     * @param me description of the mouse-event.
	     */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		// get the actual mouse coordinate.
		int x = me.getX();
		int y = me.getY();

		// compute the translation move from the last move point
		int xTrans = x - _lastX;
		int yTrans = y - _lastY;

		// compute the translation move from the beginning point
		// of the move
		_transX = x - _xBegin;
		_transY = y - _yBegin;

		_rectDrawingArea.setBounds(_rectMovingArea);
		_rectDrawingArea.translate(xTrans, yTrans);
		_rectDrawingArea = _rectDrawingArea.union(_rectMovingArea);

		_rectEdgeDrawingArea.setBounds(_rectEdgeMovingArea);
		Rectangle _tmpEdgeRect = null;
		if (_implicitlySelectedEdge != null) {
			_tmpEdgeRect = new Rectangle(_implicitlySelectedEdge.getBounds());

			for (int i = 0; i < _edgeToTranslate.size(); i++) {
				Rectangle bounds = new Rectangle(((IFigEdge) _edgeToTranslate.elementAt(i)).getBounds());
				bounds.translate(_transX, _transY);
				_tmpEdgeRect = _tmpEdgeRect.union(bounds);
			}
			
			for (int i = 0; i < _edgeWithSourceMoved.size(); i++) {
				IFigEdge tmpEdge = (IFigEdge) _edgeWithSourceMoved.elementAt(i);
				_tmpEdgeRect = _tmpEdgeRect.union(tmpEdge.getShadowBounds(_transX, _transY));
			}
			
			for (int i = 0; i < _edgeWithDestMoved.size(); i++) {
				IFigEdge tmpEdge = (IFigEdge) _edgeWithDestMoved.elementAt(i);
				_tmpEdgeRect = _tmpEdgeRect.union(tmpEdge.getShadowBounds(_transX, _transY));
			}
		}

		if (_tmpEdgeRect != null) {
			_rectEdgeDrawingArea = _rectEdgeDrawingArea.union(_tmpEdgeRect);
		}

		if (_tmpEdgeRect != null) {
			_modeManager.getGraph().damaged(_rectDrawingArea.union(_rectEdgeDrawingArea));
			_rectEdgeMovingArea.setBounds(_tmpEdgeRect);
		} else {
			_modeManager.getGraph().damaged(_rectDrawingArea);
		}

		_rectMovingArea.translate(xTrans, yTrans);

		updateLastMovePoint(x, y); // setzt _yBegin und _xBegin
		_modeManager.getGraph().scrollToShow(x, y);
		me.consume();
	}

	/** Invoked when a mouse-entered event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  This method draw the shadows of selected figs and
	 *  initiate attributes.
	 *
	 * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		IFigLayered fig = me.getFigure();

		if (fig != null) {
			setBeginMovePoint(me.getX(), me.getY());
			Selection sel = _modeManager.getGraph().getSelection();
			if (fig.isSelected() == false) {
				// if the fig is not selected, deselect all the selected
				// fig and select this figure
				sel.deselectAll();
				if ((fig.isMovable() == false) || (fig.isLocked() == true)) {
					done();
					return;
				}
				sel.select(fig);
			} else {
				Vector figToDesselect = new Vector();
				Iterator it = sel.getAllFigs();
				while (it.hasNext()) {
					IFigLayered tmpFig = (IFigLayered) it.next();
					if ((tmpFig.isMovable() == false) || (tmpFig.isLocked() == true)) {
						figToDesselect.add(tmpFig);
					}
				}
				Enumeration enum = figToDesselect.elements();
				while (enum.hasMoreElements()) {
					IFigLayered tmpFig = (IFigLayered) enum.nextElement();
					sel.deselect(tmpFig);
				}
				if (sel.size() == 0) {
					done();
					return;
				}
			}

			_edgeToTranslate = new Vector();
			_edgeWithSourceMoved = new Vector();
			_edgeWithDestMoved = new Vector();
			_implicitlySelectedEdge = null;

			Iterator it = sel.getAllFigs();
			while (it.hasNext()) {
				
				IFigLayered tmpFig = (IFigLayered) it.next();
				if (tmpFig instanceof IFigNode) {
					
					Iterator it2 = ((INetNode) ((IFigNode) tmpFig).getNetObject()).getNodeEdges();
					while (it2.hasNext()) {
						INetEdge netEdge = (INetEdge) it2.next();
						
						IFigEdge figEdge =
							(IFigEdge) getModeManager().getGraph().getFigMapper().getFigFor(netEdge);

						IFigNode figNodeBegin =
							((IFigNode) getModeManager()
								.getGraph()
								.getFigMapper()
								.getFigFor(((INetEdge) figEdge.getNetObject()).getBeginNode()));
						IFigNode figNodeEnd =
							((IFigNode) getModeManager()
								.getGraph()
								.getFigMapper()
								.getFigFor(((INetEdge) figEdge.getNetObject()).getEndNode()));

						if (sel.contains(figNodeBegin) && sel.contains(figNodeEnd)) {
							if (!_edgeToTranslate.contains(figEdge)) {
								_edgeToTranslate.add(figEdge);
								figEdge.beginShadowMoveMode(IFigEdge.BOTH_PORTS_SHADOW_MOVE);
								_implicitlySelectedEdge = figEdge;
							}

						} else if (sel.contains(figNodeBegin)) {
							if (!_edgeWithSourceMoved.contains(figEdge)) {
								_edgeWithSourceMoved.add(figEdge);
								figEdge.beginShadowMoveMode(IFigEdge.SRC_PORTS_SHADOW_MOVE);
								_implicitlySelectedEdge = figEdge;
							}

						} else if (sel.contains(figNodeEnd)) {
							if (!_edgeWithDestMoved.contains(figEdge)) {
								_edgeWithDestMoved.add(figEdge);
								figEdge.beginShadowMoveMode(IFigEdge.DST_PORTS_SHADOW_MOVE);
								_implicitlySelectedEdge = figEdge;
							}

						} else {
							throw new MCGEInternalError("figedge....");
						}

					}
				}
			}

			_rectInitialMovingArea = sel.getBounds();
			if (_implicitlySelectedEdge != null) {
				_rectEdgeInitialMovingArea = _implicitlySelectedEdge.getBounds();
				for (int i = 0; i < _edgeToTranslate.size(); i++) {
					_rectEdgeInitialMovingArea =
						_rectEdgeInitialMovingArea.union(
							((IFigEdge) _edgeToTranslate.elementAt(i)).getBounds());
				}
				for (int i = 0; i < _edgeWithSourceMoved.size(); i++) {
					_rectEdgeInitialMovingArea =
						_rectEdgeInitialMovingArea.union(
							((IFigEdge) _edgeWithSourceMoved.elementAt(i)).getShadowBounds(0, 0));
				}
				for (int i = 0; i < _edgeWithDestMoved.size(); i++) {
					_rectEdgeInitialMovingArea =
						_rectEdgeInitialMovingArea.union(
							((IFigEdge) _edgeWithDestMoved.elementAt(i)).getShadowBounds(0, 0));
				}
			}
			
			_rectMovingArea.setBounds(_rectInitialMovingArea);
			_rectEdgeMovingArea.setBounds(_rectEdgeInitialMovingArea);
			_drawMovingSelection = true;

			//_modeManager.getGraph().damageAll() ;
			if (_implicitlySelectedEdge != null) {
				_modeManager.getGraph().damaged(_rectInitialMovingArea.union(_rectEdgeInitialMovingArea));
			} else {
				_modeManager.getGraph().damaged(_rectInitialMovingArea);
			}
			setExclusive(true);
			me.consume();
		}
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  This method changes the bounds of selected figs
	     *  by using the current shadow bounds.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		int x = me.getX();
		int y = me.getY();

		// compute the translation point from the begining
		// of the move
		_transX = x - _xBegin;
		_transY = y - _yBegin;

		// If the translations are null, do nothing
		if (_transX == 0 && _transY == 0) {
			done();
			me.consume();
			return;
		}

		// Create the set of undoable commands.
		CommandSet undoCmds = new CommandSet();

		undoCmds.add(new UndoSubMoveEdges(_edgeToTranslate, _edgeWithSourceMoved, _edgeWithDestMoved, false));

		// compute the translation move from the last move point
		int xTrans = x - _lastX;
		int yTrans = y - _lastY;

		_drawMovingSelection = false;

		_rectDrawingArea.setBounds(_rectMovingArea);
		_rectDrawingArea.translate(xTrans, yTrans);
		_rectDrawingArea = _rectDrawingArea.union(_rectMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectInitialMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectEdgeMovingArea);
		_rectDrawingArea = _rectDrawingArea.union(_rectEdgeInitialMovingArea);

		Selection sel = _modeManager.getGraph().getSelection();
		Iterator it = sel.getAllFigs();
		while (it.hasNext()) {
			IFig fig = (IFig) it.next();
			// Save this move in the undoable command list
			undoCmds.add(new UndoSubMove(fig, _transX, _transY));
			// do the move
			fig.translate(_transX, _transY);
		}

		for (int i = 0; i < _edgeToTranslate.size(); i++) {
			IFigEdge figEdge = (IFigEdge) _edgeToTranslate.elementAt(i);
			// Save this move in the undoable command list
			undoCmds.add(new UndoSubMove(figEdge, _transX, _transY));
			// do the movement
			figEdge.endShadowMoveMode(_transX, _transY);
			_rectDrawingArea = _rectDrawingArea.union(figEdge.getBounds());

		}
		for (int i = 0; i < _edgeWithSourceMoved.size(); i++) {
			IFigEdge figEdge = (IFigEdge) _edgeWithSourceMoved.elementAt(i);
			figEdge.endShadowMoveMode(_transX, _transY);
			tryToRemoveTheCurrentCtrlPoint(figEdge, 1);
			_rectDrawingArea = _rectDrawingArea.union(figEdge.getBounds());
		}
		for (int i = 0; i < _edgeWithDestMoved.size(); i++) {
			IFigEdge figEdge = (IFigEdge) _edgeWithDestMoved.elementAt(i);
			figEdge.endShadowMoveMode(_transX, _transY);
			tryToRemoveTheCurrentCtrlPoint(figEdge, figEdge.getCtrlPointSize() - 2);
			_rectDrawingArea = _rectDrawingArea.union(figEdge.getBounds());
		}

		// add selection handle size to rect

		int handleSize = getPrefs().getHandleSize();

		_rectDrawingArea.x -= handleSize;
		_rectDrawingArea.y -= handleSize;
		_rectDrawingArea.width += 2 * handleSize;
		_rectDrawingArea.height += 2 * handleSize;

		// Save the moved edges into a undoable command
		undoCmds.add(new UndoSubMoveEdges(_edgeToTranslate, _edgeWithSourceMoved, _edgeWithDestMoved, true));

		_edgeToTranslate = null;
		_edgeWithSourceMoved = null;
		_edgeWithDestMoved = null;

		_modeManager.getGraph().damaged(_rectDrawingArea);
		_modeManager.getGraph().changeWorkspaceSize(_rectDrawingArea);
		done();
		me.consume();
		// Add the undoable commands
		addUndoCmd(undoCmds);
	}

	////
	// utility methods

	/** Initiate the movement point.
	 *
	 * @param x current position of the point.
	 * @param y current position of the point.
	 */
	protected void setBeginMovePoint(int x, int y) {
		_xBegin = _lastX = x;
		_yBegin = _lastY = y;
		_transX = _transY = 0;
	}

	/** Update the movement point.
	     *
	     * @param x current position of the point.
	     * @param y current position of the point.
	     */
	protected void updateLastMovePoint(int x, int y) {
		_lastX = x;
		_lastY = y;
	}

	/** Paint the shadow objects if appropriate.
	 *
	 * @param g2d the graphic context.
	 */
	public void paint(Graphics2D g2d) {
		if (_drawMovingSelection) {

			Selection sel = _modeManager.getGraph().getSelection();
			Iterator it = sel.getAllFigs();
			while (it.hasNext()) {
				IFigLayered layeredFig = (IFigLayered) it.next();
				layeredFig.paintShadow(g2d, _transX, _transY, getPrefs());
			}
			
			// die nachfolgenden Schleifen scheinen verantwortlich 
			// fr das Zeichnen der beteiligten Kanten zu sein
			for (int i = 0; i < _edgeToTranslate.size(); i++) {
				IFigEdge edgeFig = (IFigEdge) _edgeToTranslate.elementAt(i);
				edgeFig.paintShadow(g2d, _transX, _transY, getPrefs());
			}
			
			for (int i = 0; i < _edgeWithSourceMoved.size(); i++) {
				IFigEdge edgeFig = (IFigEdge) _edgeWithSourceMoved.elementAt(i);
				edgeFig.paintShadow(g2d, _transX, _transY, getPrefs());
			}
			
			for (int i = 0; i < _edgeWithDestMoved.size(); i++) {
				IFigEdge edgeFig = (IFigEdge) _edgeWithDestMoved.elementAt(i);
				edgeFig.paintShadow(g2d, _transX, _transY, getPrefs());
			}
		}

	}

	/** Try to remove control points that are in the same location.
	     *
	     * @param undoCmds a reference to a undoing commands in which
	     *                 this method store the undo actions.
	     */
	protected void tryToRemoveTheCurrentCtrlPoint(IFigEdge edge, int CtrlPointIndex) {
		// TODO: undo
		edge.tryToRemoveTheCurrentCtrlPoint(edge.getCtrlPointAt(CtrlPointIndex));
	}
}