package mygale.mcge.mode;

import javax.swing.JPopupMenu;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.IFigLayered;
import mygale.mcge.sel.Selection;

/** This class implements a Mode that permits to show a popup
 *  menu when the right button of the mouse was clicked.
 *  <p>
 *  If the user click with the right button on a
 *  {@link Fig figure}, this mode open a
 *  {@link mygale.mcge.fig.Fig#getPopupMenu(boolean) popup menu}
 *  given by the figure.
 *  <p>
 *  If the user click with the right button on the
 *  background of the {@link mygale.mcge.core.JGraph editor},
 *  this mode open a
 *  {@link mygale.mcge.core.JGraph#getPopupMenu() popup menu}
 *  given by the editor.
 *  <p><center>
 *    <img src="doc-files/ModePopup_uml.gif" width="363" height="143" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="Mode.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/Fig.html">
 *      <area shape="Rect" coords="77,114,176,141"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="212,114,311,141"  href="../sel/Selection.html">
 *    </map>
 *  </center></p>
 *
 */
public class ModePopup extends Mode {

	/** Construct a new ModePopup. The
	     *  {@link ModeManager ModeManager} parent
	     *  should be set before using this object.
	     */
	public ModePopup() {
	}

	/** Construct a new ModePopup with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModePopup(ModeManager modeManager) {
		super(modeManager);
	}

	/** Always <code>false</code> because I never want
	 *  to get out of popup mode.
	 *
	@link ModeManager ModeManager} stack,
	 *         otherwise <code>false</code>.
	 * @return <code>true</code> if this mode can be dropped from
	 *         the { */
	public boolean canExit() {
		return false;
	}

	/** Invoked when a mouse-clicked event was received.
	 *  <p>
	 *  If the user click with the right button on a
	 *  {@link Fig figure}, this mode open a popup menu
	 *  given by the figure.
	 *  <p>
	 *  If the user click with the right button on the
	 *  background of the {@link mygale.mcge.core.JGraph editor},
	 *  this mode open a popup menu given by the editor.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.fig.Fig#getPopupMenu(boolean)
	 * @see mygale.mcge.core.JGraph#getPopupMenu()
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		if (!me.isOnlyRightButtonDown()) {
			return;
		}

		IFigLayered fig = me.getFigure();
		JPopupMenu popupMenu = null;
		JGraph graph = getModeManager().getGraph();

		if (fig != null) {
			// Get the Popup menu of the figure
			Selection sl = graph.getSelection();
			if (!sl.contains(fig)) {
				sl.deselectAll();
				sl.select(fig);
			}
			popupMenu = fig.getPopupMenu(graph, sl.size() == 1);
		} else {
			// Get the Popup menu of the editor
			popupMenu = graph.getPopupMenu();
		}

		// display the popup
		if (popupMenu != null) {
			me.consume();
			popupMenu.setVisible(true);
			popupMenu.show(graph, me.getRealX(), me.getRealY());
		}
	}
}
