package mygale.mcge.mode;

import java.awt.*;
import java.util.Iterator;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.*;
import mygale.mcge.net.*;
import mygale.mcge.sel.Selection;
import mygale.mcge.undoredo.*;
import mygale.mcge.util.prefs.Prefs;

/** This class implements a Mode that interperts user inputs as
 *  reshaping one Fig.
 *  <p>
 *  A reshape can be do if a {@link mygale.mcge.fig.Fig figure}
 *  authorizes this mode to reshape. A reshape is made by
 *  dragging reshaping handles.
 *  <p><center>
 *    <img src="doc-files/ModeReshape_uml.gif" width="363" height="143" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="ModeUndoable.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/Fig.html">
 *      <area shape="Rect" coords="77,114,176,141"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="212,114,311,141"  href="../sel/Selection.html">
 *    </map>
 *  </center></p>
 *
 * @see mygale.mcge.fig.Fig#getReshapeDirection(int,int)
 */
public class ModeReshape extends ModeUndoable {

	/** The direction of the reshaping.
	 * @see mygale.mcge.fig.Fig#getReshapeDirection(int,int)
	 */
	protected int _direction = IFigLayered.RESHAPE_NONE;

	/** This is the {@link mygale.mcge.fig.Fig Fig}
	 *  that will be reshaped.
	 */
	protected IFigLayered _figToReshape = null;

	/** This is the x coordonate of the initial point of the selection rectangle.
	 */
	protected int _xAnchor;
	/** This is the y coordonate of the initial point of the selection rectangle.
	 */
	protected int _yAnchor;

	/** This is the last x coordonate of the mouse.
	 */
	protected int _lastDraggedX;
	/** This is the last y coordonate of the mouse.
	 */
	protected int _lastDraggedY;

	/** This is the rectangle used to show the reshape area.
	 */
	protected Rectangle _reshapeRect = new Rectangle(0, 0, 0, 0);

	/** Say if the reshape rectangle could be draw.
	 */
	protected boolean _showReshapeRect = false;

	/** Say if the cursor has changed.
	 */
	protected boolean _cursorChanged = false;

	/** This class is the undo/redo command for the reshaping
	 *
		 */
	protected class UndoReshape implements UndoableCmd, RedoableCmd {

		/** This is a reference to the reshaped figure.
		 */
		protected final IFigLayered _figure;

		/** This is the old bounds of the reshaped figure.
		 */
		protected final Rectangle _oldBounds;

		/** This is the new bounds of the reshaped figure.
		 */
		protected final Rectangle _newBounds;

		/** Construct a new UndoReshape.
		 *
		 * @param figure IFigLayered
		 * @param newBounds the new bounds of this figure.
		 */
		public UndoReshape(IFigLayered figure, Rectangle newBounds) {
			_figure = figure;
			_oldBounds = new Rectangle(figure.getBounds());
			_newBounds = new Rectangle(newBounds);
		}

		/** This method was called by the undo command.
		 *  It must undo something.
		 */
		public void undoIt() {
			reshapeNow(_modeManager.getGraph(), _figure, _oldBounds);
		}

		/** This method was called by the redo command.
		 *  It must redo something.
		 * @see mygale.mcge.undoredo.RedoableCmd#redoIt()
		 */
		public void redoIt() {
			reshapeNow(_modeManager.getGraph(), _figure, _newBounds);
		}

	} /* class UndoReshape */

	/** Construct a new ModeReshape. The
	     *  {@link ModeManager ModeManager} should be
	     *  set before using this object.
	     */
	public ModeReshape() {
	}

	/** Construct a new ModeReshape with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModeReshape(ModeManager modeManager) {
		super(modeManager);
	}

	//

	/** Always <code>false</code> because I never want
	 *  to get out of this mode.
	 *
	@link ModeManager ModeManager} stack,
	 *         otherwise <code>false</code>.
	 * @return always <code>true</code> if this mode can be dropped
	 *         from the { */
	public boolean canExit() {
		return false;
	}

	/** Set the mouse cursor.
	     *  The mouse cursor was determined from the
	     *  given position of the mouse.
	     *
	     * @param x x coordinate of the mouse.
	     * @param y y coordinate of the mouse.
	     * @return <code>true</code> if the cursor was changed,
	     *         otherwise <code>false</code>.
	     */
	protected boolean changeCursor(int x, int y) {
		if ((!isExclusive()) && (getReshapeDirection(x, y) != IFigLayered.RESHAPE_NONE)) {
			if (_figToReshape != null && (!_figToReshape.isLocked())) {
				setReshapeCursor();
				return true;
			}
		}
		restoreCursor();
		return false;
	}

	/** Ask to this mode to return into a stable state.
	     */
	public void cleanMode() {
		_showReshapeRect = false;
		_modeManager.getGraph().damaged(_reshapeRect);
		changeCursor(_lastDraggedX, _lastDraggedY);
	}

	/** This method prepare the reshape rect when the mode need to stop.
	     *
	     * @param x the current position of the mouse.
	     * @param y the current position of the mouse.
	     */
	void endReshapeRect(int x, int y) {
		updateReshapeRect(x, y);
		_showReshapeRect = false;
	}

	/** Reply the reshape direction for the selected
	 *  {@link mygale.mcge.fig.Fig Fig}.
	 *  <p>
	 *  This method ask to the figure what is the
	 *  reshape direction for it if the mouse are
	 *  at the given position.
	 *  <p>
	 *  We assume that this method set the attributes
	 *  <var>_figToReshape</var> and
	 *  <var>_direction</var>.
	 *
	 * @param x x coordinate of the mouse.
	 * @param y y coordinate of the mouse.
	 * @return an integer that correspond to the reshape direction
	 *         when the mouse are on the given point.
	 */
	protected int getReshapeDirection(int x, int y) {
		Selection sel = _modeManager.getGraph().getSelection();

		if ((sel.size() == 1)) {
			IFigLayered fig = (IFigLayered)sel.getAllFigs().next();
			if (fig.isReshapable()) {
				_direction = fig.getReshapeDirection(getPrefs(), x, y);
				if (_direction != IFigLayered.RESHAPE_NONE) {
					_figToReshape = fig;
					return _direction;
				}
			}
		}
		_figToReshape = null;
		return (_direction = IFigLayered.RESHAPE_NONE);
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		if (isExclusive())
			return "Drag to resize the object. Type <escape> to cancel.";
		else
			return super.instructions();
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  If the user move the mouse with the button
	     *  pressed, then the mode change the reshaping
	     *  rectangle.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((me.isOnlyLeftButtonDown()) && (isExclusive()) && (_direction != IFigLayered.RESHAPE_NONE)) {
			_lastDraggedX = me.getX();
			_lastDraggedY = me.getY();
			setPredefinedCursor(Cursor.HAND_CURSOR);
			_modeManager.getGraph().damaged(_reshapeRect);
			updateReshapeRect(_lastDraggedX, _lastDraggedY);
			_modeManager.getGraph().damaged(_reshapeRect);
			me.consume();
		}
	}

	/** Invoked when a mouse-entered event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-moved event was received.
	     *  <p>
	     *  If the user pass the mouse on an reshaping handle,
	     *  the cursor was get from a
	     *  {@link mygale.mcge.fig.Fig figure}.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseMoved(MCGEMouseEvent)
	 */
	public void mcgeMouseMoved(MCGEMouseEvent me) {
		if (changeCursor(me.getX(), me.getY())) {
			me.consume();
		}
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  If the user pressed the mouse button on an
	 *  reshaping handle, then the mode begins to
	 *  reshape the figure.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	 */
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((me.isOnlyLeftButtonDown()) && (!isExclusive()) && (_direction != IFigLayered.RESHAPE_NONE)) {
			_lastDraggedX = me.getX();
			_lastDraggedY = me.getY();
			startReshapeRect();
			setExclusive(true);
			me.consume();
		}
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  If the user release the mouse button,
	     *  then the mode changes the bounds of the figure
	     *  by using the reshaping rectangle.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((me.isOnlyLeftButtonDown()) && (isExclusive()) && (_direction != IFigLayered.RESHAPE_NONE)) {
			endReshapeRect(me.getX(), me.getY());

			// Save the change into the undo list
			addUndoCmd(new UndoReshape(_figToReshape, _reshapeRect));

			// Reshape now
			reshapeNow(_modeManager.getGraph(), _figToReshape, _reshapeRect);
			setExclusive(false);
			changeCursor(me.getX(), me.getY());
			me.consume();
		}
	}

	/** Paint the selection rectangle if appropriate.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
		if (_showReshapeRect) {
			g2d.setColor(getPrefs().getShadowColor());
			g2d.drawRect(_reshapeRect.x, _reshapeRect.y, _reshapeRect.width, _reshapeRect.height);
		}
	}

	/** Reshape the figure and update the drawing area.
	     *
	     * @param a reference to the current graph.
	     * @param a reference to the reshaped figure
	     * @param bounds the new bounds.
	     */
	protected static void reshapeNow(JGraph graph, IFigLayered figure, Rectangle bounds) {
		Prefs prefs = graph.getPrefs();

		Rectangle damagedRect = new Rectangle(figure.getDamagedBounds(prefs));
		figure.setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
		damagedRect.add(bounds);
		if (figure instanceof IFigNode) {
			Iterator it = ((INetNode) ((IFigNode)figure).getNetObject()).getNodeEdges();
			while (it.hasNext()) {
				INetEdge netEdge = (INetEdge)it.next();
				IFigEdge figEdge = (IFigEdge)graph.getFigMapper().getFigFor(netEdge);
				damagedRect.add(figEdge.getDamagedBounds(prefs));
				figEdge.updateEndsLocation();
				damagedRect.add(figEdge.getDamagedBounds(prefs));
			}
		}
		graph.damaged(damagedRect);
	}

	/** Set the mouse cursor to the default.
	     */
	protected void restoreCursor() {
		if (_cursorChanged) {
			setCursor(getInitialCursor());
			_cursorChanged = false;
		}
	}

	/** This method change the cursor to be compliant with this mode.
	     *  <p>
	     *  We assume that the attribute <var>_direction</var> was
	     *  correctly set before the call to this method.
	     */
	protected void setReshapeCursor() {
		_cursorChanged = true;
		switch (_direction) {
			case IFigLayered.RESHAPE_NORTH :
				setPredefinedCursor(Cursor.N_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_NORTHEAST :
				setPredefinedCursor(Cursor.NE_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_EAST :
				setPredefinedCursor(Cursor.E_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_SOUTHEAST :
				setPredefinedCursor(Cursor.SE_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_SOUTH :
				setPredefinedCursor(Cursor.S_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_SOUTHWEST :
				setPredefinedCursor(Cursor.SW_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_WEST :
				setPredefinedCursor(Cursor.W_RESIZE_CURSOR);
				break;
			case IFigLayered.RESHAPE_NORTHWEST :
				setPredefinedCursor(Cursor.NW_RESIZE_CURSOR);
				break;
			default :
				restoreCursor();
		}
	}

	/** This method init the reshape rectangle.
	     *  <p>
	     *  We assume that the attributes <var>_direction</var> and
	     *  <var>_figToReshape</var> was correctly set before the
	     *  call to this method.
	     */
	protected void startReshapeRect() {
		_showReshapeRect = false;
		switch (_direction) {
			case IFigLayered.RESHAPE_NORTH :
				_yAnchor = _figToReshape.getY() + _figToReshape.getHeight();
				_reshapeRect.x = _figToReshape.getX();
				_reshapeRect.width = _figToReshape.getWidth();
				break;
			case IFigLayered.RESHAPE_NORTHEAST :
				_xAnchor = _figToReshape.getX();
				_yAnchor = _figToReshape.getY() + _figToReshape.getHeight();
				_reshapeRect.x = _xAnchor;
				break;
			case IFigLayered.RESHAPE_EAST :
				_xAnchor = _figToReshape.getX();
				_yAnchor = _figToReshape.getY();
				_reshapeRect.x = _xAnchor;
				_reshapeRect.y = _yAnchor;
				_reshapeRect.height = _figToReshape.getHeight();
				break;
			case IFigLayered.RESHAPE_SOUTHEAST :
				_xAnchor = _figToReshape.getX();
				_yAnchor = _figToReshape.getY();
				_reshapeRect.x = _xAnchor;
				_reshapeRect.y = _yAnchor;
				break;
			case IFigLayered.RESHAPE_SOUTH :
				_xAnchor = _figToReshape.getX();
				_yAnchor = _figToReshape.getY();
				_reshapeRect.x = _xAnchor;
				_reshapeRect.y = _yAnchor;
				_reshapeRect.width = _figToReshape.getWidth();
				break;
			case IFigLayered.RESHAPE_SOUTHWEST :
				_xAnchor = _figToReshape.getX() + _figToReshape.getWidth();
				_yAnchor = _figToReshape.getY();
				_reshapeRect.y = _yAnchor;
				break;
			case IFigLayered.RESHAPE_WEST :
				_xAnchor = _figToReshape.getX() + _figToReshape.getWidth();
				_yAnchor = _figToReshape.getY();
				_reshapeRect.y = _yAnchor;
				_reshapeRect.height = _figToReshape.getHeight();
				break;
			case IFigLayered.RESHAPE_NORTHWEST :
				_xAnchor = _figToReshape.getX() + _figToReshape.getWidth();
				_yAnchor = _figToReshape.getY() + _figToReshape.getHeight();
				break;
		}
	}

	/** This method permit to update the reshape rect.
	     *
	     * @param x the last position of the mouse.
	     * @param y the last position of the mouse.
	     */
	void updateReshapeRect(int x, int y) {
		Rectangle r = _modeManager.getGraph().getWorkspaceRect();
		if (x < r.x)
			x = r.x;
		else if (x > (r.x + r.width))
			x = r.x + r.width;

		if (y < r.y)
			y = r.y;
		else if (y > (r.y + r.height))
			y = r.y + r.height;

		switch (_direction) {
			case IFigLayered.RESHAPE_NORTH :
				_reshapeRect.y = Math.min(_yAnchor - _figToReshape.getMinimalHeight(), y);
				_reshapeRect.height = _yAnchor - _reshapeRect.y;
				break;
			case IFigLayered.RESHAPE_NORTHEAST :
				_reshapeRect.y = Math.min(_yAnchor - _figToReshape.getMinimalHeight(), y);
				_reshapeRect.width = Math.max(_figToReshape.getMinimalWidth(), x - _xAnchor);
				_reshapeRect.height = Math.max(_figToReshape.getMinimalHeight(), _yAnchor - _reshapeRect.y);
				break;
			case IFigLayered.RESHAPE_EAST :
				_reshapeRect.width = Math.max(_figToReshape.getMinimalWidth(), x - _xAnchor);
				break;
			case IFigLayered.RESHAPE_SOUTHEAST :
				_reshapeRect.width = Math.max(_figToReshape.getMinimalWidth(), x - _xAnchor);
				_reshapeRect.height = Math.max(_figToReshape.getMinimalHeight(), y - _yAnchor);
				break;
			case IFigLayered.RESHAPE_SOUTH :
				_reshapeRect.height = Math.max(_figToReshape.getMinimalHeight(), y - _yAnchor);
				break;
			case IFigLayered.RESHAPE_SOUTHWEST :
				_reshapeRect.x = Math.min(_xAnchor - _figToReshape.getMinimalWidth(), x);
				_reshapeRect.width = _xAnchor - _reshapeRect.x;
				_reshapeRect.height = Math.max(_figToReshape.getMinimalHeight(), y - _yAnchor);
				break;
			case IFigLayered.RESHAPE_WEST :
				_reshapeRect.x = Math.min(x, _xAnchor - _figToReshape.getMinimalWidth());
				_reshapeRect.width = _xAnchor - _reshapeRect.x;
				break;
			case IFigLayered.RESHAPE_NORTHWEST :
				_reshapeRect.x = Math.min(x, _xAnchor - _figToReshape.getMinimalWidth());
				_reshapeRect.y = Math.min(_yAnchor - _figToReshape.getMinimalHeight(), y);
				_reshapeRect.width = _xAnchor - _reshapeRect.x;
				_reshapeRect.height = _yAnchor - _reshapeRect.y;
				break;
		}

		_showReshapeRect = true;
	}
}