package mygale.mcge.mode;

// Copyright (C) 2000  Mahdi Hannoun, Stphane Galland
//
// This program is free software; you can redistribute it and/or modify

import java.awt.Component;

import javax.swing.*;

import mygale.mcge.core.JGraph;
import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.throwable.MCGEInternalError;

/** This class implements a Move mode, that intercept the
 *  mouse event to drag figures on the drawing area.
 *  <p>
 *  Each time a user pressed the middle button of the mouse,
 *  he can drag the drawing area by moving the mouse.
 *  The dragging is limited to the current bounds of the
 *  workspace. This mode can't extend the drawing area. It
 *  can just move the drawing area into the workspace.
 *  <p><center>
 *    <img src="doc-files/ModeScrollDrawingArea_uml.gif" width="389" height="88" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="171,0,243,28"  href="Mode.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="315,59,388,86"  href="../core/JGraph.html">
 *    </map>
 *  </center></p>
 */
public class ModeScrollDrawingArea extends Mode {

	/** The last x coordinate of the last move mouse event.
	 */
	protected int _lastX;
	/** The last y coordinate of the last move mouse event.
	 */
	protected int _lastY;

	/** A reference to the scrollbar of the editor. */
	protected transient JScrollBar hscroll = null;
	/** A reference to the scrollbar of the editor. */
	protected transient JScrollBar vscroll = null;

	/** the minimal index of the horizontal scrollbat.
	 */
	protected transient int hposMin;
	/** the maximal index of the horizontal scrollbat.
	 */
	protected transient int hposMax;

	/** the minimal index of the vertical scrollbat.
	 */
	protected transient int vposMin;
	/** the maximal index of the vertical scrollbat.
	 */
	protected transient int vposMax;

	/** Construct a new ModeScrollDrawingArea. The
	     *  {@link ModeManager ModeManager} parent should be
	     *  set before using this object.
	     */
	public ModeScrollDrawingArea() {
	}

	/** Construct a new ModeScrollDrawingArea with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModeScrollDrawingArea(ModeManager modeManager) {
		super(modeManager);
	}

	/** Always <code>false</code> because I never want
	 *  to get out of this mode.
	 *
	@link ModeManager ModeManager} stack,
	 *         otherwise <code>false</code>.
	 * @return always <code>true</code> if this mode can be dropped
	 *         from the { */
	public boolean canExit() {
		return false;
	}

	/** This method iniate the scrolling of the drawing area.
	 *
	 * @throws mygale.mcge.throwable.MCGEInternalError something wrong !
	 */
	public void initializeScrollBarValue() {
		JGraph graph = _modeManager.getGraph();
		Component c = graph.getParent();
		if (c != null && c instanceof JViewport) {
			JScrollPane scroll = graph.getGraphEditor().getScrollPane();
			hscroll = scroll.getHorizontalScrollBar();
			vscroll = scroll.getVerticalScrollBar();

			vposMin = vscroll.getMinimum();
			vposMax = vscroll.getMaximum();

			hposMin = hscroll.getMinimum();
			hposMax = hscroll.getMaximum();

		} else {
			throw new MCGEInternalError("in the mode ScrollDrawingArea");
		}
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		return "Drag the graph area with the middle button to  scroll it. Type <escape> to cancel.";
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  When the user move the mouse, then this
	     *  mode drags the drawing area.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((!me.isOnlyMiddleButtonDown()) || (!isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		// get the actual mouse coordinate.
		int x = me.getX();
		int y = me.getY();

		// compute the translation move from the last move point
		int xTrans = x - _lastX;
		int yTrans = y - _lastY;

		scroll(xTrans, yTrans);

		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *  <p>
	     *  Consume this event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		if ((!me.isOnlyMiddleButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-exited event was received.
	     *  <p>
	     *  consume this event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		if ((!me.isOnlyMiddleButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  If the user pressed the middle mouse button,
	 *  then this mode begins to drag the drawing area.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	*/
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((!me.isOnlyMiddleButtonDown()) || (isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		_lastX = me.getX();
		_lastY = me.getY();

		initializeScrollBarValue();

		setExclusive(true);
		me.consume();
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  When the mouse button was released, this mode stops
	     *  itself.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((!me.isOnlyMiddleButtonDown()) || (!isExclusive()) || (me.isModifierKeyDown())) {
			return;
		}

		done();
		me.consume();
	}

	/** This method drag the drawing area by the given
	     *  quantity along the two axes.
	     *
	     * @param xTrans the movment along the x axis.
	     * @param yTrans the movment along the y axis.
	     */
	public void scroll(int xTrans, int yTrans) {
		int vpos = vscroll.getValue();
		int hpos = hscroll.getValue();

		hpos = hpos - ((int) (xTrans * 0.5));
		if (hpos > hposMax) {
			hpos = hposMax;
		}
		if (hpos < hposMin) {
			hpos = hposMin;
		}

		vpos = vpos - ((int) (yTrans * 0.5));
		if (vpos > vposMax) {
			vpos = vposMax;
		}
		if (vpos < vposMin) {
			vpos = vposMin;
		}

		hscroll.setValue(hpos);
		vscroll.setValue(vpos);
	}
}
