package mygale.mcge.mode;

import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.util.Iterator;

import mygale.mcge.event.mouse.MCGEMouseEvent;
import mygale.mcge.fig.*;
import mygale.mcge.sel.Selection;

/** This class implements a Mode that interperts user input as
 *  selecting one or more Figs. Clicking on a
 *  {@link mygale.mcge.fig.Fig Fig} will select it.
 *  <p>
 *  Shift-clicking will toggle whether it is selected without
 *  deleting the old selection. Control-clicking will toggle
 *  whetger it is select but toggle objects that are
 *  already selected. Dragging in on space will draw a
 *  selection rectangle.  Dragging on a Fig will
 *  switch to {@link ModeMove ModeMove}.  Dragging from a
 *  {@link mygale.mcge.fig.FigEdge edge} will switch to
 *  {@link ModeModifyEdgeCtrlPoint ModeModifyEdgeCtrlPoint}.
 *  ModeSelect paints itself by displaying its
 *  selection rectangle if any.
 *  <p><center>
 *    <img src="doc-files/ModeSelect_uml.gif" width="408" height="200" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="158,0,230,28"  href="Mode.html">
 *      <area shape="Rect" coords="0,59,100,86"  href="../event/mouse/MCGEMouseEvent.html">
 *      <area shape="Rect" coords="289,59,362,86"  href="../fig/Fig.html">
 *      <area shape="Rect" coords="14,117,113,144"  href="ModeMove.html">
 *      <area shape="Rect" coords="271,117,407,144"  href="ModeModifyEdgeCtrlPoint.html">
 *      <area shape="Rect" coords="77,171,176,199"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="212,171,311,199"  href="../sel/Selection.html">
 *    </map>
 *  </center></p>

 */
public class ModeSelect extends Mode {

	/** This is the x coordonate of the initial point of the selection rectangle.
	 */
	protected int _xAnchor;
	/** This is the y coordonate of the initial point
	 *  of the selection rectangle.
	 */
	protected int _yAnchor;

	/** This is the seclection rectangle.
	 */
	protected Rectangle _selectRect = new Rectangle(0, 0, 0, 0);

	/** <code>true</code> when the selection rectangle
	 *  should be painted.
	 */
	protected boolean _showSelectRect = false;

	/** <code>true</code> when the user holds the shift
	 *  key to toggle selections.
	 */
	protected boolean _toggleSelection = false;

	/** The {@link mygale.mcge.fig.Fig fig} under
	 *  the mouse when the mouse bouton pressed.
	 */
	protected IFig _figUnderMouse = null;

	/** Construct a new ModeSelect. The
	     *  {@link ModeManager ModeManager} should be
	     *  set before using this object.
	     */
	public ModeSelect() {
	}

	/** Construct a new ModeSelect with the given parent.
	 *
	 * @param modeManager a reference to the ModeManager that
	 *                    contains this Mode.
	 */
	public ModeSelect(ModeManager modeManager) {
		super(modeManager);
	}

	//

	/** Always <code>false</code> because I never want
	 *  to get out of this mode.
	 *
	@link ModeManager ModeManager} stack,
	 *         otherwise <code>false</code>.
	 * @return always <code>true</code> if this mode can be dropped
	 *         from the { */
	public boolean canExit() {
		return false;
	}

	/** Ask to this mode to return into a stable state.
	     */
	public void cleanMode() {
		endSelectRect(_selectRect.x + _selectRect.width, _selectRect.y + _selectRect.height);
		_toggleSelection = false;
		_figUnderMouse = null;
		_modeManager.getGraph().damaged(_selectRect);
		setExclusive(false);
	}

	/** Reply a string of instructions that should be shown in the
	 *  statusbar when this mode starts.
	 *
	 * @return a string that explain how tu use this mode. Usually,
	 *         the returned string is displayed in a status bar.
	 */
	public String instructions() {
		if (isExclusive())
			return "Drag the selection rectangle to select object(s). Type <escape> to cancel.";
		else
			return super.instructions();
	}

	/** Invoked when a mouse-clicked event was received.
	     *  <p>
	     *  If the click occurs on the background, the selection
	     *  was cleaned. If the click occurs on a figure,
	     *  this one will be added to the current selection
	     *  (iff shift was pressed); will be added or toggled
	     *  from the selection (iff control was pressed); or
	     *  will be the selection (iff no key pressed).
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseClicked(MCGEMouseEvent)
	 */
	public void mcgeMouseClicked(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive())) {
			return;
		}

		Selection sl = _modeManager.getGraph().getSelection();
		IFigLayered fig = me.getFigure();
		if (fig == null) {
			sl.deselectAll();
		} else if (me.isShiftDown() || me.isControlDown()) {
			sl.toggle(fig);
		} else {
			sl.deselectAll();
			sl.select(fig);
		}
		me.consume();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  Destroy the selection.
	     *
	     * @param me description of the mouse-event.
	     */
	public void mcgeMouseDoubleClicked(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive())) {
			return;
		}

		Selection sel = _modeManager.getGraph().getSelection();
		Rectangle rect = sel.getBounds();
		sel.deselectAll();

		if (_showSelectRect == true) {
			rect = rect.union(_selectRect);
			_showSelectRect = false;
		}
		_modeManager.getGraph().damaged(rect);
		me.consume();
	}

	/** Invoked when a mouse-dragged event was received.
	     *  <p>
	     *  On mouse dragging, stretch the selection rectangle.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseDragged(MCGEMouseEvent)
	 */
	public void mcgeMouseDragged(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}

		int x = me.getX(), y = me.getY();

		_modeManager.getGraph().damaged(_selectRect);
		updateSelectRect(x, y);
		_modeManager.getGraph().damaged(_selectRect);
		_modeManager.getGraph().scrollToShow(x, y);

		me.consume();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseEntered(MCGEMouseEvent)
	 */
	public void mcgeMouseEntered(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-double-clicked event was received.
	     *  <p>
	     *  Consume the event.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseExited(MCGEMouseEvent)
	 */
	public void mcgeMouseExited(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}
		me.consume();
	}

	/** Invoked when a mouse-pressed event was received.
	 *  <p>
	 *  Handle mouse down events by preparing for a drag. If the mouse
	 *  down event happens on a handle or an already selected object, and
	 *  the shift key is not down, then go to
	 *  {@link ModeMove ModeMove}. If the mouse down event happens on a
	 *  {@link mygale.mcge.fig.FigEdge edge}, go to
	 *  {@link ModeModifyEdgeCtrlPoint ModeModifyEdgeCtrlPoint}.
	 *
	 * @param me description of the mouse-event.
	 * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMousePressed(MCGEMouseEvent)
	*/
	public void mcgeMousePressed(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (isExclusive())) {
			return;
		}

		Selection sel = _modeManager.getGraph().getSelection();
		int x = me.getX(), y = me.getY();
		_figUnderMouse = me.getFigure();

		if (_figUnderMouse instanceof IFigEdge) {
			gotoModeModifyEdgeCtrlPoint(me);
			return;
		}

		if (_figUnderMouse != null) {
			// Move mode ;
			gotoMoveMode(me);
			return;
		}

		_selectRect.x = x;
		_selectRect.y = y;
		_selectRect.width = 0;
		_selectRect.height = 0;

		setExclusive(true);
		startSelectRect(x, y);
		if (me.isControlDown()) {
			_toggleSelection = true;
		}

		if (!me.isShiftDown() && !me.isControlDown()) {
			sel.deselectAll();
		}

		me.consume();
	}

	/** Invoked when a mouse-released event was received.
	     *  <p>
	     *  On mouse up, select or toggle the selection of items under the
	     *  mouse or in the selection rectangle.
	     *
	     * @param me description of the mouse-event.
	     * @see mygale.mcge.event.mouse.MCGEMouseEventListener#mcgeMouseReleased(MCGEMouseEvent)
	 */
	public void mcgeMouseReleased(MCGEMouseEvent me) {
		if ((!me.isOnlyLeftButtonDown()) || (!isExclusive())) {
			return;
		}

		Selection sl = _modeManager.getGraph().getSelection();

		endSelectRect(me.getX(), me.getY());
		//_modeManager.getGraph().damaged(_selectRect);

		// select all the fig under the selectRectangle
		Iterator it = _modeManager.getGraph().getAllFigs();
		while (it.hasNext()) {
			IFigLayered fig = (IFigLayered)it.next();
			if (fig.within(_selectRect)) {
				if (_toggleSelection == true) {
					sl.toggle(fig);
				} else {
					sl.addFig(fig);
				}
			}
		}

		_toggleSelection = false;
		_modeManager.getGraph().damaged(_selectRect);
		setExclusive(false);
		me.consume();
	}

	/** Destroy the selection rectangle.
	     *  <p>
	     *  We assume that attributes <var>_xAnchor</var>
	     *  and <var>_yAnchor</var> are correctly was set
	     *  before the call to this method.
	     *
	     * @param y y coordinate of second corner of the rectangle.
	     * @param y y coordinate of second corner of the rectangle.
	     */
	void endSelectRect(int x, int y) {
		Rectangle r = _modeManager.getGraph().getWorkspaceRect();
		if (x < r.x)
			x = r.x;
		else if (x > (r.x + r.width))
			x = r.x + r.width;

		if (y < r.y)
			y = r.y;
		else if (y > (r.y + r.height))
			y = r.y + r.height;

		int bound_w = Math.abs(_xAnchor - x);
		int bound_h = Math.abs(_yAnchor - y);

		int bound_x = Math.min(_xAnchor, x);
		int bound_y = Math.min(_yAnchor, y);

		_selectRect.x = bound_x;
		_selectRect.y = bound_y;
		_selectRect.width = bound_w;
		_selectRect.height = bound_h;

		_showSelectRect = false;
	}

	/** Set the Editor's Mode to
	     *  {@link ModeModifyEdgeCtrlPoint ModeModifyEdgeCtrlPoint}.
	     *
	     * @param me the mouse event to forward to the ModeModifyEdgeCtrlPoint.
	     */
	protected void gotoModeModifyEdgeCtrlPoint(MCGEMouseEvent me) {
		Mode mode = new ModeModifyEdgeCtrlPoint(_modeManager);
		_modeManager.beginMode(mode);
		mode.mcgeMousePressed(me);
	}

	/** Set the Editor's Mode to
	 *  {@link ModeMove ModeMove}.
	 *
	 * @param me the mouse event to forward to the ModeMove.
	 */
	protected void gotoMoveMode(MCGEMouseEvent me) {
		Mode moveMode = new ModeMove(_modeManager);
		_modeManager.beginMode(moveMode);
		moveMode.mcgeMousePressed(me);
	}

	/** Initiate the selection rectangle.
	 *  <p>
	 *  We assume that attributes <var>_xAnchor</var>
	 *  and <var>_yAnchor</var> are correctly set by this
	 *  method.
	 *
	 * @param x x coordinate of first corner of the rectangle.
	 * @param y y coordinate of first corner of the rectangle.
	 */
	void startSelectRect(int x, int y) {
		_showSelectRect = false;

		_xAnchor = x;
		_yAnchor = y;
	}

	/** Change the selection rectangle.
	     *  <p>
	     *  We assume that attributes <var>_xAnchor</var>
	     *  and <var>_yAnchor</var> are correctly was set
	     *  before the call to this method.
	     *
	     * @param y y coordinate of second corner of the rectangle.
	     * @param y y coordinate of second corner of the rectangle.
	     */
	void updateSelectRect(int x, int y) {
		Rectangle r = _modeManager.getGraph().getWorkspaceRect();
		if (x < r.x)
			x = r.x;
		else if (x > (r.x + r.width))
			x = r.x + r.width;

		if (y < r.y)
			y = r.y;
		else if (y > (r.y + r.height))
			y = r.y + r.height;

		int bound_w = Math.abs(_xAnchor - x);
		int bound_h = Math.abs(_yAnchor - y);

		int bound_x = Math.min(_xAnchor, x);
		int bound_y = Math.min(_yAnchor, y);

		_selectRect.x = bound_x;
		_selectRect.y = bound_y;
		_selectRect.width = bound_w;
		_selectRect.height = bound_h;

		_showSelectRect = true;
	}

	/** Paint the selection rectangle if appropriate.
	 *
	 * @param g2d Graphics2D
	 */
	public void paint(Graphics2D g2d) {
		if (_showSelectRect && (_figUnderMouse == null)) {
			g2d.setColor(getPrefs().getShadowColor());
			g2d.drawRect(_selectRect.x, _selectRect.y, _selectRect.width, _selectRect.height);
		}
	}
}
