package mygale.mcge.net;

import java.util.Iterator;
import java.util.LinkedList;

import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IMapFunction;
import de.uniwue.ki.util.functions.IPredicateFunction;

/**
 * This is the semantic object for graph models.
 * <p>
 * <center>
 * <img src="doc-files/NetGraphModelImpl_uml.gif" width="291" height="174" border="0" usemap="#umlclick">
 * <map name="umlclick"><area shape="Rect" coords="55,0,127,28"  href="NetObjectImpl.html">
 * <area shape="Rect" coords="199,59,289,86"  href="NetGraphModel.html">
 * <area shape="Rect" coords="0,145,73,172"  href="NetNode.html">
 * <area shape="Rect" coords="104,145,176,172"  href="NetEdge.html"></map></center>
 * </p>
 *  
 */
public class DefaultNetGraphModel extends AbstractNetObject implements INetGraphModel {

    /**
     * This is the list of {@link NetNode nodes}.
     */
    protected LinkedList _nodes = new LinkedList();
    protected LinkedList _edges = new LinkedList();

    protected transient NetGraphModelChangedListener _netGraphModelChangedListener = null;

    /**
     * Construct a new NetGraphModelImpl.
     */
    public DefaultNetGraphModel() {
    }

    // Accessors

    /**
     * Get the node's list of this NetGraphModel
     * 
     * @return a list of {@link NetNode NetNodes}.
     */
    public Iterator getNodes() {
        return _nodes.iterator();
    }

    /**
     * Add a {@link NetEdge node}in this NetEdge
     * 
     * @param edge
     *            the new {@link NetEdge NetEdge}.
     */
    public void addEdge(INetEdge edge) {
        _edges.add(edge);

        fireAddedEdge(edge);
    }

    /**
     * @param newListener
     *            NetGraphModelChangedListener
     * @see INetGraphModel#addNetGraphModelChangedListener(NetGraphModelChangedListener)
     */
    public void addNetGraphModelChangedListener(NetGraphModelChangedListener newListener) {
        _netGraphModelChangedListener =
            NetGraphModelChangedEventMulticaster.add(_netGraphModelChangedListener, newListener);
        return;
    }

    /**
     * Add a {@link NetNode node}in this NetNode
     * 
     * @link NetNode NetNode}.
     * 
     * @param node
     *            the new { *
     * @see INetGraphModel#addNode(INetNode)
     */
    public void addNode(INetNode node) {
        if (_nodes == null) {
            _nodes = new LinkedList();
        }
        if (!_nodes.contains(node)) {
            _nodes.add(node);

            fireAddedNode(node);
        }
    }

    /**
     * Get the edge's list of this NetGraphModel
     * 
     * @return a list of {@link NetEdge NetEdges}.
     */
    public Iterator getAllEdges() {
        return CollectionsHelper.flatten(getNodes(), new IMapFunction() {
            public Object map(Object anObject) {
                INetNode netNode = (INetNode) anObject;
                return netNode.getIncomingNodeEdges();
            }
        });
    }

    public Iterator getEdges() {
        return _edges.iterator();
    }

    /**
     * Remove a edge attached to this NetGraphModel
     * 
     * @link NetEdge NetEdge} to remove.
     * 
     * @param edge
     *            the { *
     * @see INetGraphModel#removeEdge(INetEdge)
     */
    public void removeEdge(INetEdge edge) {
        _edges.remove(edge);
        edge.removed();

        fireRemovedEdge(edge);
    }

    /**
     * @param newListener
     *            NetGraphModelChangedListener
     * @see INetGraphModel#removeNetGraphModelChangedListener(NetGraphModelChangedListener)
     */
    public void removeNetGraphModelChangedListener(NetGraphModelChangedListener newListener) {
        _netGraphModelChangedListener =
            NetGraphModelChangedEventMulticaster.remove(_netGraphModelChangedListener, newListener);
        return;
    }

    /**
     * Remove a node attached to this NetGraphModel
     * 
     * @link NetNode NetNode} to remove.
     * 
     * @param node
     *            the { *
     * @see INetGraphModel#removeNode(INetNode)
     */
    public void removeNode(INetNode node) {
        if (_nodes != null && _nodes.contains(node)) {

            //            CollectionsHelper.clear(node.getNodeEdges());

            _nodes.remove(node);

            fireRemovedNode(node);

        }
    }

    /**
     * Method to support listener events.
     * 
     * @param event
     *            NetGraphModelChangedEvent
     */
    protected void fireAddedEdge(INetEdge Edge) {
        if (_netGraphModelChangedListener == null) {
            return;
        }

        _netGraphModelChangedListener.addedEdge(new NetGraphModelChangedEvent(this, Edge));
    }

    /**
     * Method to support listener events.
     * 
     * @param event
     *            NetGraphModelChangedEvent
     */
    protected void fireAddedNode(INetNode node) {
        if (_netGraphModelChangedListener == null) {
            return;
        }
        _netGraphModelChangedListener.addedNode(new NetGraphModelChangedEvent(this, node));
    }

    /**
     * Method to support listener events.
     * 
     * @param event
     *            NetGraphModelChangedEvent
     */
    protected void fireRemovedEdge(INetEdge Edge) {
        if (_netGraphModelChangedListener == null) {
            return;
        }
        _netGraphModelChangedListener.removedEdge(new NetGraphModelChangedEvent(this, Edge));
    }

    /**
     * Method to support listener events.
     * 
     * @param event
     *            NetGraphModelChangedEvent
     */
    protected void fireRemovedNode(INetNode node) {
        if (_netGraphModelChangedListener == null) {
            return;
        }
        _netGraphModelChangedListener.removedNode(new NetGraphModelChangedEvent(this, node));
    }

    public boolean contains(INetNode obj) {
        return _nodes.contains(obj);
    }

    public void removeEdgesForNode(final INetNode node) {

        LinkedList result = new LinkedList();
        try {
            CollectionsHelper.select(_edges.iterator(), new IPredicateFunction() {

                public boolean isElementOk(Object obj) throws Exception {
                    INetEdge edge = (INetEdge) obj;
                    return edge.getBeginNode() != node && edge.getEndNode() != node;
                }
            }, result);

            //TODO there should be some notification of listeners on removing edges --steffen
            _edges = result;
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
