package mygale.mcge.sel;



import java.awt.Rectangle;
import java.util.*;

import mygale.mcge.cmd.Cmd;
import mygale.mcge.core.JGraph;
import mygale.mcge.fig.IFigLayered;

import de.uniwue.ki.util.collection.CollectionsHelper;
import de.uniwue.ki.util.functions.IPredicateFunction;

/** This class handles Manager selections. It is basically a
 *  collection of Selected instances of
 *  {@link mygale.mcge.fig.FigLayered figures}.
 *  <p><center>
 *    <img src="doc-files/Selection_uml.gif" width="191" height="164" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,0,72,27"  href="../core/JGraph.html">
 *      <area shape="Rect" coords="117,68,189,95"  href="../fig/FigLayered.html">
 *      <area shape="Rect" coords="0,135,72,162"  href="../cmd/Cmd.html">
 *    </map>
 *  </center></p>

 */
public class SelectionImpl implements Selection {

	/** The collection of Selection instances.
	 */
	protected Vector _selections = new Vector();

	/** A list of Cmd that this Selection must change the enabling state.
	 */
	protected Vector _cmdToInform = new Vector();

	/** Say if cmd inform must be do.
	 */
	protected boolean _informCmd = true;

	/** This is the owner of this SelectionManager.
	 */
	protected JGraph _graph;


    /** Create a new SelectionManager.
     *
    @link mygale.mcge.core.JGraph JGraph} owner.
     * @param g the { */
    public SelectionImpl(JGraph g) {
    setGraph( g );
    }



    /** Get the {@link mygale.mcge.core.JGraph JGraph} that
     *  contains this SelectionManager.
     *
    @link mygale.mcge.core.JGraph JGraph}.
     * @return a { */
    public JGraph getGraph() { return _graph; }

/** Set the {@link mygale.mcge.core.JGraph JGraph} that
     *  contains this SelectionManager.
     *
    @link mygale.mcge.core.JGraph JGraph}.
     * @param graph the new { */
    public void setGraph(JGraph graph) { _graph = graph; }



    /** Add a {@link mygale.mcge.cmd.Cmd Cmd} to the Inform list.
     *
     * @param cmd a instance that this Selection must
     *            inform of changes.
     * @see mygale.mcge.sel.Selection#addCmdToInform(Cmd)
 */
    public void addCmdToInform(Cmd cmd) {
    _cmdToInform.addElement(cmd);
    }

/** Add a new figure to the collection of
     *  {@link mygale.mcge.fig.FigLayered Figs}.
     *
     * @pazam s the new figure.
     */
    public void addFig(IFigLayered s) {
    if ( _selections.contains( s ) ) return;
    _selections.addElement(s);
    s.setSelected( true );
    informCmd();
    }

/** Clean this collection.
     */
public void cleanUp() {
   Iterator it = getAllFigs();
   while (it.hasNext()) {
      ((IFigLayered) it.next()).cleanUp();
   }
}

/** Return <code>true</code> if the selection contains the
     *  given figure.
     */
    public boolean contains(IFigLayered fig) {
    return _selections.contains(fig) ;
    }

/** Unselect a {@link mygale.mcge.fig.FigLayered figure}.
     *
     * @param figure the figure to unselect.
     * @see mygale.mcge.sel.Selection#deselect(IFigLayered)
 */
public void deselect(IFigLayered figure) {
   removeFig(figure);
}

/** Unselect all the {@link mygale.mcge.fig.FigLayered Figs}.
     * @see mygale.mcge.sel.Selection#deselectAll()
 */
public void deselectAll() {
   boolean inform = isCmdInform();
   setCmdInform(false);
   Iterator it = getAllFigs();
   while (it.hasNext()) {
      IFigLayered figure = (IFigLayered) it.next();
      figure.setSelected(false);
   }
   _selections.clear();
   setCmdInform(inform);
   informCmd();
}

/** Get the {@link mygale.mcge.fig.FigLayered Figs} that are
     *  managed by this SelectionManager.
     *
     * @return a list of selections.
     * @see mygale.mcge.sel.Selection#getAllFigs()
 */
public Iterator getAllFigs() {
   return _selections.iterator();
}



/** This method return the rectangle that contains all the
 *  selected {@link mygale.mcge.fig.FigLayered FigLayered}.
 *
 * @return the bound of the selection.
 * @see mygale.mcge.sel.Selection#getBounds()
 */
public Rectangle getBounds() {
   Rectangle r = null;
   Iterator it = getAllFigs();
   while (it.hasNext()) {
      IFigLayered figure = (IFigLayered) it.next();
      if (r == null)
         r = figure.getBounds();
      else
         r = r.union(figure.getBounds());
   }
   return (r == null) ? new Rectangle() : r;
}

/** Get the locked selections that is managed by this
     *  SelectionManager.
     *
    @link mygale.mcge.fig.FigLayered figs}.
     * @return a list of { * @see mygale.mcge.sel.Selection#getLockedFigs()
 */
public Iterator getLockedFigs() {
   Iterator it = getAllFigs();

   return CollectionsHelper.select(it, new IPredicateFunction() {
      public boolean isElementOk(Object anObject) {
         return ((IFigLayered) anObject).isLocked();
      }
   });
}

/** Get the unlocked selections that are managed by
     *  this SelectionManager.
     *
    @link mygale.mcge.fig.FigLayered figs}.
     * @return a list of { * @see mygale.mcge.sel.Selection#getUnlockedFigs()
 */
public Iterator getUnlockedFigs() {
   Iterator it = getAllFigs();

   return CollectionsHelper.select(it, new IPredicateFunction() {
      public boolean isElementOk(Object anObject) {
         return !((IFigLayered) anObject).isLocked();
      }
   });
}

/** This method inform all {@link mygale.mcge.cmd.Cmd Cmd}
     *  about a change in this selection.
     * @see mygale.mcge.sel.Selection#informCmd()
 */
    public void informCmd() {
    if ( isCmdInform() ) {
        Enumeration enum = _cmdToInform.elements();
        while ( enum.hasMoreElements() ) {
        //((Cmd)enum.nextElement()).setEnabled( size()>0 );
        ((Cmd)enum.nextElement()).updateEnabled() ;
        }
    }
    }

/** Reply if the Cmd's information was currently used.
     *
    @link mygale.mcge.cmd.Cmd Cmds}
     *         of any change, otherwise <code>false</code>.
     * @return <code>true</code> if this Selection must
     *         inform { */
    public boolean isCmdInform() { return _informCmd; }

/** Reply the number of selected
     *  {@link mygale.mcge.fig.FigLayered FigLayered's} that
     *  are locked.
     *
     * @return the number of locked figures.
     * @see mygale.mcge.sel.Selection#lockedSize()
 */
public int lockedSize() {
   return CollectionsHelper.size(getLockedFigs());
}

/** Remove a {@link mygale.mcge.cmd.Cmd Cmd} from the Inform list.
     *
     * @param cmd a instance that this Selection must
     *            no longer inform of changes.
     */
    public void removeCmdToInform(Cmd cmd) {
    _cmdToInform.removeElement(cmd);
    }

/** Remove a figure from the collection of
     *  {@link mygale.mcge.fig.FigLayered Figs}.
     *
     * @pazam s the figure to remove.
     * @param s IFigLayered
 * @see mygale.mcge.sel.Selection#removeFig(IFigLayered)
 */
    public void removeFig(IFigLayered s) {
    if ( ! _selections.contains( s ) ) return;
    _selections.removeElement(s);
    s.setSelected( false );
    informCmd();
    }

/** Select a set of {@link mygale.mcge.fig.FigLayered figures}.
     *
     * @param figs the vector of figures to select.
     * @see mygale.mcge.sel.Selection#select(Iterator)
 */
public void select(Iterator figs) {
   boolean inform = isCmdInform();
   setCmdInform(false);
   while (figs.hasNext()) {
      IFigLayered fig = (IFigLayered) figs.next();
      addFig(fig);
   }
   setCmdInform(inform);
   informCmd();
}



/** Select a {@link mygale.mcge.fig.FigLayered figure}.
 *
 * @param figure the {@link mygale.mcge.fig.FigLayered FigLayered}
 *               to select.
 */
public void select(IFigLayered figure) {
   addFig(figure);
}

/** Enable/Disable the Cmd's information.
     *
    @link mygale.mcge.cmd.Cmd Cmds}
     *               of any change, otherwise <code>false</code>.
     * @param inform <code>true</code> if this Selection must
     *               inform { * @return the old value of the Inform flag.
     * @see mygale.mcge.sel.Selection#setCmdInform(boolean)
 */
    public boolean setCmdInform(boolean inform) {
    boolean oldInform = _informCmd;
    _informCmd = inform;
    return oldInform;
    }

/** Reply the number of selected
     *  {@link mygale.mcge.fig.FigLayered FigLayered's}.
     *
     * @return the size of the selection.
     * @see mygale.mcge.sel.Selection#size()
 */
    public int size() { return _selections.size(); }

/** Toggle a set of {@link mygale.mcge.fig.FigLayered FigLayered}.
     *
    @link mygale.mcge.fig.FigLayered FigLayered} to toggle.
     * @param figs the vector of { * @see mygale.mcge.sel.Selection#toggle(Iterator)
 */
public void toggle(Iterator figs) {
   boolean inform = isCmdInform();
   setCmdInform(false);
   while (figs.hasNext()) {
      toggle((IFigLayered) figs.next());
   }
   setCmdInform(inform);
   informCmd();
}

/** Toggle the selection of a {@link mygale.mcge.fig.FigLayered FigLayered}.
     *
    @link mygale.mcge.fig.FigLayered FigLayered} to toggle.
     * @param figure the { * @see mygale.mcge.sel.Selection#toggle(IFigLayered)
 */
    public void toggle(IFigLayered figure) {
    if (figure.isSelected()) deselect(figure);
    else select(figure);
    }

/** Reply the number of selected
     *  {@link mygale.mcge.fig.FigLayered FigLayered's} that
     *  are unlocked.
     *
     * @return the number of unlocked figures.
     * @see mygale.mcge.sel.Selection#unlockedSize()
 */
public int unlockedSize() {
   return CollectionsHelper.size(getUnlockedFigs());
}

/** Get the selections that are managed by this SelectionManager.
     *
     * @return a list of {@link mygale.mcge.fig.FigLayered figs}.
     */
protected Vector getFigs() {
   return _selections;
}
}
