package mygale.mcge.undoredo;



import java.util.*;

/** This is the manager that permits to store undo/redo commands.
 *  <p><center>
 *    <img src="doc-files/UndoRedoManagerImpl_uml.gif" width="349" height="227" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="0,0,99,28"  href="UndoRedoListener.html">
 *      <area shape="Rect" coords="248,45,347,72"  href="UndoRedoManager.html">
 *      <area shape="Rect" coords="63,140,162,167"  href="UndoRedoableCmd.html">
 *      <area shape="Rect" coords="8,198,81,226"  href="UndoableCmd.html">
 *      <area shape="Rect" coords="149,198,221,226"  href="RedoableCmd.html">
 *    </map>
 *  </center></p>
 * */
public class UndoRedoManagerImpl
	implements UndoRedoManager {

	/** This the size of undo list.
	 */
	public static final int UNDOLIST_LENGTH = 100;
	/** This the size of redo list.
	 */
	public static final int REDOLIST_LENGTH = 10;

	/** This the list of undoable commands.
	 */
	protected UndoableCmd _undoables[] = new UndoableCmd[UNDOLIST_LENGTH];

	/** This the list of redoable commands.
	 */
	protected RedoableCmd _redoables[] = new RedoableCmd[REDOLIST_LENGTH];

	/** This is the position of the first inserted undoable command.
	 */
	protected int _undobegin = 0;

	/** This is the position of the last inserted undoable command.
	 */
	protected int _undoend = -1;

	/** This is the position of the first inserted redoable command.
	 */
	protected int _redobegin = 0;

	/** This is the position of the last inserted redoable command.
	 */
	protected int _redoend = -1;

	/** Set of command to inform about any change in the undo/redo manager.
	 */
	protected Vector _listeners = new Vector();



    /** Construct a new UndoRedoManagerImpl.
     */
    public UndoRedoManagerImpl() { }


    /** Add a command to the undo list.
     *
     * @param cmd the new undoable command.
     * @see mygale.mcge.undoredo.UndoRedoManager#addUndo(UndoableCmd)
 */
    public void addUndo( UndoableCmd cmd ) {
    addUndo( cmd, true );
    }

/** Clear the redo list.
     */
    public void clearRedo() {
    _redobegin = 0;
    _redoend = -1;
    informListeners();
    }

/** Clear the undo list.
     * @see mygale.mcge.undoredo.UndoRedoManager#clearUndo()
 */
    public void clearUndo() {
    _undobegin = 0;
    _undoend = -1;
    clearRedo();
    }

/** Redo the first command in the list.
     * @see mygale.mcge.undoredo.UndoRedoManager#redo()
 */
    public void redo() {
    int size = redoSize();
    if ( size > 0 ) {
        // Remove the command from the redo list
        RedoableCmd cmd = _redoables[ _redoend ];
        boolean inform = false;
        if ( size == 1 ) {
        // remove the last
        _redoend = -1;
        inform = true;
        }
        else {
        _redoend = ( ( _redoend == 0 ) ? REDOLIST_LENGTH : _redoend ) - 1;
        }
        // Run the redoable command
        cmd.redoIt();
        // Move the redoable command to the undo list
        if ( ! addUndo( (UndoableCmd)cmd, false ) && inform) {
        informListeners();
        }
    }
    }

/** Reply the count of redoable command in the list.
     *
     * @return a integer that is the number of redoable commands.
     * @see mygale.mcge.undoredo.UndoRedoManager#redoSize()
 */
    public int redoSize() {
    if ( _redoend < 0 ) return 0;
    else if ( _redobegin <= _redoend ) return _redoend - _redobegin + 1;
    else return REDOLIST_LENGTH - _redobegin + _redoend + 1;
    }


    /** Add un {@link UndoRedoListener UndoRedoListener}
     *  into this Manager.
     *
     * @param listener the listener to add.
     * @see mygale.mcge.undoredo.UndoRedoManager#register(UndoRedoListener)
 */
    public void register( UndoRedoListener listener ) {
    if ( ! _listeners.contains( listener ) ) {
        _listeners.addElement( listener );
    }
    }

/** Undo the first command in the list.
     * @see mygale.mcge.undoredo.UndoRedoManager#undo()
 */
    public void undo() {
    int size = undoSize();
    if ( size > 0 ) {
        UndoableCmd cmd = _undoables[ _undoend ];
        boolean inform = false;
        // Move the undoable command to the reod list
        if ( cmd instanceof RedoableCmd ) {
        inform = addRedo( (RedoableCmd)cmd, false );
        }
        // Remove the command from the undo list
        if ( size == 1 ) {
        // This the last command in the list
        _undoend = -1;
        inform = true;
        }
        else {
        _undoend = ( ( _undoend == 0 ) ? UNDOLIST_LENGTH : _undoend ) - 1;
        }
        // Run the undoable command
        cmd.undoIt();
        if ( inform ) informListeners();
    }
    }

/** Reply the count of undoable command in the list.
     *
     * @return a integer that is the number of undoable commands.
     * @see mygale.mcge.undoredo.UndoRedoManager#undoSize()
 */
    public int undoSize() {
    if ( _undoend < 0 ) return 0;
    else if ( _undobegin <= _undoend ) return _undoend - _undobegin + 1;
    else return UNDOLIST_LENGTH - _undobegin + _undoend + 1;
    }

/** Remove un {@link UndoRedoListener UndoRedoListener}
     *  into this Manager.
     *
     * @param listener the listener to add.
     * @see mygale.mcge.undoredo.UndoRedoManager#unregister(UndoRedoListener)
 */
    public void unregister( UndoRedoListener listener ) {
    _listeners.removeElement( listener );
    }


    /** Add a command to the redo list and informs
     *  {@link UndoRedoListener listeners} if
     *  <var>inform</var> was <code>true</code>.
     *
     * @param cmd the redoable command to add.
     * @param inform <code>true</code> if this method
     *               can inform {@link UndoRedoListener listeners},
     *               otherwise <code>false</code>.
     * @return <code>true</code> if listeners was informed,
     *         otherwise <code>false</code>.
     */
    protected boolean addRedo( RedoableCmd cmd, boolean inform ) {
    int size = redoSize();
    if ( size <= 0 ) {
        // Add the first redoable element
        _redoend = _redobegin - 1;
        inform = true;
    }
    else if ( size >= REDOLIST_LENGTH ) {
        _redobegin = ( _redobegin + 1 ) % REDOLIST_LENGTH;
    }
    _redoend = ( _redoend + 1 ) % REDOLIST_LENGTH;
    _redoables[ _redoend ] = cmd;
    if ( inform ) informListeners();
    return inform;
    }

/** Add a command to the undo list and clear
     *  the redo list if the given parameter
     *  <var>canClearRedo</var> was <code>true</code>.
     *
     * @param cmd the undoable command to add.
     * @param canClearRedo <code>true</code> if this method
     *                     can clear the redo list, otherwise
     *                     <code>false</code>.
     * @return <code>true</code> if listeners was informed,
     *         otherwise <code>false</code>.
     */
    protected boolean addUndo( UndoableCmd cmd, boolean canClearRedo ) {
    int size = undoSize();
    boolean inform = false;
    if ( size <= 0 ) {
        // Add the first command
        _undoend = _undobegin - 1;
        inform = true;
    }
    else if ( size >= UNDOLIST_LENGTH ) {
        // Delete the lastest command
        _undobegin = ( _undobegin + 1 ) % UNDOLIST_LENGTH;
    }
    _undoend = ( _undoend + 1 ) % UNDOLIST_LENGTH;
    _undoables[ _undoend ] = cmd;
    if ( canClearRedo ) {
        clearRedo();
        inform = true;
    }
    else if ( inform ) informListeners();
    return inform;
    }

/** Inform {@link UndoRedoListener listeners}
     *  about any significant change.
     */
    protected void informListeners() {
    Enumeration enum = _listeners.elements();
    while ( enum.hasMoreElements() ) {
        ((UndoRedoListener)enum.nextElement()).updateEnabled();
    }
    }
}
