package mygale.mcge.util.prefs;

import java.awt.Color;

import junit.framework.Assert;
import mygale.mcge.fig.IFigLayered;

/** This class contains preferences that control the behavior of the
 *  editor to make it the way that the user likes it.
 *  This is not a static classe because prefs could be different
 *  for each editors.
 *  <p><center>
 *    <img src="doc-files/Prefs_uml.gif" width="250" height="164" border="0" usemap="#umlclick">
 *    <map name="umlclick">
 *      <area shape="Rect" coords="18,0,90,27"  href="../../core/JGraph.html">
 *      <area shape="Rect" coords="140,68,248,95"  href="PrefChangedListener.html">
 *      <area shape="Rect" coords="0,135,108,163"  href="PrefChangedEvent.html">
 *    </map>
 *  </center></p>

 */
public class Prefs implements Cloneable {

	/** This is the current editing dialogue box.
	 */
	protected PrefChangedListener _listener;

	/** This indicate if events must be sent.
	 */
	private boolean _sendevents = true;

	/** The color of the canvas' backgound.
	 */
	protected Color _bgcolor = Color.lightGray;

	/** The color of the canvas' foregound.
	 */
	protected Color _fgcolor = Color.gray;

	/** The color of the edge' line.
	 */
	protected Color _figedgecolor = Color.black;

	/** The color of the Fig backgound.
	 */
	protected Color _figbgcolor = Color.white;

	/** The color of the Fig foregound.
	 */
	protected Color _figfgcolor = Color.black;

	/** The color of the locked-Fig's backgound.
	 */
	protected Color _lockfigbgcolor = Color.white;

	/** The color of the locked-Fig foregound.
	 */
	protected Color _lockfigfgcolor = Color.gray;

	/** The color of the shadow.
	 */
	protected Color _shadowcolor = Color.gray;

	/** The size of handles.
	 */
	protected int _handlesize = 6;

	/** The color of the handle.
	 */
	protected Color _handlecolor = Color.blue.darker();

	/** The color of the pseudo handle.
	 */
	protected Color _pseudohandlecolor = Color.green;

	/** The color of the locked handle.
	 */
	protected Color _lockhandlecolor = Color.cyan;

	/** The color of active port handles.
	 */
	protected Color _activeportcolor = Color.green.darker().darker();

	/** Construct a new Prefs instance.
	 */
	public Prefs() {
		super();
	}

	/**
	 * @param listener PrefChangedListener
	 */
	public void addPrefChangedListener(PrefChangedListener listener) {
		if (Assert.ASSERTION_ON)
			Assert.assertNull(_listener);
		_listener = listener;
	}

	/** Reply if event must be sent.
	     *
	     * @return <code>true</code> if events must be sent,
	     *         <code>false</code> otherwise.
	     */
	protected boolean canFire() {
		return _sendevents;
	}

	/** Create and return a copy of this Perfs.
	     *
	     * @return the copy.
	     * @throws CloneNotSupportedException this object can't be clone.
	     */
	public Object clone() throws CloneNotSupportedException {
		Prefs clone = (Prefs)super.clone();
		return clone;
	}

	/** Clone the given Perfs.
	     *
	     * @param prefs the Prefs to clone.
	     */
	public void clone(Prefs prefs) {
		setFire(false);
		setForegroundColor(prefs.getForegroundColor());
		setFigForegroundColor(prefs.getFigForegroundColor());
		setFigBackgroundColor(prefs.getFigBackgroundColor());
		setLockedFigForegroundColor(prefs.getLockedFigForegroundColor());
		setLockedFigBackgroundColor(prefs.getLockedFigBackgroundColor());
		setShadowColor(prefs.getShadowColor());
		setHandleSize(prefs.getHandleSize());
		setHandleColor(prefs.getHandleColor());
		setPseudoHandleColor(prefs.getPseudoHandleColor());
		setLockedHandleColor(prefs.getLockedHandleColor());
		setActivePortColor(prefs.getActivePortColor());
		setFire(true);
		setBackgroundColor(prefs.getBackgroundColor());
	}

	/** This method permit to inform that some prefs has changed.
	 * @param prefname this is the name of the changed preference.
	 * @param type this is the type of the preference.
	 * @param newvalue the new value of the preference.
	 * @param oldvalie the old value of the preference.
	 */
	protected void firePrefChanged(String prefname, int type, Object newvalue, Object oldvalue) {
		if (_listener != null && _sendevents) {
			_listener.prefChanged(new PrefChangedEvent(prefname, type, newvalue, oldvalue));
		}
	}

	/** Get the color of active port handles.
	     *
	     * @return color of active port handles.
	     */
	public Color getActivePortColor() {
		return _activeportcolor;
	}

	/** Get the color of the canvas' background.
	 *
	 * @return color of the background.
	 */
	public Color getBackgroundColor() {
		synchronized (_bgcolor) {
			return _bgcolor;
		}
	}

	/** Get the color of the Fig's background.
	     *
	     * @return color of the background.
	     */
	public Color getFigBackgroundColor() {
		return _figbgcolor;
	}

	/** Get the background color that should be apply to a Fig.
	     *
	     * @param figure the requesting figure.
	     * @return color of the background.
	     */
	public Color getFigBackgroundColorFor(IFigLayered figure) {
		return (figure.isLocked()) ? getLockedFigBackgroundColor() : getFigBackgroundColor();
	}

	/** Get the color of the fig edge.
	     *
	     * @return color of the edge.
	     */
	public Color getFigEdgeColor() {
		return _figedgecolor;
	}

	/** Get the color of the Fig's foreground.
	     *
	     * @return color of the foreground.
	     */
	public Color getFigForegroundColor() {
		return _figfgcolor;
	}

	/** Get the foreground color that should be apply to a Fig.
	     *
	     * @param figure the requesting figure.
	     * @return color of the foreground.
	     */
	public Color getFigForegroundColorFor(IFigLayered figure) {
		return (figure.isLocked()) ? getLockedFigForegroundColor() : getFigForegroundColor();
	}

	/** Get the color of the canvas' foreground.
	     *
	     * @return color of the foreground.
	     */
	public Color getForegroundColor() {
		synchronized (_fgcolor) {
			return _fgcolor;
		}
	}

	/** Get the color of handles.
	     *
	     * @return color of handles.
	     */
	public Color getHandleColor() {
		return _handlecolor;
	}

	/** Get the handle color that should be apply to a Fig.
	     *
	     * @param figure the requesting figure.
	     * @return color of the foreground.
	     */
	public Color getHandleColorFor(IFigLayered figure) {
		return (figure.isLocked()) ? getLockedHandleColor() : getHandleColor();
	}

	/** Get the size of handles.
	     *
	     * @return size of handles.
	     */
	public int getHandleSize() {
		return _handlesize;
	}

	/** Get the color of the locked-Fig's background.
	     *
	     * @return color of the background.
	     */
	public Color getLockedFigBackgroundColor() {
		return _lockfigbgcolor;
	}

	/** Get the color of the locked-Fig's foreground.
	     *
	     * @return color of the foreground.
	     */
	public Color getLockedFigForegroundColor() {
		return _lockfigfgcolor;
	}

	/** Get the color of locked handles.
	     *
	     * @return color of locked handles.
	     */
	public Color getLockedHandleColor() {
		return _lockhandlecolor;
	}

	/** Get the color of pseudo handles.
	     *
	     * @return color of handles.
	     */
	public Color getPseudoHandleColor() {
		return _pseudohandlecolor;
	}

	/** Get the color of shadow
	     *
	     * @return color of shadows.
	     */
	public Color getShadowColor() {
		return _shadowcolor;
	}

	/**
	 * @param listener PrefChangedListener
	 */
	public void removePrefChangedListener(PrefChangedListener listener) {

		Assert.assertSame(listener, _listener);

		_listener = null;
	}

	/** Set the color of active port handles.
	     *
	     * @param color color of active port handles.
	     */
	public void setActivePortColor(Color color) {
		Color oldvalue = _activeportcolor;
		_activeportcolor = color;
		firePrefChanged("color_activeport", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of the canvas' background.
	     *
	     * @param color color of the background.
	     */
	public void setBackgroundColor(Color color) {
		synchronized (_bgcolor) {
			Color oldvalue = _bgcolor;
			_bgcolor = color;
			firePrefChanged("color_background", PrefChangedEvent.GUI, color, oldvalue);
		}
	}

	/** Set the color of the Fig's background.
	     *
	     * @param color color of the background. */
	public void setFigBackgroundColor(Color color) {
		Color oldvalue = _figbgcolor;
		_figbgcolor = color;
		firePrefChanged("color_background_figure", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of the fig edge.
	     *
	     * @param color color of Fig edges.
	     */
	public void setFigEdgeColor(Color color) {
		Color oldvalue = _figedgecolor;
		_figedgecolor = color;
		firePrefChanged("color_edge_figure", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of the Fig's background.
	     *
	     * @param color color of the background.
	     */
	public void setFigForegroundColor(Color color) {
		Color oldvalue = _figfgcolor;
		_figfgcolor = color;
		firePrefChanged("color_foreground_figure", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the flag that say if event must be sent.
	     *
	     * @param fire <code>true</code> if events must be sent,
	     *             <code>false</code> otherwise.
	     */
	protected void setFire(boolean fire) {
		_sendevents = fire;
	}

	/** Set the color of the canvas' foreground.
	     *
	     * @param color color of the foreground.
	     */
	public void setForegroundColor(Color color) {
		Color oldvalue = _fgcolor;
		_fgcolor = color;
		firePrefChanged("color_foreground", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of handles.
	     *
	     * @param color color of handles.
	     */
	public void setHandleColor(Color color) {
		Color oldvalue = _handlecolor;
		_handlecolor = color;
		firePrefChanged("color_handle", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the size of the handles.
	     *
	     * @param size size of handles.
	     */
	public void setHandleSize(int size) {
		Integer oldvalue = new Integer(_handlesize);
		_handlesize = size;
		firePrefChanged("size_handle", PrefChangedEvent.GUI, new Integer(size), oldvalue);
	}

	/** Set the color of the locked-Fig's background.
	     *
	     * @param color color of the background.
	     */
	public void setLockedFigBackgroundColor(Color color) {
		Color oldvalue = _lockfigbgcolor;
		_lockfigbgcolor = color;
		firePrefChanged("color_background_lockedfigure", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of the locked-Fig's background.
	     *
	     * @param color color of the background.
	     */
	public void setLockedFigForegroundColor(Color color) {
		Color oldvalue = _lockfigfgcolor;
		_lockfigfgcolor = color;
		firePrefChanged("color_foreground_lockedfigure", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of locked handles.
	     *
	     * @param color color of locked handles.
	     */
	public void setLockedHandleColor(Color color) {
		Color oldvalue = _lockhandlecolor;
		_lockhandlecolor = color;
		firePrefChanged("color_lockedhandle", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of pseudo handles.
	     *
	     * @param color color of handles.
	     */
	public void setPseudoHandleColor(Color color) {
		Color oldvalue = _pseudohandlecolor;
		_pseudohandlecolor = color;
		firePrefChanged("color_pseudohandle", PrefChangedEvent.GUI, color, oldvalue);
	}

	/** Set the color of shadows.
	     *
	     * @param color color of the shadows.
	     */
	public void setShadowColor(Color color) {
		Color oldvalue = _shadowcolor;
		_shadowcolor = color;
		firePrefChanged("color_shadow", PrefChangedEvent.GUI, color, oldvalue);
	}
}