package org.sadun.util;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListResourceBundle;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * A class to handle command line options.
 * <p>
 * Client code either uses the {@link #setObjectOnOption(java.lang.String, java.lang.Object, java.lang.String, java.lang.Object)
 * setObjectOnOption()} and/or <b>setOnOption</b> methods; or subclasses
 * {@link CmdLineOptions.Handler CmdLineOptions.Handler} by implementing a method with
 * signature
 * <pre>
 *    public void option(String option, CmdLineOptions.Handler.ParamSet ps, boolean value);
 * </pre>
 * with callback code for each option and registers the subclasses with an instance of this class.
 * <p>
 * Invoking the {@link #parse(java.lang.String[]) parse()} method examines the command line
 * and executes the callback code.
 * <p>
 * Other features are the possibility of definining option aliases, option descriptions
 * an automatic usage printout build up over this information.
 *
 * @author C. Sadun
 * @version 1.6
 */
public class CmdLineOptions {

	private static String optPrefix = "-";
	private boolean allowNegatedOptions;
	private boolean failOnUnrecognizedOption;
	private boolean failOnMixedHelpOptionEncountered = true;
	private Map handlers = new HashMap();
	private Set mandatoryOptions = new HashSet();
	private Map aliasesTable = new HashMap();
	private String currentContext = ""; // set by parse()

	// Message bundles
	private static ResourceBundle msgsBundle =
		ResourceBundle.getBundle(
			"org.sadun.util.CmdLineOptions$Resources",
			Locale.getDefault());

	// A description of the tool embedding this CmdLineOptions, may be null
	private String usageDescription;

	// The "help options", i.e. the options which is supposed to show help (like -help)
	private String helpOption;

	/**
	 * A base class for options-related exceptions. Note that this class
	 * extends <tt>RuntimeException</tt>.
	 */
	public class OptionException extends RuntimeException {

		private String context;

		public OptionException(String msg) {
			this(msg, currentContext);
		}

		private OptionException(String msg, String context) {
			super(msg);
			this.context = context;
		}

		public String getMessage() {
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			pw.print(super.getMessage());
			if (context.length() > 0)
				pw.println(" in " + context);
			else
				pw.println();
			pw.println(usageString());
			return sw.toString();
		}

		public String usageString() {
			return CmdLineOptions.this.usage();
		}

	}

	private static class ProgrammingError extends Error {
		public ProgrammingError(String msg) {
			super("clopt library programming error: " + msg);
		}
	}

	/**
	 * This class handles an option and can be subclassed to define the
	 * reaction to the discovery of an option.
	 */
	public static abstract class Handler implements Comparable {

		/**
		 * The option literal
		 */
		protected String option;

		/**
		 * The description for the option
		 */
		protected String description;

		/**
		 * The aliases for the option
		 */
		protected String[] aliases;

		/**
		 * A set of parmeters for an option
		 */
		public static class ParamSet {

			private CmdLineOptions c;
			private int i;
			private Vector params;
			private String option;

			private ParamSet(CmdLineOptions c, String option, Vector params) {
				this.c = c;
				this.option = option;
				this.params = params;
				i = 0;
			}

			/**
			 * Return the next parameter.
			 * @return the next parameter
			 * @exception OptionException if there are no more parameters.
			 */
			public String next() {
				if (!hasNext())
					throw c.new OptionException(
						MessageFormat.format(
							msgsBundle.getString("oe1"),
							new Object[] { option }));
				return (String) params.elementAt(i++);
			}

			/**
			 * Return <b>true</b> if there is a next parameter.
			 * @return <b>true</b> if there is a next parameter.
			 */
			public boolean hasNext() {
				if (i >= params.size())
					return false;
				return true;
			}

			/**
			 * Return the number of parameters.
			 * @return the number of parameters.
			 */
			public int size() {
				return params.size();
			}

			/**
				 * Return a string representation of the parameters.
				 * @return a string representation of the parameters.
				 */
			public String toString() {
				Enumeration e = params.elements();
				StringBuffer sb = new StringBuffer("{ ");
				while (e.hasMoreElements()) {
					sb.append(e.nextElement());
					if (e.hasMoreElements())
						sb.append(", ");
				}
				sb.append(" }");
				return sb.toString();
			}
		}

		/**
		 * Build an handler which recognizes the given option literal,
		 * the given aliases and the given description.
		 * <p>
		 * @param option the option literal.
		 * @param description a description for the option.
		 * @param aliases the option aliases, in comma-separated format.
		 */
		public Handler(String option, String description, String aliases) {
			if (!option.startsWith(optPrefix))
				option = optPrefix + option;
			this.option = option;
			this.aliases =
				aliases == null
					? new String[0]
					: CmdLineOptions.getOptArrayFromString(aliases);
			if (description == null)
				this.description = "no description";
			else
				this.description = description;
		}

		/**
			  * Build an handler which recognizes the given option literal.
			  * <p>
			  * @param option the option literal.
			  */
		public Handler(String option, String description) {
			this(option, description, null);
		}

		/**
		 * Build an handler which recognizes the given option literal.
		 * <p>
		 * @param option the option literal.
		 */
		public Handler(String option) {
			this(option, null);
		}

		/**
		 * Set a description for the handled option - which is included in the {@link #usage() usage()} 
		 * string.
		 * @param description a short description of the tool.
		 */
		public void setDescription(String description) {
			if (description == null)
				throw new IllegalArgumentException("Description can't be null");
			this.description = description;
		}

		/**
		 * Return a description for the handled option  - which is included in the {@link #usage() usage()} 
		  * string.
		 * @return a description for the handled option - which is included in the {@link #usage() usage()} 
		  * string.
		 */
		public String getDescription() {
			return description;
		}

		/**
		 * Return a description of the handler.
		 * 
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			String name = getClass().getName();
			int i;
			if ((i = name.lastIndexOf(".")) != -1)
				name = name.substring(i + 1);
			if ((i = name.lastIndexOf("$")) != -1)
				name = name.substring(i + 1);
			return name + " on option '" + option + "'";
		}

		/**
		 * Return the array of aliases for this handler.
		 * When registering the handler, it will be registered for the aliases
		 * too.
		 *
		 * @return the array of aliases for this handler.
		 */
		protected String[] getAliases() {
			return aliases;

		}

		/**
		 * Adds aliases for the option handled by this handler.
		 *
		 * @param aliases the option aliases.
		 */
		public void addAliases(String[] aliases) {
			Set s = new HashSet();
			for (int i = 0; i < this.aliases.length; i++)
				s.add(this.aliases[i]);
			for (int i = 0; i < aliases.length; i++)
				s.add(aliases[i]);
			this.aliases = new String[s.size()];
			s.toArray(this.aliases);
		}

		/**
		* Adds aliases for the option handled by this handler, in comma-separated format.
		*
		* @param aliases the option aliases, in comma-separated format.
		*/
		public void addAliases(String aliases) {
			addAliases(getOptArrayFromString(aliases));
		}

		/**
		* Sets the aliases for the option handled by this handler, in comma-separated format.
		*
		* @param aliases the option aliases, in comma-separated format.
		*/
		public void setAliases(String aliases) {
			setAliases(getOptArrayFromString(aliases));
		}

		/**
		* Sets the aliases for the option handled by this handler.
		*
		* @param aliases the option aliases.
		*/
		public void setAliases(String[] aliases) {
			this.aliases = new String[0];
			addAliases(aliases);
		}

		/**
		 * This method must be overridden by an implementation to define
		 * the behaviour of the handler when an option is found. For example,
		 * an handler may set a member variable, load a file, etc.
		 * <p>
		 * It receives the option literal (stripped of the option prefix and postfix),
		 * a ParamSet object with additional non-optional parameters and
		 * the "set/unset" value of the option.
		 * <p>
		 * @param option the option literal.
		 * @param ps a ParamSet object, to browse option parameters.
		 * @param value specifies if the option has been set or unset.
		 */
		protected abstract void option(
			String option,
			ParamSet ps,
			boolean value);
		/**
		 * @see java.lang.Comparable#compareTo(Object)
		 */
		public int compareTo(Object obj) {
			if (obj instanceof Handler) {
				Handler handler = (Handler) obj;
				return option.compareTo(handler.option);
			}
			return 0;
		}

	}

	/**
	 * A chained handler: executes two handlers one after the other.
	 * The two handlers must handle the same option; the description of the
	 * option is taken by the first handler.
	 */
	public static class ChainedHandler extends CmdLineOptions.Handler {

		private CmdLineOptions.Handler h1;
		private CmdLineOptions.Handler h2;

		/**
		 * Build an handlers which chains the two given ones.
		 * @param h1 the handler executed first.
		 * @param h2 the handler executed second.
		 */
		public ChainedHandler(
			CmdLineOptions.Handler h1,
			CmdLineOptions.Handler h2) {
			super(h1.option);
			if (!h1.option.equals(h2.option))
				throw new ProgrammingError(
					"chained handlers can be used only with the same option. The chained handlers handle respectively the '"
						+ h1.option
						+ "' and '"
						+ h2.option
						+ "' options");
			this.h1 = h1;
			this.h2 = h2;
			ChainedHandler.this.setAliases(h1.getAliases());
			ChainedHandler.this.addAliases(h2.getAliases());
		}

		//public String[] getAliases() { return h1.getAliases(); }

		/**
		 * Executes the two chained handlers one after the other.
		 */
		public final void option(String option, ParamSet ps, boolean value) {
			h1.option(option, ps, value);
			h2.option(option, ps, value);
		}

		/**
		 * Sets the description of both the chained handlers.
		 * @param the description of the handled option.
		 */
		public void setDescription(String description) {
			h1.setDescription(description);
			h2.setDescription(description);
		}

		/**
		 * Returns the description as provided by the first of the two chained handlers.
		 * @return the description as provided by the first of the two chained handlers.
		 */
		public String getDescription() {
			return h1.getDescription();
		}
	}

	/**
	 * An handler which reacts to an option by setting other options as
	 * mandatory.
	 */
	private class MandatoryControllerHandler extends InnerHandler {

		private Set mandatorySet;
		private boolean alternative;

		public MandatoryControllerHandler(
			String option,
			Set mandatorySet,
			boolean alternative) {
			super(option);
			this.mandatorySet = new HashSet(mandatorySet);
			this.alternative = alternative;
			checkMandatorySet();
		}

		public MandatoryControllerHandler(
			String option,
			String[] mandatorySet,
			boolean alternative) {
			super(option);
			this.mandatorySet = new HashSet();
			this.alternative = alternative;
			for (int i = 0; i < mandatorySet.length; i++) {
				this.mandatorySet.add(mandatorySet[i]);
			}
			checkMandatorySet();
		}

		public MandatoryControllerHandler(
			String option,
			String mandatorySet,
			boolean alternative) {
			super(option);
			StringTokenizer st = new StringTokenizer(mandatorySet, ";, ");
			this.mandatorySet = new HashSet();
			this.alternative = alternative;
			while (st.hasMoreTokens()) {
				String opt = st.nextToken();
				this.mandatorySet.add(st.nextToken());
			}
			checkMandatorySet();
		}

		private void checkMandatorySet() {
			for (Iterator i = mandatorySet.iterator(); i.hasNext();) {
				String option = (String) i.next();
				if (handlers.get(option) == null)
					throw new ProgrammingError(
						"The option '"
							+ option
							+ "' has no handler, so it can't be defined "
							+ (alternative ? "alternatively-" : "")
							+ "mandatory.");
			}
		}

		public final void option(String option, ParamSet ps, boolean value) {
			if (!alternative)
				// Add the *elements*
				mandatoryOptions.addAll(mandatorySet);
			else
				// Add the *set*
				mandatoryOptions.add(mandatorySet);
		}

	}

	/**
	 * A base class for predefined handlers.
	 */
	private abstract class InnerHandler extends Handler {

		public InnerHandler(String option) {
			super(option);
		}

		public InnerHandler(
			String option,
			String description,
			String aliases) {
			super(option, description, aliases);
		}

	}

	/**
	 * The reflectionhandler reacts to an option by attempting to set
	 * the associated referenced member with a correct value.
	 */
	private class ReflectionHandler extends InnerHandler {

		private Object target;
		private java.lang.reflect.Field field;
		private Object defaultValue;
		private boolean defaultOnly;

		public ReflectionHandler(String option, Object target, String member) {
			this(option, target, member, null);
		}

		public ReflectionHandler(
			String option,
			Object target,
			String member,
			Object defaultValue) {
			this(option, target, member, defaultValue, false);
		}

		public ReflectionHandler(
			String option,
			Object target,
			String member,
			Object defaultValue,
			boolean defaultOnly) {
			this(option, null, null, target, member, defaultValue, defaultOnly);
		}

		public ReflectionHandler(
			String option,
			String description,
			String aliases,
			Object target,
			String member,
			Object defaultValue,
			boolean defaultOnly) {
			super(option, description, aliases);
			this.target = target;
			this.defaultValue = defaultValue;
			this.defaultOnly = defaultOnly;

			// Verify that if defaultOnly's true, there exists a default value
			if (defaultValue == null && defaultOnly == true)
				throw new ProgrammingError(
					"option '"
						+ option
						+ "' registered with defaultOnly=true, but null default value.");

			// Verify that the member's of primitive type or an array of primitive type
			// or has a constructor accepting a String
			try {

				// Allow static members, if the target is a Class object
				Class targetClass = target.getClass();
				if (target instanceof Class)
					targetClass = (Class) target;

				field = getField(targetClass, member);

				Class fType = field.getType();
				if (fType.isArray())
					fType = fType.getComponentType();

				if (!fType.isPrimitive()
					&& !fType.isAssignableFrom(String.class)
					&& !hasStringConstructor(fType))
					throw new ProgrammingError(
						"the field '"
							+ member
							+ "' in the class '"
							+ target.getClass().getName()
							+ "' to be set by the option '"
							+ option
							+ "' is not of primitive/String type or array of primitive/String type, "
							+ "nor has a String-based constructor.");

			} catch (NoSuchFieldException e) {
				throw new ProgrammingError(
					"there's no field named '"
						+ member
						+ "' in the class '"
						+ target.getClass().getName()
						+ "' to be set by the option '"
						+ option
						+ "'.");
			}
		}

		private Field getField(Class cls, String member)
			throws NoSuchFieldException {
			Field f;
			while (cls.getSuperclass() != null) {
				try {
					f = cls.getDeclaredField(member);
					return f;
				} catch (NoSuchFieldException e) {
					cls = cls.getSuperclass();
				}
			}
			throw new NoSuchFieldException();
		}

		/**
		 * Return true if the class has constructor accepting
		 * a single string.
		 */
		private boolean hasStringConstructor(Class cls) {
			try {
				Constructor ctor =
					cls.getConstructor(new Class[] { String.class });
				return true;
			} catch (NoSuchMethodException e) {
				return false;
			}
		}

		/**
		 * Implements prescribed behaviour with respect to setting the target
		 * object's field.
		 */
		public void option(String option, ParamSet ps, boolean value) {

			Class fType = field.getType();
			Object valueToSet;

			try {
				if (fType.equals(Boolean.TYPE)) { // Handle boolean case
					if (ps.hasNext())
						throw new OptionException(
							MessageFormat.format(
								msgsBundle.getString("oe2"),
								new Object[] { option }));
					valueToSet = Boolean.valueOf(value);
				} else {

					// A "don't set" value doesn't make sense for parametrized
					// options
					if (!value)
						throw new OptionException(
							MessageFormat.format(
								msgsBundle.getString("oe3"),
								new Object[] { option }));

					if (fType.isArray()) { // Handle arrays
						// Get array's component type
						Class acType = fType.getComponentType();

						// Create an array of proper component type and sized with
						// to contain the passed parameters
						valueToSet = Array.newInstance(acType, ps.size());

						// Fill it in with parameters
						int index = 0;
						while (ps.hasNext()) {
							String param = ps.next();
							Array.set(
								valueToSet,
								index++,
								getSingleValueToSet(acType, param));
						}

					} else { // Handle single var case

						// Expect exactly one parameter if no default's defined
						if (defaultValue == null) {
							if (ps.size() != 1)
								throw new OptionException(
									MessageFormat.format(
										msgsBundle.getString("oe4"),
										new Object[] { option }));
							String optValue = ps.next();
							valueToSet = getSingleValueToSet(fType, optValue);
						} else {
							if (ps.size() == 1) {
								String optValue = ps.next();
								if (defaultOnly)
									throw new OptionException(
										MessageFormat.format(
											msgsBundle.getString("oe5"),
											new Object[] { option, optValue }));
								else
									valueToSet =
										getSingleValueToSet(fType, optValue);
							} else {
								// Check type of the default value, to prevent programming errors
								if (!fType
									.isAssignableFrom(defaultValue.getClass()))
									throw new ProgrammingError(
										"The default value "
											+ defaultValue
											+ "for the option '"
											+ option
											+ "' has invalid type: expected <"
											+ fType.getName()
											+ ">, found <"
											+ defaultValue.getClass().getName()
											+ ">");
								valueToSet = defaultValue;
							}
						}
					}
				}
			} catch (NumberFormatException e) {
				throw new OptionException(
					MessageFormat.format(
						msgsBundle.getString("oe6"),
						new Object[] { option, e.getMessage()}));
			}

			doSetField(field, valueToSet);
		}

		/**
		 * Sets the field, regardless of its accessibility state.
		 */
		private void doSetField(Field field, Object valueToSet) {
			try {
				boolean originalAccessibility = field.isAccessible();
				field.setAccessible(true);
				field.set(target, valueToSet);
				field.setAccessible(originalAccessibility);
			} catch (IllegalAccessException e) {
				e.printStackTrace();
				throw new ProgrammingError(
					"the member '"
						+ field.getName()
						+ "' in the class '"
						+ target.getClass().getName()
						+ "' associated to the option "
						+ option
						+ " is not accessible");
			}
		}

		/**
		 * Transforms the cmdline option value (a String) in an primitive-wrapping object
		 * of class appropriate for assignment to a reference of primitive type, via
		 * reflection.
		 *
		 * @param fType the class of the reference to assign.
		 *              Must be primitive or primitive-wrapping type.
		 * @param optValue the string object denoting a value of type fType.
		 * @return an Object of primitive-wrapping class suitable for assignment
		 *         via reflection.
		 */
		private Object getSingleValueToSet(Class fType, String optValue) {
			Object valueToSet;
			if (fType.equals(String.class)) {
				valueToSet = optValue;
			} else if (
				fType.equals(Character.TYPE)
					|| fType.equals(Character.class)) {
				if (optValue.length() > 1)
					throw new OptionException(
						MessageFormat.format(
							msgsBundle.getString("oe7"),
							new Object[] { option }));
				valueToSet = new Character(optValue.charAt(0));
			} else if (fType.equals(Byte.TYPE) || fType.equals(Byte.class)) {
				valueToSet = new Byte(optValue);
			} else if (fType.equals(Short.TYPE) || fType.equals(Short.class)) {
				valueToSet = new Short(optValue);
			} else if (
				fType.equals(Integer.TYPE) || fType.equals(Integer.class)) {
				valueToSet = new Integer(optValue);
			} else if (fType.equals(Long.TYPE) || fType.equals(Long.class)) {
				valueToSet = new Long(optValue);
			} else if (fType.equals(Float.TYPE) || fType.equals(Float.class)) {
				valueToSet = new Float(optValue);
			} else if (
				fType.equals(Double.TYPE) || fType.equals(Double.class)) {
				valueToSet = new Double(optValue);
			} else if (hasStringConstructor(fType)) {
				try {
					Constructor ctor =
						fType.getConstructor(new Class[] { String.class });
					valueToSet = ctor.newInstance(new Object[] { optValue });
				} catch (Exception e) {
					throw new OptionException(
						MessageFormat.format(
							msgsBundle.getString("oe11"),
							new Object[] {
								fType,
								e.getClass().getName(),
								e.getMessage()}));
				}
			} else
				throw new ProgrammingError(
					"unexpected member type " + fType.getName());
			return valueToSet;
		}
	}

	/**
	 * Build an handler which recognizes options with the given prefix and
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself).
	 * <p>
	 * @param failOnUnrecognizedOption if true, a ProgrammingError will occur if an option is not recognized
	 *        by {@link #parse(java.lang.String[]) parse()}.
	 * @param optPrefix the prefix which identifies an option.
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 */
	public CmdLineOptions(
		boolean allowNegatedOptions,
		String optPrefix,
		boolean failOnUnrecognizedOption) {
		this(null, allowNegatedOptions, optPrefix, failOnUnrecognizedOption);
	}

	/**
	 * Build an handler with the given usage description, which recognizes options with the given prefix and
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself).
	 * <p>
	 * @param usageDescription a human-readable description of the tool embedding this object
	 * @param failOnUnrecognizedOption if true, a ProgrammingError will occur if an option is not recognized
	 *        by {@link #parse(java.lang.String[]) parse()}.
	 * @param optPrefix the prefix which identifies an option.
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 */
	public CmdLineOptions(
		String usageDescription,
		boolean allowNegatedOptions,
		String optPrefix,
		boolean failOnUnrecognizedOption) {
		this.usageDescription = usageDescription;
		CmdLineOptions.optPrefix = optPrefix;
		this.allowNegatedOptions = allowNegatedOptions;
		this.failOnUnrecognizedOption = failOnUnrecognizedOption;
	}

	/**
	 * Build an handler which recognizes options with the given prefix,
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself), and fails if an option is not recognized.
	 * <p>
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 * @param optPrefix the prefix which identifies an option.
	 */
	public CmdLineOptions(boolean allowNegatedOptions, String optPrefix) {
		this(allowNegatedOptions, optPrefix, true);
	}

	/**
	 * Build an handler which recognizes options with the given prefix,
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself), and fails if an option is not recognized.
	 * <p>
	 * @param usageDescription a human-readable description of the tool embedding this object.
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 * @param optPrefix the prefix which identifies an option.
	 */
	public CmdLineOptions(
		String usageDescription,
		boolean allowNegatedOptions,
		String optPrefix) {
		this(usageDescription, allowNegatedOptions, optPrefix, true);
	}

	/**
	 * Build an handler which recognizes options with the "-" prefix,
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself), and fails if an option is not recognized.
	 * <p>
	 * @param usageDescription a human-readable description of the tool embedding this object.
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 */
	public CmdLineOptions(
		String usageDescription,
		boolean allowNegatedOptions) {
		this(usageDescription, allowNegatedOptions, "-");
	}

	/**
	 * Build an handler which recognizes options with the "-" prefix,
	 * allows only plain options, and fails if an option is not recognized.
	 * <p>
	 * @param usageDescription a human-readable description of the tool embedding this object.
	 */
	public CmdLineOptions(String usageDescription) {
		this(usageDescription, false);
	}

	/**
	 * Build an handler which recognizes options with the "-" prefix,
	 * specifies whether to allow or no the negation of options(by using
	 * +/- after the option itself), and fails if an option is not recognized.
	 * <p>
	 * @param allowNegatedOptions if <b>true</b>, options can be followed by +/- to switch them on or off.
	 */
	public CmdLineOptions(boolean allowNegatedOptions) {
		this(allowNegatedOptions, "-");
	}

	/**
	 * Build an handler which recognizes options with the "-" prefix,
	 * allows only plain options, and fails if an option is not recognized.
	 */
	public CmdLineOptions() {
		this(false);
	}

	/**
	 * Parse the cmdline arguments (usually received by <b>main</b>) looking for the
	 * options managed by the handler.
	 * <br>
	 * If the handlers has been built with the failOnUnrecognizedOption parameter true,
	 * this method will throw an {@link CmdLineOptions.OptionException OptionException} on unrecognized
	 * options.
	 *
	 * @param args the cmdline arguments.
	 * @return the cmdline arguments stripped by the options found.
	 */
	public String[] parse(String[] args) {
		return parse(args, null);
	}

	/**
	 * Parses the cmdline arguments (usually received by <b>main</b>) looking for the
	 * options managed by the handler.
	 * <br>
	 * If the handlers has been built with the failOnUnrecognizedOption parameter true,
	 * this method will throw an {@link CmdLineOptions.OptionException OptionException} on unrecognized
	 * options.
	 *
	 * @param args the cmdline arguments.
	 * @param context in case of parsing problem, it's used in the diagnostics
	 *        to tell the user the context of the problem (for example, if the parsing
	 *        is done on a String[] synthetized by the program and not directly on
	 *        command line arguments).
	 * @return the cmdline arguments stripped by the options found.
	 */
	public String[] parse(String[] args, String context) {

		if (context == null)
			context = "";
		currentContext = context;
		Vector v = new Vector();
		Set encounteredOptions = new HashSet();
		for (int i = 0; i < args.length; i++) {
			String option;
			if ((option = getCanonicalOption(args[i])) != null) {

				// See if it's an alias
				if (aliasesTable.containsKey(option))
					option = (String) aliasesTable.get(option);

				Handler handler = (Handler) handlers.get(option);
				if (handler == null)
					if (failOnUnrecognizedOption)
						throw new OptionException(
							MessageFormat.format(
								msgsBundle.getString("oe8"),
								new Object[] { optPrefix + option }));
					else
						continue;

				// Determine option "negated" value
				boolean optValue = true;
				if (allowNegatedOptions)
					if (args[i].endsWith("-"))
						optValue = false;

				// Build ParamSet
				Vector v2 = new Vector();
				for (int j = i + 1; j < args.length; j++) {
					if (getCanonicalOption(args[j]) == null) {
						v2.addElement(args[j]);
					} else
						break;
					i++;
				}
				Handler.ParamSet ps =
					new Handler.ParamSet(CmdLineOptions.this, option, v2);

				// Invoke callback
				handler.option(option, ps, optValue);
				encounteredOptions.add(option);
			} else
				v.addElement(args[i]);
		}

		// Do the 'mandatory' check
		checkEncounteredOptions(encounteredOptions);

		String[] opts = new String[v.size()];
		v.copyInto(opts);
		currentContext = "";
		return opts;
	}

	private Handler.ParamSet find(String opt, String[] args, String context) {

		if (context == null)
			context = "";
		currentContext = context;
		for (int i = 0; i < args.length; i++) {
			String option;
			if ((option = getCanonicalOption(args[i])) != null) {

				if (!(opt.equals(option) || aliasesTable.containsKey(option)))
					continue;

				// See if it's an alias
				if (aliasesTable.containsKey(option))
					option = (String) aliasesTable.get(option);

				Handler handler = (Handler) handlers.get(option);
				if (handler == null)
					return null;

				// Build ParamSet
				Vector v2 = new Vector();
				for (int j = i + 1; j < args.length; j++) {
					if (getCanonicalOption(args[j]) == null) {
						v2.addElement(args[j]);
					} else
						break;
					i++;
				}
				Handler.ParamSet ps =
					new Handler.ParamSet(CmdLineOptions.this, option, v2);

				// Invoke callback
				return ps;
			}
		}

		return null;
	}

	/**
	 * Creates an usage string by combining the usage description
	 * provided at construction (if any) and the descriptions/aliases
	 * of every option.
	 */
	public String usage() {
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		pw.println();
		if (usageDescription != null) {
			pw.println(usageDescription);
			pw.println();
		}
		pw.print(msgsBundle.getString("us1") + ": ");
		if (mandatoryOptions.size() > 0)
			pw.println(msgsBundle.getString("us2"));
		else
			pw.println();
		pw.println();

		List l = new ArrayList(handlers.values());

		Collections.sort(l);

		for (Iterator i = handlers.values().iterator(); i.hasNext();) {
			Handler handler = (Handler) i.next();
			pw.print(handler.option);
			if (isMandatory(handler.option))
				pw.print("(*)");
			pw.println("\t" + handler.getDescription());
			if (handler.getAliases().length > 0) {
				pw.print("\t(aliases:");
				for (int j = 0; j < handler.aliases.length; j++) {
					pw.print(optPrefix + handler.aliases[j]);
					if (j < handler.aliases.length - 1)
						pw.print(", ");
				}
				pw.println(")");
			}
		}
		pw.println();
		return sw.toString();
	}

	/**
	 * Throws an OptionException if the given set does not contain
	 * all the options in the mandatoryOptions set.
	 * <p>
	 * This set may contain either directly an option string,
	 * or a Set of options strings, in which case only one
	 * is considered mandatory.
	 * <p>
	 * However, if the set contains the single "help option" (or mixed help and nonhelp options and 
	 * {@link #isFailOnMixedHelpOptionEncountered() isFailOnMixedHelpOptionEncountered()} is <b>true</b>)
	 * (see {@link #setHelpOption(java.lang.String) setHelpOption()}, no exception is thrown. 
	 * It's client code responsability to act properly when help and nonhelp options 
	 * are mixed.
	 */
	private void checkEncounteredOptions(Set encounteredOptions) {
		Set missingOptions = new HashSet();

		// If the help option is there, the parsing is only supposed
		// to setup the context for the help	    
		if (helpOption != null) {
			if (encounteredOptions.contains(helpOption)) {
				if (isFailOnMixedHelpOptionEncountered())
					if (encounteredOptions.size() != 1)
						throw new OptionException(
							MessageFormat.format(
								msgsBundle.getString("oe12"),
								new Object[] { helpOption }));
				return;
			}
		}

		for (Iterator i = mandatoryOptions.iterator(); i.hasNext();) {
			Object obj = i.next();
			if (obj instanceof Set) {
				// Verify that at least one is encountered. Quick&dirty impl at
				// the moment. Let's build up the error diagnostics since
				// we're at it
				String alternativeFound = null;
				StringBuffer sb = new StringBuffer("one of { ");
				for (Iterator j = ((Set) obj).iterator(); j.hasNext();) {
					String option = (String) j.next();
					if (alternativeFound == null) {
						if (encounteredOptions.contains(option)) {
							alternativeFound = option;
						}
					} else {
						if (encounteredOptions.contains(option))
							throw new OptionException(
								MessageFormat.format(
									msgsBundle.getString("oe9"),
									new Object[] {
										optPrefix + option,
										optPrefix + alternativeFound }));
					}
					sb.append(optPrefix + option);
					if (j.hasNext())
						sb.append(", ");
				}

				if (alternativeFound == null) {
					sb.append(" }");
					missingOptions.add(sb.toString());
				}
			} else {
				// Plain single option, just verify it exits
				if (!encounteredOptions.contains((String) obj)) {
					missingOptions.add(obj);
				}
			}

			// Let's throw the exception if necessary
			if (!missingOptions.isEmpty()) {
				StringBuffer sb = new StringBuffer();
				for (i = missingOptions.iterator(); i.hasNext();) {
					sb.append(i.next());
					if (i.hasNext())
						sb.append(", ");
				}
				throw new OptionException(
					MessageFormat.format(
						msgsBundle.getString("oe10"),
						new Object[] { sb.toString()}));
			}
		}
	}

	/**
	 * Set or unset the 'mandatory' state of an option. The option must have
	 * an handler. Mandatory options will generate an OptionException at
	 * runtime on {@link #parse(java.lang.String[]) parse()} if they're not
	 * included in the parsed arguments.
	 *
	 * @param option the option whose mandatory state is to be set.
	 * @param mandatory the mandatory state to set.
	 */
	public void setMandatory(String option, boolean mandatory) {
		option = getCanonicalOption(optPrefix + option);
		if (handlers.get(option) == null)
			throw new ProgrammingError(
				"Option '"
					+ option
					+ "' can't be set as mandatory since there's no handler for it");
		if (mandatory)
			mandatoryOptions.add(option);
		else
			mandatoryOptions.remove(option);
	}

	/**
	 * Set an option as mandatory. The option must have
	 * an handler. Mandatory options will generate an OptionException at
	 * runtime on {@link #parse(java.lang.String[]) parse()} if they're not
	 * included in the parsed arguments.
	 *
	 * @param option the option whose mandatory state is to be set.
	 */
	public void setMandatory(String option) {
		setMandatory(option, true);
	}

	/**
	 * Set/unset an option as the "help option". 
	 * <p>
	 * An "help option" is an option which the user uses when he wants only to 
	 * receive help. 
	 * <p>
	 * It affects the parsing in that, when {@link #parse(java.lang.String[]) 
	 * parse()} is invoked, the parser will not complain about mandatory options 
	 * missing if the help option is the only one present. 
	 * <p> 
	 * If the help option is mixed with other options, the parsing result 
	 * (exceptional or correct) depend on the result of {@link #isFailOnMixedHelpOptionEncountered() 
	 * isFailOnMixedHelpOptionEncountered()} method (which defaults to <b>true</b>).
	 * <p>
	 * It is s client code responsability to react properly (for example displaying the 
	 * {@link #usage() usage()} string) when the user specifies an help option.
	 * 
	 * @param option  the option to be set or unset as help option.
	 * @param value if <b>true</b>, the given option is set as help option. 
	 *         If <b>false</b>, the given option (which must be the current help
	 *         option) is unset as help option.
	 */
	public void setHelpOption(String option, boolean value) {
		if (handlers.get(option) == null)
			throw new ProgrammingError(
				"Option '"
					+ option
					+ "' can't be set as help option since there's no handler for it");
		if (value) {
			helpOption = option;
		} else {
			if (!option.equals(helpOption)) {
				if (helpOption != null)
					throw new ProgrammingError(
						"Can't unset '"
							+ option
							+ "' as help option since the current help option is "
							+ helpOption);
				else
					throw new ProgrammingError(
						"Can't unset '"
							+ option
							+ "' as help option since no current help option is defined");
			} else
				helpOption = null;
		}
	}

	/**
	 * Set the given option an help option. The option must have an handler.
	 * <p>
	 * See {@link setHelpOption(java.lang.String, boolean) setHelpOption()}
	 * for details on help options.
	 * @param option  the option to be set as help option.
	 */
	public void setHelpOption(String option) {
		setHelpOption(option, true);
	}

	/**
	 * Return the current help option See {@link setHelpOption(java.lang.String, boolean) setHelpOption()}
	 * or <b>null</b> if no help option is set.
	 * @return the current help option or <b>null</b> if no help option is set.
	 */
	public String getHelpOption() {
		return helpOption;
	}

	/**
	 * Returns the mandatory state of the given option. The option must have
	 * an handler.
	 * @param option the option whose mandatory state is to be verified.
	 * @return the mandatory state of the given option.
	 */
	public boolean isMandatory(String option) {
		option = getCanonicalOption(option);
		if (handlers.get(option) == null)
			throw new ProgrammingError(
				"Option '"
					+ option
					+ "' mandatory state can't be requested since there's no handler for it");
		return mandatoryOptions.contains(option);
	}

	/**
	 * Sets/removes the mandatory state of a group of mutually exclusive options, for
	 * which already exist handlers, where at least one is mandatory.
	 * Alternative mandatory options will
	 * generate a ProgrammingError at runtime on {@link #parse(java.lang.String[])
	 * parse()} if at least one is included in the parsed arguments.
	 *
	 * @param options an array of option names.
	 * @param mandatory the mandatory state.
	 */
	public void setAlternativeMandatory(String[] options, boolean mandatory) {
		Set alternatives = new HashSet();
		for (int i = 0; i < options.length; i++) {
			if (handlers.get(options[i]) == null)
				throw new ProgrammingError(
					"Option '"
						+ options[i]
						+ "' can't be set as mandatory since there's no handler for it");
			alternatives.add(options[i]);
		}
		if (mandatory)
			mandatoryOptions.add(alternatives);
		else
			mandatoryOptions.remove(alternatives);
	}

	/**
	 * Specifies that the given options set are mandatory only if one other
	 * option is encountered.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory.
	 * @param alternative if <b>true</b> only one amont the options to set becomes mandatory.
	 */
	private void setMandatoryOnOption(
		String option,
		String[] optionsToSet,
		boolean alternative) {
		Handler optHandler = (Handler) handlers.get(option);
		if (optHandler == null)
			throw new ProgrammingError(
				"there's no handler for option '"
					+ optPrefix
					+ option
					+ "', so no other options can be set as mandatory when the former is encountered");

		// Replace the existing handler with a chained one which adds the given options
		// to the mandatory bunch. Also replace the description.
		handlers.remove(optHandler);
		Handler mHandler =
			new MandatoryControllerHandler(option, optionsToSet, alternative);
		mHandler.setDescription(optHandler.getDescription());
		mHandler.setAliases(optHandler.getAliases());
		handlers.put(option, new ChainedHandler(mHandler, optHandler));
	}

	/**
	 * Specifies that all the options in the given set become mandatory only if one other
	 * option is encountered
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory.
	 */
	public void setMandatoryOnOption(String option, String[] optionsToSet) {
		setMandatoryOnOption(option, optionsToSet, false);
	}

	/**
	 * Specifies that the given options set are mandatory only if one other
	 * option is encountered. The options to set as mandatory are specified
	 * as a comma separated string.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory, as a comma separated string.
	 */
	public void setMandatoryOnOption(String option, String optionsToSet) {
		setMandatoryOnOption(option, getOptArrayFromString(optionsToSet));
	}

	/**
	 * Specifies that the given options set are mandatory only if one in another
	 * set of options is encountered.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory.
	 */
	public void setMandatoryOnOptions(String options, String[] optionsToSet) {
		setMandatoryOnOptions(options, optionsToSet, false);
	}

	/**
	 * Specifies that the given options set are mandatory only if one in another
	 * set of options is encountered. The options to set as mandatory are specified
	 * as a comma separated string.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory, as a comma separated string.
	 */
	public void setMandatoryOnOptions(String options, String optionsToSet) {
		setMandatoryOnOptions(options, getOptArrayFromString(optionsToSet));
	}

	/**
	 * Specifies that the given options set are mandatory (or alternatively mandatory)
	 * only if one in another set of options is encountered.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become mandatory.
	 * @param alternative the given option set is made alternative-mandatory.
	 */
	private void setMandatoryOnOptions(
		String options,
		String[] optionsToSet,
		boolean alternative) {
		String[] opts = getOptArrayFromString(options);
		for (int i = 0; i < opts.length; i++)
			setMandatoryOnOption(opts[i], optionsToSet, alternative);
	}

	/**
	 * Specifies that one among the given options set become mandatory only if one
	 * option is encountered.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become alternative-mandatory.
	 */
	public void setAlternativeMandatoryOnOption(
		String options,
		String[] optionsToSet) {
		// It's now just really a frontend to the setMandatoryOnOption method
		setMandatoryOnOption(options, optionsToSet, true);
	}

	/**
	 * Specifies that one among the given options set become mandatory only if one
	 * option is encountered. The options to set as mandatory are
	 * specified as a comma separated string.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become alternative-mandatory, as a comma separated string.
	 */
	public void setAlternativeMandatoryOnOption(
		String options,
		String optionsToSet) {
		setAlternativeMandatoryOnOption(
			options,
			getOptArrayFromString(optionsToSet));
	}

	/**
	 * Specifies that one among the given options set become mandatory only if one
	 * in another set of options is encountered.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become alternative-mandatory.
	 */
	public void setAlternativeMandatoryOnOptions(
		String options,
		String[] optionsToSet) {
		// It's now just really a frontend to the setMandatoryOnOption method
		setMandatoryOnOptions(options, optionsToSet, true);
	}

	/**
	 * Specifies that one among the given options set become mandatory only if one
	 * in another set of options is encountered. The options to set as mandatory are
	 * specified as a comma separated string.
	 *
	 * @param option the option with triggers the mandatory state of the given others.
	 * @param optionsToSet the options which become alternative-mandatory, as a comma separated string.
	 */
	public void setAlternativeMandatoryOnOptions(
		String options,
		String optionsToSet) {
		setAlternativeMandatoryOnOptions(
			options,
			getOptArrayFromString(optionsToSet));
	}

	private static String[] getOptArrayFromString(String options) {
		StringTokenizer st = new StringTokenizer(options, ",; ");
		String opts[] = new String[st.countTokens()];

		for (int i = 0; st.hasMoreTokens(); i++) {
			opts[i] = st.nextToken();
		}
		return opts;
	}

	/**
	 * Sets up a group of options, for which already exist
	 * handlers, as mutually exclusive  and where at least one is mandatory.
	 * <p>
	 * Alternative-mandatory options will
	 * generate a ProgrammingError at runtime on {@link #parse(java.lang.String[])
	 * parse()} if at least one is included in the parsed arguments.
	 *
	 * @param options an array of option names.
	 * @param mandatory the mandatory state.
	 */
	public void setAlternativeMandatory(String[] options) {
		setAlternativeMandatory(options, true);
	}

	/**
	 * Returns the canonical form of an option, if the given string
	 * denotes an option as of the class conventions; else, <b>null</b>.
	 *
	 * @return the canonical form of an option, if the given string
	 *          denotes an option as of the class conventions; else, <b>null</b>.
	 */
	private String getCanonicalOption(String arg) {
		if (!arg.startsWith(optPrefix))
			return null;
		// Check whether it's a negative number. Use the double
		// as most general number denotation
		try {
			Double.parseDouble(arg);
			// It's a number
			return null;
		} catch (NumberFormatException e) {
			// It's okay, we were just checking
		}
		if (arg.length() == optPrefix.length())
			throw new ProgrammingError("'" + arg + "' is not a valid argument");
		String option = arg.substring(optPrefix.length());
		if (allowNegatedOptions) {
			if (option.endsWith("+") || option.endsWith("-")) {
				if (option.length() == 1)
					throw new ProgrammingError(
						"'" + arg + "' is not a valid argument");
				option = option.substring(0, option.length() - 1);
			}
		}
		return option;
	}

	/**
	 * Registers an handler. Only one handler for a given option literal may be registered.
	 * <p>
	 * @param handler the option handler to register.
	 */
	public void registerHandler(Handler handler) {
		if (handlers.get(getCanonicalOption(handler.option)) != null)
			throw new ProgrammingError(
				"An handler for '" + handler.option + "' already exists");
		handlers.put(getCanonicalOption(handler.option), handler);

		// Add aliases
		String[] aliases = handler.getAliases();

		for (int i = 0; i < aliases.length; i++) {
			String alias = aliases[i];
			if (handlers.containsKey(alias))
				throw new ProgrammingError(
					"'"
						+ alias
						+ "' is already defined as an alias/option, thus can't be defined as alias for '"
						+ handler.option
						+ "'");
			handlers.put(alias, handler);
		}
	}

	/*
	 * OVERLOADS FOR REFLECTION HANDLER
	 */

	/**
	 * Registers an handler which sets a member variable of String type,
	 * on the target object when encountered, having the same name as the
	 * option.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(String option, Object target) {
		setObjectOnOption(option, target, option, null, false);
	}

	/**
	 * Registers an handler which sets a member variable (of Object type) of
	 * on the target object when encountered, having the same name as the
	 * option.
	 * <p>
	 * If the member type is <tt>boolean</tt>, the member will be set to <tt>true</tt>
	 * if the option exists. For other types, a single following parameter is required,
	 * unless a default is specified.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member - may be <b>null</b>.
	 */
	public void setObjectOnOption(
		String option,
		Object target,
		Object defaultValue) {
		setObjectOnOption(option, target, option, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable (of Object type) of
	 * on the target object when encountered, having the same name as the
	 * option, and no default value (<b>null</b>).
	 * <p>
	 *
	 * @param option the option and member name to be set.
	 * @param target the object which contains the member to be set.
		 */
	public void setObjectOnOption(String option, Object target) {
		setObjectOnOption(option, target, option, null, false);
	}

	/**
	 * Registers an handler which sets a member variable of String type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(String option, Object target, String memberName) {
		setObjectOnOption(option, target, memberName, null, false);
	}

	/**
	 * Registers an handler which sets a member variable (of Object type) on
	 * the target object when encountered.
	 * If the member type is <tt>boolean</tt>, the member will be set to <tt>true</tt>
	 * if the option exists. For other types, a single following parameter is required,
	 * unless a default is specified.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member - may be <b>null</b>.
	 */
	public void setObjectOnOption(
		String option,
		Object target,
		String memberName,
		Object defaultValue) {
		setObjectOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of String type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		String defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		boolean defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		char defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		byte defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		short defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		int defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		long defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		float defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		double defaultValue) {
		setOnOption(option, target, memberName, defaultValue, false);
	}

	/**
	 * Registers an handler which sets a member variable (of Object type) of
	 * on the target object when encountered.
	 * If the member type is <tt>boolean</tt>, the member will be set to <tt>true</tt>
	 * if the option exists. For other types, a single following parameter is required,
	 * unless a default is specified.
	 *
	 * @param option the option to be set.
	 * @param description a description for the option.
		 * @param aliases the option aliases, in comma-separated format.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member - may be <b>null</b>.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setObjectOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		Object defaultValue,
		boolean defaultOnly) {
		if (memberName == null)
			throw new ProgrammingError("Member name is null. If the member is of 'String' class, you may have inadvertely passed a reference to the member instead of a reference to its name");
		registerHandler(
			new ReflectionHandler(
				option,
				description,
				aliases,
				target,
				memberName,
				defaultValue,
				defaultOnly));
	}

	/**
	 * Registers an handler which sets a member variable (of Object type) of
	 * the target object when found (without defining any description nor aliases).
	 * If the member type is <tt>boolean</tt>, the member will be set to <tt>true</tt>
	 * if the option exists. For other types, a single following parameter is required,
	 * unless a default is specified.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member - may be <b>null</b>.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setObjectOnOption(
		String option,
		Object target,
		String memberName,
		Object defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			null,
			target,
			memberName,
			defaultValue,
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		String defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			defaultValue,
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		boolean defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			Boolean.valueOf(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		char defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Character(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		byte defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Byte(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		short defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Short(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		int defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Integer(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>) (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		long defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Long(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		float defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Float(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set.
	 * @param target the object which contains the member to be set.
	 * @param memberName the name of the member to be set.
	 * @param defaultValue a default value of the member.
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>).
	 */
	public void setOnOption(
		String option,
		Object target,
		String memberName,
		double defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Double(defaultValue),
			defaultOnly);
	}

	/* -Overloads with description and aliases ----------------------------------- */

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		String defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			defaultValue,
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		boolean defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			Boolean.valueOf(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		char defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Character(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		byte defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Byte(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		short defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Short(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		int defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Integer(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>) (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		long defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Long(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		float defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Float(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		double defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Double(defaultValue),
			defaultOnly);
	}

	/* --- (with defaultOnly == false) --- */

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		String defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			defaultValue,
			false);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		boolean defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			Boolean.valueOf(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		char defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Character(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		byte defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Byte(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		short defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Short(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		int defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Integer(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>) (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		long defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Long(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		float defaultValue) {
		setObjectOnOption(
			option,
			description,
			aliases,
			target,
			memberName,
			new Float(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String description,
		String aliases,
		Object target,
		String memberName,
		double defaultValue) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Double(defaultValue),
			false);
	}

	/* -End Overloads with description and aliases ------------------------------- */

	/* -Overloads with aliases only ---------------------------------------------- */

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		String defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			defaultValue,
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		boolean defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			Boolean.valueOf(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		char defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Character(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		byte defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Byte(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		short defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Short(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		int defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Integer(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>) (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		long defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Long(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		float defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Float(defaultValue),
			defaultOnly);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param defaultOnly if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		double defaultValue,
		boolean defaultOnly) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Double(defaultValue),
			defaultOnly);
	}

	/* --- (with defaultOnly == false) --- */

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		String defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			defaultValue,
			false);
	}

	/**
	 * Registers an handler which sets a member variable of boolean primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		boolean defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			Boolean.valueOf(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>char</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		char defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Character(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>boolean</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		byte defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Byte(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>short</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		short defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Short(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>int</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		int defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Integer(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>long</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>) (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		long defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Long(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>float</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		float defaultValue) {
		setObjectOnOption(
			option,
			null,
			aliases,
			target,
			memberName,
			new Float(defaultValue),
			false);
	}

	/**
	 * Registers an handler which sets a member variable of <tt>double</tt> primitive type
	 * on the target object when encountered.
	 *
	 * @param option the option to be set
	 * @param target the object which contains the member to be set
	 * @param memberName the name of the member to be set
	 * @param defaultValue a default value of the member
	 * @param false if <b>true</b>, the option will not allow parameters, but the associated member
	 *                    will be set to the default if passed on the command line (defaultValue must be not <b>null</b>)
	 */
	public void setOnOption(
		String option,
		String aliases,
		Object target,
		String memberName,
		double defaultValue) {
		setObjectOnOption(
			option,
			target,
			memberName,
			new Double(defaultValue),
			false);
	}

	/* -End overloads with aliases only ---------------------------------------- */

	/**
	 * Sets the same member with different values when different options
	 * are encountered. When a user specifies an option, the member
	 * is set with the corresponding value. The user will not be able/need
	 * to specify any parameter for the option.
	 * <p>
	 * It's just a shortcut to multiple <tt>setOnOptions()</tt> calls referring to
	 * the same member in the target object and allowing only the default value to
	 * be set.
	 * <p>
	 * An example:
	 * <p>
	 * <code>
	 *   setValuesOnOptions(&quot;doThis, doThat&quot;, &quot;this, that&quot;, this, &quot;thingToDo&quot;);
	 * </code>
	 * <p>
	 * sets the member &quot;thingToDo&quot; in the target (<code>this</code>) to
	 * either <tt>this</tt> or <tt>that</tt> when the user specifies either <tt>doThis</tt>
	 * or <tt>doThat</tt>.
	 *
	 * @param options the options which trigger the assignment, in comma separated format
	 * @param values the values to assign the target object's member, in comma separated format,
	 *               in 1-1 correspondence with the options
	 * @param target the target object, owning the member to be set
	 * @param member the member variable to be set in the target object when one of the options
	 *               is encountered
	 *
	 */
	public void setValuesOnOptions(
		String options,
		String values,
		Object target,
		String member) {
		setValuesOnOptions(
			getOptArrayFromString(options),
			getOptArrayFromString(values),
			target,
			member);
	}

	/**
	 * Sets the same member with different values when different options
	 * are encountered. When a user specifies an option, the member
	 * is set with the corresponding value. The user will not be able/need
	 * to specify any parameter for the option.
	 * <p>
	 * It's just a shortcut to multiple <tt>setOnOptions()</tt> calls referring to
	 * the same member in the target object and allowing only the default value to
	 * be set.
	 * <p>
	 * An example:
	 * <p>
	 * <code>
	 *   setValuesOnOptions(new String[] { &quotdoThis&quot, &quotdoThat&quot; }, new String[] { &quothis&quot, &quotthat&quot; }, this, &quot;thingToDo&quot;);
	 * </code>
	 * <p>
	 * sets the member &quot;thingToDo&quot; in the target (<code>this</code>) to
	 * either <tt>this</tt> or <tt>that</tt> when the user specifies either <tt>doThis</tt>
	 * or <tt>doThat</tt>.
	 *
	 * @param options the options which trigger the assignment
	 * @param values the values to assign the target object's member, in 1-1 correspondence
	 *               with the options
	 * @param target the target object, owning the member to be set
	 * @param member the member variable to be set in the target object when one of the options
	 *               is encountered
	 */
	public void setValuesOnOptions(
		String[] options,
		String[] values,
		Object target,
		String member) {
		if (options.length != values.length) {
			StringWriter sw;
			PrintWriter pw = new PrintWriter(sw = new StringWriter());
			pw.println(
				"the number of values passed must match the number of options");
			pw.println(
				"The following "
					+ options.length
					+ " options have been specified:");
			for (int i = 0; i < options.length; i++)
				pw.println("\t" + options[i]);

			pw.println(
				"While the following "
					+ values.length
					+ " values have been specified:");
			for (int i = 0; i < values.length; i++)
				pw.println("\t" + values[i]);
			throw new ProgrammingError(sw.toString());
		}

		for (int i = 0; i < options.length; i++) {
			setOnOption(options[i], target, member, values[i], true);
		}
	}

	/**
	 * An overload for {@link #setValuesOnOptions(java.lang.String, java.lang.String, java.lang.Object,
	 * java.lang.String) setValuesOnOptions()} for String target members, that uses the
	 * option literals themselves as option values.
	 * The net effect is that the member target is set with the name of the option, whenever
	 * one of the given options is specified by the user.
	 *
	 * @param options the options which trigger the assignment, in comma separated format
	 * @param target the target object, owning the member to be set
	 * @param member the member variable to be set in the target object when one of the options
	 *               is encountered
	 */
	public void setValuesOnOptions(
		String options,
		Object target,
		String member) {
		setValuesOnOptions(options, options, target, member);
	}

	/**
	 * An overload for {@link #setValuesOnOptions(java.lang.String[], java.lang.String[], java.lang.Object,
	 * java.lang.String) setValuesOnOptions()} for String target members, that uses the
	 * option literals themselves as option values.
	 * The net effect is that the member target is set with the name of the option, whenever
	 * one of the given options is specified by the user.
	 *
	 * @param options the options which trigger the assignment, in comma separated format
	 * @param target the target object, owning the member to be set
	 * @param member the member variable to be set in the target object when one of the options
	 *               is encountered
	 */
	public void setValuesOnOptions(
		String[] options,
		Object target,
		String member) {
		setValuesOnOptions(options, options, target, member);
	}

	/**
	 * Sets the description for a given option.
	 *
	 * @param option the option (for which an handler must already exist)
	 * @param description the description of the option
	 */
	public void setDescription(String option, String description) {
		Handler handler = (Handler) handlers.get(option);
		if (handler == null)
			throw new ProgrammingError(
				"there is no option '"
					+ option
					+ "' of which to set the description");
		handler.setDescription(description);
	}

	public Handler.ParamSet hasOption(String opt, String[] args) {
		return find(opt, args, "searching for option \"" + opt + "\"");
	}

	/**
	 * Retrieves the description for a given option.
	 *
	 * @param option the option (for which an handler must exist)
	 * @retrun the description of the option
	 */
	public String getDescription(String option) {
		Handler handler = (Handler) handlers.get(option);
		if (handler == null)
			throw new ProgrammingError(
				"there is no option '"
					+ option
					+ "' of which to set the description");
		return handler.description;
	}

	/**
	 * Adds aliases for a given option.
	 *
	 * @param option the option (for which an handler must already exist)
	 * @param aliases the new aliases for the option. If an alias is already set
	 *        it's ignored.
	 */
	public void addAliases(String option, String aliases) {
		Handler handler = (Handler) handlers.get(option);
		if (handler == null)
			throw new ProgrammingError(
				"there is no option '" + option + "' to which to add aliases");
		handler.addAliases(aliases);
		// Update aliases table
		String[] a = getOptArrayFromString(aliases);
		for (int i = 0; i < a.length; i++) {
			if (aliasesTable.containsKey(a[i]))
				throw new ProgrammingError(
					"The alias "
						+ a[i]
						+ "is already defined for the option "
						+ aliasesTable.get(a[i]));
			aliasesTable.put(a[i], option);
		}

	}

	/**
	 * String resources for {@link CmdLineOptions.OptionException OptionException} messages.
	 */
	public final static class Resources extends ListResourceBundle {

		static final Object[][] msgs = {
			// Can be localized
			{ "oe1", "''{0}'': missing option argument" }, {
				"oe2", "Unexpected parameters for boolean option ''{0}''" }, {
				"oe3",
					"Option '{0}' is not boolean, thus can't be negated" },
					{
				"oe4", "One parameter expected for option ''{0}''" }, {
				"oe5",
					"Option ''{0}'' does not expect parameters (found '{1}')" },
					{
				"oe6",
					"Expected numeric value for option ''{0}'' (found ''{1}'' instead)" },
					{
				"oe7",
					"The parameter for option ''{0}'' must be a single charachter" },
					{
				"oe8", "Unrecognized option: {0}" }, {
				"oe9",
					"Option ''{0}'' specified, conflicting with preceeding ''{1}''" },
					{
				"oe10", "Mandatory option(s) missing: ''{0}''" }, {
				"oe11",
					"Could not instantiate object of class <{0}> using the single string constructor! The exception is {1}:{2}" },
					{
				"oe12",
					"The option ''{0}'' cannot be mixed with other options" },
					{
				"us1", "Usage" }, {
				"us2", "(options marked with (*) are mandatory)" }
			// Can be localized - end
		};

		/**
		 * Overrides {@link ListResourceBundle ListResourceBundle} method.
		 */
		public Object[][] getContents() {
			return msgs;
		}
	}

	/**
	 * Return the failOnUnrecognizedOption.
	 * @return boolean
	 */
	public boolean isFailOnUnrecognizedOption() {
		return failOnUnrecognizedOption;
	}

	/**
	 * Set the failOnUnrecognizedOption.
	 * @param failOnUnrecognizedOption The failOnUnrecognizedOption to set
	 */
	public void setFailOnUnrecognizedOption(boolean failOnUnrecognizedOption) {
		this.failOnUnrecognizedOption = failOnUnrecognizedOption;
	}

	/**
	 * Return the value for the 'failOnMixedHelpOptionEncountered' flag.
	 * If the flag is <b>true</b>, encountering an 'help option' when
	 * parsing will result in an {@link OptionException OptionException}.
	 * <p>
	 * See {@link setHelpOption(java.lang.String, boolean) setHelpOption()}
	 * for details on help options.
	 * <p>
	 * 
	 * @return boolean
	 */
	public boolean isFailOnMixedHelpOptionEncountered() {
		return failOnMixedHelpOptionEncountered;
	}

	/**
	 * Set the value for the 'failOnMixedHelpOptionEncountered' flag.
	 * If the flag is <b>true</b>, encountering an 'help option' when
	 * parsing will result in an {@link OptionException OptionException}.
	 * <p>
	 * See {@link setHelpOption(java.lang.String, boolean) setHelpOption()}
	 * for details on help options.
	 * <p>
	 * @param failOnMixedHelpOptionEncountered The failOnMixedHelpOptionEncountered to set
	 */
	public void setFailOnMixedHelpOptionEncountered(boolean failOnMixedHelpOptionEncountered) {
		this.failOnMixedHelpOptionEncountered =
			failOnMixedHelpOptionEncountered;
	}

	public static void main(String args[]) {
		
		CmdLineOptions co = new CmdLineOptions("Test");
		co.setOnOption("failOnUnrecognizedOption", co);
		co.setOnOption("allowNegatedOptions", co);
		
		System.out.println(co.usage());

	}

}
